/*
!+ (*currently this program is just a stub*)
   Attempts to configure Makefiles for the current system type (takes into 
   account CPU, F90 and C compiler(s), and the presence of NCAR Graphics,
   MPI libraries, etc.).

!~ 2013-04-05 (last modified; TIM)
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#define MAX_FIO_CHAR_LEN 512
#define N_COMPARG_ENTRY_FIELDS 9

#define N_COMPILER_TYPES 2
enum compiler_type_idx { IDX_F90, IDX_C };

#define N_SPECARG_TYPES 4  /* compiler and loader: optimal, debug */
enum usetype_idx { COMPILE_OPTIMAL, COMPILE_DEBUG,
                        LOADER_OPTIMAL, LOADER_DEBUG };
struct usetype_nstr {
  char name_str[MAX_FIO_CHAR_LEN];
} usetype[N_SPECARG_TYPES] = { {"comp_opt"},{"comp_dbg"},
                               {"load_opt"},{"load_dbg"} };

struct comparg_fs_param {
  char comploc_str[MAX_FIO_CHAR_LEN];
  char compvers_fn[MAX_FIO_CHAR_LEN];
};
struct comparg_fs_param comparg_fs_set[N_COMPILER_TYPES];

struct specific_args {
  char specargs_str[MAX_FIO_CHAR_LEN];
};

struct comparg_entry {
  char compname_str[MAX_FIO_CHAR_LEN];
  char compvers_str[MAX_FIO_CHAR_LEN];
  char lib64type_str[MAX_FIO_CHAR_LEN];
  char compusetype_str[MAX_FIO_CHAR_LEN];
  struct specific_args csargs_str[N_SPECARG_TYPES];
};
struct comparg_entry compargE_raw,compargE_set[N_COMPILER_TYPES];

struct cpuinfo_entry {
  char vendorid_str[MAX_FIO_CHAR_LEN];
  char cpufamily_str[MAX_FIO_CHAR_LEN];
  char cpumodel_str[MAX_FIO_CHAR_LEN];
  char cpumodname_str[MAX_FIO_CHAR_LEN];
  char cpufeatflg_str[MAX_FIO_CHAR_LEN];
  float cpuMHz;
};
struct cpuinfo_entry cpuinfoE_raw, cpuinfoE_set;

char *compargE_raw_field[N_COMPARG_ENTRY_FIELDS];
char *compargE_set_field[N_COMPARG_ENTRY_FIELDS];

char lval[MAX_FIO_CHAR_LEN], lval0[MAX_FIO_CHAR_LEN];
char syscanon_cputype[MAX_FIO_CHAR_LEN]="",
     syscanon_vendor[MAX_FIO_CHAR_LEN]="",
     syscanon_os[MAX_FIO_CHAR_LEN]="",
     syscanon_endianness[MAX_FIO_CHAR_LEN]="",
     compargs_fn[MAX_FIO_CHAR_LEN];

/* Compiler argument table parser vars */

enum token_types {DELIMITER, CONJUNCTION, REL_OPERATOR, ITEM};
struct token_type_name {
  char name[20];
} ttnames[] = { {"Delimiter"}, {"Conjunction"}, {"Relational Operator"},
                {"Item"} };

enum rel_ops { EQ, NE, LT, GT, LE, GE };

char token[80];
int tok_type, op_type;
char *prog;   /* holds expression to be analyzed */


char *afill_char_fbuffer(char *input_fn);
int proc_cpuinfo(char *input_fn);
void proc_syscfg(char *input_fn);
void proc_compversf(void);
int proc_compargtable(int icomp, int iarg, char *input_fn);
void init_comparg_entry_field(char *field_p[],
			      struct comparg_entry *input_compargE,
			      struct cpuinfo_entry *input_cpuinfoE);

int eval_exp_parse_ca(char *raw, char *tstr, char *undef);
void eval_exp1_parse_ca(int *answer, char *tstr);
void eval_exp2_parse_ca(int *answer, char *tstr);
void get_token_parse_ca(void);
int isdelim_parse_ca(char c);
void numcmpchk_parse_ca(int *answer, char *str1, char *str2);


/*```````````````````````````````````````````````````````````````````*/
int main()
{

/*
!+

!~ 2013-04-05 (last modified; TIM)
*/

  int ierr, CPU_cnt, icomp, iarg, i;

/*-------------------------*/

  /* Read/process system config file */
/*  proc_syscfg("build_templates_config-util_c.cfg");*/

  /* Display selected machine information (if available) */
/*  printf("Selected machine information:\n\n");
    if(strstr(syscanon_os,"linux") != NULL) {*/
    /* For Linux systems, this program can do more in depth processing */

    /* Read/process cpuinfo file */
/*    ierr = system("cat /proc/cpuinfo > cpuinfo.tmp");
    CPU_cnt = proc_cpuinfo("cpuinfo.tmp");
    ierr = system("rm -f cpuinfo.tmp");*/

    /* Display CPU/system info to user: */
/*    printf("%d CPU core(s)\n",CPU_cnt);
    printf("%s\n",cpuinfoE_set.vendorid_str);  
    printf("%s\n",cpuinfoE_set.cpumodname_str);  
    printf("%f MHz (%s)\n",cpuinfoE_set.cpuMHz,syscanon_cputype);
    ierr = system("ls /usr/lib64 >/dev/null 2>&1");
    if(ierr == 0) {
      printf("64-bit OS (%s) architecture\n",syscanon_os);
      strcpy(compargE_set[IDX_F90].lib64type_str,"lib64-e64");
      strcpy(compargE_set[IDX_C].lib64type_str,"lib64-e64");
    }
    else {
      printf("32-bit OS (%s) architecture\n",syscanon_os);
      strcpy(compargE_set[IDX_F90].lib64type_str,"lib32-e32");
      strcpy(compargE_set[IDX_C].lib64type_str,"lib32-e32");
    }
    printf("Byte-order type: %s-endian\n\n",syscanon_endianness);
  }
  else {
    printf("Only very limited information available\n");
    printf("CPU type: %s\n",syscanon_cputype);
    printf("System vendor: %s\n",syscanon_vendor);
    printf("System OS type: %s\n",syscanon_os);
    printf("Byte-order type: %s-endian\n\n",syscanon_endianness);
    strcpy(compargE_set[IDX_F90].lib64type_str,"unknown");
    strcpy(compargE_set[IDX_C].lib64type_str,"unknown");
    strcpy(cpuinfoE_set.vendorid_str,"unknown");
    strcpy(cpuinfoE_set.cpufamily_str,"unknown");
    strcpy(cpuinfoE_set.cpumodel_str,"unknown");
    strcpy(cpuinfoE_set.cpumodname_str,"unknown");
    strcpy(cpuinfoE_set.cpufeatflg_str,"unknown");
    }*/

  /* Parse compiler version info file(s) */
/*  proc_compversf();*/

/*-------------------------*/

  /* Search for a compatible entry in the compiler argument table */

/*  init_comparg_entry_field(compargE_raw_field,&compargE_raw,&cpuinfoE_raw);

  for(icomp=0; icomp<N_COMPILER_TYPES; icomp++) {
    init_comparg_entry_field(compargE_set_field,&compargE_set[icomp],
			     &cpuinfoE_set);
    for(iarg=0; iarg<N_SPECARG_TYPES; iarg++) {
      strcpy(compargE_set[icomp].compusetype_str,usetype[iarg].name_str);

      if(!proc_compargtable(icomp,iarg,compargs_fn)) {
	fprintf(stderr,"\nERROR: Relevant (%s: %s) compiler arguments not found.\n",
		compargE_set[icomp].compname_str,usetype[iarg].name_str);
	fprintf(stderr,"       Contact the maintainer of this program code.\n");
	exit(99);
      }
    }
    }*/

/*-------------------------*/

/* Recommend compiler and loader/linker flags based on
   system/compiler/CPU type */

/*  for(i=0; i<N_COMPILER_TYPES; i++) {
    switch(i) {
    case IDX_F90:
      strcpy(lval,"F90"); break;
    case IDX_C:
      strcpy(lval,"C"); break;
    }

    printf("=== Recommended COMPILER flags ===\n");
    printf("%s OPTIMIZE: (%s)\n",
       lval,compargE_set[i].csargs_str[COMPILE_OPTIMAL].specargs_str);
    printf("%s DEBUG: (%s)\n",
       lval,compargE_set[i].csargs_str[COMPILE_DEBUG].specargs_str);
    printf("=== Recommended LOADER/LINKER flags ===\n");
    printf("%s OPTIMIZE: (%s)\n",
       lval,compargE_set[i].csargs_str[LOADER_OPTIMAL].specargs_str);
    printf("%s DEBUG: (%s)\n\n",
       lval,compargE_set[i].csargs_str[LOADER_DEBUG].specargs_str);
       }*/

/*-------------------------*/

  /* Finished */
  return 0;
}


/*```````````````````````````````````````````````````````````````````*/
void init_comparg_entry_field(char *field_p[],
			      struct comparg_entry *input_compargE,
			      struct cpuinfo_entry *input_cpuinfoE)
{

/*
!+ Initializes the 'field_p' pointer table to the memory addresses
!   of the relevant individual components of various structures

!~ 28.04.2010 (last modified; TIM)
*/
 
  field_p[0] = input_cpuinfoE->vendorid_str;
  field_p[1] = input_cpuinfoE->cpufamily_str;
  field_p[2] = input_cpuinfoE->cpumodel_str;
  field_p[3] = input_cpuinfoE->cpumodname_str;
  field_p[4] = input_cpuinfoE->cpufeatflg_str;
  field_p[5] = input_compargE->compname_str;
  field_p[6] = input_compargE->compvers_str;
  field_p[7] = input_compargE->lib64type_str;
  field_p[8] = input_compargE->compusetype_str;
}


/*```````````````````````````````````````````````````````````````````*/
char* afill_char_fbuffer(char *input_fn)
{

/*
!+ IN: 'input_fn' is the input filename string
!  OUT: returns a pointer to the (filled) char buffer

!~ 28.04.2010 (last modified; TIM)
*/

  struct stat f_stats;
  FILE *fileh;

  char *loc_p, *beg_p;
  int ierr;


  /* Allocate buffer for file contents */
  ierr = stat(input_fn, &f_stats);
  if((beg_p=(char *) malloc(f_stats.st_size+1)) == NULL) {
    printf("Buffer allocation failure for %s",input_fn);
    exit(99);
  }

  fileh = fopen(input_fn, "rb");

  /* Read and store contents of entire file */

  loc_p = beg_p;
  do {
    *loc_p = fgetc(fileh);
    loc_p++;
  } while(!feof(fileh));
  *(loc_p-1) = '\0';  /* Terminate file data with NULL char */

  fclose(fileh);

  return(beg_p);
}


/*```````````````````````````````````````````````````````````````````*/
int proc_compargtable(int icomp, int iarg, char *input_fn)
{

/*
!+ Parses compiler argument table.
!  IN: 'input_fn' is the input filename string 

!~ 28.04.2010 (last modified; TIM)
*/

  char *infile_buf, *loc_p, *temp_p;
  int i, j;

  int debug = 0;


  /* Allocate/fill buffer for compiler argument table */
  infile_buf = afill_char_fbuffer(input_fn);

  /* Mildly parse and store relevant file contents */

  strcpy(compargE_set[icomp].csargs_str[iarg].specargs_str,""); /* Empty */

  loc_p = infile_buf;  /* Set to beginning of file data buffer */
  do {
    if(!isspace(*loc_p)) {  /* Skip whitespace characters */
 
      if(*loc_p == '!') {   /* Comment line */
	for( ; *loc_p != '\n'; loc_p++);   /* Skip to end of comment line */ 
      }

      else if(*loc_p == '#') {    /* Specifier block */
	i = 4; do {    /* Determine specifier rank */
	  i++; loc_p++;
	} while(*loc_p == '#');
	if(debug) { printf("## specifier block, rank %d\n",i); fflush(NULL); }
	for( ; isspace(*loc_p); loc_p++);   /* Skip any whitespace */

	sscanf(loc_p,"%s",compargE_raw_field[i]);  /* Read specifier string */
	for( ; !isspace(*loc_p); loc_p++);   /* Advance past specifier */

	if(!eval_exp_parse_ca(compargE_raw_field[i], /* Specifier fails test */
			      compargE_set_field[i],"XXX")) {
	  if(debug) {
	    printf("-- specifier test failed (%s vs. %s)\n",
		   compargE_raw_field[i],compargE_set_field[i]); fflush(NULL);
	  }

	  /* Search for another specifier with <= rank */
	  for( ; (*loc_p != '@') && (*loc_p != '\0'); loc_p++) {
	    if(*loc_p == '#') {
	      
	      temp_p = loc_p;
	      j = 4; do {    /* Determine specifier rank */
	      	j++; loc_p++;
	      } while(*loc_p == '#');
	      
              if(debug) {
		sscanf(temp_p,"%s%s",lval,lval0);
		printf("rank %d vs %d: %s\n",j,i,lval0); fflush(NULL);
	      }

	      if(j <= i) {  
		loc_p = temp_p;
		break;
	      }
	    }
	  }
	  loc_p--;
	  
	}

	else {   /* Specifier test succeeded */
	  if(debug) {
	    printf("-- specifier test succeeded (%s vs. %s)\n",
		   compargE_raw_field[i],compargE_set_field[i]); fflush(NULL);
	  }

          if(i == 8) {       /* lowest-rank specifier test has succeeded  */
	    for( ; *loc_p != '{'; loc_p++);   /* Skip to beginning of args */
	    for( ; !isspace(*loc_p); loc_p++); /* Skip leading whitespace */
	    loc_p++;

	    temp_p = lval;
	    while(*loc_p != '}') {  /* Stop at end of arg block */ 
	      if(*loc_p == '\n') {   /* Provide line continuation sequence */
		*temp_p = ' '; temp_p++;
		*temp_p = '\\'; temp_p++;
	      }
	      *temp_p = *loc_p;
	      temp_p++;
	      loc_p++;
	    } 
	    *temp_p = '\0';  /* terminate string var */ 

	    /* Add arguments to this argument list */
	    strcat(compargE_set[icomp].csargs_str[iarg].specargs_str,lval);

	    if(debug) { printf("-- args read (%s)\n",lval); fflush(NULL); }
      	  }

	  /* if(i==8), then advance to <= rank
	     if(i<8), then advance to > rank */

	}

      }

      else if(*loc_p == '@') {  /* This is a CPU parameter block */
	if(debug) { printf("## CPU parameter block\n"); fflush(NULL); }

	loc_p++;
	sscanf(loc_p,"%s%s%s%s%s",cpuinfoE_raw.vendorid_str,
	       cpuinfoE_raw.cpufamily_str,cpuinfoE_raw.cpumodel_str,
	       cpuinfoE_raw.cpumodname_str,cpuinfoE_raw.cpufeatflg_str);

	for(j=0; j<5; j++) {   /* Test relevance to current CPU */
	  if(debug) { printf("%s vs. %s \n",compargE_raw_field[j],compargE_set_field[j]); fflush(NULL); }

	  /* If not relevant (even partially), search for next CPU block */
	  if(!eval_exp_parse_ca(compargE_raw_field[j],
				compargE_set_field[j],"XXX")) {
	    for( ; (*loc_p != '@') && (*loc_p != '\0'); loc_p++); loc_p--;    
	    break;
	  }
	}
	if(debug) { printf("\n"); fflush(NULL); }

	/* If fully relevant, search for first specifier block */
	if(j == 5) { for( ; *loc_p != '#'; loc_p++); loc_p--; }
      }

    }

    loc_p++;
  } while(*loc_p != '\0');

  free(infile_buf);

  if(*compargE_set[icomp].csargs_str[iarg].specargs_str == '\0') return 0;
  else return 1;
}


/*```````````````````````````````````````````````````````````````````*/
int proc_cpuinfo(char *input_fn)
{

/*
!+ Reads and parses /proc/cpuinfo file (or similar) for CPU specifications. 
!  IN: 'input_fn' is the input filename string
!  OUT: returns the number of CPU entries processed

!~ 28.04.2010 (last modified; TIM)
*/

  char *infile_buf, *loc_p, *temp_p;
  int i;

  int debug = 0;


  /* Allocate/fill buffer for cpuinfo data */
  infile_buf = afill_char_fbuffer(input_fn);

  /* Mildly parse and store file contents */

  i = -1;
  loc_p = infile_buf;  /* Reset to beginning of file data buffer */
  do {
    if(strchr("\n \t\r",*loc_p) == NULL) {  /* Not whitespace */

      /* Obtain field name (without surrounding whitespace) */

      temp_p = lval;
      for( ; *loc_p != ':'; loc_p++) {   /* find delimiter */
	*temp_p = *loc_p; temp_p++;
      }
      *temp_p = '\0'; temp_p--;  /* terminate string, "backspace" */

      for( ; strchr(" \t",*temp_p) != NULL; temp_p--);
      *(temp_p+1) = '\0';  /* terminate trimmed string var */ 

      /* Obtain field value (without surrounding whitespace) */

      loc_p++;  /* Advance past delimiter */
      for( ; *loc_p == ' '; loc_p++);  /* Skip preceding whitespace */

      temp_p = lval0;
      for( ; *loc_p != '\n'; loc_p++) {   /* find newline */
	*temp_p = *loc_p; temp_p++;
      }
      *temp_p = '\0'; temp_p--;  /* terminate string, "backspace" */

      for( ; strchr(" \t",*temp_p) != NULL; temp_p--);
      *(temp_p+1) = '\0';  /* terminate trimmed string var */ 

      /* Fill appropriate variables */

      if(strcmp(lval,"processor") == 0) {
	i++;
	if(debug) printf("proc#%d\n",i); fflush(NULL);
      }
      else if(strcmp(lval,"vendor_id") == 0) {
	strcpy(cpuinfoE_set.vendorid_str,lval0);
	if(debug) printf("%s %s\n",lval,lval0); fflush(NULL);
      }
      else if(strcmp(lval,"cpu family") == 0) {
	strcpy(cpuinfoE_set.cpufamily_str,lval0);
	if(debug) printf("%s %s\n",lval,lval0); fflush(NULL);
      }
      else if(strcmp(lval,"model") == 0) {
	strcpy(cpuinfoE_set.cpumodel_str,lval0);
	if(debug) printf("%s %s\n",lval,lval0); fflush(NULL);
      }
      else if(strcmp(lval,"model name") == 0) {
	if(i == 0) strcpy(cpuinfoE_set.cpumodname_str,lval0);
	if(debug) printf("%s %s\n",lval,lval0); fflush(NULL);
      }
      else if(strcmp(lval,"cpu MHz") == 0) {
	sscanf(lval0,"%f",&cpuinfoE_set.cpuMHz);
	if(debug) printf("%s %f\n",lval,cpuinfoE_set.cpuMHz); fflush(NULL);
      }
      else if(strcmp(lval,"flags") == 0) {
	strcpy(cpuinfoE_set.cpufeatflg_str,lval0);
	if(debug) printf("%s %s\n",lval,lval0); fflush(NULL);
      }

    } 
    loc_p++;
  } while(*loc_p != '\0');

  free(infile_buf);

  return(i+1);
}


/*```````````````````````````````````````````````````````````````````*/
void proc_syscfg(char *input_fn)
{

/*
!+ Reads "system" configuration values from a file.
!  IN: 'input_fn' is the input filename string

!~ 06.04.2011 (last modified; TIM)
*/

  FILE *fileh;

  fileh = fopen(input_fn, "rb");  

  while(fscanf(fileh,"%s%s",lval,lval0) != EOF) {
    if(strcmp(lval,"sys_canonical_cputype") == 0) {
      strcpy(syscanon_cputype,lval0);
    }
    else if(strcmp(lval,"sys_canonical_vendor") == 0) {
      strcpy(syscanon_vendor,lval0);
    }
    else if(strcmp(lval,"sys_canonical_os") == 0) {
      strcpy(syscanon_os,lval0);
    }
    else if(strcmp(lval,"sys_canonical_endianness") == 0) {
      strcpy(syscanon_endianness,lval0);
    }
    else if(strcmp(lval,"compiler_name_Fortran") == 0) {
      strcpy(compargE_set[IDX_F90].compname_str,lval0);
    }
    else if(strcmp(lval,"compiler_name_C") == 0) {
      strcpy(compargE_set[IDX_C].compname_str,lval0);
    }
    else if(strcmp(lval,"compiler_Fortran") == 0) {
      strcpy(comparg_fs_set[IDX_F90].comploc_str,lval0);
    }
    else if(strcmp(lval,"compiler_C") == 0) {
      strcpy(comparg_fs_set[IDX_C].comploc_str,lval0);
    }
    else if(strcmp(lval,"comp_Fortran_versionf") == 0) {
      strcpy(comparg_fs_set[IDX_F90].compvers_fn,lval0);
    }
    else if(strcmp(lval,"comp_C_versionf") == 0) {
      strcpy(comparg_fs_set[IDX_C].compvers_fn,lval0);
    }
    else if(strcmp(lval,"compargs_cfg_file") == 0) {
      strcpy(compargs_fn,lval0);
    }
  }

  fclose(fileh);
}


/*```````````````````````````````````````````````````````````````````*/
void proc_compversf(void)
{

/*
!+ Parse compiler version info file(s).

!~ 28.04.2010 (last modified; TIM)
*/

  FILE *fileh;
  int i;

  for(i=0; i<N_COMPILER_TYPES; i++) {
    fileh = fopen(comparg_fs_set[i].compvers_fn, "rb");

    if(strcmp(compargE_set[i].compname_str,"pgf90") == 0 ||
       strcmp(compargE_set[i].compname_str,"pgf95") == 0) {
      fscanf(fileh,"%s%s",lval,compargE_set[i].compvers_str);
    }
    else if(strcmp(compargE_set[i].compname_str,"gfortran") == 0) {
      fscanf(fileh,"%s%s%s%s%s",lval,lval,lval,lval,
                                compargE_set[i].compvers_str);
    }
    else if(strcmp(compargE_set[i].compname_str,"ifc") == 0 ||
	    strcmp(compargE_set[i].compname_str,"ifort") == 0) {
      /* Version output format currently not known */
      strcpy(compargE_set[i].compvers_str,"unknown");
    }
    else if(strcmp(compargE_set[i].compname_str,"gcc") == 0) {
      do {
	fscanf(fileh,"%s",lval);
      } while(strcmp(lval,"version") != 0);
      fscanf(fileh,"%s",compargE_set[i].compvers_str);
    }
    else if(strcmp(compargE_set[i].compname_str,"cc") == 0) {
      /* Version output format currently not known */
      strcpy(compargE_set[i].compvers_str,"unknown");
    }
    else {
      /* Version output format currently not known */
      strcpy(compargE_set[i].compvers_str,"unknown");
    }

    fclose(fileh);
  }
}


/*```````````````````````````````````````````````````````````````````*/
int eval_exp_parse_ca(char *raw, char *tstr, char *undef)
{

/*
!+ Parser entry point

!~ 28.04.2010 (last modified; TIM)
*/

  int answer;

  prog = raw;
  if(!strcmp(undef,prog)) return(1);  /* if undefined value, return true */
  answer = 0;  /* Assume false at init */

  get_token_parse_ca();
  if(*token == '\0') {
    printf("\nERROR: parser given empty expression\n");
    exit(99);
  }
  
  eval_exp1_parse_ca(&answer,tstr);
  
  if(*token != '\0') {   /* Last token must be null */
    printf("\nERROR: last token parser sees must be a null char\n");
    exit(99);
  }
  
  return(answer);
}


/*```````````````````````````````````````````````````````````````````*/
void eval_exp1_parse_ca(int *answer, char *tstr)
{

/*
!+ Evaluate relational operators and substring matching

!~ 28.04.2010 (last modified; TIM)
*/

  if(tok_type == REL_OPERATOR) {
    get_token_parse_ca();
    /*    switch(op_type) {
    case NE:
      *answer = strcmp(tstr,token) != 0;
      break;
    case LT:
      printf("LT (<): %s vs. %s",tstr,token);
      *answer = strcmp(tstr,token) < 0;
      break;
    case LE:
      *answer = strcmp(tstr,token) <= 0;
      break;
    case GT:
      *answer = strcmp(tstr,token) > 0;
      break;
    case GE:
      *answer = strcmp(tstr,token) >= 0;
      break;
      }*/
    numcmpchk_parse_ca(answer,token,tstr);
    get_token_parse_ca();
  }
  else if(tok_type == ITEM) {
    *answer = strstr(tstr,token) != NULL;
    get_token_parse_ca();
  }

  eval_exp2_parse_ca(answer,tstr);
}


/*```````````````````````````````````````````````````````````````````*/
void eval_exp2_parse_ca(int *answer, char *tstr)
{

/*
!+ Process conjunctional expressions

!~ 28.04.2010 (last modified; TIM)
*/

  int atemp;
  char op;

  if(tok_type == CONJUNCTION) {
    op = *token;
    get_token_parse_ca();
    eval_exp1_parse_ca(&atemp,tstr);

    if(op == '|') *answer = *answer || atemp;
    else if (op == ',') *answer = *answer && atemp;

    get_token_parse_ca();
    eval_exp1_parse_ca(answer,tstr);
  }
  else return;

}


/*```````````````````````````````````````````````````````````````````*/
void get_token_parse_ca(void)
{

/*
!+ Return the next token (with token and/or operator type)

!~ 28.04.2010 (last modified; TIM)
*/

  register char *temp;
  char op;


  tok_type = -1;
  temp = token;
  *temp = '\0';

  while(isspace(*prog)) ++prog;  /* skip over white space */
  if(*prog == '\0') return;   /* at end of expression */

  if(isdelim_parse_ca(*prog)) {
    tok_type = DELIMITER;
    op = *prog;
    *temp++ = *prog++;      /* advance to next char */
    if(strchr("<>!",op)) {
      tok_type = REL_OPERATOR;
      if(*prog == '=') {
	if(op == '<') op_type = LE;
        else if(op == '>') op_type = GE;
        else if(op == '!') op_type = NE;
	*temp++ = *prog++; 
      }
      else if(op == '<') op_type = LT;
      else if(op == '>') op_type = GT;
    }
    else if(strchr("|,",op)) {
      tok_type = CONJUNCTION;
    }
  }
  else {
    tok_type = ITEM;
    while(!isdelim_parse_ca(*prog) && !isspace(*prog)) *temp++ = *prog++;
  }

  *temp = '\0';  /* terminate token string */

  /*  if(tok_type >= 0) {
    printf("%s : %s#\n",ttnames[tok_type].name,token);fflush(NULL);
    }*/
}


/*```````````````````````````````````````````````````````````````````*/
int isdelim_parse_ca(char c)
{

/*
!+ Return true if c is a "valid" delimiter

!~ 28.04.2010 (last modified; TIM)
*/

  if(strchr("!,|<>=",c) != NULL) return 1;
  return 0;
}


/*```````````````````````````````````````````````````````````````````*/
void numcmpchk_parse_ca(int *answer, char *str1, char *str2)
{

/*
!+ Check if relational operation comparison should be "numerical" or
!   "lexicographical"; return result of comparison.

!~ 28.04.2010 (last modified; TIM)
*/

  char *tmp1,*tmp2,*tmp;
  char substr[64];
  int chk_num,i1,i2,i;

  tmp1 = str1;
  tmp2 = str2;

  /* Check if strings are composed only of numbers and certain symbols: */
  do {
    chk_num = isdigit(*tmp1) || (strchr("_.-",*tmp1) != NULL);
    chk_num = isdigit(*tmp2) || (strchr("_.-",*tmp2) != NULL);
    tmp1++;
    tmp2++;
    if(*tmp1 == '\0' || *tmp2 == '\0') break;
  } while(chk_num);

  tmp1 = str1;
  tmp2 = str2;

  if(chk_num) {   /* Perform "arithmetic" comparison */

    for( ; ; ) {
      i = 0;
      for(tmp=tmp1; ((strchr("_.-",*tmp) == NULL) && *tmp != '\0'); tmp++) {
	substr[i] = *tmp;
	i++;
      }
      tmp1 = tmp;
      substr[i] = '\0';   /* Terminate substring */
      i1 = atoi(substr);

      i = 0;
      for(tmp=tmp2; ((strchr("_.-",*tmp) == NULL) && *tmp != '\0'); tmp++) {
	substr[i] = *tmp;
	i++;
      }
      tmp2 = tmp;
      substr[i] = '\0';   /* Terminate substring */
      i2 = atoi(substr);

      switch(op_type) {
      case LT:
	*answer = i2 < i1;
	break;
      case LE:
	*answer = i2 <= i1;
	break;
      case GT:
	*answer = i2 > i1;
	break;
      case GE:
	*answer = i2 >= i1;
	break;
      }

      if(!*answer) return;  /* Exit if any numerical comparison is false */
      if(*tmp1 == '\0' || *tmp2 == '\0') return;
      tmp1++;
      tmp2++;
    }

  }
  else {     /* Perform "lexicographic" comparison */

    switch(op_type) {
    case NE:
      *answer = strcmp(str2,str1) != 0;
      break;
    case LT:
      *answer = strcmp(str2,str1) < 0;
      break;
    case LE:
      *answer = strcmp(str2,str1) <= 0;
      break;
    case GT:
      *answer = strcmp(str2,str1) > 0;
      break;
    case GE:
      *answer = strcmp(str2,str1) >= 0;
      break;
    }

  }

  return;
}
