!```````````````````````````````````````````````````````````````````
module PRAMS_chkcfg_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine chkcfg1

!+ Check for some initial namelist value conflicts

use univ_kind_defs_mod, only: f8
use main_memory_mod, only: run_is_purely_2D, n_grids, naddsc, maximum_sim_ET,  &
     run_type, gd, there_is_an_error, start_sim_ET, dt_dynamics, RT_paradigm,  &
     run_is_parallel, initialization_type
use main_memory_bspec_mod, only: dust_tog, H2Ov_tog, H2Oi_tog, CO2i_tog,  &
     dust_nb, H2Oi_nb, CO2i_nb, dust_tog, bdst_tog, RTc_tog, RT_flux_trace_np
use aermic_rad_mem_mod, only: iaertog

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: n, icm, ifm

!=== Executable statements:

! Check nxtnest values for validity:

if (gd(1)%nxtnest /= 0) then
   write (*, '(/,a,/,a)')  & 
        'ERROR: Grid #1 must have its parent grid set as grid #0',  &
        '   [ nxtnest(1) = 0) ]'  
   there_is_an_error = .TRUE.
end if

n = 0
do ifm=1,n_grids
   icm = gd(ifm)%nxtnest

   if (icm >= ifm) then
      write (*, '(/,a,i3,a,/,a)')  & 
           'ERROR: Nested grid #', ifm, ' has specified parent mesh of ',  &
           '     equal or higher number.'  
      there_is_an_error = .TRUE.
   end if

   if (icm < 0) then
      write (*, '(/,a,i3,a,/,a)')  & 
           'ERROR: Nested grid #', ifm, ' has specified parent grid of ',  &
           '     a number less than zero.'  
      there_is_an_error = .TRUE.
   end if
  
   if (icm == 0) then
      n = n+1
   end if

end do

if (n > 1) then
   write (*, '(/,a)')  & 
        'ERROR: More than one grid has grid #0 specified as its parent.'  
   there_is_an_error = .TRUE.
end if

! Ascertain if the run and which grids *APPEAR* to be 2D in nature:
run_is_purely_2D = .FALSE.
do ifm=1,n_grids
   icm = gd(ifm)%nxtnest

   gd(ifm)%grid_is_2D = ( gd(ifm)%nyp == 1 )
   run_is_purely_2D = ( icm == 0 .and. gd(ifm)%grid_is_2D ) 
end do

! If run appears to be 2D, check that stuff is configured correctly:
if (run_is_purely_2D) then
   do icm=1,n_grids
      if (.NOT.gd(icm)%grid_is_2d) then
         write (*, '(/,a,/,a)')  & 
              'ERROR: Appears to be a fully 2D run, but all grids are not ',  &
              'configured correctly [ nnyp(ngrid) = 1 ]'
         there_is_an_error = .TRUE.
         return   ! Error is grave enough to exit
      end if
   end do
end if

! Check for validity of 2D nested grids (nested within 2D or 3D grids):
do ifm=1,n_grids
   icm = gd(ifm)%nxtnest

   if (icm /= 0) then  ! Only look at nested grids

      if (gd(icm)%grid_is_2D .and. (.NOT.gd(ifm)%grid_is_2D)) then
         write (*, '(/,a,/,a,i4,a,i4,a)')  & 
              'ERROR: A 3D grid may not be NESTED within a 2D coarser grid.',  &
              '( offending grids: parent =', icm, 'nest =', ifm, ' )'
         there_is_an_error = .TRUE.
         return   ! Error is grave enough to exit
      else
         if (gd(ifm)%grid_is_2D .and.  &
              (gd(ifm)%ninest < 3 .or. gd(ifm)%njnest < 1)) then
            write (*, '(/,2a)')  &
                 'ERROR: Nested 2D grid must have NINEST > 2 ',  &
                 'and NJNEST = 1 (in model namelist).'
            there_is_an_error = .TRUE.
         end if
      end if

      if (gd(ifm)%ninest > gd(icm)%nxp) then
         write (*, '(/,a,i3,a)') 'ERROR: NINEST for grid #', ifm,  &
           ' is greater than the NXP of its parent grid.'
         there_is_an_error = .TRUE.
      end if
      if (gd(ifm)%njnest > gd(icm)%nyp) then
         write (*, '(/,a,i3,a)') 'ERROR: NJNEST for grid #', ifm,  &
           ' is greater than the NYP of its parent grid.'
         there_is_an_error = .TRUE.
      end if

   end if
   
   if (gd(ifm)%grid_is_2D .and. gd(ifm)%nstraty /= 1) then
      write (*, '(/,a,i3,a,/,a)') 'ERROR: Grid #', ifm,  &
           ' appears to be 2D, but is incorrectly configured.',  &
           '   [ nstraty(ngrid) must equal 1]'
      there_is_an_error = .TRUE.
   end if

end do

! Check to make sure that the coarsest grids have NSTTOP and NSTBOT set to 1:

if (gd(1)%nsttop /= 1) then
   write (*, '(/,a)')  &
        'ERROR: NSSTOP must be set to 1 for the coarsest grid.'
   there_is_an_error = .TRUE.
end if

if (gd(1)%nstbot /= 1) then
   write (*, '(/,a)')  &
        'ERROR: NSSBOT must be set to 1 for the coarsest grid.'
   there_is_an_error = .TRUE.
end if

! "Sanity" check for restarts:
if (run_type == 'RESTART') then
   if (start_sim_ET >= maximum_sim_ET+real(dt_dynamics, kind=f8)) then
      write (*, '(/,a)')  &
           'ERROR: state file time must be less than MAXIMUM_SIM_ET.'
      there_is_an_error = .TRUE.
   end if
end if

if (run_type == 'MAKE_VAR_FILES' .and. initialization_type < 2) then
   write (*, '(/,a)')  &
        'ERROR: INITIALIZATION_TYPE must >= 2 for RUNTYPE="MAKE_VAR_FILES".'
   there_is_an_error = .TRUE.
end if

! Check aerosol microphysics model options for compatibility with each other:

if (H2Ov_tog == 0 .and. H2Oi_tog >= 1) then
   write (*, '(/,a,/)')  &
  'ERROR: aerosol options incompatible (no water vapor to freeze).'
   there_is_an_error = .TRUE.
end if

if (dust_tog == 1 .and. (H2Oi_tog >= 1 .or. CO2i_tog >= 1)) then
   write (*, '(/,a,/)')  &
  'ERROR: aerosol options incompatible (cannot nucleate static dust).'
   there_is_an_error = .TRUE.
end if

if (H2Ov_tog == 1 .and. H2Oi_tog >= 1) then
   write (*, '(/,a,/)')  &
  'ERROR: aerosol options incompatible (cannot freeze static water vapor).'
   there_is_an_error = .TRUE.
end if

if (H2Ov_tog >= 1 .and. CO2i_tog >= 1 .and. H2Oi_tog == 0) then
   write (*, '(/,a,/)')  &
  'ERROR: aerosol options incompatible (CO2 ice and H2O vapor need H2O ice).'
   there_is_an_error = .TRUE.
end if

if (CO2i_tog >= 1 .and. naddsc == 0) then
   write (*, '(/,a,/)')  &
  'ERROR: CO2I_TOG >= 1 requires NADDSC >= 1 (for non-condensible enrichment).'
   there_is_an_error = .TRUE.
end if

!  Force aerosol bin counts to zero if necessary; set aertog value: 

iaertog = 3
if (dust_tog == 0) then
   dust_nb = 0
   iaertog = iaertog-1
end if
if (H2Oi_tog == 0) then
   H2Oi_nb = 0
   iaertog = iaertog-1
end if
if (CO2i_tog == 0) then
   CO2i_nb = 0
   iaertog = iaertog-1
end if

! Check for "background dust" configuration requirements:

if (RT_paradigm == 2) then
   if ((RTc_tog == 2 .or. RTc_tog == 3) .and.  &
        (bdst_tog > 0 .and. dust_tog == 0)) then
      write(*, '(/,a)') 'ERROR: BDST_TOG > 0 requires that DUST_TOG > 0'
      there_is_an_error = .TRUE.
   end if
end if

! Disallow RT_flux_trace operations for a parallel run:
if (run_is_parallel .and. RT_flux_trace_np > 0) then
   write(*, '(/,a)')  &
        'ERROR: RT_FLUX_TRACE_NP > 0 cannot be used for a parallel run'
   there_is_an_error = .TRUE.
end if

end subroutine chkcfg1


!```````````````````````````````````````````````````````````````````
subroutine chkcfg2

!+ Check that nested grids are placed/dimensioned correctly, etc.

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use main_memory_mod, only: n_grids, there_is_an_error, gd, nestz1
use aermic_rad_mem_mod, only: aer_inj_gpbeg, aer_gpbeg, aer_csr_gpbeg

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'chkcfg2'

integer :: ifm, icm, ncx, ncy, nfx, nfxp, nfy, nfyp, nesta, nfz, kc

!=== Executable statements:

do ifm=1,n_grids
   icm = gd(ifm)%nxtnest

   if (icm >= 1) then
      ncx = (gd(ifm)%nxp-2)/gd(ifm)%nstratx
      ncy = (gd(ifm)%nyp-2)/gd(ifm)%nstraty

      if ((gd(ifm)%nyp == 1 .and. gd(icm)%nyp /= 1) .or.  &
           (gd(ifm)%nyp /= 1 .and. gd(icm)%nyp == 1)) then
         write (*, '(/,a)') 'ERROR: grids must be either all 3D or 2D'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%ninest < 3) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' is too close to W boundary of coarser grid.'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%njnest < 3 .and. .not.gd(ifm)%grid_is_2D) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' is too close to S boundary of coarser grid.'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%nknest < 3 .and. gd(ifm)%nstbot == 0) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' is too close to lower boundary of coarser grid.'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%nknest /= 1 .and. gd(ifm)%nstbot == 1) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' does not go to lower boundary of coarser grid.'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%ninest+ncx > gd(icm)%nxp-3) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' is too close to E boundary of coarser grid.'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%njnest+ncy > gd(icm)%nyp-3 .and. .not.gd(ifm)%grid_is_2D) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' is too close to N boundary of coarser grid.'
         there_is_an_error = .TRUE.
      end if

      if (ncx < 2 .or. (ncy < 2 .and. .not.gd(ifm)%grid_is_2D)) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' dimensioned too small in x- and/or y- direction.'
         there_is_an_error = .TRUE.
      end if

      nfx = ncx*gd(ifm)%nstratx+2
      if (gd(ifm)%nxp /= nfx) then
         nfxp = nfx+gd(ifm)%nstratx
         write (*, '(/,a,i2,a,/,a,i4,a,i4)') 'ERROR: Nested grid #', ifm,  &
              ' NXP incompatible with NSTRATX (nest ratio):',  &
              '       May increase NXP to ', nfxp, ' or decrease it to ', nfx
         there_is_an_error = .TRUE.
      end if

      nfy = ncy*gd(ifm)%nstraty+2
      if (gd(ifm)%nyp /= nfy .and. .not.gd(ifm)%grid_is_2D) then
         nfyp = nfy+gd(ifm)%nstraty
         write (*, '(/,a,i2,a,/,a,i4,a,i4)') 'ERROR: Nested grid #', ifm,  &
              ' NYP incompatible with NSTRATY (nest ratio):',  &
              '       May increase NYP to ', nfyp, ' or decrease it to ', nfy
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%nstbot == 1 .and. gd(icm)%nstbot /= 1) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' NSTBOT value incompatible with coarser mesh value.'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%nsttop == 1 .and. gd(icm)%nsttop /= 1) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' NSTTOP value incompatible with coarser mesh value.'
         there_is_an_error = .TRUE.
      end if

   end if
end do

nesta = abs(nestz1)
if (nestz1 /= 0 .and. nesta <= n_grids) then
   nfz = gd(nesta)%nzp-2
   kc = gd(nesta)%nknest
   do while(nfz > 0)
      kc = kc+1
      nfz = nfz-gd(nesta)%nrz(kc)
      if (nfz < 0) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', nesta,  &
              ' has illegal number of levels for given NSTRATZ values.'
        there_is_an_error = .TRUE.
      end if
   end do
   if (nfz == 0) then
      if (kc > gd(gd(nesta)%nxtnest)%nzp-3 .and. gd(nesta)%nsttop /= 1) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', nesta,  &
              ' top is too high or NNSTTOP set incorrectly.'
         there_is_an_error = .TRUE.
      end if
     
      if (kc /= gd(gd(nesta)%nxtnest)%nzp-1 .and. gd(nesta)%nsttop == 1) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', nesta,  &
              ' top not the same as top of coarser grid (check NNSTTOP).'
         there_is_an_error = .TRUE.   
      end if

   end if
end if

do ifm=1,n_grids
   icm = gd(ifm)%nxtnest

   if (ifm /= nesta .and. icm >= 1) then
      if (gd(ifm)%nzp > gd(icm)%nzp-gd(ifm)%nknest-1 .and.  &
           gd(ifm)%nsttop == 0) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' NZP incompatible with parent grid (check NSTTOP too).'
         there_is_an_error = .TRUE.
      end if

      if (gd(ifm)%nzp /= gd(icm)%nzp-gd(ifm)%nknest+1 .and.  &
           gd(ifm)%nsttop == 1) then
         write (*, '(/,a,i2,a)') 'ERROR: Nested grid #', ifm,  &
              ' not to top of coarser grid (check NNSTTOP too).'
         there_is_an_error = .TRUE.
      end if
   end if
end do

if (aer_inj_gpbeg(1) < aer_gpbeg(1)) then
   call exception(procedure_name, FATAL,  &
        'DUST_JBG must be >= DUST_BEG')
end if

if (aer_csr_gpbeg(1) > aer_inj_gpbeg(1)) then
   call exception(procedure_name, FATAL,  &
        'DUST_JSR must be <= DUST_JPG')
end if

end subroutine chkcfg2


!```````````````````````````````````````````````````````````````````
subroutine chkcfg3

!+ Check that configured model options do not conflict, etc.

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use univ_numtype_constants_mod, only: epsilon_f4
use main_memory_mod, only: there_is_an_error, n_grids, dzmax, deltaz_g1,  &
     dzrat, nfpt, iadvl, iadvf, ihorgrad, nzg, nzs, initialization_type,  &
     distim, dt_dynamics, nudlat, tnudlat, tnudcent, tnudtop,  &
     max_dt_dynamics, radfrq, gd, dt_mode, run_is_parallel
use main_memory_bspec_mod, only: slz, H2Ov_tog, H2Ov_ivr

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: igd, k

!=== Executable statements:

do igd=1,n_grids
   if ((abs(gd(igd)%gridu) > 1.e-20_f4 .or. abs(gd(igd)%gridv) > 1.e-20_f4)  &
        .and. gd(igd)%itoptflg /= 0 .and. gd(igd)%itinflg /= 0  &
        .and. gd(igd)%ialbflg /= 0 ) then
      write (*, '(/,a,/,a)')   &
           'ERROR: nested grid topography/sfc_chars must be interpolated',  &
           ' from parent grid if nested grid moves.'
      there_is_an_error = .TRUE.
   end if
end do

! Check horizontal and vertical grid spacings:

IF (DZMAX < deltaz_g1) THEN
   write (*, '(/,a)') 'WARNING: dz is being REDUCED with height (down to DZMAX)'
END IF

IF (DZRAT > 1.2_f4) THEN
   write (*, '(/,a,/,a)')  &
        'WARNING: Large vertical stretch ratio (will sacrifice ',  &
        ' second order accuracy in vertical differencing).'
END IF

! Check numerical schemes:

IF (NFPT > gd(1)%NZP) THEN
   write(*, '(/,a)') 'ERROR: NFPT must be less than NZP(1)'
   there_is_an_error = .TRUE.
END IF

IF (IADVL /= 2 .AND. IADVL /= 4) then
   write (*, '(/,a)') 'ERROR: IADVL must be 2 or 4'
   there_is_an_error = .TRUE.
END IF

IF (IADVF /= 2 .AND. IADVF /= 6) then
   write (*, '(/,a)') 'ERROR: IADVF must be 2 or 6'
   there_is_an_error = .TRUE.
END IF

! DT_MODE < 0 does not currently work correctly in parallel:
if (dt_mode < 0 .and. run_is_parallel) then
   write (*, '(/,a)')  &
        'ERROR: DT_MODE < 0 currently does not work correctly in parallel'
   there_is_an_error = .TRUE.
end if

! Where relevant (dt_mod < 0), require max_dtlong to be < radfrq (try to
!  minimize RT calls that are too far apart):

if (dt_mode < 0 .and. max_dt_dynamics >= radfrq) then
   write (*, '(/,a)') 'ERROR: MAX_DT_DYNAMICS must be < RADFRQ'
   there_is_an_error = .TRUE.
end if

! Check that diffusion flags are compatible if using ihorgrad=2:

if (ihorgrad == 2) then
   do igd=1,n_grids
      if (gd(igd)%IDIFFK == 3 .or. gd(igd)%IDIFFK == 4 .or.  &
           gd(igd)%IDIFFK == 5) then
         write(*, '(/,a)') 'ERROR: Can''t use IHORGRAD=2 if IDIFFK=3 or 4 or 5'
         there_is_an_error = .TRUE.
      end if
   end do
end if

! Check whether the soil model will be run and make sure that the
!   number of soil levels are correct.(severity - F,I )

IF (NZG < 5) THEN
   write(*, '(/,a)') 'ERROR: SBPS model needs at least 5 subsurface levels'
   there_is_an_error = .TRUE.
END IF

if (nzs < 1) then
   write(*, '(/,a)') 'ERROR: SBPS model needs at least 1 supersurface level'
   there_is_an_error = .TRUE.
end if

do k=1,nzg
   if (slz(k) > -0.0001_f4) then
      write(*, '(/,a,i2,a)') 'ERROR: Subsurface level ',k,' not below surface'
      there_is_an_error = .TRUE.
   end if
end do

! Make sure that if nudging, nudging time scales are greater than
! the model coarse grid timestep, and that Rayleigh friction nudging
! is not done with variable initialization:

if (initialization_type == 1) then
   if (nfpt > 0 .and. distim > ZERO_f4 .and. distim < dt_dynamics) then
      write(*, '(/,a,/,a)') 'ERROR: Rayleigh friction nudging is being done,'  &
           ,' and DISTIM is less than DT_DYNAMICS. This nudging is too strong.'
      there_is_an_error = .TRUE.
   end if
end if

if (initialization_type == 2) then

   if (nfpt > 0 .and. distim > ZERO_f4) then
      write(*, '(/,a,/,a)')   &
           'ERROR: Rayleigh friction nudging may not be used when',  &
           ' variable initialization is used.'
      there_is_an_error = .TRUE.
   end if

   if (nudlat >= 1 .and. tnudlat > epsilon_f4 .and. tnudlat < dt_dynamics) then
      write(*, '(/,a,/,a)') 'ERROR: Lateral boundary nudging is being done',  &
           ' and TNUDLAT is less than DT_DYNAMICS. This nudging is too strong.'
      there_is_an_error = .TRUE.
   end if

   if (tnudcent > epsilon_f4 .and. tnudcent < dt_dynamics) then
      write(*, '(/,a,/,a)') 'ERROR: Interior (center) nudging is being done',  &
           ' and TNUDCENT is less than DT_DYNAMICS. This nudging is too strong.'
      there_is_an_error = .TRUE.
   end if

   if (tnudtop > epsilon_f4 .and. tnudtop < dt_dynamics) then
      write(*, '(/,a,/,a)') 'ERROR: Top boundary nudging is being done',  &
           ' and TNUDTOP is less than DT_DYNAMICS. This nudging is too strong.'
      there_is_an_error = .TRUE.
   end if

end if

if (H2Ov_tog > 0 .and. initialization_type == 1 .and. H2Ov_ivr == 0) then
   write(*, '(/,a)')  &
        'For H2OV_TOG > 0 and INITIALIZATION_TYPE = 1, H2OV_IVR must be > 0'
   there_is_an_error = .TRUE.
end if

end subroutine chkcfg3


!```````````````````````````````````````````````````````````````````
subroutine chkcfg4

!+ Check the namelist PIPP configuration for conflicts

use main_memory_mod, only: there_is_an_error, n_grids, nigrids

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

! NONE

!=== Executable statements:

! Don't allow NIGRIDS /= N_GRIDS:

if (nigrids /= n_grids) then
   write (*, '(/,a)') 'ERROR: NIGRIDS must be equal to N_GRIDS.'
   there_is_an_error = .TRUE.
end if

end subroutine chkcfg4


end module PRAMS_chkcfg_mod
