!```````````````````````````````````````````````````````````````````
module PRAMS_graddiv_mod

!~ 10.10.2008 (last modified; TIM)

!+

implicit none

private

public :: gradient, divcart, gradxu, gradxt, gradyt, gradyv, gradzw, gradzt

contains


!```````````````````````````````````````````````````````````````````
subroutine gradient(m1,m2,m3,ia,iz,ja,jz,array_A,array_B,direction,gdpt_type)

!~ 19.11.2008 (last modified; TIM)

!+ Driver to compute a gradient

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz
real(f4), dimension(m1,m2,m3), intent(IN) :: array_A
character(len=*), intent(IN) :: direction,gdpt_type

real(f4), dimension(m1,m2,m3), intent(INOUT) :: array_B

!== Local declarations:

character(len=6) :: optype

!=== Executable statements:

optype = 'GRADNT'

call gradq(m1,m2,m3,ia,iz,ja,jz,array_A,array_B,direction,gdpt_type,optype)

end subroutine gradient


!```````````````````````````````````````````````````````````````````
subroutine divcart(m1,m2,m3,ia,iz,ja,jz,array_A,array_B,direction,gdpt_type)

!~ 19.11.2008 (last modified; TIM)

!+ Driver to compute divergence in Cartesian coordinates

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz
real(f4), dimension(m1,m2,m3), intent(IN) :: array_A
character(len=*), intent(IN) :: direction,gdpt_type

real(f4), dimension(m1,m2,m3), intent(INOUT) :: array_B

!== Local declarations:

character(len=6) :: optype

!=== Executable statements:

optype = 'DIVCRT'

call gradq(m1,m2,m3,ia,iz,ja,jz,array_A,array_B,direction,gdpt_type,optype)

end subroutine divcart


!```````````````````````````````````````````````````````````````````
subroutine gradxu(m1,m2,m3,ia,iz,ja,jz,optype,VC3DA,VC3DB,VC1DA,RTGE,RTGC  &
     ,DX,DZ,FMAPI,FMAP,FQ,HQ,HQ4,LEV,JD)

!~ 19.11.2008 (last modified; TIM)

!+ Compute a gradient

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: VC3DA
real(f4), dimension(m1), intent(IN) :: DZ,HQ
real(f4), dimension(m2,m3), intent(IN) :: RTGE,RTGC,DX,FMAP,FMAPI,FQ
character(len=*), intent(IN) :: optype,LEV

real(f4), dimension(m1), intent(INOUT) :: HQ4,VC1DA
real(f4), dimension(m1,m2,m3), intent(INOUT) :: VC3DB

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

IF(optype == 'GRADNT') THEN
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J)=(VC3DA(K,I,J)*RTGE(I,J)  &
                 -VC3DA(K,I-1,J)*RTGE(I-1,J))  &
                 *DX(I,J)/RTGC(I,J)
         ENDDO
      ENDDO
   ENDDO
ELSE
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J)=(VC3DA(K,I,J)*RTGE(I,J)  &
                 *FMAPI(I,J)  &
                 -VC3DA(K,I-1,J)*RTGE(I-1,J)  &
                 *FMAPI(I-1,J))  &
                 *DX(I,J)/RTGC(I,J)*FMAP(I,J)
         ENDDO
      ENDDO
   ENDDO
ENDIF

IF(optype /= 'DIVSTR') THEN
   IF(LEV == 'W') THEN
      DO K=1,m1
         HQ4(K) = 0.25_f4*HQ(K)
      ENDDO
   ELSE
      DO K=2,m1
         HQ4(K) = 0.25_f4*HQ(K-1)
      ENDDO
   ENDIF
   
   DO J=ja,jz
      DO I=ia,iz
         DO K=2,m1
            VC1DA(K) = HQ4(K)*(VC3DA(K,I,J)+VC3DA(K-1,I,J)  &
                 +VC3DA(K,I-1,J)+VC3DA(K-1,I-1,J))
         ENDDO
         DO K=2,m1-1
            VC3DB(K,I,J) = VC3DB(K,I,J)  &
                 +FQ(I,J)*DZ(K)*(VC1DA(K+1)-VC1DA(K))
         ENDDO
         VC3DB(1,I,J) = VC3DB(2,I,J)
         IF(LEV == 'W') then
            VC3DB(m1-1,I,J) = VC3DB(m1-2,I,J)
         elseif(LEV == 'T') then
            VC3DB(m1,I,J) = VC3DB(m1-1,I,J)
         endif
      ENDDO
   ENDDO
ENDIF

end subroutine gradxu


!```````````````````````````````````````````````````````````````````
subroutine gradxt(m1,m2,m3,ia,iz,ja,jz,optype,VC3DA,VC3DB,VC1DA,RTGE,RTGC  &
     ,DX,DZ,FMAPI,FMAP,FQ,HQ,HQ4,LEV,JD)

!~ 19.11.2008 (last modified; TIM)

!+ Compute a gradient

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: VC3DA
real(f4), dimension(m1), intent(IN) :: DZ,HQ
real(f4), dimension(m2,m3), intent(IN) :: RTGE,RTGC,DX,FMAP,FMAPI,FQ
character(len=*), intent(IN) :: optype,LEV

real(f4), dimension(m1), intent(INOUT) :: HQ4,VC1DA
real(f4), dimension(m1,m2,m3), intent(INOUT) :: VC3DB

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

IF(optype == 'GRADNT') THEN
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J) = (VC3DA(K,I+1,J)*RTGE(I+1,J)  &
                 -VC3DA(K,I,J)*RTGE(I,J))  &
                 *DX(I,J)/RTGC(I,J)
         ENDDO
      ENDDO
   ENDDO
ELSE
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J) = (VC3DA(K,I+1,J)*RTGE(I+1,J)  &
                 *FMAPI(I+1,J)  &
                 -VC3DA(K,I,J)*RTGE(I,J)  &
                 *FMAPI(I,J))  &
                 *DX(I,J)/RTGC(I,J)*FMAP(I,J)
         ENDDO
      ENDDO
   ENDDO
ENDIF

IF(optype /= 'DIVSTR') THEN
   IF(LEV == 'W') THEN
      DO K=1,m1
         HQ4(K) = 0.25_f4*HQ(K)
      ENDDO
   ELSE
      DO K=2,m1
         HQ4(K) = 0.25_f4*HQ(K-1)
      ENDDO
   ENDIF

   DO J=ja,jz
      DO I=ia,iz
         DO K=2,m1
            VC1DA(K) = HQ4(K)*(VC3DA(K,I,J)+VC3DA(K-1,I,J)  &
                 +VC3DA(K,I+1,J)+VC3DA(K-1,I+1,J))
         ENDDO
         DO K=2,m1-1
            VC3DB(K,I,J) = VC3DB(K,I,J)  &
                 +FQ(I,J)*DZ(K)*(VC1DA(K+1)-VC1DA(K))
         ENDDO
         VC3DB(1,I,J) = VC3DB(2,I,J)
         IF(LEV == 'W') then
            VC3DB(m1-1,I,J) = VC3DB(m1-2,I,J)
         elseif(LEV == 'T') then
            VC3DB(m1,I,J) = VC3DB(m1-1,I,J)
         endif
      ENDDO
   ENDDO
ENDIF

end subroutine gradxt


!```````````````````````````````````````````````````````````````````
subroutine gradyv(m1,m2,m3,ia,iz,ja,jz,optype,VC3DA,VC3DB,VC1DA,RTGE,RTGC  &
     ,DY,DZ,FMAPI,FMAP,FQ,HQ,HQ4,LEV,JD)

!~ 19.11.2008 (last modified; TIM)

!+ Compute a gradient

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: VC3DA
real(f4), dimension(m1), intent(IN) :: DZ,HQ
real(f4), dimension(m2,m3), intent(IN) :: RTGE,RTGC,DY,FMAP,FMAPI,FQ
character(len=*), intent(IN) :: optype,LEV

real(f4), dimension(m1), intent(INOUT) :: HQ4,VC1DA
real(f4), dimension(m1,m2,m3), intent(INOUT) :: VC3DB

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

IF(optype == 'GRADNT') THEN
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J) = (VC3DA(K,I,J)*RTGE(I,J)  &
                 -VC3DA(K,I,J-jd)*RTGE(I,J-jd))  &
                 *DY(I,J)/RTGC(I,J)
         ENDDO
      ENDDO
   ENDDO
ELSE
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J) = (VC3DA(K,I,J)*RTGE(I,J)  &
                 *FMAPI(I,J)  &
                 -VC3DA(K,I,J-jd)*RTGE(I,J-jd)  &
                 *FMAPI(I,J-jd))  &
                 *DY(I,J)/RTGC(I,J)*FMAP(I,J)
         ENDDO
      ENDDO
   ENDDO
ENDIF

IF(optype /= 'DIVSTR') THEN
   IF(LEV == 'W') THEN
      DO K=1,m1
         HQ4(K) = 0.25_f4*HQ(K)
      ENDDO
   ELSE
      DO K=2,m1
         HQ4(K) = 0.25_f4*HQ(K-1)
      ENDDO
   ENDIF

   DO J=ja,jz
      DO I=ia,iz
         DO K=2,m1
            VC1DA(K) = HQ4(K)*(VC3DA(K,I,J)+VC3DA(K-1,I,J)  &
                 +VC3DA(K,I,J-jd)+VC3DA(K-1,I,J-jd))
         ENDDO
         DO K=2,m1-1
            VC3DB(K,I,J) = VC3DB(K,I,J)  &
                 +FQ(I,J)*DZ(K)*(VC1DA(K+1)-VC1DA(K))
         ENDDO
         VC3DB(1,I,J) = VC3DB(2,I,J)
         IF(LEV == 'W') then
            VC3DB(m1-1,I,J) = VC3DB(m1-2,I,J)
         elseif(LEV == 'T') then
            VC3DB(m1,I,J) = VC3DB(m1-1,I,J)
         endif
      ENDDO
   ENDDO
ENDIF

end subroutine gradyv


!```````````````````````````````````````````````````````````````````
subroutine gradyt(m1,m2,m3,ia,iz,ja,jz,optype,VC3DA,VC3DB,VC1DA,RTGE,RTGC  &
     ,DY,DZ,FMAPI,FMAP,FQ,HQ,HQ4,LEV,JD)

!~ 19.11.2008 (last modified; TIM)

!+ Compute a gradient

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: VC3DA
real(f4), dimension(m1), intent(IN) :: DZ,HQ
real(f4), dimension(m2,m3), intent(IN) :: RTGE,RTGC,DY,FMAP,FMAPI,FQ
character(len=*), intent(IN) :: optype,LEV

real(f4), dimension(m1), intent(INOUT) :: HQ4,VC1DA
real(f4), dimension(m1,m2,m3), intent(INOUT) :: VC3DB

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

if(optype == 'GRADNT') THEN
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J) = (VC3DA(K,I,J+jd)*RTGE(I,J+jd)  &
                 -VC3DA(K,I,J)*RTGE(I,J))  &
                 *DY(I,J)/RTGC(I,J)
         ENDDO
      ENDDO
   ENDDO
ELSE
   DO J=ja,jz
      DO I=ia,iz
         DO K=1,m1
            VC3DB(K,I,J) = (VC3DA(K,I,J+jd)*RTGE(I,J+jd)  &
                 *FMAPI(I,J+jd)  &
                 -VC3DA(K,I,J)*RTGE(I,J)  &
                 *FMAPI(I,J))  &
                 *DY(I,J)/RTGC(I,J)*FMAP(I,J)
         ENDDO
      ENDDO
   ENDDO
ENDIF

IF(optype /= 'DIVSTR') THEN
   IF(LEV == 'W') THEN
      DO K=1,m1
         HQ4(K) = 0.25_f4*HQ(K)
      ENDDO
   ELSE
      DO K=2,m1
         HQ4(K) = 0.25_f4*HQ(K-1)
      ENDDO
   ENDIF

   DO J=ja,jz
      DO I=ia,iz
         DO K=2,m1
            VC1DA(K) = HQ4(K)*(VC3DA(K,I,J)+VC3DA(K-1,I,J)  &
                 +VC3DA(K,I,J+jd)+VC3DA(K-1,I,J+jd))
         ENDDO
         DO K=2,m1-1
            VC3DB(K,I,J) = VC3DB(K,I,J)  &
                 +FQ(I,J)*DZ(K)*(VC1DA(K+1)-VC1DA(K))
         ENDDO
         VC3DB(1,I,J) = VC3DB(2,I,J)
         IF(LEV == 'W') then
            VC3DB(m1-1,I,J) = VC3DB(m1-2,I,J)
         elseif(LEV == 'T') then
            VC3DB(m1,I,J) = VC3DB(m1-1,I,J)
         endif
      ENDDO
   ENDDO
ENDIF

end subroutine gradyt


!```````````````````````````````````````````````````````````````````
subroutine gradzw(m1,m2,m3,ia,iz,ja,jz,VC3DA,VC3DB,RTGC,DZ)

!~ 19.11.2008 (last modified; TIM)

!+ Compute a gradient

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz
real(f4), dimension(m1,m2,m3), intent(IN) :: VC3DA
real(f4), dimension(m2,m3), intent(IN) :: RTGC
real(f4), dimension(m1), intent(IN) :: DZ

real(f4), dimension(m1,m2,m3), intent(INOUT) :: VC3DB

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

DO J=ja,jz
   DO I=ia,iz
      DO K=2,m1
         VC3DB(K,I,J) = (VC3DA(K,I,J)-VC3DA(K-1,I,J))*DZ(K)/RTGC(I,J)
      ENDDO
   ENDDO
ENDDO

end subroutine gradzw


!```````````````````````````````````````````````````````````````````
subroutine gradzt(m1,m2,m3,ia,iz,ja,jz,VC3DA,VC3DB,RTGC,DZ)

!~ 19.11.2008 (last modified; TIM)

!+ Compute a gradient

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz
real(f4), dimension(m1,m2,m3), intent(IN) :: VC3DA
real(f4), dimension(m2,m3), intent(IN) :: RTGC
real(f4), dimension(m1), intent(IN) :: DZ

real(f4), dimension(m1,m2,m3), intent(INOUT) :: VC3DB

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

DO J=ja,jz
   DO I=ia,iz
      DO K=1,m1-1
         VC3DB(K,I,J) = (VC3DA(K+1,I,J)-VC3DA(K,I,J))*DZ(K)/RTGC(I,J)
      ENDDO
   ENDDO
ENDDO

end subroutine gradzt


end module PRAMS_graddiv_mod


!```````````````````````````````````````````````````````````````````
subroutine gradq(m1,m2,m3,ia,iz,ja,jz,array_A,array_B,direction,gdpt_type  &
     ,optype)

!+ Compute a gradient

!~ 2013-05-17 (last modified; TIM)

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: a, jdim, vctr1, vctr2, c_gd
use vtables_mod, only: irtgu,irtgt,idxt,ifmapui,ifmapt,if13t,irtgv,irtgm,idxm  &
     ,ifmapvi,ifmapm,if13m,idxu,ifmapti,ifmapu,if13u,idxv,ifmapmi,ifmapv  &
     ,if13v,idym,if23m,idyt,if23t,idyv,if23v,idyu,if23u
use PRAMS_graddiv_mod, only: gradxu, gradxt, gradyt, gradyv, gradzw, gradzt

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz
real(f4), dimension(m1,m2,m3), intent(IN) :: array_A
character(len=*), intent(IN) :: direction,gdpt_type
character(len=6), intent(IN) :: optype

real(f4), dimension(m1,m2,m3), intent(INOUT) :: array_B

!== Local declarations:

integer :: jaa, jzz, i, j

!=== Executable statements:

if(jdim == 0) then
   jaa = 1
   jzz = 1
else
   jaa = ja
   jzz = jz
endif

IF(direction == 'XDIR') THEN
   IF(gdpt_type == 'UPNT') THEN
      CALL GRADXU(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGU),A(IRTGT),A(IDXT),c_gd%DZT,A(IFMAPUI),A(IFMAPT),A(IF13T)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'VPNT') THEN
      CALL GRADXT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGV),A(IRTGM),A(IDXM),c_gd%DZT,A(IFMAPVI),A(IFMAPM),A(IF13M)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'WPNT') THEN
      CALL GRADXT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGT),A(IRTGU),A(IDXU),c_gd%DZM,A(IFMAPTI),A(IFMAPU),A(IF13U)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ELSEIF(gdpt_type == 'TPNT') THEN
      CALL GRADXT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGT),A(IRTGU),A(IDXU),c_gd%DZT,A(IFMAPTI),A(IFMAPU),A(IF13U)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'NPNT') THEN
      CALL GRADXT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGV),A(IRTGM),A(IDXM),c_gd%DZM,A(IFMAPVI),A(IFMAPM),A(IF13M)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ELSEIF(gdpt_type == 'OPNT') THEN
      CALL GRADXU(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGU),A(IRTGT),A(IDXT),c_gd%DZM,A(IFMAPUI),A(IFMAPT),A(IF13T)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ELSEIF(gdpt_type == 'PPNT') THEN
      CALL GRADXU(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGM),A(IRTGV),A(IDXV),c_gd%DZT,A(IFMAPMI),A(IFMAPV),A(IF13V)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'MPNT') THEN
      CALL GRADXU(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGM),A(IRTGV),A(IDXV),c_gd%DZM,A(IFMAPMI),A(IFMAPV),A(IF13V)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ENDIF
ELSEIF(direction == 'YDIR') THEN
   IF(gdpt_type == 'UPNT') THEN
      CALL GRADYT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGU),A(IRTGM),A(IDYM),c_gd%DZT,A(IFMAPUI),A(IFMAPM),A(IF23M)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'VPNT') THEN
      CALL GRADYV(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGV),A(IRTGT),A(IDYT),c_gd%DZT,A(IFMAPVI),A(IFMAPT),A(IF23T)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'WPNT') THEN
      CALL GRADYT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGT),A(IRTGV),A(IDYV),c_gd%DZM,A(IFMAPTI),A(IFMAPV),A(IF23V)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ELSEIF(gdpt_type == 'TPNT') THEN
      CALL GRADYT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGT),A(IRTGV),A(IDYV),c_gd%DZT,A(IFMAPTI),A(IFMAPV),A(IF23V)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'NPNT') THEN
      CALL GRADYV(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGV),A(IRTGT),A(IDYT),c_gd%DZM,A(IFMAPVI),A(IFMAPT),A(IF23T)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ELSEIF(gdpt_type == 'OPNT') THEN
      CALL GRADYT(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGU),A(IRTGM),A(IDYM),c_gd%DZM,A(IFMAPUI),A(IFMAPM),A(IF23M)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ELSEIF(gdpt_type == 'PPNT') THEN
      CALL GRADYV(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGM),A(IRTGU),A(IDYU),c_gd%DZT,A(IFMAPMI),A(IFMAPU),A(IF23U)  &
           ,c_gd%HW,VCTR2,'T',JDIM)
   ELSEIF(gdpt_type == 'MPNT') THEN
      CALL GRADYV(m1,m2,m3,ia,iz,jaa,jzz,optype,array_A,array_B,VCTR1  &
           ,A(IRTGM),A(IRTGU),A(IDYU),c_gd%DZM,A(IFMAPMI),A(IFMAPU),A(IF23U)  &
           ,c_gd%HT,VCTR2,'W',JDIM)
   ENDIF
ELSEIF(direction == 'ZDIR') THEN
   IF(gdpt_type == 'UPNT') THEN
      CALL GRADZT(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGU),c_gd%DZM)
   ELSEIF(gdpt_type == 'VPNT') THEN
      CALL GRADZT(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGV),c_gd%DZM)
   ELSEIF(gdpt_type == 'WPNT') THEN
      CALL GRADZW(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGT),c_gd%DZT)
   ELSEIF(gdpt_type == 'TPNT') THEN
      CALL GRADZT(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGT),c_gd%DZM)
   ELSEIF(gdpt_type == 'NPNT') THEN
      CALL GRADZW(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGV),c_gd%DZT)
   ELSEIF(gdpt_type == 'OPNT') THEN
      CALL GRADZW(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGU),c_gd%DZT)
   ELSEIF(gdpt_type == 'PPNT') THEN
      CALL GRADZT(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGM),c_gd%DZM)
   ELSEIF(gdpt_type == 'MPNT') THEN
      CALL GRADZW(m1,m2,m3,ia,iz,jaa,jzz,array_A,array_B,A(IRTGM),c_gd%DZT)
   ENDIF
ENDIF

! Impose lateral boundary conditions on the gradient:

if (ja > 1) then
   do j=1,ja-jdim
      do i=ia,iz
         array_B(:,i,j) = array_B(:,i,ja)
      end do
   end do
end if
if (jz < m3) then
   do j=jz+jdim,m3
      do i=ia,iz
         array_B(:,i,j) = array_B(:,i,jz)
      end do
   end do
end if
if (ia > 1) then
   do j=1,m3
      do i=1,ia
         array_B(:,i,j) = array_B(:,ia,j)
      end do
   end do
end if
if (iz < m2) then
   do j=1,m3
      do i=iz+1,m2
         array_B(:,i,j) = array_B(:,iz,j)
      end do
   end do
end if

end subroutine gradq
