!```````````````````````````````````````````````````````````````````
module PRAMS_init_mem_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine dims_calc

!+ Calculates/sets a variety of relevant model variable dimensions 

use main_memory_mod, only: n_grids, nzg, nzs, npatch, jdim, gd

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: igd

!=== Executable statements:

if (.NOT.gd(1)%grid_is_2D) then
   jdim = 1   ! If coarsest grid is 3D
else
   jdim = 0   ! If coarsest grid is 2D
end if

do igd=1,n_grids

   gd(igd)%nx = gd(igd)%nxp-1   !
   gd(igd)%nx1 = gd(igd)%nxp-2  ! x-direction 
   gd(igd)%nx2 = gd(igd)%nxp-3  !

   if (jdim == 1) then
      gd(igd)%ny = gd(igd)%nyp-1   !
      gd(igd)%ny1 = gd(igd)%nyp-2  ! y-direction 
      gd(igd)%ny2 = gd(igd)%nyp-3  !
   else
      gd(igd)%ny = 1           !
      gd(igd)%ny1 = 1          ! y-direction 
      gd(igd)%ny2 = 1          !
   end if

   gd(igd)%nz = gd(igd)%nzp-1   ! z-direction 
   gd(igd)%nz1 = gd(igd)%nzp-2  ! 

end do

do igd=1,n_grids
   gd(igd)%nxyzp = gd(igd)%nxp*gd(igd)%nyp*gd(igd)%nzp
   gd(igd)%nxyp = gd(igd)%nxp*gd(igd)%nyp
   gd(igd)%nxys3p = (nzg+nzs)*gd(igd)%nxp*gd(igd)%nyp*npatch
   gd(igd)%nxys2p = gd(igd)%nxp*gd(igd)%nyp*npatch
   gd(igd)%nxyzorsp = gd(igd)%nxyp*max((nzg+nzs), gd(igd)%nzp)
end do

end subroutine dims_calc


!```````````````````````````````````````````````````````````````````
subroutine refs3D(n1, n2, n3, igd, pi0, dn0, dn0u, dn0v, th0, topt, rtgt,  &
     iq_ref)

!+ Initializes the 3D reference state arrays from the 1D reference state

use univ_kind_defs_mod, only: f4
use univ_interpolation_mod, only: intrp_1D_arr_type01a
use main_memory_mod, only: vctr2, vctr1, topref, gd, rtgref, igdref
use body_specific_constants_mod, only: cvor, cp, p00, R_air, g_Xoid

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3, igd, iq_ref
real(f4), dimension(n2,n3), intent(IN) :: topt, rtgt

real(f4), dimension(n1,n2,n3), intent(INOUT) :: pi0, dn0, dn0u, dn0v, th0

!== Local declarations:

integer :: i, j, k
real(f4) :: c1, c2

!=== Executable statements:

do j=1,n3
   do i=1,n2

      ! Calculate interpolation heights:

      do k=1,n1
         vctr1(k) = gd(igdref(iq_ref))%zt(k)*rtgref(iq_ref) ! 1D refstate z_AGL
         vctr2(k) = gd(igd)%zt(k)*rtgt(i,j)+  &   ! 1D refstate height (AGL) to
              (topt(i,j)-topref(iq_ref))     !  interpolate 3D column from
      end do

      ! Interpolate PI0 and TH0 from 1D refstate:

      pi0(1:n1,i,j) = intrp_1D_arr_type01a(gd(igd)%pi01d(1:n1), vctr1(1:n1),  &
           vctr2(1:n1))
!      call htint(n1,pi01dn(1,igd),vctr1,n1,pi0(1,i,j),vctr2)
      th0(1:n1,i,j) = intrp_1D_arr_type01a(gd(igd)%th01d(1:n1), vctr1(1:n1),  &
           vctr2(1:n1))
!      call htint(n1,th01dn(1,igd),vctr1,n1,th0(1,i,j),vctr2)

      ! Hydrostatic integration to adjust PI0 values:

      c1 = g_Xoid*2._f4*rtgt(i,j)

      do k=gd(igd)%nz,1,-1
         pi0(k,i,j) = pi0(k+1,i,j)+c1/((th0(k,i,j)+th0(k+1,i,j))*gd(igd)%dzm(k))
      end do

      ! Calculate 3D refstate density from PI0 and TH0:

      c2 = p00/(R_air*(cp**cvor))
      do k=1,n1
         dn0(k,i,j) = c2*(pi0(k,i,j)**cvor)/th0(k,i,j)
      end do

   end do
end do

call fill_dn0uv(n1, n2, n3, dn0, dn0u, dn0v)

end subroutine refs3D


!```````````````````````````````````````````````````````````````````
subroutine fill_dn0uv(n1,n2,n3,dn0,dn0u,dn0v)

!+ Fills the 3D u,v density reference state arrays from the 1D reference state

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1,n2,n3
real(f4), dimension(n1,n2,n3), intent(IN) :: dn0

real(f4), dimension(n1,n2,n3), intent(INOUT) :: dn0u,dn0v

!== Local declarations:

integer :: i,j,k,j1,i1

!=== Executable statements:

do j=1,n3
   j1 = min(j+1,n3)
   do i=1,n2
      i1 = min(i+1,n2)
      do k=1,n1
         dn0u(k,i,j) = 0.5_f4*(dn0(k,i,j)+dn0(k,i1,j))
         dn0v(k,i,j) = 0.5_f4*(dn0(k,i,j)+dn0(k,i,j1))
      enddo
   enddo
enddo

end subroutine fill_dn0uv


!```````````````````````````````````````````````````````````````````
subroutine sndref_print(this_is_a_restart)

!+ Displays 1D reference state AND/OR initial sounding

use main_memory_mod, only: initialization_type, vctr1, vctr2, iref, jref,  &
     topref, gd, igdref
use main_memory_bspec_mod, only: nsndp, hsnd, psnd, tsnd, usnd, vsnd, rtsnd,  &
     thdsnd, gd_bd
use body_specific_constants_mod, only: p00, cpor, cp

implicit none

!=== Argument declarations:

logical, intent(IN) :: this_is_a_restart

!== Local declarations:

integer :: k, igd, iq_ref

!=== Executable statements:

if (initialization_type == 1 .and. .NOT.this_is_a_restart) then   ! Display input sounding

   write (*, '(/,/,25x,a)') 'MODEL INPUT SOUNDING'
   write (*, '(/,a,/)')  &
        '    Z        U        V         P       T       RT       TH'
   do k=1,nsndp
      write (*, '(f8.1,2x,2(f7.2,2x),f9.2,2x,f6.2,2x,e8.2,2x,f8.2)')  &
           hsnd(k), usnd(k), vsnd(k), psnd(k), tsnd(k), rtsnd(k), thdsnd(k)
   end do
   
end if

iq_ref = 1
igd = igdref(iq_ref)

do k=1,gd(igd)%nzp
   vctr1(k) = p00*(gd(igd)%pi01d(k)/cp)**cpor    ! Pressure
   vctr2(k) = gd(igd)%th01d(k)*(gd(igd)%pi01d(k)/cp)    ! Temperature
end do

write (*, '(/,/,20x,a,i2,a)') 'MODEL 1D REFERENCE STATE (grid #', igd, ')'
write (*, '(8x,a,i4,a,i4,a,f8.2,a)') '[ at gridpoint (', iref(iq_ref),  &
     ',', jref(iq_ref), ') and sfc elevation of ', topref(iq_ref), ' m ]'
write (*, '(/,a,/)')  &
     '    Z        U        V         P       T       RT       TH       PI'
do k=1,gd(igd)%nzp
   write (*, '(f8.1,2x,2(f7.2,2x),f9.2,2x,f6.2,2x,e8.2,2x,f8.2,2x,f7.2)')  &
        gd(igd)%zt(k), gd(igd)%u01d(k), gd(igd)%v01d(k), vctr1(k),  &
        vctr2(k), gd_bd(igd)%rt01d(k), gd(igd)%th01d(k), gd(igd)%pi01d(k)
end do

end subroutine sndref_print


end module PRAMS_init_mem_mod
