!```````````````````````````````````````````````````````````````````
subroutine toptinit_user(n2, n3, igd, topt, fmapt, xt, yt)

!+ Where user may customize topography values

use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_OPS_proj_mod, only: ll_xy_OPS
use univ_interpolation_mod, only: intrp_1D_arr_type01a
use univ_num_search_mod, only: locidx_1D_mono_type01
use univ_exception_mod, only: exception, FATAL
use main_memory_mod, only: gd, initialization_type, run_toptinit_user,  &
     polelat, polelon
use body_specific_constants_mod, only: radius


implicit none

!=== Argument declarations:

integer, intent(IN) :: n2, n3, igd
real(f4), dimension(n2), intent(IN) :: xt
real(f4), dimension(n3), intent(IN) :: yt

real(f4), dimension(n2,n3), intent(INOUT) :: topt, fmapt

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'toptinit_user'

!  This subroutine is the intended location for a user to customize TOPT,
!  the surface topography array.  It is called after all other types of
!  initialization of this field, so this subroutine has the last word.
!  By default the subroutine makes no change to the field.  The commented
!  lines below serve as a template for user-designed changes.
!
!  *NOTE* that this routine is called for each grid separately, so attention to
!  the current value of 'igd' in this routine may be required by the user:

   ! For Example D, below:
integer, parameter :: n_ic_pts = 53
real(f4), dimension(n_ic_pts) :: r_ic, t_ic

real(f4), dimension(1) :: tmp_f_1D

real(f4) :: half_width, height, halfw2, r
integer :: i, j

real(f4) :: diameter, depth, height_ag, c1, c2, c3, xbasedz, ybasedz, tgt_cx,  &
     tgt_cy
integer :: cent_i, cent_j

!=== Executable statements:

if (.not.run_toptinit_user) return   ! Not desired, currently

!=== Example D -- idealized larger crater (d>10 km?), based on Moreux crater
!                 (d~138km) and Gale crater (d~154km) on Mars (see
!                 Borlina et al. 2015 in JGR), *will* work with topographic
!                 shadowing:
!
!if (igd == 1) then
!      ! Set feature center (provide lat [deg_N] and lon [deg_E] coords, obtain
!      !  OPS x,y coords).  This is necessary if topographic
!      !  slope/aspect/shadowing effects are desired, since a separate
!      !  asymmetrically-augmented (extra points added to each side) topography
!      !  grid is used for those calculations:
!   call ll_xy_OPS(-5.40001_f4, 137.80001_f4, polelat, polelon, tgt_cx,  &
!        tgt_cy, radius)     ! lat/lon to OPS coords
!
!      ! Determine a grid coordinate near the feature center:
!   cent_i = locidx_1D_mono_type01(xt, tgt_cx, (/1,n2/))
!   cent_j = locidx_1D_mono_type01(yt, tgt_cy, (/1,n3/))
!   if (cent_i > n2 .or. cent_j > n3) call exception(procedure_name, FATAL,  &
!        'Feature center not found on this grid.')
!
!      ! Idealized crater profile radius (from feature center) [m]:
!   r_ic = (/ 0.0000_f4, 12643.6797_f4, 14080.4609_f4, 16091.9531_f4,  &
!        18390.8047_f4, 20977.0078_f4, 22988.5039_f4, 25000.0000_f4,  &
!        26724.1367_f4, 30154.3516_f4, 32440.0625_f4, 33582.9219_f4,  &
!        34725.7773_f4, 35868.6367_f4, 37011.4922_f4, 40154.3477_f4,  &
!        41582.9219_f4, 43011.4922_f4, 44440.0625_f4, 46154.3516_f4,  &
!        47868.6328_f4, 49297.2070_f4, 50440.0625_f4, 51582.9219_f4,  &
!        52725.7773_f4, 53868.6328_f4, 54440.0625_f4, 55582.9219_f4,  &
!        56725.7812_f4, 57868.6328_f4, 59011.4922_f4, 60440.0625_f4,  &
!        61297.2070_f4, 62440.0625_f4, 63297.2070_f4, 64154.3477_f4,  &
!        65297.2070_f4, 66154.3516_f4, 66725.7812_f4, 67297.2031_f4,  &
!        67868.6328_f4, 68440.0625_f4, 69011.4922_f4, 70154.3516_f4,  &
!        71011.4922_f4, 72154.3516_f4, 72725.7812_f4, 73582.9219_f4,  &
!        74440.0625_f4, 75011.4922_f4, 75297.2031_f4, 75868.6328_f4,  &
!        77011.4922_f4 /)
!
!      ! Idealized crater profile *relative* topography [m]:
!    t_ic = (/ -4774.6479_f4, -4774.6479_f4, -4774.6479_f4, -4774.6479_f4,  &
!         -4774.6479_f4, -4774.6479_f4, -4774.6479_f4, -4774.6479_f4,  &
!         -4774.6479_f4, -4774.6479_f4, -4774.6479_f4, -4774.6479_f4,  &
!         -4760.5635_f4, -4732.3945_f4, -4690.1406_f4, -4690.1406_f4,  &
!         -4661.9717_f4, -4619.7183_f4, -4549.2959_f4, -4464.7886_f4,  &
!         -4338.0283_f4, -4197.1831_f4, -4056.3381_f4, -3929.5774_f4,  &
!         -3746.4790_f4, -3535.2112_f4, -3309.8594_f4, -3000.0000_f4,  &
!         -2680.5557_f4, -2319.4446_f4, -2041.6666_f4, -1816.9014_f4,  &
!         -1647.8873_f4, -1436.6196_f4, -1295.7747_f4, -1112.6760_f4,  &
!         -971.8310_f4, -830.9859_f4, -676.0563_f4, -535.2112_f4,  &
!         -309.8592_f4, -183.0986_f4, -56.3380_f4, 42.2535_f4, 140.8451_f4,  &
!         197.1831_f4, 169.0141_f4, 154.9296_f4, 140.8451_f4, 140.8451_f4,  &
!         84.5070_f4, 0.0000_f4, 0.0000_f4 /)
!
!   c2 = 77000._f4  ! Desired feature (crater) radius [m]
!
!   c1 = c2/r_ic(n_ic_pts)  ! Scaling factor
!   r_ic = r_ic*c1; t_ic = t_ic*c1   ! Rescale idealized crater profile
!
!!@   topt = 0._f4  ! Default topography value [m]
!   c3 = topt(cent_i,cent_j)  ! Pre-existing "centroid" topography value
!
!   do j=1,n3
!      do i=1,n2
!            ! Radial distance from the feature's center [m], adjusting by an
!            !  approximate map project factor:
!         r = sqrt((xt(i)-tgt_cx)**2+(yt(j)-tgt_cy)**2)/fmapt(cent_i,cent_j)
!
!         if (r < r_ic(n_ic_pts)) then
!          ! Add topographic feature (relative to c3):
!            tmp_f_1D = intrp_1D_arr_type01a(t_ic, r_ic, (/r/))
!            topt(i,j) = c3+tmp_f_1D(1)
!         else if (r < 1.75_f4*r_ic(n_ic_pts)) then
!          ! Blend topographic feature into surrounding topography:
!            c1 = (r-r_ic(n_ic_pts))/(1.75_f4*r_ic(n_ic_pts)-r_ic(n_ic_pts))
!            topt(i,j) = c3*(1._f4-c1)+topt(i,j)*c1
!         end if
!      end do
!   end do
!end if

!=== Example C -- symmetric linear crater, *will* work with topographic
!                 shadowing:
!
!if (igd == 1) then
!      ! Set feature center (provide lat [deg_N] and lon [deg_E] coords, obtain
!      !  OPS x,y coords).  This is necessary if topographic
!      !  slope/aspect/shadowing effects are desired, since a separate
!      !  asymmetrically-augmented (extra points added to each side) topography
!      !  grid is used for those calculations:
!   call ll_xy_OPS(-7.00001_f4, -72.50001_f4, polelat, polelon, tgt_cx,  &
!        tgt_cy, radius)     ! lat/lon to OPS coords
!
!   c1 = -2300._f4  ! Feature's nominal minimum topography value [m]
!   c2 = 60000._f4    ! The radius (from feature's center) at which topo=0. [m]
!   do j=1,n3
!      do i=1,n2
!            ! xbasedz is radius:
!         xbasedz = sqrt((xt(i)-tgt_cx)**2+(yt(j)-tgt_cy)**2)
!
!            ! ybasedz is topo [m]; b+(m*xbasedz):
!         ybasedz = c1-c1/c2*xbasedz
!
!            ! Clip topo values to desired range:
!         ybasedz = min(ybasedz, 0._f4)
!         topt(i,j) = max(ybasedz, -1000._f4)
!      end do
!   end do
!end if

!=== Example B -- Parabolic crater, will not work with topographic shadowing:
!
!if (initialization_type == 1 .and. igd == 1) then
!
!   diameter = 3500.  ! (measured from the top of one rim to the other) [m]    
!   depth = 0.12*diameter**0.96 ! Total depth (top of rim to floor bottom); [m]
!                               !  functional form from Matias_etal (LPSC 2000)
!   height_ag = depth/4.  ! Height (above surrounding plain) of top of rim [m]
!   cent_i = n2/2      ! Center grid point coords
!   cent_j = max(n3/2, 1)   !
!
!   tgt_cx = xt(cent_i)
!   tgt_cy = yt(cent_j)
!   c1 = 4.*depth/(diameter*diameter)
!   c2 = height_ag-depth
!   c3 = diameter*0.5-sqrt(-c2/c1)
!
!   do j=1,n3
!      do i=1,n2
!         r = sqrt((xt(i)-tgt_cx)**2+(yt(j)-tgt_cy)**2)
!         if (abs(r) <= diameter*0.5) then
!            topt(i,j) = c1*r*r+c2
!         else if (abs(r) <= diameter*0.5+c3) then
!            r = diameter-r
!            topt(i,j) = c1*r*r+c2
!         else
!            topt(i,j) = 0.
!         end if
!      end do
!   end do
!end if

!=== Example A -- a "witch of Agnesi" mountain, will not work with topographic
!                 shadowing:
!
!if (igd == 1) then
!   if (initialization_type == 1) then
!      half_width = 15000.  ! [m]
!      height = 1000.       !
!      halfw2 = half_width*half_width
!
!      do j=1,n3
!         do i=1,n2
!            r = sqrt(xt(i)**2+yt(j)**2)
!            topt(i,j) = height*halfw2/(halfw2+r**2) !xt(i)**2)
!         end do
!      end do
!
!else if (igd == 2) then
!
!   Whatever you want here
!
!end if

!*** UNCOMMENT the below statement to zero out topography on all grids:
!topt(1:n2,1:n3) = 0.

end subroutine toptinit_user


!```````````````````````````````````````````````````````````````````
subroutine sfcinit_user(n1, n2, n3, n4, n5, n6, igd, vnam, glat, glon,  &
     TI_static, albd_static)

!+ Where user may customize SBPS characteristics' values

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: a, initialization_type, TI_val, run_sfcinit_user

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3, n4, n5, n6, igd
character(len=*), intent(IN) :: vnam
real(f4), dimension(n2,n3), intent(IN) :: glat, glon

real(f4), dimension(n2,n3,n5), intent(INOUT) :: TI_static, albd_static

!== Local declarations:

!  This subroutine is the intended location for a user to customize any
!  sub/super surface (SBPS) characteristic array.  It is called after all
!  other types of initialization of this field, so this subroutine has the
!  last word.  By default the subroutine makes no change to the field.  The 
!  commented lines below serve as a template for user-designed changes;
!  Note that this routine is called for each grid separately, so attention to
!  the current value of NG in this routine may be required by the user:
!
!------------------------------------------------------------------------
!  *** IMPORTANT ***: Do not alter SBPSTMCD here, as it will not be used
!     properly; instead, alter TISTAT appropriately.
!------------------------------------------------------------------------

integer :: i, j, k, ip

!=== Executable statements:

if (.not.run_sfcinit_user) return   ! Not desired, currently

if (vnam == 'FILEDATA') then

   if (initialization_type == 2) then

!      do ip=1,n5  ! Loop over all patches

!         do j=1,n3
!            do i=1,n2

!               if(glat(i,j) < -79.9499) then   ! South polar TI kludge
!                  TI_static(i,j,ip) = 100. 
!               elseif(glat(i,j) > 79.9499) then   ! North polar TI kludge
!                  TI_static(i,j,ip) = TI_val
!               endif

!            enddo
      
!         enddo
!      enddo

   end if
 
else if (vnam == 'NOFILE') then

   if (igd == 1) then

      !***** WHATEVER ******

   end if

end if

end subroutine sfcinit_user


!```````````````````````````````````````````````````````````````````
subroutine LES_iperturb(m1,m2,m3,thp,lev)

!+ Slightly randomly perturb thermal LES init field (so not so homogeneous)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,lev

real(f4), dimension(m1,m2,m3), intent(INOUT) :: thp

!== Local declarations:

real(f4), dimension(m2,m3) :: harvest

!=== Executable statements:

call random_number(harvest)   ! Get a horizontal grid of numbers between 0 & 1 
thp(lev,:,:) = thp(lev,:,:)+(harvest(:,:)/5._f4-0.1_f4) ! Add random perturbation 
                                          ! to desired theta level (+/- 0.1 K)

end subroutine LES_iperturb


!```````````````````````````````````````````````````````````````````
subroutine tracer_init(n1,n2,n3,ng,tID,tracer,topo,rtgt,zt,this_is_a_restart)

!+ Initialize some tracer quantities

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1,n2,n3,ng,tID
real(f4), dimension(n2,n3), intent(IN) :: topo,rtgt
real(f4), dimension(n1), intent(IN) :: zt
logical, intent(IN) :: this_is_a_restart

real(f4), dimension(n1,n2,n3), intent(INOUT) :: tracer  ! Is actually: a(isclpn(X,ng))

!== Local declarations:

integer :: i,j,ie,ib,je,jb,k
real(f4) :: az,z_agl

!=== Executable statements:

if(this_is_a_restart) then

!@   tracer = 0.    ! Make sure everything else is zeroed out

!@   if(ng == 3) then    ! Only want to do this on grid 3
!      jb = 50   ! frames Olympus Mons (and some surrounding plain)
!      je = 82   !
!      ib = 10   !
!      ie = 42   !
!@      select case(tID)      ! Operating on which tracer?

!@      case(1)

!@         jb = 50   ! western half of Olympus Mons (and some surrounding plain)
!@         je = 82   !
!@         ib = 10   !
!@         ie = 28   !
!@         do j=jb,je
!@            do i=ib,ie  ! If topo is < 1500m, tracer at or below 200m AGL
!@               if(topo(i,j) < 1500.) then
!@                  do k=1,n1
!@                     z_agl = zt(k)*rtgt(i,j)
!@                     if(z_agl <= 200.) then
!@                        tracer(k,i,j) = 1.
!@                     endif
!@                  enddo
!@               endif
!@            enddo
!@         enddo

!@      case(2)

!@         jb = 50   ! western half of Olympus Mons (and some surrounding plain)
!@         je = 82   !
!@         ib = 10   !
!@         ie = 28   !
!@         do j=jb,je
!@            do i=ib,ie  ! tracer if az <= 12km and z_AGL > 200m
!@               do k=1,n1
!@                  z_agl = zt(k)*rtgt(i,j)
!@                  if(z_agl > 200. .and. z_agl+topo(i,j) <= 12000.) then
!@                     tracer(k,i,j) = 1.
!@                  endif
!@               enddo
!@            enddo
!@         enddo

!@      case(3)

!@         jb = 50   ! western half of Olympus Mons (and some surrounding plain)
!@         je = 82   !
!@         ib = 10   !
!@         ie = 28   !
!@         do j=jb,je
!@            do i=ib,ie  ! tracer if 12km<az<=20km
!@               do k=1,n1
!@                  az = zt(k)*rtgt(i,j)+topo(i,j)
!@                  if(az > 12000. .and. az <= 20000.) then
!@                     tracer(k,i,j) = 1.
!@                  endif
!@               enddo
!@            enddo
!@         enddo

!@      case(4)

!@         jb = 50   ! eastern half of Olympus Mons (and some surrounding plain)
!@         je = 82   !
!@         ib = 29   !
!@         ie = 42   !
!@         do j=jb,je
!@            do i=ib,ie  ! If topo is < 1500m, tracer at or below 200m AGL
!@               if(topo(i,j) < 1500.) then
!@                  do k=1,n1
!@                     z_agl = zt(k)*rtgt(i,j)
!@                     if(z_agl <= 200.) then
!@                        tracer(k,i,j) = 1.
!@                     endif
!@                  enddo
!@               endif
!@            enddo
!@         enddo

!@      case(5)


!@         jb = 50   ! eastern half of Olympus Mons (and some surrounding plain)
!@         je = 82   !
!@         ib = 29   !
!@         ie = 42   !
!@         do j=jb,je
!@            do i=ib,ie  ! tracer if az <= 12km and z_AGL > 200m
!@               do k=1,n1
!@                  z_agl = zt(k)*rtgt(i,j)
!@                  if(z_agl > 200. .and. z_agl+topo(i,j) <= 12000.) then
!@                     tracer(k,i,j) = 1.
!@                  endif
!@               enddo
!@            enddo
!@         enddo

!@      case(6)

!@         jb = 50   ! eastern half of Olympus Mons (and some surrounding plain)
!@         je = 82   !
!@         ib = 29   !
!@         ie = 42   !
!@         do j=jb,je
!@            do i=ib,ie  ! tracer if 12km<az<=20km
!@               do k=1,n1
!@                  az = zt(k)*rtgt(i,j)+topo(i,j)
!@                  if(az > 12000. .and. az <= 20000.) then
!@                     tracer(k,i,j) = 1.
!@                  endif
!@               enddo
!@            enddo
!@         enddo

!@      case(7)

!@         jb = 74   ! eastern half of Olympus Mons (and some surrounding plain)
!@         je = 84   !
!@         ib = 50   !
!@         ie = 60   !
!@         do j=jb,je
!@            do i=ib,ie  ! Tracer at or below 200m AGL
!@               do k=1,n1
!@                  z_agl = zt(k)*rtgt(i,j)
!@                  if(z_agl <= 200.) then
!@                     tracer(k,i,j) = 1.
!@                  endif
!@               enddo
!@            enddo
!@         enddo
!@
!@      end select

!@   endif

else    ! Just initialize to zero

   do j=1,n3
      do i=1,n2
         do k=1,n1
            tracer(k,i,j) = ZERO_f4
         enddo
      enddo
   enddo

   select case(tID)      ! Operating on which tracer?
      
   case(1)

      do j=1,n3
         do i=1,n2
            do k=1,n1
               tracer(k,i,j) = 0.9532_f4    ! CO2 gas fraction
            enddo
         enddo
      enddo
      
   end select

endif

end subroutine tracer_init
