!```````````````````````````````````````````````````````````````````
subroutine RT_driver(mzp, mxp, myp, ia, iz, ja, jz)

!+ Top-level radiative transfer driver

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ZERO_f4
use univ_integer_time_mod, only: time_obj, fp_seconds_eqv_Time
use univ_namelist_utils_mod, only: Time_to_gen_time_strrep
use main_memory_mod, only: a, i_grid, radfrq, npatch, c_gd, gd,  &
     current_sim_ET, module_tps, RT_paradigm, mnsdy_Time, morbpd_Time
use main_memory_bspec_mod, only: RTc_tog 
use vtables_mod, only: itht, ifthrd, ipi0, ipp, irshort, irlong, irlongup
use RT_utils_mod, only: dT2dtheta, ae3p3_TIM

implicit none

!=== Argument declarations:

integer, intent(IN) :: mzp,mxp,myp,ia,iz,ja,jz

!== Local declarations:

integer :: ic, mxyp_1
type(time_obj) :: tmp_Time
character(len=40) :: gen_time_strrep

!=== Executable statements:

! If this is grid #1, check if RT for all grids "should" occur closer to the
!  beginning of the current grid #1 timestep or the next one -- otherwise,
!  just use the value RT_call(1) as a guide:
if (i_grid == 1) then
   ic = int((current_sim_ET+real(c_gd%dt_dyn, kind=f8)*0.5_f8)/  &
        real(radfrq, kind=f8))
else
   ic = gd(1)%RT_call
end if

if (ic > c_gd%RT_call) then
   c_gd%RT_call = ic
   mxyp_1 = mxp*myp-1

   ! Run the configured radiative transfer scheme:

   select case (RT_paradigm)
   case (0)       ! No RT calculations
      a(irshort:irshort+mxyp_1) = ZERO_f4
      a(irlong:irlong+mxyp_1) = ZERO_f4
      a(irlongup:irlongup+mxyp_1) = ZERO_f4  ! filled by SBPS model
      a(ifthrd:ifthrd+mxyp_1) = ZERO_f4

   case(1)        ! RT, as if there were no atmosphere at all
      a(irlongup:irlongup+mxyp_1) = ZERO_f4  ! filled by SBPS model
      a(ifthrd:ifthrd+mxyp_1) = ZERO_f4  ! No atmospheric dT/dt due to RT

      call no_atm_RT_interface(mzp, mxp, myp, npatch, ia, iz, ja, jz)

   case(2)        ! Full RT (atmospheric + surface)
      a(irlongup:irlongup+mxyp_1) = ZERO_f4  ! filled by SBPS model

      select case (RTc_tog)
      case (1)  ! Use scheme based on legacy NASA Ames MGCM RT code
         call radn_interface(ia, iz, ja, jz, mzp, mxp, myp)

      case (2,3)  ! Use scheme based on CARMA RT code (2-stream)
         call aerad_interface(mzp, mxp, myp, npatch, ia, iz, ja, jz)

      case (4)  ! Use scheme based on NASA ARC RT code (2-stream)
         call ARCnRT_interface(mzp, mxp, myp, npatch, ia, iz, ja, jz)
      end select

      ! Convert temperature tendency (heating rate) to a theta tendency:
      call dT2dtheta(mzp, mxp, myp, ia, iz, ja, jz, a(ipp), a(ipi0),  &
           a(ifthrd))

      tmp_Time = fp_seconds_eqv_Time(current_sim_ET, module_tps)
      call Time_to_gen_time_strrep(tmp_Time, gen_time_strrep,  &
           len(gen_time_strrep), mnsdy_Time, morbpd_Time,  &
           short_unit_style=.TRUE.)

      write (*, '(/,a,i2,3a,/)')  &
           'Radiation tendencies updated on grid # ',  &
           i_grid, ', at time = [', trim(gen_time_strrep), ']'
   end select

end if

! Add radiative heating tendency to THETA tendency:
if (RT_paradigm /= 0) call ae3p3_TIM(mzp, mxp, myp, ia, iz, ja, jz, 1, mzp,  &
     a(itht), a(itht), a(ifthrd))

end subroutine RT_driver
