!```````````````````````````````````````````````````````````````````
module aeX_dust_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine dust_optdepth(m1,m2,m3,mia,miz,mja,mjz,dust_nb,tau,dust_mrat  &
     ,air_density,rtgt,zt)

!+ Calculate dust optical depth (supports multiple radius bins)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,dust_nb,mia,miz,mja,mjz
  ! Passed-in air_density should be in CGS units:
real(f4), dimension(m1,m2,m3), intent(IN) :: air_density,tau
real(f4), dimension(m1), intent(IN) :: zt
real(f4), dimension(m1,m2,m3,dust_nb), intent(IN) :: dust_mrat
real(f4), dimension(m2,m3), intent(IN) :: rtgt

!== Local declarations:

integer, parameter :: dust_group = 1   ! Index of dust aerosol group

!%integer :: i,ib,k,j
!%real(f4) :: c1,dust_mass,tau_bin
!%real(f4), save, dimension(NBIN) :: Qe

!%data Qe/1.0,1.25,1.85,2.0,2.2,2.2/  ! Approximate extinction efficiency for
                                    !   aerosol bins (at solar wavelengths)

!=== Executable statements:

!%tau(1:PRAMS_n1) = 0.  ! Initialize optical depth array

!%do j=mja,mjz
!%   do i=mia,miz

!%do ib=1,NBIN   ! Loop over all dust bins

   ! Constant for optical depth calc below
!%   c1 = Qe(ib)*r(ib,dust_group)*r(ib,dust_group)*PI/rmass(ib,dust_group)

   ! Obtain dust (tracer) optical depth values:

!%   tau_bin = 0.   ! Set optical depth at top of model to zero
!%   do k=PRAMS_n1-1,2,-1

!%      dust_mass = dust_mrat(k,ib)*air_density(k)   ! [g]
!%      tau_bin = max(1.e-20,tau_bin+dust_mass*c1*  &
!%           rtgt*(zt(k+1)-zt(k))*m_to_cm)    ! Optical depth for this z,bin
!%      tau(k) = tau(k)+tau_bin   ! Total optical depth
!%   enddo
!%   tau(1) = tau(2)  ! Set level 1 value to avoid interp problems

!%enddo

end subroutine dust_optdepth


!```````````````````````````````````````````````````````````````````
subroutine dust_massdist(new_mass)!%,bin_mass)

!+ Distribute a given mass of dust into radius bins (lognormal distribution)

use univ_kind_defs_mod, only: f4

implicit none

!%include 'globaer.h90'
!%include 'aerconstants.h90'

!=== Argument declarations:

real(f4), intent(IN) :: new_mass   ! [g]
!%real(f4), dimension(NBIN) :: bin_mass   ! [g]

!== Local declarations:

integer, parameter :: dust_group = 1,  &
                      dist_subbins = 10000

! (rel_contrib_s+rel_contrib_e) must = 1. = 100%
real(f4), parameter :: rel_contrib_s = 0.5_f4,  &  ! Dust contrib due to saltation
                   rel_contrib_e = 0.5_f4,  &  ! Other dust source contribution
                   dist_subbin_width = 1.5e-3_f4   ! [cm] = 15 um  

!%real(f4) :: rsig_s,r0_s,rsig_e,r0_e,c1_s,c2_s,c2_e  &
!%     ,subbin_dr,subbin_rad  &
!%     ,sum_bfrac,cnt_to_mass_factor
!%real(f4) :: c1_e,bin_massf_e,bin_end,bin_massf_s,bin_rad,bin_start
!%integer :: ib,isb
!%real(f4), save :: bin_frac(NBIN)
integer, save :: ncall

data ncall/0/

!=== Executable statements:

if(ncall == 0) then   ! Only do this once (it's expensive)

   ! Modal radius <r0_s> and deviation <rsig_s> of dust from saltation:
!%   rsig_s = 1.2    ! unitless scaling parameter
!%   cnt_to_mass_factor = exp(3.*log(rsig_s)**2)  ! # -> mass distribution
!%   r0_s = 2.5e-6*m_to_cm*cnt_to_mass_factor   ! 2.5 um

   ! Modal radius <r0_e> and deviation <rsig_e> of dust from "else":
!%   rsig_e = 1.6    ! unitless scaling parameter
!%   cnt_to_mass_factor = exp(3.*log(rsig_e)**2)  ! # -> mass distribution
!%   r0_e = 0.8e-6*m_to_cm*cnt_to_mass_factor   ! 0.8 um

!%   subbin_dr = dist_subbin_width/real(dist_subbins)  ! width of a subbin

!%   ib = 1   ! start with first bin
!%   bin_rad = r(ib,dust_group)  ! mean radius (by volume) of current bin
!%   bin_start = rlow(ib,dust_group)  ! lower radius bound of bin
!%   bin_end = rup(ib,dust_group)   ! upper radius bound of bin

!%   distsubstep: do isb=1,dist_subbins
!%      subbin_rad = (real(isb)-0.5)*subbin_dr  ! mean radius of current subbin

!%      if(subbin_rad > bin_end .or. isb == dist_subbins) then  ! Go to next bin
!%         bin_frac(ib) = sum_bfrac   ! Save finished result for current bin
!%         sum_bfrac = 0.  ! Reset sum
!%         ib = ib+1  ! Increment bin index
!%         if(ib > NBIN) then    ! Don't continue past last existing bin
!%            exit distsubstep   !
!%         endif                 !
!%         bin_rad = r(ib,dust_group)  ! mean radius (by volume) of current bin
!%         bin_start = rlow(ib,dust_group)  ! lower radius bound of bin
!%         bin_end = rup(ib,dust_group)   ! upper radius bound of bin
!%      endif

!%      if(subbin_rad >= bin_start .and. subbin_rad <= bin_end) then
!%         c1_s = rel_contrib_s/(sqrt(2.*PI)*subbin_rad*log(rsig_s))
!%         c2_s = -0.5*(log(subbin_rad/r0_s)/log(rsig_s))**2
!%         c1_e = rel_contrib_e/(sqrt(2.*PI)*subbin_rad*log(rsig_e))
!%         c2_e = -0.5*(log(subbin_rad/r0_e)/log(rsig_e))**2
      
         ! Fraction of total particles that lie in current subbin (well,
         ! almost... need to multiply by a factor of dr() first): 
!%         bin_massf_e = c1_e*exp(c2_e)
!%         bin_massf_s = c1_s*exp(c2_s)

         ! Sum to obtain fraction of total particles within radius bin
!%         sum_bfrac = sum_bfrac+(bin_massf_s+bin_massf_e)*subbin_dr
!%      else
!%         continue  ! Keep looping if not currently within a bin
!%      endif
!%   enddo distsubstep

   ncall = 1
endif

! Distribute given mass over bins (done every time routine is called):

!%do ib=1,NBIN
!%   bin_mass(ib) = bin_frac(ib)*new_mass
!%enddo 

end subroutine dust_massdist


end module aeX_dust_mod


!```````````````````````````````````````````````````````````````````
subroutine dust_init(m1, m2, m3, ia, iz, ja, jz, dust_nb, dustnc, tempk,  &
     press, zt, rtgt, i_grid)

!+ Initialize vertical/spatial dust distribution

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use aermic_rad_mem_mod, only: mvar_dust_params
use main_memory_mod, only: gd, map_projection, a
use main_memory_bspec_mod, only: dust_imr, dust_igd, dust_inn, dust_ivr,  &
     lonrad, dust_icn, dust_irt, dust_iht, dust_itt, dust_ctau_refp
use vtables_mod, only: iglatn, iglonn, ivt2dan
use body_specific_constants_mod, only: R_air, g_Xoid
use za_vardust_spec_mod, only: za_vardust_spec
use RT_utils_mod, only: stellar_angles
use aemic_drvint_mod, only: aer_ndist_logn
use aerad_aeropd_mod, only: calc_abp_opd_specific

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,dust_nb,ia,iz,ja,jz, i_grid
real(f4), dimension(m1,m2,m3), intent(IN) :: tempk,press   ! [K], [Pa]
real(f4), dimension(m1), intent(IN) :: zt     ! [m]
real(f4), dimension(m2,m3), intent(IN) :: rtgt

real(f4), dimension(m1,m2,m3,dust_nb), intent(INOUT) :: dustnc

!== Local declarations:

integer, parameter :: tgt_bin = 4    ! For monodisperse dust init

real(f4), parameter :: H = 5000._f4    ! Scaling height [m] for exp profile
                     
integer, parameter :: init_tog = 2   ! 1=monodisperse, 2=polydisperse

integer :: i,j,k,ib,k_norm
!%real(f4), dimension(m1,m2,m3) :: tau
real(f4) :: dust_num,N_0,xsdist,Ls_curr,lyr_press,q_q0,N_raw,ftmp  &
     ,air_density,conrath_nu,column_tau,tau_tgt,N_fact,haze_top  &
     ,p_top,p_bot
real(f4), dimension(dust_nb) :: aer_ndist_bfrac
real(f4), allocatable, dimension(:,:) :: N_scr
real(f4), allocatable, dimension(:) :: dz_scr

real(f4), dimension(m2,m3) :: tmp2D_a, tmp2D_b

!=== Executable statements:

if(dust_ivr == 1) then    ! "Standard" init

   N_0 = dust_inn

   if(init_tog == 1) then
      
      do j=1,m3
         do i=1,m2
            do k=2,m1
               
               dust_num = N_0*exp(-zt(k)*rtgt(i,j)/H)   ! Dust # conc [#/cm^3]
               
               !%         ! Convert to dust mixing ratio [g/g (dust/air)]:
               !%         dustmr(k,i,j,tgt_bin) = dust_num/density(k,i,j)

               dustnc(k,i,j,tgt_bin) = dust_num

            enddo

            dustnc(1,i,j,tgt_bin) = dustnc(2,i,j,tgt_bin)   ! Bottom BC

         enddo
      enddo
   
   elseif(init_tog == 2) then

      call aer_ndist_logn(dust_imr*1.e-6_f4,dust_igd,dust_nb,1,aer_ndist_bfrac)

      do j=1,m3
         do i=1,m2
            do ib=1,dust_nb            ! Initialize all dust bins
               
               do k=2,m1
                  dust_num = N_0*exp(-zt(k)*rtgt(i,j)/H)  ! Total #conc [#/cm^3]
                  dustnc(k,i,j,ib) = aer_ndist_bfrac(ib)*dust_num
               enddo
               
               dustnc(1,i,j,ib) = dustnc(2,i,j,ib)   ! Bottom BC
               
            enddo
            
         enddo
      enddo

   endif

else if (dust_ivr == 4 .or. dust_ivr == 5) then   ! "haze_top" profile type

   if (dust_ivr == 4) then   ! User-specified parameters

      tmp2D_b(:,:) = dust_iht
      tmp2D_a(:,:) = dust_itt
      
   else if (dust_ivr == 5) then     ! "vtau" init

      ! Obtain current Ls (info for grid #1 is sufficient) [deg]:

      call stellar_angles(gd(1)%nzp, gd(1)%nxp, gd(1)%nyp, 1, 1, 1, 1,  &
           a(iglatn(1)), a(iglonn(1)), tmp2D_a, xsdist, gd(1)%centlat,  &
           gd(1)%centlon, lonrad, tmp2D_b, Ls_curr, .FALSE.)
 
      ! Calculate "more variable" radiatively-active dust loading parameters:
      !  (tmp2D_a holds tautot, tmp2D_b holds conrath_nu)

      if (map_projection == 0) then
         a(ivt2dan(i_grid):ivt2dan(i_grid)+m2*m3-1) = gd(1)%centlat
         call za_vardust_spec(m2*m3, Ls_curr, a(ivt2dan(i_grid)), tmp2D_a,  &
              tmp2D_b)
      else
         call za_vardust_spec(m2*m3, Ls_curr, a(iglatn(i_grid)), tmp2D_a,  &
              tmp2D_b)
      end if

   end if

   ! Determine size/mass distribution parameters:

   call aer_ndist_logn(dust_imr*1.e-6_f4,dust_igd,dust_nb,1,aer_ndist_bfrac)

   ! Construct vertical profiles of dust:

   allocate(N_scr(-8:m1,dust_nb),dz_scr(-8:m1))

   k_norm = 0

   do j=1,m3
      do i=1,m2
         
         haze_top = tmp2D_b(i,j)

         do k=2,m1
            lyr_press = press(k,i,j) 

            if (lyr_press < dust_ctau_refp) then
               q_q0 = max(1.e-30_f4,  &
                    exp(0.007_f4*(ONE_f4-(dust_ctau_refp/lyr_press)**  &
                    (70._f4/haze_top))))
            else
               q_q0 = ONE_f4
            end if

            air_density = press(k,i,j)/(R_air*tempk(k,i,j))
            N_raw = q_q0*air_density   ! dust N proportional to mass
            
            do ib=1,dust_nb            ! Initialize all dust bins
               N_scr(k,ib) = aer_ndist_bfrac(ib)*N_raw  
            enddo
         enddo
         N_scr(1,1:dust_nb) = N_scr(2,1:dust_nb)   ! Bottom BC

         dz_scr(1:m1) = rtgt(i,j)/gd(i_grid)%dzt(1:m1)

         ! Attempt to determine layer that includes 'dust_ctau_refp' value:

         k_norm = 0
         if (dust_ctau_refp < press(1,i,j) .and.  & ! Only do if value in column
              dust_ctau_refp > press(m1-1,i,j)) then
            do k=2,m1-1
               p_bot = (press(k,i,j)+press(k-1,i,j))*0.5_f4
               p_top = (press(k,i,j)+press(k+1,i,j))*0.5_f4
               if(p_bot >= dust_ctau_refp .and. p_top < dust_ctau_refp) then
                  k_norm = k
                  ftmp = R_air*tempk(k,i,j)/g_Xoid*log(dust_ctau_refp/p_top)
                  dz_scr(k) = min(dz_scr(k), ftmp)
                  exit
               end if
            end do
         end if

         if(k_norm == 0) then   ! Need to invent some atmosphere
            
            p_bot = dust_ctau_refp
            ftmp = (dust_ctau_refp-(press(2,i,j)+press(1,i,j))*0.5_f4)/10._f4
            k_norm = -8

            do k=-8,1
               p_top = p_bot-ftmp
               lyr_press = 0.5_f4*(p_bot+p_top)
               
               if (lyr_press < dust_ctau_refp) then
                  q_q0 = exp(0.007_f4*  &
                       (ONE_f4-(dust_ctau_refp/lyr_press)**(70._f4/haze_top)))
               else
                  q_q0 = ONE_f4
               end if

               dz_scr(k) = R_air*tempk(2,i,j)/g_Xoid*log(p_bot/p_top)

               air_density = lyr_press/(R_air*tempk(2,i,j))
               N_raw = q_q0*air_density   ! dust N proportional to mass
   
               do ib=1,dust_nb
                  N_scr(k,ib) = aer_ndist_bfrac(ib)*N_raw 
               enddo

               p_bot = p_top
            enddo
           
         endif

         tau_tgt = tmp2D_a(i,j)   ! Target opacity in column above 'k_norm'

         ! Initial guess at value of 'N_fact' resulting in 'tau_tgt' opacity
         !  (at "9um"):
         N_fact = ONE_f4
         call calc_abp_opd_specific(-8, m1, dust_nb, 'dust_9um', k_norm, m1,  &
              N_scr, dz_scr, N_fact, column_tau)

         N_fact = tau_tgt/column_tau   ! Final normalization value

         ! Scale dust number concentrations:

         do ib=1,dust_nb
            dustnc(2:m1,i,j,ib) = N_scr(2:m1,ib)*N_fact   ! [#/cm^3]
         enddo
         dustnc(1,i,j,1:dust_nb) = dustnc(2,i,j,1:dust_nb)

      enddo
   enddo

   deallocate(N_scr,dz_scr)

elseif(dust_ivr == 2 .or. dust_ivr == 3) then   ! "Conrath nu" profile type

   if(dust_ivr == 2) then   ! User-specified parameters

      tmp2D_b(:,:) = dust_icn
      tmp2D_a(:,:) = dust_irt
      
   elseif(dust_ivr == 3) then     ! "var_tau" init

      ! Obtain current Ls (info for grid #1 is sufficient) [deg]:

      call stellar_angles(gd(1)%nzp, gd(1)%nxp, gd(1)%nyp, 1, 1, 1, 1,  &
           a(iglatn(1)), a(iglonn(1)), tmp2D_a, xsdist, gd(1)%centlat,  &
           gd(1)%centlon, lonrad, tmp2D_b, Ls_curr, .FALSE.)
 
      ! Calculate "more variable" radiatively-active dust loading parameters:
      !  (tmp2D_a holds tautot, tmp2D_b holds conrath_nu)

      if (map_projection == 0) then
         call mvar_dust_params(m1,m2,m3,1,m2,1,m3,Ls_curr,a(iglatn(i_grid)),  &
              tmp2D_a, tmp2D_b, constant_lat=gd(1)%centlat)
      else
         call mvar_dust_params(m1,m2,m3,1,m2,1,m3,Ls_curr,a(iglatn(i_grid)),  &
              tmp2D_a, tmp2D_b)
      end if

   end if

   ! Determine size/mass distribution parameters:

   call aer_ndist_logn(dust_imr*1.e-6_f4,dust_igd,dust_nb,1,aer_ndist_bfrac)

   ! Construct vertical profiles of dust:

   allocate(N_scr(-8:m1,dust_nb),dz_scr(-8:m1))

   k_norm = 0

   do j=1,m3
      do i=1,m2
         
         conrath_nu = tmp2D_b(i,j)

         do k=2,m1
            lyr_press = press(k,i,j) 

            if (lyr_press < dust_ctau_refp) then
             q_q0 = max(1.e-30_f4,exp(conrath_nu*(ONE_f4-dust_ctau_refp/  &
                  lyr_press)))
            else
               q_q0 = ONE_f4
            end if

            air_density = press(k,i,j)/(R_air*tempk(k,i,j))
            N_raw = q_q0*air_density   ! dust N proportional to mass
            
            do ib=1,dust_nb            ! Initialize all dust bins
               N_scr(k,ib) = aer_ndist_bfrac(ib)*N_raw  
            enddo
         enddo
         N_scr(1,1:dust_nb) = N_scr(2,1:dust_nb)   ! Bottom BC

         dz_scr(1:m1) = rtgt(i,j)/gd(i_grid)%dzt(1:m1)

         ! Attempt to determine layer that includes 'dust_ctau_refp' value:

         k_norm = 0
         if (dust_ctau_refp < press(1,i,j) .and.  & ! Only do if value in column
              dust_ctau_refp > press(m1-1,i,j)) then
            do k=2,m1-1
               p_bot = (press(k,i,j)+press(k-1,i,j))*0.5_f4
               p_top = (press(k,i,j)+press(k+1,i,j))*0.5_f4
               if (p_bot >= dust_ctau_refp .and. p_top < dust_ctau_refp) then
                  k_norm = k
                  ftmp = R_air*tempk(k,i,j)/g_Xoid*log(dust_ctau_refp/p_top)
                  dz_scr(k) = min(dz_scr(k), ftmp)
                  exit
               end if
            end do
         end if

         if(k_norm == 0) then   ! Need to invent some atmosphere
            
            p_bot = dust_ctau_refp
            ftmp = (dust_ctau_refp-(press(2,i,j)+press(1,i,j))*0.5_f4)/10._f4
            k_norm = -8

            do k=-8,1
               p_top = p_bot-ftmp
               lyr_press = 0.5_f4*(p_bot+p_top)
               
               if (lyr_press < dust_ctau_refp) then
                  q_q0 = exp(conrath_nu*(ONE_f4-dust_ctau_refp/lyr_press))
               else
                  q_q0 = ONE_f4
               end if

               dz_scr(k) = R_air*tempk(2,i,j)/g_Xoid*log(p_bot/p_top)

               air_density = lyr_press/(R_air*tempk(2,i,j))
               N_raw = q_q0*air_density   ! dust N proportional to mass
   
               do ib=1,dust_nb
                  N_scr(k,ib) = aer_ndist_bfrac(ib)*N_raw 
               enddo

               p_bot = p_top
            enddo
           
         endif

         tau_tgt = tmp2D_a(i,j)   ! Target opacity in column above 'k_norm'

         ! Initial guess at value of 'N_fact' resulting in 'tau_tgt' opacity (at
         !  "visible" wavelengths):
         N_fact = ONE_f4
         call calc_abp_opd_specific(-8, m1, dust_nb, 'dust_vis', k_norm, m1,  &
              N_scr, dz_scr, N_fact, column_tau)

         N_fact = tau_tgt/column_tau   ! Final normalization value

         ! Scale dust number concentrations:

         do ib=1,dust_nb
            dustnc(2:m1,i,j,ib) = N_scr(2:m1,ib)*N_fact   ! [#/cm^3]
         enddo
         dustnc(1,i,j,1:dust_nb) = dustnc(2,i,j,1:dust_nb)

      enddo
   enddo

   deallocate(N_scr,dz_scr)

endif

end subroutine dust_init
