!```````````````````````````````````````````````````````````````````
module aeX_water_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine H2Ov_init(m1,m2,m3,ia,iz,ja,jz,rv,density,zt,zm,rtgt)

!+ Initialize water vapor profile

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz
real(f4), dimension(m1,m2,m3), intent(IN) :: density    ! [kg/m^3]
real(f4), dimension(m1), intent(IN) :: zt,zm
real(f4), dimension(m2,m3), intent(IN) :: rtgt

real(f4), dimension(m1,m2,m3), intent(OUT) :: rv    ! [g/cm^3]

!== Local declarations:

real(f4), parameter :: kgm3_to_gcm3 = 1.e-3_f4  & ! Density conversion
                            ,H = 5000._f4  &   ! Scaling height for exp profile
                            ,m_0 = 1.6e-4_f4  ! H2Ov profile normalization factor

integer :: i,j,k
!%real(f4), dimension(m1) :: tau
real(f4) :: H2Ov_mass,rho_air_sfc_cgs
!real(f4) :: H2Ov_pd

!=== Executable statements:

do j=1,m3
   do i=1,m2

      do k=2,m1

!print *,i,j,k,density(1,i,j),density(2,i,j),kgm3_to_gcm3,zt(k),rtgt(i,j),H
         rho_air_sfc_cgs = 0.5_f4*(density(2,i,j)+density(1,i,j))*  &
              kgm3_to_gcm3               ! [g/cm^3]
!         print *,rho_air_sfc_cgs,rho_air_sfc_cgs/kgm3_to_gcm3
!print *,'sdff'
         ! H2O vapor mass concentration [g/cm^3]:
         H2Ov_mass = m_0*rho_air_sfc_cgs*exp(-zt(k)*rtgt(i,j)/H)
!print *,'www'
!%         ! Convert to water vapor mixing ratio [g/g (H2Ov/air)]:
!%         rv(k,i,j) = H2Ov_mass/density(k,i,j)

         ! "Sculpt" the water profile as desired:

         if(rtgt(i,j)*zt(k) <= H .and. k > 2) then
            rv(k,i,j) = rv(2,i,j)
         elseif(k == 2) then 
            rv(k,i,j) = H2Ov_mass
         else
!print *,'dede',k
            rv(k,i,j) = H2Ov_mass/real(k, kind=f4)
         endif

      enddo

      rv(1,i,j) = rv(2,i,j)           ! Bottom BC

!      if(i == 10 .and. j == 22) then
!         call H2Ov_cond_coldepth(m1,H2Ov_pd,rv(1:m1,i,j),density(1:m1,i,j)  &
!              ,rtgt(i,j),zm)
!         print *,'pd',i,j,H2Ov_pd,H2Ov_pd*1.e4
!         pause
!      endif

!%      call dust_optdepth(m1,tau,dust_mrat,density(1:m1,i,j),rtgt(i,j),zt)
!%      print *,'od',i,j,tau(2)
!%      pause

   enddo
enddo

end subroutine H2Ov_init


!```````````````````````````````````````````````````````````````````
subroutine H2Ov_bound(m1,m2,m3,mia,miz,mja,mjz,H2Ovmr,nudlat,BCtype)

!+ Boundary conditions for water vapor profile (w/o MGCM water BCs)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,mia,miz,mja,mjz,nudlat
character(len=8), intent(IN) :: BCtype

real(f4), dimension(m1,m2,m3), intent(INOUT) :: H2Ovmr

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

if(BCtype == 'zero_grd') then   ! Zero gradient BC for H2O vapor
   
   do j=mja,mja+nudlat-1
      do i=mia,miz
         do k=1,m1
            H2Ovmr(k,i,j) = H2Ovmr(k,i,mja+nudlat)
         enddo
      enddo
   enddo

   do j=mjz-nudlat+1,mjz
      do i=mia,miz
         do k=1,m1
            H2Ovmr(k,i,j) = H2Ovmr(k,i,mjz-nudlat)
         enddo
      enddo
   enddo

   do j=mja,mjz
      do i=mia,mia+nudlat-1
         do k=1,m1
            H2Ovmr(k,i,j) = H2Ovmr(k,mia+nudlat,j)
         enddo
      enddo
   enddo
   
   do j=mja,mjz
      do i=miz-nudlat+1,miz
         do k=1,m1
            H2Ovmr(k,i,j) = H2Ovmr(k,miz-nudlat,j)
         enddo
      enddo
   enddo

endif

end subroutine H2Ov_bound


!```````````````````````````````````````````````````````````````````
subroutine H2Ov_cond_coldepth(m1,H2Ov_pd,H2Ovmr,density,rtgt,zm)

!+ Compute depth [cm] of precipitable water in vapor column  

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4, ONE_f4
use aermicro_mem_mod, only: m_to_cm

implicit none

integer, intent(IN) :: m1
real(f4), dimension(m1), intent(IN) :: H2Ovmr,density,zm
real(f4), intent(IN) :: rtgt

real(f4), intent(OUT) :: H2Ov_pd

!== Local declarations:

real(f4), parameter :: H2O_rho = ONE_f4  ! Density of water [g/cm^3]

integer :: k
real(f4) :: H2Ov_lyrmass!,fd

!=== Executable statements:

H2Ov_pd = ZERO_f4
!fd = 0.
do k=2,m1
!mixrat%   H2Ov_lyrmass = H2Ovmr(k)*density(k)*rtgt*(zm(k)-zm(k-1))*m_to_cm
   H2Ov_lyrmass = H2Ovmr(k)*rtgt*(zm(k)-zm(k-1))*m_to_cm
!   fd = fd+(zm(k)-zm(k-1))*m_to_cm
!   print *,'dfgft',k,H2Ov_lyrmass,H2Ovmr(k),rtgt,zm(k),zm(k-1),(zm(k)-zm(k-1)),m_to_cm,fd,H2Ov_pd+H2Ov_lyrmass/H2O_rho
!   print *,k,H2Ovmr(k),H2Ovmr(k)/density(k)
   H2Ov_pd = H2Ov_pd+H2Ov_lyrmass/H2O_rho
enddo

end subroutine H2Ov_cond_coldepth


end module aeX_water_mod
