#!/usr/bin/env python
"""
Runs an NCL script that visualizes a set of PRAMS grids (in the horizontal).

This program requires either Python version 2.6+ or 3.2+, and is importable as a
python module.
"""

from __future__ import print_function, absolute_import, division, unicode_literals

import sys
import os
import subprocess
import getopt

#------------------------------------------------------------------------------
def main(vis_dir):
    """Driver routine."""

    test_fpath = os.path.join(vis_dir, "grid_display-data.nc")
    if os.path.isfile(test_fpath):
        cmd_list = []
        cmd_list.append("ncl")
        cmd_list.append('vis_dir="'+vis_dir+'"')
        cmd_list.append(os.path.join(anchor_path, "display_PRAMS_grids.ncl"))
        subprocess.check_call(cmd_list)
    else:
        print("ERROR: data file ("+test_fpath+ \
              ") was not found or is unreadable.")
        sys.exit(1)
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------
def usage():
    usage_msg = """
usage: run_display_PRAMS_grids.py [-h]  VIS_DIR

Runs an NCL script that visualizes a set of PRAMS grids (in the horizontal).

mandatory argument:
  VIS_DIR   The path to the 'vis' directory that contains the
            'grid_display-data.nc' file to visualize (default:
            ./output/MPF_test-01/vis"); any output files will also be placed
            in this directory.

optional arguments:
  -h, --help   show this help message and exit
"""
    print(usage_msg)
#------------------------------------------------------------------------------


if __name__ == "__main__":
    # Determine fully-qualified filesystem location of this script:
    anchor_path = os.path.abspath(os.path.dirname(sys.argv[0]))

    # Process arguments:
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'h', ["help"])
    except getopt.GetoptError as err:
        print(err)
        usage()
        sys.exit(1)
      # Default option values:
    # NONE
      # Act according to the options given:
    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit(1)
        else:
            assert False, "unhandled option"
      # Act according to the arguments given:
    if (len(args) > 1 or len(args) < 1):
        print("An incorrect number of arguments were provided.")
        usage()
        sys.exit(1)
    vis_dir = args[0]

    # Run driver:
    main(vis_dir)
