!```````````````````````````````````````````````````````````````````
subroutine grid_set(tog_grid)

!+ Calculate gridspacings on staggered grids, etc.

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4, ONE_f4
use univ_OPS_proj_mod, only: ll_xy_OPS
use body_specific_constants_mod, only: radius
use main_memory_mod, only: gd, jdim, ztop, n_grids, polelon, polelat,  &
     deltax_g1, deltay_g1, maxnzp, nestz1, nstratz1,  &
     run_is_purely_2D, deltaz_g1, zz, dzrat, dzmax
use nest_mod, only: coeff_nest

implicit none

!=== Argument declarations:

integer, intent(IN) :: tog_grid

!== Local declarations:

integer :: icm, ifm, nestza1, k, iinc, icnt, if, jinc, jcnt, jf, kinc, kcnt,  &
     kf, nrat, ngtmp, kcy, kcw, i, j, kk
real(f4) :: centx, centy, dzr, dsum, dzrcm, dzrfm, tsum
real(f4), dimension(-1:maxnzp+1,n_grids) :: zmnvc

!=== Executable statements:

! If tog_grid = 1, the model is being initialized.
! If tog_grid > 1, this routine is being called for moving a nested grid.

if (tog_grid == 1) then    ! Only if model being initialized

   ! Fill platn and plonn arrays (value of polelat and polelon for each grid)

   do ifm=1,n_grids
      icm = gd(ifm)%nxtnest
      if (ifm == 1) then
         gd(ifm)%plat = polelat
         gd(ifm)%plon = polelon
      else
         gd(ifm)%plat = gd(icm)%plat
         gd(ifm)%plon = gd(icm)%plon
      end if
   end do

! Fill NRZ with the variable nest ratios:

   do ifm=1,n_grids
      do k=1,maxnzp
         gd(ifm)%nrz(k) = 1
      end do
   end do

   nestza1 = abs(nestz1)

   if (nestza1 > 1 .and. nestza1 <= n_grids) then
      do k=2,gd(1)%nzp
         gd(nestza1)%nrz(k) = max(1, nstratz1(k))
      end do
      gd(nestza1)%nrz(1) = nstratz1(2)
   end if

! Fill all DELTAX(:) and DELTAY(:) values; find NINEST(:) and NJNEST(:) values
!  if set to zero in namelist

   gd(1)%deltax = deltax_g1
   gd(1)%deltay = deltay_g1
   call ll_xy_OPS(gd(1)%centlat, gd(1)%centlon, gd(1)%plat, gd(1)%plon,  &
        centx, centy, radius)
   gd(1)%xm(1) = centx-0.5_f4*real(gd(1)%nxp-2, kind=f4)*gd(1)%deltax
   gd(1)%ym(1) = centy-0.5_f4*real(gd(1)%nyp-2, kind=f4)*gd(1)%deltay
   if (run_is_purely_2D) gd(1)%ym(1) = centy

   do ifm=1,n_grids
      icm = gd(ifm)%nxtnest
      if (icm >= 1) then
         gd(ifm)%deltax = gd(icm)%deltax/real(gd(ifm)%nstratx, kind=f4)
         gd(ifm)%deltay = gd(icm)%deltay/real(gd(ifm)%nstraty, kind=f4)
         if (gd(ifm)%ninest <= 0 .or. gd(ifm)%njnest <= 0) then
            call ll_xy_OPS(gd(ifm)%centlat, gd(ifm)%centlon, gd(ifm)%plat,  &
                 gd(ifm)%plon, centx, centy, radius)
         end if
         if (gd(ifm)%ninest <= 0) then
            gd(ifm)%xm(1) = centx-0.5_f4*real(gd(ifm)%nxp-2, kind=f4)*  &
                 gd(ifm)%deltax
            gd(ifm)%ninest = int((gd(ifm)%xm(1)-  &
                 gd(icm)%xm(1))/gd(icm)%deltax+1.5_f4)
         end if
         gd(ifm)%xm(1) = gd(icm)%xm(1)+gd(icm)%deltax*  &
              real(gd(ifm)%ninest-1, kind=f4)
         if (gd(ifm)%njnest <= 0) then
            gd(ifm)%ym(1) = centy-0.5_f4*real(gd(ifm)%nyp-2, kind=f4)*  &
                 gd(ifm)%deltay
            gd(ifm)%njnest = int((gd(ifm)%ym(1)-  &
                 gd(icm)%ym(1))/gd(icm)%deltay+1.5_f4)
         end if
         gd(ifm)%ym(1) = gd(icm)%ym(1)+gd(icm)%deltay*  &
              real(gd(ifm)%njnest-1, kind=f4)
      end if

   end do

end if

! Fill IPM, JPM, and KPM arrays with parent grid index values (all fine grids):

do ifm=2,n_grids
   icm = gd(ifm)%nxtnest
   if (icm >= 1) then
      gd(ifm)%ipm(1) = gd(ifm)%ninest
      IINC = 1
      ICNT = 0
      DO IF=2,gd(ifm)%NXP
         gd(ifm)%IPM(IF) = gd(ifm)%IPM(IF-1)+IINC
         ICNT = ICNT+1
         IF (ICNT >= gd(ifm)%NSTRATX) THEN
            ICNT = 0
            IINC = 1
         ELSE
            IINC = 0
         END IF
      END DO

      gd(ifm)%JPM(1) = gd(ifm)%NJNEST
      JINC = 1
      JCNT = 0
      DO JF=2,gd(ifm)%NYP
         gd(ifm)%JPM(JF) = gd(ifm)%JPM(JF-1)+JINC
         JCNT = JCNT+1
         IF (JCNT >= gd(ifm)%NSTRATY) THEN
            JCNT = 0
            JINC = 1
         ELSE
            JINC = 0
         END IF
      END DO

      IF (gd(ifm)%NKNEST == 1) THEN
         gd(ifm)%KPM(1) = 2
         KINC = 0
      ELSE
         gd(ifm)%KPM(1) = gd(ifm)%NKNEST
         KINC = 1
      END IF
      KCNT = 0
      DO KF=2,gd(ifm)%NZP
         gd(ifm)%KPM(KF) = gd(ifm)%KPM(KF-1)+KINC
         NRAT = gd(ifm)%NRZ(gd(ifm)%KPM(KF))
         KCNT = KCNT+1
         IF (KCNT >= NRAT .AND. (KF < gd(ifm)%NZP-1 .OR.  &
              gd(ifm)%KPM(KF) < gd(icm)%NZP-1)) THEN
            KCNT = 0
            KINC = 1
         ELSE
            KINC = 0
         END IF
      END DO

   end if
end do

! Calculate xm(1) and ym(1) for the coarse grid for an initial start:

if (tog_grid == 1) then

   DO I=2,gd(1)%NXP
      gd(1)%XM(I) = gd(1)%XM(I-1)+gd(1)%DELTAX
   END DO

   DO J=2,gd(1)%NYP
      gd(1)%YM(J) = gd(1)%YM(J-1)+gd(1)%DELTAY
   END DO

end if

! Compute xm(:) and ym(:) for any required NESTED grids, and xt(:) and yt(:) for
!  any required grids:

if (tog_grid == 1) then
   ngtmp = n_grids
else
   ngtmp = tog_grid
end if

do ifm=tog_grid,ngtmp
   icm = gd(ifm)%nxtnest
   if (icm >= 1) then
      gd(ifm)%XM(1) = gd(icm)%XM(gd(ifm)%NINEST)
      gd(ifm)%YM(1) = gd(icm)%YM(gd(ifm)%NJNEST)
   end if
   DO I=2,gd(ifm)%NXP
      gd(ifm)%XM(I) = gd(ifm)%XM(I-1)+gd(ifm)%DELTAX
      gd(ifm)%XT(I) = 0.5_f4*(gd(ifm)%XM(I)+gd(ifm)%XM(I-1))
   END DO
   gd(ifm)%XT(1) = 1.5_f4*gd(ifm)%XM(1)-0.5_f4*gd(ifm)%XM(2)

   IF (JDIM == 1) THEN
      DO J=2,gd(ifm)%NYP
         gd(ifm)%YM(J) = gd(ifm)%YM(J-1)+gd(ifm)%DELTAY
         gd(ifm)%YT(J) = 0.5_f4*(gd(ifm)%YM(J)+gd(ifm)%YM(J-1))
      END DO
      gd(ifm)%YT(1) = 1.5_f4*gd(ifm)%YM(1)-0.5_f4*gd(ifm)%YM(2)
   ELSE
      gd(ifm)%YT(1) = gd(ifm)%YM(1)
   END IF

END DO

! Return if tog_grid > 1 since all remaining operations in this routine are
! required for initialization only:

if (tog_grid > 1) return

! Calculate zmn for the coarse grid(s):

IF (abs(DELTAZ_g1) <= 1.e-4_f4) THEN
   DO K=1,gd(1)%NZP
      gd(1)%ZM(K) = ZZ(K)
   END DO
ELSE
   gd(1)%ZM(1) = ZERO_f4
   gd(1)%ZM(2) = DELTAZ_g1

   ! Contort vertical gridspacing of coarser grid if desired:
   
   if (nestza1 > 1) then

      do k=3,gd(1)%nzp
         DZR = DZRAT

         IF (NESTZ1 < -1 .AND. K > gd(nestza1)%KPM(2) .AND.  &
              K <= gd(nestza1)%KPM(gd(nestza1)%NZP-1) .AND.  &
              gd(nestza1)%NRZ(K) /= gd(nestza1)%NRZ(K-1)) THEN
            IF (MAX(gd(nestza1)%NRZ(K),gd(nestza1)%NRZ(K-1)) == 2) then
               DZR = 1.325_f4**(gd(nestza1)%NRZ(K)-gd(nestza1)%NRZ(K-1))
            end if
            IF (MAX(gd(nestza1)%NRZ(K),gd(nestza1)%NRZ(K-1)) == 3) then
               DZR = 1.124_f4**(gd(nestza1)%NRZ(K)-gd(nestza1)%NRZ(K-1))
            end if
            IF (MAX(gd(nestza1)%NRZ(K),gd(nestza1)%NRZ(K-1)) == 4)  then
               DZR = 1.066_f4**(gd(nestza1)%NRZ(K)-gd(nestza1)%NRZ(K-1))
            end if
            IF (MAX(gd(nestza1)%NRZ(K),gd(nestza1)%NRZ(K-1)) == 5) then
               DZR = 1.042_f4**(gd(nestza1)%NRZ(K)-gd(nestza1)%NRZ(K-1))
            end if
         end if

         gd(1)%ZM(K) = gd(1)%ZM(K-1)+MIN(DZR*(gd(1)%ZM(K-1)-gd(1)%ZM(K-2)),  &
              DZMAX)
      END DO
   
   else      ! No contortion of coarse grid vertical gridspacing

      DO K=3,gd(1)%NZP
         DZR = DZRAT
         gd(1)%ZM(K) = gd(1)%ZM(K-1)+MIN(DZR*(gd(1)%ZM(K-1)-gd(1)%ZM(K-2)),  &
              DZMAX)
      END DO
      gd(1)%DELTAZ = DELTAZ_g1
      
   end if
end if

! Fill scratch array znmvc with ZM(:) for coarse grid(s):

ZTOP = gd(1)%ZM(gd(1)%NZP-1)

DO K=1,gd(1)%NZP
   ZMNVC(K,1) = gd(1)%ZM(K)
END DO
ZMNVC(0,1) = -(ZMNVC(2,1)-ZMNVC(1,1))**2/(ZMNVC(3,1)-ZMNVC(2,1))
ZMNVC(-1,1) = ZMNVC(0,1)-(ZMNVC(1,1)-ZMNVC(0,1))**2/(ZMNVC(2,1)-ZMNVC(1,1))
ZMNVC(gd(1)%NZP+1,1) = ZMNVC(gd(1)%NZP,1)  &
     +(ZMNVC(gd(1)%NZP,1)-ZMNVC(gd(1)%NZP-1,1))**2  &
     /(ZMNVC(gd(1)%NZP-1,1)-ZMNVC(gd(1)%NZP-2,1))

! Get zm(:) coordinates for all nested grids:

do ifm=2,N_GRIDS
   icm = gd(ifm)%nxtnest
   if (icm >= 1) then
      KCY = MAX(-1, gd(ifm)%NRZ(gd(ifm)%KPM(2))-3)
      DO K=-1,gd(ifm)%NZP+1
         IF (K <= 0) THEN
            NRAT = gd(ifm)%NRZ(gd(ifm)%KPM(2))
            KCW = gd(ifm)%NKNEST-1
            IF (K == -1 .AND. NRAT == 1) KCW = gd(ifm)%NKNEST - 2
         ELSE IF (K == gd(ifm)%NZP+1) THEN
            NRAT = gd(ifm)%NRZ(gd(ifm)%KPM(gd(ifm)%NZP-1))
         ELSE
            NRAT = gd(ifm)%NRZ(gd(ifm)%KPM(K))
         END IF
         KCY = MOD(KCY+1, NRAT)

         IF (KCY == 0) THEN
            IF (K >= 1) KCW = KCW+1
            ZMNVC(K,ifm) = ZMNVC(KCW,icm)
         ELSE
            IF (KCY == 1 .OR. K == -1) THEN
               DSUM = 0.
               DZRCM = SQRT((ZMNVC(KCW+2,icm)-ZMNVC(KCW+1,icm))  &
                    *gd(ifm)%NRZ(MAX(1, KCW))/((ZMNVC(KCW,icm)-ZMNVC(KCW-1,icm))  &
                    *gd(ifm)%NRZ(KCW+2)))
               DZRFM = DZRCM**(1./real(NRAT, kind=f4))
               TSUM = 0.
               DO KK=1,NRAT
                  TSUM = TSUM+DZRFM**(KK-1)
               END DO
            END IF
            IF (K == -1) THEN
               DO KK=1,KCY-1
                  DSUM = DSUM+((ZMNVC(KCW+1,icm)-ZMNVC(KCW,icm))  &
                       /TSUM)*DZRFM**(KK-1)
               END DO
            END IF
            DSUM = DSUM+(ZMNVC(KCW+1,icm)-ZMNVC(KCW,icm))  &
                 /TSUM*DZRFM**(KCY-1)
            ZMNVC(K,ifm) = ZMNVC(KCW,icm)+DSUM
         END IF

         IF (K >= 1 .AND. K <= gd(ifm)%NZP) gd(ifm)%ZM(K) = ZMNVC(K,ifm)
      END DO
   end if
END DO

! Compute ZT(:) values for all grids by geometric interpolation:

DO ifm=1,N_GRIDS
   DO K=1,gd(ifm)%NZP
      DZRFM = SQRT(SQRT((ZMNVC(K+1,ifm)-ZMNVC(K,ifm))  &
         /(ZMNVC(K-1,ifm)-ZMNVC(K-2,ifm))))
      gd(ifm)%ZT(K) = ZMNVC(K-1,ifm)+(ZMNVC(K,ifm)-ZMNVC(K-1,ifm))/(ONE_f4+DZRFM)
   END DO

   ! Compute other arrays based on the vertical grid:

   DO K = 1,gd(ifm)%NZP-1
      gd(ifm)%DZM(K) = ONE_f4/(gd(ifm)%ZT(K+1)-gd(ifm)%ZT(K))
   END DO
   DO K=2,gd(ifm)%NZP
      gd(ifm)%DZT(K)=ONE_f4/(gd(ifm)%ZM(K)-gd(ifm)%ZM(K-1))
   END DO
   DO K = 2,gd(ifm)%NZP-1
      gd(ifm)%DZM2(K) = ONE_f4/(gd(ifm)%ZM(K+1)-gd(ifm)%ZM(K-1))
      gd(ifm)%DZT2(K) = ONE_f4/(gd(ifm)%ZT(K+1)-gd(ifm)%ZT(K-1))
   END DO

   gd(ifm)%DZM(gd(ifm)%NZP) = gd(ifm)%DZM(gd(ifm)%NZP-1)  &
        *gd(ifm)%DZM(gd(ifm)%NZP-1)/gd(ifm)%DZM(gd(ifm)%NZP-2)
   gd(ifm)%DZT(1) = gd(ifm)%DZT(2)*gd(ifm)%DZT(2)/gd(ifm)%DZT(3)

   gd(ifm)%DZM2(1) = gd(ifm)%DZM2(2)
   gd(ifm)%DZM2(gd(ifm)%NZP) = gd(ifm)%DZM2(gd(ifm)%NZP-1)
   gd(ifm)%DZT2(1) = gd(ifm)%DZT2(2)
   gd(ifm)%DZT2(gd(ifm)%NZP) = gd(ifm)%DZT2(gd(ifm)%NZP-1)

   gd(ifm)%deltaz = gd(ifm)%zm(2)-gd(ifm)%zm(1)

END DO

!do ifm=1,ngrids
!   do i=1,nnxp(ifm)
!      print *,'x',ifm,xmn(i,ifm),xtn(i,ifm)
!   enddo
!enddo
!do j=1,nnyp(1)
!   print *,'y',ymn(j,1),ytn(j,1)
!enddo
!pause
!do k=1,nnzp(1)
!   print *,'z',zmn(k,1),ztn(k,1)
!enddo
!pause

call coeff_nest    ! Calculate nesting interpolation/feedback coefficients

end subroutine grid_set


!```````````````````````````````````````````````````````````````````
subroutine print_gridloc

!+ Print grid configuration, etc.

use univ_kind_defs_mod, only: f4
use univ_OPS_proj_mod, only: xy_ll_OPS
use main_memory_mod, only: pbanner, pbnlen, a, gd, n_grids
use vtables_mod, only: iglatn, iglonn
use body_specific_constants_mod, only: radius
use schedule_mod, only: value_gp

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: igd, midx, midy, i, j, n2, n3
real(f4) :: centx, centy
real(f4), dimension(19) :: gscr

!=== Executable statements:

write (*, '(/,2a)') pbanner(1:pbnlen),  &
     '************* Grid locations and dimensions *************'

do igd=1,n_grids

   n2 = gd(igd)%nxp; n3 = gd(igd)%nyp

   gscr(1) = value_gp(1, n2, n3, 1, 1, 1, a(iglatn(igd)))
   gscr(2) = value_gp(1, n2, n3, 1, 1, 1, a(iglonn(igd)))
   gscr(3) = value_gp(1, n2, n3, 1, n2, 1, a(iglatn(igd)))
   gscr(4) = value_gp(1, n2, n3, 1, n2, 1, a(iglonn(igd)))
   gscr(5) = value_gp(1, n2, n3, 1, n2, n3, a(iglatn(igd)))
   gscr(6) = value_gp(1, n2, n3, 1, n2, n3, a(iglonn(igd)))
   gscr(7) = value_gp(1, n2, n3, 1, 1, n3, a(iglatn(igd)))
   gscr(8) = value_gp(1, n2, n3, 1, 1, n3, a(iglonn(igd)))
   
   midx = (n2+1)/2
   midy = (n3+1)/2
   if (n2/2 == midx) then
      centx = gd(igd)%xm(n2/2)
   else
      centx = gd(igd)%xt(midx)
   end if
   if (n3/2 == midy) then
      centy = gd(igd)%ym(n3/2)
   else
      centy = gd(igd)%yt(midy)
   end if
   call xy_ll_OPS(gscr(9), gscr(10), gd(igd)%plat, gd(igd)%plon, centx,  &
        centy, radius)

   gscr(12) = -1000._f4
   gscr(13) = -1000._f4
   gscr(14) = 1000._f4
   gscr(15) = 1000._f4
   gscr(16) = 1000._f4
   gscr(17) = 1000._f4
   gscr(18) = -1000._f4
   gscr(19) = -1000._f4

   do j=1,n3
      do i=1,n2
         if (j == n3) then
            gscr(12) = max(gscr(12), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglatn(igd)))) ! maxN
            gscr(13) = max(gscr(13), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglonn(igd)))) !
         end if
         if (j == 1) then
            gscr(14) = min(gscr(14), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglatn(igd)))) ! maxS
            gscr(15) = min(gscr(15), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglonn(igd)))) !
         end if
         if (i == 1) then
            gscr(16) = min(gscr(16), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglatn(igd)))) ! maxW
            gscr(17) = min(gscr(17), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglonn(igd)))) !
         end if
         if (i == n2) then
            gscr(18) = max(gscr(18), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglatn(igd)))) ! maxE
            gscr(19) = max(gscr(19), value_gp(1, n2, n3, 1, i, j,  &
                 a(iglonn(igd)))) !
         end if
      end do
   end do

   write (*, '(/,a,i2,a)') '----------------- GRID #', igd,  &
        ': -------------------'

   write (*, '(/,a,i4,a,i4,a)') '(NINEST,NJNEST) = (', gd(igd)%ninest, ', ',  &
        gd(igd)%njnest, ')'

   write (*, '(/,a,3(3x,a1,f8.3,a1,f8.3,a1))') 'NW  maxN  NE',  &
        '(', gscr(7), ',', gscr(8), ')', '(', gscr(12), ',', gscr(13), ')',  &
        '(', gscr(5), ',', gscr(6), ')'
   write (*, '(/,a,3(3x,a1,f8.3,a1,f8.3,a1))') 'maxW ** maxE',  &
        '(', gscr(16), ',', gscr(17), ')', '(', gscr(9), ',', gscr(10), ')',  &
        '(', gscr(18), ',', gscr(19), ')'
   write (*, '(/,a,3(3x,a1,f8.3,a1,f8.3,a1))') 'SW  maxS  SE',  &
        '(', gscr(1), ',', gscr(2), ')', '(', gscr(14), ',', gscr(15), ')',  &
        '(', gscr(3), ',', gscr(4), ')'
   write (*, '(/,20x,a,f6.1,a,4x,a,f9.1,a)') 'Bottom =', gd(igd)%zm(1), ' m',  &
        'Top =', gd(igd)%zm(gd(igd)%nzp), ' m'

   write (*, '(/,10x,a,f9.1,a,3x,a,f9.1,a,3x,a,f6.1,a)')  &
        'DX =', gd(igd)%deltax, ' m', 'DY =', gd(igd)%deltay, ' m',  &
        'bottom DZ =', gd(igd)%deltaz, ' m'

end do

write (*, '(/,80(a1),/)') ('-', igd=1,80)

end subroutine print_gridloc
