!```````````````````````````````````````````````````````````````````
module grid_info_file_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
recursive subroutine rank_grids(ngrids, ngrids_parent_of, daughter_grid,  &
     parent_grid, grid_rank, grid_mask, this_rank)

!+ Recursive routine that determines grid relationships for GRID_INFO_FILE mode

implicit none

!=== Argument declarations:

integer, intent(IN) :: ngrids, parent_grid, this_rank
integer, dimension(ngrids), intent(IN) :: ngrids_parent_of
integer, dimension(ngrids,ngrids), intent(IN) :: daughter_grid

integer, dimension(ngrids), intent(INOUT) :: grid_rank
logical, dimension(ngrids,ngrids), intent(INOUT) :: grid_mask

!== Local declarations:

integer :: iq

!=== Executable statements:

do iq=1,ngrids_parent_of(parent_grid)
   grid_rank(daughter_grid(iq,parent_grid)) = this_rank
   grid_mask(daughter_grid(iq,parent_grid),parent_grid) = .TRUE.
   call rank_grids(ngrids, ngrids_parent_of, daughter_grid,  &
        daughter_grid(iq,parent_grid), grid_rank, grid_mask, this_rank+1)
end do

end subroutine rank_grids


!```````````````````````````````````````````````````````````````````
subroutine grid_display_data_write(output_dir)

! Writes data file(s) intended to enable plotting of the model grid location(s)

use univ_kind_defs_mod, only: f4
use univ_numtype_constants_mod, only: huge_f4
use univ_maxlen_string_mod, only: MAX_FILE_PATH_LEN, MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use manage_NetCDF_classic_mod, only: mNc_MAX_NAME_LEN, mNc_file_md_Att_obj,  &
     mNc_file_metadata_obj, mNc_define_f_md_Att_obj, mNc_populate_f_md_obj_w,  &
     mNc_FLOAT_type, mNc_recycle_f_md_Att_obj, mNc_recycle_f_md_obj,  &
     mNc_sync_with_file, mNc_INT_type, mNc_f_md_Att_populated
use manage_NetCDF_classic_wv_mod, only: mNc_write_var
use main_memory_mod, only: a, gd, n_grids, PRAMS_version,  &
     default_input_data_dir, simulation_ID_w_sfx, polelat, polelon
use vtables_mod, only: iglatn, iglonn, ifmaptn
use body_specific_constants_mod, only: dist_deglat

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: output_dir

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'grid_display_data_write'

character(len=MAX_FILE_PATH_LEN) :: fpath_md
character(len=3), dimension(n_grids) :: cgrid
integer :: igd, a_err, dim1_n, dim2_n, i, i_d1, i_d2, ic, ia, ab, ae, nxyp
logical :: initialize

real(f4), allocatable, dimension(:,:) :: a_tmp_2d

type(mNc_file_metadata_obj) :: output_file_obj

! Items for storage/processing of NetCDF metadata:
integer :: n_dims, n_vars, n_g_Atts, max_var_ndims, max_var_nAtts
character(len=mNc_MAX_NAME_LEN), allocatable, dimension(:) :: dim_name,  &
     var_name
real(f4), allocatable, dimension(:) :: var_f4_missing_value
integer, allocatable, dimension(:) :: dimv, var_type, var_ndims, var_nAtts
integer, allocatable, dimension(:,:) :: var_dimids
type(mNc_file_md_Att_obj), allocatable, dimension(:) :: g_Att
type(mNc_file_md_Att_obj), allocatable, dimension(:,:) :: var_Att
character(len=mNc_MAX_NAME_LEN) :: tmp_mNc_str

!=== Executable statements:

initialize = .TRUE.

! Total number of dimensions, variables, and global attributes:
n_dims = 2*n_grids+1; n_vars = 3*n_grids+5; n_g_Atts = 3

! Allocate and define "dimensions":
if (initialize) then
   allocate(dim_name(n_dims), dimv(n_dims), stat=a_err)
   if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
        procedure_name, 'dim_n_dims')
end if
i = 1; dim_name(i) = 'n_grids'; dimv(i) = n_grids
do igd=1,n_grids
   write (cgrid(igd), '(a1,i2.2)') 'g', igd 
   i = i+1; dim_name(i) = 'x_'//cgrid(igd); dimv(i) = gd(igd)%nxp
   i = i+1; dim_name(i) = 'y_'//cgrid(igd); dimv(i) = gd(igd)%nyp
end do

! Allocate some variable- and global_attribute-related items:
if (initialize) then
   allocate(var_name(n_vars), var_type(n_vars), var_ndims(n_vars),  &
        var_nAtts(n_vars), var_f4_missing_value(n_vars), stat=a_err) 
   if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
        procedure_name, 'dim_n_vars')
   allocate(g_Att(n_g_Atts), stat=a_err) 
   if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
        procedure_name, 'dim_n_g_Atts')    
end if

! Define "global attributes":
i = 1; tmp_mNc_str = trim(PRAMS_version); ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(g_Att(i), 'PRAMS_version', i, ic, tmp_mNc_str)
i = 2; tmp_mNc_str = trim(simulation_ID_w_sfx); ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(g_Att(i), 'run_moniker', i, ic, tmp_mNc_str)
i = 3; tmp_mNc_str = trim(default_input_data_dir); ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(g_Att(i), 'input_sfc_data_dir', i, ic,  &
     tmp_mNc_str)

! Allocate variable-related items:
max_var_ndims = 2; max_var_nAtts = 4
if (initialize) then
   allocate(var_dimids(max_var_ndims, n_vars), stat=a_err) 
   if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
        procedure_name, 'dim_max_var_ndims')
   allocate(var_Att(max_var_nAtts, n_vars), stat=a_err) 
   if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
        procedure_name, 'dim_max_var_nAtts')
end if

! Define "variables":

i = 0
do igd=1,n_grids
   i = i+1; var_name(i) = 'lat_'//cgrid(igd); var_type(i) = mNc_FLOAT_type
   var_ndims(i) = 2; var_nAtts(i) = 4
   var_dimids(1:var_ndims(i),i) = (/ (igd-1)*2+2, (igd-1)*2+3 /)
   ia = 1; tmp_mNc_str = 'latitude of mass-points '//cgrid(igd)
   ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
        tmp_mNc_str)
   ia = 2; tmp_mNc_str = 'degrees_N'; ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
   var_f4_missing_value(i) = -181._f4
   ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
        1, (/ var_f4_missing_value(i) /))
   ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
        (/ var_f4_missing_value(i) /))
   
   i = i+1; var_name(i) = 'lon_'//cgrid(igd); var_type(i) = mNc_FLOAT_type
   var_ndims(i) = 2; var_nAtts(i) = 4
   var_dimids(1:var_ndims(i),i) = (/ (igd-1)*2+2, (igd-1)*2+3 /)
   ia = 1; tmp_mNc_str = 'longitude of mass-points '//cgrid(igd)
   ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
        tmp_mNc_str)
   ia = 2; tmp_mNc_str = 'degrees_E'; ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
   var_f4_missing_value(i) = -361._f4
   ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
        1, (/ var_f4_missing_value(i) /))
   ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
        (/ var_f4_missing_value(i) /))

   i = i+1; var_name(i) = 'pt_scale_f_'//cgrid(igd)
   var_type(i) = mNc_FLOAT_type
   var_ndims(i) = 2; var_nAtts(i) = 4
   var_dimids(1:var_ndims(i),i) = (/ (igd-1)*2+2, (igd-1)*2+3 /)
   ia = 1; tmp_mNc_str = 'point scale factor '//cgrid(igd)
   ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
        tmp_mNc_str)
   ia = 2; tmp_mNc_str = 'none'; ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
   var_f4_missing_value(i) = -huge_f4
   ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
        1, (/ var_f4_missing_value(i) /))
   ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
        (/ var_f4_missing_value(i) /))
end do

i = i+1; var_name(i) = 'nxtnest'; var_type(i) = mNc_INT_type
var_ndims(i) = 1; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ 1 /)
ia = 1; tmp_mNc_str = 'parent grid ID'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 'none'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ -1 /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ -1 /))

i = i+1; var_name(i) = 'grid_sp_nom_deg'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 1; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ 1 /)
ia = 1; tmp_mNc_str = 'nominal gridspacing'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
     tmp_mNc_str)
ia = 2; tmp_mNc_str = 'degrees_latitude'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
var_f4_missing_value(i) = -361._f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
     1, (/ var_f4_missing_value(i) /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ var_f4_missing_value(i) /))

i = i+1; var_name(i) = 'grid_sp_nom_m'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 1; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ 1 /)
ia = 1; tmp_mNc_str = 'nominal gridspacing'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
     tmp_mNc_str)
ia = 2; tmp_mNc_str = 'm'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
var_f4_missing_value(i) = -1._f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
     1, (/ var_f4_missing_value(i) /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ var_f4_missing_value(i) /))

i = i+1; var_name(i) = 'polelat'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 0; var_nAtts(i) = 4
ia = 1; tmp_mNc_str = 'pole latitude of OPS map projection'
ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 'degrees_N'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
var_f4_missing_value(i) = -181._f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ var_f4_missing_value(i) /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ var_f4_missing_value(i) /))

i = i+1; var_name(i) = 'polelon'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 0; var_nAtts(i) = 4
ia = 1; tmp_mNc_str = 'pole longitude of OPS map projection'
ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 'degrees_E'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
var_f4_missing_value(i) = -361._f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ var_f4_missing_value(i) /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ var_f4_missing_value(i) /))

! Create output data file name:
fpath_md = trim(output_dir)//'/grid_display-data.nc'

! Begin writing the output file:
call mNc_populate_f_md_obj_w(fpath_md, output_file_obj, n_dims, n_vars,  &
     n_g_Atts, dim_name, dimv, g_Att, var_name, var_type, var_ndims,  &
     var_nAtts, var_dimids, var_Att)

! Write variables to the output file:

do igd=1,n_grids
   dim1_n = gd(igd)%nxp; dim2_n = gd(igd)%nyp; nxyp = gd(igd)%nxyp

   if (allocated(a_tmp_2d)) then
      if (size(a_tmp_2d, 1) /= dim1_n .or. size(a_tmp_2d, 2) /= dim2_n) then
         deallocate(a_tmp_2d, stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'a_tmp_2d')
         allocate(a_tmp_2d(dim1_n,dim2_n), stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
              procedure_name, 'a_tmp_2d')
      end if
   else
      allocate(a_tmp_2d(dim1_n,dim2_n), stat=a_err)
      if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
           procedure_name, 'a_tmp_2d')
   end if

   ab = iglatn(igd); ae = ab+nxyp-1
   a_tmp_2d = reshape(a(ab:ae), (/ dim1_n, dim2_n /))
   tmp_mNc_str = 'lat_'//cgrid(igd)
   call mNc_write_var(output_file_obj, trim(tmp_mNc_str), a_tmp_2d)

   ab = iglonn(igd); ae = ab+nxyp-1
   a_tmp_2d = reshape(a(ab:ae), (/ dim1_n, dim2_n /))
   tmp_mNc_str = 'lon_'//cgrid(igd)
   call mNc_write_var(output_file_obj, trim(tmp_mNc_str), a_tmp_2d)

   ab = ifmaptn(igd); ae = ab+nxyp-1
   a_tmp_2d = reshape(a(ab:ae), (/ dim1_n, dim2_n /))
   tmp_mNc_str = 'pt_scale_f_'//cgrid(igd)
   call mNc_write_var(output_file_obj, trim(tmp_mNc_str), a_tmp_2d)
end do

call mNc_write_var(output_file_obj, 'nxtnest', gd(1:n_grids)%nxtnest)
call mNc_write_var(output_file_obj, 'grid_sp_nom_deg',  &
     gd(1:n_grids)%deltax/dist_deglat)
call mNc_write_var(output_file_obj, 'grid_sp_nom_m',  &
     gd(1:n_grids)%deltax)
call mNc_write_var(output_file_obj, 'polelat', polelat)
call mNc_write_var(output_file_obj, 'polelon', polelon)

call mNc_sync_with_file(output_file_obj)   ! Optional

! Finalize file, free object memory:
call mNc_recycle_f_md_obj(output_file_obj)

! Must "recycle" the attribute objects before reusing:
do i_d2=1,size(var_Att,2)
   do i_d1=1,size(var_Att,1)
      if (mNc_f_md_Att_populated(var_Att(i_d1,i_d2)))  &
           call mNc_recycle_f_md_Att_obj(var_Att(i_d1,i_d2))
   end do
end do
do i_d1=1,size(g_Att,1)
   if (mNc_f_md_Att_populated(g_Att(i_d1)))  &
        call mNc_recycle_f_md_Att_obj(g_Att(i_d1))
end do

deallocate(a_tmp_2d, stat=a_err)
if (a_err /= 0) call illegal_alloc_stop(for_deallocation,  &
     procedure_name, 'a_tmp_2d')

deallocate(dim_name, dimv, stat=a_err)
if (a_err /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_n_dims')
deallocate(var_name, var_type, var_ndims, var_nAtts, var_f4_missing_value,  &
     stat=a_err) 
if (a_err /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_n_vars')
deallocate(g_Att, stat=a_err) 
if (a_err /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_n_g_Atts')    
deallocate(var_dimids, stat=a_err) 
if (a_err /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_max_var_ndims')
deallocate(var_Att, stat=a_err) 
if (a_err /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_max_var_nAtts')

write (*, '(/,a,/)') '*** GRID_INFO_FILE run completed successfully ***'

stop '*** Normal completion ***'

end subroutine grid_display_data_write


end module grid_info_file_mod
