!```````````````````````````````````````````````````````````````````
subroutine initialize(nlst_fn)

!+ Main PRAMS initialization driver

use univ_kind_defs_mod, only: f4, f8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_math_constants_mod, only: ZERO_f8
use univ_exception_mod, only: exception, FATAL
use univ_std_fileIO_mod, only: mkdir_p, find_file_IOunit
use univ_maxlen_string_mod, only: MAX_FILE_PATH_LEN
use main_memory_mod, only: there_is_an_error, pbanner, pbnlen, gd, a,  &
     iscr1, sfw_call, run_type, n_grids,  &
     initialization_type, naddsc, nzg, nzs, npatch,  &
     target_Ls, start_UT, input_varfile_pfx, start_sim_ET,  &
     current_sim_ET, perturb_nsfc_theta,  &
     state_output_fpfx, output_dir_root_w_sim_ID, simulation_ID_w_sfx,  &
     orbital_info_table, RT_paradigm, igdref
use main_memory_bspec_mod, only: dust_beg, H2Ov_beg, H2Oi_beg, CO2i_beg,  &
     H2Ov_tog, dust_rtt, H2Oi_rtt, CO2i_rtt, dust_ivr, H2Ov_ivr, RTc_tog,  &
     dust_jbg, dust_jsr, Ls_of_periapse, orbital_eccentricity,  &
     axial_obliquity, AU_to_m_knob
use vtables_mod, only: iglatn, iglonn, ifmapun, ifmapvn, ifmaptn, ifmapmn,  &
     ifmapuin, ifmapvin, ifmaptin, ifmapmin, idxun, idxvn, idxtn, idxmn,  &
     idyun, idyvn, idytn, idymn, ifcorun, ifcorvn, itoptn, itopun, itopvn,  &
     itopmn, irtgtn, irtgun, irtgvn, irtgmn, if13un, if13vn, if13tn, if13mn,  &
     if23un, if23vn, if23tn, if23mn, ithpn, ithetan, irvn, ipi0n,  &
     idn0n, idn0un, idn0vn, ith0n, isclpn, naersclr, iaersclp, ind3dn
use aermic_rad_mem_mod, only: aer_gpbeg, aer_rvbeg, iaertog, aer_active,  &
     aer_rad_gpbeg, aer_inj_gpbeg, aer_csr_gpbeg
use PRAMS_init_mem_mod, only: sndref_print, refs3d
use state_IO_mod, only: n_grids_st
use PRAMS_chkcfg_mod, only: chkcfg3, chkcfg2
use aerad_interface_mod, only: aerad_setup
use ARCnRT_interface_mod, only: ARCnRT_setup
use radn_interface_mod, only: radn_setup
use PRAMS_thermo_mod, only: dry_thermo
use nest_interpolation_mod, only: fmrefs1d
use nest_mod, only: fdback
use aeX_water_mod, only: H2Ov_init
use varfile_init_mod, only: varfile_find, refstate_loc
use aemic_drvint_mod, only: aerproc_check
use grid_mem_mod, only: coord_transforms, grid_spacing, map_facts
use PRAMS_coriolis_mod, only: f_coriolis
use surface_file_mod, only: sfc_files
use simple_astron_scheme_mod, only: populate_sas_table, set_sloc_sas_table,  &
     get_current_sas_vals
use body_specific_constants_mod, only: morbpd, smaxis, mass_bodysys,  &
     massratio_Sun2bodysys, mnsdy
use grid_info_file_mod, only: grid_display_data_write

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: nlst_fn

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'initialize'

integer :: ifm, icm, itracerm, n1, n2, n3, i, IO_unit, igd, iic, iif
logical :: this_is_a_restart
character(len=MAX_FILE_PATH_LEN) :: fpath_tmp
character(len=4) :: junkc
real(f4) :: tmp1
real(f8) :: tmp1_f8, tmp2_f8, tmp3_f8

!=== Executable statements:

write (*, '(/,2a)') pbanner(1:pbnlen), 'Starting model initialization...'

! Orbital initialization:
call populate_sas_table(orbital_info_table, morbpd, 1000._f8, smaxis,  &
     real(AU_to_m_knob, kind=f8), mass_bodysys, massratio_Sun2bodysys,  &
     real(orbital_eccentricity, kind=f8), real(Ls_of_periapse, kind=f8),  &
     real(axial_obliquity, kind=f8))

! Set some time parameters accordingly:

aer_gpbeg(1) = dust_beg      !
aer_rvbeg = H2Ov_beg         ! Aerosol micro/rad start times
aer_gpbeg(2) = H2Oi_beg      !
aer_gpbeg(3) = CO2i_beg      !
aer_rad_gpbeg(1) = dust_rtt  !
aer_rad_gpbeg(2) = H2Oi_rtt  !
aer_rad_gpbeg(3) = CO2i_rtt  !
aer_inj_gpbeg(1) = dust_jbg  !
aer_csr_gpbeg(1) = dust_jsr  !

! Calculate basic parameters that define the model grid(s):
call grid_set(1)

if (run_type /= 'RESTART') then   ! This is *MAY* be an initial start of time
                                !  integration, OR may be another intermediate
                                !  runtype

   ! Set some control parameters appropriately:

   if (run_type == 'INITIAL_START') then    ! Varfile timefile should exist

      ! Initialize variables that record the write status of file(s):
      sfw_call = -1

      gd(1:n_grids)%RT_call = -1

      if (initialization_type == 2) then
         IO_unit = find_file_IOunit()
         fpath_tmp = trim(input_varfile_pfx)//'-Vtime.tag' ! Timefile filename
         open(IO_unit, file=fpath_tmp(1:len_trim(fpath_tmp)), status='old',  &
              form='formatted')  ! Open timefile
         read (IO_unit, '(a4,e14.7,2x,a4,e14.7)') junkc, tmp1, junkc, start_UT 
         close(IO_unit)    ! Close timefile

         if (target_Ls <= -tiny(target_Ls)) target_Ls = tmp1  ! Set from file
      end if
     
      aer_active = .FALSE.   ! Aerosol microphysics has not been activated yet

      ! Find and record starting location in orbital/time table:
      tmp1_f8 = real(target_Ls, kind=f8); tmp2_f8 = real(start_UT, kind=f8)
      call set_sloc_sas_table(orbital_info_table, tmp2_f8, mnsdy, tmp1_f8)
      target_Ls = real(tmp1_f8, kind=f4)

   else       ! Not 'RESTART' or 'INITIAL_START'

      ! Find and record starting location in orbital/time table (using
      !  arbitrary starting Ls and UT values):
      tmp1_f8 = 142._f8
      call set_sloc_sas_table(orbital_info_table, 5._f8, mnsdy, tmp1_f8)

   end if

   ! Set other bookkeeping variables:

   n_grids_st = 0   ! Not a restart run (needed for completeness)
   this_is_a_restart = .FALSE.
   
   current_sim_ET = ZERO_f8   ! Set "current" and "run-start" sim ET to *zero*
   start_sim_ET = ZERO_f8     !

else   ! This is a restart

   call restart      ! Check state file compatibility with current config;
                     !    read in data, fill model fields

   this_is_a_restart = .TRUE.

   current_sim_ET = start_sim_ET ! Set "current"  sim ET to "run-start" time

   ! Find and record starting location in orbital/time table:
   tmp1_f8 = real(target_Ls, kind=f8); tmp2_f8 = real(start_UT, kind=f8)
   call set_sloc_sas_table(orbital_info_table, tmp2_f8, mnsdy, tmp1_f8)
   target_Ls = real(tmp1_f8, kind=f4)

end if

! Fill latitude-longitude, map factor, and Coriolis arrays:

do ifm=1,n_grids

   gd(ifm)%ngbegun = 0
   
   n1 = gd(ifm)%nzp; n2 = gd(ifm)%nxp; n3 = gd(ifm)%nyp
   
   call map_facts(n2, n3, a(iglatn(ifm)), a(iglonn(ifm)), a(ifmapun(ifm)),  &
        a(ifmapvn(ifm)), a(ifmaptn(ifm)), a(ifmapmn(ifm)), a(ifmapuin(ifm)),  &
        a(ifmapvin(ifm)), a(ifmaptin(ifm)), a(ifmapmin(ifm)), gd(ifm)%xm,  &
        gd(ifm)%xt, gd(ifm)%ym, gd(ifm)%yt, gd(ifm)%plat, gd(ifm)%plon)
    
   call grid_spacing(n2, n3, gd(ifm)%nx, gd(ifm)%ny, a(idxun(ifm)),  &
        a(idxvn(ifm)), a(idxtn(ifm)), a(idxmn(ifm)), a(idyun(ifm)),  &
        a(idyvn(ifm)), a(idytn(ifm)), a(idymn(ifm)), a(ifmapun(ifm)),  &
        a(ifmapvn(ifm)), a(ifmaptn(ifm)), a(ifmapmn(ifm)), gd(ifm)%xt,  &
        gd(ifm)%xm, gd(ifm)%yt, gd(ifm)%ym, gd(ifm)%deltay)

   call f_coriolis(n2, n3, gd(ifm)%nx, gd(ifm)%ny, a(ifcorun(ifm)),  &
        a(ifcorvn(ifm)), a(iglatn(ifm)))

end do

! Print grid locations, size, etc.:

call print_gridloc

! Check grid placement, option specifications, etc.:

there_is_an_error = .FALSE.
call chkcfg2
if (there_is_an_error) call exception(procedure_name, FATAL, 'chkcfg2')

! If this is a "GRID_INFO_FILE" run (PRAMS will exit within this routine):
fpath_tmp = trim(output_dir_root_w_sim_ID)//'/vis'
call mkdir_p(trim(fpath_tmp))  ! Create any needed dirs
if (run_type == 'GRID_INFO_FILE') call grid_display_data_write(fpath_tmp)

! Read and check surface files; remake them if necessary:

call sfc_files(this_is_a_restart)
if (run_type == 'MAKE_SFC_FILES' .or. run_type == 'CHECK_SFC_FILES') then
   write (*, '(/,a)') '*** MAKE_SFC_FILES/CHECK_SFC_FILES run complete ***'
   return
end if

! Define transformation Jacobians, etc. for all grids:

do ifm=1,n_grids

   n1 = gd(ifm)%nzp; n2 = gd(ifm)%nxp; n3 = gd(ifm)%nyp

   call coord_transforms(n1, n2, n3, gd(ifm)%nx, gd(ifm)%ny, ifm,  &
        a(itoptn(ifm)), a(itopun(ifm)), a(itopvn(ifm)), a(itopmn(ifm)),  &
        a(irtgtn(ifm)), a(irtgun(ifm)), a(irtgvn(ifm)), a(irtgmn(ifm)),  &
        a(if13un(ifm)),  &
        a(if13vn(ifm)), a(if13tn(ifm)), a(if13mn(ifm)), a(if23un(ifm)),  &
        a(if23vn(ifm)), a(if23tn(ifm)), a(if23mn(ifm)), a(idxun(ifm)),  &
        a(idxvn(ifm)), a(idxtn(ifm)), a(idxmn(ifm)), a(idyun(ifm)),  &
        a(idyvn(ifm)), a(idytn(ifm)), a(idymn(ifm)), gd(ifm)%xm,  &
        gd(ifm)%xt, gd(ifm)%ym, gd(ifm)%yt, gd(ifm)%zm, gd(ifm)%zt)

end do

! If a MAKE_VAR_FILES run, execute IPP, then return:

if (run_type == 'MAKE_VAR_FILES') then
   call PIPP_driver(nlst_fn)
   write (*, '(/,a,/)') '*** IPP complete ***'
   return
end if

call aerproc_check(.TRUE.)  ! Set model process overrides for aerosol vars

if (.NOT.this_is_a_restart) then    ! runtype = 'INITIAL_START' case

   if (initialization_type == 1) then   ! Horizontally homogeneous atm init

      call init_HH

   else if (initialization_type == 2) then   ! Fully heterogeneous atm init

      call varfile_find
      
      ! Determine master 1D reference state location/grid:
      do igd=1,n_grids
         call refstate_loc(gd(igd)%nxp, gd(igd)%nyp, igd, a(itoptn(igd)),  &
              a(irtgtn(igd)), 1, igd == 1)
      end do
      
      igd = igdref(1); call initv(igd, .TRUE., .TRUE.)

      igd = 1; call initv(igd, .TRUE., .FALSE.)
      write (*, '(/,a,i2,a,/)')  &
           'INFO: Initial varfile data input successfully (grid #', igd, ')'

   end if

  ! On all fine grids, compute the 1D and 3D reference state arrays
  !  and the prognostic atmospheric fields by interpolation:

   call fmrefs1D(2, n_grids)

   do ifm=2,n_grids
      icm = gd(ifm)%nxtnest

      if (icm >= 1) then

         call fmrefs3D(ifm)

         call prgintrp(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(icm)%nzp,  &
              gd(icm)%nxp, gd(icm)%nyp, 0, 0, ifm, 1, 0)

         if (initialization_type == 2) then
            call initv(ifm, .TRUE., .FALSE.)
            write (*, '(/,a,i2,a,/)')  &
                 'INFO: Initial varfile data input successfully (grid #',  &
                 ifm, ')'
            call vfintrp(ifm)
         end if

         call fmdn0(ifm)   ! Finish fine grid dn0

      end if

   end do

  ! Initialize past time level velocity and perturbation Exner function
  !  on all grids; possibly perturb potential temperature, modify gaseous fields
  !  and/or set user-defined tracer fields:

   do ifm=1,n_grids

      call fldinit(ifm, 1)    ! Initialize "past" fields

      ! For initial start of LES and some horizontally homogeneous runs,
      !   randomly perturb lowest level potential temperature of all grids
      !   that extend to the surface:

      if (perturb_nsfc_theta) then   ! User wants this treatment
         if (gd(ifm)%nstbot == 1) then   ! k=2 is lowest "real" level
            call LES_iperturb(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp,  &
                 a(ithpn(ifm)), 2)
            call botset(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, 1,  &
                 gd(ifm)%nxp, 1, gd(ifm)%nyp, ifm, a(ithpn(ifm)), 'T') ! Change k=1 accordingly
         end if
      end if

      ! Initialize water vapor:
      if (H2Ov_ivr > 0 .and. H2Ov_tog > 0) then
         call H2Ov_init(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, 1,  &
              gd(ifm)%nxp, 1, gd(ifm)%nyp, a(irvn(ifm)), a(idn0n(ifm)),  &
              gd(ifm)%zt, gd(ifm)%zm, a(irtgtn(ifm)))
      end if

      if (naddsc > 0) then
         ! Init tracer "mass"
         do itracerm=1,naddsc
            call tracer_init(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm,  &
                 itracerm, a(isclpn(ifm,itracerm)), a(itoptn(ifm)),  &
                 a(irtgtn(ifm)), gd(ifm)%zt, this_is_a_restart)
         end do
      end if

      ! Ensure that water vapor is positive-definite:
      call negadj(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm, 'RV')

   end do

  ! Feed back gas-phase field(s) and/or user-defined tracer field(s) from
  !  finer meshes to coarser meshes; enforce top/bottom boundary conditions:

   do ifm=n_grids,2,-1
      icm = gd(ifm)%nxtnest

      ! Operate on gas-phase field(s):
      if (H2Ov_ivr > 0 .and. H2Ov_tog > 0) then
         call fdback(a(irvn(icm)), a(irvn(ifm)), a(idn0n(icm)),  &
              a(idn0n(ifm)), gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
              gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm, 't', a(iscr1))
         if (gd(icm)%nstbot == 1) then
            call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                 gd(icm)%nxp, 1, gd(icm)%nyp, icm, a(irvn(icm)), 'T')
         end if
         if (gd(icm)%nsttop == 1) then
            call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                 gd(icm)%nxp, 1, gd(icm)%nyp, icm, a(irvn(icm)), 'S')
         end if
      end if

      ! Operate on user-defined tracer field(s):
      if (naddsc > 0) then
         do itracerm=1,naddsc
            call fdback(a(isclpn(icm,itracerm)), a(isclpn(ifm,itracerm)),  &
                 a(idn0n(icm)), a(idn0n(ifm)), gd(icm)%nzp, gd(icm)%nxp,  &
                 gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm,  &
                 't', a(iscr1))
            if (gd(icm)%nstbot == 1) then
               call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                    gd(icm)%nxp, 1, gd(icm)%nyp, icm,  &
                    a(isclpn(icm,itracerm)), 'T')
            end if
            if (gd(icm)%nsttop == 1) then
               call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                    gd(icm)%nxp, 1, gd(icm)%nyp, icm,  &
                    a(isclpn(icm,itracerm)), 'S')
            end if
         end do
      end if
   end do

  ! Gas-phase field(s) are now finished, allowing the diagnosis of the "dry"
  !  potential temperature field:

   do ifm=1,n_grids
      ! Fill THETA variable from THP:
      if (H2Ov_tog > 0) then
         call dry_thermo(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, 1,  &
              gd(ifm)%nxp, 1, gd(ifm)%nyp, a(ithpn(ifm)), a(ithetan(ifm)),  &
              a(irvn(ifm)), iaertog)
      else
         call dry_thermo(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, 1,  &
              gd(ifm)%nxp, 1, gd(ifm)%nyp, a(ithpn(ifm)), a(ithetan(ifm)),  &
              a(iscr1), iaertog)
      end if
   end do

  ! Initialize items related to radiative transfer:

   if (RT_paradigm == 2) then  ! Full RT initialization
      select case (RTc_tog)
      case (1)
         call radn_setup
      case (2,3)
         call aerad_setup(2, gd(1)%nzp, naersclr(1))
      case (4)
         call ARCnRT_setup(2, gd(1)%nzp, naersclr(1))
      end select
   else if (dust_ivr >= 2) then  ! Items needed for column opacity init calcs
      select case (RTc_tog)
      case (2,3)
         call aerad_setup(2, gd(1)%nzp, naersclr(1))
      case (4)
         call ARCnRT_setup(2, gd(1)%nzp, naersclr(1))
      end select
   end if

  ! Fill land surface data for all grids that have no standard input files
  !  (NOTE: this call is located here because of a dependency on "dry" potential
  !  temperature):

   call surface_nest(1, n_grids, nzg+nzs, npatch, 'NOFILE')

  ! Possibly modify airborne-particulate field(s); check whether all fields are
  !  relevant fields are positive-definite:

   do ifm=1,n_grids
      ! (NOTE: `aerad_setup` must be called before this):
      call aer_fldinit(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, 1, gd(ifm)%nxp,  &
           1, gd(ifm)%nyp, ifm)

      call negadj(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm, 'ALL')
   end do

  ! Feed back airborne-particulate field(s) from finer meshes to coarser
  !  meshes; enforce top/bottom boundary conditions:

   do ifm=n_grids,2,-1
      icm = gd(ifm)%nxtnest
      do iif=1,naersclr(ifm)
         do iic=1,naersclr(icm)
            if (iaersclp(iif,ifm) == iaersclp(iic,icm)) then
               call fdback(a(ind3dn(icm,iaersclp(iic,icm))),  &
                    a(ind3dn(ifm,iaersclp(iif,ifm))), a(idn0n(icm)),  &
                    a(idn0n(ifm)), gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
                    gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm, 't', a(iscr1))

               if (gd(icm)%nstbot == 1) then
                  call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                       gd(icm)%nxp, 1, gd(icm)%nyp, icm,  &
                       a(ind3dn(icm,iaersclp(iic,icm))), 'A')
               end if
               if (gd(icm)%nsttop == 1) then
                  call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                       gd(icm)%nxp, 1, gd(icm)%nyp, icm,  &
                       a(ind3dn(icm,iaersclp(iic,icm))), 'A')
               end if
            end if
         end do
      end do
   end do

  ! The below are only **dummy values**, solely to allow `switch_to_grid` within
  !  `state_output` to work correctly:

   do ifm=1,n_grids
      gd(ifm)%dt_dyn = -1._f4
      gd(ifm)%n_dta_per_dtdyn = 1
   end do

!**************************************************************************
!****  IF THIS IS A RESTART:

else           ! runtype = 'RESTART' case

   if (initialization_type == 2) call varfile_find

   do ifm=1,n_grids
      icm = gd(ifm)%nxtnest
      if (icm == 0) call refs3D(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm,  &
           a(ipi0n(ifm)), a(idn0n(ifm)), a(idn0un(ifm)), a(idn0vn(ifm)),  &
           a(ith0n(ifm)), a(itoptn(ifm)), a(irtgtn(ifm)), 1)
   end do

   do ifm = 1,min(n_grids, n_grids_st)
      icm = gd(ifm)%nxtnest
      if (icm >= 1) call fmrefs3D(ifm)

      if (initialization_type == 2) then
         call initv(ifm, .FALSE., .FALSE.)
         write (*, '(/,a,i2,a,/)')  &
              'INFO: (Restart) varfile data input successfully (grid #',  &
              ifm, ')'
         if (icm >= 1) call vfintrp(ifm)
      end if

      if (icm >= 1) call fmdn0(ifm)
!%      CALL NEGADJ1(A,nzp,nxp,nyp,1,nxp,1,nyp)
   end do

  ! If any grids are being added for this run, initialize their
  !  surface layer variables, 1-D reference state variables, and
  !  prognostic atmospheric and soil model fields:

   if (n_grids > n_grids_st) then

      write (*, '(/,a)') 'Adding new grid(s) for this restart...' 

      call fmrefs1D(n_grids_st+1, n_grids)

      do ifm = n_grids_st+1,n_grids
         icm = gd(ifm)%nxtnest
         if (icm == 0) call exception(procedure_name, FATAL,  &
              'Cannot have more than one coarsest grid.')
    
         call fmrefs3D(ifm)
         call prgintrp(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(icm)%nzp,  &
              gd(icm)%nxp, gd(icm)%nyp, 0, 0, ifm, 1, 0)
         
         if (initialization_type == 2) then
            call initv(ifm, .FALSE., .FALSE.)  ! Only fill nudging arrays
            write (*, '(/,a,i2,a,/)')  &
                 'INFO: Initial varfile data input successfully (grid #',  &
                 ifm, ')'
            call vfintrp(ifm)
         end if

         call fmdn0(ifm)
         call fldinit(ifm, 0)
 
         ! Make sure certain fields are positive definite:

         call negadj(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm, 'ALL')

         call dry_thermo(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, 1,  &
              gd(ifm)%nxp, 1, gd(ifm)%nyp, a(ithpn(ifm)), a(ithetan(ifm)),  &
              a(irvn(ifm)), iaertog)

      end do

      ! Fill land surface data for all grids that have no standard input files:

      call surface_nest(n_grids_st+1, n_grids, nzg+nzs, npatch, 'NOFILE')

   else if (n_grids < n_grids_st) then

      write (*, '(/,a)') 'Subtracting grid(s) for this restart...' 

   end if

  ! Set, feed back (from finer meshes to coarser meshes), and bound user-defined
  !  tracer field(s):

   if (naddsc > 0) then
      do ifm=1,n_grids
         ! Init tracer "mass"
         do itracerm=1,naddsc
            call tracer_init(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm,  &
                 itracerm, a(isclpn(ifm,itracerm)), a(itoptn(ifm)),  &
                 a(irtgtn(ifm)), gd(ifm)%zt, this_is_a_restart)
         end do
      end do

      do ifm=n_grids,2,-1
         icm = gd(ifm)%nxtnest
         ! Feed back user-defined tracer field(s) from a finer mesh to a coarser
         !  mesh; enforce top/bottom boundary conditions:
         do itracerm=1,naddsc
            call fdback(a(isclpn(icm,itracerm)), a(isclpn(ifm,itracerm)),  &
                 a(idn0n(icm)), a(idn0n(ifm)), gd(icm)%nzp, gd(icm)%nxp,  &
                 gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, ifm,  &
                 't', a(iscr1))
            if (gd(icm)%nstbot == 1) then
               call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                    gd(icm)%nxp, 1, gd(icm)%nyp, icm,  &
                    a(isclpn(icm,itracerm)), 'T')
            end if
            if (gd(icm)%nsttop == 1) then
               call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                    gd(icm)%nxp, 1, gd(icm)%nyp, icm,  &
                    a(isclpn(icm,itracerm)), 'S')
            end if
         end do
      end do
   end if

  ! Initialize items related to radiative transfer:

   if (RT_paradigm == 2) then  ! Full RT initialization
      select case (RTc_tog)
      case (1)
         call radn_setup
      case (2,3)
         call aerad_setup(2, gd(1)%nzp, naersclr(1))
      case (4)
         call ARCnRT_setup(2, gd(1)%nzp, naersclr(1))
      end select
   end if

end if

! Check configured model options, etc.:

there_is_an_error = .FALSE.
call chkcfg3
if (there_is_an_error) call exception(procedure_name, FATAL, 'chkcfg3')

! Save initial fields in state file(s), ensuring that the recipient
!  directories exist:

state_output_fpfx = trim(output_dir_root_w_sim_ID)//'/state/'//  &
     trim(simulation_ID_w_sfx)
i = index(state_output_fpfx, '/', back=.TRUE.)-1  ! Determine dir name length
call mkdir_p(state_output_fpfx(1:i))  ! Create any needed directories

call state_output(.FALSE., this_is_a_restart)

! Display 1D reference state and/or initial sounding values:

call sndref_print(this_is_a_restart)

! Alert user to time/Ls used:

call get_current_sas_vals(orbital_info_table, current_sim_ET, tmp1_f8,  &
     tmp2_f8, tmp3_f8, tmp3_f8, tmp3_f8)
write (*, '(/,/,a,f8.4,a,f7.4,/)')  &
     '****** PRAMS start Ls = ', tmp2_f8, ', and UT = ', tmp1_f8

end subroutine initialize


!```````````````````````````````````````````````````````````````````
subroutine aer_fldinit(mzp, mxp, myp, ia, iz, ja, jz, i_grid)

!+ Initializes aerosol fields according to the namelist configuration

use main_memory_mod, only: initialization_type, a, gd
use main_memory_bspec_mod, only: dust_tog, dust_nb, dust_ivr
use vtables_mod, only: idustncn, irtgtn, ipi0n, ippn, ithpn, ivt3da, ivt3db
use aerad_interface_mod, only: aerad_get_pt

implicit none

!=== Argument declarations:

integer, intent(IN) :: mzp, mxp, myp, ia, iz, ja, jz, i_grid

!== Local declarations:

! NONE

!=== Executable statements:

if (initialization_type == 1) then ! Horizontally homogeneous atm initialization

   ! Initialize dust:      
   if (dust_ivr > 0 .and. dust_tog > 0) then

      ! Calculate atmospheric temp (K) and pressure (Pa)
      !  (vt3da holds T, vt3db holds P):
      call aerad_get_pt(mzp, mxp, myp, 1, mxp, 1, myp, a(ippn(i_grid)),  &
           a(ipi0n(i_grid)), a(ithpn(i_grid)), a(ivt3da), a(ivt3db))

      call dust_init(mzp, mxp, myp, ia, iz, ja, jz, dust_nb,  &
           a(idustncn(i_grid, 1)), a(ivt3da), a(ivt3db), gd(i_grid)%zt,  &
           a(irtgtn(i_grid)), i_grid)

   end if

else if (initialization_type == 2) then   ! Variable atm initialization
      
   ! Possibly initialize dust:
   if (dust_ivr > 0 .and. dust_tog > 0) then

      ! Calculate atmospheric temp (K) and pressure (Pa)
      !  (vt3da holds T, vt3db holds P):
      call aerad_get_pt(mzp, mxp, myp, 1, mxp, 1, myp, a(ippn(i_grid)),  &
           a(ipi0n(i_grid)), a(ithpn(i_grid)), a(ivt3da), a(ivt3db))

      call dust_init(mzp, mxp, myp, ia, iz, ja, jz, dust_nb,  &
           a(idustncn(i_grid, 1)), a(ivt3da), a(ivt3db), gd(i_grid)%zt,  &
           a(irtgtn(i_grid)), i_grid)

   end if

end if
   
end subroutine aer_fldinit


!```````````````````````````````````````````````````````````````````
subroutine fldinit(igd, initflg)

!+ Fills the "past" time level arrays from the "current" ones

use univ_math_constants_mod, only: ZERO_f4
use main_memory_mod, only: a, gd
use vtables_mod, only: iucn, iupn, ivcn, ivpn, iwcn, iwpn, ipcn, ippn, itkepn
use turbulent_K_coeffs_mod, only: tkeinit

implicit none

!=== Argument declarations:

integer, intent(IN) :: initflg, igd

!== Local declarations:

integer :: n1, n2, n3

!=== Executable statements:

! Finish initializing past time level variables:

a(iupn(igd):iupn(igd)+gd(igd)%nxyzp-1) = a(iucn(igd):iucn(igd)+gd(igd)%nxyzp-1)
a(ivpn(igd):ivpn(igd)+gd(igd)%nxyzp-1) = a(ivcn(igd):ivcn(igd)+gd(igd)%nxyzp-1)
a(iwpn(igd):iwpn(igd)+gd(igd)%nxyzp-1) = a(iwcn(igd):iwcn(igd)+gd(igd)%nxyzp-1)
a(ippn(igd):ippn(igd)+gd(igd)%nxyzp-1) = a(ipcn(igd):ipcn(igd)+gd(igd)%nxyzp-1)

if (initflg == 1) then
   a(iwpn(igd):iwpn(igd)+gd(igd)%nxyzp-1) = ZERO_f4   ! Initialize vertical velocity
end if

n1 = gd(igd)%nzp; n2 = gd(igd)%nxp; n3 = gd(igd)%nyp

if (gd(igd)%idiffk == 1 .or. gd(igd)%idiffk == 4 .or.  &
     gd(igd)%idiffk == 5 .or. gd(igd)%idiffk == 7 .or. gd(igd)%idiffk == 9) then
   call tkeinit(n1, n2, n3, a(itkepn(igd)))   ! Init TKE
end if

! Set boundary values for velocities:

call dumset(n1, n2, n3, 1, n2, 1, n3, 15, a(iucn(igd)), 'U')
call dumset(n1, n2, n3, 1, n2, 1, n3, 15, a(ivcn(igd)), 'V')
call dumset(n1, n2, n3, 1, n2, 1, n3, 15, a(iwcn(igd)), 'W')
call dumset(n1, n2, n3, 1, n2, 1, n3, 15, a(iupn(igd)), 'U')
call dumset(n1, n2, n3, 1, n2, 1, n3, 15, a(ivpn(igd)), 'V')
call dumset(n1, n2, n3, 1, n2, 1, n3, 15, a(iwpn(igd)), 'W')

end subroutine fldinit
