!```````````````````````````````````````````````````````````````````
module interpolation_mod

!~ 07.10.2008 (last modified; TIM)

!+ Module that contains interpolation routines

implicit none

private

public :: zV_interp

contains


!```````````````````````````````````````````````````````````````````
subroutine zV_interp(nzz1,vctra,eleva,nzz2,vctrb,elevb,topo)

!+ Interpolates a point vertically in Z (replacement/enhancement of HTINT)

!~ 16.03.2012 (last modified; TIM)

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: undef_value_f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: nzz1,nzz2
real(f4), dimension(nzz1), intent(IN) :: vctra,eleva
real(f4), dimension(nzz2), intent(IN) :: elevb
real(f4), intent(IN) :: topo

real(f4), dimension(nzz2), intent(OUT) :: vctrb

!== Local declarations:

integer :: q,k,kk
real(f4) :: wt

!=== Executable statements:

!do k=1,nzz1
!   print *,'k1',k,eleva(k)
!enddo
!do k=1,nzz2
!   print *,'k2',k,elevb(k)
!enddo

q = 1
do k=1,nzz2
!   print *,k,'beg'

   if(elevb(k) <= topo) then  ! Below ground; just assign an undefined value
      vctrb(k) = undef_value_f4
!      print *,k,'BGL',elevb(k),topo
   else     ! Interpolate
      if(elevb(k) > eleva(nzz1)) then
         wt = (elevb(k)-eleva(nzz1))/(eleva(nzz1-1)-eleva(nzz1))
         vctrb(k) = vctra(nzz1)+(vctra(nzz1-1)-vctra(nzz1))*wt
!         print *,k,'interp',elevb(k),eleva(nzz1),topo
      else
         do while((.not.(elevb(k) < eleva(1))) .and.  &
              (.not.(elevb(k) >= eleva(q) .and. elevb(k) <= eleva(q+1))))
!            print *,k,'interpwl',elevb(k),q,eleva(q),topo
            q = q+1
            if(q == nzz1) then
               print *,'zV_interp:  ERROR - Cannot inter/extra-polate'
               do kk=1,q
                  print *,'k,eleva(k),elevb(k)',eleva(kk),elevb(kk)
               enddo
               stop
            endif
         enddo
!         print *,'through'
         wt = (elevb(k)-eleva(q))/(eleva(q+1)-eleva(q))
         vctrb(k) = vctra(q)+(vctra(q+1)-vctra(q))*wt
      endif
   endif

enddo

end subroutine zV_interp


end module interpolation_mod
