!```````````````````````````````````````````````````````````````````
module main_memory_mod

!+ 

use univ_kind_defs_mod, only: f4, f8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_NAME_LEN,  &
     MAX_FILE_PATH_LEN
use univ_integer_time_mod, only: time_obj
use namelist_maxdims_mod, only: nml__strtoklen_max, nml_fn_strtoklen_max
use simple_astron_scheme_mod,  only: sas_table_obj

implicit none

type trace_coord_obj
   integer :: igd, ix, iy
end type trace_coord_obj

!== Grid-INDEPENDENT namelist parameters, sorted by namelist block:

! &static_parameters :
integer :: impl, ntopsmth, izflat, iadvl, iadvf

! &basic_control :
integer :: initialization_type, n_grids, map_projection
character(len=nml__strtoklen_max) :: run_type
character(len=nml_fn_strtoklen_max) :: simulation_description

! &time_config :
real(f4) :: target_Ls, start_UT, dt_dynamics, min_ssfrac,  &
     max_dt_dynamics, dt_max_frac_increase_g1, radfrq
real(f8) :: maximum_sim_ET ! Elapsed sim time [s] that model should stop
integer :: dt_mode, n_dta_per_dtdyn, n_substeps_max,  &
     n_dtdyn_per_state_copy
logical :: orbital_progression, diurnal_progression

! &spatial_config :
integer ::  nzg, nzs, nestz1
real(f4) :: deltax_g1, deltay_g1, deltaz_g1, dzrat, dzmax, polelat,  &
     polelon
real(f4), allocatable, dimension(:) :: zz
integer, allocatable, dimension(:) :: nstratz1

! &inout_config :
character(len=nml_fn_strtoklen_max) :: output_dir_root, pnLES_fn,  &
     input_varfile_pfx, input_state_file_pfx, input_sfcfile_pfx,  &
     default_input_data_dir
character(len=nml__strtoklen_max) :: simulation_ID, run_suffix,  &
     simulation_ID_w_sfx, output_dir_root_w_sim_ID
integer :: ioutput, nudlat, pnLES_np
logical :: stop_if_file_exists, preserve_dt_on_restart,  &
     run_sfcinit_user, run_toptinit_user
real(f4) :: wfreq_state_file, tnudlat, tnudcent, tnudtop, znudtop, z0fact
character(len=nml__strtoklen_max), allocatable, dimension(:) :: pnLES_pt

! &model_options :
integer :: naddsc, icorflg, ibnd, jbnd, lsflg, nfpt, islastog, npatch,  &
     ihorgrad, RT_paradigm
real(f4) :: cphas, distim, zrough, albedo, TI_val, emissivity
logical :: perturb_nsfc_theta, enable_advection, prognose_uvwp,  &
     enable_turb_diffusion, prognose_sbps, enable_aerosol_microphysics

! &PIPP_control :
integer :: nigrids, PIPP_nvf, nf_span
logical :: gen_vf_input_comparison
character(len=nml__strtoklen_max) :: guess1st
character(len=nml_fn_strtoklen_max) :: varpfx
real(f4) :: vf_Ls, tgt_time, tgt_tlon
character(len=nml_fn_strtoklen_max), allocatable, dimension(:) :: iapr

!== Non-namelist grid-INDEPENDENT variables and control parameters:

type(sas_table_obj), save :: orbital_info_table

   ! "Standard" undefined (folating-point) values and corresponding
   !  floating-point comparison/check values:
real(f4), parameter :: undef_value_f4 = 1.e34_f4,  &
                       check_undef_val_f4 = undef_value_f4*0.99_f4
real(f8), parameter :: undef_value_f8 = 1.e34_f8,  &
                       check_undef_val_f8 = undef_value_f8*0.99_f8

   ! Default ticks/second:
integer, parameter :: module_tps = 1000000000,  &
                              module_tps_log10 = 9  ! log10(module_tps)

type(time_obj) :: mnsdy_Time, morbpd_Time, zero_Time

integer, parameter :: n_Cn_records = 2, n_isched_entries = 5,  &
     n_vf_refstates = 1
real(f4), parameter :: Cn_warning_threshold = 0.90_f4,  &
                       Cn_violation_threshold = 0.91_f4

integer :: maxnxp, maxnyp, maxnzp, maxnnp, i_grid, pbnlen,  &
     n_vars, iscr1, iscr2, memsizeA, memsizeB, jdim, itopo,  &
     sfw_call, ngridsv, n_grids_v, n_varf, n_varftot, tstep,  &
     isstp, n_substeps_max_curr, nsubs, n_isched_steps, nscl, naerscl
real(f4) :: sspct, dts, dtlv, ztop, substep_austerity_level, Ls_of_RT_call,  &
     UT_of_RT_call, EOT_of_RT_call
logical :: there_is_an_error, run_is_purely_2D, terminate_run_now
character(len=MAX_PROC_NAME_LEN) :: pbanner
character(len=MAX_PROC_NAME_LEN) :: PRAMS_version
character(len=MAX_FILE_PATH_LEN) :: state_output_fpfx
real(f8) ::  &
     start_sim_ET,  &  ! Sim elapsed time [s] when *run* started
     current_sim_ET,  & ! Sim time [s] elapsed since *simulation* start
     begin_dtdyn_sim_ET,  & ! Sim elapsed time [s] when latest grid #1 dt began
     epsilon_sim_ET  ! Tiny time increment [s] for comparison of sim ET values

real(f4), allocatable, dimension(:) :: a, b   ! {?}

character(len=MAX_FILE_PATH_LEN), allocatable, dimension(:) ::  &
     varfile_name  ! {n_varftot}
real(f4), allocatable, dimension(:) :: vtime   ! {n_varftot}

real(f4), allocatable, dimension(:) :: vctr1, vctr2,  &  ! {maxnnp}
     vctr3, vctr4, vctr5, vctr6, vctr7, vctr8, vctr9, vctr10, vctr11, vctr12,  &
     vctr13, vctr14, vctr15, vctr16, vctr17, vctr18, vctr19, vctr20, vctr21,  &
     vctr22, vctr23, vctr24, vctr25

real(f4), allocatable, dimension(:) :: sbps_Tlev  ! {nzg+nzs}

integer, allocatable, dimension(:,:) ::  &
     isched  ! {n_substeps_max**2,n_isched_entries}

real(f4), allocatable, dimension(:) :: topref, rtgref     ! {n_vf_refstates}
integer, allocatable, dimension(:) :: iref, jref, igdref    ! {n_vf_refstates}

!== Parallel-computing grid-INDEPENDENT variables and control parameters:

!------------------ used by: root PE *and* leaf PE ---------------------------

logical :: run_is_parallel

integer :: my_rank, my_ID, rootPE_rank, n_PE_total, n_PE_leaf

integer :: mxp, myp, mzp, ia, iz, ja, jz, i0, j0, ibndc,  &
     ia_1, ia_2, ia_3, ia1, ia2, ia3, ja_1, ja_2, ja_3, ja1, ja2, ja3,  &
     iz_1, iz_2, iz_3, iz1, iz2, iz3, jz_1, jz_2, jz_3, jz1, jz2, jz3,  &
     izu, jzv, maxMxp, maxMyp, maxMzp

integer, allocatable, dimension(:) :: PE_rank     ! {n_PE_leaf}

integer, parameter :: n_comm_path_entries_max = 5
integer, parameter :: n_comm_path_optypes_max = 7,  &
     i_long_dt_lBC_comm = 1,  &   ! Long timestep lateral boundary conditions
     i_ac_dt_u_lBC_comm = 2,  &   ! Acoustic timestep lateral BCs for u
     i_ac_dt_v_lBC_comm = 3,  &   ! Acoustic timestep lateral BCs for v
     i_ac_dt_pp_lBC_comm = 4,  &    ! Acoustic timestep lateral BCs for pp
     i_c_to_f_grid_comm = 5,  &   ! Coarse-to-fine (downscale) grid feedback
     i_f_to_c_grid_comm = 6,  &  ! Fine-to-coarse (upscale) grid feedback (comm)
     i_f_to_c_grid_info = 7      ! Fine-to-coarse (upscale) grid feedback (info)

integer, parameter :: n_cyclic_entries = 8
integer :: n_pts_cyclic
integer, allocatable, dimension(:,:) :: cyclic_paths_t,  &
     cyclic_paths_u, cyclic_paths_v  ! {n_cyclic_entries,n_pts_cyclic}

!------------------ root_PE-only variables -----------------------------

! NONE, currently

!------------------ leaf_PE-only variables -----------------------------

integer, allocatable, dimension(:) :: ivar_ID, ivar_type ! {n_vars}

integer :: dnscale_nBC_sd_max_pB, upscale_nBC_sd_max_pB
integer, allocatable, dimension(:) :: dnscale_nBC_packed_pos,  &
     upscale_nBC_packed_pos            ! {n_PE_leaf}

integer, parameter :: n_acoustic_lBC_optypes = 4

integer :: acoustic_lBC_sd_max_pB
integer, allocatable, dimension(:) ::  &
     acoustic_lBC_packed_pos            ! {n_PE_leaf}

integer, allocatable, dimension(:) ::  &
     longdt_lBC_packed_pos            ! {n_PE_leaf}

!== Grid-DEPENDENT variables and control parameters:

type grid_obj_type
   !== Grid-dependent namelist parameters, sorted by namelist block:

   ! &time_config :
   integer :: n_dt_per_pgdt

   ! &spatial_config :
   integer :: nxp, nyp, nzp, nxtnest, nstratx, nstraty, nndtrat, ninest,  &
        njnest, nknest, nsttop, nstbot
   real(f4) :: centlat, centlon, gridu, gridv

   ! &inout_config :
   integer :: itoptflg, itopsflg, iz0flg, itinflg, ialbflg
   real(f4) :: toptenh, toptwvlx, toptwvly, tinwvlx, tinwvly, albwvlx,  &
        albwvly, z0max, z0_of_topolevel, topolevel_for_z0
   character(len=nml_fn_strtoklen_max) :: itoptfn, itinfn, ialbfn
   logical :: grid_in_state_outp

   ! &model_options :
   integer :: nofilflg, idiffk
   real(f4) :: csx, csz, xkhkm, zkhkm, akmin

   !== Non-namelist grid-dependent variables and control parameters:

   real(f4) :: dt_dyn, plat, plon, deltax, deltay, deltaz, dimove,  &
        djmove, Courant_number_xyz, ss_over_dxy, dxyt_max, dt_sbps,  &
        dt_sbps_estimate
   logical :: grid_is_2D
   integer :: nx, nx1, nx2, ny, ny1, ny2, nz, nz1, nxyzp, nxyp, nxys3p,  &
        nxys2p, nxyzorsp, ngbegun, n_dta_per_dtdyn, RT_call, n_iter_sbps

   real(f4), dimension(n_Cn_records) :: Cn_value_record, Cn_change_record
   logical, dimension(n_Cn_records) :: timestep_OK 

   real(f4), pointer, dimension(:) :: ht => NULL(), ht2 => NULL(),  &
        ht4 => NULL(), hw => NULL(), hw2 => NULL(), hw4 => NULL(),  & ! {maxnzp}
        zt => NULL(), zm => NULL(), dzm => NULL(), dzt => NULL(),  &
        dzm2 => NULL(), dzt2 => NULL(), ek1 => NULL(), ek2 => NULL(),  &
        ek3 => NULL(), ek4 => NULL(), ek5 => NULL(), ek6 => NULL(),  &
        ek7 => NULL(), th01d => NULL(), pi01d => NULL(), u01d => NULL(),  &
        v01d => NULL(), dn01d => NULL()
   integer, pointer, dimension(:) :: nrz => NULL(), kpm => NULL()  ! {maxnzp}
   real(f4), pointer, dimension(:,:) :: fbcf => NULL()  ! {maxnzp,4}

   real(f4), pointer, dimension(:) :: xt => NULL(), xm => NULL(),  &  ! {maxnxp}
        ei1 => NULL(), ei2 => NULL(), ei3 => NULL(), ei4 => NULL(),  &
        ei5 => NULL(), ei6 => NULL(), ei7 => NULL()  
   integer, pointer, dimension(:) :: ipm => NULL()  ! {maxnxp}

   real(f4), pointer, dimension(:) :: yt => NULL(), ym => NULL(),  &  ! {maxnyp}
        ej1 => NULL(), ej2 => NULL(), ej3 => NULL(), ej4 => NULL(),  &
        ej5 => NULL(), ej6 => NULL(), ej7 => NULL()
   integer, pointer, dimension(:) :: jpm => NULL()  ! {maxnyp}

   logical, pointer, dimension(:,:) :: Cn_mask => NULL()    ! {maxnxp,maxnyp}

   integer, pointer, dimension(:,:) :: idx_mask_xy => NULL()   ! {x,y}

   ! Normal 3D atmospheric arrays {z,x,y}:
   real(f4), pointer, dimension(:) :: up_backup => NULL(),  &
        uc_backup => NULL(), vp_backup => NULL(), vc_backup => NULL(),  &
        wp_backup => NULL(), wc_backup => NULL(), pp_backup => NULL(),  &
        pc_backup => NULL(), thp_backup => NULL(), theta_backup => NULL(),  &
        tkep_backup => NULL(), fthrd_backup => NULL(), hkm_backup => NULL(),  &
        vkm_backup => NULL(), vkh_backup => NULL()

   ! Generic scalar 3D atmospheric arrays {z,x,y,scalar_idx}:
   real(f4), pointer, dimension(:) :: sclp_backup => NULL()

   ! Normal 2D "atmospheric" arrays {x,y}:
   real(f4), pointer, dimension(:) :: uw_backup => NULL(),  &
        vw_backup => NULL(), wfz_backup => NULL(), tfz_backup => NULL(),  &
        rshort_backup => NULL(), rlong_backup => NULL(),  &
        rlongup_backup => NULL(), albedt_backup => NULL(), cosz_backup => NULL()
   real(f4), pointer, dimension(:,:) :: bottom_other_sflux_down => NULL(),  &
        bottom_other_sflux_down_backup => NULL(),  &
        bottom_other_sflux_up => NULL(),  &
        bottom_other_sflux_up_backup => NULL(), bottom_SWflux_down => NULL(),  &
        bottom_SWflux_down_backup => NULL(), bottom_SWflux_up => NULL(),  &
        bottom_SWflux_up_backup => NULL(), emisst => NULL(),  &
        emisst_backup => NULL(), topo_factor_sflux => NULL(),  &
        topo_factor_sflux_backup => NULL()

   ! Normal 3D sub/super-surface arrays {z,x,y,patch}:
   real(f4), pointer, dimension(:) :: sbpsdens_backup => NULL(),  &
        sbpstmcp_backup => NULL(), sbpstmcd_backup => NULL(),  &
        sbpslthk_backup => NULL(), sbpsq_backup => NULL()
   
   ! Normal 2D sub/super-surface arrays {x,y,patch}:
   real(f4), pointer, dimension(:) :: patchfcv_backup => NULL(),  &
        z0net_backup => NULL(), sfctempk_backup => NULL()
   real(f4), pointer, dimension(:,:,:) :: emissivity_static => NULL()

   ! Topographic shadowing 3D arrays {x,y,azm}:
   real(f4), pointer, dimension(:,:,:) :: cosz_horizon => NULL(),  &
        TCx_obstruct => NULL(), TCz_obstruct => NULL()

   !== Parallel-computing grid-dependent variables and control parameters:
   
   !------------------ used by: root PE *and* leaf PE ------------------------

   ! {n_PE_leaf}:
   integer, pointer, dimension(:) :: leafPE_ia => NULL(),  &
        leafPE_iz => NULL(), leafPE_ja => NULL(), leafPE_jz => NULL(),  &
        leafPE_i0 => NULL(), leafPE_j0 => NULL(), leafPE_ibndc => NULL(),  &
        leafPE_mxp => NULL(), leafPE_myp => NULL(), leafPE_mzp => NULL()

   integer :: mmxp, mmyp, mmzp, mia, miz, mja, mjz, mi0, mj0, mibndc

   !------------------ root_PE-only variables --------------------------------

   ! {n_comm_path_entries_max,n_comm_path_optypes_max,n_PE_leaf,n_PE_leaf} orig:{5,7,n_grids,n_PE_leaf,n_PE_leaf}
   integer, pointer, dimension(:,:,:,:) :: tx_paths_all => NULL()

   ! {n_comm_path_optypes_max,n_PE_leaf,n_PE_leaf} orig:{6,n_grids,n_PE_leaf,n_PE_leaf}
   integer, pointer, dimension(:,:,:) :: rx_paths_all => NULL()

   ! {n_PE_leaf}
   integer, pointer, dimension(:) :: leafPE_xb => NULL(),  &
        leafPE_xe => NULL(), leafPE_yb => NULL(), leafPE_ye => NULL(),  &
        leafPE_xy => NULL(), ixb => NULL(), ixe => NULL(), iyb => NULL(),  &
        iye => NULL()

   !------------------ leaf_PE-only variables --------------------------------

   ! {n_comm_path_entries_max,n_comm_path_optypes_max,n_PE_leaf} orig:{5,7,n_grids,n_PE_leaf}
   integer, pointer, dimension(:,:,:) :: tx_paths_leaf => NULL()

   ! {n_comm_path_optypes_max,n_PE_leaf} orig:{6,n_grids,n_PE_leaf}
   integer, pointer, dimension(:,:) :: rx_paths_leaf => NULL()

   ! {n_grids,n_PE_leaf} orig:{n_grids,n_grids,n_PE_leaf}
   integer, pointer, dimension(:,:) :: dnscale_nBC_sd_packed_B => NULL(),  &
        upscale_nBC_sd_packed_B => NULL()

   ! {n_acoustic_lBC_optypes,n_PE_leaf} orig:{n_acoustic_lBC_optypes,n_grids,n_PE_leaf}
   integer, pointer, dimension(:,:) :: acoustic_lBC_sd_packed_B => NULL()

   ! {n_PE_leaf} orig:{n_grids,n_PE_leaf}
   integer, pointer, dimension(:) :: longdt_lBC_sd_packed_B => NULL()

end type grid_obj_type

type(grid_obj_type), allocatable, target, dimension(:) :: gd
type(grid_obj_type), pointer :: c_gd => NULL()


contains


!```````````````````````````````````````````````````````````````````
subroutine manage_nml_var_memory(dim_n_grids, dim_maxnzp, dim_pnLES_np,  &
     dim_nf_span)

!+ Allocates or deallocates namelist-related arrays with various dimensions

use univ_IO_mem_status_mod, only: illegal_alloc_stop
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

logical, optional, intent(IN) :: dim_n_grids, dim_maxnzp, dim_pnLES_np,  &
     dim_nf_span

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'manage_nml_var_memory'

integer :: a_err, igd

!=== Executable statements:

if (present(dim_n_grids)) then
   if (dim_n_grids) then
      allocate(gd(n_grids), stat=a_err)
   else
      deallocate(gd, stat=a_err)
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_n_grids, procedure_name,  &
        'dim_n_grids')

   ! Nullify all component pointer elements of each grid object:
   if (dim_n_grids) then
      do igd=1,n_grids
         call manage_grid_object(gd(igd), nullify_pointers=.TRUE.)
      end do
   end if
end if

if (present(dim_maxnzp)) then
   if (dim_maxnzp) then
      allocate(nstratz1(maxnzp), zz(maxnzp), stat=a_err)  
   else
      deallocate(nstratz1, zz, stat=a_err)
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_maxnzp, procedure_name,  &
        'dim_maxnzp')
end if

if (present(dim_pnLES_np)) then
   if (dim_pnLES_np) then
      allocate(pnLES_pt(pnLES_np), stat=a_err)  
   else
      deallocate(pnLES_pt, stat=a_err)
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_pnLES_np, procedure_name,  &
        'dim_pnLES_np')
end if

if (present(dim_nf_span)) then
   if (dim_nf_span) then
      allocate(iapr(nf_span), stat=a_err)  
   else
      deallocate(iapr, stat=a_err)
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_nf_span, procedure_name,  &
        'dim_nf_span')
end if

end subroutine manage_nml_var_memory


!```````````````````````````````````````````````````````````````````
subroutine manage_misc_memory(dim_n_varftot, dim_maxnnp, dim_sbps,  &
     dim_isched, dim_vf_refs)

!+ Allocates or deallocates miscellaneous arrays with various dimensions

use univ_IO_mem_status_mod, only: illegal_alloc_stop
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

logical, optional, intent(IN) :: dim_n_varftot, dim_maxnnp, dim_sbps,  &
     dim_vf_refs, dim_isched

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'manage_misc_memory'

integer :: a_err

!=== Executable statements:

if (present(dim_n_varftot)) then
   a_err = 0   ! Initialize, in case the following if-else does not set 'a_err'
   if (dim_n_varftot) then
      allocate(varfile_name(n_varftot), vtime(n_varftot), stat=a_err) 
   else
      if (allocated(varfile_name)) deallocate(varfile_name, vtime, stat=a_err)
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_n_varftot, procedure_name,  &
        'dim_n_varftot')
end if

if (present(dim_sbps)) then
   if (dim_sbps) then
      allocate(sbps_Tlev(nzg+nzs), stat=a_err) 
   else
      deallocate(sbps_Tlev, stat=a_err) 
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_sbps, procedure_name,  &
        'dim_sbps')
end if

if (present(dim_isched)) then
   if (dim_isched) then
      allocate(isched(n_isched_steps,n_isched_entries), stat=a_err) 
   else
      deallocate(isched, stat=a_err) 
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_isched, procedure_name,  &
        'dim_isched')
end if

if (present(dim_vf_refs)) then
   if (dim_vf_refs) then
      allocate(iref(n_vf_refstates), jref(n_vf_refstates),  &
           topref(n_vf_refstates), igdref(n_vf_refstates),  &
           rtgref(n_vf_refstates), stat=a_err) 
   else
      deallocate(iref, jref, topref, igdref, rtgref, stat=a_err) 
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_vf_refs, procedure_name,  &
        'dim_vf_refs')
end if

if (present(dim_maxnnp)) then
   if (dim_maxnnp) then
      allocate(vctr1(maxnnp), vctr2(maxnnp), vctr3(maxnnp), vctr4(maxnnp),  &
           vctr5(maxnnp), vctr6(maxnnp), vctr7(maxnnp), vctr8(maxnnp),  &
           vctr9(maxnnp), vctr10(maxnnp), vctr11(maxnnp), vctr12(maxnnp),  &
           vctr13(maxnnp), vctr14(maxnnp), vctr15(maxnnp), vctr16(maxnnp),  &
           vctr17(maxnnp), vctr18(maxnnp), vctr19(maxnnp), vctr20(maxnnp),  &
           vctr21(maxnnp), vctr22(maxnnp), vctr23(maxnnp), vctr24(maxnnp),  &
           vctr25(maxnnp), stat=a_err) 
   else
      deallocate(vctr1, vctr2, vctr3, vctr4, vctr5, vctr6, vctr7, vctr8,  &
           vctr9, vctr10, vctr11, vctr12, vctr13, vctr14, vctr15, vctr16,  &
           vctr17, vctr18, vctr19, vctr20, vctr21, vctr22, vctr23, vctr24,  &
           vctr25, stat=a_err) 
   end if
   if (a_err /= 0) call illegal_alloc_stop(dim_maxnnp, procedure_name,  &
        'dim_maxnnp')
end if

end subroutine manage_misc_memory


!```````````````````````````````````````````````````````````````````
subroutine manage_grid_object(o, nullify_pointers, recycle_pointers,  &
     allocate_pd_ind_pointers, allocate_pd_dep_pointers)

!+ Manages the component pointer elements of an object of type 'grid_obj_type' 

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_pointer_utils_mod, only: ptr_release
use topo_RT_effects_dims_mod, only: nazm

implicit none

!=== Argument declarations:

type(grid_obj_type), intent(INOUT) :: o  ! Object to be operated on

! Nullify ALL pointers (no allocation check)?
logical, optional, intent(IN) :: nullify_pointers

! "Recycle" (deallocate if allocated, then nullify) ALL pointers?
logical, optional, intent(IN) :: recycle_pointers

! Allocate all pointers that are INDEPENDENT of the parallel decomposition?
logical, optional, intent(IN) :: allocate_pd_ind_pointers

! Allocate all pointers that are DEPENDENT on the parallel decomposition?
logical, optional, intent(IN) :: allocate_pd_dep_pointers

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'manage_grid_object'

integer :: e, dim_z, dim_o, dim_om, dim_em, dim_ao, dim_x, dim_y, dim_g
logical :: np, rp, e_check, ai_pr, ad_pr, ai_p, ad_p, ai_pl, ad_pl, ai,  &
     ad, ad_tc

!=== Executable statements:

! * Parallel-decomposition DEPENDENT arrays change shape/size depending on
!   whether this code is executed by the root PE or a leaf PE.
! * In general, arrays allocated only by the root PE are pd-INDEPENDENT. 

! Initialize control parameters and process optional arguments:
np = .FALSE.; rp = .FALSE.; ai = .FALSE.; ad = .FALSE.
if (present(nullify_pointers)) np = nullify_pointers
if (present(recycle_pointers)) rp = recycle_pointers
if (present(allocate_pd_ind_pointers)) ai = allocate_pd_ind_pointers
if (present(allocate_pd_dep_pointers)) ad = allocate_pd_dep_pointers

! Is this a parallel run?
ai_p = ai .and. run_is_parallel
ad_p = ad .and. run_is_parallel

! Is this the root PE of a parallel run?
ai_pr = ai_p .and. my_rank == rootPE_rank
ad_pr = ad_p .and. my_rank == rootPE_rank

! Is this a leaf PE of a parallel run?
ai_pl = ai_p .and. my_rank /= rootPE_rank
ad_pl = ad_p .and. my_rank /= rootPE_rank

! Allocate timestep model state copy memory (serial root or leaf PE, only)?
ad_tc = ad .and. dt_mode < 0 .and. (.NOT.ad_p .or. ad_pl)

! Initially there are no errors:
e_check = .FALSE.; e = 0

!== {pd-INDEPENDENT} "Vertical" 1D arrays {maxnzp}:

!VARY: use correct size(s)?
if (ai) dim_z = maxnzp

if (np) nullify(o%ht); if (rp) call ptr_release(o%ht, e)
if (ai) allocate(o%ht(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ht2); if (rp) call ptr_release(o%ht2, e)
if (ai) allocate(o%ht2(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ht4); if (rp) call ptr_release(o%ht4, e)
if (ai) allocate(o%ht4(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%hw); if (rp) call ptr_release(o%hw, e)
if (ai) allocate(o%hw(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%hw2); if (rp) call ptr_release(o%hw2, e)
if (ai) allocate(o%hw2(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%hw4); if (rp) call ptr_release(o%hw4, e)
if (ai) allocate(o%hw4(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%zt); if (rp) call ptr_release(o%zt, e)
if (ai) allocate(o%zt(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%zm); if (rp) call ptr_release(o%zm, e)
if (ai) allocate(o%zm(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%dzt); if (rp) call ptr_release(o%dzt, e)
if (ai) allocate(o%dzt(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%dzm); if (rp) call ptr_release(o%dzm, e)
if (ai) allocate(o%dzm(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%dzt2); if (rp) call ptr_release(o%dzt2, e)
if (ai) allocate(o%dzt2(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%dzm2); if (rp) call ptr_release(o%dzm2, e)
if (ai) allocate(o%dzm2(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ek1); if (rp) call ptr_release(o%ek1, e)
if (ai) allocate(o%ek1(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ek2); if (rp) call ptr_release(o%ek2, e)
if (ai) allocate(o%ek2(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ek3); if (rp) call ptr_release(o%ek3, e)
if (ai) allocate(o%ek3(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ek4); if (rp) call ptr_release(o%ek4, e)
if (ai) allocate(o%ek4(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ek5); if (rp) call ptr_release(o%ek5, e)
if (ai) allocate(o%ek5(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ek6); if (rp) call ptr_release(o%ek6, e)
if (ai) allocate(o%ek6(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ek7); if (rp) call ptr_release(o%ek7, e)
if (ai) allocate(o%ek7(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%th01d); if (rp) call ptr_release(o%th01d, e)
if (ai) allocate(o%th01d(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%pi01d); if (rp) call ptr_release(o%pi01d, e)
if (ai) allocate(o%pi01d(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%u01d); if (rp) call ptr_release(o%u01d, e)
if (ai) allocate(o%u01d(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%v01d); if (rp) call ptr_release(o%v01d, e)
if (ai) allocate(o%v01d(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%dn01d); if (rp) call ptr_release(o%dn01d, e)
if (ai) allocate(o%dn01d(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%nrz); if (rp) call ptr_release(o%nrz, e)
if (ai) allocate(o%nrz(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%kpm); if (rp) call ptr_release(o%kpm, e)
if (ai) allocate(o%kpm(dim_z), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-INDEPENDENT} Vertical 1D array {maxnzp}')

!== {pd-INDEPENDENT} "Vertical" 2D array {maxnzp,4}:

!VARY: use correct size(s)?
if (ai) then
   dim_z = maxnzp
   dim_o = 4
end if

if (np) nullify(o%fbcf); if (rp) call ptr_release(o%fbcf, e)
if (ai) allocate(o%fbcf(dim_z,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-INDEPENDENT} Vertical 2D array {maxnzp,4}')

!== {pd-DEPENDENT} "x-direction" 1D array {maxnxp}:

!VARY some of these with pd?
if (ad) dim_x = maxnxp

if (np) nullify(o%xt); if (rp) call ptr_release(o%xt, e)
if (ad) allocate(o%xt(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%xm); if (rp) call ptr_release(o%xm, e)
if (ad) allocate(o%xm(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ei1); if (rp) call ptr_release(o%ei1, e)
if (ad) allocate(o%ei1(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ei2); if (rp) call ptr_release(o%ei2, e)
if (ad) allocate(o%ei2(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ei3); if (rp) call ptr_release(o%ei3, e)
if (ad) allocate(o%ei3(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ei4); if (rp) call ptr_release(o%ei4, e)
if (ad) allocate(o%ei4(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ei5); if (rp) call ptr_release(o%ei5, e)
if (ad) allocate(o%ei5(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ei6); if (rp) call ptr_release(o%ei6, e)
if (ad) allocate(o%ei6(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ei7); if (rp) call ptr_release(o%ei7, e)
if (ad) allocate(o%ei7(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ipm); if (rp) call ptr_release(o%ipm, e)
if (ad) allocate(o%ipm(dim_x), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} "x-direction" 1D array {maxnxp}')

!== {pd-DEPENDENT} "y-direction" 1D array {maxnyp}:

!VARY some of these with pd?
if (ad) dim_y = maxnyp

if (np) nullify(o%yt); if (rp) call ptr_release(o%yt, e)
if (ad) allocate(o%yt(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ym); if (rp) call ptr_release(o%ym, e)
if (ad) allocate(o%ym(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ej1); if (rp) call ptr_release(o%ej1, e)
if (ad) allocate(o%ej1(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ej2); if (rp) call ptr_release(o%ej2, e)
if (ad) allocate(o%ej2(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ej3); if (rp) call ptr_release(o%ej3, e)
if (ad) allocate(o%ej3(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ej4); if (rp) call ptr_release(o%ej4, e)
if (ad) allocate(o%ej4(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ej5); if (rp) call ptr_release(o%ej5, e)
if (ad) allocate(o%ej5(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ej6); if (rp) call ptr_release(o%ej6, e)
if (ad) allocate(o%ej6(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ej7); if (rp) call ptr_release(o%ej7, e)
if (ad) allocate(o%ej7(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%jpm); if (rp) call ptr_release(o%jpm, e)
if (ad) allocate(o%jpm(dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} "y-direction" 1D array {maxnyp}')

!== {pd-DEPENDENT} "Horizontal" 2D array {maxnxp,maxnyp}:

!VARY some of these with pd?
if (ad) then
   dim_x = maxnxp
   dim_y = maxnyp
end if

if (np) nullify(o%Cn_mask); if (rp) call ptr_release(o%Cn_mask, e)
if (ad) allocate(o%Cn_mask(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} Horizontal 2D array {maxnxp,maxnyp}')

!== {pd-DEPENDENT} "Horizontal" 2D array {x,y}:

if (ad) then
   if (my_rank == rootPE_rank) then
      dim_x = o%nxp; dim_y = o%nyp
   else
      dim_x = o%mmxp; dim_y = o%mmyp
   end if
end if

if (np) nullify(o%idx_mask_xy); if (rp) call ptr_release(o%idx_mask_xy, e)
if (ad) allocate(o%idx_mask_xy(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_other_sflux_down)
if (rp) call ptr_release(o%bottom_other_sflux_down, e)
if (ad) allocate(o%bottom_other_sflux_down(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_other_sflux_up)
if (rp) call ptr_release(o%bottom_other_sflux_up, e)
if (ad) allocate(o%bottom_other_sflux_up(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_SWflux_down)
if (rp) call ptr_release(o%bottom_SWflux_down, e)
if (ad) allocate(o%bottom_SWflux_down(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_SWflux_up)
if (rp) call ptr_release(o%bottom_SWflux_up, e)
if (ad) allocate(o%bottom_SWflux_up(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%topo_factor_sflux)
if (rp) call ptr_release(o%topo_factor_sflux, e)
if (ad) allocate(o%topo_factor_sflux(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%emisst)
if (rp) call ptr_release(o%emisst, e)
if (ad) allocate(o%emisst(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} Horizontal 2D array {x,y}')

!== {pd-DEPENDENT} State copy 3D atmospheric arrays {nxyzp}:

if (ad_tc) then
   if (my_rank == rootPE_rank) then
      dim_o = o%nxyzp
   else
      dim_o = o%mmzp*o%mmxp*o%mmyp
   end if
end if

if (np) nullify(o%up_backup); if (rp) call ptr_release(o%up_backup, e)
if (ad_tc) allocate(o%up_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%uc_backup); if (rp) call ptr_release(o%uc_backup, e)
if (ad_tc) allocate(o%uc_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%vp_backup); if (rp) call ptr_release(o%vp_backup, e)
if (ad_tc) allocate(o%vp_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%vc_backup); if (rp) call ptr_release(o%vc_backup, e)
if (ad_tc) allocate(o%vc_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%wp_backup); if (rp) call ptr_release(o%wp_backup, e)
if (ad_tc) allocate(o%wp_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%wc_backup); if (rp) call ptr_release(o%wc_backup, e)
if (ad_tc) allocate(o%wc_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%pp_backup); if (rp) call ptr_release(o%pp_backup, e)
if (ad_tc) allocate(o%pp_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%pc_backup); if (rp) call ptr_release(o%pc_backup, e)
if (ad_tc) allocate(o%pc_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%thp_backup); if (rp) call ptr_release(o%thp_backup, e)
if (ad_tc) allocate(o%thp_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%theta_backup); if (rp) call ptr_release(o%theta_backup, e)
if (ad_tc) allocate(o%theta_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%fthrd_backup); if (rp) call ptr_release(o%fthrd_backup, e)
if (ad_tc) allocate(o%fthrd_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%hkm_backup); if (rp) call ptr_release(o%hkm_backup, e)
if (ad_tc) allocate(o%hkm_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%vkm_backup); if (rp) call ptr_release(o%vkm_backup, e)
if (ad_tc) allocate(o%vkm_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%vkh_backup); if (rp) call ptr_release(o%vkh_backup, e)
if (ad_tc) allocate(o%vkh_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%tkep_backup); if (rp) call ptr_release(o%tkep_backup, e)
if (o%idiffk == 1 .or. o%idiffk == 4 .or. o%idiffk == 5 .or.  &
     o%idiffk == 7 .or. o%idiffk == 9) then
   if (ad_tc) allocate(o%tkep_backup(dim_o), stat=e)
end if
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} State copy 3D atmospheric arrays {nxyzp}')

!== {pd-DEPENDENT} State copy generic scalar 3D array {nxyzp,*}:

if (ad_tc) then
   if (my_rank == rootPE_rank) then
      dim_o = o%nxyzp*max(1, naddsc)
   else
      dim_o = o%mmzp*o%mmxp*o%mmyp*max(1, naddsc)
   end if
end if

if (np) nullify(o%sclp_backup); if (rp) call ptr_release(o%sclp_backup, e)
if (naddsc >= 1) then
   if (ad_tc) allocate(o%sclp_backup(dim_o), stat=e)
end if
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} State copy generic scalar 3D array {nxyzp,*}')

!== {pd-DEPENDENT} State copy 2D "atmospheric" arrays {x,y}:

if (ad_tc) then
   if (my_rank == rootPE_rank) then
      dim_x = o%nxp; dim_y = o%nyp
   else
      dim_x = o%mmxp; dim_y = o%mmyp
   end if
end if

if (np) nullify(o%emisst_backup); if (rp) call ptr_release(o%emisst_backup, e)
if (ad_tc) allocate(o%emisst_backup(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_other_sflux_down_backup)
if (rp) call ptr_release(o%bottom_other_sflux_down_backup, e)
if (ad_tc) allocate(o%bottom_other_sflux_down_backup(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_other_sflux_up_backup)
if (rp) call ptr_release(o%bottom_other_sflux_up_backup, e)
if (ad_tc) allocate(o%bottom_other_sflux_up_backup(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_SWflux_down_backup)
if (rp) call ptr_release(o%bottom_SWflux_down_backup, e)
if (ad_tc) allocate(o%bottom_SWflux_down_backup(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%bottom_SWflux_up_backup)
if (rp) call ptr_release(o%bottom_SWflux_up_backup, e)
if (ad_tc) allocate(o%bottom_SWflux_up_backup(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%topo_factor_sflux_backup)
if (rp) call ptr_release(o%topo_factor_sflux_backup, e)
if (ad_tc) allocate(o%topo_factor_sflux_backup(dim_x,dim_y), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} State copy 2D "atmospheric" arrays {x,y}')

!== {pd-DEPENDENT} State copy 2D "atmospheric" arrays {nxyp}:

if (ad_tc) then
   if (my_rank == rootPE_rank) then
      dim_o = o%nxyp
   else
      dim_o = o%mmxp*o%mmyp
   end if
end if

if (np) nullify(o%uw_backup); if (rp) call ptr_release(o%uw_backup, e)
if (ad_tc) allocate(o%uw_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%vw_backup); if (rp) call ptr_release(o%vw_backup, e)
if (ad_tc) allocate(o%vw_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%wfz_backup); if (rp) call ptr_release(o%wfz_backup, e)
if (ad_tc) allocate(o%wfz_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%tfz_backup); if (rp) call ptr_release(o%tfz_backup, e)
if (ad_tc) allocate(o%tfz_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%rshort_backup); if (rp) call ptr_release(o%rshort_backup, e)
if (ad_tc) allocate(o%rshort_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%rlong_backup); if (rp) call ptr_release(o%rlong_backup, e)
if (ad_tc) allocate(o%rlong_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%rlongup_backup); if (rp) call ptr_release(o%rlongup_backup, e)
if (ad_tc) allocate(o%rlongup_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%albedt_backup); if (rp) call ptr_release(o%albedt_backup, e)
if (ad_tc) allocate(o%albedt_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%cosz_backup); if (rp) call ptr_release(o%cosz_backup, e)
if (ad_tc) allocate(o%cosz_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} State copy 2D atmospheric arrays {nxyp}')

!== {pd-DEPENDENT} "Horizontal" 2D sub-/super-surface arrays {x,y,patch}:

if (ad) then
   if (my_rank == rootPE_rank) then
      dim_x = o%nxp; dim_y = o%nyp; dim_o = npatch
   else
      dim_x = o%mmxp; dim_y = o%mmyp; dim_o = npatch
   end if
end if

if (np) nullify(o%emissivity_static)
if (rp) call ptr_release(o%emissivity_static, e)
if (ad) allocate(o%emissivity_static(dim_x,dim_y,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} Horizontal 2D sub-/super-surface arrays {x,y,patch}')

!== {pd-DEPENDENT} State copy 3D sub-/super-surface arrays {nxys3p,patch}:

if (ad_tc) then
   if (my_rank == rootPE_rank) then
      dim_o = o%nxys3p
   else
      dim_o = (nzg+nzs)*o%mmxp*o%mmyp*npatch
   end if
end if

if (np) nullify(o%sbpsdens_backup); if (rp) call ptr_release(o%sbpsdens_backup, e)
if (ad_tc) allocate(o%sbpsdens_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%sbpstmcp_backup); if (rp) call ptr_release(o%sbpstmcp_backup, e)
if (ad_tc) allocate(o%sbpstmcp_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%sbpstmcd_backup); if (rp) call ptr_release(o%sbpstmcd_backup, e)
if (ad_tc) allocate(o%sbpstmcd_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%sbpslthk_backup); if (rp) call ptr_release(o%sbpslthk_backup, e)
if (ad_tc) allocate(o%sbpslthk_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%sbpsq_backup); if (rp) call ptr_release(o%sbpsq_backup, e)
if (ad_tc) allocate(o%sbpsq_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} State copy 3D sub-/super-surface arrays {nxys3p,patch}')

!== {pd-DEPENDENT} State copy 2D sub-/super-surface arrays {nxys2p,patch}:

if (ad_tc) then
   if (my_rank == rootPE_rank) then
      dim_o = o%nxys2p
   else
      dim_o = o%mmxp*o%mmyp*npatch
   end if
end if

if (np) nullify(o%patchfcv_backup); if (rp) call ptr_release(o%patchfcv_backup, e)
if (ad_tc) allocate(o%patchfcv_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%z0net_backup); if (rp) call ptr_release(o%z0net_backup, e)
if (ad_tc) allocate(o%z0net_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%sfctempk_backup); if (rp) call ptr_release(o%sfctempk_backup, e)
if (ad_tc) allocate(o%sfctempk_backup(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} State copy 2D sub-/super-surface arrays {nxys2p,patch}')

!== {pd-DEPENDENT} Topographic shadowing 3D arrays {*xp,*yp,nazm}:

if (ad_pl) then    ! Leaf PE
   dim_x = o%mmxp
   dim_y = o%mmyp
else if (ad) then  ! Root PE
   dim_x = o%nxp
   dim_y = o%nyp
end if
if (ad) dim_o = nazm

if (np) nullify(o%cosz_horizon); if (rp) call ptr_release(o%cosz_horizon, e)
if (ad) allocate(o%cosz_horizon(dim_x, dim_y, dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%TCx_obstruct); if (rp) call ptr_release(o%TCx_obstruct, e)
if (ad) allocate(o%TCx_obstruct(dim_x, dim_y, dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%TCz_obstruct); if (rp) call ptr_release(o%TCz_obstruct, e)
if (ad) allocate(o%TCz_obstruct(dim_x, dim_y, dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-DEPENDENT} Topographic shadowing 3D arrays {*xp,*yp,nazm}')

!== {pd-INDEPENDENT} Parallel control parameters (1D arrays) {n_PE_leaf}:

if (ai_p) dim_o = n_PE_leaf

if (np) nullify(o%leafPE_ia); if (rp) call ptr_release(o%leafPE_ia, e)
if (ai_p) allocate(o%leafPE_ia(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_iz); if (rp) call ptr_release(o%leafPE_iz, e)
if (ai_p) allocate(o%leafPE_iz(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_ja); if (rp) call ptr_release(o%leafPE_ja, e)
if (ai_p) allocate(o%leafPE_ja(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_jz); if (rp) call ptr_release(o%leafPE_jz, e)
if (ai_p) allocate(o%leafPE_jz(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_i0); if (rp) call ptr_release(o%leafPE_i0, e)
if (ai_p) allocate(o%leafPE_i0(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_j0); if (rp) call ptr_release(o%leafPE_j0, e)
if (ai_p) allocate(o%leafPE_j0(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_ibndc); if (rp) call ptr_release(o%leafPE_ibndc, e)
if (ai_p) allocate(o%leafPE_ibndc(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_mxp); if (rp) call ptr_release(o%leafPE_mxp, e)
if (ai_p) allocate(o%leafPE_mxp(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_myp); if (rp) call ptr_release(o%leafPE_myp, e)
if (ai_p) allocate(o%leafPE_myp(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_mzp); if (rp) call ptr_release(o%leafPE_mzp, e)
if (ai_p) allocate(o%leafPE_mzp(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-INDEPENDENT} Parallel control parameters (1D arrays) {n_PE_leaf}')

!== {pd-INDEPENDENT} Root PE parallel control parameter arrays {[*],n_PE_leaf}:

if (ai_pr) dim_o = n_PE_leaf

if (np) nullify(o%leafPE_xb); if (rp) call ptr_release(o%leafPE_xb, e)
if (ai_pr) allocate(o%leafPE_xb(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_xe); if (rp) call ptr_release(o%leafPE_xe, e)
if (ai_pr) allocate(o%leafPE_xe(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_yb); if (rp) call ptr_release(o%leafPE_yb, e)
if (ai_pr) allocate(o%leafPE_yb(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_ye); if (rp) call ptr_release(o%leafPE_ye, e)
if (ai_pr) allocate(o%leafPE_ye(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%leafPE_xy); if (rp) call ptr_release(o%leafPE_xy, e)
if (ai_pr) allocate(o%leafPE_xy(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ixb); if (rp) call ptr_release(o%ixb, e)
if (ai_pr) allocate(o%ixb(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%ixe); if (rp) call ptr_release(o%ixe, e)
if (ai_pr) allocate(o%ixe(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%iyb); if (rp) call ptr_release(o%iyb, e)
if (ai_pr) allocate(o%iyb(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%iye); if (rp) call ptr_release(o%iye, e)
if (ai_pr) allocate(o%iye(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (ai_pr) then
   dim_om = n_comm_path_optypes_max
   dim_em = n_comm_path_entries_max
end if

if (np) nullify(o%tx_paths_all); if (rp) call ptr_release(o%tx_paths_all, e)
if (ai_pr) allocate(o%tx_paths_all(dim_em,dim_om,dim_o,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%rx_paths_all); if (rp) call ptr_release(o%rx_paths_all, e)
if (ai_pr) allocate(o%rx_paths_all(dim_om,dim_o,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-INDEPENDENT} Root PE parallel control parameter arrays {[*],n_PE_leaf}')

!== {pd-INDEPENDENT} Leaf PE parallel control parameter arrays {[*],n_PE_leaf}:

if (ai_pl) dim_o = n_PE_leaf

if (np) nullify(o%longdt_lBC_sd_packed_B); if (rp) call ptr_release(o%longdt_lBC_sd_packed_B, e)
if (ai_pl) allocate(o%longdt_lBC_sd_packed_B(dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (ai_pl) then
   dim_om = n_comm_path_optypes_max
   dim_em = n_comm_path_entries_max
end if

if (np) nullify(o%tx_paths_leaf); if (rp) call ptr_release(o%tx_paths_leaf, e)
if (ai_pl) allocate(o%tx_paths_leaf(dim_em,dim_om,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%rx_paths_leaf); if (rp) call ptr_release(o%rx_paths_leaf, e)
if (ai_pl) allocate(o%rx_paths_leaf(dim_om,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (ai_pl) dim_g = n_grids

if (np) nullify(o%dnscale_nBC_sd_packed_B); if (rp) call ptr_release(o%dnscale_nBC_sd_packed_B, e)
if (ai_pl) allocate(o%dnscale_nBC_sd_packed_B(dim_g,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (np) nullify(o%upscale_nBC_sd_packed_B); if (rp) call ptr_release(o%upscale_nBC_sd_packed_B, e)
if (ai_pl) allocate(o%upscale_nBC_sd_packed_B(dim_g,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (ai_pl) dim_ao = n_acoustic_lBC_optypes

if (np) nullify(o%acoustic_lBC_sd_packed_B); if (rp) call ptr_release(o%acoustic_lBC_sd_packed_B, e)
if (ai_pl) allocate(o%acoustic_lBC_sd_packed_B(dim_ao,dim_o), stat=e)
e_check = e_check .or. (e /= 0)

if (e_check) call exception(procedure_name, FATAL,  &
     '{pd-INDEPENDENT} Leaf PE parallel control parameter arrays {[*],n_PE_leaf}')

end subroutine manage_grid_object


!```````````````````````````````````````````````````````````````````
subroutine manage_parallel_memory(dim_n_PE_leaf, dim_misc)

!+ Allocates or deallocates module arrays

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop

implicit none

!=== Argument declarations:

logical, optional, intent(IN) :: dim_n_PE_leaf, dim_misc

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'manage_parallel_memory'

integer :: a_err

!=== Executable statements:

if (my_rank == rootPE_rank) then   ! For root PE operations

   if (present(dim_n_PE_leaf)) then
      if (dim_n_PE_leaf) then
         allocate(PE_rank(n_PE_leaf), stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_n_PE_leaf,  &
              procedure_name, 'dim_n_PE_leaf root')
      else
         deallocate(PE_rank, stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_n_PE_leaf,  &
              procedure_name, 'dim_n_PE_leaf root')
      end if
   end if

   if (present(dim_misc)) then
      if (dim_misc) then
         if (ibnd == 4 .or. jbnd == 4) then  ! Size of cyclic paths array
            n_pts_cyclic = n_cyclic_entries*n_PE_leaf
            if (ibnd == 4) n_pts_cyclic = n_pts_cyclic+gd(1)%nyp*4
            if (jbnd == 4) n_pts_cyclic = n_pts_cyclic+gd(1)%nxp*4
            if (ibnd == 4 .and. jbnd == 4) n_pts_cyclic = n_pts_cyclic+16
         end if

         allocate(cyclic_paths_t(n_cyclic_entries,n_pts_cyclic),  &
              cyclic_paths_u(n_cyclic_entries,n_pts_cyclic),  &
              cyclic_paths_v(n_cyclic_entries,n_pts_cyclic), stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_misc, procedure_name,  &
              'dim_misc root')    
      else
         deallocate(cyclic_paths_t, cyclic_paths_u, cyclic_paths_v,  &
              stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_misc, procedure_name,  &
              'dim_misc root')
      end if
   end if

else                               ! For leaf PE operations

   if (present(dim_n_PE_leaf)) then
      if (dim_n_PE_leaf) then
         allocate(dnscale_nBC_packed_pos(n_PE_leaf),  &
              upscale_nBC_packed_pos(n_PE_leaf), PE_rank(n_PE_leaf),  &
              acoustic_lBC_packed_pos(n_PE_leaf),  &
              longdt_lBC_packed_pos(n_PE_leaf), stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_n_PE_leaf,  &
              procedure_name, 'dim_n_PE_leaf leaf')
      else
         deallocate(dnscale_nBC_packed_pos, PE_rank,  &
              upscale_nBC_packed_pos, acoustic_lBC_packed_pos,  &
              longdt_lBC_packed_pos, stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_n_PE_leaf,  &
              procedure_name, 'dim_n_PE_leaf leaf')
      end if
   end if

   if (present(dim_misc)) then
      if (dim_misc) then
         if (ibnd == 4 .or. jbnd == 4) then  ! Size of cyclic paths array
            n_pts_cyclic = n_cyclic_entries*n_PE_leaf
            if (ibnd == 4) n_pts_cyclic = n_pts_cyclic+gd(1)%nyp*4
            if (jbnd == 4) n_pts_cyclic = n_pts_cyclic+gd(1)%nxp*4
            if (ibnd == 4 .and. jbnd == 4) n_pts_cyclic = n_pts_cyclic+16
         end if

         allocate(ivar_ID(n_vars), ivar_type(n_vars),  &
              cyclic_paths_t(n_cyclic_entries,n_pts_cyclic),  &
              cyclic_paths_u(n_cyclic_entries,n_pts_cyclic),  &
              cyclic_paths_v(n_cyclic_entries,n_pts_cyclic), stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_misc, procedure_name,  &
              'dim_misc leaf')
      else
         deallocate(ivar_ID, ivar_type, cyclic_paths_t, cyclic_paths_u,  &
              cyclic_paths_v, stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(dim_misc, procedure_name,  &
              'dim_misc leaf')
      end if
   end if

end if

end subroutine manage_parallel_memory


end module main_memory_mod
