!```````````````````````````````````````````````````````````````````
module mod_Euler_rotation_mod

!+ Module containing TC->AC and AC->TC coordinate transformation matrices

use univ_kind_defs_mod, only: f4, f8
use univ_math_constants_mod, only: ZERO_f8

implicit none

!-----------------------------------------------------------------------
! *** For AC --> TC:
!
! ALPHA : rotation of axes about z-axis (longitude)
! BETA  : rotation of axes about new y-axis (latitude) 
! GAMMA : rotation of axes about new z-axis (azimuth angle) 
!
! OC_to_NC : coordinate transformation matrix (bodycentric -> topocentric )
! NC_to_OC : coordinate transformation matrix (topocentric -> bodycentric )
!-----------------------------------------------------------------------

real(f8) :: alpha = ZERO_f8, beta = ZERO_f8, gamma = ZERO_f8

real(f4), dimension(3,3) :: OC_to_NC,NC_to_OC


contains


!```````````````````````````````````````````````````````````````````
subroutine Euler_Rotation_Matrix

!+

use univ_kind_defs_mod, only: f8

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

real(f8) :: cosA,sinA,cosB,sinB,cosG,sinG,tx,ty,tz

!=== Executable statements:

! Compute trigonometric quantities of rotation angles:

cosA = cos(alpha)
sinA = sin(alpha)
cosB = cos(beta)
sinB = sin(beta)
cosG = cos(gamma)
sinG = sin(gamma) 

! Compute coordinate transformation matrix:
  
tx = cosA*cosB*cosG-sinA*sinG
ty = cosA*sinG+sinA*cosB*cosG
tz = -sinB*cosG
  
OC_to_NC(1,1) = real(tx, kind=f4)
OC_to_NC(1,2) = real(ty, kind=f4)
OC_to_NC(1,3) = real(tz, kind=f4)
NC_to_OC(1,1) = real(tx, kind=f4)
NC_to_OC(2,1) = real(ty, kind=f4)
NC_to_OC(3,1) = real(tz, kind=f4)
  
tx = -cosA*cosB*sinG-sinA*cosG
ty = cosA*cosG-sinA*cosB*sinG
tz = sinB*sinG

OC_to_NC(2,1) = real(tx, kind=f4)
OC_to_NC(2,2) = real(ty, kind=f4)
OC_to_NC(2,3) = real(tz, kind=f4)
NC_to_OC(1,2) = real(tx, kind=f4)
NC_to_OC(2,2) = real(ty, kind=f4)
NC_to_OC(3,2) = real(tz, kind=f4)

tx = cosA*sinB
ty = sinA*sinB
tz = cosB

OC_to_NC(3,1) = real(tx, kind=f4)
OC_to_NC(3,2) = real(ty, kind=f4)
OC_to_NC(3,3) = real(tz, kind=f4)
NC_to_OC(1,3) = real(tx, kind=f4)
NC_to_OC(2,3) = real(ty, kind=f4)
NC_to_OC(3,3) = real(tz, kind=f4)
 
end subroutine Euler_Rotation_Matrix


end module mod_Euler_rotation_mod
