!```````````````````````````````````````````````````````````````````
module namelist_mod

!+ Module that contains routines and variables related to the PRAMS namelist

use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_MESSAGE_LEN 
use univ_integer_time_mod, only: time_obj
use namelist_maxdims_mod, only: nml_max_n_grids, nml_max_n_z,  &
     nml_max_n_zsubsfc, nml_max_n_zsupersfc, nml_max_n_sounding,  &
     nml_max_n_input_files, nml__strtoklen_max, nml_fn_strtoklen_max
use main_memory_mod, only: simulation_description, run_type,  &
     initialization_type, target_Ls, start_UT, dt_dynamics, min_ssfrac,  &
     max_dt_dynamics, dt_max_frac_increase_g1, dt_mode, diurnal_progression,  &
     n_dta_per_dtdyn, n_substeps_max, n_grids, map_projection,  &
     deltax_g1, deltay_g1, deltaz_g1, dzrat, dzmax, polelat, nzg, nzs,  &
     nestz1, polelon, output_dir_root, simulation_ID, run_suffix, ioutput,  &
     stop_if_file_exists, wfreq_state_file, orbital_progression,  &
     input_varfile_pfx, nudlat, tnudlat, tnudcent, tnudtop,  &
     znudtop, preserve_dt_on_restart, input_state_file_pfx, prognose_uvwp,  &
     input_sfcfile_pfx, z0fact, naddsc, icorflg, ibnd, jbnd, cphas, lsflg,  &
     nfpt, distim, pnLES_np, pnLES_fn, pnLES_pt, radfrq, perturb_nsfc_theta,  &
     enable_advection, islastog, npatch, zrough, albedo, TI_val, ihorgrad,  &
     nigrids, PIPP_nvf, guess1st, vf_LS, tgt_time, tgt_tlon, nf_span, varpfx,  &
     gen_vf_input_comparison, impl, ntopsmth, izflat, iadvl, iadvf,  &
     default_input_data_dir, run_sfcinit_user, run_toptinit_user,  &
     enable_turb_diffusion, enable_aerosol_microphysics, RT_paradigm,  &
     prognose_SBPS, n_dtdyn_per_state_copy, islastog, emissivity

implicit none

character(len=18), parameter :: model_nml_fn = 'auto_gen-PRAMS.nml'

integer, dimension(nml_max_n_grids) :: tmp_ngd_i

!== &static_parameters (currently not actually read from namelist)
namelist /static_parameters/ impl, ntopsmth, izflat, iadvl, iadvf

!== &namelist_info
integer, parameter :: n_chars_nml_version = 4, n_chars_nml_is_for = 5
character(len=n_chars_nml_is_for) :: namelist_is_for
character(len=n_chars_nml_version) :: namelist_version_common
character(len=MAX_MESSAGE_LEN) :: namelist_notes_common
namelist /namelist_info/ namelist_is_for, namelist_version_common,  &
     namelist_notes_common 

!== &basic_control
namelist /basic_control/ simulation_description, run_type, n_grids,  &
     map_projection, initialization_type

!== &time_config
character(len=nml__strtoklen_max) :: maximum_sim_ET_
type(time_obj) :: max_sim_ET_Time
integer, dimension(nml_max_n_grids) :: n_dt_per_pgdt_
namelist /time_config/ orbital_progression, diurnal_progression, target_Ls,  &
     start_UT, maximum_sim_ET_, dt_mode, dt_dynamics, n_dta_per_dtdyn,  &
     min_ssfrac, radfrq, max_dt_dynamics, dt_max_frac_increase_g1,  &
     n_substeps_max, n_dtdyn_per_state_copy, n_dt_per_pgdt_

!== &spatial_config
integer, dimension(nml_max_n_grids) :: nxp_, nyp_, nzp_, nxtnest_,  &
     nstratx_, nstraty_, ninest_, njnest_, nknest_, nsttop_,  &
     nstbot_
real(f4), dimension(nml_max_n_grids) :: centlat_, centlon_
real(f4), dimension(nml_max_n_z) :: zz_
integer, dimension(nml_max_n_z) :: nstratz1_
namelist /spatial_config/ deltax_g1, deltay_g1, deltaz_g1, dzrat,  &
     dzmax, nxp_, nyp_, nzp_, nzg, nzs,  &
     zz_, nxtnest_, nstratx_, nstraty_, nestz1, nstratz1_, polelat,  &
     polelon, centlat_, centlon_, ninest_, njnest_, nknest_, nsttop_,  &
     nstbot_

!== &inout_config
integer, dimension(nml_max_n_grids) :: itoptflg_, itopsflg_, itinflg_,  &
     ialbflg_, iz0flg_
real(f4), dimension(nml_max_n_grids) :: toptenh_, toptwvlx_, toptwvly_,  &
     tinwvlx_, tinwvly_, albwvlx_, albwvly_, z0max_, z0_of_topolevel_,  &
     topolevel_for_z0_
character(len=nml_fn_strtoklen_max), dimension(nml_max_n_grids) :: itoptfn_,  &
     itinfn_, ialbfn_
character(len=nml__strtoklen_max) :: state_file_grids_
character(len=nml__strtoklen_max), dimension(nml_max_n_sounding) :: pnLES_pt_
namelist /inout_config/ output_dir_root, simulation_ID, run_suffix, ioutput,  &
     stop_if_file_exists, state_file_grids_, wfreq_state_file,  &
     input_varfile_pfx, nudlat, tnudlat, tnudcent, tnudtop, znudtop,  &
     input_state_file_pfx, preserve_dt_on_restart,  &
     default_input_data_dir, input_sfcfile_pfx, itoptflg_, itoptfn_,  &
     itopsflg_, toptenh_, toptwvlx_, toptwvly_, itinflg_, itinfn_, tinwvlx_,  &
     tinwvly_, ialbflg_, ialbfn_, albwvlx_, albwvly_, iz0flg_, z0max_,  &
     z0fact, z0_of_topolevel_, topolevel_for_z0_, pnLES_np, pnLES_fn,  &
     pnLES_pt_, run_sfcinit_user, run_toptinit_user

!== &model_options
integer, dimension(nml_max_n_grids) :: nofilflg_, idiffk_
real(f4), dimension(nml_max_n_grids) :: csx_, csz_, xkhkm_, zkhkm_, akmin_,  &
     gridu_, gridv_
namelist /model_options/ naddsc, icorflg, enable_advection,   &
     enable_turb_diffusion, enable_aerosol_microphysics, RT_paradigm,  &
     prognose_SBPS, prognose_uvwp, gridu_, gridv_, ibnd, jbnd, cphas,  &
     lsflg, nfpt, distim, perturb_nsfc_theta, islastog, idiffk_, ihorgrad,  &
     csx_, csz_, xkhkm_, zkhkm_, akmin_, npatch, nofilflg_, zrough, albedo,  &
     TI_val, emissivity
     
!== &PIPP_control
character(len=nml_fn_strtoklen_max), dimension(nml_max_n_input_files) :: iapr_
namelist /PIPP_control/ nigrids, PIPP_nvf, guess1st, vf_Ls, tgt_time,  &
     tgt_tlon, nf_span, iapr_, varpfx, gen_vf_input_comparison


contains


!```````````````````````````````````````````````````````````````````
subroutine nml_info_check(namelist_fh)

!+ Check '&namelist_info' namelist block for errors 

use univ_std_fileIO_mod, only: find_file_IOunit
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_MESSAGE_LEN, MAX_PROC_NAME_LEN
use univ_namelist_utils_mod, only: nml_num_value_check

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh    ! File I/O unit of open namelist file

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'nml_info_check'

character(len=n_chars_nml_is_for) :: base_app_to_match = 'PRAMS'
character(len=n_chars_nml_version) :: base_version_to_match = '0006'

integer :: cn, IO_status
character(len=MAX_MESSAGE_LEN) :: msg

!=== Executable statements:

! Input &namelist_info namelist block; check for errors:

read(namelist_fh, nml=namelist_info, iostat=IO_status)
if (IO_status /= 0) call exception(procedure_name, FATAL,  &
     'Namelist block &namelist_info not found or is damaged.')

cn = max(1, len_trim(namelist_is_for))
if (namelist_is_for(1:cn) /= base_app_to_match) then
   write(msg, '(5a)')  'Input namelist application (', namelist_is_for(1:cn),  &
        ') is incompatible with the necessary application (',  &
        base_app_to_match, ').'
   call exception(procedure_name, FATAL, trim(msg))
end if
cn = max(1, len_trim(namelist_version_common))
if (namelist_version_common(1:cn) /= base_version_to_match) then
   write(msg, '(5a)')  'Input namelist version (',  &
        namelist_version_common(1:cn),  &
        ') is incompatible with the necessary version (',  &
        base_version_to_match, ').'
   call exception(procedure_name, FATAL, trim(msg))
end if

end subroutine nml_info_check


!```````````````````````````````````````````````````````````````````
subroutine proc_nml_static_params(namelist_fh)

!+ Processing related to items from namelist &static_parameters

!@use univ_exception_mod, only: exception, FATAL
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'proc_nml_static_params'

!@integer :: IO_status

!=== Executable statements:

!== Initialize namelist variables with out of range values (where possible) for
!    checking purposes:

! For this routine (only), set these to their proper values:
!
! This set should be changed only for special tests or with code modification:
!  IMPL     = 1   ! Implicit flag for acoustic model  -  0=off, 1=on
!  NTOPSMTH = 0   ! Number of passes for topography smoother
!  IZFLAT   = 0   ! Width of flat margin around domain (in grid points)
!  IADVL    = 2   ! Order of advection - Leapfrog - 2 or 4 - only 2 now
!  IADVF    = 2   ! Order of advection - Forward - 2 or 6 - only 2 now

impl = 1
ntopsmth = 0
izflat = 0
iadvl = 2; iadvf = 2

!== Attempt to read the namelist:

!@read(namelist_fh, nml=static_parameters, iostat=IO_status)
!@if (IO_status /= 0) call exception(procedure_name, FATAL,  &
!@     'Namelist block &static_parameters not found or is damaged.')

!== Check whether some namelist values are within an allowable range:

! Nothing to do here at the moment

end subroutine proc_nml_static_params


!```````````````````````````````````````````````````````````````````
subroutine proc_nml_basic_control(namelist_fh)

!+ Processing related to items from namelist &basic_control

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_namelist_utils_mod, only: nml_num_value_check
use main_memory_mod, only:  &
     manage_nml_var_memory  ! and others already used by this module

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'proc_nml_basic_control'

integer :: IO_status

!=== Executable statements:

!== Initialize namelist variables with out of range values (where possible) for
!    checking purposes:

simulation_description = 'No description provided.'
run_type = '__NULL__'
n_grids = -1
map_projection = -1
initialization_type = -1

!== Attempt to read the namelist:

read(namelist_fh, nml=basic_control, iostat=IO_status)
if (IO_status /= 0) call exception(procedure_name, FATAL,  &
     'Namelist block &basic_control not found or is damaged.')

!== Check whether some namelist values are within an allowable range:

!... 'simulation_description' may contain nearly anything

select case (run_type)
case ('GRID_INFO_FILE', 'MAKE_SFC_FILES', 'CHECK_SFC_FILES',  &
     'MAKE_VAR_FILES', 'INITIAL_START', 'RESTART')
   continue
case default
   call exception(procedure_name, FATAL,  &
        'Improper value for namelist variable RUN_TYPE')
end select

call nml_num_value_check(n_grids, 'N_GRIDS', 'and', ge=1, le=nml_max_n_grids)
call manage_nml_var_memory(dim_n_grids=.TRUE.) ! Allocate grid object(s)

call nml_num_value_check(map_projection, 'MAP_PROJECTION', 'and', ge=0, le=1)
call nml_num_value_check(initialization_type, 'INITIALIZATION_TYPE', 'and',  &
     ge=1, le=2)

end subroutine proc_nml_basic_control


!```````````````````````````````````````````````````````````````````
subroutine proc_nml_time_config(namelist_fh)

!+ Processing related to items from namelist &time_config

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ONE_f4
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_integer_time_mod, only: time_eqv_fp_seconds, fp_seconds_eqv_time,  &
     set_time
use univ_numtype_constants_mod, only: epsilon_f4, epsilon_f8, tiny_f4
use univ_namelist_utils_mod, only: gen_time_strrep_to_Time, nml_num_value_check
use main_memory_mod, only:  &  ! and others already used by this module
     maximum_sim_ET, gd, epsilon_sim_ET, n_isched_steps, mnsdy_Time,  &
     morbpd_Time, zero_Time, module_tps
use body_specific_constants_mod, only: morbpd, mnsdy

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'proc_nml_time_config'

integer :: IO_status
logical :: error

!=== Executable statements:

!== Initialize namelist variables with out of range values (where possible) for
!    checking purposes:

orbital_progression = .TRUE.; diurnal_progression = .TRUE.
target_Ls = -ONE_f4
start_UT = -ONE_f4
maximum_sim_ET_ = '__NULL__'
dt_mode = 1
dt_dynamics = -ONE_f4
n_dta_per_dtdyn = -1
min_ssfrac = -ONE_f4
radfrq = -ONE_f4
max_dt_dynamics = -ONE_f4
dt_max_frac_increase_g1 = -ONE_f4
n_substeps_max = -1
n_dtdyn_per_state_copy = -1
n_dt_per_pgdt_(:) = -1

!== Attempt to read the namelist:

read(namelist_fh, nml=time_config, iostat=IO_status)
if (IO_status /= 0) call exception(procedure_name, FATAL,  &
     'Namelist block &time_config not found or is damaged.')

!== Check whether some namelist values are within an allowable range:

   ! Compute/set some important time parameters:
zero_Time = set_time(0, 0, module_tps)
mnsdy_Time = fp_seconds_eqv_time(mnsdy, module_tps)
morbpd_Time = fp_seconds_eqv_time(morbpd, module_tps)

if (initialization_type == 1) then   ! These must be >= 0. in this case
   call nml_num_value_check(target_Ls, 'TARGET_LS', gt=(-tiny_f4))
   call nml_num_value_check(start_UT, 'START_UT', gt=(-tiny_f4))  
end if

call gen_time_strrep_to_Time(trim(maximum_sim_ET_), max_sim_ET_Time, error,  &
     mnsdy_Time, morbpd_Time, module_tps)
if (error) call exception(procedure_name, FATAL,  &
     'Improper value for namelist variable MAXIMUM_SIM_ET_')
maximum_sim_ET = time_eqv_fp_seconds(max_sim_ET_Time)

epsilon_sim_ET = 100._f8*epsilon_f8 ! Sim ET comparison parameter

call nml_num_value_check(dt_mode, 'DT_MODE', le=0) ! Use only implemented range
call nml_num_value_check(dt_dynamics, 'DT_DYNAMICS', ge=tiny_f4)
call nml_num_value_check(n_dta_per_dtdyn, 'N_DTA_PER_DTDYN', ge=1)
call nml_num_value_check(min_ssfrac, 'MIN_SSFRAC', 'and', gt=(-tiny_f4),  &
     lt=ONE_f4+epsilon_f4)
call nml_num_value_check(radfrq, 'RADFRQ', ge=tiny_f4)
call nml_num_value_check(max_dt_dynamics, 'MAX_DT_DYNAMICS', gt=(-tiny_f4))

call nml_num_value_check(dt_max_frac_increase_g1, 'DT_MAX_FRAC_INCREASE_G1',  &
     gt=(-tiny_f4))
call nml_num_value_check(n_substeps_max, 'N_SUBSTEPS_MAX', ge=1)
n_isched_steps = n_substeps_max**2  ! Set related parameter
call nml_num_value_check(n_dtdyn_per_state_copy, 'N_DTDYN_PER_STATE_COPY', ge=1)

call nml_num_value_check(n_dt_per_pgdt_(1:n_grids),  &
     'N_DT_PER_PGDT_', ge=1)
gd(1:n_grids)%n_dt_per_pgdt = n_dt_per_pgdt_(1:n_grids)

end subroutine proc_nml_time_config


!```````````````````````````````````````````````````````````````````
subroutine proc_nml_spatial_config(namelist_fh)

!+ Processing related to items from namelist &spatial_config

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_numtype_constants_mod, only: tiny_f4, epsilon_f4
use univ_namelist_utils_mod, only: nml_num_value_check
use main_memory_mod, only:  &   ! and others already used by this module
     gd, maxnxp, maxnyp, maxnzp, manage_nml_var_memory, zz, nstratz1

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'proc_nml_spatial_config'

integer :: IO_status

!=== Executable statements:

!== Initialize namelist variables with out of range values (where possible) for
!    checking purposes:

nxp_(:) = -1; nyp_(:) = -1; nzp_(:) = -1
nzg = -1; nzs = -1
deltax_g1 = -ONE_f4; deltay_g1 = -ONE_f4; deltaz_g1 = ZERO_f4
zz_(:) = -ONE_f4
dzrat = -ONE_f4
dzmax = -ONE_f4
nxtnest_(:) = -1; nstratx_(:) = -1; nstraty_(:) = -1
nestz1 = nml_max_n_grids+1
nstratz1_(:) = -1
nknest_(:) = -1
nsttop_(:) = -1; nstbot_(:) = -1
polelat = 999._f4; polelon = 999._f4
centlat_(:) = 999._f4; centlon_(:) = 999._f4 
ninest_(:) = -1; njnest_(:) = -1

!== Attempt to read the namelist:

read(namelist_fh, nml=spatial_config, iostat=IO_status)
if (IO_status /= 0) call exception(procedure_name, FATAL,  &
     'Namelist block &spatial_config not found or is damaged.')

!== Check whether some namelist values are within an allowable range:

call nml_num_value_check(nxp_(1:n_grids), 'NXP_', ge=3)
gd(1:n_grids)%nxp = nxp_(1:n_grids)
maxnxp = maxval(nxp_(1:n_grids))

call nml_num_value_check(nyp_(1:n_grids), 'NYP_', ge=1)
gd(1:n_grids)%nyp = nyp_(1:n_grids)
maxnyp = maxval(nyp_(1:n_grids))

call nml_num_value_check(nzp_(1:n_grids), 'NZP_', 'and', ge=1, le=nml_max_n_z)
gd(1:n_grids)%nzp = nzp_(1:n_grids)
maxnzp = maxval(nzp_(1:n_grids))
call manage_nml_var_memory(dim_maxnzp=.TRUE.) ! Allocate nml vars w/dim maxnzp

call nml_num_value_check(nzg, 'NZG', 'and', ge=1, le=nml_max_n_zsubsfc)
call nml_num_value_check(nzs, 'NZS', 'and', ge=1, le=nml_max_n_zsupersfc)

call nml_num_value_check(deltax_g1, 'DELTAX_G1', ge=tiny_f4)
call nml_num_value_check(deltay_g1, 'DELTAY_G1', ge=tiny_f4)
call nml_num_value_check(deltaz_g1, 'DELTAZ_G1', 'or', le=(-tiny_f4),  &
     ge=tiny_f4)

if (deltaz_g1 <= -tiny_f4) call nml_num_value_check(zz_(1:maxnzp),  &
     'ZZ_', ge=tiny_f4)    ! Only check if this parameter will be used
zz(1:maxnzp) = zz_(1:maxnzp)

call nml_num_value_check(dzrat, 'DZRAT', ge=tiny_f4)
call nml_num_value_check(dzmax, 'DZMAX', ge=tiny_f4)

call nml_num_value_check(nxtnest_(1:n_grids), 'NXTNEXT_', 'and', ge=0,  &
     le=nml_max_n_grids)
gd(1:n_grids)%nxtnest = nxtnest_(1:n_grids)
call nml_num_value_check(nstratx_(1:n_grids), 'NSTRATX_', ge=1)
gd(1:n_grids)%nstratx = nstratx_(1:n_grids)
call nml_num_value_check(nstraty_(1:n_grids), 'NSTRATY_', ge=1)
gd(1:n_grids)%nstraty = nstraty_(1:n_grids)

call nml_num_value_check(nestz1, 'NESTZ1', 'and', ge=-nml_max_n_grids,  &
     le=nml_max_n_grids)
if (nestz1 /= 0) call nml_num_value_check(nstratz1_(1:maxnzp), 'NSTRATZ1_',  &
     ge=1)      ! Only check if this parameter will be used
nstratz1(1:maxnzp) = nstratz1_(1:maxnzp)
call nml_num_value_check(nknest_(1:n_grids), 'NKNEST_', ge=1)
gd(1:n_grids)%nknest = nknest_(1:n_grids)
call nml_num_value_check(nsttop_(1:n_grids), 'NSTTOP_', 'and', ge=0, le=1)
gd(1:n_grids)%nsttop = nsttop_(1:n_grids)
call nml_num_value_check(nstbot_(1:n_grids), 'NSTBOT_', 'and', ge=0, le=1)
gd(1:n_grids)%nstbot = nstbot_(1:n_grids)

call nml_num_value_check(polelat, 'POLELAT', 'and', gt=-90._f4-epsilon_f4,  &
     lt=90._f4+epsilon_f4)
call nml_num_value_check(polelon, 'POLELON', 'and', gt=-180._f4-epsilon_f4,  &
     lt=180._f4+epsilon_f4)

call nml_num_value_check(centlat_(1:n_grids), 'CENTLAT_', 'and',  &
     gt=-90._f4-epsilon_f4, lt=90._f4+epsilon_f4)
gd(1:n_grids)%centlat = centlat_(1:n_grids)
call nml_num_value_check(centlon_(1:n_grids), 'CENTLON_', 'and',  &
     gt=-180._f4-epsilon_f4, lt=180._f4+epsilon_f4)
gd(1:n_grids)%centlon = centlon_(1:n_grids)

call nml_num_value_check(ninest_(1:n_grids), 'NINEST_', ge=0)
gd(1:n_grids)%ninest = ninest_(1:n_grids)
call nml_num_value_check(njnest_(1:n_grids), 'NJNEST_', ge=0)
gd(1:n_grids)%njnest = njnest_(1:n_grids)

end subroutine proc_nml_spatial_config


!```````````````````````````````````````````````````````````````````
subroutine proc_nml_inout_config(namelist_fh)

!+ Processing related to items from namelist &inout_config

use univ_math_constants_mod, only: ONE_f4
use univ_exception_mod, only: exception, FATAL
use univ_std_fileIO_mod, only: find_file_IOunit
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN
use univ_numtype_constants_mod, only: tiny_f4
use univ_namelist_utils_mod, only: nml_num_value_check
use main_memory_mod, only:  &  ! and others already used by this module
     simulation_ID_w_sfx, gd, output_dir_root_w_sim_ID, pnLES_pt,  &
     manage_nml_var_memory
use state_IO_mod, only: cfgctl_rw
use postp_init_mod, only: process_dim_str

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'proc_nml_inout_config'

integer :: IO_status, IO_unit
character(len=MAX_FILE_PATH_LEN) :: fn_tmp

!=== Executable statements:

!== Initialize namelist variables with out of range values (where possible) for
!    checking purposes:

output_dir_root = '__NULL__'
simulation_ID = '__NULL__'
run_suffix = '__NULL__'
ioutput = -1
stop_if_file_exists = .TRUE.; state_file_grids_ = '__NULL__'
wfreq_state_file = 0._f4
input_varfile_pfx = '__NULL__'
nudlat = -1
tnudlat = -ONE_f4; tnudcent = -ONE_f4; tnudtop = -ONE_f4
znudtop = -ONE_f4
input_state_file_pfx = '__NULL__'
preserve_dt_on_restart = .FALSE.
default_input_data_dir = '__NULL__'
input_sfcfile_pfx = '__NULL__'
run_sfcinit_user = .FALSE.
run_toptinit_user = .FALSE.
itoptflg_(:) = -1; itoptfn_(:) = '__NULL__'; itopsflg_(:) = -1
toptenh_(:) = -ONE_f4; toptwvlx_(:) = -ONE_f4; toptwvly_(:) = -ONE_f4
itinflg_(:) = -1; itinfn_(:) = '__NULL__'
tinwvlx_(:) = -ONE_f4; tinwvly_(:) = -ONE_f4
ialbflg_(:) = -1; ialbfn_(:) = '__NULL__'
albwvlx_(:) = -ONE_f4; albwvly_(:) = -ONE_f4
iz0flg_(:) = -1; z0max_(:) = -ONE_f4; z0_of_topolevel_(:) = -99999._f4
topolevel_for_z0_(:) = -99999._f4; z0fact = -ONE_f4
pnLES_np = -1; pnLES_fn = '__NULL__'; pnLES_pt_(:) = '__NULL__'

!== Attempt to read the namelist:

read(namelist_fh, nml=inout_config, iostat=IO_status)
if (IO_status /= 0) call exception(procedure_name, FATAL,  &
     'Namelist block &inout_config not found or is damaged.')

!== Check whether some namelist values are within an allowable range:

if (trim(output_dir_root) == '__NULL__') call exception(procedure_name,  &
     FATAL, 'Improper value for namelist variable OUTPUT_DIR_ROOT')
if (trim(simulation_ID) == '__NULL__') call exception(procedure_name,  &
     FATAL, 'Improper value for namelist variable SIMULATION_ID')
output_dir_root_w_sim_ID = trim(output_dir_root)//'/'//trim(simulation_ID)

if (trim(run_suffix) == '__NULL__') then
   simulation_ID_w_sfx = trim(simulation_ID)
else
   simulation_ID_w_sfx = trim(simulation_ID)//trim(run_suffix)
end if

call nml_num_value_check(ioutput, 'IOUTPUT', 'and', ge=0, le=1)
call nml_num_value_check(wfreq_state_file, 'WFREQ_STATE_FILE', 'or',  &
     le=(-tiny_f4), ge=tiny_f4)

if (trim(state_file_grids_) == '__NULL__') call exception(procedure_name,  &
     FATAL, 'Improper value for namelist variable STATE_FILE_GRIDS_')
if (trim(state_file_grids_) == 'ALL') then
   gd(1:n_grids)%grid_in_state_outp = .TRUE.
else
   tmp_ngd_i(1:n_grids) = -1   ! Initialize with sentinel value
   call process_dim_str('STATE_FILE_GRIDS', state_file_grids_, tmp_ngd_i)
   gd(1:n_grids)%grid_in_state_outp = ( tmp_ngd_i(1:n_grids) > 0 )
end if

if (trim(input_varfile_pfx) == '__NULL__') then   ! Must set default
   if (trim(run_suffix) == '__NULL__') then
      input_varfile_pfx = trim(output_dir_root_w_sim_ID)//  &
           '/intermediary/varfile'
   else
      input_varfile_pfx = trim(output_dir_root_w_sim_ID)//  &
           '/intermediary/varfile'//trim(run_suffix)
   end if
end if
call nml_num_value_check(nudlat, 'NUDLAT', gt=0)
call nml_num_value_check(tnudlat, 'TNUDLAT', gt=(-tiny_f4))
call nml_num_value_check(tnudcent, 'TNUDCENT', gt=(-tiny_f4))
call nml_num_value_check(tnudtop, 'TNUDTOP', gt=(-tiny_f4))
call nml_num_value_check(znudtop, 'ZNUDTOP', gt=(-tiny_f4))

if (trim(input_state_file_pfx) == '__NULL__') call exception(  &
     procedure_name, FATAL,  &
     'Improper value for namelist variable INPUT_STATE_FILE_PFX')
if (run_type == 'RESTART') then    ! Extract restart time from state file
   IO_unit = find_file_IOunit() ! Obtain an available file unit
   fn_tmp = trim(input_state_file_pfx)//'-hdr.vfm'  ! State header filename
   open(IO_unit, file=trim(fn_tmp), form='unformatted', status='old')
   read (IO_unit) IO_status; read (IO_unit) IO_status  ! Skip two records
   call cfgctl_rw(IO_unit, 'R', 'Q0S')  ! Fills 'start_sim_ET'
   close(IO_unit)
end if

if (trim(default_input_data_dir) == '__NULL__') call exception(  &
     procedure_name, FATAL,  &
     'Improper value for namelist variable DEFAULT_INPUT_DATA_DIR')

if (trim(input_sfcfile_pfx) == '__NULL__') then   ! Must set default
   if (trim(run_suffix) == '__NULL__') then
      input_sfcfile_pfx = trim(output_dir_root_w_sim_ID)//  &
           '/intermediary/sfcfile'
   else
      input_sfcfile_pfx = trim(output_dir_root_w_sim_ID)//  &
           '/intermediary/sfcfile'//trim(run_suffix)
   end if
end if

call nml_num_value_check(itoptflg_(1:n_grids), 'ITOPTFLG_', 'and', ge=0, le=2)
gd(1:n_grids)%itoptflg = itoptflg_(1:n_grids)
gd(1:n_grids)%itoptfn = itoptfn_(1:n_grids)
call nml_num_value_check(itopsflg_(1:n_grids), 'ITOPSFLG_', 'and', ge=0, le=3)
gd(1:n_grids)%itopsflg = itopsflg_(1:n_grids)
call nml_num_value_check(toptenh_(1:n_grids), 'TOPTENH_', ge=tiny_f4)
gd(1:n_grids)%toptenh = toptenh_(1:n_grids)
call nml_num_value_check(toptwvlx_(1:n_grids), 'TOPTWVLX_', ge=tiny_f4)
gd(1:n_grids)%toptwvlx = toptwvlx_(1:n_grids)
call nml_num_value_check(toptwvly_(1:n_grids), 'TOPTWVLY_', ge=tiny_f4)
gd(1:n_grids)%toptwvly = toptwvly_(1:n_grids)

call nml_num_value_check(itinflg_(1:n_grids), 'ITINFLG_', 'and', ge=0, le=2)
gd(1:n_grids)%itinflg = itinflg_(1:n_grids)
gd(1:n_grids)%itinfn = itinfn_(1:n_grids)
call nml_num_value_check(tinwvlx_(1:n_grids), 'TINWVLX_', ge=tiny_f4)
gd(1:n_grids)%tinwvlx = tinwvlx_(1:n_grids)
call nml_num_value_check(tinwvly_(1:n_grids), 'TINWVLY_', ge=tiny_f4)
gd(1:n_grids)%tinwvly = tinwvly_(1:n_grids)

call nml_num_value_check(ialbflg_(1:n_grids), 'IALBFLG_', 'and', ge=0, le=2)
gd(1:n_grids)%ialbflg = ialbflg_(1:n_grids)
gd(1:n_grids)%ialbfn = ialbfn_(1:n_grids)
call nml_num_value_check(albwvlx_(1:n_grids), 'ALBWVLX_', ge=tiny_f4)
gd(1:n_grids)%albwvlx = albwvlx_(1:n_grids)
call nml_num_value_check(albwvly_(1:n_grids), 'ALBWVLY_', ge=tiny_f4)
gd(1:n_grids)%albwvly = albwvly_(1:n_grids)

call nml_num_value_check(iz0flg_(1:n_grids), 'IZ0FLG_', 'and', ge=0, le=2)
gd(1:n_grids)%iz0flg = iz0flg_(1:n_grids)
call nml_num_value_check(z0max_(1:n_grids), 'Z0MAX_', ge=tiny_f4)
gd(1:n_grids)%z0max = z0max_(1:n_grids)
call nml_num_value_check(z0fact, 'Z0FACT', ge=tiny_f4)
gd(1:n_grids)%z0_of_topolevel = z0_of_topolevel_(1:n_grids)   ! No value check
gd(1:n_grids)%topolevel_for_z0 = topolevel_for_z0_(1:n_grids) !

call nml_num_value_check(pnLES_np, 'PNLES_NP', 'and', ge=0,  &
     le=nml_max_n_sounding)
call manage_nml_var_memory(dim_pnLES_np=.TRUE.)   ! Allocate 'pnLES_pt'
if (trim(pnLES_fn) == '__NULL__') call exception(procedure_name, FATAL,  &
     'Improper value for namelist variable PNLES_FN')
pnLES_pt(1:pnLES_np) = pnLES_pt_(1:pnLES_np)

end subroutine proc_nml_inout_config


!```````````````````````````````````````````````````````````````````
subroutine proc_nml_model_options(namelist_fh)

!+ Processing related to items from namelist &model_options

use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_numtype_constants_mod, only: tiny_f4
use univ_namelist_utils_mod, only: nml_num_value_check
use main_memory_mod, only:  &  ! and others already used by this module
     gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'proc_nml_model_options'

integer :: IO_status

!=== Executable statements:

!== Initialize namelist variables with out of range values (where possible) for
!    checking purposes:

naddsc = -1; icorflg = -1;
enable_advection = .TRUE.; enable_turb_diffusion = .TRUE.
enable_aerosol_microphysics = .TRUE.; RT_paradigm = -1
prognose_SBPS = .TRUE.; prognose_uvwp = .TRUE.
gridu_(:) = ZERO_f4; gridv_(:) = ZERO_f4
ibnd = -1; jbnd = -1; cphas = -ONE_f4; lsflg = -1
nfpt = -(nml_max_n_z+1); distim = -ONE_f4; perturb_nsfc_theta = .FALSE.
islastog = -1; idiffk_(:) = -1; ihorgrad = -1
csx_(:) = -ONE_f4; csz_(:) = -ONE_f4
xkhkm_(:) = -ONE_f4; zkhkm_(:) = -ONE_f4; akmin_(:) = -ONE_f4
npatch = -1; nofilflg_(:) = -1; zrough = -ONE_f4; albedo = -ONE_f4
TI_val = -ONE_f4; emissivity = -ONE_f4

!== Attempt to read the namelist:

read(namelist_fh, nml=model_options, iostat=IO_status)
if (IO_status /= 0) call exception(procedure_name, FATAL,  &
     'Namelist block &model_options not found or is damaged.')

!== Check whether some namelist values are within an allowable range:

call nml_num_value_check(naddsc, 'NADDSC', ge=0)
call nml_num_value_check(icorflg, 'ICORFLG', 'and', ge=0, le=2)
call nml_num_value_check(RT_paradigm, 'RT_PARADIGM', 'and', ge=0, le=2)

call nml_num_value_check(gridu_(1:n_grids), 'GRIDU_')
gd(1:n_grids)%gridu = gridu_(1:n_grids)
call nml_num_value_check(gridv_(1:n_grids), 'GRIDV_')
gd(1:n_grids)%gridv = gridv_(1:n_grids)

call nml_num_value_check(ibnd, 'IBND', 'and', ge=1, le=4)
call nml_num_value_check(jbnd, 'JBND', 'and', ge=1, le=4)
call nml_num_value_check(cphas, 'CPHAS', ge=tiny_f4)
call nml_num_value_check(lsflg, 'LSFLG', 'and', ge=0, le=3)
call nml_num_value_check(nfpt, 'NFPT', 'and', ge=-nml_max_n_z, le=nml_max_n_z)
call nml_num_value_check(distim, 'DISTIM', ge=tiny_f4)
call nml_num_value_check(islastog, 'ISLASTOG', 'and', ge=0, le=1)

call nml_num_value_check(idiffk_(1:n_grids), 'IDIFFK_', 'and', ge=1, le=9)
gd(1:n_grids)%idiffk = idiffk_(1:n_grids)
call nml_num_value_check(ihorgrad, 'IHORGRAD', 'and', ge=1, le=2)
call nml_num_value_check(csx_(1:n_grids), 'CSX_', ge=tiny_f4)
gd(1:n_grids)%csx = csx_(1:n_grids)
call nml_num_value_check(csz_(1:n_grids), 'CSZ_', ge=tiny_f4)
gd(1:n_grids)%csz = csz_(1:n_grids)
call nml_num_value_check(xkhkm_(1:n_grids), 'XKHKM_', ge=tiny_f4)
gd(1:n_grids)%xkhkm = xkhkm_(1:n_grids)
call nml_num_value_check(zkhkm_(1:n_grids), 'ZKHKM_', ge=tiny_f4)
gd(1:n_grids)%zkhkm = zkhkm_(1:n_grids)
call nml_num_value_check(akmin_(1:n_grids), 'AKMIN_', ge=tiny_f4)
gd(1:n_grids)%akmin = akmin_(1:n_grids)

call nml_num_value_check(npatch, 'NPATCH', eq=1)
call nml_num_value_check(nofilflg_(1:n_grids), 'NOFILFLG_', 'and', ge=0, le=2)
gd(1:n_grids)%nofilflg = nofilflg_(1:n_grids)

call nml_num_value_check(zrough, 'ZROUGH', ge=tiny_f4)
call nml_num_value_check(albedo, 'ALBEDO', 'and', ge=tiny_f4, le=ONE_f4)
call nml_num_value_check(TI_val, 'TI_VAL', ge=tiny_f4)
call nml_num_value_check(emissivity, 'EMISSIVITY', 'and', ge=tiny_f4, le=ONE_f4)

end subroutine proc_nml_model_options


!```````````````````````````````````````````````````````````````````
subroutine proc_nml_PIPP_control(namelist_fh)

!+ Processing related to items from namelist &PIPP_control

use univ_exception_mod, only: exception, FATAL
use univ_math_constants_mod, only: ONE_f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_numtype_constants_mod, only: epsilon_f4, tiny_f4
use univ_namelist_utils_mod, only: nml_num_value_check
use main_memory_mod, only:  &
     manage_nml_var_memory, iapr,  &  ! and others already used by this module
     output_dir_root_w_sim_ID

implicit none

!=== Argument declarations:

integer, intent(IN) :: namelist_fh

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'proc_nml_PIPP_control'

integer :: IO_status

!=== Executable statements:

!== Initialize namelist variables with out of range values (where possible) for
!    checking purposes:

nigrids = -1; PIPP_nvf = -1; guess1st = '__NULL__'; vf_Ls = -ONE_f4
tgt_time = -ONE_f4; tgt_tlon = -181._f4; nf_span = -1; iapr_(:) = '__NULL__'
varpfx = '__NULL__'; gen_vf_input_comparison = .FALSE.

!== Attempt to read the namelist:

read(namelist_fh, nml=PIPP_control, iostat=IO_status)
if (IO_status /= 0) call exception(procedure_name, FATAL,  &
     'Namelist block &PIPP_control not found or is damaged.')

!== Check whether some namelist values are within an allowable range:

call nml_num_value_check(nigrids, 'NIGRIDS', 'and', ge=1, le=nml_max_n_grids)
call nml_num_value_check(PIPP_nvf, 'PIPP_NVF', ge=1)

select case (guess1st)
case ('GCM', 'PRAMS')
   continue
case default
   call exception(procedure_name, FATAL,  &
        'Improper value for namelist variable GUESS1ST')
end select

call nml_num_value_check(vf_Ls, 'VF_LS', 'and', gt=(-tiny_f4),  &
     lt=360._f4+epsilon_f4)
call nml_num_value_check(tgt_time, 'TGT_TIME', 'and', gt=(-tiny_f4),  &
     lt=24._f4+epsilon_f4)
call nml_num_value_check(tgt_tlon, 'TGT_TLON', 'and', gt=-180._f4-epsilon_f4,  &
     lt=180._f4+epsilon_f4)

call nml_num_value_check(nf_span, 'NF_SPAN', 'and', ge=1,  &
     le=nml_max_n_input_files)
call manage_nml_var_memory(dim_nf_span = .TRUE.)
iapr(1:nf_span) = iapr_(1:nf_span)

if (trim(varpfx) == '__NULL__') then   ! Must set default
   if (trim(run_suffix) == '__NULL__') then
      varpfx = trim(output_dir_root_w_sim_ID)//  &
           '/intermediary/varfile'
   else
      varpfx = trim(output_dir_root_w_sim_ID)//  &
           '/intermediary/varfile'//trim(run_suffix)
   end if
end if

!... GEN_VF_INPUT_COMPARISON (logical) only has two possible values

end subroutine proc_nml_PIPP_control


end module namelist_mod
