!```````````````````````````````````````````````````````````````````
module namelist_maxdims_mod

!+ Module that contains maximum dimensions/sizes of various namelist-related
!   items.
!
!  As is indicated below, many of these values are never used in the actual
!   model -- instead they are for temporary namelist input purposes (due to
!   Fortran features/limitations).

!~ 14.01.2012 (last modified; TIM)

use univ_maxlen_string_mod, only: MAX_FILE_NAME_LEN, MAX_FILE_PATH_LEN

implicit none

private

!============== May be used in the actual model code ===============

! Maximum length of any individual NON-FILENAME namelist string token
!  (currently = 256):
integer, parameter, public :: nml__strtoklen_max = MAX_FILE_NAME_LEN

! Maximum length of any individual FILENAME namelist string token
!  (currently = 4096):
integer, parameter, public :: nml_fn_strtoklen_max = MAX_FILE_PATH_LEN

! Maximum number of separate grids in a single PRAMS run:
integer, parameter, public :: nml_max_n_grids = 99
!===================================================================

!======= ONLY use in namelist code for temporary variables =========

! Since Fortran namelist input does not work with dynamically-allocated
!  objects, define (maximum) dimensions for static objects to use in the said
!  input process:
integer, parameter, public :: nml_max_n_z = 10000,  &
                              nml_max_n_zsubsfc = 1000,  &
                              nml_max_n_zsupersfc = 1000,  &
                              nml_max_n_sounding = 10000,  &
                              nml_max_n_input_files = 1000
!===================================================================

end module namelist_maxdims_mod
