!```````````````````````````````````````````````````````````````````
module namelist_utils_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine strrep_to_trace_coord(strrep, resultant_trace_coord, syntax_error)

!+ Convert a string representation into an object of type 'trace_coord_obj',
!   checking for consistency with the configured grid.

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use main_memory_mod, only: trace_coord_obj, n_grids, gd

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: strrep
type(trace_coord_obj), intent(OUT) :: resultant_trace_coord
logical, intent(OUT) :: syntax_error

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'strrep_to_trace_coord'

integer :: nc, ic_gs, ic_xs, ic_ys, iy, ix, igd

!=== Executable statements:

syntax_error = .FALSE.  ! No syntax error yet

nc = len_trim(strrep)    ! Total length of string representation

! Extract grid index specifier:

ic_gs = index(strrep(1:nc), ':')   ! Location of separator
if (ic_gs-1 < 2) then   ! Invalid strrep
   write (*, '(a)') trim(procedure_name)//': Grid specification incorrect'
   syntax_error = .TRUE.
   return
end if
read (strrep(2:ic_gs-1), *) igd

if (igd > n_grids .or. igd < 1) then
   write (*, '(a)') trim(procedure_name)//': Grid specification out-of-bounds'
   syntax_error = .TRUE.
   return
end if
resultant_trace_coord%igd = igd

! Extract x coordinate index:

ic_xs = ic_gs+index(strrep(ic_gs+1:nc), ',')   ! Location of separator
if (ic_xs-ic_gs < 3) then   ! Invalid strrep
   write (*, '(a)') trim(procedure_name)//': x-index specification incorrect'
   syntax_error = .TRUE.
   return
end if
read (strrep(ic_gs+2:ic_xs-1), *) ix

if (ix > gd(igd)%nxp .or. ix < 1) then
   write (*, '(a)')  &
        trim(procedure_name)//': x-index specification out-of-bounds'
   syntax_error = .TRUE.
   return
end if
resultant_trace_coord%ix = ix

! Extract y coordinate index:

ic_ys = ic_xs+index(strrep(ic_xs+1:nc), ')')   ! Location of separator
if (ic_ys-ic_xs < 2) then   ! Invalid strrep
   write (*, '(a)') trim(procedure_name)//': y-index specification incorrect'
   syntax_error = .TRUE.
   return
end if
read (strrep(ic_xs+1:ic_ys-1), *) iy

if (iy > gd(igd)%nyp .or. iy < 1) then
   write (*, '(a)')  &
        trim(procedure_name)//': y_index specification out-of-bounds'
   syntax_error = .TRUE.
   return
end if
resultant_trace_coord%iy = iy

end subroutine strrep_to_trace_coord


end module namelist_utils_mod
