!```````````````````````````````````````````````````````````````````
module nest_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine coeff_nest

!+ Calculate nesting interpolation/feedback coefficients

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4, ONE_f4
use main_memory_mod, only: jdim,  vctr1, vctr2, vctr3, gd, n_grids

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: ifm, icm, nrat, i_f, j_f, k_f, k_c
real(f4) :: alpha, et, ev

!=== Executable statements:

do ifm=2,n_grids
   icm = gd(ifm)%nxtnest

   if (icm /= 0) then
     
      nrat = gd(ifm)%nstratx
      alpha = ((ONE_f4/real(nrat, kind=f4))**2-ONE_f4)/24._f4
      do i_f=1,gd(ifm)%nxp
         et = -0.5_f4+real(2*mod(i_f+nrat-2, nrat)+1, kind=f4)/  &
              (2._f4*real(nrat, kind=f4))
         ev = -0.5_f4+real(mod(i_f+nrat-2,nrat)+1, kind=f4)/real(nrat, kind=f4)
         
         gd(ifm)%ei1(i_f) = et*(et-ONE_f4)*0.5_f4+alpha
         gd(ifm)%ei2(i_f) = (ONE_f4-et*et)-2._f4*alpha
         gd(ifm)%ei3(i_f) = et*(et+ONE_f4)*0.5_f4+alpha
         gd(ifm)%ei4(i_f) = (ev*ev-0.25_f4)*(1.5_f4-ev)/6._f4
         gd(ifm)%ei5(i_f) = (0.5_f4-ev)*(2.25_f4-ev*ev)*0.5_f4
         gd(ifm)%ei6(i_f) = (0.5_f4+ev)*(2.25_f4-ev*ev)*0.5_f4
         gd(ifm)%ei7(i_f) = (ev*ev-0.25_f4)*(1.5_f4+ev)/6._f4
      end do
      
      if (jdim == 1) then
         nrat = gd(ifm)%nstraty
         alpha = ((ONE_f4/real(nrat, kind=f4))**2-ONE_f4)/24._f4
         do j_f=1,gd(ifm)%nyp
            et = -0.5_f4+real(2*mod(j_f+nrat-2,nrat)+1, kind=f4)/  &
                 (2._f4*real(nrat, kind=f4))
            ev = -0.5_f4+real(mod(j_f+nrat-2,nrat)+1, kind=f4)/  &
                 real(nrat, kind=f4)
            
            gd(ifm)%ej1(j_f) = et*(et-ONE_f4)*0.5_f4+alpha
            gd(ifm)%ej2(j_f) = (ONE_f4-et*et)-2._f4*alpha
            gd(ifm)%ej3(j_f) = et*(et+ONE_f4)*0.5_f4+alpha
            gd(ifm)%ej4(j_f) = (ev*ev-0.25_f4)*(1.5_f4-ev)/6._f4
            gd(ifm)%ej5(j_f) = (0.5_f4-ev)*(2.25_f4-ev*ev)*0.5_f4
            gd(ifm)%ej6(j_f) = (0.5_f4+ev)*(2.25_f4-ev*ev)*0.5_f4
            gd(ifm)%ej7(j_f) = (ev*ev-0.25_f4)*(1.5_f4+ev)/6._f4
         end do
      end if
      
      do k_c=1,gd(icm)%nzp
         vctr1(k_c) = ZERO_f4
         vctr2(k_c) = ZERO_f4
         vctr3(k_c) = ZERO_f4
      end do
      
      do k_f=1,gd(ifm)%nzp
         k_c = gd(ifm)%kpm(k_f)
         gd(ifm)%ek1(k_f) = (gd(ifm)%zt(k_f)-gd(icm)%zt(k_c))*  &
              (gd(ifm)%zt(k_f)-gd(icm)%zt(k_c+1))/  &
              ((gd(icm)%zt(k_c-1)-gd(icm)%zt(k_c))*  &
              (gd(icm)%zt(k_c-1)-gd(icm)%zt(k_c+1)))
         gd(ifm)%ek2(k_f) = (gd(ifm)%zt(k_f)-gd(icm)%zt(k_c-1))*  &
              (gd(ifm)%zt(k_f)-gd(icm)%zt(k_c+1))/  &
              ((gd(icm)%zt(k_c)-gd(icm)%zt(k_c-1))*  &
              (gd(icm)%zt(k_c)-gd(icm)%zt(k_c+1)))
         gd(ifm)%ek3(k_f) = (gd(ifm)%zt(k_f)-gd(icm)%zt(k_c-1))*  &
              (gd(ifm)%zt(k_f)-gd(icm)%zt(k_c))/  &
              ((gd(icm)%zt(k_c+1)-gd(icm)%zt(k_c-1))*  &
              (gd(icm)%zt(k_c+1)-gd(icm)%zt(k_c)))
         if (k_f >= 2 .and. k_f <= gd(ifm)%nzp-1) then
            vctr1(k_c) = vctr1(k_c)+(gd(ifm)%zm(k_f)-gd(ifm)%zm(k_f-1))*  &
                 gd(ifm)%ek1(k_f)
            vctr2(k_c) = vctr2(k_c)+(gd(ifm)%zm(k_f)-gd(ifm)%zm(k_f-1))*  &
                 gd(ifm)%ek2(k_f)
            vctr3(k_c) = vctr3(k_c)+(gd(ifm)%zm(k_f)-gd(ifm)%zm(k_f-1))*  &
                 gd(ifm)%ek3(k_f)
         end if
      end do
      
      do k_f=2,gd(ifm)%nzp-1
         k_c = gd(ifm)%kpm(k_f)
         gd(ifm)%ek1(k_f) = gd(ifm)%ek1(k_f)-gd(icm)%dzt(k_c)*vctr1(k_c)
         gd(ifm)%ek2(k_f) = gd(ifm)%ek2(k_f)-gd(icm)%dzt(k_c)*vctr2(k_c)+ONE_f4
         gd(ifm)%ek3(k_f) = gd(ifm)%ek3(k_f)-gd(icm)%dzt(k_c)*vctr3(k_c)
      end do
      
      do k_f=1,gd(ifm)%nzp-1
         k_c = gd(ifm)%kpm(k_f+1)
         if (k_f == 1 .or. k_f == gd(ifm)%nzp-1 .or.  &
              k_c /= gd(ifm)%kpm(k_f)) then
            gd(ifm)%ek4(k_f) = ZERO_f4
            gd(ifm)%ek5(k_f) = ONE_f4
            gd(ifm)%ek6(k_f) = ZERO_f4
            gd(ifm)%ek7(k_f) = ZERO_f4
         else
            gd(ifm)%ek4(k_f) = gd(ifm)%ek4(k_f-1)+  &
                 (gd(ifm)%zm(k_f)-gd(ifm)%zm(k_f-1))  &
                 *(-gd(ifm)%ek1(k_f))*gd(icm)%dzt(k_c-1)
            gd(ifm)%ek5(k_f) = gd(ifm)%ek5(k_f-1)+  &
                 (gd(ifm)%zm(k_f)-gd(ifm)%zm(k_f-1))  &
                 *(gd(ifm)%ek1(k_f)*gd(icm)%dzt(k_c-1)-  &
                 gd(ifm)%ek2(k_f)*gd(icm)%dzt(k_c))
            gd(ifm)%ek6(k_f) = gd(ifm)%ek6(k_f-1)+  &
                 (gd(ifm)%zm(k_f)-gd(ifm)%zm(k_f-1))  &
                 *(gd(ifm)%ek2(k_f)*gd(icm)%dzt(k_c)-  &
                 gd(ifm)%ek3(k_f)*gd(icm)%dzt(k_c+1))
            gd(ifm)%ek7(k_f) = gd(ifm)%ek7(k_f-1)+  &
                 (gd(ifm)%zm(k_f)-gd(ifm)%zm(k_f-1))  &
                 *gd(ifm)%ek3(k_f)*gd(icm)%dzt(k_c+1)
         end if
      end do
      do k_f=1,gd(ifm)%nzp
         gd(ifm)%fbcf(k_f,1) = gd(icm)%dzt(gd(ifm)%kpm(k_f))/  &
              (real(gd(ifm)%nstraty)*gd(ifm)%dzt(k_f))
         gd(ifm)%fbcf(k_f,2) = gd(icm)%dzt(gd(ifm)%kpm(k_f))/  &
              (real(gd(ifm)%nstratx)*gd(ifm)%dzt(k_f))
         gd(ifm)%fbcf(k_f,3) = ONE_f4/  &
              real(gd(ifm)%nstratx*gd(ifm)%nstraty, kind=f4)
         gd(ifm)%fbcf(k_f,4) = gd(icm)%dzt(gd(ifm)%kpm(k_f))/  &
              (real(gd(ifm)%nstratx*gd(ifm)%nstraty, kind=f4)*gd(ifm)%dzt(k_f))
      end do
      
   end if
end do

end subroutine coeff_nest


!```````````````````````````````````````````````````````````````````
subroutine fdback(ac, af, dc, df, nzc, nxc, nyc, nzf, nxf, nyf, ifm, vnam,  &
     sumflg)

!+ Nested grid to parent grid feedback calculations - uses NG density
!   (serial run version)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4, ONE_f4
use main_memory_mod, only: jdim, gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: nzc, nxc, nyc, nzf, nxf, nyf, ifm
real(f4), dimension(nzc,nxc,nyc), intent(IN) :: dc
real(f4), dimension(nzf,nxf,nyf), intent(IN) :: af, df
character(len=*), intent(IN) :: vnam

real(f4), dimension(nzc,nxc,nyc), intent(INOUT) :: ac, sumflg

!== Local declarations:

integer, parameter :: scheme = 0   ! 0 = original, 1 = new 

integer :: ibeg, jbeg, kbeg, iend, jend, kend, icm, iinc, jinc, kv, ifbcf,  &
     i_c, j_c, k_c, i_f, j_f, k_f
real(f4), dimension(nzc,nxc,nyc) :: adf  ! Avg NG density (each PG box)

!=== Executable statements:

icm = gd(ifm)%nxtnest
sumflg(:,:,:) = ZERO_f4
adf(:,:,:) = ZERO_f4
ibeg = 2
jbeg = 1+jdim
kbeg = 2
iend = nxf-1
jend = nyf-jdim
kend = nzf-1
iinc = 1
jinc = 1
kv = 0

if (vnam == 'u') then
   ibeg = 1 + gd(ifm)%nstratx
   iend = nxf - 1 - gd(ifm)%nstratx
   iinc = gd(ifm)%nstratx
   ifbcf = 1
else if (vnam == 'v') then
   jbeg = 1 + gd(ifm)%nstraty * jdim
   jend = nyf - (1 + gd(ifm)%nstraty) * jdim
   jinc = gd(ifm)%nstraty
   ifbcf = 2
else if (vnam == 'w' .or. vnam == 'terr') then
   if (vnam == 'w') then
      kbeg = 1 + gd(ifm)%nrz(gd(ifm)%kpm(2))
      kend = nzf - 1 - gd(ifm)%nrz(gd(ifm)%kpm(nzf-1))
      kv = 1
   else
      kbeg = 1
      kend = 1
   end if
   ifbcf = 3
else
   ifbcf = 4
end if

k_f = kbeg
do while (k_f <= kend)
   k_c = gd(ifm)%kpm(k_f)
   if (vnam == 'terr') k_c = 1

   do j_f = jbeg,jend,jinc
      j_c = gd(ifm)%jpm(j_f)
      if (vnam == 'p' .or. vnam == 'terr') then
         do i_f = ibeg,iend,iinc
            i_c = gd(ifm)%ipm(i_f)
            ac(k_c,i_c,j_c) = ac(k_c,i_c,j_c) * sumflg(k_c,i_c,j_c)  &
                 + af(k_f,i_f,j_f) * gd(ifm)%fbcf(k_f,ifbcf)
            sumflg(k_c,i_c,j_c) = ONE_f4
         end do
      else if (vnam == 'w') then
         do i_f = ibeg,iend,iinc
            i_c = gd(ifm)%ipm(i_f)
            ac(k_c,i_c,j_c) = ac(k_c,i_c,j_c) * sumflg(k_c,i_c,j_c)  &
                 + af(k_f,i_f,j_f) * gd(ifm)%fbcf(k_f,ifbcf)  &
                 * (df(k_f,i_f,j_f) + df(k_f+kv,i_f,j_f))
            adf(k_c,i_c,j_c) = adf(k_c,i_c,j_c)+  &
                 gd(ifm)%fbcf(k_f,ifbcf)*(df(k_f,i_f,j_f)+df(k_f+kv,i_f,j_f))
            sumflg(k_c,i_c,j_c) = ONE_f4
         end do
      else
         if (scheme == 1) then
            do i_f = ibeg,iend,iinc
               i_c = gd(ifm)%ipm(i_f)
               ac(k_c,i_c,j_c) = ac(k_c,i_c,j_c) * sumflg(k_c,i_c,j_c)  &
                    + af(k_f,i_f,j_f) * gd(ifm)%fbcf(k_f,ifbcf)*df(k_f,i_f,j_f)
               adf(k_c,i_c,j_c) = adf(k_c,i_c,j_c)+  &
                    gd(ifm)%fbcf(k_f,ifbcf)*df(k_f,i_f,j_f)
               sumflg(k_c,i_c,j_c) = ONE_f4
            end do
         else
            do i_f = ibeg,iend,iinc
               i_c = gd(ifm)%ipm(i_f)
               ac(k_c,i_c,j_c) = ac(k_c,i_c,j_c) * sumflg(k_c,i_c,j_c)  &
                    + af(k_f,i_f,j_f) * gd(ifm)%fbcf(k_f,ifbcf)*df(k_f,i_f,j_f)
               sumflg(k_c,i_c,j_c) = ONE_f4
            end do
         end if

      end if
   end do
   k_f = k_f+1
   if (vnam == 'w') then
      k_f = k_f + gd(ifm)%nrz(gd(ifm)%kpm(k_f)) - 1
   end if
end do

do k_c = gd(ifm)%kpm(kbeg),gd(ifm)%kpm(kend)
   do j_c = gd(ifm)%jpm(jbeg),gd(ifm)%jpm(jend)
      
      if (scheme == 1) then

         if (vnam /= 'p' .and. vnam /= 'terr') then
            do i_c = gd(ifm)%ipm(ibeg),gd(ifm)%ipm(iend) 
               ac(k_c,i_c,j_c) = ac(k_c,i_c,j_c)/adf(k_c,i_c,j_c)
            end do
         end if
         
      else

         if (vnam == 'w') then
            
            do i_c = gd(ifm)%ipm(ibeg),gd(ifm)%ipm(iend)
               ac(k_c,i_c,j_c) = ac(k_c,i_c,j_c)/  &
                    (dc(k_c,i_c,j_c)+dc(k_c+kv,i_c,j_c))
            end do
            
         else if (vnam /= 'p' .and. vnam /= 'terr') then
           
            do i_c = gd(ifm)%ipm(ibeg),gd(ifm)%ipm(iend) 
               ac(k_c,i_c,j_c) = ac(k_c,i_c,j_c)/dc(k_c,i_c,j_c)
            end do

         end if
     
      end if

   end do
end do

end subroutine fdback


!```````````````````````````````````````````````````````````````````
subroutine fdbackp(ivarn, af, ac, mtp, df, dfu, dfv, m1, m2, m3, ifm, icm,  &
     ia, iz, ja, jz, i0, j0, ibndc, nestratx, nestraty, mynum)

!+ Nested grid to parent grid feedback calculations - uses NG density
!   (parallel run version)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4
use main_memory_mod, only: gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: ivarn, mtp, m1, m2, m3, ifm, icm, ia, iz, ja, jz, i0,  &
     j0, ibndc, nestratx, nestraty, mynum
real(f4), dimension(m1,m2,m3), intent(IN) :: af, df, dfu, dfv

real(f4), dimension(mtp), intent(INOUT) :: ac

!== Local declarations:

integer :: kptscm, ikptscm, ibeg, jbeg, kbeg, iend, jend, kend, iinc, jinc,  &
     kv, ifbcf, i_f, j_f, j_c, k_f, k_c, indcf

!=== Executable statements:

! ivarn = variable types 1 = u
!                        2 = v
!                        3 = w
!                        4 = p
!                        5 = scalar

! Local variables ia, iz, ja, and jz in this subroutine refer not to the
! limits of prognostic points but to the limits of points on this fm node
! to be averaged to a given cm node.

kptscm = gd(ifm)%kpm(m1-1)-gd(ifm)%kpm(2)+1
ikptscm = kptscm*(gd(ifm)%ipm(iz+i0)-gd(ifm)%ipm(ia+i0)+1)
ibeg = ia
jbeg = ja
kbeg = 2
iend = iz
jend = jz
kend = m1 - 1
iinc = 1
jinc = 1
kv = 0

ac(1:mtp) = ZERO_f4

if (ivarn == 1) then
   do ibeg=ia+i0,ia+i0+nestratx
      if (mod(ibeg-1, nestratx) == 0) exit
   end do
   ibeg = ibeg-i0
   iinc = nestratx
   ifbcf = 1
else if (ivarn == 2) then
   do jbeg=ja+j0,ja+j0+nestraty
      if (mod(jbeg-1, nestraty) == 0) exit
   end do
   jbeg = jbeg-j0
   jinc = nestraty
   ifbcf = 2
else if (ivarn == 3) then
   kbeg = 1+gd(ifm)%nrz(gd(ifm)%kpm(2))
   kend = m1-2
   kv = 1
   ifbcf = 3
else
   ifbcf = 4
end if

k_f = kbeg
1 continue
k_c = gd(ifm)%kpm(k_f)-gd(ifm)%kpm(2)+1
if (ivarn == 4) then
   do j_f=jbeg,jend,jinc
      j_c=(gd(ifm)%jpm(j_f+j0)-gd(ifm)%jpm(ja+j0))*ikptscm
      do i_f=ibeg,iend,iinc
         indcf = k_c+j_c+(gd(ifm)%ipm(i_f+i0)-gd(ifm)%ipm(ia+i0))*kptscm
         ac(indcf) = ac(indcf)+af(k_f,i_f,j_f)*gd(ifm)%fbcf(k_f,ifbcf)
      end do
   end do
else if (ivarn == 3) then
   do j_f=jbeg,jend,jinc
      j_c=(gd(ifm)%jpm(j_f+j0)-gd(ifm)%jpm(ja+j0))*ikptscm
      do i_f=ibeg,iend,iinc
         indcf = k_c+j_c+(gd(ifm)%ipm(i_f+i0)-gd(ifm)%ipm(ia+i0))*kptscm
         ac(indcf) = ac(indcf)+af(k_f,i_f,j_f)*gd(ifm)%fbcf(k_f,ifbcf)  &
              *(df(k_f,i_f,j_f)+df(k_f+kv,i_f,j_f))
      end do
   end do
else if (ivarn == 2) then
   do j_f=jbeg,jend,jinc
      j_c=(gd(ifm)%jpm(j_f+j0)-gd(ifm)%jpm(ja+j0))*ikptscm
      do i_f=ibeg,iend,iinc
         indcf = k_c+j_c+(gd(ifm)%ipm(i_f+i0)-gd(ifm)%ipm(ia+i0))*kptscm
         ac(indcf) = ac(indcf)+af(k_f,i_f,j_f)*gd(ifm)%fbcf(k_f,ifbcf)*  &
              dfv(k_f,i_f,j_f)
      end do
   end do
else if (ivarn == 1) then
   do j_f=jbeg,jend,jinc
      j_c=(gd(ifm)%jpm(j_f+j0)-gd(ifm)%jpm(ja+j0))*ikptscm
      do i_f=ibeg,iend,iinc
         indcf = k_c+j_c+(gd(ifm)%ipm(i_f+i0)-gd(ifm)%ipm(ia+i0))*kptscm
         ac(indcf) = ac(indcf)+af(k_f,i_f,j_f)*gd(ifm)%fbcf(k_f,ifbcf)*  &
              dfu(k_f,i_f,j_f)
      end do
   end do
else
   do j_f=jbeg,jend,jinc
      j_c=(gd(ifm)%jpm(j_f+j0)-gd(ifm)%jpm(ja+j0))*ikptscm
      do i_f=ibeg,iend,iinc
         indcf = k_c+j_c+(gd(ifm)%ipm(i_f+i0)-gd(ifm)%ipm(ia+i0))*kptscm
         ac(indcf) = ac(indcf)+af(k_f,i_f,j_f)*gd(ifm)%fbcf(k_f,ifbcf)*  &
              df(k_f,i_f,j_f)
      end do
   end do
end if
k_f = k_f+1
if (ivarn == 3) k_f = k_f+gd(ifm)%nrz(gd(ifm)%kpm(k_f))-1

if (k_f <= kend) goto 1

end subroutine fdbackp


end module nest_mod


!```````````````````````````````````````````````````````````````````
subroutine nest_feedback(ifm, icm)

!+ Feeds back finer mesh to a coarser mesh

use main_memory_mod, only: a, iscr1, gd, ia, iz, ja, jz, run_is_parallel
use vtables_mod, only: iucn, idn0un, ivcn, idn0vn, iwcn, idn0n, ipcn, nsclr,  &
     isclrp, iscp, naersclr, iaersclp, iaerscp, ipi0n, ithpn, irtgtn, THP_vidx

use nest_mod, only: fdback
use grid_mem_mod, only: switch_to_Nscalar, switch_to_Ascalar

implicit none

!=== Argument declarations:

integer, intent(IN) :: ifm, icm

!== Local declarations:

integer :: nf, nc, isc1, isc2

!=== Executable statements:

! Feed back the finer mesh to the coarser mesh:

if (icm == 0) return    ! Return if not a nested grid

call fdback(a(iucn(icm)), a(iucn(ifm)), a(idn0un(icm)), a(idn0un(ifm)),  &
     gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp,  &
     gd(ifm)%nyp, ifm, 'u', a(iscr1))

if (gd(icm)%nstbot == 1) then
   if (run_is_parallel) then
      call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(icm)%mia,  &
           gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm, a(iucn(icm)), 'U')
   else
      call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja, jz, icm,  &
           a(iucn(icm)), 'U')
   end if
end if

if (gd(icm)%nsttop == 1) then
   if (run_is_parallel) then
      call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(icm)%mia,  &
           gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm, a(iucn(icm)), 'U')
   else
      call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja, jz, icm,  &
           a(iucn(icm)), 'U')
   end if
end if

call fdback(a(ivcn(icm)), a(ivcn(ifm)), a(idn0vn(icm)), a(idn0vn(ifm)),  &
     gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp,  &
     gd(ifm)%nyp, ifm, 'v', a(iscr1))

if (gd(icm)%nstbot == 1) then
   if (run_is_parallel) then
      call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(icm)%mia,  &
           gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm, a(ivcn(icm)), 'V')
   else
      call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja, jz, icm,  &
           a(ivcn(icm)), 'V')
   end if
end if

if (gd(icm)%nsttop == 1) then
   if (run_is_parallel) then
      call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(icm)%mia,  &
           gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm, a(ivcn(icm)), 'V')
   else
      call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja, jz, icm,  &
           a(ivcn(icm)), 'V')
   end if
end if

call fdback(a(iwcn(icm)), a(iwcn(ifm)), a(idn0n(icm)), a(idn0n(ifm)),  &
     gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp,  &
     gd(ifm)%nyp, ifm, 'w', a(iscr1))

call fdback(a(ipcn(icm)), a(ipcn(ifm)), a(idn0n(icm)), a(idn0n(ifm)),  &
     gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp,  &
     gd(ifm)%nyp, ifm, 'p', a(iscr1))

if (gd(icm)%nstbot == 1) then
   if (run_is_parallel) then
      call botset_atm_mass(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
           gd(icm)%mia, gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm,  &
           a(ipcn(icm)), a(ipi0n(icm)), 'P')
   else
      call botset_atm_mass(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja,  &
           jz, icm, a(ipcn(icm)), a(ipi0n(icm)), 'P')
   end if
end if

if (gd(icm)%nsttop == 1) then
   if (run_is_parallel) then
      call topset_atm_mass(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
           gd(icm)%mia, gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm,  &
           a(ipcn(icm)), a(ipi0n(icm)), a(ithpn(icm)), a(irtgtn(icm)), 'P')
   else
      call topset_atm_mass(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja,  &
           jz, icm, a(ipcn(icm)), a(ipi0n(icm)), a(ithpn(icm)),  &
           a(irtgtn(icm)), 'P')
   end if
end if

! Feed back "normal" scalars:

do nf=1,nsclr(ifm)
   do nc=1,nsclr(icm)
      if (isclrp(nf,ifm) == isclrp(nc,icm)) then
         call switch_to_Nscalar(icm, nc)
         isc1 = iscp
         call switch_to_Nscalar(ifm, nf)
         isc2 = iscp

         call fdback(a(isc1), a(isc2), a(idn0n(icm)), a(idn0n(ifm)),  &
              gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp,  &
              gd(ifm)%nxp, gd(ifm)%nyp, ifm, 't', a(iscr1))

         if (gd(icm)%nstbot == 1) then
            if (run_is_parallel) then
               call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
                    gd(icm)%mia, gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm,  &
                    a(isc1), 'T')
            else
               call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja,  &
                    jz, icm, a(isc1), 'T')
            end if
         end if

         if (gd(icm)%nsttop == 1) then
            if (run_is_parallel) then
               if (isclrp(nf,ifm) == THP_vidx) then     ! Scalar is theta
                  call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
                       gd(icm)%mia, gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz,  &
                       icm, a(isc1), 'T')
               else       ! Other scalar
                  call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
                       gd(icm)%mia, gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz,  &
                       icm, a(isc1), 'S')
               end if
            else
               if (isclrp(nf,ifm) == THP_vidx) then     ! Scalar is theta
                  call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz,  &
                       ja, jz, icm, a(isc1), 'T')
               else       ! Other scalar
                  call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz,  &
                       ja, jz, icm, a(isc1), 'S')
               end if
            end if
         end if

      end if
   end do
end do

! Feed back "aerosol" scalars:

do nf=1,naersclr(ifm)
   do nc=1,naersclr(icm)
      if (iaersclp(nf,ifm) == iaersclp(nc,icm)) then
         call switch_to_Ascalar(icm, nc)
         isc1 = iaerscp
         call switch_to_Ascalar(ifm, nf)
         isc2 = iaerscp

         call fdback(a(isc1), a(isc2), a(idn0n(icm)), a(idn0n(ifm)),  &
              gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp,  &
              gd(ifm)%nxp, gd(ifm)%nyp, ifm, 't', a(iscr1))

         if (gd(icm)%nstbot == 1) then
            if (run_is_parallel) then
               call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
                    gd(icm)%mia, gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm,  &
                    a(isc1), 'A')
            else
               call botset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja,  &
                    jz, icm, a(isc1), 'A')
            end if
         end if

         if (gd(icm)%nsttop == 1) then
            if (run_is_parallel) then
               call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp,  &
                    gd(icm)%mia, gd(icm)%miz, gd(icm)%mja, gd(icm)%mjz, icm,  &
                    a(isc1), 'A')
            else
               call topset(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, ia, iz, ja,  &
                    jz, icm, a(isc1), 'A')
            end if
         end if

      end if
   end do
end do

end subroutine nest_feedback
