!```````````````````````````````````````````````````````````````````
module nest_interpolation_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine eintp(ac, as, n1m, n2m, n3m, n1f, n2f, n3f, ifm, ndim, ispnt, i0, j0)

!+ Interpolates nested grid values from coarser grid values

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: jdim, gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1m, n2m, n3m, ifm, ndim, n1f, n2f, n3f, i0, j0
character(len=*), intent(IN) :: ispnt

real(f4), dimension(n1m,n2m,n3m), intent(INOUT) :: ac, as

!== Local declarations:

integer :: icm, k1, k2, jc1, jc, kc, if, jf, kf

!=== Executable statements:

icm = gd(ifm)%nxtnest
if (icm == 0) return  ! Don't execute for coarsest grid (not a nest)

k1 = max(1, gd(ifm)%kpm(2)-2)
k2 = min(gd(icm)%nzp, gd(ifm)%kpm(max(1, n1f-1))+2)

if (ndim > 1) then

   if (ndim == 2) then
      k1 = 1
      k2 = 1
   end if
   
   if (ispnt == 'u') then
      do jc1=gd(ifm)%jpm(1)-jdim,gd(ifm)%jpm(n3f)+jdim
         jc = jc1-j0
         do if=1,n2f-1
            do kc=k1,k2
               as(kc,if,jc) = gd(ifm)%ei4(if)*ac(kc,gd(ifm)%ipm(if)-i0-2,jc)  &
                    +gd(ifm)%ei5(if)*ac(kc,gd(ifm)%ipm(if)-i0-1,jc)  &
                    +gd(ifm)%ei6(if)*ac(kc,gd(ifm)%ipm(if)-i0,  jc)  &
                    +gd(ifm)%ei7(if)*ac(kc,gd(ifm)%ipm(if)-i0+1,jc)
            end do
         end do
      end do
   else
      do jc1=gd(ifm)%jpm(1)-jdim,gd(ifm)%jpm(n3f)+jdim
         jc = jc1-j0
         do if=1,n2f
            do kc=k1,k2
               as(kc,if,jc) = gd(ifm)%ei1(if)*ac(kc,gd(ifm)%ipm(if)-i0-1,jc)  &
                    +gd(ifm)%ei2(if)*ac(kc,gd(ifm)%ipm(if)-i0  ,jc)  &
                    +gd(ifm)%ei3(if)*ac(kc,gd(ifm)%ipm(if)-i0+1,jc)
            end do
         end do
      end do
   end if

   if (ispnt == 'v' .and. jdim == 1) then
      do jf=1,n3f-1
         do if=1,n2f
            do kc=k1,k2
               ac(kc,if,jf) = gd(ifm)%ej4(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0-2)  &
                    +gd(ifm)%ej5(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0-1)  &
                    +gd(ifm)%ej6(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0  )  &
                    +gd(ifm)%ej7(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0+1)
            end do
         end do
      end do
   else if (jdim == 1) then
      if (ndim == 2) then
         do jf=1,n3f
            do if=1,n2f
               do kc=k1,k2
                  ac(kc,if,jf) = gd(ifm)%ej1(jf)*  &
                       as(kc,if,gd(ifm)%jpm(jf)-j0-1)  &
                       +gd(ifm)%ej2(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0  )  &
                       +gd(ifm)%ej3(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0+1)
               end do
            end do
         end do

         do jf=1,n3f
            do if=1,n2f
               do kf=k1,k2
                  as(kf,if,jf) = ac(kf,if,jf)
               end do
            end do
         end do
         return
      else
         do jf=1,n3f
            do if=1,n2f
               do kc=k1,k2
                  ac(kc,if,jf) = gd(ifm)%ej1(jf)*  &
                       as(kc,if,gd(ifm)%jpm(jf)-j0-1)  &
                       +gd(ifm)%ej2(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0  )  &
                       +gd(ifm)%ej3(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0+1)
               end do
            end do
         end do
      end if
   else
      if (ndim == 2) then
!            do if=1,n2f
!               do kc=k1,k2
!                  ac(kc,if,1)=as(kc,if,1)
!               enddo
!            enddo
         return
      else
         do if=1,n2f
            do kc=k1,k2
               ac(kc,if,1) = as(kc,if,1)
            end do
         end do
      end if
   end if

end if

if (ispnt == 'w') then
   do jf=1,n3f
      do if=1,n2f
         do kf=1,n1f
            kc = gd(ifm)%kpm(min(n1f, kf+1))
            as(kf,if,jf) = gd(ifm)%ek4(kf)*ac(max(1, kc-2),if,jf)  &
                          +gd(ifm)%ek5(kf)*ac(                kc-1 ,if,jf)  &
                          +gd(ifm)%ek6(kf)*ac(                kc   ,if,jf)  &
                          +gd(ifm)%ek7(kf)*ac(min(gd(icm)%nzp-1, kc+1),if,jf)
         end do
      end do
   end do
else
   do jf=1,n3f
      do if=1,n2f
         do kf=1,n1f
            as(kf,if,jf) = gd(ifm)%ek1(kf)*ac(gd(ifm)%kpm(kf)-1,if,jf)  &
                          +gd(ifm)%ek2(kf)*ac(gd(ifm)%kpm(kf)  ,if,jf)  &
                          +gd(ifm)%ek3(kf)*ac(gd(ifm)%kpm(kf)+1,if,jf)
         end do
      end do
   end do
end if

end subroutine eintp


!```````````````````````````````````````````````````````````````````
subroutine eintp_2Dsfc(ac, as, n2m, n3m, n2f, n3f, ifm, i0, j0)

!+ Interpolates nested grid values from coarser grid values (2D vars)

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: jdim, gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: n2m, n3m, ifm, n2f, n3f, i0, j0

real(f4), dimension(n2m,n3m), intent(INOUT) :: ac, as

!== Local declarations:

integer :: icm, jc1, jc, if, jf

!=== Executable statements:

icm = gd(ifm)%nxtnest
if (icm == 0) return  ! Don't execute for coarsest grid (not a nest)

do jc1=gd(ifm)%jpm(1)-jdim,gd(ifm)%jpm(n3f)+jdim
   jc = jc1-j0
   do if=1,n2f
      as(if,jc) = gd(ifm)%ei1(if)*ac(gd(ifm)%ipm(if)-i0-1,jc)  &
           +gd(ifm)%ei2(if)*ac(gd(ifm)%ipm(if)-i0,jc)  &
           +gd(ifm)%ei3(if)*ac(gd(ifm)%ipm(if)-i0+1,jc)
   end do
end do

if (jdim == 1) then
   
   do jf=1,n3f
      do if=1,n2f
         ac(if,jf) = gd(ifm)%ej1(jf)*  &
              as(if,gd(ifm)%jpm(jf)-j0-1)  &
              +gd(ifm)%ej2(jf)*as(if,gd(ifm)%jpm(jf)-j0  )  &
              +gd(ifm)%ej3(jf)*as(if,gd(ifm)%jpm(jf)-j0+1)
      end do
   end do
   
   do jf=1,n3f
      do if=1,n2f
         as(if,jf) = ac(if,jf)
      end do
   end do

end if

end subroutine eintp_2Dsfc


!```````````````````````````````````````````````````````````````````
subroutine eintp_3Dsfc(ac, as, n1m, n2m, n3m, n1f, n2f, n3f, ifm, i0, j0)

!+ Interpolates nested grid values from coarser grid values (3D vars)

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: jdim, gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1m, n2m, n3m, ifm, n1f, n2f, n3f, i0, j0
real(f4), dimension(n1m,n2m,n3m), intent(INOUT) :: ac, as

!== Local declarations:

integer :: icm, k1, k2, jc1, jc, kc, if, jf, kf

!=== Executable statements:

icm = gd(ifm)%nxtnest
if (icm == 0) return  ! Don't execute for coarsest grid (not a nest)

k1 = 1
k2 = n1f
   
do jc1=gd(ifm)%jpm(1)-jdim,gd(ifm)%jpm(n3f)+jdim
   jc = jc1-j0
   do if=1,n2f
      do kc=k1,k2
         as(kc,if,jc) = gd(ifm)%ei1(if)*ac(kc,gd(ifm)%ipm(if)-i0-1,jc)  &
              +gd(ifm)%ei2(if)*ac(kc,gd(ifm)%ipm(if)-i0  ,jc)  &
              +gd(ifm)%ei3(if)*ac(kc,gd(ifm)%ipm(if)-i0+1,jc)
      end do
   end do
end do

if (jdim == 1) then
  
   do jf=1,n3f
      do if=1,n2f
         do kc=k1,k2
            ac(kc,if,jf) = gd(ifm)%ej1(jf)*  &
                 as(kc,if,gd(ifm)%jpm(jf)-j0-1)  &
                 +gd(ifm)%ej2(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0)  &
                 +gd(ifm)%ej3(jf)*as(kc,if,gd(ifm)%jpm(jf)-j0+1)
         end do
      end do
   end do

   do jf=1,n3f
      do if=1,n2f
         do kf=k1,k2
            as(kf,if,jf) = ac(kf,if,jf)
         end do
      end do
   end do

end if

end subroutine eintp_3Dsfc


!```````````````````````````````````````````````````````````````````
subroutine fmrefs1D(ngbegin, ngend)

!+ Interpolates fine mesh 1D reference state variables from coarse grid vars

use univ_kind_defs_mod, only: f4
use body_specific_constants_mod, only: R_air, cp, p00
use main_memory_mod, only: vctr1, vctr2, vctr3, vctr4, vctr5, vctr6, vctr7,  &
     vctr8, gd
use main_memory_bspec_mod, only: gd_bd

implicit none

!=== Argument declarations:

integer, intent(IN) :: ngbegin, ngend

!== Local declarations:

integer :: ifm, icm, k
real(f4) :: c1, c2

!=== Executable statements:

c1 = R_air/(cp-R_air)
c2 = cp*(R_air/p00)**c1
do ifm=ngbegin,ngend
   icm = gd(ifm)%nxtnest
   if (icm >= 1) then

      do k=1,gd(icm)%nzp
         vctr1(k) = gd(icm)%th01d(k)*gd(icm)%dn01d(k)
         vctr2(k) = gd(icm)%u01d(k)*gd(icm)%dn01d(k)
         vctr3(k) = gd(icm)%v01d(k)*gd(icm)%dn01d(k)
         vctr4(k) = gd_bd(icm)%rt01d(k)*gd(icm)%dn01d(k)
      end do

      call eintp(gd(icm)%dn01d, gd(ifm)%dn01d, gd(icm)%nzp, 1, 1,  &
           gd(ifm)%nzp, 1, 1, ifm, 1, 't', 0, 0)
      call eintp(vctr1, vctr5, gd(icm)%nzp, 1, 1, gd(ifm)%nzp, 1, 1, ifm, 1,  &
           't', 0, 0)
      call eintp(vctr2, vctr6, gd(icm)%nzp, 1, 1, gd(ifm)%nzp, 1, 1, ifm, 1,  &
           't', 0, 0)
      call eintp(vctr3, vctr7, gd(icm)%nzp, 1, 1, gd(ifm)%nzp, 1, 1, ifm, 1,  &
           't', 0, 0)
      call eintp(vctr4, vctr8, gd(icm)%nzp, 1, 1, gd(ifm)%nzp, 1, 1, ifm, 1,  &
           't', 0, 0)

      do k=1,gd(ifm)%nzp
         gd(ifm)%th01d(k) = vctr5(k)/gd(ifm)%dn01d(k)
         gd(ifm)%u01d(k) = vctr6(k)/gd(ifm)%dn01d(k)
         gd(ifm)%v01d(k) = vctr7(k)/gd(ifm)%dn01d(k)
         gd_bd(ifm)%rt01d(k) = vctr8(k)/gd(ifm)%dn01d(k)
         gd(ifm)%pi01d(k) = c2*(gd(ifm)%dn01d(k)*gd(ifm)%th01d(k))**c1
      end do

   end if
end do

end subroutine fmrefs1D


!```````````````````````````````````````````````````````````````````
subroutine fmint3D(m1c, m2c, m3c, n1f, n2f, n3f, maxnzp, maxnxp, maxnyp,  &
     ifm, icm, nstbot, nsttop, jdim, initflg, idwt, irtgflg, vnam, varc,  &
     varf, dn0xc, dn0xf, scr1, scr2, toptf, vt2da, bx, by, bz)

!+ Interpolates 3D fine mesh variables

use univ_kind_defs_mod, only: f4
use nest_lib_mod, only: fillscr, fillvar, dnswt2
use parallel_nesting_mod, only: interpolate_dnscale_nBC

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1c, m2c, m3c, n1f, n2f, n3f, maxnzp, maxnxp, maxnyp,  &
     ifm, icm, nstbot, nsttop, jdim, initflg, idwt, irtgflg
real(f4), dimension(m1c,m2c,m3c), intent(IN) :: varc, dn0xc
real(f4), dimension(n1f,n2f,n3f), intent(IN) :: dn0xf
real(f4), dimension(n2f,n3f), intent(IN) :: toptf, vt2da
character(len=*), intent(IN) :: vnam

real(f4), dimension(n1f,n3f,2), intent(INOUT) :: bx
real(f4), dimension(n1f,n2f,2), intent(INOUT) :: by
real(f4), dimension(n2f,n3f,2), intent(INOUT) :: bz
real(f4), dimension(maxnzp,maxnxp,maxnyp), intent(INOUT) :: scr1, scr2
real(f4), dimension(n1f,n2f,n3f), intent(INOUT) :: varf

!== Local declarations:

integer :: ivartyp2
  
!=== Executable statements:

if (icm == 0) return   ! Return if not a nested grid

call fillscr(maxnzp, maxnxp, maxnyp, m1c, m2c, m3c, 1, m1c, scr1, varc)

if (idwt == 1) call dnswt2(maxnzp, maxnxp, maxnyp, m1c, m2c, m3c, scr1,  &
     dn0xc, vnam, 1)

!bob 16 may 2000:  Call parallel routine PAR_BINTP in place of sequential
!bob    routine EINTP for nesting interpolation if simulation is 3D and
!bob    for timestepping, not initialization.  This is to help eliminate
!bob    machine epsilon differences between parallel and sequential runs.
  
if (jdim == 1 .and. initflg == 0) then
 
   if (vnam == 'u') then
      ivartyp2 = 1
   else if (vnam == 'v') then
      ivartyp2 = 2
   else if (vnam == 'w') then
      ivartyp2 = 3
   else if (vnam == 't' .and. idwt == 0) then
      ivartyp2 = 4
   else if (vnam == 't' .and. idwt == 1) then
      ivartyp2 = 5
   end if

   call interpolate_dnscale_nBC(scr1, scr2, dn0xf, maxnzp, maxnxp, maxnyp,  &
        n1f, n1f, n2f, n3f, ifm, ivartyp2, 0, 0, 15, bx, by, bz)

else

   call eintp(scr1, scr2, maxnzp, maxnxp, maxnyp, n1f, n2f, n3f, ifm, 3,  &
        vnam, 0, 0)

   if (idwt == 1 .and. initflg >= 0) then
      call dnswt2(maxnzp, maxnxp, maxnyp, n1f, n2f, n3f, scr2, dn0xf, vnam, 2)
   end if

   if (initflg == 1) then
      call fillvar(maxnzp, maxnxp, maxnyp, n1f, n2f, n3f, 1, n1f, scr2, varf)
      if (irtgflg == 1) call rtgintrp(n1f, n2f, n3f, varf, vt2da, toptf, ifm,  &
           vnam)
   else
      call nstb(maxnzp, maxnxp, maxnyp, n1f, n2f, n3f, scr2, bx, by, bz,  &
           vnam, nstbot, nsttop, jdim)
   end if

end if

end subroutine fmint3D


!```````````````````````````````````````````````````````````````````
subroutine rtgintrp(n1, n2, n3, fld, vt2da, topt, igd, vpnt)

!+ Do special vertical interpolation (for nested grids)

use univ_kind_defs_mod, only: f4
use univ_interpolation_mod, only: inextrp_1D_arr_type01a
use univ_math_constants_mod, only: ONE_f4
use main_memory_mod, only: vctr1, vctr2, vctr3, ztop, gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3, igd
real(f4), dimension(n2,n3), intent(IN) :: vt2da, topt
character(len=*), intent(IN) :: vpnt

real(f4), dimension(n1,n2,n3), intent(INOUT) :: fld

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

! Do special vertical interpolation in case terrain on this grid
!  (topt) is different from what would be interpolated from the
!  coarser grid (vt2da):

do j=1,n3
   do i=1,n2
      if (vpnt == 'W') then
         do k=1,n1
            vctr1(k) = gd(igd)%zm(k)*(ONE_f4-vt2da(i,j)/ztop)+vt2da(i,j)
            vctr2(k) = gd(igd)%zm(k)*(ONE_f4-topt(i,j)/ztop)+topt(i,j)
            vctr3(k) = fld(k,i,j)
         end do
      else
         do k=1,n1
            vctr1(k) = gd(igd)%zt(k)*(ONE_f4-vt2da(i,j)/ztop)+vt2da(i,j)
            vctr2(k) = gd(igd)%zt(k)*(ONE_f4-topt(i,j)/ztop)+topt(i,j)
            vctr3(k) = fld(k,i,j)
         end do
      end if

      fld(1:n1,i,j) = inextrp_1D_arr_type01a(vctr3(1:n1), vctr1(1:n1),  &
                                             vctr2(1:n1))

   end do
end do

end subroutine rtgintrp


!```````````````````````````````````````````````````````````````````
subroutine nstb(mx1, mx2, mx3, n1, n2, n3, vt3da, bx, by, bz, vnam, nstbot,  &
     nsttop, jdim)

!+ Fills nested grid boundary points

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: mx1, mx2, mx3, n1, n2, n3, nstbot, nsttop, jdim
real(f4), dimension(mx1,mx2,mx3), intent(IN) :: vt3da
character(len=*), intent(IN) ::  vnam

real(f4), dimension(n1,n3,2), intent(INOUT) :: bx
real(f4), dimension(n1,n2,2), intent(INOUT) :: by
real(f4), dimension(n2,n3,2), intent(INOUT) :: bz

!== Local declarations:

integer :: nxfm, nyfm, nzfm, i, j, k

!=== Executable statements:

nxfm = n2
nyfm = n3
nzfm = n1
if (vnam == 'u') then
   nxfm = n2-1
else if (vnam == 'v') then
   nyfm = (n3-2)*jdim+1
else if (vnam == 'w') then
   nzfm = n1-1
end if

do j=1,n3
   do k=1,n1
      bx(k,j,1) = vt3da(k,1,j)
      bx(k,j,2) = vt3da(k,nxfm,j)
   end do
end do

if (jdim == 1) then
   do i=1,n2
      do k=1,n1
         by(k,i,1) = vt3da(k,i,1)
         by(k,i,2) = vt3da(k,i,nyfm)
      end do
   end do
end if

if (nstbot == 0) then
   do j=1,n3
      do i=1,n2
         bz(i,j,1) = vt3da(1,i,j)
      end do
   end do
end if

if (nsttop == 0) then
   do j=1,n3
      do i=1,n2
         bz(i,j,2) = vt3da(nzfm,i,j)
      end do
   end do
end if

end subroutine nstb


!```````````````````````````````````````````````````````````````````
subroutine nstbtnd(m1, m2, m3, ia, iz, ja, jz, ibcon, scp, sct, bx, by, bz,  &
     vnam, tymeinv, nstbot, nsttop, jdim)

!+ Sets nested grid boundaries

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz, ibcon, nstbot, nsttop, jdim
real(f4), intent(IN) :: tymeinv
real(f4), dimension(m1,m2,m3), intent(IN) :: scp
real(f4), dimension(m1,m3,2), intent(IN) :: bx
real(f4), dimension(m1,m2,2), intent(IN) :: by
real(f4), dimension(m2,m3,2), intent(IN) :: bz
character(len=*), intent(IN) :: vnam

real(f4), dimension(m1,m2,m3), intent(INOUT) :: sct

!== Local declarations:

integer :: i, j, k, nzfm, nxfm, nyfm, incia, inciz, incja, incjz

!=== Executable statements:

nzfm = m1
nxfm = iz+1
nyfm = jz+1
if (vnam == 'w') then
   nzfm = nzfm-1
else if (vnam == 'u') then
   nxfm = nxfm-1
else if (vnam == 'v') then
   nyfm = nyfm-1
end if

if (jdim == 0) nyfm = 1

incia = 0
inciz = 0
if (iand(ibcon, 1) /= 0) incia = 1
if (iand(ibcon, 2) /= 0 .and. vnam /= 'u') inciz = 1
incja = 0
incjz = 0
if (iand(ibcon, 4) /= 0) incja = jdim
if (iand(ibcon, 8) /= 0 .and. vnam /= 'v') incjz = jdim

if (nstbot == 0) then
   do j=ja,jz
      do i=ia,iz
         sct(1,i,j) = (bz(i,j,1) - scp(1,i,j)) * tymeinv
      end do
   end do
end if
if (nsttop == 0) then
   do j=ja,jz
      do i=ia,iz
         sct(nzfm,i,j) = (bz(i,j,2) - scp(nzfm,i,j)) * tymeinv
      end do
   end do
end if

if (iand(ibcon, 1) /= 0) then
   do j=ja-incja,jz+incjz
      do k=1,m1
         sct(k,1,j) = (bx(k,j,1) - scp(k,1,j)) * tymeinv
      end do
   end do
end if

if (iand(ibcon, 2) /= 0) then
   do j=ja-incja,jz+incjz
      do k=1,m1
         sct(k,nxfm,j) = (bx(k,j,2) - scp(k,nxfm,j)) * tymeinv
      end do
   end do
end if

if (jdim == 1) then
   if(iand(ibcon, 4) /= 0) then
      do i=ia-incia,iz+inciz
         do k=1,m1
            sct(k,i,1) = (by(k,i,1) - scp(k,i,1)) * tymeinv
         end do
      end do
   end if
   if (iand(ibcon, 8) /= 0) then
      do i=ia-incia,iz+inciz
         do k=1,m1
            sct(k,i,nyfm) = (by(k,i,2) - scp(k,i,nyfm)) * tymeinv
         end do
      end do
   end if
end if

end subroutine nstbtnd


!```````````````````````````````````````````````````````````````````
subroutine nstbtndA(m1, m2, m3, ia, iz, ja, jz, ibcon, scp, sct, bx, by, bz,  &
     vnam, tymeinv, nstbot, nsttop, jdim)

!+ Sets nested grid boundaries

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz, ibcon, nstbot, nsttop, jdim
real(f4), intent(IN) :: tymeinv
real(f4), dimension(m1,m2,m3), intent(IN) :: scp
real(f4), dimension(m1,m3,2), intent(IN) :: bx
real(f4), dimension(m1,m2,2), intent(IN) :: by
real(f4), dimension(m2,m3,2), intent(IN) :: bz
character(len=*), intent(IN) :: vnam

real(f4), dimension(m1,m2,m3), intent(INOUT) :: sct

!== Local declarations:

integer :: i, j, k, nzfm, nxfm, nyfm, incia, inciz, incja, incjz

!=== Executable statements:

nzfm = m1
nxfm = iz+1
nyfm = jz+1
if (vnam == 'w') then
   nzfm = nzfm-1
else if (vnam == 'u') then
   nxfm = nxfm-1
else if (vnam == 'v') then
   nyfm = nyfm-1
end if

if (jdim == 0) nyfm = 1

incia = 0
inciz = 0
if (iand(ibcon, 1) /= 0) incia = 1
if (iand(ibcon, 2) /= 0 .and. vnam /= 'u') inciz = 1
incja = 0
incjz = 0
if (iand(ibcon, 4) /= 0) incja = jdim
if (iand(ibcon, 8) /= 0 .and. vnam /= 'v') incjz = jdim

if (nstbot == 0) then
   do j=ja,jz
      do i=ia,iz
         sct(1,i,j) = (bz(i,j,1) - scp(1,i,j)) * tymeinv
      end do
   end do
end if
if (nsttop == 0) then
   do j=ja,jz
      do i=ia,iz
         sct(nzfm,i,j) = (bz(i,j,2) - scp(nzfm,i,j)) * tymeinv
      end do
   end do
end if

if (iand(ibcon, 1) /= 0) then
   do j=ja-incja,jz+incjz
      do k=1,m1
         sct(k,1,j) = (bx(k,j,1) - scp(k,1,j)) * tymeinv
      end do
   end do
end if

if (iand(ibcon, 2) /= 0) then
   do j=ja-incja,jz+incjz
      do k=1,m1
         sct(k,nxfm,j) = (bx(k,j,2) - scp(k,nxfm,j)) * tymeinv
      end do
   end do
end if

if (jdim == 1) then
   if (iand(ibcon, 4) /= 0) then
      do i=ia-incia,iz+inciz
         do k=1,m1
            sct(k,i,1) = (by(k,i,1) - scp(k,i,1)) * tymeinv
         end do
      end do
   end if
   if (iand(ibcon, 8) /= 0) then
      do i=ia-incia,iz+inciz
         do k=1,m1
            sct(k,i,nyfm) = (by(k,i,2) - scp(k,i,nyfm)) * tymeinv
         end do
      end do
   end if
end if

end subroutine nstbtndA


end module nest_interpolation_mod


!```````````````````````````````````````````````````````````````````
subroutine fmrefs3D(ifm)

!+ Interpolates fine mesh 3D reference state variables from coarse grid vars

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: gd, maxnzp, maxnxp, maxnyp, a, b, jdim, iscr1,  &
     iscr2
use vtables_mod, only: idn0n, itoptn, ivt2dan, ith0n, ipi0n, idn0un, idn0vn
use body_specific_constants_mod, only: R_air, cp, p00
use nest_lib_mod, only: fillscr, fillvar
use nest_interpolation_mod, only: eintp, rtgintrp, fmint3D
use PRAMS_init_mem_mod, only: fill_dn0uv

implicit none

!=== Argument declarations:

integer, intent(IN) :: ifm

!== Local declarations:

integer :: icm, i
real(f4) :: c1, c2

!=== Executable statements:

icm = gd(ifm)%nxtnest
if (icm == 0) return ! Do not execute this routine for a grid that is not nested

call fmint3D(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp,  &
     gd(ifm)%nyp, maxnzp, maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot,  &
     gd(ifm)%nsttop, jdim, 1, 0 ,0, 't', a(idn0n(icm)), a(idn0n(ifm)),  &
     a(idn0n(icm)), a(idn0n(ifm)), a(iscr1), a(iscr2), a(itoptn(ifm)),  &
     a(ivt2dan(ifm)), b(1), b(1), b(1))

call fmint3D(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp,  &
     gd(ifm)%nyp, maxnzp, maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot,  &
     gd(ifm)%nsttop, jdim, 1, 1 , 0, 't', a(ith0n(icm)), a(ith0n(ifm)),  &
     a(idn0n(icm)), a(idn0n(ifm)), a(iscr1), a(iscr2), a(itoptn(ifm)),  &
     a(ivt2dan(ifm)), b(1), b(1), b(1))

c1 = R_air/(cp-R_air)
c2 = cp*(R_air/p00)**c1
do i=0,gd(ifm)%nxyzp-1
   a(ipi0n(ifm)+i) = c2*(a(idn0n(ifm)+i)*a(ith0n(ifm)+i))**c1
end do

call fillscr(1, maxnxp, maxnyp, 1, gd(icm)%nxp, gd(icm)%nyp, 1, 1, a(iscr1),  &
     a(itoptn(icm)))
call eintp(a(iscr1), a(iscr2), 1, maxnxp, maxnyp, 1, gd(ifm)%nxp,  &
     gd(ifm)%nyp, ifm, 2, 't', 0, 0)
call fillvar(1, maxnxp, maxnyp, 1, gd(ifm)%nxp, gd(ifm)%nyp, 1, 1, a(iscr2),  &
     a(iscr1))

call rtgintrp(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, a(ith0n(ifm)), a(iscr1),  &
     a(itoptn(ifm)), ifm, 't')
call rtgintrp(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, a(ipi0n(ifm)), a(iscr1),  &
     a(itoptn(ifm)), ifm, 't')

! Define dn0u and dn0v:

call fill_dn0uv(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, a(idn0n(ifm)),  &
     a(idn0un(ifm)), a(idn0vn(ifm)))

end subroutine fmrefs3D


!```````````````````````````````````````````````````````````````````
subroutine fmdn0(ifm)

!+ Special interpolation of 3D density reference state

use main_memory_mod, only: a, iscr1, iscr2, maxnyp, maxnxp, gd
use vtables_mod, only: idn0n, idn0un, idn0vn, itoptn
use nest_lib_mod, only: fillscr, fillvar
use nest_interpolation_mod, only: eintp, rtgintrp
use PRAMS_init_mem_mod, only: fill_dn0uv

implicit none

!=== Argument declarations:

integer, intent(IN) :: ifm

!== Local declarations:

integer :: icm

!=== Executable statements:

! Special vertical interpolation of DN0 must be done after all other
! 3-D reference state and prognostic variables are interpolated;

icm = gd(ifm)%nxtnest
if (icm == 0) return   ! Return if not a nested grid

call fillscr(1, maxnxp, maxnyp, 1, gd(icm)%nxp, gd(icm)%nyp, 1, 1, a(iscr1),  &
     a(itoptn(icm)))
call eintp(a(iscr1), a(iscr2), 1, maxnxp, maxnyp, 1, gd(ifm)%nxp,  &
     gd(ifm)%nyp, ifm, 2, 't', 0, 0)
call fillvar(1, maxnxp, maxnyp, 1, gd(ifm)%nxp, gd(ifm)%nyp, 1, 1, a(iscr2),  &
     a(iscr1))

call rtgintrp(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, a(idn0n(ifm)), a(iscr1),  &
     a(itoptn(ifm)), ifm, 't')

! Define dn0u and dn0v:

call fill_dn0uv(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, a(idn0n(ifm)),  &
     a(idn0un(ifm)), a(idn0vn(ifm)))

end subroutine fmdn0


!```````````````````````````````````````````````````````````````````
subroutine prgintrp(n1c, n2c, n3c, m1c, m2c, m3c, i0c, j0c, ifm, initflg, mynum)

!+ Interpolate atmospheric variables from parent to nested grid boundaries

use main_memory_mod, only: icorflg, maxnzp, maxnxp, maxnyp, a, b, jdim,  &
     iscr2, iscr1, gd
use vtables_mod, only: iucn, idn0un, ibuy, ibux, ibuz, ibvx, ibvy, ibvz,  &
     ibwx, ibwy, ibwz, idn0n, ipcn, iwcn, ibpx, ibpy, ibpz, ivcn, idn0vn,  &
     itoptn, ivt2dan, ibsx, ibsy, ibsz, ind3dn, nsclr, isclrp, naersclr,  &
     iaersclp, ibax, ibay, ibaz
use nest_lib_mod, only: fillscr, fillvar
use nest_interpolation_mod, only: eintp, fmint3D

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1c, n2c, n3c, m1c, m2c, m3c, i0c, j0c, ifm, initflg,  &
     mynum

!== Local declarations:

integer :: icm, nf, nc

!=== Executable statements:

! Temporarily fill VT2DA with interpolated topography from coarser grid:

icm = gd(ifm)%nxtnest
if (icm == 0) return   ! Return if not a nested grid

if (initflg == 1) then
   call fillscr(1, maxnxp, maxnyp, 1, n2c, n3c, 1, 1, a(iscr1), a(itoptn(icm)))
   call eintp(a(iscr1), a(iscr2), 1, maxnxp, maxnyp, 1, gd(ifm)%nxp,  &
        gd(ifm)%nyp, ifm, 2, 't', 0, 0)
   call fillvar(1, maxnxp, maxnyp, 1, gd(ifm)%nxp, gd(ifm)%nyp, 1, 1,  &
        a(iscr2), a(ivt2dan(ifm)))
end if

!*** Interpolate atmospheric variables:

! For velocity components:

call fmint3D(m1c, m2c, m3c, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, maxnzp,  &
     maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot, gd(ifm)%nsttop, jdim, initflg,  &
     1, 1, 'u', a(iucn(icm)), a(iucn(ifm)), a(idn0un(icm)), a(idn0un(ifm)),  &
     a(iscr1), a(iscr2), a(itoptn(ifm)), a(ivt2dan(ifm)), b(ibux(ifm)),  &
     b(ibuy(ifm)), b(ibuz(ifm)))

if (jdim == 1 .or. icorflg >= 1) then
   call fmint3D(m1c, m2c, m3c, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, maxnzp,  &
        maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot, gd(ifm)%nsttop, jdim,  &
        initflg, 1, 1, 'v', a(ivcn(icm)), a(ivcn(ifm)), a(idn0vn(icm)),  &
        a(idn0vn(ifm)), a(iscr1), a(iscr2), a(itoptn(ifm)), a(ivt2dan(ifm)),  &
        b(ibvx(ifm)), b(ibvy(ifm)), b(ibvz(ifm)))
end if
call fmint3D(m1c, m2c, m3c, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, maxnzp,  &
     maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot, gd(ifm)%nsttop, jdim, initflg,  &
     1, 1, 'w', a(iwcn(icm)), a(iwcn(ifm)), a(idn0n(icm)), a(idn0n(ifm)),  &
     a(iscr1), a(iscr2), a(itoptn(ifm)), a(ivt2dan(ifm)), b(ibwx(ifm)),  &
     b(ibwy(ifm)), b(ibwz(ifm)))
call fmint3D(m1c, m2c, m3c, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, maxnzp,  &
     maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot, gd(ifm)%nsttop, jdim, initflg,  &
     0, 1, 't', a(ipcn(icm)), a(ipcn(ifm)), a(idn0n(icm)), a(idn0n(ifm)),  &
     a(iscr1), a(iscr2), a(itoptn(ifm)), a(ivt2dan(ifm)), b(ibpx(ifm)),  &
     b(ibpy(ifm)), b(ibpz(ifm)))

! For "normal" scalars:

do nf=1,nsclr(ifm)
   do nc=1,nsclr(icm)
      if (isclrp(nf,ifm) == isclrp(nc,icm)) then
         call fmint3D(m1c, m2c, m3c, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp,  &
              maxnzp, maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot,  &
              gd(ifm)%nsttop, jdim, initflg, 1, 1, 't',  &
              a(ind3dn(icm,isclrp(nc,icm))),  &
              a(ind3dn(ifm,isclrp(nf,ifm))), a(idn0n(icm)), a(idn0n(ifm)),  &
              a(iscr1), a(iscr2), a(itoptn(ifm)), a(ivt2dan(ifm)),  &
              b(ibsx(nf,ifm)), b(ibsy(nf,ifm)), b(ibsz(nf,ifm)))
      end if
   end do
end do

! For "aerosol" scalars:

do nf=1,naersclr(ifm)
   do nc=1,naersclr(icm)
      if (iaersclp(nf,ifm) == iaersclp(nc,icm)) then
         call fmint3D(m1c, m2c, m3c, gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp,  &
              maxnzp, maxnxp, maxnyp, ifm, icm, gd(ifm)%nstbot,  &
              gd(ifm)%nsttop, jdim, initflg, 1, 1, 't',  &
              a(ind3dn(icm,iaersclp(nc,icm))),  &
              a(ind3dn(ifm,iaersclp(nf,ifm))), a(idn0n(icm)), a(idn0n(ifm)),  &
              a(iscr1), a(iscr2), a(itoptn(ifm)), a(ivt2dan(ifm)),  &
              b(ibax(nf,ifm)), b(ibay(nf,ifm)), b(ibaz(nf,ifm)))
      end if
   end do
end do

end subroutine prgintrp


!```````````````````````````````````````````````````````````````````
subroutine vfintrp(ifm)

!+ Interpolate parent grid reference state to nested grid reference state

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: tnudcent, tnudtop, gd, maxnxp, maxnyp, a, iscr1,  &
     iscr2
use vtables_mod, only: ivt2dan, ivarwtsn, idn0n, itoptn
use nest_lib_mod, only: fillscr, fillvar
use nest_interpolation_mod, only: eintp

implicit none

!=== Argument declarations:

integer, intent(IN) :: ifm

!== Local declarations:

integer :: icm

!=== Executable statements:

! If not nudging nested grids, return immediately:
if (tnudcent+tnudtop < 0.001_f4) return

icm = gd(ifm)%nxtnest
if (icm == 0) return

! Temporarily fill VT2DA with interpolated topography from coarser grid:

call fillscr(1, maxnxp, maxnyp, 1, gd(icm)%nxp, gd(icm)%nyp, 1, 1, a(iscr1),  &
     a(itoptn(icm)))
call eintp(a(iscr1), a(iscr2), 1, maxnxp, maxnyp, 1, gd(ifm)%nxp,  &
     gd(ifm)%nyp, ifm, 2, 't', 0, 0)
call fillvar(1, maxnxp, maxnyp, 1, gd(ifm)%nxp, gd(ifm)%nyp, 1, 1, a(iscr2),  &
     a(ivt2dan(ifm)))

! Interpolate varwts:

call fmint4(gd(icm)%nzp, gd(icm)%nxp, gd(icm)%nyp, gd(ifm)%nzp, gd(ifm)%nxp,  &
     gd(ifm)%nyp, a(ivarwtsn(icm)), a(ivarwtsn(ifm)), a(idn0n(icm)),  &
     a(idn0n(ifm)), a(ivt2dan(ifm)), ifm, icm, 't', 0)

end subroutine vfintrp


!```````````````````````````````````````````````````````````````````
subroutine fmint4(n1c, n2c, n3c, n1f, n2f, n3f, var1, var2, dn0xc, dn0xf,  &
     vt2da, ifm, icm, vpnt, idwt)

!+ Special nesting interpolations

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: a, iscr1, iscr2, maxnzp, maxnxp, maxnyp
use vtables_mod, only: itoptn
use nest_lib_mod, only: fillscr, dnswt2, fillvar
use nest_interpolation_mod, only: rtgintrp, eintp

implicit none

!=== Argument declarations:

integer, intent(IN) :: ifm, icm, idwt, n1c, n2c, n3c, n1f, n2f, n3f
real(f4), dimension(n1c,n2c,n3c), intent(IN) :: var1, dn0xc
real(f4), dimension(n1f,n2f,n3f), intent(IN) :: dn0xf
real(f4), dimension(n2f,n3f), intent(IN) :: vt2da
character(len=*), intent(IN) :: vpnt

real(f4), dimension(n1f,n2f,n3f), intent(INOUT) :: var2

!== Local declarations:

! NONE

!=== Executable statements:

if (icm == 0) return   ! Return if not nested grid

call fillscr(maxnzp, maxnxp, maxnyp, n1c, n2c, n3c, 1, n1c, a(iscr1), var1)

if (idwt == 1) call dnswt2(maxnzp, maxnxp, maxnyp, n1c, n2c, n3c, a(iscr1),  &
     dn0xc, vpnt, 1)

call eintp(a(iscr1), a(iscr2), maxnzp, maxnxp, maxnyp, n1f, n2f, n3f, ifm, 3,  &
     vpnt, 0, 0)

call fillvar(maxnzp, maxnxp, maxnyp, n1f, n2f, n3f, 1, n1f, a(iscr2), var2)

if (idwt == 1) call dnswt2(n1f, n2f, n3f, n1f, n2f, n3f, var2, dn0xf, vpnt, 2)

call rtgintrp(n1f, n2f, n3f, var2, vt2da, a(itoptn(ifm)), ifm, vpnt)

end subroutine fmint4


!```````````````````````````````````````````````````````````````````
subroutine nstbdriv

!+ Sets nested grid boundaries (driver)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use main_memory_mod, only: a, b, dtlv, i_grid, isstp, c_gd, jdim, nscl,  &
     naerscl, mzp, mxp, myp, ia, iz, ja, jz, ibndc
use vtables_mod, only: iup, iut, ibux, ibuy, ibuz, ivp, ivt, ibvx, ibvy,  &
     ibvz, iwp, iwt, ibwx, ibwy, ibwz, ipp, ipt, ibpx, ibpy, ibpz, iscp,  &
     isct, ibsx, ibsy, ibsz, iaerscp, iaersct, ibax, ibay, ibaz
use nest_interpolation_mod, only: nstbtnd
use grid_mem_mod, only: switch_to_Ascalar, switch_to_Nscalar

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: n
real(f4) :: tymeinvv, tymeinvs

!=== Executable statements:

tymeinvv = ONE_f4 / (dtlv * 0.5_f4 * real(c_gd%n_dt_per_pgdt+2-isstp, kind=f4))
tymeinvs = ONE_f4 / (c_gd%dt_dyn*real(c_gd%n_dt_per_pgdt+1-isstp, kind=f4))

call nstbtnd(mzp, mxp, myp, ia, iz, ja, jz, ibndc, a(iup), a(iut),  &
     b(ibux(i_grid)), b(ibuy(i_grid)), b(ibuz(i_grid)), 'u', tymeinvv,  &
     c_gd%nstbot, c_gd%nsttop, jdim)

call nstbtnd(mzp, mxp, myp, ia, iz, ja, jz, ibndc, a(ivp), a(ivt),  &
     b(ibvx(i_grid)), b(ibvy(i_grid)), b(ibvz(i_grid)), 'v', tymeinvv,  &
     c_gd%nstbot, c_gd%nsttop, jdim)

call nstbtnd(mzp, mxp, myp, ia, iz, ja, jz, ibndc, a(iwp), a(iwt),  &
     b(ibwx(i_grid)), b(ibwy(i_grid)), b(ibwz(i_grid)), 'w', tymeinvv,  &
     c_gd%nstbot, c_gd%nsttop, jdim)

call nstbtnd(mzp, mxp, myp, ia, iz, ja, jz, ibndc, a(ipp), a(ipt),  &
     b(ibpx(i_grid)), b(ibpy(i_grid)), b(ibpz(i_grid)), 'p', tymeinvv,  &
     c_gd%nstbot, c_gd%nsttop, jdim)

do n=1,nscl    ! "Normal" scalars
   call switch_to_Nscalar(i_grid, n)

   call nstbtnd(mzp, mxp, myp, ia, iz, ja, jz, ibndc, a(iscp), a(isct),  &
        b(ibsx(n,i_grid)), b(ibsy(n,i_grid)), b(ibsz(n,i_grid)), 't',  &
        tymeinvs, c_gd%nstbot, c_gd%nsttop, jdim)
end do

do n=1,naerscl    ! "Aerosol" scalars
   call switch_to_Ascalar(i_grid, n)
   call nstbtnd(mzp, mxp, myp, ia, iz, ja, jz, ibndc, a(iaerscp), a(iaersct),  &
       b(ibax(n,i_grid)), b(ibay(n,i_grid)), b(ibaz(n,i_grid)), 't',  &
       tymeinvs, c_gd%nstbot, c_gd%nsttop, jdim)
end do

end subroutine nstbdriv
