! Based on code with:
!###########################################################################
!  Copyright (C)  1990, 1995, 1999, 2000 - All Rights Reserved
!  Regional Atmospheric Modeling System - RAMS
!  Mission Research Corporation / *ASTeR Division
!###########################################################################

!```````````````````````````````````````````````````````````````````
module nest_lib_mod

!~ 23.09.2008 (last modified; TIM)

!+

implicit none

private

public :: fillvar, fillscr, dnswt2

contains


!```````````````````````````````````````````````````````````````````
subroutine fillscr_den(m1,m2,m3,n1,n2,n3,scr,var,i0,j0)

!~ 20.11.2008 (last modified; TIM)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, n1, n2, n3, i0, j0
real(f4), dimension(n1,n2,n3), intent(IN) :: var

real(f4), dimension(m1,m2,m3), intent(INOUT) :: scr

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

do j=1,n3
   do i=1,n2
      do k=1,n1
         scr(k,i,j)=var(k,i+i0,j+j0)
      enddo
   enddo
enddo

end subroutine fillscr_den


!```````````````````````````````````````````````````````````````````
subroutine fillscr(m1,m2,m3,n1,n2,n3,k1,k2,scr,var)

!~ 20.11.2008 (last modified; TIM)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, n1, n2, n3, k1, k2
real(f4), dimension(n1,n2,n3), intent(IN) :: var

real(f4), dimension(m1,m2,m3), intent(INOUT) :: scr

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

do j = 1,n3
   do i = 1,n2
      do k = k1,k2
         scr(k,i,j) = var(k,i,j)
      enddo
   enddo
enddo

end subroutine fillscr


!```````````````````````````````````````````````````````````````````
subroutine fillvar(m1,m2,m3,n1,n2,n3,k1,k2,scr,var)

!~ 20.11.2008 (last modified; TIM)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, n1, n2, n3, k1, k2
real(f4), dimension(m1,m2,m3), intent(IN) :: scr

real(f4), dimension(n1,n2,n3), intent(INOUT) :: var

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

do j = 1,n3
   do i = 1,n2
      do k = k1,k2
         var(k,i,j) = scr(k,i,j)
      enddo
   enddo
enddo

end subroutine fillvar


!```````````````````````````````````````````````````````````````````
subroutine dnswt(nzmx,nxmx,nymx,n1,n2,n3,var1,dn0,vpnt  &
     ,jdim,idir)

!~ 20.11.2008 (last modified; TIM)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: nzmx, nxmx, nymx, n1, n2, n3, jdim, idir
real(f4), dimension(n1,n2,n3), intent(IN) :: dn0
character(len=*), intent(IN) :: vpnt

real(f4), dimension(nzmx,nxmx,nymx), intent(INOUT) :: var1

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

if(idir == 1) then
   if(vpnt == 't') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*dn0(k,i,j)
            enddo
         enddo
      enddo
   elseif(vpnt == 'u') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*0.5_f4*(dn0(k,i,j)+dn0(k,i+1,j))
            enddo
         enddo
      enddo
   elseif(vpnt == 'v') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*0.5_f4*(dn0(k,i,j)+dn0(k,i,j+jdim))
            enddo
         enddo
      enddo
   elseif(vpnt == 'w') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*0.5_f4*(dn0(k,i,j)+dn0(k+1,i,j))
            enddo
         enddo
      enddo
   endif
elseif(idir == 2) then
   if(vpnt == 't') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/dn0(k,i,j)
            enddo
         enddo
      enddo
   elseif(vpnt == 'u') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/(0.5_f4*(dn0(k,i,j)+dn0(k,i+1,j)))
            enddo
         enddo
      enddo
   elseif(vpnt == 'v')then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/(0.5_f4*(dn0(k,i,j)+dn0(k,i,j+jdim)))
            enddo
         enddo
      enddo
   elseif(vpnt == 'w')then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/(0.5_f4*(dn0(k,i,j)+dn0(k+1,i,j)))
            enddo
         enddo
      enddo
   endif
endif

end subroutine dnswt


!```````````````````````````````````````````````````````````````````
subroutine dnswt_off(nzmx,nxmx,nymx,n1,n2,n3,i0,j0  &
     ,var1,dn0,vpnt,jdim,idir)

!~ 20.11.2008 (last modified; TIM)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: nzmx, nxmx, nymx, n1, n2, n3, i0, j0, jdim, idir
real(f4), dimension(n1,n2,n3), intent(IN) :: dn0
character(len=*), intent(IN) :: vpnt

real(f4), dimension(nzmx,nxmx,nymx), intent(INOUT) :: var1

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

if(idir == 1) then
   if(vpnt == 't') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*dn0(k,i+i0,j+j0)
            enddo
         enddo
      enddo
   elseif(vpnt == 'u') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*0.5_f4*(dn0(k,i+i0,j+j0)+dn0(k,i+i0+1,j+j0))
            enddo
         enddo
      enddo
   elseif(vpnt == 'v') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*0.5_f4*(dn0(k,i+i0,j+j0)+dn0(k,i+i0,j+jdim+j0))
            enddo
         enddo
      enddo
   elseif(vpnt == 'w') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)*0.5_f4*(dn0(k,i+i0,j+j0)+dn0(k+1,i+i0,j+j0))
            enddo
         enddo
      enddo
   endif
elseif(idir == 2) then
   if(vpnt == 't') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/dn0(k,i+i0,j+j0)
            enddo
         enddo
      enddo
   elseif(vpnt == 'u') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/(0.5_f4*(dn0(k,i+i0,j+j0)+dn0(k,i+i0+1,j+j0)))
            enddo
         enddo
      enddo
   elseif(vpnt == 'v') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/(0.5_f4*(dn0(k,i+i0,j+j0)+dn0(k,i+i0,j+jdim+j0)))
            enddo
         enddo
      enddo
   elseif(vpnt == 'w') then
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j)/(0.5_f4*(dn0(k,i+i0,j+j0)+dn0(k+1,i+i0,j+j0)))
            enddo
         enddo
      enddo
   endif
endif

end subroutine dnswt_off


!```````````````````````````````````````````````````````````````````
subroutine dnswt2(nzmx,nxmx,nymx,n1,n2,n3,var1,dn0x,vnam,idir)

!~ 20.11.2008 (last modified; TIM)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: nzmx, nxmx, nymx, n1, n2, n3, idir
real(f4), dimension(n1,n2,n3), intent(IN) :: dn0x
character(len=*), intent(IN) :: vnam

real(f4), dimension(nzmx,nxmx,nymx), intent(INOUT) :: var1

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

if(idir == 1) then

   if (vnam == 'w') then
      do j=1,n3
         do i=1,n2
            do k=1,n1-1
               var1(k,i,j) = var1(k,i,j)*0.5_f4 * (dn0x(k,i,j) + dn0x(k+1,i,j))
            enddo
            k = n1
            var1(k,i,j) = var1(k,i,j) * dn0x(k,i,j)
         enddo
      enddo
   else
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j) * dn0x(k,i,j)
            enddo
         enddo
      enddo
   endif

elseif (idir == 2) then

   if (vnam == 'w') then
      do j=1,n3
         do i=1,n2
            do k=1,n1-1
               var1(k,i,j) = var1(k,i,j)/ (0.5_f4 * (dn0x(k,i,j) + dn0x(k+1,i,j)))
            enddo
            k = n1
            var1(k,i,j) = var1(k,i,j) / dn0x(k,i,j)
         enddo
      enddo
   else
      do j=1,n3
         do i=1,n2
            do k=1,n1
               var1(k,i,j) = var1(k,i,j) / dn0x(k,i,j)
            enddo
         enddo
      enddo
   endif

endif

end subroutine dnswt2


end module nest_lib_mod
