!```````````````````````````````````````````````````````````````````
module parallel_nesting_mod

!+

use main_memory_mod, only: i_c_to_f_grid_comm

implicit none

integer, parameter ::  &
     dnscale_nBC_intra_msgtag = i_c_to_f_grid_comm*10+1,  &
     dnscale_nBC_inter_msgtag = i_c_to_f_grid_comm*10+2
logical :: dnscale_nBC_is_uninitialized = .true.

logical :: upscale_nBC_is_uninitialized = .true.


contains


!```````````````````````````````````````````````````````````````````
subroutine leafPE_dnscale_nBC_init

!+ Initializes some parameters relevant to the "leafPE_dnscale_nBC" 
!   procedure

use univ_kind_defs_mod, only: f4_B
use main_memory_mod, only: n_grids, n_PE_leaf, gd, i_c_to_f_grid_comm

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: n, i1, i2, j1, j2, k1, k2, n_subdomain_W, ng_c, ng_f, igd

!=== Executable statements:

! Determine byte-length for packed data:

do igd=1,n_grids
   gd(igd)%dnscale_nBC_sd_packed_B(:,:) = 0
end do
do n=1,n_PE_leaf
   do ng_c=1,n_grids
      do ng_f=1,n_grids
         if (gd(ng_f)%tx_paths_leaf(1,i_c_to_f_grid_comm,n) /= 0) then
            i1 = gd(ng_f)%tx_paths_leaf(1,i_c_to_f_grid_comm,n)
            i2 = gd(ng_f)%tx_paths_leaf(2,i_c_to_f_grid_comm,n)
            j1 = gd(ng_f)%tx_paths_leaf(3,i_c_to_f_grid_comm,n)
            j2 = gd(ng_f)%tx_paths_leaf(4,i_c_to_f_grid_comm,n)

            k1 = 1
            k2 = gd(1)%mmzp   ! gd(ng_c)%mmzp

            n_subdomain_W = (i2-i1+1)*(j2-j1+1)*(k2-k1+1)
            
            gd(ng_f)%dnscale_nBC_sd_packed_B(ng_c,n) = n_subdomain_W*f4_B
         end if
      end do
   end do 
end do

dnscale_nBC_is_uninitialized = .false.

end subroutine leafPE_dnscale_nBC_init


!```````````````````````````````````````````````````````````````````
subroutine leafPE_upscale_nBC_init

!+ Initializes some parameters relevant to the "leafPE_upscale_nBC" 
!   procedure

use univ_kind_defs_mod, only: f4_B
use main_memory_mod, only: n_grids, n_PE_leaf, gd, i_f_to_c_grid_comm

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: n, i1, i2, j1, j2, k1, k2, n_subdomain_W, ng_c, ng_f, igd

!=== Executable statements:

! Determine byte-length for packed data:

do igd=1,n_grids
   gd(igd)%upscale_nBC_sd_packed_B(:,:) = 0
end do
do n=1,n_PE_leaf
   do ng_c=1,n_grids
      do ng_f=1,n_grids
         if (gd(ng_f)%tx_paths_leaf(1,i_f_to_c_grid_comm,n) /= 0) then
            i1 = gd(ng_f)%tx_paths_leaf(1,i_f_to_c_grid_comm,n)
            i2 = gd(ng_f)%tx_paths_leaf(2,i_f_to_c_grid_comm,n)
            j1 = gd(ng_f)%tx_paths_leaf(3,i_f_to_c_grid_comm,n)
            j2 = gd(ng_f)%tx_paths_leaf(4,i_f_to_c_grid_comm,n)

            k1 = 2  ! kpm(2,ifm)
            k2 = gd(1)%mmzp-1  ! gd(ng_f)%kpm(gd(ng_f)%mmzp-1)

            n_subdomain_W = (i2-i1+1)*(j2-j1+1)*(k2-k1+1)

            gd(ng_f)%upscale_nBC_sd_packed_B(ng_c,n) = n_subdomain_W*f4_B
         end if
      end do
   end do 
end do

upscale_nBC_is_uninitialized = .false.

end subroutine leafPE_upscale_nBC_init


!```````````````````````````````````````````````````````````````````
subroutine fill_dnscale_nBC_buffer(v_type, ac, acs, den, m1, m2, m3, i0, j0,  &
     i1, i2, j1, j2, k1, k2)

!+ Fills nest buffer with variable data

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: v_type, m1, m2, m3, i0, j0, i1, i2, j1, j2, k1, k2
real(f4), dimension(m1,m2,m3), intent(IN) :: ac, den

real(f4), dimension(0:(k2-k1),0:(i2-i1),0:(j2-j1)), intent(OUT) :: acs

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

! v_type:  1 = u, 2 = v, 3 = w, 4 = p, 5 = normal scalar, 6 = aerosol scalar

select case(v_type)

case(1)
   do j=j1,j2
      do i=i1,i2
         do k=k1,k2
            acs(k-k1,i-i1,j-j1) = ac(k,i-i0,j-j0)*  &
                 ((den(k,i-i0,j-j0)+den(k,i+1-i0,j-j0))*0.5_f4)
         end do
      end do
   end do
case(2)
   do j=j1,j2
      do i=i1,i2
         do k=k1,k2
            acs(k-k1,i-i1,j-j1) = ac(k,i-i0,j-j0)*  &
                 ((den(k,i-i0,j-j0)+den(k,i-i0,j+1-j0))*0.5_f4)
         end do
      end do
   end do
case(3)
   do j=j1,j2
      do i=i1,i2
         do k=k1,k2
            acs(k-k1,i-i1,j-j1) = ac(k,i-i0,j-j0)*  &
                 ((den(k,i-i0,j-j0)+den(min(k+1, m1),i-i0,j-j0))*0.5_f4)
         end do
      end do
   end do
case(4)
   do j=j1,j2
      do i=i1,i2
         do k=k1,k2
            acs(k-k1,i-i1,j-j1) = ac(k,i-i0,j-j0)
         end do
      end do
   end do
case(5)
   do j=j1,j2
      do i=i1,i2
         do k=k1,k2
            acs(k-k1,i-i1,j-j1) = ac(k,i-i0,j-j0)*den(k,i-i0,j-j0)
         end do
      end do
   end do
case(6)
   do j=j1,j2
      do i=i1,i2
         do k=k1,k2
            acs(k-k1,i-i1,j-j1) = ac(k,i-i0,j-j0)*den(k,i-i0,j-j0)
         end do
      end do
   end do

end select

end subroutine fill_dnscale_nBC_buffer


!```````````````````````````````````````````````````````````````````
subroutine empty_dnscale_nBC_buffer(ac, buff, max1, max2, max3, m1, m2, m3,  &
     i1, i2, j1, j2, k1, k2)

!+ Fills array with nest buffer data

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: max1, max2, max3, m1, m2, m3, i1, i2, j1, j2, k1, k2
real(f4), dimension(0:m1-1,0:m2-1,0:m3-1), intent(IN) :: buff

real(f4), dimension(max1,max2,max3), intent(OUT) :: ac

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

do j=j1,j2
   do i=i1,i2
      do k=k1,k2
         ac(k,i,j) = buff(k-k1,i-i1,j-j1)
      end do
   end do
end do

end subroutine empty_dnscale_nBC_buffer


!```````````````````````````````````````````````````````````````````
subroutine zero_upscale_nBC_tgt(ac, m1, m2, m3, i0, j0, i1z, i2z, j1z, j2z,  &
     k1z, k2z)

!+ Zeroes out nesting feedback target array subsection

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, i0, j0, i1z, i2z, j1z, j2z, k1z, k2z

real(f4), dimension(m1,m2,m3), intent(OUT) :: ac

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

do j=j1z,j2z
   do i=i1z,i2z
      do k=k1z,k2z
         ac(k,i-i0,j-j0) = ZERO_f4
      end do
   end do
end do

end subroutine zero_upscale_nBC_tgt


!```````````````````````````````````````````````````````````````````
subroutine copy_upscale_nBC_data(v_type, ac, acf, den, denu, denv, m1, m2,  &
     m3, i0, j0, i1s, i2s, j1s, j2s, k1s, k2s, i2u, j2v, k2w, nf1, nf2, nf3)

!+ Fill array with data from nesting feedback buffer

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: v_type, m1, m2, m3, i0, j0, i1s, i2s, j1s, j2s, k1s,  &
     k2s, i2u, j2v, k2w, nf1, nf2, nf3
real(f4), dimension(m1,m2,m3), intent(IN) :: denu, denv, den
real(f4), dimension(nf1,nf2,nf3), intent(IN) :: acf

real(f4), dimension(m1,m2,m3), intent(INOUT) :: ac

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

! v_type:  1 = u, 2 = v, 3 = w, 4 = p, 5 = normal scalar, 6 = aerosol scalar

select case(v_type)

case(1)
   do j=j1s,j2s
      do i=i1s,i2u
         do k=k1s,k2s
            ac(k,i-i0,j-j0) = ac(k,i-i0,j-j0)+acf(k-k1s+1,i-i1s+1,j-j1s+1)/  &
                 denu(k,i-i0,j-j0)
         end do
      end do
   end do
case(2)
   do j=j1s,j2v
      do i=i1s,i2s
         do k=k1s,k2s
            ac(k,i-i0,j-j0) = ac(k,i-i0,j-j0)+acf(k-k1s+1,i-i1s+1,j-j1s+1)/  &
                 denv(k,i-i0,j-j0)
         end do
      end do
   end do
case(3)
   do j=j1s,j2s
      do i=i1s,i2s
         do k=k1s,k2w
            ac(k,i-i0,j-j0)=ac(k,i-i0,j-j0)+acf(k-k1s+1,i-i1s+1,j-j1s+1)/  &
                 (den(k,i-i0,j-j0)+den(k+1,i-i0,j-j0))
         end do
      end do
   end do
case(4)
   do j=j1s,j2s
      do i=i1s,i2s
         do k=k1s,k2s
            ac(k,i-i0,j-j0)=ac(k,i-i0,j-j0)+acf(k-k1s+1,i-i1s+1,j-j1s+1)
         end do
      end do
   end do
case(5)
   do j=j1s,j2s
      do i=i1s,i2s
         do k=k1s,k2s
            ac(k,i-i0,j-j0) = ac(k,i-i0,j-j0)+acf(k-k1s+1,i-i1s+1,j-j1s+1)/  &
                 den(k,i-i0,j-j0)
         end do
      end do
   end do
case(6)
   do j=j1s,j2s
      do i=i1s,i2s
         do k=k1s,k2s
            ac(k,i-i0,j-j0) = ac(k,i-i0,j-j0)+acf(k-k1s+1,i-i1s+1,j-j1s+1)/  &
                 den(k,i-i0,j-j0)
         end do
      end do
   end do
end select

end subroutine copy_upscale_nBC_data


!```````````````````````````````````````````````````````````````````
subroutine interpolate_dnscale_nBC(ac, as, dn0f, n1m, n2m, n3m, m1f, m1, m2,  &
     m3, ifm, iv_type, i0, j0, ibndc, bx, by, bz)

!+ Interpolates nested grid BC values from coarser grid values

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: gd, c_gd, ia, iz, ja, jz

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1m, n2m, n3m, m1f, m1, m2, m3, ifm, iv_type, i0, j0,  &
     ibndc
real(f4), dimension(m1,m2,m3), intent(IN) :: dn0f

real(f4), dimension(n1m,n2m,n3m), intent(INOUT) :: ac, as
real(f4), dimension(m1,m3,2), intent(INOUT) :: bx
real(f4), dimension(m1,m2,2), intent(INOUT) :: by
real(f4), dimension(m2,m3,2), intent(INOUT) :: bz

!== Local declarations:

integer :: icm, iqf, jqf, kqf, kc, jc, ic, im, jm, k1, k2

!=== Executable statements:

! Set various indices. DO NOT CHANGE THE ORDER OF THESE STATEMENTS!!

ia = m2; if (iv_type == 1) ia = m2-1
iz = 1

ja = m3; if (iv_type == 2) ja = m3-1
jz = 1

if (ibndc /= 0) then
   if (ibndc /= 1) iz = ia
   if (ibndc /= 2) ia = 1
   if (ibndc /= 4) jz = ja
   if (ibndc /= 8) ja = 1
end if

! i0 and j0 now passed in as fm leaf PE offsets
! ia, iz, ja, and jz refer to locations where the "b" arrays need to
!   be filled; they are not the usual limits of prognostic points.
! The nstbot=0 and nsttop=0 options do not work with the current tx_paths_*
!   array definitions.

icm = gd(ifm)%nxtnest
k1 = max(1, gd(ifm)%kpm(2)-2); k2 = min(gd(icm)%nzp, gd(ifm)%kpm(m1f-1)+2)

if (iv_type == 1) then
   do jc=gd(ifm)%jpm(ja+j0)-2,gd(ifm)%jpm(jz+j0)+1
      do iqf=ia,iz
         im = iqf+i0; ic = gd(ifm)%ipm(im)
         do kc=k1,k2
            as(kc,iqf,jc) = gd(ifm)%ei4(im)*ac(kc,ic-2,jc)+  &
                 gd(ifm)%ei5(im)*ac(kc,ic-1,jc)+  &
                 gd(ifm)%ei6(im)*ac(kc,ic  ,jc)+  &
                 gd(ifm)%ei7(im)*ac(kc,ic+1,jc)
         end do
      end do
   end do
else
   do jc=gd(ifm)%jpm(ja+j0)-2,gd(ifm)%jpm(jz+j0)+1
      do iqf=ia,iz
         im = iqf+i0; ic = gd(ifm)%ipm(im)
         do kc=k1,k2
            as(kc,iqf,jc) = gd(ifm)%ei1(im)*ac(kc,ic-1,jc)+  &
                 gd(ifm)%ei2(im)*ac(kc,ic  ,jc)+  &
                 gd(ifm)%ei3(im)*ac(kc,ic+1,jc)
         end do
      end do
   end do
end if

if (iv_type == 2) then
   do jqf=ja,jz
      jm = jqf+j0; jc = gd(ifm)%jpm(jm)
      do iqf=ia,iz
         do kc=k1,k2
            ac(kc,iqf,jqf) = gd(ifm)%ej4(jm)*as(kc,iqf,jc-2)+  &
                 gd(ifm)%ej5(jm)*as(kc,iqf,jc-1)+  &
                 gd(ifm)%ej6(jm)*as(kc,iqf,jc  )+  &
                 gd(ifm)%ej7(jm)*as(kc,iqf,jc+1)
         end do
      end do
   end do
else
   do jqf=ja,jz
      jm = jqf+j0; jc = gd(ifm)%jpm(jm)
      do iqf=ia,iz
         do kc=k1,k2
            ac(kc,iqf,jqf) = gd(ifm)%ej1(jm)*as(kc,iqf,jc-1)+  &
                 gd(ifm)%ej2(jm)*as(kc,iqf,jc  )+  &
                 gd(ifm)%ej3(jm)*as(kc,iqf,jc+1)
         end do
      end do
   end do
end if

if (iv_type == 3) then
   
   if (iand(ibndc, 1) /= 0) then
      iqf = ia
      do jqf=ja,jz
         do kqf=1,m1f-1
            kc = gd(ifm)%kpm(kqf+1)
            bx(kqf,jqf,1) =  &
                 (gd(ifm)%ek4(kqf)*ac(max(1         , kc-2),iqf,jqf)+  &
                 gd(ifm)%ek5(kqf)*ac(               kc-1 ,iqf,jqf)+  &
                 gd(ifm)%ek6(kqf)*ac(               kc   ,iqf,jqf)+  &
                 gd(ifm)%ek7(kqf)*ac(min(gd(icm)%nzp-1, kc+1),iqf,jqf))/  &
                 (0.5_f4*(dn0f(kqf,iqf,jqf)+dn0f(kqf+1,iqf,jqf)))
         end do
      end do
   end if

   if (iand(ibndc, 2) /= 0) then
      iqf = iz
      do jqf=ja,jz
         do kqf=1,m1f-1
            kc = gd(ifm)%kpm(kqf+1)
            bx(kqf,jqf,2) =  &
                 (gd(ifm)%ek4(kqf)*ac(max(1        , kc-2),iqf,jqf)+  &
                 gd(ifm)%ek5(kqf)*ac(               kc-1 ,iqf,jqf)+  &
                 gd(ifm)%ek6(kqf)*ac(               kc   ,iqf,jqf)+  &
                 gd(ifm)%ek7(kqf)*ac(min(gd(icm)%nzp-1, kc+1),iqf,jqf))/  &
                 (0.5_f4*(dn0f(kqf,iqf,jqf)+dn0f(kqf+1,iqf,jqf)))
         end do
      end do
   end if

   if (iand(ibndc, 4) /= 0) then
      jqf = ja
      do iqf=ia,iz
         do kqf=1,m1f-1
            kc = gd(ifm)%kpm(kqf+1)
            by(kqf,iqf,1) =  &
                 (gd(ifm)%ek4(kqf)*ac(max(1         , kc-2),iqf,jqf)+  &
                 gd(ifm)%ek5(kqf)*ac(               kc-1 ,iqf,jqf)+  &
                 gd(ifm)%ek6(kqf)*ac(               kc   ,iqf,jqf)+  &
                 gd(ifm)%ek7(kqf)*ac(min(gd(icm)%nzp-1, kc+1),iqf,jqf))/  &
                 (0.5_f4*(dn0f(kqf,iqf,jqf)+dn0f(kqf+1,iqf,jqf)))
         end do
      end do
   end if

   if (iand(ibndc, 8) /= 0) then
      jqf = jz
      do iqf=ia,iz
         do kqf=1,m1f-1
            kc = gd(ifm)%kpm(kqf+1)
            by(kqf,iqf,2) =  &
                 (gd(ifm)%ek4(kqf)*ac(max(1         , kc-2),iqf,jqf)+  &
                 gd(ifm)%ek5(kqf)*ac(               kc-1 ,iqf,jqf)+  &
                 gd(ifm)%ek6(kqf)*ac(               kc   ,iqf,jqf)+  &
                 gd(ifm)%ek7(kqf)*ac(min(gd(icm)%nzp-1, kc+1),iqf,jqf))/  &
                 (0.5_f4*(dn0f(kqf,iqf,jqf)+dn0f(kqf+1,iqf,jqf)))
         end do
      end do
   end if
   
   if (c_gd%nstbot == 0) then
      kqf = 1
      do jqf=ja,jz
         do iqf=ia,iz
            kc = gd(ifm)%kpm(kqf+1)
            bz(iqf,jqf,1) =  &
                 (gd(ifm)%ek4(kqf)*ac(max(1         , kc-2),iqf,jqf)+  &
                 gd(ifm)%ek5(kqf)*ac(               kc-1 ,iqf,jqf)+  &
                 gd(ifm)%ek6(kqf)*ac(               kc   ,iqf,jqf)+  &
                 gd(ifm)%ek7(kqf)*ac(min(gd(icm)%nzp-1, kc+1),iqf,jqf))/  &
                 (0.5_f4*(dn0f(kqf,iqf,jqf)+dn0f(kqf+1,iqf,jqf)))
         end do
      end do
   end if

   if (c_gd%nsttop == 0) then
      kqf = m1f-1
      do jqf=ja,jz
         do iqf=ia,iz
            kc = gd(ifm)%kpm(kqf+1)
            bz(iqf,jqf,2) =  &
                 (gd(ifm)%ek4(kqf)*ac(max(1         , kc-2),iqf,jqf)+  &
                 gd(ifm)%ek5(kqf)*ac(               kc-1 ,iqf,jqf)+  &
                 gd(ifm)%ek6(kqf)*ac(               kc   ,iqf,jqf)+  &
                 gd(ifm)%ek7(kqf)*ac(min(gd(icm)%nzp-1, kc+1),iqf,jqf))/  &
                 (0.5_f4*(dn0f(kqf,iqf,jqf)+dn0f(kqf+1,iqf,jqf)))
         end do
      end do
   end if

else if (iv_type == 4) then

   if (iand(ibndc, 1) /= 0) then
      iqf = ia
      do jqf=ja,jz
         do kqf=1,m1f
            bx(kqf,jqf,1) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))
         end do
      end do
   end if
   
   if (iand(ibndc, 2) /= 0) then
      iqf = iz
      do jqf=ja,jz
         do kqf=1,m1f
            bx(kqf,jqf,2) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))
         end do
      end do
   end if
   
   if (iand(ibndc, 4) /= 0) then
      jqf = ja
      do iqf=ia,iz
         do kqf=1,m1f
            by(kqf,iqf,1) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))
         end do
      end do
   end if

   if (iand(ibndc, 8) /= 0) then
      jqf = jz
      do iqf=ia,iz
         do kqf=1,m1f
            by(kqf,iqf,2) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))
         end do
      end do
   end if

   if (c_gd%nstbot == 0) then
      kqf = 1
      do jqf=ja,jz
         do iqf=ia,iz
            bz(iqf,jqf,1) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))
         end do
      end do
   end if

   if (c_gd%nsttop == 0) then
      kqf = m1f
      do jqf=ja,jz
         do iqf=ia,iz
            bz(iqf,jqf,2) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))
         end do
      end do
   end if

else

   if (iand(ibndc, 1) /= 0) then
      iqf = ia
      do jqf=ja,jz
         do kqf=1,m1f
            bx(kqf,jqf,1) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))/  &
                 dn0f(kqf,iqf,jqf)
         end do
      end do
   end if

   if (iand(ibndc, 2) /= 0) then
      iqf = iz
      do jqf=ja,jz
         do kqf=1,m1f
            bx(kqf,jqf,2) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))/  &
                 dn0f(kqf,iqf,jqf)
         end do
      end do
   end if

   if (iand(ibndc, 4) /= 0) then
      jqf = ja
      do iqf=ia,iz
         do kqf=1,m1f
            by(kqf,iqf,1) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))/  &
                 dn0f(kqf,iqf,jqf)
         end do
      end do
   end if
   
   if (iand(ibndc, 8) /= 0) then
      jqf = jz
      do iqf=ia,iz
         do kqf=1,m1f
            by(kqf,iqf,2) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))/  &
                 dn0f(kqf,iqf,jqf)
         end do
      end do
   end if

   if (c_gd%nstbot == 0) then
      kqf = 1
      do jqf=ja,jz
         do iqf=ia,iz
            bz(iqf,jqf,1) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))/  &
                 dn0f(kqf,iqf,jqf)
         end do
      end do
   end if

   if (c_gd%nsttop == 0) then
      kqf = m1f
      do jqf=ja,jz
         do iqf=ia,iz
            bz(iqf,jqf,2) =  &
                 (gd(ifm)%ek1(kqf)*ac(gd(ifm)%kpm(kqf)-1,iqf,jqf)+  &
                 gd(ifm)%ek2(kqf)*ac(gd(ifm)%kpm(kqf)  ,iqf,jqf)+  &
                 gd(ifm)%ek3(kqf)*ac(gd(ifm)%kpm(kqf)+1,iqf,jqf))/  &
                 dn0f(kqf,iqf,jqf)
         end do
      end do
   end if

end if

end subroutine interpolate_dnscale_nBC


end module parallel_nesting_mod


!```````````````````````````````````````````````````````````````````
subroutine leafPE_dnscale_nBC(ifm, icm, tx_bp, rx_bp, tx_ireq, rx_ireq,  &
     n_PE_leaf)

!+ Transmits BC from a fine grid's parent (for use in downscale nesting
!   operations) to another leaf PE; also receives the same from other leaf PEs

use univ_kind_defs_mod, only: f4_B, i4_B
use univ_math_constants_mod, only: ZERO_f4
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_receive_nonb_packed, mpiq_wait,  &
     mpiq_transmit_nonb_packed, mpiq_pack_buffer, mpiq_unpack_buffer,  &
     mpiq_check_buffer_packed, MPI_buffer_packed_obj, mpiq_waitall    
#endif  /* end:USE_MPI */
use parallel_nesting_mod, only: dnscale_nBC_is_uninitialized,  &
     fill_dnscale_nBC_buffer,  &
     empty_dnscale_nBC_buffer, interpolate_dnscale_nBC,  &
     dnscale_nBC_intra_msgtag, dnscale_nBC_inter_msgtag
use main_memory_mod, only: n_grids, maxnzp, maxnyp, maxnxp, iscr1, iscr2, gd,  &
     a, b, ivar_ID, ivar_type, PE_rank, dnscale_nBC_packed_pos, my_rank,  &
     dnscale_nBC_sd_max_pB, i_c_to_f_grid_comm
use vtables_mod, only: nsclr, isclrp, naersclr, iaersclp, ind3dn, idn0n,  &
     ibux, ibuy, ibuz, ibvx, ibvy, ibvz, ibwx, ibwy, ibwz, ibpx, ibpy,  &
     ibpz, ibsx, ibsy, ibsz, ibax, ibay, ibaz,ithpn

implicit none

!=== Argument declarations:

integer, intent(IN) :: ifm, icm, n_PE_leaf
integer, dimension(n_PE_leaf), intent(INOUT) :: tx_ireq, rx_ireq

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), dimension(n_PE_leaf), intent(INOUT) :: tx_bp, rx_bp
#else  /* Faux interface */
integer, dimension(n_PE_leaf), intent(INOUT) :: tx_bp, rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer, parameter :: n_spec_vars = 4

integer :: n, nf, nc, ng_c, ng_f, i1, i2, j1, j2, k1, k2, n_subdomain_W,  &
     n_packed_B, nW_per_var, source_rank, nv, n_relevant_vars,  &
     dnscale_nBC_misc_packed_B, i

integer, dimension(n_PE_leaf) :: i1l, i2l, j1l, j2l, k1l, k2l, idxl, nW_per_varl
integer, dimension(2*n_PE_leaf) :: ireq_list

!=== Executable statements:

#if defined(USE_MPI)

!--- Construct table of variables being transmitted:

! Add IDs and arbitrary types of UC, VC, WC, and PC:
n_relevant_vars = n_spec_vars
ivar_ID(1:n_relevant_vars) = (/ 2, 4, 6, 8 /)
ivar_type(1:n_relevant_vars) = (/ 1, 2, 3, 4 /)

! Add any matching normal scalar variables:
do nf=1,nsclr(ifm)
   do nc=1,nsclr(icm)
      if (isclrp(nf,ifm) == isclrp(nc,icm)) then
         n_relevant_vars = n_relevant_vars+1
         ivar_ID(n_relevant_vars) = isclrp(nf,ifm)
         ivar_type(n_relevant_vars) = 5
      end if
   end do
end do

! Add any matching aerosol scalar variables:
do nf=1,naersclr(ifm)
   do nc=1,naersclr(icm)
      if (iaersclp(nf,ifm) == iaersclp(nc,icm)) then
         n_relevant_vars = n_relevant_vars+1
         ivar_ID(n_relevant_vars) = iaersclp(nf,ifm)
         ivar_type(n_relevant_vars) = 6
      end if
   end do
end do

!--- Before anything is transmitted, prepost the receives:

! Determine receive buffer size; assume that 'n_relevant_vars'
!  is constant for all leaf PEs (and time):

dnscale_nBC_misc_packed_B = i4_B*(9 + 2*n_relevant_vars)
n_packed_B = dnscale_nBC_sd_max_pB*n_relevant_vars+dnscale_nBC_misc_packed_B

! Prepost receives from leaf PEs:

do n=1,n_PE_leaf
   rx_ireq(n) = 0
   if (gd(ifm)%rx_paths_leaf(i_c_to_f_grid_comm,n) /= 0) then
      call mpiq_check_buffer_packed(rx_bp(n), n_packed_B)
      call mpiq_receive_nonb_packed(rx_bp(n), PE_rank(n), rx_ireq(n), 0,   &
           gd(ifm)%rx_paths_leaf(i_c_to_f_grid_comm,n)*10000+my_rank)  !@@
#if defined(MPI_DEBUG)
      write (*,'(a,i2,a,i2,a,i2,a,i2,a,i11)')  &
           'RX-prepost intra:', my_rank,  &
           ' rx msg from ',  &
           gd(ifm)%rx_paths_leaf(i_c_to_f_grid_comm,n),  &
           '; grid ', ifm, ', pass ', n,  &
           ', req ', rx_ireq(n); flush 6
#endif  /* end:MPI_DEBUG */
   end if
end do

!--- Transmit coarse grid points necessary for fine grid boundary interpolation
!     to fine grid PEs. Note that even though coarse grid points are
!     transmitted, comm_sched_leafPE is referenced by the fine grid, since all
!     nests only have one parent, not vice versa:

do n=1,n_PE_leaf
   tx_ireq(n) = 0
   if (gd(ifm)%tx_paths_leaf(1,i_c_to_f_grid_comm,n) /= 0) then

      ! Prepare transmit buffer; assume that 'n_relevant_vars' is constant for
      !  all leaf PEs (and time):
      n_packed_B = gd(ifm)%dnscale_nBC_sd_packed_B(icm,n)*n_relevant_vars+  &
           dnscale_nBC_misc_packed_B
      call mpiq_check_buffer_packed(tx_bp(n), n_packed_B)

      i1 = gd(ifm)%tx_paths_leaf(1,i_c_to_f_grid_comm,n)
      i2 = gd(ifm)%tx_paths_leaf(2,i_c_to_f_grid_comm,n)
      j1 = gd(ifm)%tx_paths_leaf(3,i_c_to_f_grid_comm,n)
      j2 = gd(ifm)%tx_paths_leaf(4,i_c_to_f_grid_comm,n)
      k1 = 1; k2 = gd(icm)%nzp

      call mpiq_pack_buffer(i1, tx_bp(n))
      call mpiq_pack_buffer(i2, tx_bp(n))
      call mpiq_pack_buffer(j1, tx_bp(n))
      call mpiq_pack_buffer(j2, tx_bp(n))
      call mpiq_pack_buffer(k1, tx_bp(n))
      call mpiq_pack_buffer(k2, tx_bp(n))
      call mpiq_pack_buffer(my_rank, tx_bp(n))
      call mpiq_pack_buffer(n_relevant_vars, tx_bp(n))

      nW_per_var = gd(ifm)%dnscale_nBC_sd_packed_B(icm,n)/f4_B
      call mpiq_pack_buffer(nW_per_var, tx_bp(n))

      call mpiq_pack_buffer(ivar_ID, n_relevant_vars, tx_bp(n))
      call mpiq_pack_buffer(ivar_type, n_relevant_vars, tx_bp(n))

      ! Sequentially add each relevant variable to tx buffer:
      do nv=1,n_relevant_vars
         ! Weight variable (in some type-dependent way) by coarse grid density:
         call fill_dnscale_nBC_buffer(ivar_type(nv),  &
              a(ind3dn(icm,ivar_ID(nv))), a(iscr1), a(idn0n(icm)),  &
              gd(icm)%mmzp, gd(icm)%mmxp, gd(icm)%mmyp, gd(icm)%mi0,  &
              gd(icm)%mj0, i1, i2, j1, j2, k1, k2)

         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n))
      end do

      call mpiq_transmit_nonb_packed(tx_bp(n),  &  ! Transmit message to PE
           gd(ifm)%tx_paths_leaf(5,i_c_to_f_grid_comm,n), tx_ireq(n),  &
           my_rank*10000+gd(ifm)%tx_paths_leaf(5,i_c_to_f_grid_comm,n)) !@@
#if defined(MPI_DEBUG)
      write (*,'(a,i2,a,i2,a,i2,a,i2,a,i11)') 'TX-post intra:', my_rank,  &
           ' tx msg to ',  &
           gd(ifm)%tx_paths_leaf(5,i_c_to_f_grid_comm,n),  &
           '; grid ', ifm, ', pass ', n,  &
           ', req ', tx_ireq(n); flush 6
#endif  /* end:MPI_DEBUG */
   end if

end do

!--- Ensure that all of our transmissions and receives have completed:

i = 0
do n=1,n_PE_leaf      ! Transmissions
   if (gd(ifm)%tx_paths_leaf(1,i_c_to_f_grid_comm,n) /= 0)  &
        call mpiq_wait(tx_ireq(n))
!##        i = i+1; ireq_list(i) = tx_ireq(n)
!##        call mpiq_wait(tx_ireq(n))
end do

do n=1,n_PE_leaf      ! Receives
   if (gd(ifm)%rx_paths_leaf(i_c_to_f_grid_comm,n) /= 0)  &
        call mpiq_wait(rx_ireq(n))
!##        i = i+1; ireq_list(i) = rx_ireq(n)
!##        call mpiq_wait(rx_ireq(n))
end do

!##call mpiq_waitall(ireq_list, i)

!--- Unpack the relevant coarse grid data (from other leaf PEs),
!     interpolate the fine grid boundaries, and put them in the "B" array:

   ! Unpack/store received message bookkeeping information from all nodes:
do n=1,n_PE_leaf
   if (gd(ifm)%rx_paths_leaf(i_c_to_f_grid_comm,n) /= 0) then
#if defined(MPI_DEBUG)
      write (*,'(a,i2,a,i2,a,i2,a,i2)') 'RX:', my_rank, ' rx msg from ',  &
           gd(ifm)%rx_paths_leaf(i_c_to_f_grid_comm,n), '; grid ', ifm,  &
           ', pass ', n; flush 6
#endif  /* end:MPI_DEBUG */
      call mpiq_unpack_buffer(i1l(n), rx_bp(n))
      call mpiq_unpack_buffer(i2l(n), rx_bp(n))
      call mpiq_unpack_buffer(j1l(n), rx_bp(n))
      call mpiq_unpack_buffer(j2l(n), rx_bp(n))
      call mpiq_unpack_buffer(k1l(n), rx_bp(n))
      call mpiq_unpack_buffer(k2l(n), rx_bp(n))
      call mpiq_unpack_buffer(source_rank, rx_bp(n))
      call mpiq_unpack_buffer(n_relevant_vars, rx_bp(n))
      call mpiq_unpack_buffer(nW_per_varl(n), rx_bp(n))

      call mpiq_unpack_buffer(ivar_ID, n_relevant_vars, rx_bp(n))
      call mpiq_unpack_buffer(ivar_type, n_relevant_vars, rx_bp(n))
   end if
end do

   ! Loop through each variable:
do nv=1,n_relevant_vars

   ! Unpack and (re)construct coarse grid variable (put in 'scr1'):
   a(iscr1:iscr1+maxnzp*maxnxp*maxnyp-1) = ZERO_f4
   do n=1,n_PE_leaf
      if (gd(ifm)%rx_paths_leaf(i_c_to_f_grid_comm,n) /= 0) then
         call mpiq_unpack_buffer(a(iscr2:iscr2+nW_per_varl(n)-1),  &
              nW_per_varl(n), rx_bp(n))

         call empty_dnscale_nBC_buffer(a(iscr1), a(iscr2), maxnzp, maxnxp,  &
              maxnyp, k2l(n)-k1l(n)+1, i2l(n)-i1l(n)+1, j2l(n)-j1l(n)+1,  &
              i1l(n), i2l(n), j1l(n), j2l(n), k1l(n), k2l(n))
      end if
   end do

   ! Interpolate to the fine grid boundaries and put result into the
   !  "B" array:

   select case(ivar_type(nv))

   case(1)
      call interpolate_dnscale_nBC(a(iscr1), a(iscr2), a(idn0n(ifm)),  &
           maxnzp, maxnxp, maxnyp, gd(ifm)%nzp, gd(ifm)%mmzp,  &
           gd(ifm)%mmxp, gd(ifm)%mmyp, ifm, ivar_type(nv), gd(ifm)%mi0,  &
           gd(ifm)%mj0, gd(ifm)%mibndc, b(ibux(ifm)), b(ibuy(ifm)),  &
           b(ibuz(ifm)))
   case(2)
      call interpolate_dnscale_nBC(a(iscr1), a(iscr2), a(idn0n(ifm)),  &
           maxnzp, maxnxp, maxnyp, gd(ifm)%nzp, gd(ifm)%mmzp,  &
           gd(ifm)%mmxp, gd(ifm)%mmyp, ifm, ivar_type(nv), gd(ifm)%mi0,  &
           gd(ifm)%mj0, gd(ifm)%mibndc, b(ibvx(ifm)), b(ibvy(ifm)),  &
           b(ibvz(ifm)))
   case(3)
      call interpolate_dnscale_nBC(a(iscr1), a(iscr2), a(idn0n(ifm)),  &
           maxnzp, maxnxp, maxnyp, gd(ifm)%nzp, gd(ifm)%mmzp,  &
           gd(ifm)%mmxp, gd(ifm)%mmyp, ifm, ivar_type(nv), gd(ifm)%mi0,  &
           gd(ifm)%mj0, gd(ifm)%mibndc, b(ibwx(ifm)), b(ibwy(ifm)),  &
           b(ibwz(ifm)))
   case(4)
      call interpolate_dnscale_nBC(a(iscr1), a(iscr2), a(idn0n(ifm)),  &
           maxnzp, maxnxp, maxnyp, gd(ifm)%nzp, gd(ifm)%mmzp,  &
           gd(ifm)%mmxp, gd(ifm)%mmyp, ifm, ivar_type(nv), gd(ifm)%mi0,  &
           gd(ifm)%mj0, gd(ifm)%mibndc, b(ibpx(ifm)), b(ibpy(ifm)),  &
           b(ibpz(ifm)))
   case(5)
      call interpolate_dnscale_nBC(a(iscr1), a(iscr2), a(idn0n(ifm)),  &
           maxnzp, maxnxp, maxnyp, gd(ifm)%nzp, gd(ifm)%mmzp,  &
           gd(ifm)%mmxp, gd(ifm)%mmyp, ifm, ivar_type(nv), gd(ifm)%mi0,  &
           gd(ifm)%mj0, gd(ifm)%mibndc, b(ibsx(nv-n_spec_vars,ifm)),  &
           b(ibsy(nv-n_spec_vars,ifm)), b(ibsz(nv-n_spec_vars,ifm)))
   case(6)
      call interpolate_dnscale_nBC(a(iscr1), a(iscr2), a(idn0n(ifm)),  &
           maxnzp, maxnxp, maxnyp, gd(ifm)%nzp, gd(ifm)%mmzp,  &
           gd(ifm)%mmxp, gd(ifm)%mmyp, ifm, ivar_type(nv), gd(ifm)%mi0,  &
           gd(ifm)%mj0, gd(ifm)%mibndc,  &
           b(ibax(nv-n_spec_vars-nsclr(ifm),ifm)),  &
           b(ibay(nv-n_spec_vars-nsclr(ifm),ifm)),  &
           b(ibaz(nv-n_spec_vars-nsclr(ifm),ifm)))
   end select
end do

#endif  /* end:USE_MPI */

end subroutine leafPE_dnscale_nBC


!```````````````````````````````````````````````````````````````````
subroutine leafPE_upscale_nBC(icm, ifm, tx_bp, rx_bp, tx_ireq, rx_ireq,  &
     n_PE_leaf)

!+ Support for upscale nest operations; Transmits new values for coarser grid
!   points (created from finer grid point values) from one leaf PE to another;
!   also receives the same from other leaf PEs

use univ_kind_defs_mod, only: f4_B, i4_B
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_receive_nonb_packed, mpiq_wait,  &
     mpiq_transmit_nonb_packed, mpiq_pack_buffer, mpiq_unpack_buffer,  &
     mpiq_check_buffer_packed, MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use main_memory_mod, only: n_grids, a, iscr1, iscr2, gd, i0, j0, my_rank,  &
     PE_rank, upscale_nBC_packed_pos, upscale_nBC_sd_max_pB, ivar_ID,  &
     ivar_type, i_f_to_c_grid_comm, i_f_to_c_grid_info
use vtables_mod, only: nsclr, isclrp, naersclr, iaersclp, ind3dn, idn0n,  &
     idn0un, idn0vn, ipi0n, ithpn, irtgtn
use nest_mod, only: fdbackp
use parallel_nesting_mod, only: zero_upscale_nBC_tgt, copy_upscale_nBC_data,  &
     upscale_nbc_is_uninitialized

implicit none

!=== Argument declarations:

integer, intent(IN) :: icm, ifm, n_PE_leaf
integer, dimension(n_PE_leaf), intent(INOUT) :: tx_ireq, rx_ireq

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), dimension(n_PE_leaf), intent(INOUT) :: tx_bp, rx_bp
#else  /* Faux interface */
integer, dimension(n_PE_leaf), intent(INOUT) :: tx_bp, rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer, parameter :: n_spec_vars = 4

integer :: n_relevant_vars, upscale_nBC_misc_packed_B, n_subdomain_W, n,  &
     ng_c, ng_f, i1, i2, j1, j2, k1, k2, nf, nc, n_packed_B, nW_per_var, nv,  &
     i1B, i2B, j1B, j2B, i1z, i2z, i2zu, j1z, j2z, j2zv, k1z, k2z, k2zw,  &
     source_rank
logical :: is_not_initialized

!=== Executable statements:

#if defined(USE_MPI)

!--- Construct table of variables being transmitted:

! Add IDs and arbitrary types of UC, VC, WC, and PC:
n_relevant_vars = n_spec_vars
ivar_ID(1:n_relevant_vars) = (/ 2, 4, 6, 8 /)
ivar_type(1:n_relevant_vars) = (/ 1, 2, 3, 4 /)

! Add any matching normal scalar variables:
do nf=1,nsclr(ifm)
   do nc=1,nsclr(icm)
      if (isclrp(nf,ifm) == isclrp(nc,icm)) then
         n_relevant_vars = n_relevant_vars+1
         ivar_ID(n_relevant_vars) = isclrp(nf,ifm)
         ivar_type(n_relevant_vars) = 5
      end if
   end do
end do

! Add any matching aerosol scalar variables:
do nf=1,naersclr(ifm)
   do nc=1,naersclr(icm)
      if (iaersclp(nf,ifm) == iaersclp(nc,icm)) then
         n_relevant_vars = n_relevant_vars+1
         ivar_ID(n_relevant_vars) = iaersclp(nf,ifm)
         ivar_type(n_relevant_vars) = 6
      end if
   end do
end do

!--- Before anything is transmitted, prepost the receives:

! Determine receive buffer size; assume that 'n_relevant_vars'
!  is constant for all leaf PEs (and time):

upscale_nBC_misc_packed_B = i4_B*(9 + 2*n_relevant_vars)
n_packed_B = upscale_nBC_sd_max_pB*n_relevant_vars+upscale_nBC_misc_packed_B

! Prepost receives from leaf PEs:

do n=1,n_PE_leaf
   rx_ireq(n) = 0
   if (gd(ifm)%rx_paths_leaf(i_f_to_c_grid_comm,n) /= 0) then
      call mpiq_check_buffer_packed(rx_bp(n), n_packed_B)
      call mpiq_receive_nonb_packed(rx_bp(n), PE_rank(n), rx_ireq(n), 0)
   end if
end do

!--- Transmit "replacement" coarse grid points (created by averaging fine
!     grid points) to other leaf PEs: 

do n=1,n_PE_leaf

   tx_ireq(n) = 0
   if (gd(ifm)%tx_paths_leaf(1,i_f_to_c_grid_comm,n) /= 0) then
    
      ! Prepare transmit buffer; assume that 'n_relevant_vars' is constant for
      !  all leaf PEs (and time):
      n_packed_B = gd(ifm)%upscale_nBC_sd_packed_B(icm,n)*n_relevant_vars+  &
           upscale_nBC_misc_packed_B
      call mpiq_check_buffer_packed(tx_bp(n), n_packed_B)

      i1 = gd(ifm)%tx_paths_leaf(1,i_f_to_c_grid_comm,n)
      i2 = gd(ifm)%tx_paths_leaf(2,i_f_to_c_grid_comm,n)
      j1 = gd(ifm)%tx_paths_leaf(3,i_f_to_c_grid_comm,n)
      j2 = gd(ifm)%tx_paths_leaf(4,i_f_to_c_grid_comm,n)
      k1 = gd(ifm)%kpm(2); k2 = gd(ifm)%kpm(gd(ifm)%nzp-1)

      call mpiq_pack_buffer(i1, tx_bp(n))
      call mpiq_pack_buffer(i2, tx_bp(n))
      call mpiq_pack_buffer(j1, tx_bp(n))
      call mpiq_pack_buffer(j2, tx_bp(n))
      call mpiq_pack_buffer(k1, tx_bp(n))
      call mpiq_pack_buffer(k2, tx_bp(n))
      call mpiq_pack_buffer(my_rank, tx_bp(n))
      call mpiq_pack_buffer(n_relevant_vars, tx_bp(n))

      nW_per_var = gd(ifm)%upscale_nBC_sd_packed_B(icm,n)/f4_B
      call mpiq_pack_buffer(nW_per_var, tx_bp(n))

      call mpiq_pack_buffer(ivar_ID, n_relevant_vars, tx_bp(n))
      call mpiq_pack_buffer(ivar_type, n_relevant_vars, tx_bp(n))

      i1B = gd(ifm)%tx_paths_leaf(1,i_f_to_c_grid_info,n)
      i2B = gd(ifm)%tx_paths_leaf(2,i_f_to_c_grid_info,n)
      j1B = gd(ifm)%tx_paths_leaf(3,i_f_to_c_grid_info,n)
      j2B = gd(ifm)%tx_paths_leaf(4,i_f_to_c_grid_info,n)

      do nv=1,n_relevant_vars
         call fdbackp(ivar_type(nv), a(ind3dn(ifm,ivar_ID(nv))),  &
              a(iscr1), nW_per_var, a(idn0n(ifm)), a(idn0un(ifm)),  &
              a(idn0vn(ifm)), gd(ifm)%mmzp, gd(ifm)%mmxp, gd(ifm)%mmyp, ifm,  &
              icm, i1B-i0, i2B-i0, j1B-j0, j2B-j0, i0, j0, gd(ifm)%mibndc,  &
              gd(ifm)%nstratx, gd(ifm)%nstraty, my_rank)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n))
      end do

      call mpiq_transmit_nonb_packed(tx_bp(n),  &     ! Transmit message to PE
           gd(ifm)%tx_paths_leaf(5,i_f_to_c_grid_comm,n), tx_ireq(n))

   endif
enddo

!--- Ensure that all of our transmissions and receives have completed:

do n=1,n_PE_leaf      ! Transmissions
   if (gd(ifm)%tx_paths_leaf(1,i_f_to_c_grid_comm,n) /= 0)  &
        call mpiq_wait(tx_ireq(n))
end do

do n=1,n_PE_leaf      ! Receives
   if (gd(ifm)%rx_paths_leaf(i_f_to_c_grid_comm,n) /= 0)  &
        call mpiq_wait(rx_ireq(n))
end do

!--- Unpack the relevant coarse grid data (from other leaf PEs); process it:

   ! Unpack/store received message bookkeeping information from all nodes:
is_not_initialized = .TRUE.
do n=1,n_PE_leaf
   if (gd(ifm)%rx_paths_leaf(i_f_to_c_grid_comm,n) /= 0) then
      call mpiq_unpack_buffer(i1, rx_bp(n))
      call mpiq_unpack_buffer(i2, rx_bp(n))
      call mpiq_unpack_buffer(j1, rx_bp(n))
      call mpiq_unpack_buffer(j2, rx_bp(n))
      call mpiq_unpack_buffer(k1, rx_bp(n))
      call mpiq_unpack_buffer(k2, rx_bp(n))
      call mpiq_unpack_buffer(source_rank, rx_bp(n))
      call mpiq_unpack_buffer(n_relevant_vars, rx_bp(n))
      call mpiq_unpack_buffer(nW_per_var, rx_bp(n))

      call mpiq_unpack_buffer(ivar_ID, n_relevant_vars, rx_bp(n))
      call mpiq_unpack_buffer(ivar_type, n_relevant_vars, rx_bp(n))

      if (is_not_initialized) then
         i1z = max(gd(ifm)%ipm(2), 1+gd(icm)%mi0)
         j1z = max(gd(ifm)%jpm(2),1+gd(icm)%mj0)
         k1z = gd(ifm)%kpm(2)

         i2zu = 0; j2zv = 0; k2zw = 0

         do nv=1,n_relevant_vars

            ! Set the portion of this coarse grid subdomain that will be filled
            !  with fine grid info to zero:

            i2z = min(gd(ifm)%ipm(gd(ifm)%nxp-1), gd(icm)%mmxp+gd(icm)%mi0)
            j2z = min(gd(ifm)%jpm(gd(ifm)%nyp-1), gd(icm)%mmyp+gd(icm)%mj0)
            k2z = gd(ifm)%kpm(gd(ifm)%nzp-1)

            select case(ivar_type(nv))
            case(1)
               i2z = min(gd(ifm)%ipm(gd(ifm)%nxp-1-gd(ifm)%nstratx),  &
                    gd(icm)%mmxp+gd(icm)%mi0)
               i2zu = i2z
            case(2)
               j2z = min(gd(ifm)%jpm(gd(ifm)%nyp-1-gd(ifm)%nstraty),  &
                    gd(icm)%mmyp+gd(icm)%mj0)
               j2zv = j2z
            case(3)
               k2z = gd(ifm)%kpm(gd(ifm)%nzp-1-  &
                    gd(ifm)%nrz(gd(ifm)%kpm(gd(ifm)%nzp-1)))
               k2zw = k2z
            end select

            call zero_upscale_nBC_tgt(a(ind3dn(icm,ivar_ID(nv))),  &
                 gd(icm)%mmzp, gd(icm)%mmxp, gd(icm)%mmyp, gd(icm)%mi0,  &
                 gd(icm)%mj0, i1z, i2z, j1z, j2z, k1z, k2z)

         end do
         is_not_initialized = .FALSE.
      end if

      do nv=1,n_relevant_vars
         call mpiq_unpack_buffer(a(iscr2:iscr2+nW_per_var-1), nW_per_var,  &
              rx_bp(n))

         call copy_upscale_nBC_data(ivar_type(nv),  &
              a(ind3dn(icm,ivar_ID(nv))), a(iscr2), a(idn0n(icm)),  &
              a(idn0un(icm)), a(idn0vn(icm)), gd(icm)%mmzp, gd(icm)%mmxp,  &
              gd(icm)%mmyp, gd(icm)%mi0, gd(icm)%mj0, i1, i2, j1, j2,  &
              k1, k2, min(i2, i2zu), min(j2, j2zv), min(k2, k2zw), k2-k1+1,  &
              i2-i1+1, j2-j1+1)
      end do

   end if
end do

do nv=1,n_relevant_vars
   select case(ivar_type(nv))
   case(4)
      if (gd(icm)%nstbot == 1) call botset_atm_mass(gd(icm)%mmzp,  &
           gd(icm)%mmxp, gd(icm)%mmyp, 1, gd(icm)%mmxp, 1,  &
           gd(icm)%mmyp, icm, a(ind3dn(icm,ivar_ID(nv))),  &
           a(ipi0n(icm)), 'P')
      if (gd(icm)%nsttop == 1) call topset_atm_mass(gd(icm)%mmzp,  &
           gd(icm)%mmxp, gd(icm)%mmyp, 1, gd(icm)%mmxp, 1,  &
           gd(icm)%mmyp, icm, a(ind3dn(icm,ivar_ID(nv))),  &
           a(ipi0n(icm)), a(ithpn(icm)), a(irtgtn(icm)), 'P')
   case(1)
      if (gd(icm)%nstbot == 1) call botset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'U')
      if (gd(icm)%nsttop == 1) call topset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'U')
   case(2)
      if (gd(icm)%nstbot == 1) call botset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'V')
      if (gd(icm)%nsttop == 1) call topset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'V')
   case(5)
      if (gd(icm)%nstbot == 1) call botset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'T')
      if (gd(icm)%nsttop == 1) call topset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'T')
   case(6)
      if (gd(icm)%nstbot == 1) call botset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'A')
      if (gd(icm)%nsttop == 1) call topset(gd(icm)%mmzp, gd(icm)%mmxp,  &
           gd(icm)%mmyp, 1, gd(icm)%mmxp, 1, gd(icm)%mmyp, icm,  &
           a(ind3dn(icm,ivar_ID(nv))), 'A')
   end select
end do

#endif  /* end:USE_MPI */

end subroutine leafPE_upscale_nBC
