!```````````````````````````````````````````````````````````````````
module parallel_results_mod

!+

implicit none

integer, parameter :: n_infotags = 4

integer, dimension(n_infotags) :: infotags


contains


!```````````````````````````````````````````````````````````````````
subroutine esdbuff_3D(a, b, n1, n2, n3, m2, m3, i0, j0, ic1, ic2, jc1, jc2)

!+ Copy subdomain data (3D normal) from leaf PE buffer, filling root PE array

use univ_kind_defs_mod, only: f4

implicit none

integer, intent(IN) :: n1, n2, n3, m2, m3, i0, j0, ic1, ic2, jc1, jc2
real(f4), dimension(n1,m2,m3), intent(IN) :: b

real(f4), dimension(n1,n2,n3), intent(INOUT) :: a

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

do j=jc1,jc2
   do i=ic1,ic2
      do k=1,n1
         a(k,i+i0,j+j0) = b(k,i,j)
      end do
   end do
end do

end subroutine esdbuff_3D


!```````````````````````````````````````````````````````````````````
subroutine esdbuff_2D(a, b, n2, n3, m2, m3, i0, j0, ic1, ic2, jc1, jc2)

!+ Copy subdomain data (2D normal) from leaf PE buffer, filling root PE array

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n2, n3, m2, m3, i0, j0, ic1, ic2, jc1, jc2
real(f4), dimension(m2,m3), intent(IN) :: b

real(f4), dimension(n2,n3), intent(INOUT) :: a

!== Local declarations:

integer :: i, j

!=== Executable statements:

do j=jc1,jc2
   do i=ic1,ic2
      a(i+i0,j+j0) = b(i,j)
   end do
end do

end subroutine esdbuff_2D


!```````````````````````````````````````````````````````````````````
subroutine esdbuff_3Ds(a, b, n4, n2, n3, n5, m2, m3, i0, j0, ic1, ic2, jc1, jc2)

!+ Copy subdomain data (3D SBPS) from leaf PE buffer, filling root PE array

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n2, n3, n4, n5, m2, m3, i0, j0, ic1, ic2, jc1, jc2
real(f4), dimension(n4,m2,m3,n5), intent(IN) :: b

real(f4), dimension(n4,n2,n3,n5), intent(INOUT) :: a

!== Local declarations:

integer :: i, j, k, ip

!=== Executable statements:

do ip=1,n5
   do j=jc1,jc2
      do i=ic1,ic2
         do k=1,n4
            a(k,i+i0,j+j0,ip) = b(k,i,j,ip)
         end do
      end do
   end do
end do

end subroutine esdbuff_3Ds


!```````````````````````````````````````````````````````````````````
subroutine esdbuff_2Ds(a, b, n2, n3, n5, m2, m3, i0, j0, ic1, ic2, jc1, jc2)

!+ Copy subdomain data (2D SBPS) from leaf PE buffer, filling root PE array

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n2, n3, n5, m2, m3, i0, j0, ic1, ic2, jc1, jc2
real(f4), dimension(m2,m3,n5), intent(IN) :: b

real(f4), dimension(n2,n3,n5), intent(INOUT) :: a

!== Local declarations:

integer :: i, j, ip

!=== Executable statements:

do ip=1,n5
   do j=jc1,jc2
      do i=ic1,ic2
         a(i+i0,j+j0,ip) = b(i,j,ip)
      end do
   end do
end do

end subroutine esdbuff_2Ds


end module parallel_results_mod


!```````````````````````````````````````````````````````````````````
subroutine leafPE_put_all_results

!+ Transmits all relevant fields to the root PE

use ISO_C_binding, only: C_F_pointer, C_loc
use univ_kind_defs_mod, only: f4
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_transmit_plain
#endif  /* end:USE_MPI */
use main_memory_mod, only: n_grids, nzg, nzs, npatch, a, my_rank, gd,  &
     rootPE_rank
use parallel_results_mod, only: n_infotags, infotags
use vtables_mod, only: nmp33, ind3dn, mp3tab3, nmp23, ind2dn, mp2tab3,  &
     nmps23, ind2dsn, mp2stab3, nmps33, ind3dsn, mp3stab3

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: igd, mxyp, mxyzp, mxys3p, mxys2p, idx, nmp
real(f4), pointer, dimension(:) :: ptr_1D => NULL()

!=== Executable statements:

#if defined(USE_MPI)

infotags(1) = my_rank

do igd=1,n_grids

   infotags(2) = igd

   mxyp = gd(igd)%mmxp*gd(igd)%mmyp; mxyzp = mxyp*gd(igd)%mmzp
   mxys3p = mxyp*(nzg+nzs)*npatch; mxys2p = mxyp*npatch

   infotags(4) = 3
   do nmp=1,nmp33(igd)
      infotags(3) = mp3tab3(nmp,igd)

      call mpiq_transmit_plain(infotags, n_infotags, rootPE_rank) ! To root PE

      ! REAL (4-byte) data:
      idx = ind3dn(igd,mp3tab3(nmp,igd))
      call mpiq_transmit_plain(a(idx:idx+mxyzp-1), mxyzp, rootPE_rank)
   end do

   infotags(4) = 2
   do nmp=1,nmp23(igd)
      infotags(3) = mp2tab3(nmp,igd)
      call mpiq_transmit_plain(infotags, n_infotags, rootPE_rank) ! To root PE

      ! REAL (4-byte) data:
      idx = ind2dn(igd,mp2tab3(nmp,igd))
      call mpiq_transmit_plain(a(idx:idx+mxyp-1), mxyp, rootPE_rank)
   end do

   infotags(4) = 4
   do nmp=1,nmps23(igd)
      infotags(3) = mp2stab3(nmp,igd)
      call mpiq_transmit_plain(infotags, n_infotags, rootPE_rank) ! To root PE

      ! REAL (4-byte) data:
      idx = ind2dsn(igd,mp2stab3(nmp,igd))
      call mpiq_transmit_plain(a(idx:idx+mxys2p-1), mxys2p, rootPE_rank)
   end do

   infotags(4) = 5
   do nmp=1,nmps33(igd)
      infotags(3) = mp3stab3(nmp,igd)
      call mpiq_transmit_plain(infotags, n_infotags, rootPE_rank) ! To root PE

      ! REAL (4-byte) data:
      idx = ind3dsn(igd,mp3stab3(nmp,igd))
      call mpiq_transmit_plain(a(idx:idx+mxys3p-1), mxys3p, rootPE_rank)
   end do

   infotags(4) = 6
   call mpiq_transmit_plain(infotags, n_infotags, rootPE_rank) ! To root PE
      ! REAL (4-byte) data:
   call C_F_pointer(C_loc(gd(igd)%emisst), ptr_1D, (/mxyp/))
   call mpiq_transmit_plain(ptr_1D, mxyp, rootPE_rank)
   
end do

#endif  /* end:USE_MPI */

end subroutine leafPE_put_all_results


!```````````````````````````````````````````````````````````````````
subroutine rootPE_get_all_results

!+ Receives all relevant fields from leaf PEs

#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_receive_plain
#endif  /* end:USE_MPI */
use main_memory_mod, only: n_grids, nzg, nzs, npatch, gd, iscr1, a,  &
     n_PE_leaf, mxp, myp, PE_rank
use vtables_mod, only: nmp33, nmp23, nmps33, nmps23, ind3dn, ind2dn,   &
     ind3dsn, ind2dsn
use parallel_results_mod, only: esdbuff_2d, esdbuff_3d, esdbuff_2ds,  &
     esdbuff_3ds, infotags, n_infotags

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: igd, nvars, nv, n, il1, ir2, jb1, jt2, mxyp, mxyzp, mxys3p,  &
     mxys2p, ivar, idx, ivartype

!=== Executable statements:

#if defined(USE_MPI)

nvars = 0; do igd=1,n_grids
   nvars = nvars+nmp33(igd)+nmp23(igd)+nmps33(igd)+nmps23(igd)+1
end do

do n=1,n_PE_leaf
   do nv=1,nvars

      call mpiq_receive_plain(infotags, n_infotags, PE_rank(n))

      igd = infotags(2); ivar = infotags(3); ivartype = infotags(4)

      il1 = gd(igd)%leafPE_ia(n);
      ir2 = gd(igd)%leafPE_iz(n)

      jb1 = gd(igd)%leafPE_ja(n); jt2 = gd(igd)%leafPE_jz(n)

      if (iand(gd(igd)%leafPE_ibndc(n),1) /= 0) il1 = il1-1
      if (iand(gd(igd)%leafPE_ibndc(n),2) /= 0) ir2 = ir2+1
      if (iand(gd(igd)%leafPE_ibndc(n),4) /= 0) jb1 = jb1-1
      if (iand(gd(igd)%leafPE_ibndc(n),8) /= 0) jt2 = jt2+1

      mxp = gd(igd)%leafPE_mxp(n);  myp = gd(igd)%leafPE_myp(n)
      mxyp = mxp*myp; mxyzp = mxyp*gd(igd)%leafPE_mzp(n)
      mxys3p = mxyp*(nzg+nzs)*npatch; mxys2p = mxyp*npatch

      ! REAL (4-byte) data:

      select case(ivartype)

      case(2)
         call mpiq_receive_plain(a(iscr1:iscr1+mxyp-1), mxyp, PE_rank(n))
         idx = ind2dn(igd,ivar)
         call esdbuff_2D(a(idx:idx+gd(igd)%nxyp-1), a(iscr1:iscr1+mxyp-1),  &
              gd(igd)%nxp, gd(igd)%nyp, mxp, myp,  &
              gd(igd)%leafPE_i0(n), gd(igd)%leafPE_j0(n), il1, ir2, jb1, jt2)
      case(3)
         call mpiq_receive_plain(a(iscr1:iscr1+mxyzp-1), mxyzp, PE_rank(n))
         idx = ind3dn(igd,ivar)
         call esdbuff_3D(a(idx:idx+gd(igd)%nxyzp-1), a(iscr1:iscr1+mxyzp-1),  &
              gd(igd)%nzp, gd(igd)%nxp, gd(igd)%nyp,  &
              mxp, myp, gd(igd)%leafPE_i0(n), gd(igd)%leafPE_j0(n), il1,  &
              ir2, jb1, jt2)
      case(4)
         call mpiq_receive_plain(a(iscr1:iscr1+mxys2p-1), mxys2p, PE_rank(n))
         idx = ind2dsn(igd,ivar)
         call esdbuff_2Ds(a(idx:idx+gd(igd)%nxys2p-1),  &
              a(iscr1:iscr1+mxys2p-1), gd(igd)%nxp, gd(igd)%nyp, npatch, mxp,  &
              myp, gd(igd)%leafPE_i0(n), gd(igd)%leafPE_j0(n), il1, ir2, jb1,  &
              jt2)      
      case(5)
         call mpiq_receive_plain(a(iscr1:iscr1+mxys3p-1), mxys3p, PE_rank(n))
         idx = ind3dsn(igd,ivar)
         call esdbuff_3Ds(a(idx:idx+gd(igd)%nxys3p-1),  &
              a(iscr1:iscr1+mxys3p-1), nzg+nzs, gd(igd)%nxp, gd(igd)%nyp,  &
              npatch, mxp, myp, gd(igd)%leafPE_i0(n), gd(igd)%leafPE_j0(n),  &
              il1, ir2, jb1, jt2)
      case(6)
         call mpiq_receive_plain(a(iscr1:iscr1+mxyp-1), mxyp, PE_rank(n))
         call esdbuff_2D(gd(igd)%emisst, a(iscr1:iscr1+mxyp-1),  &
              gd(igd)%nxp, gd(igd)%nyp, mxp, myp,  &
              gd(igd)%leafPE_i0(n), gd(igd)%leafPE_j0(n), il1, ir2, jb1, jt2)
      end select

   end do
end do

#endif  /* end:USE_MPI */

end subroutine rootPE_get_all_results
