!```````````````````````````````````````````````````````````````````
module parallel_samegrid_lBC_mod

!+

implicit none

logical :: acoustic_lBC_is_uninitialized = .true.,  &
           longdt_lBC_is_uninitialized = .true.


contains


!```````````````````````````````````````````````````````````````````
subroutine lBC_fill_buffer(n1, n2, n3, nm, a, b, il, ir, jb, jt)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3, il, ir, jb, jt, nm
real(f4), dimension(n1,n2,n3), intent(IN) :: a

real(f4), dimension(nm), intent(OUT) :: b

!== Local declarations:

integer :: i, j, k, idx

!=== Executable statements:

idx = 0
do j=jb,jt
   do i=il,ir
      do k=1,n1
         idx = idx+1; b(idx) = a(k,i,j)
      end do
   end do
end do

end subroutine lBC_fill_buffer


!```````````````````````````````````````````````````````````````````
subroutine lBC_empty_buffer(n1, n2, n3, nm, a, b, il, ir, jb, jt)

!+

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3, il, ir, jb, jt, nm
real(f4), dimension(nm), intent(IN) :: b

real(f4), dimension(n1,n2,n3), intent(OUT) :: a

!== Local declarations:

integer :: i, j, k, idx

!=== Executable statements:

idx = 0
do j=jb,jt
   do i=il,ir
      do k=1,n1
         idx = idx+1; a(k,i,j) = b(idx)
      end do
   end do
end do

end subroutine lBC_empty_buffer


!```````````````````````````````````````````````````````````````````
subroutine leafPE_acoustic_lBC_init

!+ Initializes some parameters relevant to the "leafPE_acoustic_lBC_*" 
!   procedures

use univ_kind_defs_mod, only: f4_B
use main_memory_mod, only: n_grids, n_PE_leaf, n_acoustic_lBC_optypes, gd

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: n, i1, i2, j1, j2, n_subdomain_W, igd, it

!=== Executable statements:

! Determine byte-length for packed data:

do igd=1,n_grids
   gd(igd)%acoustic_lBC_sd_packed_B(:,:) = 0
end do

do n=1,n_PE_leaf
   do igd=1,n_grids
      do it=1,n_acoustic_lBC_optypes
         if (gd(igd)%tx_paths_leaf(1,it,n) /= 0) then
            i1 = gd(igd)%tx_paths_leaf(1,it,n)
            i2 = gd(igd)%tx_paths_leaf(2,it,n)
            j1 = gd(igd)%tx_paths_leaf(3,it,n)
            j2 = gd(igd)%tx_paths_leaf(4,it,n)
            
            n_subdomain_W = (i2-i1+1)*(j2-j1+1)*gd(igd)%mmzp
            
            gd(igd)%acoustic_lBC_sd_packed_B(it,n) = n_subdomain_W*f4_B
         end if
      end do
   end do
end do

acoustic_lBC_is_uninitialized = .false.

end subroutine leafPE_acoustic_lBC_init


!```````````````````````````````````````````````````````````````````
subroutine leafPE_longdt_lBC_init

!+ Initializes some parameters relevant to the "leafPE_longdt_lBC_*" procedures

use univ_kind_defs_mod, only: f4_B
use main_memory_mod, only: n_grids, n_PE_leaf, gd, i_long_dt_lBC_comm

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: i1, i2, j1, j2, n, igd, n_subdomain_W

!=== Executable statements:

! Determine byte-length for packed data:

do igd=1,n_grids
   gd(igd)%longdt_lBC_sd_packed_B(:) = 0
end do

do n=1,n_PE_leaf
   do igd=1,n_grids
      if (gd(igd)%tx_paths_leaf(1,i_long_dt_lBC_comm,n) /= 0) then
         i1 = gd(igd)%tx_paths_leaf(1,i_long_dt_lBC_comm,n)
         i2 = gd(igd)%tx_paths_leaf(2,i_long_dt_lBC_comm,n)
         j1 = gd(igd)%tx_paths_leaf(3,i_long_dt_lBC_comm,n)
         j2 = gd(igd)%tx_paths_leaf(4,i_long_dt_lBC_comm,n)
         
         n_subdomain_W = (i2-i1+1)*(j2-j1+1)*gd(igd)%mmzp
         
         gd(igd)%longdt_lBC_sd_packed_B(n) = n_subdomain_W*f4_B
      end if
   end do
end do

longdt_lBC_is_uninitialized = .false.

end subroutine leafPE_longdt_lBC_init


end module parallel_samegrid_lBC_mod


!```````````````````````````````````````````````````````````````````
subroutine leafPE_acoustic_lBC_transmit(isflag, tx_bp, rx_bp, tx_ireq,  &
     rx_ireq, n_PE_leaf)

!+ Transmits relevant BCs every acoustic timestep to other leaf PEs

use univ_kind_defs_mod, only: f4_B, i4_B
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_receive_nonb_packed,  &
     mpiq_transmit_nonb_packed, mpiq_pack_buffer, mpiq_check_buffer_packed,  &
     MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use parallel_samegrid_lBC_mod, only: lBC_fill_buffer,  &
     acoustic_lBC_is_uninitialized
use main_memory_mod, only: i_grid, a, iscr1, maxnxp, maxnyp, maxnzp,  &
     acoustic_lBC_packed_pos, acoustic_lBC_sd_max_pB, my_rank, mzp, mxp, myp,  &
     PE_rank, i0, j0, n_acoustic_lBC_optypes, gd
use vtables_mod, only: iupn, ivpn, ippn, iwpn

implicit none

!=== Argument declarations:

integer, intent(IN) :: isflag, n_PE_leaf
integer, dimension(n_PE_leaf,n_acoustic_lBC_optypes), intent(INOUT) ::  &
     tx_ireq, rx_ireq

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), target,  &
     dimension(n_PE_leaf,n_acoustic_lBC_optypes), intent(INOUT) :: tx_bp, rx_bp
#else  /* Faux interface */
integer, target,  &
     dimension(n_PE_leaf,n_acoustic_lBC_optypes), intent(INOUT) :: tx_bp, rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: n, i1, i2, j1, j2, acoustic_lBC_comm_phase_idx, nW_per_var,  &
     acoustic_lBC_misc_packed_B, n_packed_B, n_relevant_vars, p_idx

!=== Executable statements:

#if defined(USE_MPI)

!--- Set some control parameters:

select case(isflag)
case(2,3,4)
   acoustic_lBC_comm_phase_idx = isflag; n_relevant_vars = 1
case(5,6)
   acoustic_lBC_comm_phase_idx = 1; n_relevant_vars = 2
end select
p_idx = acoustic_lBC_comm_phase_idx

!--- Before anything is transmitted, prepost the receives:

! Prepare receive buffer (and position indices):

acoustic_lBC_misc_packed_B = i4_B*(6)

n_packed_B = acoustic_lBC_sd_max_pB*n_relevant_vars+acoustic_lBC_misc_packed_B

! Prepost receives from leaf PEs:

do n=1,n_PE_leaf
   rx_ireq(n,p_idx) = 0

   if (gd(i_grid)%rx_paths_leaf(acoustic_lBC_comm_phase_idx,n) /= 0) then
      call mpiq_check_buffer_packed(rx_bp(n,p_idx), n_packed_B)
      call mpiq_receive_nonb_packed(rx_bp(n,p_idx), PE_rank(n),  &
           rx_ireq(n,p_idx), 0, acoustic_lBC_comm_phase_idx)
   end if
end do

!--- Transmit data to other leaf PEs as necessary:

do n=1,n_PE_leaf

   tx_ireq(n,p_idx) = 0
   if (gd(i_grid)%tx_paths_leaf(1,acoustic_lBC_comm_phase_idx,n) /= 0) then
 
      ! Prepare transmit buffer:
      n_packed_B =  &
           gd(i_grid)%acoustic_lBC_sd_packed_B(acoustic_lBC_comm_phase_idx,n)*  &
           n_relevant_vars+acoustic_lBC_misc_packed_B
      call mpiq_check_buffer_packed(tx_bp(n,p_idx), n_packed_B)

      i1 = gd(i_grid)%tx_paths_leaf(1,acoustic_lBC_comm_phase_idx,n)
      i2 = gd(i_grid)%tx_paths_leaf(2,acoustic_lBC_comm_phase_idx,n)
      j1 = gd(i_grid)%tx_paths_leaf(3,acoustic_lBC_comm_phase_idx,n)
      j2 = gd(i_grid)%tx_paths_leaf(4,acoustic_lBC_comm_phase_idx,n)

      call mpiq_pack_buffer(i1, tx_bp(n,p_idx))
      call mpiq_pack_buffer(i2, tx_bp(n,p_idx))
      call mpiq_pack_buffer(j1, tx_bp(n,p_idx))
      call mpiq_pack_buffer(j2, tx_bp(n,p_idx))
      call mpiq_pack_buffer(my_rank, tx_bp(n,p_idx))

      nW_per_var =  &
           gd(i_grid)%acoustic_lBC_sd_packed_B(acoustic_lBC_comm_phase_idx,n)/f4_B
      call mpiq_pack_buffer(nW_per_var, tx_bp(n,p_idx))

      select case(isflag)
      case(2)
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(iupn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n,p_idx))
      case(3)
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ivpn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n,p_idx))
      case(4)
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ippn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n,p_idx))
      case(5)
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(iupn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n,p_idx))
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ivpn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n,p_idx))
      case(6)
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(iwpn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n,p_idx))
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ippn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_pack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              tx_bp(n,p_idx))

!c rlw move the following thp send back to long timestep
!c               call mkstbuff(mzp,mxp,myp,maxnxp*maxnyp*maxnzp,a(ithpn(ngrid)),a(iscr1)
!c     +              ,i1-i0,i2-i0,j1-j0,j2-j0,mtp)
!c               call par_put_float(a(iscr1),mtp)

      end select

      call mpiq_transmit_nonb_packed(tx_bp(n,p_idx),  &  ! Transmit msg to PE
           gd(i_grid)%tx_paths_leaf(5,acoustic_lBC_comm_phase_idx,n),  &
           tx_ireq(n,p_idx), acoustic_lBC_comm_phase_idx)

   end if

end do

#endif  /* end:USE_MPI */

end subroutine leafPE_acoustic_lBC_transmit


!```````````````````````````````````````````````````````````````````
subroutine leafPE_acoustic_lBC_receive(isflag, rx_bp, tx_ireq, rx_ireq,  &
     n_PE_leaf)

!+ Receives relevant BCs every acoustic timestep from other leaf PEs

#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_wait, mpiq_unpack_buffer,  &
     MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use parallel_samegrid_lBC_mod, only: lBC_empty_buffer
use main_memory_mod, only: i_grid, a, iscr1, maxnxp, maxnyp, maxnzp, mzp,  &
     mxp, myp, i0, j0, acoustic_lBC_packed_pos, n_acoustic_lBC_optypes, gd
use vtables_mod, only: iupn, ivpn, ippn, iwpn

implicit none

!=== Argument declarations:

integer, intent(IN) :: isflag, n_PE_leaf
integer, dimension(n_PE_leaf,n_acoustic_lBC_optypes), intent(IN) :: tx_ireq,  &
     rx_ireq

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), target,  &
     dimension(n_PE_leaf,n_acoustic_lBC_optypes), intent(INOUT) :: rx_bp
#else  /* Faux interface */
integer, target,  &
     dimension(n_PE_leaf,n_acoustic_lBC_optypes), intent(INOUT) :: rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: acoustic_lBC_comm_phase_idx, i1, i2, j1, j2, n, nW_per_var,  &
     source_rank, p_idx

!type(MPI_buffer_packed_obj), pointer :: rx_bp => NULL()

!=== Executable statements:

#if defined(USE_MPI)

!--- Set some control parameters:

select case(isflag)
case(2,3,4)
   acoustic_lBC_comm_phase_idx = isflag
case(5,6)
   acoustic_lBC_comm_phase_idx = 1
end select
p_idx = acoustic_lBC_comm_phase_idx

!--- Ensure that all of our transmissions and receives have completed:

do n=1,n_PE_leaf      ! Transmissions
   if (gd(i_grid)%tx_paths_leaf(1,acoustic_lBC_comm_phase_idx,n) /= 0)  &
        call mpiq_wait(tx_ireq(n,p_idx))
end do

do n=1,n_PE_leaf      ! Receives
   if (gd(i_grid)%rx_paths_leaf(acoustic_lBC_comm_phase_idx,n) /= 0)  &
        call mpiq_wait(rx_ireq(n,p_idx))
end do

!--- Unpack the data received from other leaf PEs into appropriate arrays:

do n=1,n_PE_leaf

   if (gd(i_grid)%rx_paths_leaf(acoustic_lBC_comm_phase_idx,n) /= 0) then

      call mpiq_unpack_buffer(i1, rx_bp(n,p_idx))
      call mpiq_unpack_buffer(i2, rx_bp(n,p_idx))
      call mpiq_unpack_buffer(j1, rx_bp(n,p_idx))
      call mpiq_unpack_buffer(j2, rx_bp(n,p_idx))
      call mpiq_unpack_buffer(source_rank, rx_bp(n,p_idx))
      call mpiq_unpack_buffer(nW_per_var, rx_bp(n,p_idx))

      call mpiq_unpack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
           rx_bp(n,p_idx))

      select case(isflag)
      case(2)     
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(iupn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
      case(3)
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ivpn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
      case(4)
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ippn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
      case(5)
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(iupn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_unpack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              rx_bp(n,p_idx))
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ivpn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
      case(6)
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(iwpn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)
         call mpiq_unpack_buffer(a(iscr1:iscr1+nW_per_var-1), nW_per_var,  &
              rx_bp(n,p_idx))
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ippn(i_grid)), a(iscr1:iscr1+maxnxp*maxnyp*maxnzp-1), i1-i0,  &
              i2-i0, j1-j0, j2-j0)

!c rlw move the following thp receive back to long timestep
!c               call par_get_float(a(iscr1),mtp)
!c               call exstbuff(mzp,mxp,myp,maxnxp*maxnyp*maxnzp,a(ithpn(ngrid)),a(iscr1)
!c     +              ,i1-i0,i2-i0,j1-j0,j2-j0,mtc)

      end select

   end if

end do

#endif  /* end:USE_MPI */

end subroutine leafPE_acoustic_lBC_receive


!```````````````````````````````````````````````````````````````````
subroutine leafPE_longdt_lBC_transmit(tx_bp, rx_bp, tx_ireq, rx_ireq, n_PE_leaf)

!+ Transmits relevant BCs every long timestep to other leaf PEs

use univ_kind_defs_mod, only: f4_B, i4_B
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_receive_nonb_packed,  &
     mpiq_transmit_nonb_packed, mpiq_pack_buffer, mpiq_check_buffer_packed,  &
     MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use main_memory_mod, only: i_grid, a, maxnxp, maxnyp, maxnzp, PE_rank,  &
     my_rank, mzp, mxp, myp, longdt_lBC_packed_pos, i0, j0, gd,  &
     i_long_dt_lBC_comm
use vtables_mod, only: nmp31, nmp21, ind3dn, ind2dn, mp3tab1, mp2tab1, ivt3dp
use parallel_samegrid_lBC_mod, only: lBC_fill_buffer,  &
     longdt_lBC_is_uninitialized

implicit none

!=== Argument declarations:

integer, intent(IN) :: n_PE_leaf
integer, dimension(n_PE_leaf), intent(INOUT) :: tx_ireq, rx_ireq

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), dimension(n_PE_leaf), intent(INOUT) :: tx_bp, rx_bp
#else  /* Faux interface */
integer, dimension(n_PE_leaf), intent(INOUT) :: tx_bp, rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: i1, i2, j1, j2, n, n_relevant_vars, longdt_lBC_misc_packed_B,  &
     n_packed_B, nmp, nW_per_3D_var, nW_per_2D_var

!=== Executable statements:

#if defined(USE_MPI)

!--- Before anything is transmitted, prepost the receives:

! Prepare receive buffer (and position indices):

longdt_lBC_misc_packed_B = i4_B*(7)
n_relevant_vars = nmp31(i_grid)+nmp21(i_grid)

n_packed_B = maxval(gd(i_grid)%longdt_lBC_sd_packed_B(1:n_PE_leaf))*  &
     n_relevant_vars+longdt_lBC_misc_packed_B

! Prepost receives from leaf PEs:

do n=1,n_PE_leaf
   rx_ireq(n) = 0
   if (gd(i_grid)%rx_paths_leaf(i_long_dt_lBC_comm,n) /= 0) then
      call mpiq_check_buffer_packed(rx_bp(n), n_packed_B)
      call mpiq_receive_nonb_packed(rx_bp(n), PE_rank(n), rx_ireq(n), 0)
   end if
end do

!--- Transmit data to other leaf PEs as necessary:

do n=1,n_PE_leaf

   tx_ireq(n) = 0
   if (gd(i_grid)%tx_paths_leaf(1,i_long_dt_lBC_comm,n) /= 0) then

      ! Prepare transmit buffer:
      n_packed_B =  &
           gd(i_grid)%longdt_lBC_sd_packed_B(n)*n_relevant_vars+  &
           longdt_lBC_misc_packed_B
      call mpiq_check_buffer_packed(tx_bp(n), n_packed_B)

      i1 = gd(i_grid)%tx_paths_leaf(1,i_long_dt_lBC_comm,n)
      i2 = gd(i_grid)%tx_paths_leaf(2,i_long_dt_lBC_comm,n)
      j1 = gd(i_grid)%tx_paths_leaf(3,i_long_dt_lBC_comm,n)
      j2 = gd(i_grid)%tx_paths_leaf(4,i_long_dt_lBC_comm,n)

      call mpiq_pack_buffer(i1, tx_bp(n))
      call mpiq_pack_buffer(i2, tx_bp(n))
      call mpiq_pack_buffer(j1, tx_bp(n))
      call mpiq_pack_buffer(j2, tx_bp(n))
      call mpiq_pack_buffer(my_rank, tx_bp(n))

      nW_per_3D_var = gd(i_grid)%longdt_lBC_sd_packed_B(n)/f4_B
      nW_per_2D_var = nW_per_3D_var/mzp
      call mpiq_pack_buffer(nW_per_3D_var, tx_bp(n))
      call mpiq_pack_buffer(nW_per_2D_var, tx_bp(n))

      do nmp=1,nmp31(i_grid)
         call lBC_fill_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ind3dn(i_grid,mp3tab1(nmp,i_grid))), a(ivt3dp), i1-i0, i2-i0,  &
              j1-j0, j2-j0)
         call mpiq_pack_buffer(a(ivt3dp:ivt3dp+nW_per_3D_var-1),  &
              nW_per_3D_var, tx_bp(n))
      end do

      do nmp=1,nmp21(i_grid)
         call lBC_fill_buffer(1, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ind2dn(i_grid,mp2tab1(nmp,i_grid))), a(ivt3dp), i1-i0, i2-i0,  &
              j1-j0, j2-j0)
         call mpiq_pack_buffer(a(ivt3dp:ivt3dp+nW_per_2D_var-1),  &
              nW_per_2D_var, tx_bp(n))
      end do

      call mpiq_transmit_nonb_packed(tx_bp(n),  &      ! Transmit message to PE
           gd(i_grid)%tx_paths_leaf(5,i_long_dt_lBC_comm,n), tx_ireq(n))

   end if

end do

#endif  /* end:USE_MPI */

end subroutine leafPE_longdt_lBC_transmit


!```````````````````````````````````````````````````````````````````
subroutine leafPE_longdt_lBC_receive(rx_bp, tx_ireq, rx_ireq, n_PE_leaf)

!+ Receives relevant BCs every long timestep from other leaf PEs

#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_wait, mpiq_unpack_buffer,  &
     MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use vtables_mod, only: nmp31, nmp21, ind3dn, ind2dn, mp3tab1, mp2tab1, ivt3dp
use main_memory_mod, only: i_grid, a, maxnxp, maxnyp, maxnzp, mzp, mxp, myp,  &
     i0, j0, longdt_lBC_packed_pos, i_long_dt_lBC_comm, gd
use parallel_samegrid_lBC_mod, only: lBC_empty_buffer

implicit none

!=== Argument declarations:

integer, intent(IN) :: n_PE_leaf
integer, dimension(n_PE_leaf), intent(IN) :: tx_ireq, rx_ireq

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), dimension(n_PE_leaf), intent(INOUT) :: rx_bp
#else  /* Faux interface */
integer, dimension(n_PE_leaf), intent(INOUT) :: rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: n, nmp, source_rank, nW_per_3D_var, nW_per_2D_var, i1, i2, j1, j2

!=== Executable statements:

#if defined(USE_MPI)

!--- Ensure that all of our transmissions and receives have completed:

do n=1,n_PE_leaf      ! Transmissions
   if (gd(i_grid)%tx_paths_leaf(1,i_long_dt_lBC_comm,n) /= 0)  &
        call mpiq_wait(tx_ireq(n))
end do

do n=1,n_PE_leaf      ! Receives
   if (gd(i_grid)%rx_paths_leaf(i_long_dt_lBC_comm,n) /= 0)  &
        call mpiq_wait(rx_ireq(n))
end do

!--- Unpack the data received from other leaf PEs into appropriate arrays:

do n=1,n_PE_leaf

   if (gd(i_grid)%rx_paths_leaf(i_long_dt_lBC_comm,n) /= 0) then

      call mpiq_unpack_buffer(i1, rx_bp(n))
      call mpiq_unpack_buffer(i2, rx_bp(n))
      call mpiq_unpack_buffer(j1, rx_bp(n))
      call mpiq_unpack_buffer(j2, rx_bp(n))
      call mpiq_unpack_buffer(source_rank, rx_bp(n))
      call mpiq_unpack_buffer(nW_per_3D_var, rx_bp(n))
      call mpiq_unpack_buffer(nW_per_2D_var, rx_bp(n))

      do nmp=1,nmp31(i_grid)
         call mpiq_unpack_buffer(a(ivt3dp:ivt3dp+nW_per_3D_var-1),  &
              nW_per_3D_var, rx_bp(n))
         call lBC_empty_buffer(mzp, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ind3dn(i_grid,mp3tab1(nmp,i_grid))), a(ivt3dp), i1-i0, i2-i0,  &
              j1-j0, j2-j0)
      end do

      do nmp=1,nmp21(i_grid)
         call mpiq_unpack_buffer(a(ivt3dp:ivt3dp+nW_per_2D_var-1),  &
              nW_per_2D_var, rx_bp(n))
         call lBC_empty_buffer(1, mxp, myp, maxnxp*maxnyp*maxnzp,  &
              a(ind2dn(i_grid,mp2tab1(nmp,i_grid))), a(ivt3dp), i1-i0, i2-i0,  &
              j1-j0, j2-j0)
      end do

   end if

end do

#endif  /* end:USE_MPI */

end subroutine leafPE_longdt_lBC_receive
