!```````````````````````````````````````````````````````````````````
module parallel_timing_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine rootPE_get_timings(leafPE_times, rx_bp)

!+ Receives process_timings/CFL_criteria from all leaf PEs

use univ_kind_defs_mod, only: f4, i4_B, i8_B, f4_B, f8_B, cn_B, bn_B
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_unpack_buffer, mpiq_receive_packed,  &
     mpiq_check_buffer_packed, MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use main_memory_mod, only: n_grids, gd, n_PE_leaf, PE_rank

implicit none

!=== Argument declarations:

real(f4), dimension(:,:), intent(OUT) :: leafPE_times

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), intent(INOUT) :: rx_bp
#else  /* Faux interface */
integer, intent(INOUT) :: rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: n_packed_B, n, igd
real(f4) :: ftmp

!=== Executable statements:

#if defined(USE_MPI)

! Bytecount of primary message:

n_packed_B = i4_B*0 +  &
             i8_B*0 +  &
             f4_B*(2 + 2*n_grids) +  &
             f8_B*0 +  &
             cn_B*0 +  &
             bn_B*0

! REAL (4-byte) data:

gd(1:n_grids)%dt_SBPS_estimate = 99999._f4
gd(1:n_grids)%Courant_number_xyz = -999._f4
do n=1,n_PE_leaf
   call mpiq_check_buffer_packed(rx_bp, n_packed_B)   ! Prepare receive buffer
   call mpiq_receive_packed(rx_bp, PE_rank(n))      ! Receive message

   call mpiq_unpack_buffer(leafPE_times(n,1), rx_bp)      
   call mpiq_unpack_buffer(leafPE_times(n,2), rx_bp) ! Extract from rx buffer
   do igd=1,n_grids
      call mpiq_unpack_buffer(ftmp, rx_bp)
      gd(igd)%Courant_number_xyz =  &  ! Determine max value for all leaf PEs
           max(gd(igd)%Courant_number_xyz, ftmp)
      call mpiq_unpack_buffer(ftmp, rx_bp)
      gd(igd)%dt_SBPS_estimate =  &   ! Determine min value for all leaf PEs
           min(gd(igd)%dt_SBPS_estimate, ftmp)
   end do
end do

#endif  /* end:USE_MPI */

end subroutine rootPE_get_timings


!```````````````````````````````````````````````````````````````````
subroutine leafPE_put_timings(totcpu, totwall, tx_bp)

!+ Sends leaf PE process_timings/CFL_criteria to root PE

use univ_kind_defs_mod, only: f4, i4_B, i8_B, f4_B, f8_B, cn_B, bn_B
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_check_buffer_packed,  &
     mpiq_pack_buffer, mpiq_transmit_packed, MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use main_memory_mod, only: gd, n_grids, rootPE_rank

implicit none

!=== Argument declarations:

real(f4), intent(IN) :: totcpu, totwall

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), intent(INOUT) :: tx_bp
#else  /* Faux interface */
integer, intent(INOUT) :: tx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: n_packed_B, igd

!=== Executable statements:

#if defined(USE_MPI)

! Bytecount of primary message:

n_packed_B = i4_B*0 +  &
             i8_B*0 +  &
             f4_B*(2 + 2*n_grids) +  &
             f8_B*0 +  &
             cn_B*0 +  &
             bn_B*0

! Primary message:

call mpiq_check_buffer_packed(tx_bp, n_packed_B)     ! Prepare transmit buffer

! REAL (4-byte) data:

call mpiq_pack_buffer(totcpu, tx_bp)
call mpiq_pack_buffer(totwall, tx_bp)          ! Fill transmit buffer with data
do igd=1,n_grids
   call mpiq_pack_buffer(gd(igd)%Courant_number_xyz, tx_bp)
   call mpiq_pack_buffer(gd(igd)%dt_SBPS_estimate, tx_bp)
end do

call mpiq_transmit_packed(tx_bp, rootPE_rank)     ! Transmit message to root PE

#endif  /* end:USE_MPI */

end subroutine leafPE_put_timings


!```````````````````````````````````````````````````````````````````
subroutine rootPE_put_dtsched(all_results_from_leafPE_now,  &
     all_fields_to_leafPE_now, tx_bp)

!+ Transmits timing info to all leaf PEs

use univ_kind_defs_mod, only: i4_B, i8_B, f4_B, f8_B, cn_B, bn_B
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_check_buffer_packed,  &
     mpiq_pack_buffer, mpiq_transmit_packed, MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use main_memory_mod, only: pbanner, pbnlen, isched, n_isched_steps,  &
     n_isched_entries, n_grids, maxnzp, maxnyp, maxnxp, sspct, gd, PE_rank,  &
     n_PE_leaf, nsubs

implicit none

!=== Argument declarations:

logical, intent(IN) :: all_results_from_leafPE_now, all_fields_to_leafPE_now

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), intent(INOUT) :: tx_bp
#else  /* Faux interface */
integer, intent(INOUT) :: tx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: n, n_packed_B, igd

!=== Executable statements:

#if defined(USE_MPI)

! Bytecount of primary message:

n_packed_B =   &
        i4_B*(1 + 3*n_grids + n_isched_steps*n_isched_entries) +  &
        i8_B*0 +  &
        f4_B*(1 + 2*n_grids) +  &
        f8_B*0 +  &
        cn_B*0 +  &
        bn_B*2

! Primary message:

call mpiq_check_buffer_packed(tx_bp, n_packed_B)     ! Prepare transmit buffer

! LOGICAL data:

call mpiq_pack_buffer(all_results_from_leafPE_now, tx_bp)
call mpiq_pack_buffer(all_fields_to_leafPE_now, tx_bp)

! INTEGER (4-byte) data:

call mpiq_pack_buffer(nsubs, tx_bp)

do igd=1,n_grids   ! Fill transmit buffer with data
   call mpiq_pack_buffer(gd(igd)%n_dta_per_dtdyn, tx_bp)
   call mpiq_pack_buffer(gd(igd)%n_dt_per_pgdt, tx_bp)
   call mpiq_pack_buffer(gd(igd)%n_iter_SBPS, tx_bp)
end do

call mpiq_pack_buffer(isched, n_isched_steps, n_isched_entries, tx_bp)

! REAL (4-byte) data:

call mpiq_pack_buffer(sspct, tx_bp)
do igd=1,n_grids
   call mpiq_pack_buffer(gd(igd)%dt_dyn, tx_bp)
   call mpiq_pack_buffer(gd(igd)%dt_SBPS, tx_bp)
end do

do n=1,n_PE_leaf
   call mpiq_transmit_packed(tx_bp, PE_rank(n))   ! Transmit message to leaf PE
end do

#endif  /* end:USE_MPI */

end subroutine rootPE_put_dtsched


!```````````````````````````````````````````````````````````````````
subroutine leafPE_get_dtsched(all_results_to_rootPE_now,  &
     all_fields_from_rootPE_now, rx_bp)

!+ Receives timing info from root PE

use univ_kind_defs_mod, only: i4_B, i8_B, f4_B, f8_B, cn_B, bn_B
#if defined(USE_MPI)
use univ_MPI_F90_intf_mod, only: mpiq_unpack_buffer, mpiq_receive_packed,  &
     mpiq_check_buffer_packed, MPI_buffer_packed_obj
#endif  /* end:USE_MPI */
use main_memory_mod, only: isched, n_isched_steps, n_isched_entries, n_grids,  &
     maxnzp, maxnyp, maxnxp, sspct, gd, rootPE_rank, nsubs

implicit none

!=== Argument declarations:

logical, intent(OUT) :: all_results_to_rootPE_now, all_fields_from_rootPE_now

#if defined(USE_MPI)
type(MPI_buffer_packed_obj), intent(INOUT) :: rx_bp
#else  /* Faux interface */
integer, intent(INOUT) :: rx_bp
#endif  /* end:USE_MPI */

!== Local declarations:

integer :: n_packed_B, igd

logical :: send_all_results_now

!=== Executable statements:

#if defined(USE_MPI)

! Bytecount of primary message:

n_packed_B =  &
        i4_B*(1 + 3*n_grids + n_isched_steps*n_isched_entries) +  &
        i8_B*0 +  &
        f4_B*(1 + 2*n_grids) +  &
        f8_B*0 +  &
        cn_B*0 +  &
        bn_B*2

!--- Get message body:

call mpiq_check_buffer_packed(rx_bp, n_packed_B)   ! Prepare receive buffer
call mpiq_receive_packed(rx_bp, rootPE_rank)      ! Receive message

! LOGICAL data:

call mpiq_unpack_buffer(all_results_to_rootPE_now, rx_bp)
call mpiq_unpack_buffer(all_fields_from_rootPE_now, rx_bp)

! INTEGER (4-byte) data:

call mpiq_unpack_buffer(nsubs, rx_bp)

do igd=1,n_grids   ! Extract data from receive buffer
   call mpiq_unpack_buffer(gd(igd)%n_dta_per_dtdyn, rx_bp)
   call mpiq_unpack_buffer(gd(igd)%n_dt_per_pgdt, rx_bp)
   call mpiq_unpack_buffer(gd(igd)%n_iter_SBPS, rx_bp)
end do

call mpiq_unpack_buffer(isched, n_isched_steps, n_isched_entries, rx_bp)

! REAL (4-byte) data:

call mpiq_unpack_buffer(sspct, rx_bp)
do igd=1,n_grids
   call mpiq_unpack_buffer(gd(igd)%dt_dyn, rx_bp)
   call mpiq_unpack_buffer(gd(igd)%dt_SBPS, rx_bp)
end do

#endif  /* end:USE_MPI */

end subroutine leafPE_get_dtsched


end module parallel_timing_mod
