!```````````````````````````````````````````````````````````````````
module search_mod

!~ 24.09.2008 (last modified; TIM)

!+

implicit none

private

public :: jsrchgt, jsrchge, srchgt_dblinp

contains


!```````````````````````````````````````````````````````````````````
function jsrchgt(n,sx,inc,target) result(answer)

!~ 18.11.2008 (last modified; TIM)

!+ Find first array element greater than target value

!      Find the first array element that is greater than the 
!      TARGET value.  If N < 1, then 0 is returned.  If no
!      value is found, N+1 is returned.  Replaces Cray version
!      on the workstation.
!      Bisection search implemented 09/29/93 to improve speed.
!
!      Started: 08/23/93
!
!      Input:
!        N      - Number of elements in array SX.
!        SX     - Array of numbers to be searched.  Assumed to be an
!                 ordered array.
!        INC    - Increment between elements of searched array.  Kept
!                 for compatibility with Cray call.
!        TARGET - Value searched for in array SX.
!
!      Output:
!        JSRCHGT - location in array SX where value of SX is first
!                   greater than the TARGET value.

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer :: answer

integer, intent(IN) :: n,inc
real(f4), dimension(n), intent(IN) :: SX
real(f4), intent(IN) :: target

!== Local declarations:

integer :: ians,jl,jh,jm

!=== Executable statements:

if(n < 1) then
   ians = 0
elseif(target > sx(n)) then
   ians = n+1
elseif(TARGET < SX(1)) then
   ians = 1
else
   JL = 1
   JH = N

   do while(JH-JL > 1)
      JM = (JL+JH)/2
      
      if(TARGET > SX(JM)) then
         JL = JM
         JM = (JL+JH)/2
      else
         JH = JM
         JM = (JL+JH)/2
      endif   
   enddo

   if(TARGET == SX(JH)) JH = JH+1

   ians = JH
endif

answer = ians

end function jsrchgt


!```````````````````````````````````````````````````````````````````
function jsrchge(N,SX,INC,TARGET) result(answer)

!~ 18.11.2008 (last modified; TIM)

!+ Find first array element greater than or equal to target value

!      Find the first array element that is greater than or equal to
!      the TARGET value.  If N < 1, then 0 is returned.  If no value
!      is found, N+1 is returned.  Replaces Cray version on the 
!      workstation.
!
!      Started: 08/23/93
!
!      Input:
!        N      - Number of elements in array SX.
!        SX     - Array of numbers to be searched.  Assumed to be an
!                 ordered array.
!        INC    - Increment between elements of searched array.  Kept
!                 for compatibility with Cray call.
!        TARGET - Value searched for in array SX.
!
!      Output:
!        JSRCHGT - location in array SX where value of SX is first
!                   greater than or equal to the TARGET value.

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer :: answer

integer, intent(IN) :: n,inc
real(f4), dimension(n), intent(IN) :: SX
real(f4), intent(IN) :: target

!== Local declarations:

integer :: ians,i

!=== Executable statements:

if(n < 1) then
   ians = 0
else
   ians = n+1
   do i=1,n
      if(sx(i) >= target) then
         ians = i
         exit
      endif
   enddo
endif

answer = ians

end function jsrchge


!```````````````````````````````````````````````````````````````````
function srchgt_dblinp(n,sx,inc,target) result(answer)

!~ 18.11.2008 (last modified; TIM)

!+ Find first array element greater than target value (8-byte input)

!      Find the first array element that is greater than the 
!      TARGET value.  If N < 1, then 0 is returned.  If no
!      value is found, N+1 is returned.  Replaces Cray version
!      on the workstation.
!      Bisection search implemented 09/29/93 to improve speed.
!
!      Started: 08/23/93
!
!      Input:
!        N      - Number of elements in array SX.
!        SX     - Array of numbers to be searched.  Assumed to be an
!                 ordered array.
!        INC    - Increment between elements of searched array.  Kept
!                 for compatibility with Cray call.
!        TARGET - Value searched for in array SX.
!
!      Output:
!        JSRCHGT - location in array SX where value of SX is first
!                   greater than the TARGET value.

use univ_kind_defs_mod, only: f8

implicit none

!=== Argument declarations:

integer :: answer

integer, intent(IN) :: n,inc
real(f8), dimension(n), intent(IN) :: sx
real(f8), intent(IN) :: target

!== Local declarations:

integer :: ians,lwr,upr,mid

!=== Executable statements:

if(target > sx(n)) then
   ians = n+1
elseif(target < sx(1)) then
   ians = 1
else
   lwr = 1
   upr = n

   do while(upr-lwr > 1)
      mid = (lwr+upr)/2
      if(target > sx(mid)) then
         lwr = mid
         mid = (lwr+upr)/2
      else
         upr = mid
         mid = (lwr+upr)/2
      endif
   enddo

   if(target == sx(upr)) upr = upr+1
   ians = upr
endif

answer = ians

end function srchgt_dblinp


end module search_mod
