!```````````````````````````````````````````````````````````````````
module setup_model_storage_mod

!+

implicit none

integer, parameter :: a_idx_default = -1000000000


contains


!```````````````````````````````````````````````````````````````````
subroutine mem_setupA(I_want_to_display)

!+ Calculate memory usage for A array (serial run)

use univ_kind_defs_mod, only: f4_B, i8, f8
use univ_convert_units_mod, only: distill_bytes
use main_memory_mod, only: n_grids, maxnxp, maxnyp, maxnzp, npatch, nzg, nzs,  &
     iscr1, iscr2, memsizeA, n_vars, gd, maxMxp, maxMyp, maxMzp
use vtables_mod, only: nv3dmt, ind3dm, ivtab3m, nv2d, ind2dn, ivtab2d,  &
     nv2ds, ind2dsn, ivtab2ds, nv3ds, ind3dsn, ivtab3ds, nv3d, ind3dn,  &
     ivtab3d, ivchr3dm, ivchr3d, ivchr2d, ivchr3ds, ivchr2ds, i_state_chr3D,  &
     i_state_chr2D, i_state_chr3Ds, i_state_chr2Ds, n_state_3D, n_state_2D,  &
     n_state_3Ds, n_state_2Ds, nsclr, ischrp,  &
     mp3chr1, mp2chr1, mp3schr1, mp2schr1, nmp31, nmp21, nmps31, nmps21,  &
     mp3chr2, mp2chr2, mp3schr2, mp2schr2, nmp32, nmp22, nmps32, nmps22,  &
     mp3chr3, mp2chr3, mp3schr3, mp2schr3, nmp33, nmp23, nmps33, nmps23,  &
     mp3chri, mp2chri, mp3schri, mp2schri, nmp3i, nmp2i, nmps3i, nmps2i,  &
     ind3d, ind2d, ind3ds, ind2ds, vtable_memidx

implicit none

!=== Argument declarations:

logical, intent(IN) :: I_want_to_display

!== Local declarations:

integer :: igd, iv, maxnxyzp, maxnxysp, n2, n3, ns2, ns3, memscr1, memval,  &
     ms_state
character(len=3) :: bunits
real(f8) :: bsize

logical, parameter :: show_debugging_output = .FALSE.

!=== Executable statements:

! Initialize all A-array index values to a_idx_default (to help detect coding
!  errors):

do iv=0,n_vars
   ind3dm(iv) = a_idx_default
   ind3d(iv) = a_idx_default; ind2d(iv) = a_idx_default
   ind3ds(iv) = a_idx_default; ind2ds(iv) = a_idx_default
   do igd=1,n_grids
      ind3dn(igd,iv) = a_idx_default; ind2dn(igd,iv) = a_idx_default
      ind3dsn(igd,iv) = a_idx_default; ind2dsn(igd,iv) = a_idx_default
   end do
end do

maxnxyzp = maxnxp*maxnyp*maxnzp  ! Maximum # of points needed for any grid var
maxnxysp = (nzg+nzs)*maxnxp*maxnyp*npatch

maxMxp = maxnxp; maxMyp = maxnyp; maxMzp = maxnzp

! Two scratch arrays are needed that are a maximum 3D variable long;
!  these will be the first part of the A array:

memscr1 = max(maxnxyzp, maxnxysp)
iscr1 = 1; iscr2 = iscr1+memscr1

! Assign grid-independent variable indices:

memval = iscr2+memscr1

do iv=1,nv3dmt
   ind3dm(ivtab3m(iv)) = memval
   memval = memval+maxnxyzp
end do
call vtable_memidx(0, .FALSE.)   ! Assign named indices

do igd=1,n_grids

   do iv=1,nv2d(igd)
      ind2dn(igd,ivtab2d(iv,igd)) = memval    ! 2D "normal"
      memval = memval+gd(igd)%nxyp
   end do

   do iv=1,nv2ds(igd)
      ind2dsn(igd,ivtab2ds(iv,igd)) = memval   ! 2D "sub/super surface"
      memval = memval+gd(igd)%nxys2p
   end do

   do iv=1,nv3ds(igd)
      ind3dsn(igd,ivtab3ds(iv,igd)) = memval   ! 3D "sub/super surface"
      memval = memval+gd(igd)%nxys3p
   end do

   do iv=1,nv3d(igd)
      ind3dn(igd,ivtab3d(iv,igd)) = memval   ! 3D "normal"
      memval = memval+gd(igd)%nxyzp
   end do

   call vtable_memidx(igd, .FALSE.)    ! Assign named indices

end do

memsizeA = memval   ! Total size {words} of "A" array

! Possibly print information about variables, memory usage, etc.:

if (I_want_to_display) then

   write (*, '(/,/,78a)') '+', ('-', igd=1,77) 
   write (*, '(25x,a)') '"A" ARRAY:  MEMORY STATISTICS'

   ms_state = 0
   do igd=1,n_grids
      n3 = gd(igd)%nxyzp; n2 = gd(igd)%nxyp
      ns2 = gd(igd)%nxys2p; ns3 = gd(igd)%nxys3p

      ms_state = ms_state+(n_state_3D(igd)*n3+n_state_2D(igd)*n2+  &
           n_state_3Ds(igd)*ns3+n_state_2D(igd)*ns2)
   end do
   
   call distill_bytes(int(memsizeA*f4_B, kind=i8), bsize, bunits)
   write (*, '(/,/,5x,a,f7.2,1x,a)') '***** Total memory to be allocated: ',  &
        bsize, trim(bunits)
   call distill_bytes(int(ms_state*f4_B, kind=i8), bsize, bunits)
   write (*, '(5x,a,f7.2,1x,a)')   &
        '***** State file size (includes all grids): ', bsize, trim(bunits)
   
   if (show_debugging_output) then
      write (*, '(/,/,a)')   &
           '========= Memory table information for grid-independent variables:'
      write (*, '(/,5x,a,/)') 'Name of table             #Vars      Size'
   end if

   call distill_bytes(int(nv3dmt*maxnxyzp*f4_B, kind=i8), bsize, bunits)
   if (show_debugging_output) then
      write (*, '(a,3x,i3,5x,f7.2,1x,a)') 'grid-independent............ ',  &
           nv3dmt, bsize, trim(bunits)
      write (*, '(/,(t6,(5(a8,3x))))') ivchr3dm(1:nv3dmt,1)
   end if

   do igd=1,n_grids    ! Loop over all grids

      n3 = gd(igd)%nxyzp; n2 = gd(igd)%nxyp
      ns2 = gd(igd)%nxys2p; ns3 = gd(igd)%nxys3p

      if (show_debugging_output) then
         write (*, '(/,/,a,i2,a)')   &
              '========= Memory table information on grid #', igd, ':'
         write (*, '(/,5x,a,/)') 'Name of table             #Vars      Size'
      end if

      call distill_bytes(int(nv3d(igd)*n3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D variable................. ', nv3d(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') ivchr3d(1:nv3d(igd),igd)
      end if

      call distill_bytes(int(nv2d(igd)*n2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D variable................. ', nv2d(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') ivchr2d(1:nv2d(igd),igd)
      end if

      call distill_bytes(int(nv3ds(igd)*ns3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D sub/super sfc variable... ', nv3ds(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') ivchr3ds(1:nv3ds(igd),igd)
      end if

      call distill_bytes(int(nv2ds(igd)*ns2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D sub/super sfc variable... ', nv2ds(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') ivchr2ds(1:nv2ds(igd),igd)
      end if

      call distill_bytes(int(n_state_3D(igd)*n3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D state................. ', n_state_3D(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') i_state_chr3D(1:n_state_3D(igd),igd)
      end if

      call distill_bytes(int(n_state_2D(igd)*n2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D state................. ', n_state_2D(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') i_state_chr2D(1:n_state_2D(igd),igd)
      end if

      call distill_bytes(int(n_state_3Ds(igd)*ns3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D sub/super sfc state... ', n_state_3Ds(igd), bsize,  &
              trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') i_state_chr3Ds(1:n_state_3Ds(igd),igd)
      end if

      call distill_bytes(int(n_state_2Ds(igd)*ns2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D sub/super sfc state... ', n_state_2Ds(igd), bsize,  &
              trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') i_state_chr2Ds(1:n_state_2Ds(igd),igd)
      end if

      call distill_bytes(int(nsclr(igd)*n3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D scalar variable........... ', nsclr(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') ischrp(1:nsclr(igd),igd)
      end if

      call distill_bytes(int(nmp31(igd)*n3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D MP dtL bnd N->N.......... ', nmp31(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3chr1(1:nmp31(igd),igd)
      end if

      call distill_bytes(int(nmp21(igd)*n2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D MP dtL bnd N->N.......... ', nmp21(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2chr1(1:nmp21(igd),igd)
      end if

      call distill_bytes(int(nmps31(igd)*ns3*f4_B, kind=i8), bsize, bunits)
      if(show_debugging_output) then
         write (*,' (/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D sub/supsfc dtL bnd N->N.. ', nmps31(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3schr1(1:nmps31(igd),igd)
      end if

      call distill_bytes(int(nmps21(igd)*ns2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D sub/supsfc dtL bnd N->N.. ', nmps21(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2schr1(1:nmps21(igd),igd)
      end if

      call distill_bytes(int(nmp32(igd)*n3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D dtS bnd N->N............. ', nmp32(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3chr2(1:nmp32(igd),igd)
      end if

      call distill_bytes(int(nmp22(igd)*n2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D dtS bnd N->N............. ', nmp22(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2chr2(1:nmp22(igd),igd)
      end if

      call distill_bytes(int(nmps32(igd)*ns3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D sub/supsfc dtS bnd N->N.. ', nmps32(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3schr2(1:nmps32(igd),igd)
      end if

      call distill_bytes(int(nmps22(igd)*ns2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D sub/supsfc dtS bnd N->N.. ', nmps22(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2schr2(1:nmps22(igd),igd)
      end if

      call distill_bytes(int(nmp33(igd)*n3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D output N->M.............. ', nmp33(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3chr3(1:nmp33(igd),igd)
      end if

      call distill_bytes(int(nmp23(igd)*n2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D output N->M.............. ', nmp23(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2chr3(1:nmp23(igd),igd)
      end if

      call distill_bytes(int(nmps33(igd)*ns3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D sub/supsfc output N->M... ', nmps33(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3schr3(1:nmps33(igd),igd)
      end if

      call distill_bytes(int(nmps23(igd)*ns2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D sub/supsfc output N->M... ', nmps23(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2schr3(1:nmps23(igd),igd)
      end if

      call distill_bytes(int(nmp3i(igd)*n3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D init M->N................ ', nmp3i(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3chri(1:nmp3i(igd),igd)
      end if

      call distill_bytes(int(nmp2i(igd)*n2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D init M->N................ ', nmp2i(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2chri(1:nmp2i(igd),igd)
      end if

      call distill_bytes(int(nmps3i(igd)*ns3*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '3D sub/supsfc init M->N..... ', nmps3i(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp3schri(1:nmps3i(igd),igd)
      end if

      call distill_bytes(int(nmps2i(igd)*ns2*f4_B, kind=i8), bsize, bunits)
      if (show_debugging_output) then
         write (*, '(/a,3x,i3,5x,f7.2,1x,a)')  &
              '2D sub/supsfc init M->N..... ', nmps2i(igd), bsize, trim(bunits)
         write (*, '(/,(t6,(5(a8,3x))))') mp2schri(1:nmps2i(igd),igd)
      end if

   end do

end if

end subroutine mem_setupA


!```````````````````````````````````````````````````````````````````
subroutine mem_setupB(I_want_to_display)

!+ Calculate memory usage for B array (nesting boundary tendencies)

use univ_kind_defs_mod, only: f4_B, i8, f8
use univ_convert_units_mod, only: distill_bytes
use main_memory_mod, only: gd, n_grids, memsizeB
use vtables_mod, only: nsclr, ibux, ibuy, ibuz, ibvx, ibvy, ibvz, ibwx,  &
     ibwy, ibwz, ibpx, ibpy, ibpz, ibsx, ibsy, ibsz, ibax, ibay, ibaz,  &
     naersclr

implicit none

!=== Argument declarations:

logical, intent(IN) :: I_want_to_display

!== Local declarations:

integer :: igd, ind, jsc
real(f8) :: bsize
character(len=3) :: bunits

!=== Executable statements:

! Compute memory space needed for nesting boundary tendencies:

if (n_grids == 1) then
   memsizeB = 1
else
   memsizeB = 0
   do igd=2,n_grids
      if (gd(igd)%nxtnest >= 1) then
         memsizeB = memsizeB+(gd(igd)%nxp*gd(igd)%nyp+gd(igd)%nxp*gd(igd)%nzp  &
              +gd(igd)%nyp*gd(igd)%nzp)*2*(4+nsclr(igd)+naersclr(igd))
      end if
   end do
end if

if (I_want_to_display) then

   write (*, '(/,/,78a)') '+', ('-', igd=1,77) 
   write (*, '(25x,a)') '"B" ARRAY:  MEMORY STATISTICS'

   call distill_bytes(int(memsizeB*f4_B, kind=i8), bsize, bunits)
   write (*, '(/,/,5x,a,f7.2,1x,a)') '***** Total memory to be allocated: ',  &
        bsize, trim(bunits)

end if

! Fill index arrays:

ind = 1
do igd=2,n_grids

   if (gd(igd)%nxtnest >= 1) then

      ibux(igd) = ind

      ind = ind+2*gd(igd)%nzp*gd(igd)%nyp; ibuy(igd) = ind
      ind = ind+2*gd(igd)%nzp*gd(igd)%nxp; ibuz(igd) = ind
      ind = ind+2*gd(igd)%nxp*gd(igd)%nyp; ibvx(igd) = ind
      ind = ind+2*gd(igd)%nzp*gd(igd)%nyp; ibvy(igd) = ind
      ind = ind+2*gd(igd)%nzp*gd(igd)%nxp; ibvz(igd) = ind
      ind = ind+2*gd(igd)%nxp*gd(igd)%nyp; ibwx(igd) = ind
      ind = ind+2*gd(igd)%nzp*gd(igd)%nyp; ibwy(igd) = ind
      ind = ind+2*gd(igd)%nzp*gd(igd)%nxp; ibwz(igd) = ind
      ind = ind+2*gd(igd)%nxp*gd(igd)%nyp; ibpx(igd) = ind
      ind = ind+2*gd(igd)%nzp*gd(igd)%nyp; ibpy(igd) = ind
      ind = ind+2*gd(igd)%nzp*gd(igd)%nxp; ibpz(igd) = ind

      ind = ind+2*gd(igd)%nxp*gd(igd)%nyp

      do jsc=1,nsclr(igd)      ! "normal" scalars
         ibsx(jsc,igd) = ind
         ind = ind+2*gd(igd)%nzp*gd(igd)%nyp; ibsy(jsc,igd) = ind
         ind = ind+2*gd(igd)%nzp*gd(igd)%nxp; ibsz(jsc,igd) = ind
         ind = ind+2*gd(igd)%nxp*gd(igd)%nyp
      end do

      do jsc=1,naersclr(igd)      ! "aerosol" scalars
         ibax(jsc,igd) = ind
         ind = ind+2*gd(igd)%nzp*gd(igd)%nyp; ibay(jsc,igd) = ind
         ind = ind+2*gd(igd)%nzp*gd(igd)%nxp; ibaz(jsc,igd) = ind
         ind = ind+2*gd(igd)%nxp*gd(igd)%nyp
      end do

   end if
end do

end subroutine mem_setupB


!```````````````````````````````````````````````````````````````````
subroutine rootPE_memsetA(I_want_to_display)

!+ Calculate memory usage for A array (parallel; root PE)

use univ_kind_defs_mod, only: f4_B, i8, f8
use univ_convert_units_mod, only: distill_bytes
use main_memory_mod, only: n_grids, maxnxp, maxnyp, maxnzp, npatch, nzg, nzs,  &
     iscr1, iscr2, memsizeA, n_vars, gd, maxMxp, maxMyp, maxMzp
use vtables_mod, only: nv3dmt, ind3dm, ivtab3m, nv2d, ind2dn, ivtab2d,  &
     nv2ds, ind2dsn, ivtab2ds, nv3ds, ind3dsn, ivtab3ds, nv3d, ind3dn,  &
     ivtab3d, n_state_3D, n_state_2D, n_state_3Ds,  &
     ind3d, ind2d, ind3ds, ind2ds, vtable_memidx, ivopt3m, ivopt3d, ivopt2d,  &
     ivopt3ds, ivopt2ds

implicit none

!=== Argument declarations:

logical, intent(IN) :: I_want_to_display

!== Local declarations:

integer :: igd, iv, maxnxyzp, maxnxysp, n2, n3, ns2, ns3, memscr1, memval,  &
     ms_state
character(len=3) :: bunits
real(f8) :: bsize

!=== Executable statements:

! Initialize all A-array index values to a_idx_default (to help detect coding
!  errors):

do iv=0,n_vars
   ind3dm(iv) = a_idx_default
   ind3d(iv) = a_idx_default; ind2d(iv) = a_idx_default
   ind3ds(iv) = a_idx_default; ind2ds(iv) = a_idx_default
   do igd = 1,n_grids
      ind3dn(igd,iv) = a_idx_default; ind2dn(igd,iv) = a_idx_default
      ind3dsn(igd,iv) = a_idx_default; ind2dsn(igd,iv) = a_idx_default
   end do
end do

maxnxyzp = maxnxp*maxnyp*maxnzp  ! Maximum # of points needed for any grid var
maxnxysp = (nzg+nzs)*maxnxp*maxnyp*npatch

maxMxp = maxnxp; maxMyp = maxnyp; maxMzp = maxnzp

! Two scratch arrays are needed that are a maximum 3D variable long.
!  These will be the first part of the A array:

memscr1 = max(maxnxyzp, maxnxysp)
iscr1 = 1; iscr2 = iscr1+memscr1

! Assign grid-independent variable indices:

memval = iscr2+memscr1

do iv=1,nv3dmt
   if (ivopt3m(iv)(2:2) == '1') then
      ind3dm(ivtab3m(iv)) = memval
      memval = memval+maxnxyzp
   end if
end do
call vtable_memidx(0, .FALSE.)   ! Assign named indices

do igd=1,n_grids

   do iv=1,nv2d(igd)
      if (ivopt2d(iv,igd)(2:2) == '1') then
         ind2dn(igd,ivtab2d(iv,igd)) = memval    ! 2D "normal"
         memval = memval+gd(igd)%nxyp
      end if
   end do

   do iv=1,nv2ds(igd)
      if (ivopt2ds(iv,igd)(2:2) == '1') then
         ind2dsn(igd,ivtab2ds(iv,igd)) = memval   ! 2D "sub/super surface"
         memval = memval+gd(igd)%nxys2p
      end if
   end do

   do iv=1,nv3ds(igd)
      if (ivopt3ds(iv,igd)(2:2) == '1') then
         ind3dsn(igd,ivtab3ds(iv,igd)) = memval   ! 3D "sub/super surface"
         memval = memval+gd(igd)%nxys3p
      end if
   end do

   do iv=1,nv3d(igd)
      if (ivopt3d(iv,igd)(2:2) == '1') then
         ind3dn(igd,ivtab3d(iv,igd)) = memval   ! 3D "normal"
         memval = memval+gd(igd)%nxyzp
      end if
   end do

   call vtable_memidx(igd, .FALSE.)    ! Assign named indices

end do

memsizeA = memval   ! Total size [words] of "A" array

! Possibly print information about variables, memory usage, etc.:

if (I_want_to_display) then
   write (*, '(/,/,78a)') '+', ('-', igd=1,77)
   write (*, '(25x,a)') '"A" ARRAY:  MEMORY STATISTICS'

   ms_state = 0
   do igd=1,n_grids
      n3 = gd(igd)%nxyzp; n2 = gd(igd)%nxyp
      ns2 = gd(igd)%nxys2p; ns3 = gd(igd)%nxys3p

      ms_state = ms_state+(n_state_3D(igd)*n3+n_state_2D(igd)*n2+  &
           n_state_3Ds(igd)*ns3+n_state_2D(igd)*ns2)
   end do
   
   call distill_bytes(int(memsizeA*f4_B, kind=i8), bsize, bunits)
   write (*, '(/,/,5x,a,f7.2,1x,a)') '***** Total memory to be allocated: ',  &
        bsize, trim(bunits)
   call distill_bytes(int(ms_state*f4_B, kind=i8), bsize, bunits)
   write (*, '(5x,a,f7.2,1x,a)')   &
        '***** State file size (includes all grids): ', bsize, trim(bunits)

end if

end subroutine rootPE_memsetA


!```````````````````````````````````````````````````````````````````
subroutine leafPE_memsetA(I_want_to_display)

!+ Calculate memory usage for A array (parallel; leaf PE)

use univ_kind_defs_mod, only: f4_B, i8, f8
use univ_convert_units_mod, only: distill_bytes
use main_memory_mod, only: n_grids, maxnxp, maxnyp, maxnzp, npatch, nzg, nzs,  &
     iscr1, iscr2, memsizeA, n_vars, gd, maxMxp, maxMyp, maxMzp
use vtables_mod, only: nv3dmt, ind3dm, ivtab3m, nv2d, ind2dn, ivtab2d,  &
     nv2ds, ind2dsn, ivtab2ds, nv3ds, ind3dsn, ivtab3ds, nv3d, ind3dn,  &
     ivtab3d,  &
     ind3d, ind2d, ind3ds, ind2ds, vtable_memidx, ivopt3m, ivopt3d, ivopt2d,  &
     ivopt3ds, ivopt2ds

implicit none

!=== Argument declarations:

logical, intent(IN) :: I_want_to_display

!== Local declarations:

integer :: igd, iv, maxnxyzp, maxnxysp, memscr1, memval, mmxyp, mmxyzp,  &
     mmxys2p, mmxys3p, maxMxyzp, maxMxysp
character(len=3) :: bunits
real(f8) :: bsize

!=== Executable statements:

! Initialize all A-array index values to a_idx_default (to help detect coding
!  errors):

do iv = 0,n_vars
   ind3dm(iv) = a_idx_default
   ind3d(iv) = a_idx_default; ind2d(iv) = a_idx_default
   ind3ds(iv) = a_idx_default; ind2ds(iv) = a_idx_default
   do igd = 1,n_grids
      ind3dn(igd,iv) = a_idx_default; ind2dn(igd,iv) = a_idx_default
      ind3dsn(igd,iv) = a_idx_default; ind2dsn(igd,iv) = a_idx_default
   end do
end do

maxnxyzp = maxnxp*maxnyp*maxnzp  ! Maximum # of points needed for any grid var
maxnxysp = (nzg+nzs)*maxnxp*maxnyp*npatch

! Find max number of points needed for any variable in this node's subdomain:

maxMxp = maxval(gd(:)%mmxp); maxMyp = maxval(gd(:)%mmyp)
maxMzp = maxval(gd(:)%mmzp)
maxMxyzp = maxMxp*maxMyp*maxMzp
maxMxysp = (nzg+nzs)*maxMxp*maxMyp*npatch

! Two scratch arrays are needed that are a maximum 3D variable long.
! These will be the first part of the A array:

memscr1 = max(maxnxyzp,maxnxysp)
iscr1 = 1; iscr2 = iscr1 + memscr1

! Assign grid-independent variable indices:

memval = iscr2+memscr1

do iv=1,nv3dmt
   if (ivopt3m(iv)(1:1) == '1') then
      ind3dm(ivtab3m(iv)) = memval
      memval = memval+maxMxyzp
   end if
end do
call vtable_memidx(0, .FALSE.)   ! Assign named indices

do igd=1,n_grids

   mmxyp = gd(igd)%mmxp*gd(igd)%mmyp
   mmxyzp = gd(igd)%mmxp*gd(igd)%mmyp*gd(igd)%mmzp
   mmxys2p = gd(igd)%mmxp*gd(igd)%mmyp*npatch
   mmxys3p = gd(igd)%mmxp*gd(igd)%mmyp*npatch*(nzg+nzs)

   do iv=1,nv2d(igd)
      if (ivopt2d(iv,igd)(1:1) == '1') then
         ind2dn(igd,ivtab2d(iv,igd)) = memval    ! 2D "normal"
         memval = memval+mmxyp
      end if
   end do

   do iv=1,nv2ds(igd)
      if (ivopt2ds(iv,igd)(1:1) == '1') then
         ind2dsn(igd,ivtab2ds(iv,igd)) = memval   ! 2D "sub/super surface"
         memval = memval+mmxys2p
      end if
   end do

   do iv=1,nv3ds(igd)
      if (ivopt3ds(iv,igd)(1:1) == '1') then
         ind3dsn(igd,ivtab3ds(iv,igd)) = memval   ! 3D "sub/super surface"
         memval = memval+mmxys3p
      end if
   end do

   do iv=1,nv3d(igd)
      if (ivopt3d(iv,igd)(1:1) == '1') then
         ind3dn(igd,ivtab3d(iv,igd)) = memval   ! 3D "normal"
         memval = memval+mmxyzp
      end if
   end do

   call vtable_memidx(igd, .FALSE.)    ! Assign named indices

end do

memsizeA = memval   ! Total size [words] of "A" array

! Possibly print information about variables, memory usage, etc.:

if (I_want_to_display) then

   write (*, '(/,/,78a)') '+', ('-', igd=1,77) 
   write (*, '(25x,a)') '"A" ARRAY:  MEMORY STATISTICS'

   call distill_bytes(int(memsizeA*f4_B, kind=i8), bsize, bunits)
   write (*, '(/,/,5x,a,f7.2,1x,a)') '***** Total memory to be allocated: ',  &
        bsize, trim(bunits)

end if

end subroutine leafPE_memsetA


end module setup_model_storage_mod
