!```````````````````````````````````````````````````````````````````
module state_IO_mod

!+

use univ_kind_defs_mod, only: f4

implicit none

integer :: n_vars_st, n_grids_st
integer, allocatable, dimension(:) :: nstzp, nstxp, nstyp   ! {n_grids_st}
integer, allocatable, dimension(:) :: nst3D, nst2D,  &  ! {n_grids_st}
     nst3Ds, nst2Ds, nst3Dm
character(len=8), allocatable, dimension(:,:) :: istchr3D, istchr2D,  &
     istchr3Ds, istchr2Ds, istchr3Dm    ! {n_vars_st,n_grids_st}

integer, parameter :: n_chars_cfgctl_version = 4
character(len=n_chars_cfgctl_version) :: cfgctl_version = '0003'

integer :: n_v_stfp
real(f4) :: sdist_curr, Ls_curr, UT_curr, EOT_curr, sdec_curr


contains


!```````````````````````````````````````````````````````````````````
subroutine cfgctl_rw(io_unit, action, calltype)

!+ Read/write model config and control variables (state/var file part) 

use univ_kind_defs_mod, only: f4, f8
use univ_exception_mod, only: exception, FATAL
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_math_constants_mod, only: ONE_f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_MESSAGE_LEN
use main_memory_mod, only: ztop, iref, jref, topref, PRAMS_version,  &
     n_vars, SBPS_Tlev, n_grids_v, n_substeps_max_curr, tstep,  &
     sfw_call, gd, n_Cn_records, start_sim_ET, zz, nstratz1,  &
     simulation_description, n_grids, nzg, nzs, naddsc, polelat, polelon,  &
     dzrat, dzmax, nestz1, maxnzp, initialization_type, impl, iadvl, iadvf,  &
     lsflg, ibnd, jbnd, icorflg, map_projection, n_dta_per_dtdyn,  &
     izflat, start_UT, npatch, radfrq, sspct, target_Ls, deltax_g1,  &
     deltay_g1, deltaz_g1, ntopsmth, dt_mode, n_substeps_max,  &
     dt_max_frac_increase_g1, current_sim_ET, preserve_dt_on_restart,  &
     min_ssfrac, max_dt_dynamics, maximum_sim_ET, igdref, rtgref,  &
     orbital_progression, diurnal_progression
use main_memory_bspec_mod, only: thdsnd, usnd, vsnd, tsnd, psnd, hsnd, rtsnd,  &
     slz, nsndp, slmstr, stgoff, H2Ov_tog, dust_tog, dust_nb, H2Oi_tog,  &
     H2Oi_nb, CO2i_tog, CO2i_nb, Ls_of_periapse, orbital_eccentricity,  &
     axial_obliquity, AU_to_m_knob, gd_bd, dust_op_case, RTc_tog, bdst_tog,  &
     bdst_imr, bdst_igd, bdst_icn, bdst_irt, bdst_iht, bdst_itt, dust_ctau_refp
use aermic_rad_mem_mod, only: aer_active, iaertog
use aermicro_mem_mod, only: nelements, ngroups, r, rup, rlow, rmass, rhoelem,  &
     nbins, groupname, elemname, max_nbins, ienconc

implicit none

!=== Argument declarations:

integer, intent(IN) :: io_unit
character(len=1), intent(IN) :: action
character(len=3), intent(IN) :: calltype  ! If ACTION='R': 'Q0S', 'Q1S', 'Q2S',
                                          !                'Q1V', or 'Q2V'

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'cfgctl_rw'

integer :: junk1_i0, junk2_i0, junk3_i0
real(f4) :: junk1_f0, junk2_f0, junk3_f0, junk4_f0, junk5_f0, dx, dy, dz,  &
     radfrq_st
integer, allocatable, dimension(:) :: junk1_i1, junk2_i1, junk3_i1
!%integer, allocatable, dimension(:,:) :: junk1_i2, junk2_i2, junk3_i2
real(f4), allocatable, dimension(:) :: junk1_f1, junk2_f1, junk3_f1
real(f4), allocatable, dimension(:,:) :: junk1_f2, junk2_f2 !% ,junk3_f2
logical, allocatable, dimension(:) :: junk1_b1
logical :: junk1_b0, junk2_b0
character(len=8) :: junk1_c8
character(len=n_chars_cfgctl_version) :: junk1_cv
character(len=MAX_MESSAGE_LEN) :: msg
real(f8) :: junk_f8

integer :: n, igd, ming, n1, n2, n3, a_err, maxg, igd_from
logical :: want_to_write, this_is_vfchk

!=== Executable statements:

if (action == 'W') then
   want_to_write = .TRUE.
else if (action == 'R') then
   want_to_write = .FALSE.
else
   return
end if

if (.NOT.want_to_write) then    ! Only check if ACTION='R'
   if (calltype(3:3) == 'V') then
      this_is_vfchk = .TRUE.
   else if (calltype(3:3) == 'S') then
      this_is_vfchk = .FALSE.
   else   
      call exception(procedure_name, FATAL,  &
           'Incorrect ''calltype'' arg value to `cfgctl_rw`.')
   end if
end if

if (want_to_write) then    ! Write info

   write (io_unit) PRAMS_version, simulation_description
   write (io_unit) cfgctl_version
   write (io_unit) n_grids, nzg, nzs, npatch, naddsc
   write (io_unit) current_sim_ET, start_sim_ET, maximum_sim_ET
   write (io_unit) target_Ls, start_UT, Ls_curr, UT_curr, EOT_curr,  &
        sdec_curr, sdist_curr, orbital_progression, diurnal_progression
   write (io_unit) n_vars, initialization_type, map_projection, polelat, polelon
   write (io_unit) gd(1)%deltax, gd(1)%deltay, gd(1)%deltaz, dzrat, dzmax
   write (io_unit) ztop, nestz1
   write (io_unit) H2Ov_tog, dust_tog, dust_nb, aer_active
   write (io_unit) H2Oi_tog, H2Oi_nb, CO2i_tog, CO2i_nb

   write (io_unit) gd(1:n_grids)%nxp, gd(1:n_grids)%nyp, gd(1:n_grids)%nzp
   write (io_unit) gd(1:n_grids)%nstratx, gd(1:n_grids)%nstraty,  &
        gd(1:n_grids)%nxtnest
   write (io_unit) gd(1:n_grids)%nsttop, gd(1:n_grids)%nstbot
   write (io_unit) gd(1:n_grids)%ninest, gd(1:n_grids)%njnest,  &
        gd(1:n_grids)%nknest
   write (io_unit) gd(1:n_grids)%idiffk, gd(1:n_grids)%gridu,  &
        gd(1:n_grids)%gridv
   write (io_unit) gd(1:n_grids)%centlat, gd(1:n_grids)%centlon
   write (io_unit) gd(1:n_grids)%plat, gd(1:n_grids)%plon
   write (io_unit) gd(1:n_grids)%deltax, gd(1:n_grids)%deltay,  &
        gd(1:n_grids)%deltaz
   write (io_unit) gd(1:n_grids)%dt_dyn

   write (io_unit) zz(1:maxnzp), nstratz1(1:maxnzp)
   
   write (io_unit) slz(1:nzg), slmstr(1:nzg), stgoff(1:nzg)
   write (io_unit) SBPS_Tlev(1:nzg+nzs)

   do igd=1,n_grids
      n1 = gd(igd)%nzp; n2 = gd(igd)%nxp; n3 = gd(igd)%nyp
      write (io_unit) gd(igd)%xm(1:n2), gd(igd)%xt(1:n2)
      write (io_unit) gd(igd)%ym(1:n3), gd(igd)%yt(1:n3)
      write (io_unit) gd(igd)%zm(1:n1), gd(igd)%zt(1:n1)
      write (io_unit) gd(igd)%dzm(1:n1), gd(igd)%dzt(1:n1)
      write (io_unit) gd(igd)%u01d(1:n1), gd(igd)%v01d(1:n1)
      write (io_unit) gd(igd)%pi01d(1:n1), gd(igd)%th01d(1:n1)
      write (io_unit) gd(igd)%dn01d(1:n1), gd_bd(igd)%rt01d(1:n1)
   end do

   write (io_unit) impl, iadvl, iadvf, sspct
   write (io_unit) ntopsmth, izflat
   write (io_unit) iref(:), jref(:), topref(:), igdref(:), rtgref(:)
   write (io_unit) ibnd, jbnd, lsflg, icorflg, bdst_tog, bdst_imr, bdst_igd
   write (io_unit) n_dta_per_dtdyn, radfrq, bdst_icn, bdst_irt, bdst_iht,  &
        bdst_itt, dust_ctau_refp

   write (io_unit) nsndp
   
   write (io_unit) hsnd(1:nsndp)
   write (io_unit) usnd(1:nsndp), vsnd(1:nsndp), psnd(1:nsndp)
   write (io_unit) tsnd(1:nsndp), thdsnd(1:nsndp), rtsnd(1:nsndp)

   write (io_unit) iaertog
   if (iaertog > 0) then
      write (io_unit) nelements, ngroups, max_nbins, RTc_tog, dust_op_case
      write (io_unit) nbins(1:ngroups)
      write (io_unit) groupname(1:ngroups)
      write (io_unit) ienconc(1:ngroups)
      write (io_unit) rhoelem(1:nelements)
      write (io_unit) elemname(1:nelements)
      write (io_unit) rmass(1:max_nbins,1:ngroups)
      write (io_unit) r(1:max_nbins,1:ngroups)
      write (io_unit) rup(1:max_nbins,1:ngroups)
      write (io_unit) rlow(1:max_nbins,1:ngroups)
   end if

   ! Timestep-related stuff (dt_dyn, sspct, n_dta_per_dtdyn, and radfrq have
   !  already been taken care of above):
   write (io_unit) dt_mode, n_substeps_max, dt_max_frac_increase_g1
   write (io_unit) max_dt_dynamics, min_ssfrac, n_substeps_max_curr, tstep
   write (io_unit) sfw_call
   write (io_unit) gd(1:n_grids)%n_dta_per_dtdyn, gd(1:n_grids)%n_dt_per_pgdt
   write (io_unit) gd(1:n_grids)%RT_call
   do igd=1,n_grids
      write (io_unit) gd(igd)%Cn_value_record(1:n_Cn_records),  &
           gd(igd)%Cn_change_record(1:n_Cn_records),  &
           gd(igd)%timestep_OK(1:n_Cn_records)
   end do

   ! "Orbital/astronomy"-related parameters:
   write (io_unit) Ls_of_periapse, orbital_eccentricity, axial_obliquity,  &
        AU_to_m_knob

else    ! Read info for state / varfile ingestion purposes

   if (calltype(1:2) == 'Q0') then    ! "Query" read for time info

      read (io_unit) junk1_c8   ! PRAMS_version, etc.
      read (io_unit) junk1_cv   ! cfgctl_version
      read (io_unit) junk1_i0   ! n_grids, etc.
      read (io_unit) start_sim_ET  ! Some other items after this one
      return    ! Done here

   else if (calltype(1:2) == 'Q1') then    ! "Query" read/check (first stage)

      read (io_unit) junk1_c8   ! PRAMS_version
      read (io_unit) junk1_cv   ! cfgctl_version

      if (junk1_cv /= cfgctl_version) then
         call exception(procedure_name,  FATAL,  &
          'State/varfile version is incompatible with running version.')
      end if

      if (.NOT.this_is_vfchk) then   ! Restart init call
         read (io_unit) n_grids_st, junk1_i0, junk2_i0
      else               ! Varfile init call
         read (io_unit) n_grids_v, junk1_i0, junk2_i0
         if (n_grids_v < n_grids) then
            write (*, '(/,a,/,a,i2,a)')  &
                 'ERROR: N_GRIDS in varfiles less than N_GRIDS in namelist.',  &
                 '       Remake varfiles with at least ', n_grids, ' grids.'
            call exception(procedure_name, FATAL)
         end if
      end if
      if (junk1_i0 /= nzg .or. junk2_i0 /= nzs) then
         call exception(procedure_name, FATAL,  &
              'NZG and/or NZS has changed.')
      end if

      read (io_unit) junk_f8   ! Skip (current_sim_ET ...) 

      if (.NOT.this_is_vfchk) then
         read (io_unit) target_Ls, start_UT, junk1_f0, junk2_f0, junk3_f0,  &
              junk4_f0, junk5_f0, junk1_b0, junk2_b0
         if (junk1_b0 .neqv. orbital_progression .or.  &
              junk2_b0 .neqv. diurnal_progression) then
            call exception(procedure_name, FATAL,  &
                 'ORBITAL_PROGRESSION and DIURNAL_PROGRESSION must not '//  &
                 'change upon a restart.')
         end if
      else
         read (io_unit) junk1_f0, junk2_f0
      end if

      read (io_unit) n_vars_st, junk1_i0, junk2_i0, junk1_f0, junk2_f0
      if (junk1_i0 /= initialization_type .or.  &
           junk2_i0 /= map_projection) then
         call exception(procedure_name, FATAL,  &
              'INITIALIZATION_TYPE and/or MAP_PROJECTION has changed.')
      end if
      if (abs(junk1_f0-polelat) > 0.0001_f4 .or.  &
           abs(junk2_f0-polelon) > 0.0001_f4) then
         call exception(procedure_name, FATAL,  &
              'POLELAT and/or POLELON has changed.')
      end if

      read (io_unit) junk1_f0, junk2_f0, junk3_f0, junk4_f0, junk5_f0
      if (.NOT.this_is_vfchk) then
         dx = deltax_g1; dy = deltay_g1; dz = deltaz_g1
      else
         dx = gd(1)%deltax; dy = gd(1)%deltay; dz = gd(1)%deltaz
      end if
      if (abs(junk1_f0-dx) > 0.01_f4 .or. abs(junk2_f0-dy) > 0.01_f4 .or.  &
           abs(junk3_f0-dz) > 0.01_f4) then
         call exception(procedure_name, FATAL,  &
              'DELTAX, DELTAY, and/or DELTAZ has changed.')
      end if
      if (abs(junk4_f0-dzrat) > 0.00001_f4 .or.  &
           abs(junk5_f0-dzmax) > 0.01_f4) then
         call exception(procedure_name, FATAL,  &
              'DZRAT and/or DZMAX has changed.')
      end if

      read (io_unit) junk1_f0  ! Skip this record (ztop ...)
      if (.NOT.this_is_vfchk) then     ! (H2Ov_tog ...) 
         read (io_unit) junk1_i0, junk2_i0, junk3_i0, aer_active
      else
         read (io_unit) junk1_i0, junk2_i0, junk3_i0
      end if
      read (io_unit) junk1_i0  ! Skip this record (H2Oi_tog ...)

   else if (calltype(1:2) == 'Q2') then      ! "Query" read/check (second stage)

      ming = min(n_grids_st, n_grids)
      maxg = max(n_grids_st, n_grids_v, n_grids)

      allocate(junk1_i1(maxg), junk1_f1(maxg), junk2_i1(maxg),  &
           junk2_f1(maxg), junk3_i1(maxg), junk3_f1(maxg), stat=a_err)
      if (a_err /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
           'cfgctl_rw01')

      if (.NOT.this_is_vfchk) then
         read (io_unit) nstxp(1:n_grids_st), nstyp(1:n_grids_st),  &
              nstzp(1:n_grids_st)
         junk1_i1(1:n_grids_st) = nstxp(1:n_grids_st)
         junk2_i1(1:n_grids_st) = nstyp(1:n_grids_st)
         junk3_i1(1:n_grids_st) = nstzp(1:n_grids_st)
      else  ! nxp(:) ...
         read (io_unit) junk1_i1(1:n_grids_v), junk2_i1(1:n_grids_v),  &
              junk3_i1(1:n_grids_v)
      end if
      do igd=1,ming
         if (junk1_i1(igd) /= gd(igd)%nxp .or.  &
              junk2_i1(igd) /= gd(igd)%nyp .or.  &
              junk3_i1(igd) /= gd(igd)%nzp) then
            write (*, '(/,a,i2)')   &
                 'ERROR: NXP, NYP, and/or NZP has changed on grid #', igd
            call exception(procedure_name, FATAL)
         end if
      end do

      if (.NOT.this_is_vfchk) then
         read (io_unit) junk1_i1(1:n_grids_st), junk2_i1(1:n_grids_st),  &
              junk3_i1(1:n_grids_st)
      else   ! nstratx(:) ...
         read (io_unit) junk1_i1(1:n_grids_v), junk2_i1(1:n_grids_v),  &
              junk3_i1(1:n_grids_v)
      end if
      do igd=1,ming
         if (junk1_i1(igd) /= gd(igd)%nstratx .or.  &
              junk2_i1(igd) /= gd(igd)%nstraty .or.  &
              junk3_i1(igd) /= gd(igd)%nxtnest) then
            write (*, '(/,a,i2)')   &
            'ERROR: NSTRATX, NSTRATY, and/or NXTNEST has changed on grid #', igd
            call exception(procedure_name, FATAL)
         end if
      end do

      if (.NOT.this_is_vfchk) then
         read (io_unit) junk1_i1(1:n_grids_st), junk2_i1(1:n_grids_st)
      else  ! nsttop(:) ...
         read (io_unit) junk1_i1(1:n_grids_v), junk2_i1(1:n_grids_v)
      end if
      do igd=1,ming
         if (junk1_i1(igd) /= gd(igd)%nsttop .or.  &
              junk2_i1(igd) /= gd(igd)%nstbot) then
            write (*, '(/,a,i2)')   &
            'ERROR: NSTTOP and/or NSTBOT has changed on grid #', igd
            call exception(procedure_name, FATAL)
         end if
      end do

      if (.NOT.this_is_vfchk) then
         read (io_unit) junk1_i1(1:n_grids_st), junk2_i1(1:n_grids_st),  &
              junk3_i1(1:n_grids_st)
      else   ! ninest(:) ...
         read (io_unit) junk1_i1(1:n_grids_v), junk2_i1(1:n_grids_v),  &
              junk3_i1(1:n_grids_v)
      end if
      do igd=1,ming
         if(junk1_i1(igd) /= gd(igd)%ninest .or.  &
              junk2_i1(igd) /= gd(igd)%njnest .or.  &
              junk3_i1(igd) /= gd(igd)%nknest) then
            write (*, '(/,a,i2)')   &
               'ERROR: NINEST, NJNEST, and/or NKNEST has changed on grid #', igd
            call exception(procedure_name, FATAL)
         end if
      end do

      if (.NOT.this_is_vfchk) then
         read (io_unit) junk1_i1(1:n_grids_st), junk1_f1(1:n_grids_st),  &
              junk2_f1(1:n_grids_st)
      else  ! idiffk(:) ...
         read (io_unit) junk1_i1(1:n_grids_v), junk1_f1(1:n_grids_v),  &
              junk2_f1(1:n_grids_v)
      end if
      if (.NOT.this_is_vfchk) then
         do igd=1,ming
            if (junk1_i1(igd) /= gd(igd)%idiffk .or.  &
                 abs(junk1_f1(igd)-gd(igd)%gridu) > 0.01_f4 .or.  &
                 abs(junk2_f1(igd)-gd(igd)%gridv) > 0.01_f4) then
               write (*, '(/,a,i2)')   &
                 'ERROR: IDIFFK, GRIDU, and/or GRIDV has changed on grid #', igd
               call exception(procedure_name, FATAL)
            end if
         end do
      end if

      if (.NOT.this_is_vfchk) then
         read (io_unit) junk1_f1(1:n_grids_st), junk2_f1(1:n_grids_st)
      else   ! centlat(:) ...
         read (io_unit) junk1_f1(1:n_grids_v), junk2_f1(1:n_grids_v)
      end if
      do igd=1,ming
         if (abs(junk1_f1(igd)-gd(igd)%centlat) > 0.0001_f4 .or.  &
              abs(junk2_f1(igd)-gd(igd)%centlon) > 0.0001_f4) then
            write (*, '(/,a,i2)')   &
                 'ERROR: CENTLAT and/or CENTLON has changed on grid #', igd
            call exception(procedure_name, FATAL)
         end if
      end do

      read (io_unit) junk1_f0   ! Skip this record (plat(:) ...)
      read (io_unit) junk1_f0   ! Skip this record (deltax(:) ...)

      read (io_unit) junk1_f1(1:n_grids_st)   ! dt_dyn(:)
      if (.NOT.this_is_vfchk) then
         gd(1:ming)%dt_dyn = junk1_f1(1:ming)
         do igd=ming+1,n_grids  ! Use valid parent values for any added grids
            igd_from = gd(igd)%nxtnest
            do while (igd_from > ming)         ! If the parent grid is itself an
               igd_from = gd(igd_from)%nxtnest ! added grid, use its parent
            end do
            gd(igd)%dt_dyn = gd(igd_from)%dt_dyn
         end do
      end if

      read (io_unit) junk1_f0   ! Skip this record (zz(:) ...)

      deallocate(junk1_i1, junk1_f1, junk2_i1, junk2_f1, junk3_i1, junk3_f1,  &
           stat=a_err)
      if (a_err /= 0) call illegal_alloc_stop(for_deallocation,  &
           procedure_name, 'cfgctl_rw01')
   
      allocate(junk1_f1(nzg), junk2_f1(nzg), junk3_f1(nzg), stat=a_err)
      if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
           procedure_name, 'cfgctl_rw02')

      read (io_unit) junk1_f1(1:nzg)  ! slz(:) ...
      do n=1,nzg
         if (abs(junk1_f1(n)-slz(n)) > 0.0001_f4) then
            call exception(procedure_name, FATAL,  &
                 'ERROR: SLZ has changed.')
         end if
      end do

      if (.NOT.this_is_vfchk) then

         read (io_unit) SBPS_Tlev(1:nzg+nzs)

         do igd=1,n_grids_st
            allocate(junk1_f2(nstzp(igd),n_grids_st),  &
                 junk2_f2(nstzp(igd),n_grids_st), stat=a_err)
            if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
                 procedure_name, 'cfgctl_rw03')
            
            read (io_unit) junk1_f0    ! Skip this record (xm(:) ...)
            read (io_unit) junk1_f0    ! Skip this record (ym(:) ...)
            read (io_unit) junk1_f0    ! Skip this record (zm(:) ...)
            read (io_unit) junk1_f0    ! Skip this record (dzm(:) ...)
            
            read (io_unit) junk1_f2(1:nstzp(igd),igd),  &
                 junk2_f2(1:nstzp(igd),igd)
            if (igd <= ming) then
               gd(igd)%u01d(1:nstzp(igd)) = junk1_f2(1:nstzp(igd),igd)
               gd(igd)%v01d(1:nstzp(igd)) = junk2_f2(1:nstzp(igd),igd)
            end if
            
            read (io_unit) junk1_f2(1:nstzp(igd),igd),  &
                 junk2_f2(1:nstzp(igd),igd)
            if (igd <= ming) then
               gd(igd)%pi01d(1:nstzp(igd)) = junk1_f2(1:nstzp(igd),igd)
               gd(igd)%th01d(1:nstzp(igd)) = junk2_f2(1:nstzp(igd),igd)
            end if
            
            read (io_unit) junk1_f2(1:nstzp(igd),igd),  &
                 junk2_f2(1:nstzp(igd),igd)
            if (igd <= ming) then
               gd(igd)%dn01d(1:nstzp(igd)) = junk1_f2(1:nstzp(igd),igd)
               gd_bd(igd)%rt01d(1:nstzp(igd)) = junk2_f2(1:nstzp(igd),igd)
            end if
            
            deallocate(junk1_f2, junk2_f2, stat=a_err)
            if (a_err /= 0) call illegal_alloc_stop(for_deallocation,  &
                 procedure_name, 'cfgctl_rw03')
         end do
         
         read (io_unit) junk1_i0    ! Skip this record (impl ...)
         read (io_unit) junk1_i0    ! Skip this record (ntopsmth ...)
         read (io_unit) iref(:), jref(:), topref(:), igdref(:), rtgref(:)
         read (io_unit) junk1_i0    ! Skip this record (ibnd ...) 
         read (io_unit) junk1_i0, radfrq_st  ! n_dta_per_dtdyn ...
         
         read (io_unit) junk1_i0    ! Skip this record (nsndp ...)
         
         read (io_unit) junk1_f0    ! Skip this record (hsnd(:) ...)
         read (io_unit) junk1_f0    ! Skip this record (usnd(:) ...)
         read (io_unit) junk1_f0    ! Skip this record (tsnd(:) ...)

         read (io_unit) junk1_i0  ! iaertog
         if (junk1_i0 > 0) then
            read (io_unit) junk1_i0   ! Skip this record (nelements ...)
            read (io_unit) junk1_i0   ! Skip this record (nbins(:) ...)
            read (io_unit) junk1_c8   ! Skip this record (groupname(:) ...)
            read (io_unit) junk1_i0   ! Skip this record (ienconc(:) ...)
            read (io_unit) junk1_f0   ! Skip this record (rhoelem(:) ...)
            read (io_unit) junk1_c8   ! Skip this record (elename(:) ...)
            read (io_unit) junk1_f0   ! Skip this record (rmass(:) ...)
            read (io_unit) junk1_f0   ! Skip this record (r(:) ...)
            read (io_unit) junk1_f0   ! Skip this record (rup(:) ...)
            read (io_unit) junk1_f0   ! Skip this record (rlow(:) ...)
         end if

         deallocate(junk1_f1, junk2_f1, junk3_f1, stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'cfgctl_rw02')

         allocate(junk1_i1(maxg), junk2_i1(maxg), stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
              procedure_name, 'cfgctl_rw04')

         read (io_unit) junk1_i0, junk2_i0, junk1_f0  ! dt_mode ...
         if (dt_mode < 0 .and. preserve_dt_on_restart) then
            if (junk1_i0 /= dt_mode) then
               write (msg, '(2a,i3,a1)') 'PRESERVE_DT_ON_RESTART = .TRUE., ',  &
                    'but namelist DT_MODE /= to state file DT_MODE (',  &
                    junk1_i0, ')'
               call exception(procedure_name, FATAL, msg)
            end if
            if (abs(radfrq-radfrq_st) > 2._f4*epsilon(ONE_f4)) then
               write (msg, '(2a,e15.6,a1)')  &
                    'Namelist RADFRQ /= to state file RADFRQ (',  &
                    radfrq_st, ')'
               call exception(procedure_name, FATAL, msg)
            end if
            if (junk2_i0 /= n_substeps_max) then
               write (msg, '(2a,i9,a1)')  &
               'Namelist N_SUBSTEPS_MAX /= to state file N_SUBSTEPS_MAX (',  &
                    n_substeps_max, ')'
               call exception(procedure_name, FATAL, msg)
            end if
         end if

          ! max_dt_dynamics ...
         read (io_unit) junk1_f0, junk2_f0, junk1_i0, junk2_i0
         if (dt_mode < 0 .and. preserve_dt_on_restart) then
            n_substeps_max_curr = junk1_i0
            min_ssfrac = junk2_f0
            max_dt_dynamics = junk1_f0
         end if

         read (io_unit) sfw_call   ! sfw_call ...

         ! n_dta_per_dtdyn(:) ...
         read (io_unit) junk1_i1(1:n_grids_st), junk2_i1(1:n_grids_st)
            ! Must always fill gd(:)%n_dta_per_dtdyn here so that
            !  `switch_to_grid` works for model restarts (may be overwritten
            !  soon after in `dtset`):
         gd(1:ming)%n_dta_per_dtdyn = junk1_i1(1:ming)
         do igd=ming+1,n_grids   ! Use valid parent values for any added grids
            igd_from = gd(igd)%nxtnest
            do while (igd_from > ming)         ! If the parent grid is itself an
               igd_from = gd(igd_from)%nxtnest ! added grid, use its parent
            end do
            gd(igd)%n_dta_per_dtdyn = gd(igd_from)%n_dta_per_dtdyn
         end do
         if (dt_mode < 0 .and. preserve_dt_on_restart) then
            gd(1:ming)%n_dt_per_pgdt = junk2_i1(1:ming)
            do igd=ming+1,n_grids ! Use valid parent values for any added grids
               igd_from = gd(igd)%nxtnest
               do while (igd_from > ming)      ! If the parent grid is itself an
                  igd_from = gd(igd_from)%nxtnest ! added grid, use its parent
               end do
               gd(igd)%n_dt_per_pgdt = gd(igd_from)%n_dt_per_pgdt
            end do
         end if

         ! RT_call(:) ...
         read (io_unit) junk1_i1(1:n_grids_st)
         gd(1:ming)%RT_call = junk1_i1(1:ming)
         do igd=ming+1,n_grids ! Use valid parent values for any added grids
            igd_from = gd(igd)%nxtnest
            do while (igd_from > ming)         ! If the parent grid is itself an
               igd_from = gd(igd_from)%nxtnest ! added grid, use its parent
            end do
            gd(igd)%RT_call = gd(igd_from)%RT_call
         end do

         allocate(junk1_f1(n_Cn_records), junk2_f1(n_Cn_records),  &
              junk1_b1(n_Cn_records), stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(for_allocation,  &
              procedure_name, 'cfgctl_rw05')

         ! Cn_value_record(:) ...
         do igd=1,n_grids_st
            read (io_unit) junk1_f1(1:n_Cn_records),  &
                 junk2_f1(1:n_Cn_records), junk1_b1(1:n_Cn_records)
            if (dt_mode < 0 .and. preserve_dt_on_restart) then
               if (igd <= ming) then
                  gd(igd)%Cn_value_record(1:n_Cn_records) =  &
                       junk1_f1(1:n_Cn_records)
                  gd(igd)%Cn_change_record(1:n_Cn_records) =  &
                       junk2_f1(1:n_Cn_records)
                  gd(igd)%timestep_OK(1:n_Cn_records) = junk1_b1(1:n_Cn_records)
               end if
            end if
         end do
         do igd=ming+1,n_grids ! Use valid parent values for any added grids
            igd_from = gd(igd)%nxtnest
            do while (igd_from > ming)         ! If the parent grid is itself an
               igd_from = gd(igd_from)%nxtnest ! added grid, use its parent
            end do
            gd(igd)%Cn_value_record(1:n_Cn_records) =  &
                 gd(igd_from)%Cn_value_record(1:n_Cn_records)
            gd(igd)%Cn_change_record(1:n_Cn_records) =  &
                 gd(igd_from)%Cn_change_record(1:n_Cn_records)
            gd(igd)%timestep_OK(1:n_Cn_records) =   &
                 gd(igd_from)%timestep_OK(1:n_Cn_records)
         end do

         ! Ls_of_periapse, orbital_eccentricity, axial_obliquity, AU_to_m_knob
         read (io_unit)  junk1_f0, junk2_f0, junk3_f0, junk4_f0
         if (abs(junk1_f0-Ls_of_periapse) > 1.e-5_f4)  &
              call exception(procedure_name, FATAL,  &
              'LS_OF_PERIAPSE cannot change on model restart.')
         if (abs(junk2_f0-orbital_eccentricity) > 1.e-5_f4)  &
              call exception(procedure_name, FATAL,  &
              'ORBITAL_ECCENTRICITY cannot change on model restart.')
         if (abs(junk3_f0-axial_obliquity) > 1.e-5_f4)  &
              call exception(procedure_name, FATAL,  &
              'AXIAL_OBLIQUITY cannot change on model restart.')

         deallocate(junk1_i1, junk2_i1, stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'cfgctl_rw04')

         deallocate(junk1_f1, junk2_f1, junk1_b1, stat=a_err)
         if (a_err /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'cfgctl_rw05')

      end if

   endif

endif

end subroutine cfgctl_rw


!```````````````````````````````````````````````````````````````````
subroutine gen_state_file_name(fname, prefix, type, post, fmtstr, curr_sim_ET)

!+ Create standard, timestamped, state file name

use univ_kind_defs_mod, only: f8, i8
use univ_integer_time_mod, only: time_obj, fp_seconds_eqv_Time, operator(-),  &
     get_Time_component_values, operator(*), operator(/)
use main_memory_mod, only: module_tps, module_tps_log10, mnsdy_Time

implicit none

!=== Argument declarations:

real(f8), intent(IN) :: curr_sim_ET
character(len=*), intent(IN) :: prefix, post
character(len=3), intent(IN) :: fmtstr
character(len=1), intent(IN) :: type

character(len=*), intent(OUT) :: fname

!== Local declarations:

integer :: ic, t_body_days
integer(i8) :: t_seconds, t_ticks
character(len=40) :: time_strrep, dstring
type(time_obj) :: tmp_Time1, tmp_Time2

!=== Executable statements:

tmp_Time1 = fp_seconds_eqv_Time(curr_sim_ET, module_tps)
t_body_days = tmp_Time1/mnsdy_Time
tmp_Time2 = tmp_Time1-(t_body_days*mnsdy_Time)
call get_time_component_values(tmp_Time2, t_seconds, t_ticks)
write (time_strrep, '(i4.4,a1,i8.8,a1,i4.4)') t_body_days, '_', t_seconds,  &
     '.', t_ticks/10**(module_tps_log10-4)

dstring = '-'//type//'-'//trim(time_strrep)
ic = len_trim(dstring)

fname = prefix(1:len_trim(prefix))//dstring(1:ic)
if (len_trim(post) > 0) then
   fname = fname(1:len_trim(fname))//'-'//post(1:len_trim(post))
end if

fname = fname(1:len_trim(fname))//'.'//fmtstr(1:3)    ! Final filename

end subroutine gen_state_file_name


end module state_IO_mod


!```````````````````````````````````````````````````````````````````
subroutine state_file_w(use_alt_filename_sentinel)

!+ Writes state file(s)

use ISO_C_binding, only: C_F_pointer, C_loc
use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_NAME_LEN,  &
     MAX_FILE_PATH_LEN
use univ_std_fileIO_mod, only: find_file_IOunit, determine_IO_reclen
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_integer_time_mod, only: time_obj, fp_seconds_eqv_Time
use univ_namelist_utils_mod, only: Time_to_gen_time_strrep
use main_memory_mod, only: a, n_vars, ioutput, n_grids, npatch,  &
     state_output_fpfx, current_sim_ET, gd, nzg, nzs, module_tps, mnsdy_Time,  &
     morbpd_Time
use vtables_mod, only: n_state_3D, ind3Dn, i_state_tab3D, n_state_2D, ind2Dn,  &
     i_state_tab2D, n_state_3Ds, ind3Dsn, i_state_tab3Ds, n_state_2Ds,  &
     ind2Dsn, i_state_tab2Ds, n_state_3Dm, i_state_chr3D, i_state_chr2D,  &
     i_state_chr3Ds, i_state_chr2Ds, i_state_chr3Dm
use state_IO_mod, only: gen_state_file_name, cfgctl_rw, n_v_stfp
use state_header_for_postp_mod, only: state_table_for_postp

implicit none

!=== Argument declarations:

logical, intent(IN) :: use_alt_filename_sentinel

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'state_file_w'

integer :: IO_unit, igd, iq, e, ib, ie, recn, recsizew, k, ip, nvcnt, IO_reclen
character(len=MAX_FILE_NAME_LEN) :: gen_time_strrep 
character(len=MAX_FILE_PATH_LEN) :: sfn, sfn_head
character(len=3) :: cgrid
real(f4), allocatable, dimension(:,:) :: reclen_test
type(time_obj) :: tmp_Time
real(f4), pointer, dimension(:) :: ptr_1D => NULL()

!=== Executable statements:

IO_unit = find_file_IOunit()
  
if (ioutput == 0) return   ! User does not want to write state file(s)

!=== Initialization for state file writes:

if (.NOT.associated(state_table_for_postp)) then
   ! If this has not been done yet in this run:
   !   * Find total number of fields to be written
   !   * Allocate needed book-keeping storage
   !   * Populate book-keeping table
   n_v_stfp = 0
   do igd=1,n_grids
      n_v_stfp = n_v_stfp+n_state_3D(igd)+n_state_2D(igd)+n_state_3Ds(igd)+  &
           n_state_2Ds(igd)+2
   end do
   !%      n_v_stfp = n_v_stfp+n_state_3Dm   ! Add # of grid-independent vars
   allocate(state_table_for_postp(n_v_stfp), stat=e)
   if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

   nvcnt = 0
   do igd=1,n_grids
      recn = 1
      
      ! Loop through the 3D variables:
      do iq=1,n_state_3D(igd)
         nvcnt = nvcnt+1
         state_table_for_postp(nvcnt)%beg_rec = recn
         recn = recn+gd(igd)%nzp; state_table_for_postp(nvcnt)%end_rec = recn-1
         state_table_for_postp(nvcnt)%var_type = 3
         state_table_for_postp(nvcnt)%grid = igd
         state_table_for_postp(nvcnt)%var_name = i_state_chr3D(iq,igd)
      end do

      ! Loop through the 2D variables:
      do iq=1,n_state_2D(igd)
         nvcnt = nvcnt+1
         state_table_for_postp(nvcnt)%beg_rec = recn
         recn = recn+1; state_table_for_postp(nvcnt)%end_rec = recn-1
         state_table_for_postp(nvcnt)%var_type = 2
         state_table_for_postp(nvcnt)%grid = igd
         state_table_for_postp(nvcnt)%var_name = i_state_chr2D(iq,igd)
      end do

      ! 2D non-a-array variables:

      nvcnt = nvcnt+1
      state_table_for_postp(nvcnt)%beg_rec = recn
      recn = recn+1; state_table_for_postp(nvcnt)%end_rec = recn-1
      state_table_for_postp(nvcnt)%var_type = 2
      state_table_for_postp(nvcnt)%grid = igd
      state_table_for_postp(nvcnt)%var_name = 'EMISST'

      ! Loop through the 3D SBPS variables:
      do iq=1,n_state_3Ds(igd)
         nvcnt = nvcnt+1
         state_table_for_postp(nvcnt)%beg_rec = recn
         recn = recn+(nzg+nzs)*npatch; state_table_for_postp(nvcnt)%end_rec = recn-1
         state_table_for_postp(nvcnt)%var_type = 5
         state_table_for_postp(nvcnt)%grid = igd
         state_table_for_postp(nvcnt)%var_name = i_state_chr3Ds(iq,igd)
      end do

      ! Loop through the 2D SBPS variables:
      do iq=1,n_state_2Ds(igd)
         nvcnt = nvcnt+1
         state_table_for_postp(nvcnt)%beg_rec = recn
         recn = recn+1*npatch; state_table_for_postp(nvcnt)%end_rec = recn-1
         state_table_for_postp(nvcnt)%var_type = 4
         state_table_for_postp(nvcnt)%grid = igd
         state_table_for_postp(nvcnt)%var_name = i_state_chr2Ds(iq,igd)
      end do

      ! 2D SBPS non-a-array variables:

      nvcnt = nvcnt+1
      state_table_for_postp(nvcnt)%beg_rec = recn
      recn = recn+1*npatch; state_table_for_postp(nvcnt)%end_rec = recn-1
      state_table_for_postp(nvcnt)%var_type = 4
      state_table_for_postp(nvcnt)%grid = igd
      state_table_for_postp(nvcnt)%var_name = 'EMISSTAT'

   end do
end if

!=== Write state header file (one per output time):

! Create state header file name:
if (use_alt_filename_sentinel) then
   call gen_state_file_name(sfn_head, state_output_fpfx, 'C', 'hdr',  &
        'vfm', current_sim_ET)
else
   call gen_state_file_name(sfn_head, state_output_fpfx, 'S', 'hdr',  &
        'vfm', current_sim_ET)
end if

! Open new header file:
open(IO_unit, file=sfn_head(1:len_trim(sfn_head)), form='unformatted',  &
     status='replace')

! Write information that is needed by postp:
write (IO_unit) n_v_stfp, n_grids     ! Write some dimensions first
write (IO_unit) state_table_for_postp(1:n_v_stfp)  ! Write state info table

! Write cfg/ctl info:
call cfgctl_rw(IO_unit, 'W', 'S  ')

! Write restart vartable info to header file:
write (IO_unit) n_state_3D(1:n_grids), n_state_2D(1:n_grids),  &
     n_state_3Ds(1:n_grids), n_state_2Ds(1:n_grids), n_state_3Dm(1:n_grids)
write (IO_unit) i_state_chr3D(1:n_vars,1:n_grids),  &
     i_state_chr2D(1:n_vars,1:n_grids), i_state_chr3Ds(1:n_vars,1:n_grids),  &
     i_state_chr2Ds(1:n_vars,1:n_grids), i_state_chr3Dm(1:n_vars,1:n_grids)

! Close header file:
close(IO_unit)

!=== Write state data files (one per output time and grid):

do igd=1,n_grids   ! Loop through each nest
   if (.not. gd(igd)%grid_in_state_outp) cycle ! Skip state output for this grid

   write (cgrid, '(a1,i2.2)') 'g', igd   ! Create grid ID string
   
   ! Create state file name (avoid overwriting input file when restarting):
   if (use_alt_filename_sentinel) then
      call gen_state_file_name(sfn, state_output_fpfx, 'C', cgrid, 'vfm',  &
           current_sim_ET)
   else
      call gen_state_file_name(sfn, state_output_fpfx, 'S', cgrid, 'vfm',  &
           current_sim_ET)
   end if

   ! Open new state file (direct-access binary); write variables to it:

   recsizew = gd(igd)%nxyp

   allocate(reclen_test(recsizew,1), stat=e)
   if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'reclen')
   IO_reclen = determine_IO_reclen(reclen_test)
   deallocate(reclen_test, stat=e)
   if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'reclen')

   open(IO_unit, file=sfn(1:len_trim(sfn)), form='unformatted',  &
        status='replace', access='direct', recl=IO_reclen)

   recn = 1

   ! Loop through and write the 3D variables:
   do iq=1,n_state_3D(igd)
      ib = ind3Dn(igd,i_state_tab3D(iq,igd))
      do k=1,gd(igd)%nzp
         ie = ib+recsizew-1
         write (IO_unit, rec=recn) a(ib:ie)
         recn = recn+1
         ib = ie+1
      end do
   end do

   ! Loop through and write the 2D variables:
   do iq=1,n_state_2D(igd)
      ib = ind2Dn(igd,i_state_tab2D(iq,igd))
      ie = ib+recsizew-1
      write (IO_unit, rec=recn) a(ib:ie)
      recn = recn+1
   end do

   ! 2D non-a-array variables:

   call C_F_pointer(C_loc(gd(igd)%emisst), ptr_1D, (/gd(igd)%nxyp/))
   write (IO_unit, rec=recn) ptr_1D
   recn = recn+1

   ! Loop through and write the 3D SBPS variables:
   do iq=1,n_state_3Ds(igd)
      ib = ind3Dsn(igd,i_state_tab3Ds(iq,igd))
      do ip=1,npatch
         do k=1,nzg+nzs
            ie = ib+recsizew-1
            write (IO_unit, rec=recn) a(ib:ie)
            recn = recn+1
            ib = ie+1
         end do
      end do
   end do

   ! Loop through and write the 2D SBPS variables:
   do iq=1,n_state_2Ds(igd)
      ib = ind2Dsn(igd,i_state_tab2Ds(iq,igd))
      do ip=1,npatch
         ie = ib+recsizew-1
         write (IO_unit, rec=recn) a(ib:ie)
         recn = recn+1
         ib = ie+1
      end do
   end do

   ! 2D SBPS non-a-array variables:

   do ip=1,npatch
      call C_F_pointer(C_loc(gd(igd)%emissivity_static(:,:,ip)), ptr_1D,  &
           (/gd(igd)%nxyp/))
      write (IO_unit, rec=recn) ptr_1D
      recn = recn+1
   end do

   close(IO_unit)  ! Close state file

end do

tmp_Time = fp_seconds_eqv_Time(current_sim_ET, module_tps)
call Time_to_gen_time_strrep(tmp_Time, gen_time_strrep,  &
     len(gen_time_strrep), mnsdy_Time, morbpd_Time, short_unit_style=.TRUE.)

write (*, '(/,3a,/,2a)') 'State write at time = [',  &
     trim(gen_time_strrep), ']', '  Filename sample:  ',  &
     sfn(1:len_trim(sfn)) 

end subroutine state_file_w


!```````````````````````````````````````````````````````````````````
subroutine state_output(terminate_run_now, just_restarted)

!+ PRAMS file output driver

use univ_kind_defs_mod, only: f8, f4
use univ_math_constants_mod, only: ONE_f4
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use main_memory_mod, only: ioutput, n_grids, wfreq_state_file,  &
     initialization_type, a, jdim, vtime, n_varf, sfw_call, current_sim_ET,  &
     maximum_sim_ET, gd, c_gd, i_grid
use main_memory_bspec_mod, only: H2Ov_tog, lonrad
use vtables_mod, only: ithp, itheta, irv, iglat, iglon, ivt2da
use aermic_rad_mem_mod, only: iaertog
use grid_mem_mod, only: switch_to_grid
use PRAMS_thermo_mod, only: dry_thermo
use state_IO_mod, only: sdist_curr, Ls_curr, UT_curr, EOT_curr, sdec_curr
use RT_utils_mod, only: stellar_angles

implicit none

!=== Argument declarations:

logical, intent(IN) :: terminate_run_now, just_restarted

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'state_output'

integer :: igd, irv_tmp, ic
logical :: end_of_run_reached

!=== Executable statements:

end_of_run_reached = ( current_sim_ET >= maximum_sim_ET-0.01_f4*gd(1)%dt_dyn )

! Write state file(s)?

if (wfreq_state_file > tiny(ONE_f4)) then
   ic = int(current_sim_ET/real(wfreq_state_file, kind=f8))
else   ! Every grid #1 timestep
   ic = sfw_call+1
end if
if (ic > sfw_call .or. terminate_run_now .or. end_of_run_reached .or.  &
     just_restarted) then
   if (ic > sfw_call) sfw_call = ic

   if (ioutput /= 0) then

      do igd=1,n_grids
         call switch_to_grid(igd)

         if (igd == 1) then
            call stellar_angles(c_gd%nzp, c_gd%nxp, c_gd%nyp, 1, 1, 1, 1,  &
                 a(iglat), a(iglon), a(ivt2da), sdist_curr, c_gd%centlat,  &
                 c_gd%centlon, lonrad, a(ivt2da), Ls_curr, .FALSE., UT_curr,  &
                 EOT_curr, sdec_curr)
         end if

         ! Get rid of boundary gradients in the diagnosed variables (THETA,RV):

         if (H2Ov_tog > 0) then
            irv_tmp = irv
         else
            irv_tmp = 1   ! Arbitrary, but valid index (unused array)
         end if

         call dry_thermo(c_gd%nzp, c_gd%nxp, c_gd%nyp, 1, 1, 1, c_gd%nyp,  &
              a(ithp), a(itheta), a(irv_tmp), iaertog)
         call dry_thermo(c_gd%nzp, c_gd%nxp, c_gd%nyp, c_gd%nxp, c_gd%nxp, 1,  &
              c_gd%nyp, a(ithp), a(itheta), a(irv_tmp), iaertog)
         if (jdim == 1) then
            call dry_thermo(c_gd%nzp, c_gd%nxp, c_gd%nyp, 1, c_gd%nxp, 1, 1,  &
                 a(ithp), a(itheta), a(irv_tmp), iaertog)
            call dry_thermo(c_gd%nzp, c_gd%nxp, c_gd%nyp, 1, c_gd%nxp,  &
                 c_gd%nyp, c_gd%nyp, a(ithp), a(itheta), a(irv_tmp), iaertog)
         end if

      end do

   end if

   call state_file_w(just_restarted)
end if

! Time to switch varfiles?

if (initialization_type == 2) then
   if (real(current_sim_ET, kind=f4) >= vtime(n_varf+1) .and.  &
        current_sim_ET+0.0001_f8 < maximum_sim_ET) then
   
      call switch_to_grid(1)
   
      n_varf = n_varf+1
   
      write (*, '(/,a,i2,a)') 'Switching varfiles on grid #', i_grid, ':'
      call initv(i_grid, .FALSE., .FALSE.)   ! Switch varfile
      do igd=2,n_grids
         call switch_to_grid(igd)   
         write (*, '(/,a,i2,a)') 'Switching varfiles on grid #', i_grid, ':'
         call initv(igd, .FALSE., .FALSE.)    ! Switch varfile
         call vfintrp(igd)
      end do
      write (*, '(/)')

   end if
end if

end subroutine state_output


!```````````````````````````````````````````````````````````````````
subroutine restart

!+ Do a restart (initialize model from state file data)

use univ_kind_defs_mod, only: f4
use univ_std_fileIO_mod, only: find_file_IOunit, determine_IO_reclen
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN
use main_memory_mod, only: a, gd, nzg, nzs, npatch, n_grids,  &
     input_state_file_pfx
use vtables_mod, only: nv3d, ivchr3d, ind3dn, ivtab3d, nv2d, ivchr2d, ind2dn,  &
     ivtab2d, nv3ds, ivchr3ds, ind3dsn, ivtab3ds, nv2ds, ivchr2ds, ind2dsn,  &
     ivtab2ds
use state_IO_mod, only: nstzp, nstxp, nstyp, cfgctl_rw, nst3D, nst2D, nst3Ds,  &
     nst2Ds, nst3Dm, istchr3D, istchr2D, istchr3Ds, istchr2Ds, istchr3Dm,  &
     n_grids_st, n_vars_st

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'restart'

character(len=MAX_FILE_PATH_LEN):: sfn_data, sfn_head
integer :: nc, in_unit, igd, recn, ib, ie, iv, iq, e, recsize_f, k, ip,  &
     IO_reclen
real(f4), allocatable, dimension(:,:) :: reclen_test
character(len=4) :: cgrid

!=== Executable statements:

in_unit = find_file_IOunit()

nc = len_trim(input_state_file_pfx)
sfn_head = input_state_file_pfx(1:nc)//'-hdr.vfm' ! State header file name

! Open state header file:
open(in_unit, file=sfn_head(1:len_trim(sfn_head)), form='unformatted',  &
     status='old')

read (in_unit) iq; read (in_unit) iq  ! Skip two postp-targeted records

! Read the first bit of "common" info; check for compatibility:
call cfgctl_rw(in_unit, 'R', 'Q1S')

! Allocate info arrays using number of grids specified in state file:

allocate(nstzp(n_grids_st), nstxp(n_grids_st), nstyp(n_grids_st), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

allocate(nst3D(n_grids_st), nst2D(n_grids_st), nst3Ds(n_grids_st),  &
     nst2Ds(n_grids_st), nst3Dm(n_grids_st), istchr3D(n_vars_st,n_grids_st),  &
     istchr2D(n_vars_st,n_grids_st), istchr3Ds(n_vars_st,n_grids_st),  &
     istchr2Ds(n_vars_st,n_grids_st), istchr3Dm(n_vars_st,n_grids_st), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '2')

! Read the remainder of the "common" info; check for compatibility:

call cfgctl_rw(in_unit, 'R', 'Q2S')

read (in_unit) nst3D(1:n_grids_st), nst2D(1:n_grids_st),  &
     nst3Ds(1:n_grids_st), nst2Ds(1:n_grids_st), nst3Dm(1:n_grids_st)
read (in_unit) istchr3D(1:n_vars_st,1:n_grids_st),  &
     istchr2D(1:n_vars_st,1:n_grids_st), istchr3Ds(1:n_vars_st,1:n_grids_st),  &
     istchr2Ds(1:n_vars_st,1:n_grids_st), istchr3Dm(1:n_vars_st,1:n_grids_st)

close(in_unit)     ! Close header file

! Loop through all compatible grids, checking if each variable is present
!  in the current run; if so, fill current variable with data:

do igd=1,min(n_grids, n_grids_st)

   write (cgrid, '(a2,i2.2)') '-g', igd   ! Create grid ID string

   sfn_data = input_state_file_pfx(1:nc)//cgrid//'.vfm'  ! State data file name

   ! Open state (data) file:
   recsize_f = gd(igd)%nxyp

   allocate(reclen_test(recsize_f, 1), stat=e)
   if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'reclen')
   IO_reclen = determine_IO_reclen(reclen_test)
   deallocate(reclen_test, stat=e)
   if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'reclen')

   open(in_unit, file=sfn_data(1:len_trim(sfn_data)), form='unformatted',  &
        status='old', access='direct', recl=IO_reclen)

   recn = 1

   ! Loop through the 3D variables in the file:
   do iq=1,nst3D(igd)
      ! Check if this variable is present in this run; if yes, fill with data:
      do iv=1,nv3D(igd)
         if (istchr3D(iq,igd) == ivchr3D(iv,igd)) then
            write (*, '(3a,i2,a)') 'Filling ', trim(istchr3D(iq,igd)),  &
                 ' (grid #', igd, ') from the state file(s).'
            ib = ind3Dn(igd,ivtab3d(iv,igd))
            do k=1,gd(igd)%nzp
               ie = ib+recsize_f-1
               read (in_unit, rec=recn) a(ib:ie)
               recn = recn+1
               ib = ie+1
            end do
            exit
         end if
      end do
      if (iv == nv3D(igd)+1) recn = recn+gd(igd)%nzp  ! Skip this variable
   end do

   ! Loop through the 2D variables in the file:
   do iq=1,nst2D(igd)
      ! Check if this variable is present in this run; if yes, fill with data:
      do iv=1,nv2D(igd)
         if (istchr2D(iq,igd) == ivchr2D(iv,igd)) then
            write (*, '(3a,i2,a)') 'Filling ', trim(istchr2D(iq,igd)),  &
                 ' (grid #', igd, ') from the state file(s).'
            ib = ind2Dn(igd,ivtab2d(iv,igd))
            ie = ib+recsize_f-1
            read (in_unit, rec=recn) a(ib:ie)
            recn = recn+1
            exit
         end if
      end do
      if (iv == nv2D(igd)+1) recn = recn+1  ! Skip this variable
   end do

   ! 2D non-a-array variables:

   read (in_unit, rec=recn) gd(igd)%emisst
   recn = recn+1

   ! Loop through the 3D SBPS variables in the file:
   do iq=1,nst3Ds(igd)
      ! Check if this variable is present in this run; if yes, fill with data:
      do iv=1,nv3Ds(igd)
         if (istchr3Ds(iq,igd) == ivchr3Ds(iv,igd)) then
            write (*, '(3a,i2,a)') 'Filling ', trim(istchr3Ds(iq,igd)),  &
                 ' (grid #', igd, ') from the state file(s).'
            ib = ind3Dsn(igd,ivtab3ds(iv,igd))
            do ip=1,npatch
               do k=1,nzg+nzs
                  ie = ib+recsize_f-1
                  read (in_unit, rec=recn) a(ib:ie)
                  recn = recn+1
                  ib = ie+1
               end do
            end do
            exit
         end if
      end do
      if (iv == nv3Ds(igd)+1) recn = recn+(nzg+nzs)*npatch  ! Skip this variable
   end do

   ! Loop through the 2D SBPS variables in the file:
   do iq=1,nst2Ds(igd)
      ! Check if this variable is present in this run; if yes, fill with data:
      do iv=1,nv2Ds(igd)
         if (istchr2Ds(iq,igd) == ivchr2Ds(iv,igd)) then
            write (*, '(3a,i2,a)') 'Filling ', trim(istchr2Ds(iq,igd)),  &
                 ' (grid #', igd, ') from the state file(s).'
            ib = ind2Dsn(igd,ivtab2ds(iv,igd))
            do ip=1,npatch
               ie = ib+recsize_f-1
               read (in_unit, rec=recn) a(ib:ie)
               recn = recn+1
               ib = ie+1
            end do
            exit
         end if
      end do
      if (iv == nv2Ds(igd)+1) recn = recn+npatch  ! Skip this variable
   end do

   ! 2D SBPS non-a-array variables:

   do ip=1,npatch
      read (in_unit, rec=recn) gd(igd)%emissivity_static(:,:,ip)
      recn = recn+1
   end do

end do

close(in_unit)

! Deallocate work arrays:

deallocate(nstzp, nstxp, nstyp, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

deallocate(nst3D, nst2D, nst3Ds, nst2Ds, nst3Dm, istchr3D, istchr2D,  &
     istchr3Ds, istchr2Ds, istchr3Dm, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '2')

end subroutine restart


!```````````````````````````````````````````````````````````````````
subroutine copy_model_state(copy_to_backup)

!+ Make a copy of all current necessary model variable contents

!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use main_memory_mod, only: n_grids, naddsc, a, gd, my_rank, rootPE_rank,  &
     nzg, nzs, npatch
use main_memory_bspec_mod, only: CO2i_nb, H2Oi_nb, dust_nb, CO2i_tog,  &
     H2Oi_tog, dust_tog, H2Ov_tog, gd_bd
use vtables_mod, only: iupn, iucn, ivpn, ivcn, iwpn, iwcn, ippn, ipcn, ithpn,  &
      ithetan, itkepn, ifthrdn, ihkmn, ivkmn, ivkhn, irvn, idustncn,  &
      idustfin, idustffn, idustajn, idustagn, idustrsn, iH2Oifin, iH2Oiffn,  &
      iH2Oiajn, iH2Oiagn, iH2Oirsn, iH2Odfin, iH2Odffn, iH2Odajn, iH2Odagn,  &
      iH2Odrsn, iCO2ifin, iCO2iffn, iCO2iajn, iCO2iagn, iCO2irsn, iCO2dfin,  &
      iCO2dffn, iCO2dajn, iCO2dagn, iCO2drsn, iCO2wfin, iCO2wffn, iCO2wajn,  &
      iCO2wagn, iCO2wrsn, iH2Oincn, iH2Odcrn, iCO2incn, iCO2dcrn, iCO2wcrn,  &
      isclpn, iuwn, ivwn, iwfzn, itfzn, iqfzn, irshortn, irlongn, irlongupn,  &
      ialbedtn, icoszn, iSBPSdensn, iSBPStmcpn, iSBPStmcdn, iSBPSlthkn,  &
      iSBPSqn, iSBPSH2Oan, ipatchfcvn, iz0netn, isfctempkn, isfcH2Omrn

implicit none

!=== Argument declarations:

logical, intent(IN) :: copy_to_backup

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'copy_model_state'

integer :: igd, dim_1, dim_xyz, dim_xy

!=== Executable statements:

if (copy_to_backup) then      ! Copy internal model state *TO* backup memory

   do igd=1,n_grids

      if (my_rank == rootPE_rank) then
         dim_xyz = gd(igd)%nxyzp
         dim_xy = gd(igd)%nxyp
      else
         dim_xyz = gd(igd)%mmzp*gd(igd)%mmxp*gd(igd)%mmyp
         dim_xy = gd(igd)%mmxp*gd(igd)%mmyp
      end if

      ! Normal 3D atmospheric arrays {z,x,y}:

      dim_1 = dim_xyz
      gd(igd)%up_backup(:) = a(iupn(igd):iupn(igd)+dim_1-1)
      gd(igd)%uc_backup(:) = a(iucn(igd):iucn(igd)+dim_1-1)
      gd(igd)%vp_backup(:) = a(ivpn(igd):ivpn(igd)+dim_1-1)
      gd(igd)%vc_backup(:) = a(ivcn(igd):ivcn(igd)+dim_1-1)
      gd(igd)%wp_backup(:) = a(iwpn(igd):iwpn(igd)+dim_1-1)
      gd(igd)%wc_backup(:) = a(iwcn(igd):iwcn(igd)+dim_1-1)
      gd(igd)%pp_backup(:) = a(ippn(igd):ippn(igd)+dim_1-1)
      gd(igd)%pc_backup(:) = a(ipcn(igd):ipcn(igd)+dim_1-1)
      gd(igd)%thp_backup(:) = a(ithpn(igd):ithpn(igd)+dim_1-1)
      gd(igd)%theta_backup(:) =  &
           a(ithetan(igd):ithetan(igd)+dim_1-1)
      gd(igd)%fthrd_backup(:) =  &
           a(ifthrdn(igd):ifthrdn(igd)+dim_1-1)
      gd(igd)%hkm_backup(:) = a(ihkmn(igd):ihkmn(igd)+dim_1-1)
      gd(igd)%vkm_backup(:) = a(ivkmn(igd):ivkmn(igd)+dim_1-1)
      gd(igd)%vkh_backup(:) = a(ivkhn(igd):ivkhn(igd)+dim_1-1)

      if (gd(igd)%idiffk == 1 .or. gd(igd)%idiffk == 4 .or.  &
           gd(igd)%idiffk == 5 .or. gd(igd)%idiffk == 7 .or.  &
           gd(igd)%idiffk == 9)  &
           gd(igd)%tkep_backup(:) =  &
           a(itkepn(igd):itkepn(igd)+dim_1-1)

      if (H2Ov_tog > 0) gd_bd(igd)%rv_backup(:) =  &
           a(irvn(igd):irvn(igd)+dim_1-1)

      ! Aerosol 2D and 3D "atmospheric" arrays {(z),x,y,aerosol_idx}:

      if (dust_tog > 0) then
         dim_1 = dim_xyz*max(1, dust_nb)
         gd_bd(igd)%dustnc_backup(:) =  &
              a(idustncn(igd,1):idustncn(igd,1)+dim_1-1)

         dim_1 = dim_xy*max(1, dust_nb)
         gd_bd(igd)%dustfi_backup(:) =  &
              a(idustfin(igd,1):idustfin(igd,1)+dim_1-1)
         gd_bd(igd)%dustff_backup(:) =  &
              a(idustffn(igd,1):idustffn(igd,1)+dim_1-1)
         gd_bd(igd)%dustaj_backup(:) =  &
              a(idustajn(igd,1):idustajn(igd,1)+dim_1-1)
         gd_bd(igd)%dustag_backup(:) =  &
              a(idustagn(igd,1):idustagn(igd,1)+dim_1-1)
         gd_bd(igd)%dustrs_backup(:) =  &
              a(idustrsn(igd,1):idustrsn(igd,1)+dim_1-1)
      end if

      if (H2Oi_tog > 0) then
         dim_1 = dim_xyz*max(1, H2Oi_nb)
         gd_bd(igd)%H2Oinc_backup(:) =  &
              a(iH2Oincn(igd,1):iH2Oincn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Odcr_backup(:) =  &
              a(iH2Odcrn(igd,1):iH2Odcrn(igd,1)+dim_1-1)

         dim_1 = dim_xy*max(1, H2Oi_nb)
         gd_bd(igd)%H2Oifi_backup(:) =  &
              a(iH2Oifin(igd,1):iH2Oifin(igd,1)+dim_1-1)
         gd_bd(igd)%H2Oiff_backup(:) =  &
              a(iH2Oiffn(igd,1):iH2Oiffn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Oiaj_backup(:) =  &
              a(iH2Oiajn(igd,1):iH2Oiajn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Oiag_backup(:) =  &
              a(iH2Oiagn(igd,1):iH2Oiagn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Oirs_backup(:) =  &
              a(iH2Oirsn(igd,1):iH2Oirsn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Odfi_backup(:) =  &
              a(iH2Odfin(igd,1):iH2Odfin(igd,1)+dim_1-1)
         gd_bd(igd)%H2Odff_backup(:) =  &
              a(iH2Odffn(igd,1):iH2Odffn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Odaj_backup(:) =  &
              a(iH2Odajn(igd,1):iH2Odajn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Odag_backup(:) =  &
              a(iH2Odagn(igd,1):iH2Odagn(igd,1)+dim_1-1)
         gd_bd(igd)%H2Odrs_backup(:) =  &
              a(iH2Odrsn(igd,1):iH2Odrsn(igd,1)+dim_1-1)
      end if

      if (CO2i_tog > 0) then
         dim_1 = dim_xyz*max(1, CO2i_nb)
         gd_bd(igd)%CO2inc_backup(:) =  &
              a(iCO2incn(igd,1):iCO2incn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2dcr_backup(:) =  &
              a(iCO2dcrn(igd,1):iCO2dcrn(igd,1)+dim_1-1)

         dim_1 = dim_xy*max(1, CO2i_nb)
         gd_bd(igd)%CO2ifi_backup(:) =  &
              a(iCO2ifin(igd,1):iCO2ifin(igd,1)+dim_1-1)
         gd_bd(igd)%CO2iff_backup(:) =  &
              a(iCO2iffn(igd,1):iCO2iffn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2iaj_backup(:) =  &
              a(iCO2iajn(igd,1):iCO2iajn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2iag_backup(:) =  &
              a(iCO2iagn(igd,1):iCO2iagn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2irs_backup(:) =  &
              a(iCO2irsn(igd,1):iCO2irsn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2dfi_backup(:) =  &
              a(iCO2dfin(igd,1):iCO2dfin(igd,1)+dim_1-1)
         gd_bd(igd)%CO2dff_backup(:) =  &
              a(iCO2dffn(igd,1):iCO2dffn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2daj_backup(:) =  &
              a(iCO2dajn(igd,1):iCO2dajn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2dag_backup(:) =  &
              a(iCO2dagn(igd,1):iCO2dagn(igd,1)+dim_1-1)
         gd_bd(igd)%CO2drs_backup(:) =  &
              a(iCO2drsn(igd,1):iCO2drsn(igd,1)+dim_1-1)

         if (H2Oi_tog > 0) then
            dim_1 = dim_xyz*max(1, CO2i_nb)
            gd_bd(igd)%CO2wcr_backup(:) =  &
                 a(iCO2wcrn(igd,1):iCO2wcrn(igd,1)+dim_1-1)

            dim_1 = dim_xy*max(1, CO2i_nb)
            gd_bd(igd)%CO2wfi_backup(:) =  &
                 a(iCO2wfin(igd,1):iCO2wfin(igd,1)+dim_1-1)
            gd_bd(igd)%CO2wff_backup(:) =  &
                 a(iCO2wffn(igd,1):iCO2wffn(igd,1)+dim_1-1)
            gd_bd(igd)%CO2waj_backup(:) =  &
                 a(iCO2wajn(igd,1):iCO2wajn(igd,1)+dim_1-1)
            gd_bd(igd)%CO2wag_backup(:) =  &
                 a(iCO2wagn(igd,1):iCO2wagn(igd,1)+dim_1-1)
            gd_bd(igd)%CO2wrs_backup(:) =  &
                 a(iCO2wrsn(igd,1):iCO2wrsn(igd,1)+dim_1-1)
         end if
      end if

      ! Generic scalar 3D atmospheric arrays {z,x,y,scalar_idx}:

      if (naddsc >= 1) then
         dim_1 = dim_xyz*max(1, naddsc)
         gd(igd)%sclp_backup(:) =  &
              a(isclpn(igd,1):isclpn(igd,1)+dim_1-1)
      end if

      ! Normal 2D "atmospheric" arrays {x,y}:

      dim_1 = dim_xy
      gd(igd)%uw_backup(:) = a(iuwn(igd):iuwn(igd)+dim_1-1)
      gd(igd)%vw_backup(:) = a(ivwn(igd):ivwn(igd)+dim_1-1)
      gd(igd)%wfz_backup(:) = a(iwfzn(igd):iwfzn(igd)+dim_1-1)
      gd(igd)%tfz_backup(:) = a(itfzn(igd):itfzn(igd)+dim_1-1)
      gd_bd(igd)%qfz_backup(:) = a(iqfzn(igd):iqfzn(igd)+dim_1-1)
      gd(igd)%rshort_backup(:) =  &
           a(irshortn(igd):irshortn(igd)+dim_1-1)
      gd(igd)%rlong_backup(:) =  &
           a(irlongn(igd):irlongn(igd)+dim_1-1)
      gd(igd)%rlongup_backup(:) =  &
           a(irlongupn(igd):irlongupn(igd)+dim_1-1)
      gd(igd)%albedt_backup(:) =  &
           a(ialbedtn(igd):ialbedtn(igd)+dim_1-1)
      gd(igd)%cosz_backup(:) =  &
           a(icoszn(igd):icoszn(igd)+dim_1-1)
      gd(igd)%emisst_backup = gd(igd)%emisst
      gd(igd)%bottom_other_sflux_down_backup = gd(igd)%bottom_other_sflux_down
      gd(igd)%bottom_other_sflux_up_backup = gd(igd)%bottom_other_sflux_up
      gd(igd)%bottom_SWflux_down_backup = gd(igd)%bottom_SWflux_down
      gd(igd)%bottom_SWflux_up_backup = gd(igd)%bottom_SWflux_up
      gd(igd)%topo_factor_sflux_backup = gd(igd)%topo_factor_sflux

      ! Normal 3D sub/super-surface arrays {z,x,y,patch}:

      dim_1 = (nzg+nzs)*dim_xy*npatch
      gd(igd)%SBPSdens_backup(:) =  &
           a(iSBPSdensn(igd):iSBPSdensn(igd)+dim_1-1)
      gd(igd)%SBPStmcp_backup(:) =  &
           a(iSBPStmcpn(igd):iSBPStmcpn(igd)+dim_1-1)
      gd(igd)%SBPStmcd_backup(:) =  &
           a(iSBPStmcdn(igd):iSBPStmcdn(igd)+dim_1-1)
      gd(igd)%SBPSlthk_backup(:) =  &
           a(iSBPSlthkn(igd):iSBPSlthkn(igd)+dim_1-1)
      gd(igd)%SBPSq_backup(:) =  &
           a(iSBPSqn(igd):iSBPSqn(igd)+dim_1-1)
      gd_bd(igd)%SBPSH2Oa_backup(:) =  &
           a(iSBPSH2Oan(igd):iSBPSH2Oan(igd)+dim_1-1)

      ! Normal 2D sub/super-surface arrays {x,y,patch}:

      dim_1 = dim_xy*npatch
      gd(igd)%patchfcv_backup(:) =  &
           a(ipatchfcvn(igd):ipatchfcvn(igd)+dim_1-1)
      gd(igd)%z0net_backup(:) =  &
           a(iz0netn(igd):iz0netn(igd)+dim_1-1)
      gd(igd)%sfctempk_backup(:) =  &
           a(isfctempkn(igd):isfctempkn(igd)+dim_1-1)
      gd_bd(igd)%sfcH2Omr_backup(:) =  &
           a(isfcH2Omrn(igd):isfcH2Omrn(igd)+dim_1-1)

   end do

else      ! Copy prior internal model state *FROM* backup memory

   do igd=1,n_grids

      if (my_rank == rootPE_rank) then
         dim_xyz = gd(igd)%nxyzp
         dim_xy = gd(igd)%nxyp
      else
         dim_xyz = gd(igd)%mmzp*gd(igd)%mmxp*gd(igd)%mmyp
         dim_xy = gd(igd)%mmxp*gd(igd)%mmyp
      end if

      ! Normal 3D atmospheric arrays {z,x,y}:

      dim_1 = dim_xyz

      a(iupn(igd):iupn(igd)+dim_1-1) = gd(igd)%up_backup(:)
      a(iucn(igd):iucn(igd)+dim_1-1) = gd(igd)%uc_backup(:)
      a(ivpn(igd):ivpn(igd)+dim_1-1) = gd(igd)%vp_backup(:)
      a(ivcn(igd):ivcn(igd)+dim_1-1) = gd(igd)%vc_backup(:)
      a(iwpn(igd):iwpn(igd)+dim_1-1) = gd(igd)%wp_backup(:)
      a(iwcn(igd):iwcn(igd)+dim_1-1) = gd(igd)%wc_backup(:)
      a(ippn(igd):ippn(igd)+dim_1-1) = gd(igd)%pp_backup(:)
      a(ipcn(igd):ipcn(igd)+dim_1-1) = gd(igd)%pc_backup(:)
      a(ithpn(igd):ithpn(igd)+dim_1-1) = gd(igd)%thp_backup(:)
      a(ithetan(igd):ithetan(igd)+dim_1-1) =  &
           gd(igd)%theta_backup(:)
      a(ifthrdn(igd):ifthrdn(igd)+dim_1-1) =  &
           gd(igd)%fthrd_backup(:)
      a(ihkmn(igd):ihkmn(igd)+dim_1-1) = gd(igd)%hkm_backup(:)
      a(ivkmn(igd):ivkmn(igd)+dim_1-1) = gd(igd)%vkm_backup(:)
      a(ivkhn(igd):ivkhn(igd)+dim_1-1) = gd(igd)%vkh_backup(:)

      if (gd(igd)%idiffk == 1 .or. gd(igd)%idiffk == 4 .or.  &
           gd(igd)%idiffk == 5 .or. gd(igd)%idiffk == 7 .or.  &
           gd(igd)%idiffk == 9)  &
           a(itkepn(igd):itkepn(igd)+dim_1-1) =  &
           gd(igd)%tkep_backup(:)

      if (H2Ov_tog > 0) a(irvn(igd):irvn(igd)+dim_1-1) =  &
           gd_bd(igd)%rv_backup(:)

      ! Aerosol 2D and 3D "atmospheric" arrays {(z),x,y,aerosol_idx}:

      if (dust_tog > 0) then
         dim_1 = dim_xyz*max(1, dust_nb)
         a(idustncn(igd,1):idustncn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%dustnc_backup(:)

         dim_1 = dim_xy*max(1, dust_nb)
         a(idustfin(igd,1):idustfin(igd,1)+dim_1-1) =  &
              gd_bd(igd)%dustfi_backup(:)
         a(idustffn(igd,1):idustffn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%dustff_backup(:)
         a(idustajn(igd,1):idustajn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%dustaj_backup(:)
         a(idustagn(igd,1):idustagn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%dustag_backup(:)
         a(idustrsn(igd,1):idustrsn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%dustrs_backup(:)
      end if

      if (H2Oi_tog > 0) then
         dim_1 = dim_xyz*max(1, H2Oi_nb)
         a(iH2Oincn(igd,1):iH2Oincn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Oinc_backup(:)
         a(iH2Odcrn(igd,1):iH2Odcrn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Odcr_backup(:)

         dim_1 = dim_xy*max(1, H2Oi_nb)
         a(iH2Oifin(igd,1):iH2Oifin(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Oifi_backup(:)
         a(iH2Oiffn(igd,1):iH2Oiffn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Oiff_backup(:)
         a(iH2Oiajn(igd,1):iH2Oiajn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Oiaj_backup(:)
         a(iH2Oiagn(igd,1):iH2Oiagn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Oiag_backup(:)
         a(iH2Oirsn(igd,1):iH2Oirsn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Oirs_backup(:)
         a(iH2Odfin(igd,1):iH2Odfin(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Odfi_backup(:)
         a(iH2Odffn(igd,1):iH2Odffn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Odff_backup(:)
         a(iH2Odajn(igd,1):iH2Odajn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Odaj_backup(:)
         a(iH2Odagn(igd,1):iH2Odagn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Odag_backup(:)
         a(iH2Odrsn(igd,1):iH2Odrsn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%H2Odrs_backup(:)
      end if

      if (CO2i_tog > 0) then
         dim_1 = dim_xyz*max(1, CO2i_nb)
         a(iCO2incn(igd,1):iCO2incn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2inc_backup(:)
         a(iCO2dcrn(igd,1):iCO2dcrn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2dcr_backup(:)

         dim_1 = dim_xy*max(1, CO2i_nb)
         a(iCO2ifin(igd,1):iCO2ifin(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2ifi_backup(:)
         a(iCO2iffn(igd,1):iCO2iffn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2iff_backup(:)
         a(iCO2iajn(igd,1):iCO2iajn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2iaj_backup(:)
         a(iCO2iagn(igd,1):iCO2iagn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2iag_backup(:)
         a(iCO2irsn(igd,1):iCO2irsn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2irs_backup(:)
         a(iCO2dfin(igd,1):iCO2dfin(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2dfi_backup(:)
         a(iCO2dffn(igd,1):iCO2dffn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2dff_backup(:)
         a(iCO2dajn(igd,1):iCO2dajn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2daj_backup(:)
         a(iCO2dagn(igd,1):iCO2dagn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2dag_backup(:)
         a(iCO2drsn(igd,1):iCO2drsn(igd,1)+dim_1-1) =  &
              gd_bd(igd)%CO2drs_backup(:)

         if (H2Oi_tog > 0) then
            dim_1 = dim_xyz*max(1, CO2i_nb)
            a(iCO2wcrn(igd,1):iCO2wcrn(igd,1)+dim_1-1) =  &
                 gd_bd(igd)%CO2wcr_backup(:)

            dim_1 = dim_xy*max(1, CO2i_nb)
            a(iCO2wfin(igd,1):iCO2wfin(igd,1)+dim_1-1) =  &
                 gd_bd(igd)%CO2wfi_backup(:)
            a(iCO2wffn(igd,1):iCO2wffn(igd,1)+dim_1-1) =  &
                 gd_bd(igd)%CO2wff_backup(:)
            a(iCO2wajn(igd,1):iCO2wajn(igd,1)+dim_1-1) =  &
                 gd_bd(igd)%CO2waj_backup(:)
            a(iCO2wagn(igd,1):iCO2wagn(igd,1)+dim_1-1) =  &
                 gd_bd(igd)%CO2wag_backup(:)
            a(iCO2wrsn(igd,1):iCO2wrsn(igd,1)+dim_1-1) =  &
                 gd_bd(igd)%CO2wrs_backup(:)
         end if
      end if

      ! Generic scalar 3D atmospheric arrays {z,x,y,scalar_idx}:

      if (naddsc >= 1) then
         dim_1 = dim_xyz*max(1, naddsc)
         a(isclpn(igd,1):isclpn(igd,1)+dim_1-1) =  &
              gd(igd)%sclp_backup(:)
      end if

      ! Normal 2D "atmospheric" arrays {x,y}:

      dim_1 = dim_xy
      a(iuwn(igd):iuwn(igd)+dim_1-1) = gd(igd)%uw_backup(:)
      a(ivwn(igd):ivwn(igd)+dim_1-1) = gd(igd)%vw_backup(:)
      a(iwfzn(igd):iwfzn(igd)+dim_1-1) = gd(igd)%wfz_backup(:)
      a(itfzn(igd):itfzn(igd)+dim_1-1) = gd(igd)%tfz_backup(:)
      a(iqfzn(igd):iqfzn(igd)+dim_1-1) = gd_bd(igd)%qfz_backup(:)
      a(irshortn(igd):irshortn(igd)+dim_1-1) =  &
           gd(igd)%rshort_backup(:)
      a(irlongn(igd):irlongn(igd)+dim_1-1) =  &
           gd(igd)%rlong_backup(:)
      a(irlongupn(igd):irlongupn(igd)+dim_1-1) =  &
           gd(igd)%rlongup_backup(:)
      a(ialbedtn(igd):ialbedtn(igd)+dim_1-1) =  &
           gd(igd)%albedt_backup(:)
      a(icoszn(igd):icoszn(igd)+dim_1-1) =  &
           gd(igd)%cosz_backup(:)
      gd(igd)%emisst = gd(igd)%emisst_backup
      gd(igd)%bottom_other_sflux_down = gd(igd)%bottom_other_sflux_down_backup
      gd(igd)%bottom_other_sflux_up = gd(igd)%bottom_other_sflux_up_backup
      gd(igd)%bottom_SWflux_down = gd(igd)%bottom_SWflux_down_backup
      gd(igd)%bottom_SWflux_up = gd(igd)%bottom_SWflux_up_backup
      gd(igd)%topo_factor_sflux = gd(igd)%topo_factor_sflux_backup

      ! Normal 3D sub/super-surface arrays {z,x,y,patch}:

      dim_1 = (nzg+nzs)*dim_xy*npatch
      a(iSBPSdensn(igd):iSBPSdensn(igd)+dim_1-1) =  &
           gd(igd)%SBPSdens_backup(:)
      a(iSBPStmcpn(igd):iSBPStmcpn(igd)+dim_1-1) =  &
           gd(igd)%SBPStmcp_backup(:)
      a(iSBPStmcdn(igd):iSBPStmcdn(igd)+dim_1-1) =  &
           gd(igd)%SBPStmcd_backup(:)
      a(iSBPSlthkn(igd):iSBPSlthkn(igd)+dim_1-1) =  &
           gd(igd)%SBPSlthk_backup(:)
      a(iSBPSqn(igd):iSBPSqn(igd)+dim_1-1) =  &
           gd(igd)%SBPSq_backup(:)
      a(iSBPSH2Oan(igd):iSBPSH2Oan(igd)+dim_1-1) =  &
           gd_bd(igd)%SBPSH2Oa_backup(:)

      ! Normal 2D sub/super-surface arrays {x,y,patch}:

      dim_1 = dim_xy*npatch
      a(ipatchfcvn(igd):ipatchfcvn(igd)+dim_1-1) =  &
           gd(igd)%patchfcv_backup(:)
      a(iz0netn(igd):iz0netn(igd)+dim_1-1) =  &
           gd(igd)%z0net_backup(:)
      a(isfctempkn(igd):isfctempkn(igd)+dim_1-1) =  &
           gd(igd)%sfctempk_backup(:)
      a(isfcH2Omrn(igd):isfcH2Omrn(igd)+dim_1-1) =  &
           gd_bd(igd)%sfcH2Omr_backup(:)

   end do

end if

end subroutine copy_model_state
