!```````````````````````````````````````````````````````````````````
module surface_nest_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine fillscr_sfc4D_3D(n1, n2, n3, m4, m2, m3, m5, tgtm5, newa, olda)

!+ Fill a 3D scratch array with info from a 4D sfc array (n4,n2,n3,n5)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:
  
integer, intent(IN) :: n1, n2, n3, m4, m2, m3, m5, tgtm5
real(f4), dimension(m4,m2,m3,m5), intent(IN) :: olda

real(f4), dimension(n1,n2,n3), intent(INOUT) :: newa

!== Local declarations:

! NONE

!=== Executable statements:

newa(1:m4,1:m2,1:m3) = olda(1:m4,1:m2,1:m3,tgtm5)             

end subroutine fillscr_sfc4D_3D


!```````````````````````````````````````````````````````````````````
subroutine fillscr_sfc4D_2D(n2, n3, m4, m2, m3, m5, tgtm4, tgtm5, newa, olda)

!+ Fill a 2D scratch array with info from a 4D sfc array (n4,n2,n3,n5)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:
  
integer, intent(IN) :: n2, n3, m4, m2, m3, m5, tgtm5, tgtm4
real(f4), dimension(m4,m2,m3,m5), intent(IN) :: olda

real(f4), dimension(n2,n3), intent(INOUT) :: newa

!== Local declarations:

integer :: i, j

!=== Executable statements:

do j=1,m3
   do i=1,m2
      newa(i,j) = olda(tgtm4,i,j,tgtm5)             
   end do
end do

end subroutine fillscr_sfc4D_2D


!```````````````````````````````````````````````````````````````````
subroutine fillscr_sfc3D_2D(n2, n3, m2, m3, m5, tgtm5, newa, olda)

!+ Fill a 2D scratch array with info from a 3D sfc array (n2,n3,n5)

use univ_kind_defs_mod, only: f4

implicit none
  
!=== Argument declarations:

integer, intent(IN) :: n2, n3, m2, m3, m5, tgtm5
real(f4), dimension(m2,m3,m5), intent(IN) :: olda

real(f4), dimension(n2,n3), intent(INOUT) :: newa

!== Local declarations:

! NONE

!=== Executable statements:

newa(1:m2,1:m3) = olda(1:m2,1:m3,tgtm5)             

end subroutine fillscr_sfc3D_2D


!```````````````````````````````````````````````````````````````````
subroutine fillvar_sfc3D_4D(n1, n2, n3, m4, m2, m3, m5, tgtm5, olda, newa)

!+ Fill a 4D sfc array (n4,n2,n3,n5) with info from a 3D scratch array

use univ_kind_defs_mod, only: f4

implicit none  
  
!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3, m4, m2, m3, m5, tgtm5
real(f4), dimension(n1,n2,n3), intent(IN) :: olda

real(f4), dimension(m4,m2,m3,m5), intent(INOUT) :: newa

!== Local declarations:

! NONE

!=== Executable statements:

newa(1:m4,1:m2,1:m3,tgtm5) = olda(1:m4,1:m2,1:m3) 

end subroutine fillvar_sfc3D_4D


!```````````````````````````````````````````````````````````````````
subroutine fillvar_sfc2D_4D(n2, n3, m4, m2, m3, m5, tgtm4, tgtm5, olda, newa)

!+ Fill a 4D sfc array (n4,n2,n3,n5) with info from a 2D scratch array

use univ_kind_defs_mod, only: f4

implicit none  

!=== Argument declarations:

integer, intent(IN) :: n2, n3, m4, m2, m3, m5, tgtm5, tgtm4
real(f4), dimension(n2,n3), intent(IN) :: olda

real(f4), dimension(m4,m2,m3,m5), intent(INOUT) :: newa

!== Local declarations:

integer :: i, j

!=== Executable statements:

do j=1,m3
   do i=1,m2
      newa(tgtm4,i,j,tgtm5) = olda(i,j) 
   end do
end do

end subroutine fillvar_sfc2D_4D


!```````````````````````````````````````````````````````````````````
subroutine fillvar_sfc2D_3D(n2, n3, m2, m3, m5, tgtm5, olda, newa)

!+ Fill a 3D sfc array (n2,n3,n5) with info from a 2D scratch array

use univ_kind_defs_mod, only: f4

implicit none  
  
!=== Argument declarations:

integer, intent(IN) :: n2, n3, m2, m3, m5, tgtm5
real(f4), dimension(n2,n3), intent(IN) :: olda

real(f4), dimension(m2,m3,m5), intent(INOUT) :: newa

!== Local declarations:

! NONE

!=== Executable statements:

newa(1:m2,1:m3,tgtm5) = olda(1:m2,1:m3) 

end subroutine fillvar_sfc2D_3D


!```````````````````````````````````````````````````````````````````
subroutine ae_sfc3D_3D(n2, n3, n5, i1, i2, j1, j2, tgtn5, newa, olda)

!+ Fill a *portion* of a 3D scratch (sfc) array (n2,n3,n5) with info from a
!   portion of a 3D sfc array (n2,n3,n5)

use univ_kind_defs_mod, only: f4

implicit none  
  
!=== Argument declarations:

integer, intent(IN) :: n2, n3, n5, tgtn5, i1, i2, j1, j2
real(f4), dimension(n2,n3,n5), intent(IN) :: olda

real(f4), dimension(n2,n3,n5), intent(INOUT) :: newa

!== Local declarations:

! NONE

!=== Executable statements:

newa(i1:i2,j1:j2,tgtn5) = olda(i1:i2,j1:j2,tgtn5)

end subroutine ae_sfc3D_3D


!```````````````````````````````````````````````````````````````````
subroutine ae_sfc4D_4D(n4, n2, n3, n5, k1, k2, i1, i2, j1, j2, tgtn5, newa,  &
     olda)

!+ Fill a *portion* of a 4D scratch (sfc) array (n4,n2,n3,n5) with info from a
!   portion of a 4D sfc array (n4,n2,n3,n5)

use univ_kind_defs_mod, only: f4

implicit none  
  
!=== Argument declarations:

integer, intent(IN) :: n2, n3, n4, n5, tgtn5, i1, i2, j1, j2, k1, k2
real(f4), dimension(n4,n2,n3,n5), intent(IN) :: olda

real(f4), dimension(n4,n2,n3,n5), intent(INOUT) :: newa

!== Local declarations:

! NONE

!=== Executable statements:

newa(k1:k2,i1:i2,j1:j2,tgtn5) = olda(k1:k2,i1:i2,j1:j2,tgtn5)

end subroutine ae_sfc4D_4D


!```````````````````````````````````````````````````````````````````
subroutine ae2_TIM(n2, n3, i1, i2, j1, j2, a, b)

!+ Transfer 2D data ("legacy" routine -- FIX ME!!!)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:
  
integer, intent(IN) :: n2, n3, i1, i2, j1, j2
real(f4), dimension(n2,n3), intent(IN) :: b

real(f4), dimension(n2,n3), intent(OUT) :: a

!== Local declarations:

integer :: i, j

!=== Executable statements:

do j=j1,j2
   do i=i1,i2
      a(i,j) = b(i,j)
   end do
end do

end subroutine ae2_TIM


end module surface_nest_mod


!```````````````````````````````````````````````````````````````````
subroutine surface_nest(ngra, ngrb, n4, n5, vnam)

!+ Initialize some surface quantities from files, etc.

use univ_kind_defs_mod, only: f4
use state_IO_mod, only: n_grids_st
use main_memory_mod, only: a, iscr1, iscr2, jdim, gd, npatch, nzg, nzs,  &
     maxnxp, maxnyp, albedo
use vtables_mod, only: itoptn, ipatchfcvn, itopz0n, itistatn, ialbdstatn,  &
     isbpstmcpn, isbpsdensn, isbpstmcdn, iglatn, ialbdstatn, ivt3da,  &
     ivt3db, iz0netn, iz0statn, itopz0n, itistatn, isbpslthkn, isbpsqn,  &
     ialbedtn, isfctempkn, ithetan, ipi0n, ippn, iglatn, iglonn, ifmaptn
use surface_init_mod, only: toptinit, sfcinit_kludge, TI_init, albedo_init,  &
     kludge_sbpsprop
use nest_lib_mod, only: fillscr, fillvar
use nest_interpolation_mod, only: eintp, eintp_3dsfc, eintp_2dsfc
use surface_nest_mod, only: fillscr_sfc3d_2d, fillvar_sfc2d_3d,  &
     fillscr_sfc4d_3d, fillvar_sfc3d_4d, ae_sfc3d_3d, ae_sfc4d_4d,  &
     fillscr_sfc4d_2d, fillvar_sfc2d_4d, ae2_TIM
use nest_mod, only: fdback
use surface_data_prep_mod, only: surface_data_prep

implicit none

!=== Argument declarations:

integer, intent(IN) :: ngra, ngrb, n4, n5
character(len=*), intent(IN) :: vnam

!== Local declarations:

real(f4), external :: valugp

integer :: ifm, icm, ip, nc1, ik

!=== Executable statements:

if (vnam == 'FILEDATA') then

   do ifm=ngra,ngrb
      icm = gd(ifm)%nxtnest

      ! Initialize topography and other surface characteristics:
 
      call toptinit(gd(ifm)%nxp, gd(ifm)%nyp, a(itoptn(ifm)))

      call sfcinit(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, nzg+nzs, npatch,  &
           nzg, ifm, vnam, a(ipatchfcvn(ifm)), a(isbpstmcpn(ifm)),  &
           a(isbpsdensn(ifm)), a(itistatn(ifm)), a(ialbdstatn(ifm)),  &
           a(isbpstmcdn(ifm)), a(isbpslthkn(ifm)), a(isbpsqn(ifm)),  &
           a(ialbedtn(ifm)), a(isfctempkn(ifm)), a(iz0netn(ifm)),  &
           a(iz0statn(ifm)), a(itopz0n(ifm)), a(ithetan(ifm)), a(ipi0n(ifm)),  &
           a(ippn(ifm)), a(iglatn(ifm)), a(iglonn(ifm)), a(itoptn(ifm)))

      !*** Topography:

      if (icm >= 1 .and. gd(ifm)%itoptflg == 0) then

         ! Interpolate topography from coarser grid:
          
         call fillscr(1, maxnxp, maxnyp, 1, gd(icm)%nxp, gd(icm)%nyp, 1, 1,  &
              a(iscr1), a(itoptn(icm)))
  
        call eintp(a(iscr1), a(iscr2), 1, maxnxp, maxnyp, 1, gd(ifm)%nxp,  &
              gd(ifm)%nyp, ifm, 2, 't', 0, 0)
  
        call fillvar(1, maxnxp, maxnyp, 1, gd(ifm)%nxp, gd(ifm)%nyp, 1, 1,  &
             a(iscr2), a(itoptn(ifm)))

         ! Interpolate topography-based roughness length from coarser grid:
         
         call fillscr(1, maxnxp, maxnyp, 1, gd(icm)%nxp, gd(icm)%nyp, 1, 1,  &
              a(iscr1), a(itopz0n(icm)))
         call eintp(a(iscr1), a(iscr2), 1, maxnxp, maxnyp, 1, gd(ifm)%nxp,  &
              gd(ifm)%nyp, ifm, 2, 't', 0, 0)
         call fillvar(1, maxnxp, maxnyp, 1, gd(ifm)%nxp, gd(ifm)%nyp, 1, 1,  &
              a(iscr2), a(itopz0n(ifm)))

      else if (gd(ifm)%itoptflg == 1) then

         ! Interpolate topography from standard dataset:

         call surface_data_prep(gd(ifm)%nxp, gd(ifm)%nyp, a(itoptn(ifm)),  &
              gd(ifm)%itoptfn, ifm, 'TOP')

         ! Interpolate topography-based roughness length from standard dataset:
         
         call surface_data_prep(gd(ifm)%nxp, gd(ifm)%nyp, a(itopz0n(ifm)),  &
              gd(ifm)%itoptfn, ifm, 'Z0T')

      end if

      !*** Albedo:

      if (icm >= 1 .and. gd(ifm)%ialbflg == 0) then

         do ip=1,n5    ! Loop over # of patches
         
            ! Interpolate surface albedo from coarser grid:
               
            call fillscr_sfc3D_2D(maxnxp, maxnyp, gd(icm)%nxp, gd(icm)%nyp,  &
                 n5, ip, a(iscr1), a(ialbdstatn(icm)))
            call eintp_2Dsfc(a(iscr1), a(iscr2), maxnxp, maxnyp, gd(ifm)%nxp,  &
                 gd(ifm)%nyp, ifm, 0, 0)
            call fillvar_sfc2D_3D(maxnxp, maxnyp, gd(ifm)%nxp, gd(ifm)%nyp,  &
                 n5, ip, a(iscr2), a(ialbdstatn(ifm)))
         
         end do

      else if (gd(ifm)%ialbflg == 1) then

         do ip=1,n5    ! Loop over # of patches

            ! Interpolate bare surface albedo from standard dataset:
            
            call surface_data_prep(gd(ifm)%nxp, gd(ifm)%nyp,  &
                 a(ialbdstatn(ifm)), gd(ifm)%ialbfn, ifm, 'ALB')

         end do

      end if

      !*** Thermal inertia and related quantities:

      if (icm >= 1 .and. gd(ifm)%itinflg == 0) then
         
         do ip=1,n5    ! Loop over # of patches

            ! Interpolate thermal inertia from coarser grid:
            
            call fillscr_sfc3D_2D(maxnxp, maxnyp, gd(icm)%nxp, gd(icm)%nyp,  &
                 n5, ip, a(iscr1), a(itistatn(icm)))
            call eintp_2Dsfc(a(iscr1), a(iscr2), maxnxp, maxnyp, gd(ifm)%nxp,  &
                 gd(ifm)%nyp, ifm, 0, 0)
            call fillvar_sfc2D_3D(maxnxp, maxnyp, gd(ifm)%nxp, gd(ifm)%nyp,  &
                 n5, ip, a(iscr2), a(itistatn(ifm)))

            call kludge_SBPSprop(n4, gd(ifm)%nxp, gd(ifm)%nyp, n5, nzg, ifm,  &
                 a(itistatn(ifm)), a(isbpsdensn(ifm)), a(isbpstmcpn(ifm)),  &
                 a(isbpstmcdn(ifm)), a(iglatn(ifm)), a(iglonn(ifm)),  &
                 a(ialbdstatn(ifm)))
               
!            ! Interpolate SBPS density from coarser grid:
!            
!            call fillscr_sfc4D_3D(n4,maxnxp,maxnyp,n4,nnxp(icm),nnyp(icm),n5  &
!                 ,ip,a(iscr1),a(isbpsdensn(icm)))
!            call eintp_3Dsfc(a(iscr1),a(iscr2),n4,maxnxp,maxnyp,n4,nnxp(ifm)  &
!                 ,nnyp(ifm),ifm,0,0)
!            call fillvar_sfc3D_4D(n4,maxnxp,maxnyp,n4,nnxp(ifm),nnyp(ifm),n5  &
!                 ,ip,a(iscr2),a(isbpsdensn(ifm)))
!            
!            ! Interpolate SBPS thermal capacity from coarser grid:
!            
!            call fillscr_sfc4D_3D(n4,maxnxp,maxnyp,n4,nnxp(icm),nnyp(icm),n5  &
!                 ,ip,a(iscr1),a(isbpstmcpn(icm)))
!            call eintp_3Dsfc(a(iscr1),a(iscr2),n4,maxnxp,maxnyp,n4,nnxp(ifm)  &
!                 ,nnyp(ifm),ifm,0,0)
!            call fillvar_sfc3D_4D(n4,maxnxp,maxnyp,n4,nnxp(ifm),nnyp(ifm),n5  &
!                 ,ip,a(iscr2),a(isbpstmcpn(ifm)))

         end do

      else if (gd(ifm)%itinflg == 1) then

         do ip=1,n5    ! Loop over # of patches

            ! Interpolate thermal inertia from standard dataset:
            
            call surface_data_prep(gd(ifm)%nxp, gd(ifm)%nyp,  &
                 a(itistatn(ifm)), gd(ifm)%itinfn, ifm, 'TIN')

            call kludge_SBPSprop(n4, gd(ifm)%nxp, gd(ifm)%nyp, n5, nzg, ifm,  &
                 a(itistatn(ifm)), a(isbpsdensn(ifm)), a(isbpstmcpn(ifm)),  &
                 a(isbpstmcdn(ifm)), a(iglatn(ifm)), a(iglonn(ifm)),  &
                 a(ialbdstatn(ifm)))

         end do

      end if

      ! If desired, override some current values:

      call sfcinit_kludge(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, n4, n5, nzg,  &
           ifm, vnam, a(iglatn(ifm)), a(iglonn(ifm)), a(itistatn(ifm)),  &
           a(ialbdstatn(ifm)), a(itoptn(ifm)), gd(ifm)%itinflg,  &
           gd(ifm)%ialbflg, gd(ifm)%itoptflg)

      call toptinit_user(gd(ifm)%nxp, gd(ifm)%nyp, ifm, a(itoptn(ifm)),  &
           a(ifmaptn(ifm)), gd(ifm)%xt, gd(ifm)%yt)

      call sfcinit_user(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, n4, n5, nzg,  &
           ifm, vnam, a(iglatn(ifm)), a(iglonn(ifm)), a(itistatn(ifm)),  &
           a(ialbdstatn(ifm)))

      ! Eliminate any patch area that is less than 1% of the total grid cell
      !  area; Set its area to 0, and compensate by enlarging areas of
      !  remaining patches:
!****** NEEDS WORK !!!!!!!!!!!!!!!!! ******
!%    call patch_minsize(nnxp(ifm),nnyp(ifm),nzg+nzs+3,npatch,a(ischarn(ifm)))

   end do

   ! In case surface data have been independently reassigned on any grid,
   !  average fine mesh values (and patch areas) sequentially to the coarser
   !  grids:

   if (ngra /= ngrb) then  ! Do not perform if only one grid has been specified

      do ifm=ngrb,ngra,-1

         if (gd(ifm)%nxtnest > n_grids_st .and. ifm >= 2) then

            icm = gd(ifm)%nxtnest
         
            ! Topography:
         
            call fdback(a(itoptn(icm)), a(itoptn(ifm)), a(ivt3da), a(iscr2),  &
                 1, gd(icm)%nxp, gd(icm)%nyp, 1, gd(ifm)%nxp, gd(ifm)%nyp,  &
                 ifm, 'terr', a(ivt3db))
         
            do ip=1,n5    ! Loop over # of patches

               ! Feed back thermal inertia:
            
               call fillscr_sfc3D_2D(gd(icm)%nxp, gd(icm)%nyp, gd(icm)%nxp,  &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(itistatn(icm)))
               call fillscr_sfc3D_2D(gd(ifm)%nxp, gd(ifm)%nyp, gd(ifm)%nxp,  &
                    gd(ifm)%nyp, n5, ip, a(iscr2), a(itistatn(ifm)))
               call fdback(a(iscr1), a(iscr2), a(ivt3da), a(iscr2), 1,  &
                    gd(icm)%nxp, gd(icm)%nyp, 1, gd(ifm)%nxp, gd(ifm)%nyp,  &
                    ifm, 'terr', a(ivt3db))
               call fillvar_sfc2D_3D(gd(icm)%nxp, gd(icm)%nyp, gd(icm)%nxp,  &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(itistatn(icm)))

               do ik=1,n4 

                  ! Feed back SBPS density:
            
                  call fillscr_sfc4D_2D(gd(icm)%nxp, gd(icm)%nyp, n4,  &
                       gd(icm)%nxp, gd(icm)%nyp, n5, ik, ip, a(iscr1),  &
                       a(isbpsdensn(icm)))
                  call fillscr_sfc4D_2D(gd(ifm)%nxp, gd(ifm)%nyp, n4,  &
                       gd(ifm)%nxp, gd(ifm)%nyp, n5, ik, ip, a(iscr2),  &
                       a(isbpsdensn(ifm)))
                  call fdback(a(iscr1), a(iscr2), a(ivt3da), a(iscr2), 1,  &
                       gd(icm)%nxp, gd(icm)%nyp, 1, gd(ifm)%nxp, gd(ifm)%nyp,  &
                       ifm, 'terr', a(ivt3db))
                  call fillvar_sfc2D_4D(gd(icm)%nxp, gd(icm)%nyp, n4,  &
                       gd(icm)%nxp, gd(icm)%nyp, n5, ik, ip, a(iscr1),  &
                       a(isbpsdensn(icm)))
                  
                  ! Feed back SBPS thermal capacity:
            
                  call fillscr_sfc4D_2D(gd(icm)%nxp, gd(icm)%nyp, n4,  &
                       gd(icm)%nxp, gd(icm)%nyp, n5, ik, ip, a(iscr1),  &
                       a(isbpstmcpn(icm)))
                  call fillscr_sfc4D_2D(gd(ifm)%nxp, gd(ifm)%nyp, n4,  &
                       gd(ifm)%nxp, gd(ifm)%nyp, n5, ik, ip, a(iscr2),  &
                       a(isbpstmcpn(ifm)))
                  call fdback(a(iscr1), a(iscr2), a(ivt3da), a(iscr2), 1,  &
                       gd(icm)%nxp, gd(icm)%nyp, 1, gd(ifm)%nxp, gd(ifm)%nyp,  &
                       ifm, 'terr', a(ivt3db))
                  call fillvar_sfc2D_4D(gd(icm)%nxp, gd(icm)%nyp, n4,  &
                       gd(icm)%nxp, gd(icm)%nyp, n5, ik, ip, a(iscr1),  &
                       a(isbpstmcpn(icm)))
                  
               end do

               ! Feed back static surface albedo:
            
               call fillscr_sfc3D_2D(gd(icm)%nxp, gd(icm)%nyp, gd(icm)%nxp,  &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(ialbdstatn(icm)))
               call fillscr_sfc3D_2D(gd(ifm)%nxp, gd(ifm)%nyp, gd(ifm)%nxp,  &
                    gd(ifm)%nyp, n5, ip, a(iscr2), a(ialbdstatn(ifm)))
               call fdback(a(iscr1), a(iscr2), a(ivt3da), a(iscr2), 1,  &
                    gd(icm)%nxp, gd(icm)%nyp, 1, gd(ifm)%nxp, gd(ifm)%nyp,  &
                    ifm, 'terr', a(ivt3db))
               call fillvar_sfc2D_3D(gd(icm)%nxp, gd(icm)%nyp, gd(icm)%nxp,  &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(ialbdstatn(icm)))

            end do

         end if

      end do

      ! In case terrain heights have been independently reassigned on
      ! any grid, interpolate coarse grid terrain heights to a temporary
      ! fine mesh array.  Fill the fine mesh boundary terrain heights
      ! from the temporary array:

      do ifm=ngra,ngrb
         icm = gd(ifm)%nxtnest
         
         if (icm >= 1) then

            nc1 = jdim*(gd(ifm)%nstraty+1)

            ! Topography:

            call fillscr(1, maxnxp, maxnyp, 1, gd(icm)%nxp, gd(icm)%nyp, 1,  &
                 1, a(iscr1), a(itoptn(icm)))
            call eintp(a(iscr1), a(iscr2), 1, maxnxp, maxnyp, 1, gd(ifm)%nxp,  &
                 gd(ifm)%nyp, ifm, 2, 't', 0, 0)
            call fillvar(1, maxnxp, maxnyp, 1, gd(ifm)%nxp, gd(ifm)%nyp, 1,  &
                 1, a(iscr2), a(iscr1))
            call ae2_TIM(gd(ifm)%nxp, gd(ifm)%nyp, 2+gd(ifm)%nstratx,  &
                 gd(ifm)%nxp-1-gd(ifm)%nstratx, 1+nc1, gd(ifm)%nyp-nc1,  &
                 a(iscr1), a(itoptn(ifm)))
            a(itoptn(ifm):itoptn(ifm)+gd(ifm)%nxp*gd(ifm)%nyp-1) =  &
                 a(iscr1:iscr1+gd(ifm)%nxp*gd(ifm)%nyp-1)

            do ip=1,n5    ! Loop over all patches

               ! Thermal inertia:
            
               call fillscr_sfc3D_2D(maxnxp, maxnyp, gd(icm)%nxp,  &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(itistatn(icm)))
               call eintp_2Dsfc(a(iscr1), a(iscr2), maxnxp, maxnyp,  &
                    gd(ifm)%nxp, gd(ifm)%nyp, ifm, 0, 0)
               call fillvar_sfc2D_3D(maxnxp, maxnyp, gd(ifm)%nxp,  &
                    gd(ifm)%nyp, n5, ip, a(iscr2), a(iscr1))
               call ae_sfc3D_3D(gd(ifm)%nxp, gd(ifm)%nyp, n5,  &
                    2+gd(ifm)%nstratx, gd(ifm)%nxp-1-gd(ifm)%nstratx, 1+nc1,  &
                    gd(ifm)%nyp-nc1, ip, a(iscr1), a(itistatn(ifm)))
               call ae_sfc3D_3D(gd(ifm)%nxp, gd(ifm)%nyp, n5, 1, gd(ifm)%nxp,  &
                    1, gd(ifm)%nyp, ip, a(itistatn(ifm)), a(iscr1))

               ! SBPS density:
            
               call fillscr_sfc4D_3D(n4, maxnxp, maxnyp, n4, gd(icm)%nxp, &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(isbpsdensn(icm)))
               call eintp_3Dsfc(a(iscr1), a(iscr2), n4, maxnxp, maxnyp, n4,  &
                    gd(ifm)%nxp, gd(ifm)%nyp, ifm, 0, 0)
               call fillvar_sfc3D_4D(n4, maxnxp, maxnyp, n4, gd(ifm)%nxp,  &
                    gd(ifm)%nyp, n5, ip, a(iscr2), a(iscr1))
               call ae_sfc4D_4D(n4, gd(ifm)%nxp, gd(ifm)%nyp, n5, 1, n4,  &
                    2+gd(ifm)%nstratx, gd(ifm)%nxp-1-gd(ifm)%nstratx, 1+nc1,  &
                    gd(ifm)%nyp-nc1, ip, a(iscr1), a(isbpsdensn(ifm)))
               call ae_sfc4D_4D(n4, gd(ifm)%nxp, gd(ifm)%nyp, n5, 1, n4, 1,  &
                    gd(ifm)%nxp, 1, gd(ifm)%nyp, ip, a(isbpsdensn(ifm)),  &
                    a(iscr1))

               ! SBPS thermal capacity:

               call fillscr_sfc4D_3D(n4, maxnxp, maxnyp, n4, gd(icm)%nxp,  &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(isbpstmcpn(icm)))
               call eintp_3Dsfc(a(iscr1), a(iscr2), n4, maxnxp, maxnyp, n4,  &
                    gd(ifm)%nxp, gd(ifm)%nyp, ifm, 0, 0)
               call fillvar_sfc3D_4D(n4, maxnxp, maxnyp, n4, gd(ifm)%nxp,  &
                    gd(ifm)%nyp, n5, ip, a(iscr2), a(iscr1))
               call ae_sfc4D_4D(n4, gd(ifm)%nxp, gd(ifm)%nyp, n5, 1, n4,  &
                    2+gd(ifm)%nstratx, gd(ifm)%nxp-1-gd(ifm)%nstratx, 1+nc1,  &
                    gd(ifm)%nyp-nc1, ip, a(iscr1), a(isbpstmcpn(ifm)))
               call ae_sfc4D_4D(n4, gd(ifm)%nxp, gd(ifm)%nyp, n5, 1, n4, 1,  &
                    gd(ifm)%nxp, 1, gd(ifm)%nyp, ip, a(isbpstmcpn(ifm)),  &
                    a(iscr1))
            
               ! Static surface albedo:
            
               call fillscr_sfc3D_2D(maxnxp, maxnyp, gd(icm)%nxp,  &
                    gd(icm)%nyp, n5, ip, a(iscr1), a(ialbdstatn(icm)))
               call eintp_2Dsfc(a(iscr1), a(iscr2), maxnxp, maxnyp,  &
                    gd(ifm)%nxp, gd(ifm)%nyp, ifm, 0, 0)
               call fillvar_sfc2D_3D(maxnxp, maxnyp, gd(ifm)%nxp,  &
                    gd(ifm)%nyp, n5, ip, a(iscr2), a(iscr1))
               call ae_sfc3D_3D(gd(ifm)%nxp, gd(ifm)%nyp, n5,  &
                    2+gd(ifm)%nstratx, gd(ifm)%nxp-1-gd(ifm)%nstratx, 1+nc1,  &
                    gd(ifm)%nyp-nc1, ip, a(iscr1), a(ialbdstatn(ifm)))
               call ae_sfc3D_3D(gd(ifm)%nxp, gd(ifm)%nyp, n5, 1, gd(ifm)%nxp,  &
                    1, gd(ifm)%nyp, ip, a(ialbdstatn(ifm)), a(iscr1))

            end do
            
         end if

      end do
   end if

   ! Calculate SBPS thermal conductivity from thermal inertia and volumetric
   !  heat capacity; check/repair albedo values:
   
   do ifm=ngra,ngrb
      call TI_init(ifm, gd(ifm)%nxp, gd(ifm)%nyp, n4, n5, nzg,  &
           a(itistatn(ifm)), a(isbpsdensn(ifm)), a(isbpstmcpn(ifm)),  &
           a(isbpstmcdn(ifm)))

      call albedo_init(gd(ifm)%nxp, gd(ifm)%nyp, n5, ifm, albedo,  &
           a(ialbdstatn(ifm)), a(iglatn(ifm)))
   end do

else if (vnam == 'NOFILE') then

   do ifm=ngra,ngrb
      icm = gd(ifm)%nxtnest

      ! Fill SBPS variables for which standard PRAMS datasets never exist:

      call sfcinit(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, nzg+nzs, npatch,  &
           nzg, ifm, vnam, a(ipatchfcvn(ifm)), a(isbpstmcpn(ifm)),  &
           a(isbpsdensn(ifm)), a(itistatn(ifm)), a(ialbdstatn(ifm)),  &
           a(isbpstmcdn(ifm)), a(isbpslthkn(ifm)), a(isbpsqn(ifm)),  &
           a(ialbedtn(ifm)), a(isfctempkn(ifm)), a(iz0netn(ifm)),  &
           a(iz0statn(ifm)), a(itopz0n(ifm)), a(ithetan(ifm)), a(ipi0n(ifm)),  &
           a(ippn(ifm)), a(iglatn(ifm)), a(iglonn(ifm)), a(itoptn(ifm)))

      ! If desired, override some current values:

      call sfcinit_user(gd(ifm)%nzp, gd(ifm)%nxp, gd(ifm)%nyp, nzg+nzs,  &
           npatch, nzg, ifm, vnam, a(iglatn(ifm)), a(iglonn(ifm)),  &
           a(itistatn(ifm)), a(ialbdstatn(ifm)))

   end do

end if

end subroutine surface_nest


!******************************************************************************

!%subroutine patch_minsize(n2,n3,n4,n5,schar)
!%
!%implicit none
!%integer :: n2,n3,n4,n5,i,j,ipatch,jpatch
!%
!%real :: orig_size
!%real, dimension(n2,n3,n4,n5) :: schar
!%
!%do j = 1,n3
!%   do i = 1,n2
!%      do ipatch = 2,n5
!%         if (schar(i,j,1,ipatch) .gt. 0. .and.  &
!%             schar(i,j,1,ipatch) .lt. .01) then
!%
!%            orig_size = schar(i,j,1,ipatch)
!%            schar(i,j,1,ipatch) = 0.
!%
!%            do jpatch = 1,n5
!%               if (jpatch .ne. ipatch) then
!%                  schar(i,j,1,jpatch) = schar(i,j,1,jpatch) / (1. - orig_size)
!%               endif
!%            enddo
!%         endif
!%      enddo
!%   enddo
!%enddo
!%
!%return
!%end
