!```````````````````````````````````````````````````````````````````
module PRAMS_SBPS_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine SBPS_gndice(igd, m1,m2,m3,m4,m5,m6,m7,ia,iz,ja,jz,dtSBPS,rshort  &
     ,rlong,rlongup,ustar,tstar,rstar,albedo_static,piocp_s,density_air_s  &
     ,Eflux,SBPS_dtdz,SBPS_tempK,vhc,subsfc_z,wptp,albedt,SBPS_sfcT  &
     ,SBPS_density,SBPS_thermcp,SBPS_thermcond,SBPS_lyrthkn,SBPS_Q  &
     ,net_atm_dmass,press_air_s,iter,rtgt,theta_air_s,theta_air,pp,pi0,  &
     bottom_other_sflux_down, bottom_other_sflux_up, emisst,  &
     emissivity_static, bottom_SWflux_down, bottom_SWflux_up, topo_factor_sflux)

!+ SBPS thermal energy transfer model (with/without burden of CO2 ice)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use univ_phys_constants_mod, only: stfnbltz_f4
use univ_exception_mod, only: exception, FATAL
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use main_memory_mod, only: SBPS_Tlev, c_gd, RT_paradigm, undef_value_f4
use main_memory_bspec_mod, only: CO2_sfci
use body_specific_constants_mod, only: cp, p00, cpor, R_air, cpi, rho_CO2ice,  &
     lth_CO2ice, emiss_CO2ice

implicit none

!=== Argument declarations:

! m1 = mzp, m2 = mxp, m3 = myp, m4 = nzg+nzs, m5 = npatch, m6 = nzg, m7 = nzs
integer, intent(IN) :: igd, m1,m2,m3,m4,m5,m6,m7,ia,iz,ja,jz,iter
real(f4), dimension(m6), intent(IN) :: subsfc_z
real(f4), dimension(m2,m3), intent(IN) :: piocp_s,density_air_s,rlong,wptp  &
     ,press_air_s,rtgt,theta_air_s, bottom_other_sflux_down,  &
     topo_factor_sflux, bottom_SWflux_down
real(f4), dimension(m1,m2,m3), intent(IN) :: theta_air,pp,pi0
real(f4), dimension(m2,m3,m5), intent(IN) :: albedo_static,ustar,tstar,rstar,  &
     emissivity_static
real(f4), intent(IN) :: dtSBPS

real(f4), dimension(m2,m3), intent(INOUT) :: albedt, net_atm_dmass, emisst
real(f4), dimension(m2,m3,m5), intent(INOUT) :: SBPS_sfcT
real(f4), dimension(m4,m2,m3,m5), intent(INOUT) :: SBPS_density,SBPS_thermcp  &
     ,SBPS_thermcond,SBPS_lyrthkn,SBPS_Q

real(f4), dimension(m4), intent(INOUT) :: SBPS_dtdz, vhc, SBPS_tempK, Eflux
real(f4), dimension(m2,m3), intent(INOUT) :: rlongup, bottom_other_sflux_up,  &
     rshort, bottom_SWflux_up

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'SBPS_gndice'

real(f4), parameter :: icemass_tol = 4.e-6_f4  ! [kg] Equivalent to ~2.4 nm
                                                  ! of CO2 ice (1 m^2 area) --
                                                  ! about 3 times the CO2
                                                  ! molecule diameter (~0.8nm)

real(f4), parameter :: sinksrc_tempk = 170._f4 ! [K] Temperature of "deep"
                                                  ! subsurface energy
                                                  ! sink/source layer

integer, parameter :: SBPS_paradigm = 1

integer :: k,i,j,ip,SBPS_case
real(f4) :: htflux,downradflux,Tsat,Ediff,cp_ice,dmass  &
     ,dti,icemass,Qi_new,c1,c2,fval,fval_p,p_tol,p_0  &
     ,p_new,p_2,t_2

!=== Executable statements:

!-----------------------------------------------------------------------------
!           %atmosphere%
! Kf= n6     -----------  <--sfc           NO SURFACE ICE CASE
! (Kq= n6)   ===========  
! Kf= n6-1   -----------  -
! (Kq= n6-1) ===========  | <-- SBPS layer       z='up'
! Kf= n6-2   -----------  -
! (Kq= n6-2) ===========
! Kf= 1      -----------
! (Kq= 1)    ===========  <--deep subsfc sink/source
!         %to center of Mars%
!
!
!           %atmosphere%
! Kf= n4     -----------  <--sfc           SURFACE ICE CASE
! (Kq= n4)   ===========  
! Kf= n4-1   -----------   -
! (Kq= n4-1) ===========   | <-- SBPS layer     z='up'
! Kf= n4-2   -----------   -
! (Kq= n4-2) ===========
! Kf= 1      -----------
! (Kq= 1)    ===========  <--deep subsfc sink/source
!         %to center of Mars%
!
!
! Kq = thermal (heat) storage quantity level (===========);  Variables:
!
!      SBPS_dtdz = [m4],[s/m]; ratio of SBPS timestep to SBPS layer thickness
!      vhc = [m4],[J/(K.m^3)]; volumetric thermal (heat) capacity
!      SBPS_tempK = [m4],[K]; SBPS layer temperature
!
!      SBPS_Q = [m4,m2,m3,m5],[J/(m^3)]; thermal (heat) energy storage
!
!         *IMPORTANT*: Q is the amount of internal (thermal) energy in a
!                      cubic volume of 1 m^3, NOT the energy in the volume of
!                      each SBPS layer!
!
!      SBPS_lyrthkn = [m4,m2,m3,m5],[m]; thickness of SBPS layer
!      SBPS_Tlev = [m4],[m]; depth/height of SBPS layer midpoint
!
! Kf = thermal (heat) flux level (-----------);  Variables:
!     
!      SBPS_density = [m4,m2,m3,m5],[kg/(m^3)]; density         
!      SBPS_thermcp = [m4,m2,m3,m5],[J/(kg.K)]; thermal (heat) capacity
!      Eflux = [m4],[J/(s.m^2)]; thermal (heat) energy flux
!      SBPS_thermcond = [m4,m2,m3,m5],[J/(s.m.K)]; thermal (heat) conductivity
!-----------------------------------------------------------------------------

ip = 1    ! Hardwire patch number for now

! This calc done here for now since SBPS_lyrthkn doesn't YET vary in horizontal
do k=2,m6   ! Valid at Kq 
   SBPS_dtdz(k) = dtSBPS/SBPS_lyrthkn(k,1,1,ip)  ! ratio of dt to SBPS layer dz
enddo

dti = ONE_f4/dtSBPS    ! Invert timestep

if(CO2_sfci == 1 .or. CO2_sfci == 3) then       ! With STATIC CO2 ice

   do j=ja,jz     ! Loop over all horizontal grid points
      do i=ia,iz

         ! Convert heat storage quantity to temperature (K):

         do k=2,m6    ! Valid at Kq
            ! Volumetric heat capacity of SBPS layer:
!#            vhc(k) = (SBPS_density(k-1,i,j,ip)*SBPS_thermcp(k-1,i,j,ip)+  & 
!#                 SBPS_density(k,i,j,ip)*SBPS_thermcp(k,i,j,ip))*0.5_f4
            vhc(k) = SBPS_density(k,i,j,ip)*SBPS_thermcp(k,i,j,ip)
            ! Temperature [K] of SBPS layer
            SBPS_tempK(k) = SBPS_Q(k,i,j,ip)/vhc(k)
         enddo
         vhc(1) = SBPS_density(1,i,j,ip)*SBPS_thermcp(1,i,j,ip)
         SBPS_tempK(1) = 170._f4    ! Bottom LAYER temperature constant at 170K

         ! Compute surface energy flux BC:

         ! Upward fluxes of EM radiation at the bottom (surface):
         if (RT_paradigm /= 0) then
            ! *Assumes* an "effective" surface emitting area unweighted by
            !  topographic slope (i.e., horizontal and flat):
            rlongup(i,j) = emisst(i,j)*stfnbltz_f4*(SBPS_sfcT(i,j,ip)**4)

            ! These are assumed to be simple surface reflections of the
            !  downwelling stellar flux:
            bottom_other_sflux_up(i,j) = max(ZERO_f4,  & ! use "LW" albedo
                 (ONE_f4-emisst(i,j))*bottom_other_sflux_down(i,j))
            bottom_SWflux_up(i,j) = albedt(i,j)*bottom_SWflux_down(i,j)
         end if

         ! Compute sensible energy (heat) flux from/to surface:
         !    (Note that below, WPTP is <w'theta'>; 
         !    converted to w'T' with piocp_s)
         htflux = cp*density_air_s(i,j)*wptp(i,j)*piocp_s(i,j)

         ! The downward stellar-derived flux (SW + other_sflux) *absorbed* by
         !  the surface (optionally corrected for topographic slope and
         !  azimuth):
         rshort(i,j) =  max(ZERO_f4,  &
              ((bottom_SWflux_down(i,j)-bottom_SWflux_up(i,j))+  &
              (bottom_other_sflux_down(i,j)-bottom_other_sflux_up(i,j)))*  &
              topo_factor_sflux(i,j))

         ! Net downward-directed radiative fluxes from the atmosphere:
         downradflux = rshort(i,j)+rlong(i,j)

         ! Net energy flux at surface (bare, ice-free surface)
         Eflux(m6) = htflux+rlongup(i,j)-downradflux

         icemass = SBPS_lyrthkn(m4,i,j,ip)*rho_CO2ice  ! Mass [kg] of SPS ice
         
         if(icemass < icemass_tol) then   ! Bare surface

            if (SBPS_paradigm == 1) then

               ! Compute subsurface energy fluxes:
         
               do k=1,m6-1   ! Valid at subsurface Kf

                  ![SBPS conductivity at Kf]*[-dT/dz across Kf]: 
                  Eflux(k) = SBPS_thermcond(k,i,j,ip)*  &
                       ((SBPS_tempK(k)-SBPS_tempK(k+1))/  &
                       (SBPS_Tlev(k+1)-SBPS_Tlev(k)))
               enddo

               ! Update subsurface thermal energy storage quantity:

               ! Stays constant:     SBPS_Q(1,i,j,ip) = SBPS_Q(1,i,j,ip)
         
               ! [Old SBPS_Q value]+[energy flux divergence]*[dt/dz] 

               do k=2,m6
                  SBPS_Q(k,i,j,ip) = SBPS_Q(k,i,j,ip)+  &
                       (Eflux(k-1)-Eflux(k))*SBPS_dtdz(k)
               enddo

            end if

            SBPS_sfcT(i,j,ip) = SBPS_Q(m6,i,j,ip)/vhc(m6)  ! Set new sfc temp

         else     ! CO2 ice on ground    

            Tsat = Tsat_CO2ice(press_air_s(i,j))   ! CO2 saturation temp [K]

            SBPS_sfcT(i,j,ip) = Tsat   ! Peg ice sfc temperature at Tsat
      
         endif

      enddo
   enddo

elseif(CO2_sfci == 2) then      ! With prognostic CO2 ice

   do j=ja,jz     ! Loop over all horizontal grid points
      do i=ia,iz

         ! Convert SBS heat storage quantity to temperature (K):

         do k=2,m6    ! Valid at Kq
            ! Volumetric heat capacity of SBPS layer:
!#            vhc(k) = (SBPS_density(k-1,i,j,ip)*SBPS_thermcp(k-1,i,j,ip)+  & 
!#                 SBPS_density(k,i,j,ip)*SBPS_thermcp(k,i,j,ip))*0.5_f4
            vhc(k) = SBPS_density(k,i,j,ip)*SBPS_thermcp(k,i,j,ip)
            ! Temperature [K] of SBPS layer
            SBPS_tempK(k) = SBPS_Q(k,i,j,ip)/vhc(k)
         enddo
         vhc(1) = SBPS_density(1,i,j,ip)*SBPS_thermcp(1,i,j,ip)
         SBPS_tempK(1) = sinksrc_tempk   ! Bottom LAYER temperature constant

         Tsat = Tsat_CO2ice(press_air_s(i,j)) ! Current CO2 saturation temp [K]
         cp_ice = cp_CO2ice(Tsat)  ! Current CO2 ice heat capacity
         icemass = SBPS_lyrthkn(m4,i,j,ip)*rho_CO2ice   ! Mass [kg] of SPS ice

         if(icemass > ZERO_f4) then   ! If ice exists on ground

            Qi_new = SBPS_density(m4,i,j,ip)*Tsat*cp_ice ! New Q of sfc ice

            if(iter == 1) then    ! Only do once per atmospheric timestep

               !--------------------------------------------------------------
               ! Changes in "surface" pressure have now caused Tsat and cp_ice
               !  to change, but SBPS_Q has not, and thus (*IF* the CO2 ice
               !  temperature is assumed to always be the same as the current
               !  Tsat) there is an excess/deficit of thermal energy that must
               !  be removed from the CO2 ice before continuing any further.
               ! 
               ! Since the pressure change occurs at the air/ice interface, we
               !  choose here to redistribute thermal energy by
               !  sublimating/depositing CO2 mass to/from the atmosphere,
               !  from/to the ice.
               !
               ! However, the thermal energy of the ice is so *very* strongly
               !  coupled to the air pressure that a simple "brute force"
               !  method (i.e., how much mass corresponds to needed change in
               !  ice thermal energy?; then sublimate/deposit that much mass)
               !  results in a quickly amplifying oscillation of the model
               !  solution that quickly brings down PRAMS.  It is surmised that
               !  given a small enough (likely very small) *<dynamic>*
               !  timestep, this would not occur.
               !
               ! Since we lack the luxury of being able to use such a small
               !  dynamical timestep, an iterative "equilibrium" solution to
               !  the above-mentioned thermal energy redistribution is used.
               !--------------------------------------------------------------
               
               ! CO2 ice thermal energy discrepancy between old and new
               !  (Tsat and cp); due to changes in ambient pressure due to
               !  dynamical and other physical processes:
               Ediff = Qi_new-SBPS_Q(m4,i,j,ip) 

               c1 = SBPS_Q(m4,i,j,ip)/SBPS_density(m4,i,j,ip)
               c2 = Tsat*cp_ice

               ! Evade a precision issue: c1 and c2 have values <on Mars> of
               !  about 150000.0 [J/kg], so if one "trusts" single-precision
               !  (32-bit) floating point calcs to 7 significant digits,
               !  then if the absolute difference between c1 and c2 is less
               !  than ~0.1, one can't reliably iterate using the NRI
               !  method (solution does not converge) -- instead, use
               !  "brute-force" (non-equilibrium) technique.
 
               if(abs(c1-c2) > 9.9e-2_f4) then   ! Use Newton-Raphson iteration

                  ! Make a guess at the final "equilibrium" pressure;
                  !  arbitrary, but based on the qualitative observation
                  !  that the absolute change in pressure to arrive at
                  !  the "equilibrium" state is less than that needed
                  !  to reach the brute force "non-equilibrium" state:

                  dmass = Ediff/lth_CO2ice   ! Delta-ice_mass (brute force)
                  p_2 = (pp(2,i,j)+pi0(2,i,j))*cpi   ! full_PI/cp (for now)
                  t_2 = theta_air(2,i,j)*p_2    ! Air temperature (at k=2)
                  p_2 = p00*p_2**cpor    ! Air pressure (at k=2)

                  p_0 = (p_2/(R_air*t_2)-dmass/rtgt(i,j)*c_gd%dzt(2))*  &
                       R_air*t_2    ! Resultant air pressure (brute force)

               
!^                  p_upb = p_0
!^                  print *,press_air_s(i,j),p_0,p_2,t_2,dmass,c1-c2
!^                  pause
                  p_0 = 0.08_f4*p_0+0.92_f4*press_air_s(i,j)   ! Initial guess
                  if(p_0 < press_air_s(i,j)*0.1_f4) then
                     p_0 = press_air_s(i,j)*0.1_f4   ! Don't let guess get near 0.
                  endif

                  ! Pressure value tolerance for terminating iteration [the
                  !  below factor is rather arbitrary (2.5e-6 produces
                  !  p_tol=1.e-3 at 400 Pa) -- this is probably near the limit
                  !  of 32-bit floating point precision, anyway; if set to
                  !  2.5e-5, produces usable results, but "choppy" -- a value
                  !  of 2.5e-6 yields a much "smoother" result, at a fairly
                  !  small computational cost]:
                  p_tol = 2.5e-6_f4*press_air_s(i,j)
               
                  Tscp_itlp: do k=1,50
                     
                     ! Calculate function value, etc.:
                     call Tscpfunc_stuff(p_0,c1,Tsat,cp_ice,fval,fval_p)
                        
                     p_new = p_0-fval/fval_p 
!^                     print *,'iter',k,fval,fval_p,fval/fval_p,p_tol,p_new  &
!^                          ,p_0,Tsat*cp_ice,c1
                     if(abs(p_new-p_0) < p_tol) then
                        exit Tscp_itlp    ! Iteration complete; exit loop
                     endif
                        
                     p_0 = p_new    ! Replace first guess
                        
                  enddo Tscp_itlp

                  if(k == 51) then
                     write (*,'(/a/)')  &
                  'WARNING: Tsat/cp_ice E-adjustment iteration not converging.'
                  endif

                  Tsat = Tsat_CO2ice(p_new)
                  cp_ice = cp_CO2ice(Tsat)
                  Qi_new = SBPS_density(m4,i,j,ip)*Tsat*cp_ice ! New sfc ice Q
                  Ediff = Qi_new-SBPS_Q(m4,i,j,ip) ! Energy discrepancy b/w old
                                                   !  and new Tsat

!^                  print *,'-----',k,dmass,Ediff/lth_CO2ice,p_new,p_0,p_upb  &
!^                       ,press_air_s(i,j),Tsat,cp_ice
            
               endif
                  
               dmass = Ediff/lth_CO2ice   ! Amount of ice lost/gained
               net_atm_dmass(i,j) = net_atm_dmass(i,j)-dmass ! To atm
               icemass = icemass+dmass       ! Adjust ice mass accordingly
            endif

            if(icemass <= ZERO_f4) then   ! Ice depleted ENTIRELY

               SBPS_lyrthkn(m4,i,j,ip) = ZERO_f4       ! There is no ice now
               SBPS_Q(m4,i,j,ip) = ZERO_f4             !
               SBPS_thermcp(m4,i,j,ip) = undef_value_f4    !
               SBPS_thermcond(m4,i,j,ip) = undef_value_f4  !
               SBPS_density(m4,i,j,ip) = undef_value_f4    !
               ! Change surface albedo to that of bare ground:
               albedt(i,j) = albedo_static(i,j,ip)
               emisst(i,j) = emissivity_static(i,j,ip)
               SBPS_sfcT(i,j,ip) = SBPS_tempK(m6)  ! Set to topmost SBS tempk
               SBPS_case = 0

            elseif(icemass < icemass_tol) then   ! Trace ice; will be removed

               SBPS_Q(m6,i,j,ip) = SBPS_Q(m6,i,j,ip)-  &  ! Remove energy from
                    lth_CO2ice*icemass                    !  subsurface
               SBPS_tempK(m6) = SBPS_Q(m6,i,j,ip)/vhc(m6)
               SBPS_sfcT(i,j,ip) = SBPS_tempK(m6)  ! Set to topmost SBS tempk
               net_atm_dmass(i,j) = net_atm_dmass(i,j)+icemass ! Add kg to atm
               icemass = ZERO_f4                       !
               SBPS_lyrthkn(m4,i,j,ip) = ZERO_f4       ! There is no ice now
               SBPS_Q(m4,i,j,ip) = ZERO_f4             !
               SBPS_thermcp(m4,i,j,ip) = undef_value_f4    !
               SBPS_thermcond(m4,i,j,ip) = undef_value_f4  !
               SBPS_density(m4,i,j,ip) = undef_value_f4    !
               ! Change surface albedo to that of bare ground:
               albedt(i,j) = albedo_static(i,j,ip)
               emisst(i,j) = emissivity_static(i,j,ip)
               SBPS_case = 0

            else                      ! Some ice still left

               ! New depth of ice layer ( Implicitly / by 1 m^2 ):
               SBPS_lyrthkn(m4,i,j,ip) = icemass/SBPS_density(m4,i,j,ip) 
               SBPS_Q(m4,i,j,ip) = Qi_new     ! New Q of sfc ice
               SBPS_tempK(m4) = Tsat    ! Set ice temp
               SBPS_sfcT(i,j,ip) = Tsat   ! Set ice sfc temp
               SBPS_thermcp(m4,i,j,ip) = cp_ice    ! New cp of sfc ice
               ! New CO2 ice thermal conductivity (using TES ice TI ~= 600):
               SBPS_thermcond(m4,i,j,ip) = 3.6e5_f4/  &
                    (SBPS_density(m4,i,j,ip)*SBPS_thermcp(m4,i,j,ip))
               SBPS_case = 1

            endif

         else        ! No ice

            SBPS_case = 0

         endif

         SBPS_icemode: select case (SBPS_case)

         case (0)        ! Ice-free surface
            
!            print *,'BARE SFC'
            
            !*** Compute surface energy flux BC:
            
            ! Upward fluxes of EM radiation at the bottom (surface):
            if (RT_paradigm /= 0) then
                  ! *Assumes* an "effective" surface emitting area unweighted by
                  !  topographic slope (i.e., horizontal and flat); uses
                  !  the previous surface temperature:
               rlongup(i,j) = emisst(i,j)*stfnbltz_f4*(SBPS_sfcT(i,j,ip)**4)
                  ! These are assumed to be simple surface reflections of the
                  !  downwelling stellar flux:
               bottom_other_sflux_up(i,j) = max(ZERO_f4,  & ! use "LW" albedo
                    (ONE_f4-emisst(i,j))*bottom_other_sflux_down(i,j))
               bottom_SWflux_up(i,j) = albedt(i,j)*bottom_SWflux_down(i,j) 
            end if

            ! Compute sensible energy (heat) flux from/to surface:
            !    (Note that below, WPTP is <w'theta'>; 
            !    converted to w'T' with piocp_s)
            htflux = cp*density_air_s(i,j)*wptp(i,j)*piocp_s(i,j)

            ! The downward stellar-derived flux (SW + other_sflux) *absorbed* by
            !  the surface (optionally corrected for topographic slope and
            !  azimuth):
            rshort(i,j) =  max(ZERO_f4,  &
                 ((bottom_SWflux_down(i,j)-bottom_SWflux_up(i,j))+  &
                 (bottom_other_sflux_down(i,j)-bottom_other_sflux_up(i,j)))*  &
                 topo_factor_sflux(i,j))

            ! Net downward-directed radiative fluxes from the atmosphere:
            downradflux = rshort(i,j)+rlong(i,j)

            ! Net energy flux at surface (bare, ice-free surface)
            Eflux(m6) = htflux+rlongup(i,j)-downradflux

            !*** Compute subsurface energy fluxes:
            
            do k=1,m6-1   ! Valid at subsurface Kf
               
               ![SBPS conductivity at Kf]*[-dT/dz across Kf]: 
               Eflux(k) = SBPS_thermcond(k,i,j,ip)*  &
                    ((SBPS_tempK(k)-SBPS_tempK(k+1))/  &
                    (SBPS_Tlev(k+1)-SBPS_Tlev(k)))

            enddo
            
            !*** Update subsurface thermal energy storage quantity:
            
            ! Stays constant:     SBPS_Q(1,i,j,ip) = SBPS_Q(1,i,j,ip)
            
            ! [Old SBPS_Q value]+[energy flux divergence]*[dt/dz] 
            
            do k=2,m6
               SBPS_Q(k,i,j,ip) = SBPS_Q(k,i,j,ip)+  &
                    (Eflux(k-1)-Eflux(k))*SBPS_dtdz(k)
            enddo
            
            !*** Check for CO2 frost deposition:
            
            SBPS_tempK(m6) = SBPS_Q(m6,i,j,ip)/vhc(m6)  ! New top SBS temp [K]

            if(SBPS_tempK(m6) < Tsat) then ! Surface colder than CO2 saturation
               
!               print *,'frosting sfc'
               
               ! Energy required to warm upper SBPS layer to Tsat:
               Ediff = SBPS_lyrthkn(m6,i,j,ip)*vhc(m6)*(Tsat-SBPS_tempK(m6))
               
               ! Mass of ice deposited:
               dmass = Ediff/lth_CO2ice   

               ! Subtract mass from the overlying atmosphere:
               net_atm_dmass(i,j) = net_atm_dmass(i,j)-dmass

               ! Add new ice mass to total:
               icemass = icemass+dmass

               ! Change density of Kf to that of CO2 ice:
               SBPS_density(m4,i,j,ip) = rho_CO2ice 

               ! Depth of ice layer ( Implicitly / by 1 m^2 ):
               SBPS_lyrthkn(m4,i,j,ip) = icemass/SBPS_density(m4,i,j,ip)
               
               ! Change surface albedo to that of CO2 ice:
               albedt(i,j) = albedo_CO2ice(SBPS_lyrthkn(m4,i,j,ip)  &
                    ,albedo_static(i,j,ip))
               emisst(i,j) = emiss_CO2ice

               SBPS_sfcT(i,j,ip) = Tsat  ! Set sfc temp for use elsewhere

               ! Set thermal capacity of CO2 ice:
               SBPS_thermcp(m4,i,j,ip) = cp_ice

               ! CO2 ice thermal conductivity (using ice TI ~= 600 from TES):
               SBPS_thermcond(m4,i,j,ip) = 3.6e5_f4/  &
                    (SBPS_density(m4,i,j,ip)*SBPS_thermcp(m4,i,j,ip))

               ! Set amount of thermal (non-latent) energy in ice:
               SBPS_Q(m4,i,j,ip) =  Tsat*  &
                    SBPS_density(m4,i,j,ip)*SBPS_thermcp(m4,i,j,ip)

               ! Latent heat released by CO2 deposition warms ground surface:
               !   (Added quantity implicitly * by 1 m^2)
               SBPS_Q(m6,i,j,ip) = SBPS_Q(m6,i,j,ip)+  &
                    Ediff/SBPS_lyrthkn(m6,i,j,ip)
               
            else    ! Surface is warmer than CO2 saturation temperature
               
               !print *,'no frosting'
               
               SBPS_sfcT(i,j,ip) = SBPS_tempK(m6)   ! Set sfc temperature
               
            endif

         case (1)        ! Significant CO2 ice on ground
            
!            print *,'ICE SFC'

            ! Set this better at some point?????
            SBPS_Tlev(m4) = subsfc_z(m6)+SBPS_lyrthkn(m4,i,j,ip)*0.5_f4
            
            do k=m6,m6+m7-1   ! Valid at Kf
               Eflux(k) = SBPS_thermcond(k,i,j,ip)*  &
                    ((SBPS_tempK(k)-SBPS_tempK(k+1))/  &
                    (SBPS_Tlev(k+1)-SBPS_Tlev(k)))
            enddo
            
            !*** Compute subsurface energy fluxes:
            
            do k=1,m6-1   ! Valid at subsurface Kf
               
               ![SBPS conductivity at Kf]*[-dT/dz across Kf]: 
               Eflux(k) = SBPS_thermcond(k,i,j,ip)*  &
                    ((SBPS_tempK(k)-SBPS_tempK(k+1))/  &
                    (SBPS_Tlev(k+1)-SBPS_Tlev(k)))
            enddo
            
            !*** Compute surface energy flux BC:
            
            ! Upward fluxes of EM radiation at the bottom (surface):
            if (RT_paradigm /= 0) then
                  ! *Assumes* an "effective" surface emitting area unweighted by
                  !  topographic slope (i.e., horizontal and flat):
               rlongup(i,j) = emisst(i,j)*stfnbltz_f4*(SBPS_tempK(m4)**4) 
                  ! These are assumed to be simple surface reflections of the
                  !  downwelling stellar flux:
               bottom_other_sflux_up(i,j) = max(ZERO_f4,  & ! use "LW" albedo
                    (ONE_f4-emisst(i,j))*bottom_other_sflux_down(i,j))
               bottom_SWflux_up(i,j) = albedt(i,j)*bottom_SWflux_down(i,j)
            end if

            ! Compute sensible energy (heat) flux from/to surface:
            !    (Note that below, WPTP is <w'theta'>; 
            !    converted to w'T' with piocp_s)
            htflux = cp*density_air_s(i,j)*wptp(i,j)*piocp_s(i,j)

            ! The downward stellar-derived flux (SW + other_sflux) *absorbed* by
            !  the surface (optionally corrected for topographic slope and
            !  azimuth):
            rshort(i,j) =  max(ZERO_f4,  &
                 ((bottom_SWflux_down(i,j)-bottom_SWflux_up(i,j))+  &
                 (bottom_other_sflux_down(i,j)-bottom_other_sflux_up(i,j)))*  &
                 topo_factor_sflux(i,j))

            ! Net downward-directed radiative fluxes from the atmosphere:
            downradflux = rshort(i,j)+rlong(i,j)
           
            ! Net energy flux at surface of ice
            Eflux(m4) = htflux+rlongup(i,j)-downradflux
            
            !*** Update SBPS heat storage quantity:
            
            ! Stays constant:      SBPS_Q(1,i,j,ip) = SBPS_Q(1,i,j,ip)
            
            ! [Old SBPS_Q]+[energy flux divergence]*[dt/dz] 
            
            do k=2,m6  ! Valid at Kq
               SBPS_Q(k,i,j,ip) = SBPS_Q(k,i,j,ip)+  &
                    (Eflux(k-1)-Eflux(k))*SBPS_dtdz(k)
            enddo
            
            ! Net energy (J) imparted to ice during surface timestep:
            Ediff = (Eflux(m4)-Eflux(m6))*dtSBPS  
            
            dmass = icemass*lth_CO2ice ! Energy needed to sublimate all ice
            
            if(dmass <= Ediff) then  ! All ice sublimates

               !print *,'ice melts'

               ! Warm surface with remaining energy
               SBPS_Q(m6,i,j,ip) = SBPS_Q(m6,i,j,ip)+(Ediff-dmass)/  &
                    SBPS_lyrthkn(m6,i,j,ip)

               ! Add mass to the overlying atmosphere:
               net_atm_dmass(i,j) = net_atm_dmass(i,j)+icemass

               icemass = 0.  ! Set to insignificant ice cover
               SBPS_lyrthkn(m4,i,j,ip) = ZERO_f4       ! 
               SBPS_Q(m4,i,j,ip) = ZERO_f4             !
               SBPS_thermcp(m4,i,j,ip) = undef_value_f4    ! No ice
               SBPS_thermcond(m4,i,j,ip) = undef_value_f4  !
               SBPS_density(m4,i,j,ip) = undef_value_f4    !
               ! Change surface albedo to that of bare ground:
               albedt(i,j) = albedo_static(i,j,ip)
               emisst(i,j) = emissivity_static(i,j,ip)

               SBPS_sfcT(i,j,ip) = SBPS_Q(m6,i,j,ip)/vhc(m6) ! Set sfc temp [K]
               
            else  ! Partial sublimation, accumulation, or no change
               
               !print *,'ice stays'
               
               dmass = -Ediff/lth_CO2ice
               icemass = icemass+dmass

               ! Add/subtract mass to/from the overlying atmosphere:
               net_atm_dmass(i,j) = net_atm_dmass(i,j)-dmass       
                          
               ! Implicitly / by 1 m^2 in below calc:
               SBPS_lyrthkn(m4,i,j,ip) = icemass/rho_CO2ice 

!               SBPS_Q(m4,i,j,ip) = Qi_new     ! 
!               SBPS_thermcp(m4,i,j,ip) = ?    ! Set at beginning, so no need
!               SBPS_thermcond(m4,i,j,ip) = ?  !  to reset them
!               SBPS_density(m4,i,j,ip) = ?    !
!               SBPS_sfcT(i,j,ip) = Tsat       !
               
            endif
            
         case default     ! Someone made a coding error... Tsk, tsk...
      
            call exception('SBPS_gndice', FATAL,  &
                 'Desired SBPS_icemode case does not exist.')

         end select SBPS_icemode

      enddo
   enddo

end if

end subroutine SBPS_gndice


!```````````````````````````````````````````````````````````````````
function cp_CO2ice(ice_tempk) result(cp_ice)

!+ Calculates CO2 ice heat capacity at a given temperature

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

real(f4) :: cp_ice   ! specific heat of CO2 ice [J/(kg.K)]

real(f4), intent(IN) :: ice_tempk    ! Temperature of CO2 ice [K]

!== Local declarations:

! NONE

!=== Executable statements:

! From Washburn 1948, valid for 73K < T < 200K
cp_ice = 349._f4+4.8_f4*ice_tempk  

end function cp_CO2ice


!```````````````````````````````````````````````````````````````````
function albedo_CO2ice(ice_depth,baresfc_albedo) result(albedo_ice)

!+ Calculates CO2 ice heat capacity at a given temperature

use univ_kind_defs_mod, only: f4
use body_specific_constants_mod, only: albd_CO2ice

implicit none

!=== Argument declarations:

real(f4) :: albedo_ice   ! Effective visible albedo of CO2 ice [none]

real(f4), intent(IN) :: ice_depth  &    ! Depth of CO2 ice [m]
           ,baresfc_albedo   ! Albedo of bare ("static") surface

!== Local declarations:

real(f4) :: x

! Factor that determines approximate depth that will have an 
!  albedo /= albd_CO2ice.
!  Example: maxfact = 500., baresfc_albedo = 0.3 ==> max_diff_depth ~0.001m
real(f4), parameter :: maxfact = 500._f4

!=== Executable statements:

x = maxfact*ice_depth

! Cubic Taylor series approximation of exp(x):
x = 1. + x + 0.5_f4*x*x + 0.166666667_f4*x*x*x 

albedo_ice = min(baresfc_albedo*x,albd_CO2ice)  ! Clip to <= 'albd_CO2ice'

end function albedo_CO2ice


!```````````````````````````````````````````````````````````````````
function Tsat_CO2ice(amb_press) result(Tsat_ice)

!+ Calculates CO2 saturation temperature at a given pressure (mb)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use body_specific_constants_mod, only: lth_CO2ice

implicit none

! {intent IN}:

real(f4) :: Tsat_ice  ! saturation temperature [K] for CO2 ice/vapor

real(f4), intent(IN) :: amb_press       ! Ambient pressure [Pa]

!== Local declarations:

! NONE

!=== Executable statements:

! From Clausius-Clapeyron relation for a perfect gas 
!   (Hourdin 1995; James et al., 1992)

Tsat_ice = ONE_f4/((-189._f4/lth_CO2ice*log(amb_press*0.01_f4))+  &
     (ONE_f4/136.3_f4))

! Mars book?
!Tsat_CO2ice = 3182.48/(23.3494-log(3.))

end function Tsat_CO2ice


!```````````````````````````````````````````````````````````````````
subroutine Tscpfunc_stuff(amb_press,E_i,Tsat,cp_ice,fval,ddp_fval)

!+ Returns value of the first derivative of Tsat/cp energy adjustment function

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use body_specific_constants_mod, only: lth_CO2ice

implicit none

!=== Argument declarations:

real(f4), intent(IN) :: amb_press  &      ! Ambient pressure [Pa]
                 ,E_i   ! Initial thermal E/unit_mass of CO2 ice layer [J/kg] 

! {intent OUT}:

real(f4), intent(OUT) :: Tsat  &   ! saturation temperature [K] for CO2 ice/vapor
                 ,cp_ice  &  ! specific heat of CO2 ice (ice temperature
                             !  assumed to be Tsat) [J/(kg.K)]
                 ,fval  &  ! Value of function [J/kg]
                 ,ddp_fval   ! First derivative of function [(J/kg)/Pa]

!== Local declarations:

! Function to differentiate  with respect to pressure is:
!       [ E_i-Tsat(p)*cp_ice(Tsat) ],
!  where E_i is the initial thermal energy of the CO2 ice (per unit mass),
!  Tsat(p) is the saturation temperature for CO2 ice/vapor (function of
!  amb_press), and cp_ice(Tsat) is the specific heat of CO2 ice (function of
!  ice temperature, in this case Tsat, and thus a function of amb_press as
!  well). 
!
! Analytical first derivative expression obtained using `Derive` computer
!  algebra system (CAS) software.

! Tsat formula constants [from Clausius-Clapeyron relation for a perfect gas 
!  (Hourdin 1995; James et al., 1992)]:

real(f4), parameter :: a = -189._f4/lth_CO2ice, b = ONE_f4/136.3_f4

! cp_ice formula constants (from Washburn 1948, valid for 73K < T < 200K):

real(f4), parameter :: c = 349._f4, d = 4.8_f4 

real(f4) :: p_mb,log_p,tmp1,tmp2

!=== Executable statements:

! Tsat formula is cast with amb_press in units of [mb], so convert units:

p_mb = amb_press*0.01_f4    ! [Pa] -> [mb]

log_p = log(p_mb)

! Compute first derivative value (at pressure = p_mb):

tmp1 = a*log_p+b

Tsat = ONE_f4/tmp1    ! Saturation temperature of CO2 ice/vapor

tmp2 = a*(a*c*log_p+b*c+2._f4*d)/(p_mb*(tmp1**3))   ! d/dp(func) in [(J/kg)/mb]
ddp_fval = tmp2*0.01_f4      ! Convert [(J/kg)/mb] -> [(J/kg)/Pa]

cp_ice = c+d*Tsat   ! Specific heat of CO2 ice

fval = E_i-Tsat*cp_ice   ! Function value (at presure = amb_press)

end subroutine Tscpfunc_stuff


!```````````````````````````````````````````````````````````````````
subroutine SBPS_Qprcadj(m2,m3,m4,m5,ia,iz,ja,jz,Q,Qp)

!+ Checks to see if energy in Qp needs emptied into Q (performed due to
!   precision issues with the Q array); if yes, then performs the task

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: wfreq_state_file, current_sim_ET, maximum_sim_ET, gd

implicit none

!=== Argument declarations:
 
integer, intent(IN) :: m4,m2,m3,m5,ia,iz,ja,jz

real(f4), dimension(m4,m2,m3,m5), intent(INOUT) :: Q,Qp 

!== Local declarations:

real(f4), parameter :: maxthresh = 99999._f4 ! [] -- maximum allowable value
                                                ! for Qp *before* energy is put
                                                ! in Q array.

real(f4), parameter :: minthresh = 1000._f4  ! [] -- minimum allowable value
                                                ! for Qp *after* energy is put
                                                ! in Q array.
 
integer :: i,j,k,ip
real(f4) :: Qprime

!=== Executable statements:

do ip=1,m5
   do j=ja,jz
      do i=ia,iz
         do k=1,m4

            Qprime = Qp(k,i,j,ip)
            if(Qprime > maxthresh .or.  &
                 (mod(real(current_sim_ET), wfreq_state_file) <  &
                 gd(1)%dt_dyn .or. current_sim_ET >=  &
                 maximum_sim_ET-0.01_f4*gd(1)%dt_dyn)) then
               Qprime = Qprime-minthresh  ! Subtract some energy from Qp
               Q(k,i,j,ip) = Q(k,i,j,ip)+Qprime  ! Add energy to main Q array
               Qp(k,i,j,ip) = minthresh   ! Reset to smallish value
            endif
!            print *,Q(k,i,j,ip),Qp(k,i,j,ip)
         enddo
      enddo
   enddo
enddo

end subroutine SBPS_Qprcadj


end module PRAMS_SBPS_mod
