!```````````````````````````````````````````````````````````````````
module aemic_bound_mod

!+ Module that contains aerosol microphysics routine(s)

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine limit_minconc

!+ Limits the minimum particle concentrations

use aermicro_mem_mod, only: nelements,igelem,ienconc,nbins,nz_mic,pc,SMALL_PC  &
     ,small_coremconc

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

! Limits all particle number concentrations in a grid box to SMALL_PC.  In bins
!  where this limitation results in the number particle concentration changing,
!  the core mass fraction and second moment fraction are set to <FIX_COREF>;
!  <small_coremconc> is the function of <FIX_COREF> that is actually used:

integer :: ie,ig,ip,ib,k

!=== Executable statements:
   
do ie=1,nelements
   ig = igelem(ie)
   ip = ienconc(ig)

   if(ie == ip) then    ! Element is particle concentration

      do ib=1,nbins(ig)
         do k=1,nz_mic
            pc(k,ib,ie) = max(pc(k,ib,ie),SMALL_PC)
         enddo
      enddo

   else           ! Element is core mass or core second moment
 
      do ib=1,nbins(ig)

         do k=1,nz_mic
            if(pc(k,ib,ip) <= SMALL_PC .or.  &
                 pc(k,ib,ie) < small_coremconc(ib,ie)) then
               pc(k,ib,ie) = small_coremconc(ib,ie)
            endif
         enddo

      enddo

   endif

enddo

end subroutine limit_minconc


end module aemic_bound_mod
