!```````````````````````````````````````````````````````````````````
module aemic_postDt_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine post_step

!+ Do operations required after each time-integration step

use aermicro_mem_mod, only: igridv,I_CART,dtime,time,ngroups,inucmin,inucstep

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

! Things that would appropriately be done here include:
!    Smoothing or other corrections to the latest state.
!    Possible output of print file information on current step.
!    Possible output of history file information on current step.
!    Possible output of restart file information on current step.
!    Debugging output of changes within the step.

integer :: ig

!=== Executable statements:

! Rescale vertical winds and diffusion coefficients for non-Cartesian vertical
!  coordinate systems:

!%if(igridv /= I_CART) then
!%   call rescale_vertvel
!%endif

! Update simulation time to end of current timestep:

!%time = time+dtime

! Management of index of smallest bin nucleated from a group <inucmin>
!  and nucleation update time <time_nuc>:
!
!   Whenever <inucstep> is less than <inucmin>, update <inucmin> and <time_nuc>.
!   Otherwise, update <inucmin> and <time_nuc> every <period_nuc> seconds.
!
!   <time_nuc> is the previous time that <inucmin> was updated.  
!   If <inucmin> is updated this time step, then <time_nuc>
!   becomes the next time that it will be updated.

do ig=1,ngroups
   inucmin(ig) = inucstep(ig)    ! Update inucmin after every timestep (TIM)
enddo
  
end subroutine post_step


end module aemic_postDt_mod
