!```````````````````````````````````````````````````````````````````
module aemic_preDt_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine pre_step

!+ Do operations required before each time-integration step

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4
use aermicro_mem_mod, only: nelements,nbins,igelem,nz_mic,pcl,pc,ngroups  &
     ,SMALL_PC,ienconc,pconmax,ngasses,gcl,gc,supsatlold,supsatl,supsatiold  &
     ,supsati,ptcl,ptc,told,t,pold,p,rhoaold,rhoa,coagpe,coaglg,do_coag
use aemic_bound_mod, only: limit_minconc

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

! Things that would appropriately be done here include:
!  - Input or otherwise define interface quantities from other submodels.
!  - Save any model state that is needed to compute tendencies.
!  - Save any model state that might be needed for comparison at end of step.
!  - Update timestep counter and simulation time.

!%      logical(g_normal) vap_cont
integer :: ie,ib,ig,k,igas
real(f4) :: pctmp

!=== Executable statements:

! Do not allow particle concentrations to get too small:

call limit_minconc

! Set <pcl> to <pc> from previous time step for each element:
!#### (<pcmax> is only used for variable time stepping)

do ie=1,nelements
   do ib=1,nbins(igelem(ie))
      do k=1,nz_mic
         pcl(k,ib,ie) = pc(k,ib,ie)
!#    pcmax calc
      enddo
   enddo
enddo

! Find maximum particle concentration for each spatial grid box [*/cm^3]:

do ig=1,ngroups
   ie = ienconc(ig)
   do k=1,nz_mic
      pctmp = SMALL_PC
      do ib=1,nbins(ig)
         pctmp = max(pctmp,pc(k,ib,ie))
      enddo
      pconmax(k,ig) = pctmp
   enddo
enddo

! Set <gcl> to <gc> from previous time step and store old ice and liquid
!  supersaturations for variable time stepping:

do igas=1,ngasses
   do k=1,nz_mic
      gcl(k,igas) = gc(k,igas)
      supsatlold(k,igas) = supsatl(k,igas)
      supsatiold(k,igas) = supsati(k,igas)
   enddo
enddo

! Store old values of atmospheric properties:
!#### (<told>,<pold>,<rhoaold> are only used for variable time stepping)

do k=1,nz_mic
   ptcl(k) = ptc(k)
!#   told(k) = t(k)
!#   pold(k) = p(k)
!#   rhoaold(k) = rhoa(k)
enddo

if(do_coag) then      ! Only perform every timestep if coagulation turned on

   ! Set production terms and loss rates due to slow microphysical processes
   !  (coagulation) to zero:

   do ie=1,nelements
      do ib=1,nbins(igelem(ie))
         do k=1,nz_mic
            coagpe(k,ib,ie) = ZERO_f4
         enddo
      enddo
   enddo

   do ig=1,ngroups
      do ib=1,nbins(ig)
         do k=1,nz_mic
            coaglg(k,ib,ig) = ZERO_f4
         enddo
      enddo
   enddo
   
endif

!??????????????????????????????????????????????????????????????????????

! Adjust CO2 vapor concentration based on changes to total mass change
!  at each level (wmpart [g]):
!%     
!%      vap_cont = .false.
!%		 
!%      do iz=1,NZ
!% 
!%        rmdrop = 0.
!%        rmcore_vol = 0.
!% 
!%        do ix=1,NX
!%         do iy=1,NY
!%          grid_vol = dx(ix,iy,iz)*dy(ix,iy,iz)*dz(ix,iy,iz)
!%	  
!%          do i=1,NBIN
!%           rmcore_vol = rmcore_vol + pc(ix,iy,iz,i,3)
!%
!%           rmdrop = rmdrop + pc(ix,iy,iz,i,2)*rmass(i,2)
!%	   
!%          enddo
!%	  
!%	  rmcond = rmdrop-rmcore_vol
!%	  
!%	  xyzmet = xmet(ix,iy,iz)*ymet(ix,iy,iz)*zmet(ix,iy,iz)
!%	  gcmetric = gc(ix,iy,iz,2)/xyzmet
!%	  rhoametric = rhoa(ix,iy,iz)/xyzmet
!%	  
!%	  Snd = Sqrt(R_AIR*t3(iz))
!%          FRate = Snd/zmet(ix,iy,iz)/2.
!%
!%          If(vap_cont.eq..true.) then
!%	  
!%	   grad = ((gcmetric-rmdrop)/rhoametric - 0.95)/dtime
!%           gc(ix,iy,iz,2)=(gcmetric-FRate*gcmetric*grad)*xyzmet
!%
!%	  Else    
!%	   
!%	   frac_ice = rmcond/rhoa(ix,iy,iz)/xyzmet
!%	   If(frac_ice.ge.0.95) frac_ice = 0.94
!%	   If(frac_ice.lt.0.0) frac_ice = 0.0
!%c	   frac_ice = 0.
!%c           gc(ix,iy,iz,2) = rhoa(ix,iy,iz)*(.95-frac_ice)*xyzmet
!%	   
!%          Endif
!%
!%	   If(gc(ix,iy,iz,2).le.1.e-20) gc(ix,iy,iz,2)=1.e-20
!%           If(gc(ix,iy,iz,2).gt.35.) gc(ix,iy,iz,2) = 35.
!%	   
!%      enddo
!%        enddo
!%      enddo

end subroutine pre_step


end module aemic_preDt_mod
