!```````````````````````````````````````````````````````````````````
module ARCnRT_driv_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine ARCnRT_driver(ext_n1, ext_n2, ext_n3, ext_n5, ext_ia, ext_iz,  &
     ext_ja, ext_jz, ext_ka, ext_kz, ext_naersclr, ext_aersclp, ext_maxnbins,  &
     ext_ngroups, ext_aerradius, ext_aer_ngrp, ext_aer_group, ext_aer_subgrp,  &
     ext_aer_ig, ext_aer_grpbins, ext_press, ext_tempk, ext_rv, ext_sfcT,  &
     ext_tautot, ext_cosZ, ext_msdist, ext_albedo, ext_emissivity,  &
     ext_use_gas_opacities, DUST_tau, DUST_tau_pr, DIM_dust_levs,  &
     ext_rlong, ext_fthrd, ext_dzc_static, ext_rtgt, ext_solar_azm_ang,  &
     ext_tazm_ang, ext_tslope_ang, ext_tslas_tog, ext_tshadow_tog,  &
     ext_nazm_tshadow, ext_cosz_horizon, ext_TCx_obstruct, ext_TCz_obstruct,  &
     ext_zc_static, ext_XOrad_mask, ext_XOrad, init_rad, ext_haveH2Ov,  &
     ext_vtau, ext_crnu, ext_vtau_tog, ext_flxout_imask,  &
     ext_bottom_other_sflux_down, ext_bottom_SWflux_down,  &
     ext_topo_factor_sflux, ext_fluxo_np, ext_up_fluxes, ext_dn_fluxes,  &
     ext_dT_dt, ext_solar_sw)

!+ Modular radiative transfer code driver

use univ_kind_defs_mod, only: f8, f4
use univ_math_constants_mod, only: ZERO_f8, ONE_f4, twopi_f4, ZERO_f4
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_convert_units_mod, only: g_to_kg_f4, mb_to_Pa_f8, mb_to_Pa_f4 
use body_specific_constants_mod, only: R_air, solar_mean_broadband
!#use aertxrad_mem_mod, only: nz_rad,nz1_rad,max_aerbins,naerspc_rad,i_curr,j_curr  &
!#     ,two_nz1_rad,p,t,t_lev,p_lev,qH2O_lev,sfc_tempk,sflux_frac  &
!#     ,sw_iwbeg,sw_iwend,sfc_albedo,sfc_emiss,scr_dim,sw_sfc_dnflux  &
!#     ,lw_sfc_dnflux,net_dTdt,dzc,tau_aer,aertxrad_AD,sw_on,ip_ibeg  &
!#     ,ip_iend,ntot_gpt,lw_ipbeg,lw_ipend,iw_iend,iw_ibeg,lw_iwbeg,lw_iwend  &
!#     ,ntot_wavlnrng,spc_radtx,spc_begiv,part_in_tshadow,swcut,bdst_Nconc,  &
!#     ntot_wavlnrng, want_rtflux_output, up_fluxes, dn_fluxes, tau_gas
!#use aermic_rad_mem_mod, only: faux_dust_rtxinfo_needed
!#use aerad_aeropd_mod, only: bdst_calc, bdst_calc_htop, calc_aeropd
!#use aerad_txrad_mod, only: two_stream, add_txrad, refine_lwflux
!#use aerad_precalc_mod, only: rad_precalc
!#use aerad_gasopd_mod, only: calc_gasopd
!#use aerad_postcalc_mod, only: rad_postcalc
use ARCnRT_memory_mod, only: albi, albv, tlev, plev, tmid, pmid, qH2O, sol,  &
     cos_mu, dtauv, tauv, taucumv, fmnetv, fmneti, fluxupi, fluxdni, fluxupv,  &
     fluxdnv, wbarv, detau, taugsurf, cosbv, nfluxtopv, nfluxtopi, diffvt,  &
     want_rtflux_output, sw_on, i_curr, j_curr, swcut, part_in_tshadow,  &
     wbari, taugsurfi, taucumi, dtaui, cosbi, grav, cp_air, net_dTdt,  &
     use_gas_opacities, up_fluxes, dn_fluxes, solar_sw
use ARCnRT_fccsave_h, only: QEXTREFDST, QEXTREFCLD, QXVCLD, QSVCLD, QXICLD,  &
     QSICLD, GICLD, GVCLD, QXIDST, QSIDST, GIDST, QXVDST, QSVDST, GVDST,  &
     taurefcld
use ARCnRT_radcommon_h, only: gi, gic, gv, gvc, QEXTI, QEXTV, QSCATV, QSCATI,  &
     QEXTVc, QSCATVc, QEXTIc, QSCATIc, solarf, co2v, fzerov, gweight, tauref,  &
     tgasref, wrefh2o, pfgasref, tauray, ubari, fzeroi, dwni, co2i
use ARCnRT_radinc_h, only: l_nlayrad, L_NSPECTV, L_NSPECTI 
use ARCnRT_grid_h, only: l_levels, l_layers
use ARCnRT_dustprofile_mod, only: ARCnRT_dustprofile
use ARCnRT_filltaucum_mod, only: ARCnRT_filltaucum
use ARCnRT_ini_optdst_mod, only: ARCnRT_ini_optdst
use ARCnRT_ini_optcld_mod, only: ARCnRT_ini_optcld
use ARCnRT_optcv_mod, only: ARCnRT_optcv
use ARCnRT_sfluxv_mod, only: ARCnRT_sfluxv
use ARCnRT_optci_mod, only: ARCnRT_optci
use ARCnRT_sfluxi_mod, only: ARCnRT_sfluxi

implicit none

!=== Argument declarations:

integer, intent(IN) :: ext_n1, ext_n2, ext_n3, ext_n5, ext_ia, ext_iz,  &
     ext_ja, ext_jz, ext_ka, ext_kz, DIM_dust_levs, ext_naersclr,  &
     ext_maxnbins, ext_ngroups, ext_aer_ngrp, ext_tslas_tog, ext_tshadow_tog,  &
     ext_nazm_tshadow, ext_vtau_tog
real(f4), intent(IN) :: ext_tautot, ext_msdist
real(f4), dimension(DIM_dust_levs), intent(IN) :: DUST_tau, DUST_tau_pr
real(f4), dimension(ext_n2,ext_n3,ext_n5), intent(IN) :: ext_sfcT
real(f4), dimension(ext_n2,ext_n3), intent(IN) :: ext_albedo, ext_cosZ,  &
     ext_rtgt, ext_solar_azm_ang, ext_tazm_ang, ext_tslope_ang, ext_vtau,  &
     ext_crnu, ext_emissivity
real(f4), dimension(ext_n2,ext_n3,ext_nazm_tshadow), intent(IN) ::  &
     ext_cosz_horizon, ext_TCx_obstruct, ext_TCz_obstruct
real(f4), dimension(ext_n1), intent(IN) :: ext_dzc_static, ext_zc_static
real(f4), dimension(ext_n1,ext_n2,ext_n3), intent(IN) :: ext_press, ext_tempk
real(f4), dimension(ext_maxnbins,ext_ngroups), intent(IN) :: ext_aerradius
integer, dimension(ext_naersclr), intent(IN) :: ext_aer_group,  &
     ext_aer_subgrp, ext_aer_ig
integer, dimension(ext_aer_ngrp), intent(IN) :: ext_aer_grpbins
logical, dimension(ext_naersclr), intent(IN) :: ext_XOrad, ext_XOrad_mask
logical, intent(IN) :: init_rad, ext_haveH2Ov, ext_use_gas_opacities
! ***NOTE: the following have CGS units:
real(f4), dimension(ext_n1,ext_n2,ext_n3), intent(IN) :: ext_rv
real(f4), dimension(ext_n1,ext_n2,ext_n3,ext_naersclr), intent(IN) ::  &
     ext_aersclp
integer, dimension(ext_n2,ext_n3), intent(IN) :: ext_flxout_imask
integer, intent(IN) :: ext_fluxo_np

real(f4), dimension(ext_n2,ext_n3), intent(INOUT) :: ext_rlong,  &
     ext_bottom_other_sflux_down, ext_bottom_SWflux_down, ext_topo_factor_sflux
real(f4), dimension(ext_n1,ext_n2,ext_n3), intent(INOUT) :: ext_fthrd

real(f4), dimension(ext_n1,L_NSPECTV+L_NSPECTI,ext_fluxo_np), optional,  &
     intent(INOUT) :: ext_up_fluxes, ext_dn_fluxes
real(f4), dimension(ext_n1,ext_fluxo_np), optional, intent(INOUT) :: ext_dT_dt
real(f4), optional, intent(OUT) :: ext_solar_sw

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'ARCnRT_driver'
    
real(f4) :: sinz, cosai, atmp, aint, cosz_h, x_obst, z_obst, sw_cutoff, rho_air
real(f8) :: tmp_f8
integer :: iv, ig, ik, k, i, j, iw, ip, ia0, ia1, ix_fo, iy_fo, l
logical :: horizon_extension, hextn_active, have_some_opacities
!%logical :: found

!=== Executable statements:

!%if (init_rad) then     ! Initialize radiative transfer model (do only once)

!%end if

! Check whether optional arguments are required:
if (maxval(ext_flxout_imask) > 0) then
   if (.not. (present(ext_up_fluxes) .and. present(ext_dn_fluxes) .and.  &
        present(ext_dT_dt) .and. present(ext_solar_sw)))  &
        call exception(procedure_name, FATAL,  &
        'Optional arguments not provided, but are required.')
end if

! Find and flag currently radiatively active aerosol species:

!%ig = naerspc_rad
!%if(faux_dust_rtxinfo_needed) then
!%   ig = ig-1
!%   spc_radtx(naerspc_rad) = .TRUE.
!%endif

!%do ip=1,ig     ! Loop over all *possible* real rad_active species
!%   iv = spc_begiv(ip)
!%   spc_radtx(ip) = ext_XOrad(iv)   ! .TRUE. if rad_active now, else .FALSE.
!%enddo

ix_fo = 0; iy_fo = 0

do j=ext_ja,ext_jz        !
   j_curr = j             ! Loop over all *computational* horizontal grid
   do i=ext_ia,ext_iz     !  points of external model
      i_curr = i          !

      ! Set temperature, pressure, and water vapor mixing ratio of layer
      !  MIDPOINTS [with ik=l_levels-1 being the lowest layer (in z), and
      !  k=ext_kz being remapped to ik=2]:

      ik = l_levels-1
      if (ext_haveH2Ov) then   ! Water vapor present in external model
         do k=ext_ka,ext_kz
!%            dzc(ik) = ext_rtgt(i,j)/ext_dzc_static(k) ! dz (real-world) of lyr
            tlev(ik) = real(ext_tempk(k,i,j), kind=f8)     ! [K]
            plev(ik) = real(ext_press(k,i,j), kind=f8)      ! [hPa]
            rho_air = ext_press(k,i,j)*mb_to_Pa_f4/  &
                 (ext_tempk(k,i,j)*R_air) ! rho [kg/m^3]
            tmp_f8 = real(ext_rv(k,i,j)/(rho_air*g_to_kg_f4),  &
                 kind=f8)    ! H2O vapor mconc [g/cm^3] -> mixing ratio [g/g]
!#            qH2O(ik) = max(tmp_f8, 1.e-7_f8) ! H2Ov mixing ratio [g/g = kg/kg]
            qH2O(ik) = tmp_f8 ! H2Ov mixing ratio [g/g = kg/kg]
            ik = ik-2
         end do
      else   ! No water vapor, and ext_rv() likely holds undefined values
         do k=ext_ka,ext_kz     
!%            dzc(ik) = ext_rtgt(i,j)/ext_dzc_static(k) ! dz (real-world) of lyr
            tlev(ik) = real(ext_tempk(k,i,j), kind=f8)     ! [K]
            plev(ik) = real(ext_press(k,i,j), kind=f8)      ! [hPa]
!@            rho_air = ext_press(k,i,j)/(ext_tempk(k,i,j)*R_air) ! rho [kg/m^3]
            qH2O(ik) = ZERO_f8  ! H2Ov mixing ratio [g/g = kg/kg]
            ik = ik-2
         end do
      end if

      ! Set temperature, pressure, and water vapor mixing ratio of layer
      !  BOUNDARIES:

      do ik=3,l_levels-2,2
         tlev(ik) = (tlev(ik-1)+tlev(ik+1))*0.5_f8
         plev(ik) = exp(0.5_f8*log(plev(ik+1)/plev(ik-1))+  &
              log(plev(ik-1)))  ! log(p) interp
         qH2O(ik) = (qH2O(ik-1)+qH2O(ik+1))*0.5_f8
      end do

      ik = 1   ! Uppermost layer BOUNDARY
      tlev(ik) = tlev(ik+1)
      plev(ik) = plev(ik+2)*0.5_f8
      qH2O(ik) = qH2O(ik+1)

      ik = 2   ! Uppermost layer MIDPOINT
      plev(ik) =  plev(ik-1)

      ik = 3   ! Next-to-uppermost layer BOUNDARY
      tlev(ik) = tlev(ik-1)

      ik = l_levels   ! Bottom layer BOUNDARY
      tlev(ik) = ext_sfcT(i,j,1)
      plev(ik) = real(ext_press(ext_ka-1,i,j), kind=f8)  ! [hPa]
      qH2O(ik) = qH2O(ik-1)

!  Fill the PMID & TMID arrays used by OPTCI and OPTCV subroutines.
!  TMID and PMID used to get the index for CO2 k-coefficient 
!  interpolation.

      TMID(1) = TLEV(2)
      TMID(2) = TLEV(2)
      PMID(1) = PLEV(1)
      PMID(2) = PLEV(2)

      DO L=1,L_LAYERS
        TMID(2*L+1) = TLEV(2*L+1)
        TMID(2*L+2) = TLEV(2*L+1)
        PMID(2*L+1) = PLEV(2*L+1)
        PMID(2*L+2) = PLEV(2*L+1)
      END DO

      TMID(L_LEVELS) = TLEV(L_LEVELS)
      PMID(L_LEVELS) = PLEV(L_LEVELS)

      ! Calculate "background" dust profile for this column:

      call ARCnRT_dustprofile(ext_vtau(i,j), ext_crnu(i,j))
      call ARCnRT_filltaucum
      
!%      if(ext_vtau_tog == 3 .or. ext_vtau_tog == 4) then
!%         ! Use "vtau" "background" dust loading:
!%         call bdst_calc_htop(ext_vtau(i,j),ext_crnu(i,j),ext_aer_grpbins(1))
!%      elseif(ext_vtau_tog == 1 .or. ext_vtau_tog == 2) then
!%         ! Use "background" dust loading:
!%         call bdst_calc(ext_vtau(i,j),ext_crnu(i,j),ext_aer_grpbins(1))
!%      else    ! No "background" dust
!%         bdst_Nconc = ZERO_f4   ! Fill entire array at once
!%      endif

      ! Set "longwave" surface properties:
      albi = ONE_f4-ext_emissivity(i,j)    ! [none]

      !*** Set cosine of incident solar angle (zenith angle); take into account
      !     (crudely) any horizon extension:

      cos_mu = real(ext_cosZ(i,j), kind=f8)     ! [none]

      hextn_active = .FALSE.
      if (ext_tshadow_tog == 1) then   ! Need to check for horizon extension
      
         ! Calculate cosz_horizon array location/interpolative indices (using
         !  solar azimuth values):
         atmp = ext_solar_azm_ang(i,j)/(twopi_f4/real(ext_nazm_tshadow))
         ia0 = int(atmp)
         aint = atmp-real(ia0)   ! Location crucial to avoid ia0 complications
         
         if (ia0 == 0) then
            ia0 = ext_nazm_tshadow; ia1 = 1
         else if (ia0 == ext_nazm_tshadow) then
            ia1 = 1
         else
            ia1 = ia0+1
         end if

         ! Calc horizon cosz, etc. for current solar azimuth via interpolation:
         z_obst = (ONE_f4-aint)*ext_TCz_obstruct(i,j,ia0)+  &
              aint*ext_TCz_obstruct(i,j,ia1)
         if (z_obst < epsilon(z_obst)) then   ! Horizon extension
            horizon_extension = .TRUE.
            cosz_h = (ONE_f4-aint)*ext_cosz_horizon(i,j,ia0)+  &
                 aint*ext_cosz_horizon(i,j,ia1)
            if (real(cos_mu, kind=f4) > cosz_h .and. cos_mu < 4.e-2_f8) then
               ! Sun is above extended horizon, but still low in sky;
               !  cosZ=0.04 =~ 87.71deg
               cos_mu = min(cos_mu-real(cosz_h, kind=f8), 4.e-2_f8)
               hextn_active = .TRUE.
            end if
         else
            horizon_extension = .FALSE.
         end if
      end if

      ! Only calculate "shortwave" stuff if sun is sufficiently above horizon;
      !  set probability point interval indices for use in radiative transfer
      !  calcs:

      if (cos_mu > 1.e-4_f8) then
         sw_on = .TRUE.
!#         ip_ibeg = 1          ! Loop over ALL probability points
!#         ip_iend = ntot_gpt   !  
!#         iw_ibeg = 1               ! Loop over ALL wavelength bands
!#         iw_iend = ntot_wavlnrng   !

         ! Set "shortwave" wavelength group sfc optical properties:      
         albv = ext_albedo(i,j)       ! [none]

         ! Topographic slope/aspect correction:

         if (ext_tslas_tog > 1) then
            sinz = sin(acos(real(cos_mu, kind=f4)))
            
            ! If cosai < 0., point not sunlit
            cosai = (real(cos_mu, kind=f4)*cos(ext_tslope_ang(i,j)))+  &
                 (sinz*sin(ext_tslope_ang(i,j))*  &
                 cos(ext_solar_azm_ang(i,j)-ext_tazm_ang(i,j)))
         end if

         ! Topographic shadowing:

         part_in_tshadow  = .FALSE.   ! Initialize
         if (ext_tshadow_tog == 1 .and.  &     ! Possible shadowing
              .NOT.horizon_extension) then
            cosz_h = (ONE_f4-aint)*ext_cosz_horizon(i,j,ia0)+  &
                 aint*ext_cosz_horizon(i,j,ia1)
            x_obst = (ONE_f4-aint)*ext_TCx_obstruct(i,j,ia0)+  &
                 aint*ext_TCx_obstruct(i,j,ia1)
!Calc'd          z_obst = (1.-aint)*ext_TCz_obstruct(i,j,ia0)+  &
!above                 aint*ext_TCz_obstruct(i,j,ia1)

            sw_cutoff = z_obst-(x_obst*tan(asin(real(cos_mu, kind=f4)))) 

            if (sw_cutoff > ZERO_f4 .and.  &  ! Part of column shadowed
                 sw_cutoff < ext_zc_static(ext_kz)*ext_rtgt(i,j)) then 

               part_in_tshadow = .TRUE.

               swcut = 0
               atmp = ZERO_f4
               shadow_top: do ik=ext_ka,ext_kz        ! Find top of shadow
                  aint = ext_zc_static(ik)*ext_rtgt(i,j)
                  if(sw_cutoff < aint) then
                     if(ik /= ext_ka) then
                        if(abs(aint-sw_cutoff) < abs(atmp-sw_cutoff)) then
                           swcut = ext_kz-ik+1
                        else
                           swcut = ext_kz-ik+2
                        endif
                     else
                        swcut = ext_kz-ik+1
                     endif
                     exit shadow_top
                  endif
                  atmp = aint
               end do shadow_top   ! Column sunlit above layer 'swcut'
            
               if(swcut == 0) then
                  write(*,*) 'ERROR: unable to find shadow top.'
                  stop '*** ERROR: ARCnRT_driver01 ***'
               end if
               
            end if
         end if

         ! Interplanetary stellar flux at the top of the atmosphere:
         if (part_in_tshadow) then   ! Eliminate direct SW flux within shadow
            sol = ZERO_f8
         else
            sol = solarf/real(ext_msdist, kind=f8)
         end if
 
      else
         sw_on = .FALSE.
         ! Interplanetary stellar flux at the top of the atmosphere:
         sol = ZERO_f8
!#         ip_ibeg = lw_ipbeg    ! Only loop over "longwave" pts if
!#         ip_iend = lw_ipend    !  "shortwave" is off
!#         iw_ibeg = lw_iwbeg    !
!#         iw_iend = lw_iwend    !
      end if

      ! Compute "air" (mixed CO2 *and* H2O_vapor) opacities at layer MIDPOINTS
      !  [results placed in tau_air(ig,iw,k)]:

      use_gas_opacities = ext_use_gas_opacities

      call ARCnRT_ini_optdst(QEXTV, QSCATV, GV, QEXTI, QSCATI, GI, QXVDST,  &
           QXIDST, QSVDST, QSIDST, GVDST, GIDST, QEXTREFDST)
      call ARCnRT_ini_optcld(QEXTVc, QSCATVc, GVc, QEXTIc, QSCATIc, GIc,  &
           QXVCLD, QXICLD, QSVCLD, QSICLD, GVCLD, GICLD, QEXTREFCLD)
!#      call calc_aeropd(ext_n1,ext_n2,ext_n3,ext_naersclr,ext_aersclp  &
!#           ,ext_sfcT(i,j,1),ext_kz)

      taurefcld = ZERO_f8

      if (ext_flxout_imask(i,j) > 0) then
         want_rtflux_output = .TRUE.      ! Calc/save 2-stream fluxes
      else
         want_rtflux_output = .FALSE.
      end if

      !================================================================

      ! Are there any RT-relevant opacities at all?
      have_some_opacities = (ext_vtau(i,j) > tiny(ext_vtau(i,j)) .or.  &
           (maxval(taurefcld) > tiny(taurefcld(1))) .or. use_gas_opacities)

      ! If the sun is up, calculate the solar fluxes, else set them to zero:
      up_fluxes(:,L_NSPECTI+1:L_NSPECTV+L_NSPECTI) = 0._f8
      dn_fluxes(:,L_NSPECTI+1:L_NSPECTV+L_NSPECTI) = 0._f8
      if (sw_on .and. have_some_opacities) then
         !  Calculate the optical depths in each layer, spectral interval,
         !  and Gauss point.

         call ARCnRT_optcv(DTAUV,TAUV,TAUCUMV,CO2V,PLEV,PFGASREF,  &
              TGASREF,QXVDST,QSVDST,GVDST,WBARV,COSBV,  &
              TAURAY,TAUREF,TMID,PMID,TAUGSURF,QH2O,WREFH2O,  &
              QEXTREFCLD,TAUREFCLD,QXVCLD,QSVCLD,GVCLD)

         !  Compute the visible fluxes
         call ARCnRT_sfluxv(DTAUV,TAUV,TAUCUMV,ALBV,WBARV,COSBV,  &
              cos_mu,SOL,GWEIGHT,NFLUXTOPV,FMNETV,  &
              FLUXUPV,FLUXDNV,DIFFVT,FZEROV,taugsurf,  &
              detau)
      else
         NFLUXTOPV = ZERO_f8
         do L=1,L_NLAYRAD
            FMNETV(L) = ZERO_f8
            FLUXUPV(L) = ZERO_f8
            FLUXDNV(L) = ZERO_f8
         end do
      end if

!     Set up, and solve for, the Infrared fluxes

      up_fluxes(:,1:L_NSPECTI) = 0._f8; dn_fluxes(:,1:L_NSPECTI) = 0._f8
      if (have_some_opacities) then
!  Calculate the optical depths in each layer, spectral interval,
!  and Gauss point.

         call ARCnRT_optci(DTAUI,TAUCUMI,CO2I,PLEV,PFGASREF,TGASREF,  &
              QEXTREFDST,QXIDST,QSIDST,GIDST,COSBI,WBARI,TAUREF,  &
              TMID,PMID,TAUGSURFI,QH2O,WREFH2O,  &
              QEXTREFCLD,TAUREFCLD,QXICLD,QSICLD,GICLD)
 
!  Compute the IR fluxes

         call ARCnRT_sfluxi(PLEV,TLEV,DTAUI,TAUCUMI,UBARI,ALBI,DWNI,  &
              COSBI,WBARI,GWEIGHT,NFLUXTOPI,FMNETI,  &
              fluxupi,fluxdni,FZEROI,taugsurfI)
      else
         NFLUXTOPI = ZERO_f8
         do L=1,L_NLAYRAD
            FMNETI(L) = ZERO_f8
            FLUXUPI(L) = ZERO_f8
            FLUXDNI(L) = ZERO_f8
         end do         
      end if

      !================================================================
 
!  Heating rates
  
      ik = 1
      net_dTdt(ik) = ((FMNETV(1)-NFLUXTOPV)+(FMNETI(1)-NFLUXTOPI))*  &
           grav/(cp_air*mb_to_Pa_f8*PLEV(3))

      do ik=2,L_NLAYRAD 
         net_dTdt(ik) = ((FMNETV(ik)-FMNETV(ik-1))+(FMNETI(ik)-FMNETI(ik-1)))* &
              grav/(cp_air*mb_to_Pa_f8*(PLEV(2*ik+1)-PLEV(2*ik-1)))
      end do

      ! *Incident* surface "shortwave" flux [W/m^2]:
      if (sw_on) then
         ! The portion of the stellar flux (at the body's surface) that *IS*
         !  treated by the "shortwave" atmospheric RT calculations:
         ext_bottom_SWflux_down(i,j) = real(fluxdnv(l_nlayrad), kind=f4)

         ! The portion of the stellar flux (at the body's surface) that *IS NOT*
         !  treated by the "shortwave" atmospheric RT calculations:
         ext_bottom_other_sflux_down(i,j) = (solar_mean_broadband-solar_sw)/  &
              ext_msdist*real(cos_mu, kind=f4)

         if (ext_tslas_tog > 1 .and. .NOT.hextn_active) then
            ! Multiply by cosai to take both topography orientation AND stellar
            !  incidence angle into account (after dividing by cos_mu to obtain
            !  the non-angle-corrected incident stellar flux):
            ext_topo_factor_sflux(i,j) = cosai/real(cos_mu, kind=f4)
         else
            ! Take ONLY cos(Z) into account:
            ext_topo_factor_sflux(i,j) = ONE_f4
         end if
      else
         ext_bottom_SWflux_down(i,j) = ZERO_f4
         ext_bottom_other_sflux_down(i,j) = ZERO_f4
         ext_topo_factor_sflux(i,j) = ZERO_f4
      end if

      ext_rlong(i,j) = real(fluxdni(l_nlayrad), kind=f4) ! Surface "longwave" down flux [W/m^2]

      if (want_rtflux_output) then        ! Calc/save 2-stream fluxes
         ip = ext_flxout_imask(i,j)
         do iw=1,L_NSPECTV+L_NSPECTI 
            ik = l_nlayrad
            do k=ext_ka-1,ext_kz-1
               ext_up_fluxes(k,iw,ip) = real(up_fluxes(ik,iw), kind=f4)
               ext_dn_fluxes(k,iw,ip) = real(dn_fluxes(ik,iw), kind=f4)
               ik = ik-1
            end do
            ext_up_fluxes(ext_kz,iw,ip) = ext_up_fluxes(ext_kz-1,iw,ip)
            ext_dn_fluxes(ext_kz,iw,ip) = ext_dn_fluxes(ext_kz-1,iw,ip)
         end do

         ik = l_nlayrad
         do k=ext_ka,ext_kz
            ext_dT_dt(k,ip) = real(net_dTdt(ik), kind=f4)  ! Net layer heating rate
            ik = ik-1
         end do
         
         ! The portion of the stellar flux @1 AU that *IS* treated by the
         !  "shortwave" atmospheric RT calculations:
         ext_solar_sw = solar_sw
      end if
 
      ik = l_nlayrad
      do k=ext_ka,ext_kz     
         ext_fthrd(k,i,j) = real(net_dTdt(ik), kind=f4)  ! Net layer heating rate
         ik = ik-1
      end do

   end do
end do

end subroutine ARCnRT_driver


end module ARCnRT_driv_mod
