module ARCnRT_dtridgl_mod

implicit none


contains


      SUBROUTINE ARCnRT_DTRIDGL(L,AF,BF,CF,DF,XK)

!     GCM v23   2010
!     Ames Mars GCM group
!     Jeffery Hollingsworth, PI
!     NASA Ames Research Center

      use univ_kind_defs_mod, only: f8

      implicit none

!     DOUBLE PRESCISION VERSION OF TRIDGL

      integer, intent(IN) :: L
      real(f8), intent(IN) :: AF(L), BF(L), CF(L), DF(L)
      real(f8), intent(OUT) :: XK(L)

      integer, parameter :: NMAX = 201
      integer :: I

      real(f8) :: AS(NMAX),DS(NMAX), X, XKB

!*    THIS SUBROUTINE SOLVES A SYSTEM OF TRIDIAGIONAL MATRIX
!*    EQUATIONS. THE FORM OF THE EQUATIONS ARE:
!*    A(I)*X(I-1) + B(I)*X(I) + C(I)*X(I+1) = D(I)
!*    WHERE I=1,L  LESS THAN 103.
!* ..............REVIEWED -CP........

!======================================================================C

      AS(L) = AF(L)/BF(L)
      DS(L) = DF(L)/BF(L)

      DO I=2,L
        X         = 1._f8/(BF(L+1-I) - CF(L+1-I)*AS(L+2-I))
        AS(L+1-I) = AF(L+1-I)*X
        DS(L+1-I) = (DF(L+1-I)-CF(L+1-I)*DS(L+2-I))*X
      END DO
 
      XK(1)=DS(1)
      DO I=2,L
        XKB   = XK(I-1)
        XK(I) = DS(I)-AS(I)*XKB
      END DO

      RETURN
      end subroutine ARCnRT_dtridgl


end module ARCnRT_dtridgl_mod
