module ARCnRT_dustprofile_mod

implicit none


contains


      subroutine ARCnRT_dustprofile(tautot_ij, conrath_nu)

!     GCM v23   2010
!     Ames Mars GCM group
!     Jeffery Hollingsworth, PI
!     NASA Ames Research Center

! Bob's updates 9/17/99 
! Reference the dust optical depth to RPTAU, and modify the way
! the dust mixing ratio is calculated to more accurately reflect the
! pressure-optical depth relationship.
!  Called by EMISS and GMP.

use univ_kind_defs_mod, only: f8, f4
      use ARCnRT_defines_h, only: L_NPDST
      use ARCnRT_standard_h, only: prdst, taudst, npdst
use ARCnRT_memory_mod, only: dust_tau_ref_pr_min, dust_tautot_ref_pr

      implicit none

      real(f4), intent(IN) :: tautot_ij, conrath_nu

      REAL(f8) :: QRDST(L_NPDST)

!     implicit none

!TIM      integer :: i, j, n
      integer :: n
      real(f8)  :: refpr, conr, sum, pave, qrdst0, pdif1, pdif2
!TIM      real(f8)  :: deltals, conrtmp, tetalat

!======================================================================C

! Calculate the Reference Pressure Grid (prdst)

      refpr = (5._f8*dust_tautot_ref_pr/dust_tau_ref_pr_min)**  &
           (1._f8/(real(npdst, kind=f8)-1._f8))
      prdst(1) = dust_tau_ref_pr_min
   
      do n=2,npdst
        prdst(n) = refpr*prdst(n-1)
      end do

! Calculate the Mixing Ratio at the Reference Pressure Level

!  GCM1.7  6/28/01   spatially varying dust

      conr    = conrath_nu

      sum = 0._f8
      do n = 2,npdst
        if (prdst(n).lt.dust_tautot_ref_pr) then
          pave = 0.5_f8*(prdst(n)+prdst(n-1))
          sum  = sum +  &
                exp(conr*(1._f8-(dust_tautot_ref_pr/pave)))*  &
                   (prdst(n)-prdst(n-1))
        end if
        if (prdst(n).ge.dust_tautot_ref_pr) exit
      end do

      pave = 0.5_f8*(dust_tautot_ref_pr+prdst(n-1))
      sum  = sum + exp(conr*(1._f8-(dust_tautot_ref_pr/pave)))*(dust_tautot_ref_pr-prdst(n-1))

      qrdst0 = tautot_ij/sum

! Now calculate the mixing ratio at all other levels

      do n=1,npdst-1

! Region 1: Mixing ratio changes continuously through the layer

        if (dust_tautot_ref_pr.gt.prdst(n+1)) then
          pave     = 0.5_f8*(prdst(n+1)+prdst(n))
          qrdst(n) = qrdst0*exp(conr*(1._f8-(dust_tautot_ref_pr/pave)))
        end if

! Region 2: Reference pressure level within this layer. 

        if (dust_tautot_ref_pr.le.prdst(n+1).and.dust_tautot_ref_pr.ge.prdst(n)) then
          pave     = 0.5_f8*(prdst(n)+dust_tautot_ref_pr)
          pdif1    = dust_tautot_ref_pr-prdst(n)
          pdif2    = prdst(n+1)-dust_tautot_ref_pr
          qrdst(n) = qrdst0*(  &
                    exp(conr*(1._f8-(dust_tautot_ref_pr/pave)))*pdif1+pdif2) /  &
                    (prdst(n+1)-prdst(n))
        end if

! Region 3: Mixing ratio constant

        if (dust_tautot_ref_pr.lt.prdst(n)) then
          qrdst(n) = qrdst0
        end if

      end do

! Now compute the optical depths (taudst).

      taudst(1) = 0._f8

      do n=2,npdst
        taudst(n) = taudst(n-1) +  &
                        qrdst(n-1)*(prdst(n)-prdst(n-1))
      end do

      RETURN
      END subroutine ARCnRT_dustprofile


end module ARCnRT_dustprofile_mod
