module ARCnRT_filltaucum_mod

implicit none


contains


      SUBROUTINE ARCnRT_FILLTAUCUM

!     GCM v23   2010
!     Ames Mars GCM group
!     Jeffery Hollingsworth, PI
!     NASA Ames Research Center

!
!  PURPOSE
!      FILLTAUCUM fills the TAUCUM array - the dust column density of each
!      GCM level.  NDUSK(K) is the column density of the sublayer whose
!      bottom is level K, and whose top is at K-1.
!
use univ_kind_defs_mod, only: f8
      use ARCnRT_grid_h
      use ARCnRT_defines_h
      use ARCnRT_radinc_h
      use ARCnRT_radcommon_h
      use ARCnRT_standard_h
      use ARCnRT_comp3cmn_h
use ARCnRT_memory_mod, only: plev
use ARCnRT_jsrchgt_mod, only: ARCnRT_jsrchgt

      implicit none

      INTEGER :: Nstar, N

!     implicit none

!TIM      integer :: J, I, K
      integer :: K
!      integer, external :: ARCnRT_JSRCHGT
      real(f8) :: PSTAR, PSTAR1
 
!#=====================================================================

!  Dust optical depth at the bottom of each sub-layer.

      DO N=1,3
        TAUCUM(N) = 0._f8
        TAUREF(N) = 0._f8
      END DO

      DO K=4,L_LEVELS
        PSTAR     = plev(k)   !TIM! PSF*SIGMA(K)+PTROP
        PSTAR1    = MAX(PSTAR,PRDST(1))
        NSTAR     = MIN0(ARCnRT_JSRCHGT(NPDST-1,PRDST,1,PSTAR1)-1,NPDST-1)
        TAUCUM(K) = TAUDST(NSTAR)+(PSTAR1-PRDST(NSTAR))*  &
                   (TAUDST(NSTAR+1) - TAUDST(NSTAR))/  &
                   (PRDST(NSTAR+1)-PRDST(NSTAR))
        TAUREF(K) = TAUCUM(K) - TAUCUM(K-1)
      END DO

      TAUREF(L_LEVELS+1) = 0._f8

      RETURN
      END subroutine ARCnRT_filltaucum


end module ARCnRT_filltaucum_mod
