module ARCnRT_getdetau_mod

implicit none


contains


      SUBROUTINE ARCnRT_GETDETAU(DTDEL,TDEL,TAUCUMIN,WDEL,CDEL,UBAR0,DETAU)

!     GCM v23   2010
!     Ames Mars GCM group
!     Jeffery Hollingsworth, PI
!     NASA Ames Research Center

!======================================================================!

      use ARCnRT_grid_h
      use ARCnRT_radinc_h

      implicit none

      real(f8), intent(IN) :: DTDEL(L_NLAYRAD), TDEL(L_NLEVRAD),  &
           TAUCUMIN(L_LEVELS), WDEL(L_NLAYRAD), CDEL(L_NLAYRAD), ubar0
      real(f8), intent(OUT) :: detau

      REAL(f8) :: W0(L_NLAYRAD), COSBAR(L_NLAYRAD), DTAU(L_NLAYRAD)
      REAL(f8) :: TAU(L_NLEVRAD)
      REAL(f8) :: FACTOR, TAUCUM(L_LEVELS)

      integer :: NAYER, L, K

!======================================================================C
 
      NAYER  = L_NLAYRAD
      
!  Delta-Eddington Scaling

      FACTOR    = 1._f8 - WDEL(1)*CDEL(1)**2

      TAU(1)    = TDEL(1)*FACTOR
      TAUCUM(1) = 0._f8
      TAUCUM(2) = TAUCUMIN(2)*FACTOR
      TAUCUM(3) = TAUCUM(2) +(TAUCUMIN(3)-TAUCUMIN(2))*FACTOR

      DO L=1,L_NLAYRAD-1
        FACTOR      = 1._f8 - WDEL(L)*CDEL(L)**2
        W0(L)       = WDEL(L)*(1._f8-CDEL(L)**2)/FACTOR
        COSBAR(L)   = CDEL(L)/(1._f8+CDEL(L))
        DTAU(L)     = DTDEL(L)*FACTOR
        TAU(L+1)    = TAU(L)+DTAU(L)
        K           = 2*(L+1)
        TAUCUM(K)   = TAU(L+1)
        TAUCUM(K+1) = TAUCUM(K) + (TAUCUMIN(K+1)-TAUCUMIN(K))*FACTOR
      END DO

!  Bottom layer

      L             = L_NLAYRAD
      FACTOR        = 1._f8 - WDEL(L)*CDEL(L)**2
      W0(L)         = WDEL(L)*(1._f8-CDEL(L)**2)/FACTOR
      COSBAR(L)     = CDEL(L)/(1._f8+CDEL(L))
      DTAU(L)       = DTDEL(L)*FACTOR
      TAU(L+1)      = TAU(L)+DTAU(L)
      TAUCUM(2*L+1) = TAU(L+1)
      detau         = TAUCUM(2*L+1)

      return
      end subroutine ARCnRT_getdetau


end module ARCnRT_getdetau_mod
