module ARCnRT_jsrchgt_mod

implicit none


contains


      function ARCnRT_jsrchgt(N,SX,INC,TARGET)

!     GCM v23   2010
!     Ames Mars GCM group
!     Jeffery Hollingsworth, PI
!     NASA Ames Research Center

!      Find the first array element that is greater than the 
!      TARGET value.  If N < 1, then 0 is returned.  If no
!      value is found, N+1 is returned.  Replaces Cray version
!      on the workstation.
!      Bisection search implemented 09/29/93 to improve speed.
!
!      Started: 08/23/93
!
!      Input:
!        N      - Number of elements in array SX.
!        SX     - Array of numbers to be searched.  Assumed to be an
!                 ordered array.
!        INC    - Increment between elements of searched array.  Kept
!                 for compatibility with Cray call.
!        TARGET - Value searched for in array SX.
!
!      Output:
!        JSRCHGT - location in array SX where value of SX is first
!                   greater than the TARGET value.
!----------------------------------------------------------------------C

use univ_kind_defs_mod, only: f8

      implicit none
 
      integer, intent(IN) :: n, inc
      real(f8), intent(IN) :: SX(N), target

      integer :: ARCnRT_jsrchgt, ians, jl, jh, jm

!======================================================================C

      if(N.lt.1) then
        ians = 0
      elseif(TARGET.gt.SX(N)) then
        ians = N+1
      elseif(TARGET.lt.SX(1)) then
        ians = 1
      else

        JL = 1
        JH = N

do while (JH-JL.gt.1)
          JM = (JL+JH)/2

          if(TARGET.GT.SX(JM)) then
            JL = JM
            JM = (JL+JH)/2
          else
            JH = JM
            JM = (JL+JH)/2
          end if
end do

        if(TARGET.EQ.SX(JH)) JH = JH+1

        ians = JH
      end if

      ARCnRT_JSRCHGT = ians

      end function ARCnRT_jsrchgt


end module ARCnRT_jsrchgt_mod
