module ARCnRT_lagrange_mod

implicit none


contains


      subroutine ARCnRT_lagrange(x, xi, yi, ans)

!     GCM v23   2010
!     Ames Mars GCM group
!     Jeffery Hollingsworth, PI
!     NASA Ames Research Center

!  Lagrange interpolation - Polynomial interpolation at point x 
!  xi(1) <= x <= xi(4).  Yi(n) is the functional value at XI(n).

      use univ_kind_defs_mod, only: f8

      implicit none

      real(f8), intent(IN) :: x, xi(4), yi(4)
      real(f8), intent(OUT) :: ans

      real(f8) :: fm1, fm2, fm3, fm4

!======================================================================!

      fm1 = x - XI(1)
      fm2 = x - XI(2)
      fm3 = x - XI(3)
      fm4 = x - XI(4)

!  Get the "answer" at the requested X
 
      ans = fm2*fm3*fm4*YI(1)/                                         &
                      ((XI(1)-XI(2))*(XI(1)-XI(3))*(XI(1)-XI(4)))  +   &
            fm1*fm3*fm4*YI(2)/                                         &
                      ((XI(2)-XI(1))*(XI(2)-XI(3))*(XI(2)-XI(4)))  +   &
            fm1*fm2*fm4*YI(3)/                                         &
                      ((XI(3)-XI(1))*(XI(3)-XI(2))*(XI(3)-XI(4)))  +   &
            fm1*fm2*fm3*YI(4)/                                         &
                      ((XI(4)-XI(1))*(XI(4)-XI(2))*(XI(4)-XI(3))) 

      return
      end subroutine ARCnRT_lagrange


end module ARCnRT_lagrange_mod
