!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            modules.f90
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!     RT based on GCM v23   2010
!     Ames Mars GCM group
!     Jeffery Hollingsworth, PI
!     NASA Ames Research Center

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            version_h   
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      module ARCnRT_version_h  !TIM

      character(len=9), parameter :: version = "V23 r0029"

      end module ARCnRT_version_h  !TIM
      module ARCnRT_grid_h  !TIM

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                             grid_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      implicit none

!     Number of atmospheric layers
      integer :: l_layers

!     Number of atmospheric levels:   2 * L_LAYERS + 3
      integer :: l_levels  ! = 2*L_LAYERS+3

!     Number of dust particle sizes
      integer, parameter :: NDP = 2

      end module ARCnRT_grid_h  !TIM
      module ARCnRT_defines_h  !TIM

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            defines_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use ARCnRT_grid_h  !TIM
      implicit none

!     L_LAYERS + 1
      integer :: L_LAYERP1 ! = L_LAYERS+1

!     L_LAYERS - 1
      integer :: L_LAYERM1 ! = L_LAYERS-1

!     L_LEVELS - 1
      integer :: L_LEVELM1 ! = L_LEVELS-1

!     L_LEVELS - 2
      integer :: L_LEVELM2 ! = L_LEVELS-2

!     L_LEVELS - 3
      integer :: L_LEVELM3 ! = L_LEVELS-3

!     L_LEVELS - 4
      integer :: L_LEVELM4 ! = L_LEVELS-4

!     Number of fine-mesh layers used in calculating dust profiles
      integer, PARAMETER :: L_NPDST   = 100

      end module ARCnRT_defines_h  !TIM
      module ARCnRT_constants_h  !TIM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            constants_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use univ_kind_defs_mod, only: f8

      implicit none

!     The mathematical constant PI.
      real(f8), parameter :: PI = 3.1415926535897932_f8

!     Acceleration due to gravity (mks)
!TIM      real(f8), parameter :: GRAV = 3.72D0

!     Gas constant for mars.
      real(f8), parameter :: RGAS = 1.8902e+2_f8

!     Stefan-Boltzmann constant
      real(f8), parameter :: STBO = 5.67051e-8_f8

!     Heat capacity (or specific heat) of CO2 gas.
!     ( units of joules per ( kg * degrees kelvin))
!TIM      real(f8), parameter :: Cp = 7.3594D+2

!     Factor to convert pressures from millibars to Pascals
      real(f8), parameter :: SCALEP = 1.00e+2_f8

!     A radiation code conversion factor.
      real(f8), parameter :: Cmk = 3.51e+22_f8 

      end module ARCnRT_constants_h  !TIM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            radinc_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      module ARCnRT_radinc_h  !TIM

      use univ_kind_defs_mod, only: f8
      use ARCnRT_grid_h, only: L_LAYERS  !TIM
      implicit none

!======================================================================C
!
!     RADINC.H    RADiation INCludes
!
!     Includes for the radiation code; RADIATION LAYERS, LEVELS,
!     number of spectral intervals. . .
!
!     GCM2.0  Feb 2003
! 
!======================================================================C

!     RADIATION parameters

!     In radiation code, layer 1 corresponds to the stratosphere.  Level
!     1 is the top of the stratosphere.  The dummy layer is at the same
!     temperature as the (vertically isothermal) stratosphere, and
!     any time it is explicitly needed, the appropriate quantities will
!     be dealt with (aka "top". . .)

!     L_NLEVRAD corresponds to the surface - i.e., the GCM Level that
!     is at the surface.  PLEV(L_NLEVRAD) = P(J,I)+PTROP, 
!     PLEV(2) = PTROP, PLEV(1) = ptrop

!     L_NLAYRAD is the number of radiation code layers
!     L_NLEVRAD is the number of radiation code levels.  Level N is the
!               top of layer N. 
!
!     L_NSPECTI is the number of IR spectral intervals
!     L_NSPECTV is the number of visible (or Solar) spectral intervals
!     L_NGAUSS  is the number of Gauss points for K-coefficients
!               GAUSS POINT 9 (aka the last one) is the special case
!     L_NWNGI   is L_NSPECTI*L_NGAUSS;  the total number of "intervals"
!               in the IR
!     L_NWNGV   is L_NSPECTV*L_NGAUSS;  the total number of "intervals"
!               in the VISIBLE
!
!     L_NPREF   is the number of reference pressures that the 
!               k-coefficients are calculated on
!     L_PINT    is the number of Lagrange interpolated reference
!               pressures for the CO2 k-coefficients.
!     L_NTREF   is the number of refernce temperatures for the
!               k-coefficients
!     L_TAUMAX  is the largest optical depth - larger ones are set
!               to this value.
!
!     MAXEXP    The largest value (TAU/MU) used in exp(-TAU/MU), where
!               TAU is the optical depth, and MU is the cosine of the
!               zenith angle.
!
!     L_REFH2O  The number of different water-mixing ratio values for
!               the k-coefficients that are now CO2+H2O. 
!
!     L_NREFI   The spectral interval number of the IR reference
!               wavelength (i.e. the 9 micron band) (8-12 microns)
!
!     L_NREFV   The spectral interval number of the visible reference
!               wavelength (i.e. the 0.67 micron band) 
!
!----------------------------------------------------------------------C

      integer :: L_NLAYRAD ! = L_LAYERS+1
      integer :: L_NLEVRAD ! = L_LAYERS+2
      
      integer, parameter :: L_NSPECTI =  5
      integer, parameter :: L_NSPECTV =  7
      integer, parameter :: L_NGAUSS  = 17

      integer, parameter :: L_NPREF   = 11
      integer, parameter :: L_NTREF   =  7
      integer, parameter :: L_TAUMAX  = 35

      real(f8), parameter  :: MAXEXP    = 35._f8

      integer, parameter :: L_PINT    = 51

      integer, parameter :: L_REFH2O  = 10

      integer, parameter :: L_NREFV   = 6
      integer, parameter :: L_NREFI   = 4

      end module ARCnRT_radinc_h  !TIM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            radcommon_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      module ARCnRT_radcommon_h  !TIM

      use univ_kind_defs_mod, only: f8
      use ARCnRT_grid_h, only: L_LEVELS  !TIM
      use ARCnRT_radinc_h   !TIM
      implicit none

!----------------------------------------------------------------------C
!
!                             radcommon.h
!                         FORTRAN PARAMETERS
!                          GCM2.0  Feb 2003
!
!----------------------------------------------------------------------C
!
!     WNOI       - Array of wavenumbers at the spectral interval
!                  centers for the infrared.  Array is NSPECTI
!                  elements long.
!     DWNI       - Array of "delta wavenumber", i.e., the width,
!                  in wavenumbers (cm^-1) of each IR spectral
!                  interval.  NSPECTI elements long.
!     WAVEI      - Array (NSPECTI elements long) of the wavelenght
!                  (in microns) at the center of each IR spectral
!                  interval.
!     WNOV       - Array of wavenumbers at the spectral interval
!                  center for the VISIBLE.  Array is NSPECTV
!                  elements long.
!     DWNV       - Array of "delta wavenumber", i.e., the width,
!                  in wavenumbers (cm^-1) of each VISIBLE spectral
!                  interval.  NSPECTV elements long.
!     WAVEV      - Array (NSPECTV elements long) of the wavelenght
!                  (in microns) at the center of each VISIBLE spectral
!                  interval.
!     SOLARF     - Array (NSPECTV elements) of solar flux (W/M^2) in
!                  each spectral interval.  Values are for 1 AU, and
!                  are scaled to the Mars distance elsewhere.
!     TAURAY     - Array (NSPECTV elements) of the pressure-independent
!                  part of Rayleigh scattering optical depth.
!     PTOP       - Pressure at the top of the radiation code coordinate;
!                  = smallest k-coefficient pressure (1.0E-6 mbar)
!     FZEROI     - Fraction of zeros in the IR CO2 k-coefficients, for
!                  each temperature, pressure, and spectral interval
!     FZEROV     - Fraction of zeros in the VISIBLE CO2 k-coefficients, for
!                  each temperature, pressure, and spectral interval
!
!     AEROSOL RADIATIVE OPTICAL CONSTANTS
!     Values are at the wavelenght interval center
!
!     MIE SCATTERING - Size distribution weighted
!     Qextv    - Extinction efficiency - in the visible.
!     QextREF  - Reference visible wavelength (.67 micron band)
!     Qscatv   - Scattering efficiency - in the visible.
!     WV       - Single scattering albedo - in the visible.
!     GV       - Asymmetry parameter - in the visible.
!
!     Qexti    - Extinction efficiency - in the infrared.
!     Qscati   - Scattering efficiency - in the infrared.
!     WI       - Single scattering albedo - in the infrared.
!     GI       - Asymmetry parameter - in the infrared.
!     
!  Water ice clouds
!
!     MIE SCATTERING - Size distribution weighted
!     Qextvc   - Extinction efficiency - in the visible.
!     QextREFc - Reference visible wavelength (.67 micron band)
!     Qscatvc  - Scattering efficiency - in the visible.
!     WVc      - Single scattering albedo - in the visible.
!     GVc      - Asymmetry parameter - in the visible.
!
!     Qextic   - Extinction efficiency - in the infrared.
!     Qscatic  - Scattering efficiency - in the infrared.
!     WIc      - Single scattering albedo - in the infrared.
!     GIc      - Asymmetry parameter - in the infrared.
!     

      REAL(f8) :: WNOI(L_NSPECTI), DWNI(L_NSPECTI), WAVEI(L_NSPECTI)
      REAL(f8) :: WNOV(L_NSPECTV), DWNV(L_NSPECTV), WAVEV(L_NSPECTV)
      REAL(f8) :: SOLARF(L_NSPECTV), TAURAY(L_NSPECTV)

      real(f8) :: CO2I(L_NTREF,L_PINT,L_REFH2O,L_NSPECTI,L_NGAUSS)
      real(f8) :: CO2V(L_NTREF,L_PINT,L_REFH2O,L_NSPECTV,L_NGAUSS)
      real(f8) :: FZEROI(L_NSPECTI)
      real(f8) :: FZEROV(L_NSPECTV)
      real(f8) :: PGASREF(L_NPREF), TGASREF(L_NTREF)

      real(f8) :: qextv(L_NSPECTV), qscatv(L_NSPECTV), wv(L_NSPECTV)
      real(f8) :: gv(L_NSPECTV)
      real(f8) :: QextREF

      real(f8) :: qexti(L_NSPECTI), qscati(L_NSPECTI), wi(L_NSPECTI)
      real(f8) :: gi(L_NSPECTI)

      real(f8) :: qextvc(L_NSPECTV), qscatvc(L_NSPECTV), wvc(L_NSPECTV)
      real(f8) :: gvc(L_NSPECTV)
      real(f8) :: QextREFc

      real(f8) :: qextic(L_NSPECTI), qscatic(L_NSPECTI), wic(L_NSPECTI)
      real(f8) :: gic(L_NSPECTI)

      real(f8) :: planckir(L_NSPECTI,8501)

      real(f8), allocatable, dimension(:) :: TAUREF    ! (L_LEVELS+1)
      real(f8) :: PFGASREF(L_PINT)

!     Hemispheric mean in the IR
      real(f8), parameter :: UBARI = 0.5_f8

!     These are for the Gauss-split 0.95 case

      real(f8), parameter :: GWEIGHT(L_NGAUSS) = [                       &
                      4.8083554740e-02_f8, 1.0563099137e-01_f8,              &
                      1.4901065679e-01_f8, 1.7227479710e-01_f8,              &
                      1.7227479710e-01_f8, 1.4901065679e-01_f8,              &
                      1.0563099137e-01_f8, 4.8083554740e-02_f8,              &
                      2.5307134073e-03_f8, 5.5595258613e-03_f8,              &
                      7.8426661469e-03_f8, 9.0670945845e-03_f8,              &
                      9.0670945845e-03_f8, 7.8426661469e-03_f8,              &
                      5.5595258613e-03_f8, 2.5307134073e-03_f8,  0._f8 ]  

!     These are for the CO2+H2O k-coefficients

      real(f8), parameter :: WREFCO2(L_REFH2O) = [                       &
                     9.999999e-1_f8, 9.99999e-1_f8, 9.9999e-1_f8, 9.999e-1_f8,     &
                     9.99e-1_f8, 9.9e-1_f8, 9.e-1_f8, 8.e-1_f8, 7.e-1_f8, 6.e-1_f8 ]

      real(f8), parameter :: WREFH2O(L_REFH2O) = [                       &
                     1.e-7_f8, 1.e-6_f8, 1.e-5_f8, 1.e-4_f8, 1.e-3_f8, 1.e-2_f8,   &
                     1.e-1_f8, 2.e-1_f8, 3.e-1_f8, 4.e-1_f8                  ]

!     If the CO2 optical depth (top to the surface) is less than
!     this value, we place that Gauss-point into the "zeros"
!     channel.  NRC parameter.

!  TLIMITS - TLIMIT for solar part of the spectrum
!  TLIMITI - TLIMIT for the IR

      real(f8), parameter :: TLIMITS = 1.e-32_f8
      real(f8), parameter :: TLIMITI = 5.e-32_f8

      end module ARCnRT_radcommon_h  !TIM
      module ARCnRT_fccsave_h  !TIM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                              fccsave_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!  Cloud Radiative properties

      use univ_kind_defs_mod, only: f8
      use ARCnRT_grid_h,   only: L_LEVELS, L_LAYERS  !TIM
      use ARCnRT_radinc_h, only: L_NSPECTV, L_NSPECTI, L_NGAUSS  !TIM
      implicit none

      real(f8), allocatable, dimension(:) ::  TAUREFCLD !(L_LEVELS+1)
      real(f8), allocatable, dimension(:) ::  QEXTREFCLD !(L_LEVELS+1)

      real(f8), allocatable, dimension(:,:) ::  QXVCLD !(L_LEVELS+1,L_NSPECTV)
      real(f8), allocatable, dimension(:,:) ::  QSVCLD !(L_LEVELS+1,L_NSPECTV)
      real(f8), allocatable, dimension(:,:) ::  GVCLD !(L_LEVELS+1,L_NSPECTV)

      real(f8), allocatable, dimension(:,:) ::  QXICLD !(L_LEVELS+1,L_NSPECTI)
      real(f8), allocatable, dimension(:,:) ::  QSICLD !(L_LEVELS+1,L_NSPECTI)
      real(f8), allocatable, dimension(:,:) ::  GICLD !(L_LEVELS+1,L_NSPECTI)

!  Dust Radiative properties

      real(f8), allocatable, dimension(:) ::  QEXTREFDST !(L_LEVELS+1)

      real(f8), allocatable, dimension(:,:) ::  QXVDST !(L_LEVELS+1,L_NSPECTV)
      real(f8), allocatable, dimension(:,:) ::  QSVDST !(L_LEVELS+1,L_NSPECTV)
      real(f8), allocatable, dimension(:,:) ::  GVDST !(L_LEVELS+1,L_NSPECTV)

      real(f8), allocatable, dimension(:,:) ::  QXIDST !(L_LEVELS+1,L_NSPECTI)
      real(f8), allocatable, dimension(:,:) ::  QSIDST !(L_LEVELS+1,L_NSPECTI)
      real(f8), allocatable, dimension(:,:) ::  GIDST !(L_LEVELS+1,L_NSPECTI)

      end module ARCnRT_fccsave_h   !TIM
      module ARCnRT_standard_h  !TIM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            standard_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use univ_kind_defs_mod, only: f8
      use ARCnRT_grid_h  !TIM
      use ARCnRT_defines_h  !TIM

      implicit none

      real(f8), allocatable, dimension(:) :: TAUCUM !(L_LEVELS)
!%      real(f8) :: SIGMA(L_LEVELS)
      real(f8) :: PRDST(L_NPDST), TAUDST(L_NPDST)

!     Spatially varying dust - TAUSURF(J,I) is the (visible 0.67 micron)
!     dust opacity at the surface for each grid point
!     TAUTOT added 6/28/01  GCM1.7
!     TAUTOT is the dust opacity at the RPTAU reference pressure
!     for each grid point.

      REAL(f8) :: TAUTOT, RPTAU

!     variables that were in bkdata.f
!
!   ALICEN  :  SURFACE ALBEDO FOR CO2 ICE ON GROUND - Northern cap
!   ALICES  :  SURFACE ALBEDO FOR CO2 ICE ON GROUND - Southern cap
!
!   CP      :  HEAT CAPACITY (OR SPECIFIC HEAT) OF CO2 GAS.
!                  ( UNITS OF JOULES PER ( KG * DEGREES KELVIN))
!
!   DECMAX  :  MAXIMUM SOLAR DECLINATION.
!   DSIG    :  COORDINATE FOR ATM. LAYERS.
!   DTM     :  UNROUNDED DEFAULT NUMBER OF MINUTES PER TIME STEP.
!
!   ECCN    :  ORBITAL ECCENTRICITY FOR MARS.
!   EGOGND  :  EMISSIVITY OF BARE GROUND OUTSIDE 15 MICRON BAND WIDTH.
!   EG15GND :  EMISSIVITY OF BARE GROUND INSIDE 15 MICRON BAND WIDTH.
!   EG15CO2 :  EMISSIVITY OF GROUND WITH CO2 ICE INSIDE 15 MICRON BAND
!                   WIDTH.
!   EPS     :  SHEAR < EPS  MEANS  NEAR ZERO SHEAR.
!
!   GRAV    :  ACCELERATION DUE TO GRAVITY (MKS)
!
!   IM      :  DEFAULT LONGITUDINAL DIMENSION OF GRID
!   ISIZE   :  X DIMENSION OF GRID ARRAYS
!
!   JM      :  DEFAULT LATITUDINAL DIMENSION OF GRID
!   JSIZE   :  Y DIMENSION OF GRID ARRAYS
!
!   KAPA    :  A THERMODYNAMIC CONSTANT.
!
!   LAYERS  :  Z DIMENSION OF GRID ARRAYS
!
!   NCYCLE  :  DEFAULT NUMBER OF TIME STEPS PER TIME-STEPPING CYCLE.
!                   (USED IN STEP.)
!   NC3     :  THE FULL COMP3 ROUTINE IS EXECUTED ONLY ONCE PER NC3
!                   TIME STEPS.
!   NLAY    :  DEFAULT NUMBER OF ATMOSPHERIC LAYERS
!   NSIZE2D :  SIZE OF 2-D GRID ARRAYS =ISIZE*JSIZE.
!   NTAPE   :  Number of tape volume (ex. fort.11_032   NTAPE = 32)
!              GCM1.7 update
!
!   PI      :  THE MATHEMATICAL CONSTANT PI.
!   PTROP   :  DEFAULT PRESSURE AT THE TROPOPAUSE.
!
!   RAD     :  RADIUS OF MARS IN KM. CONVERTED TO METERS IN
!                   SUBROUTINE INPUT.
!   RGAS    :  GAS CONSTANT FOR MARS.
!   ROT     :  ANGULAR (ROTATIONAL) VELOCITY OF MARS IN RADIANS
!                   PER SECOND.
!   ROTPER  :  ROTATIONAL PERIOD IN MARTIAN HOURS.
!   RPTAU   :  Reference Pressure optical depth
!
!   SCALEP  :  MULTIPLY BY 100 TO CONVERT PRESSURE FROM MILLIBARS
!                   TO PASCALS.
!   STBO    :  STEFAN-BOLTZMAN CONSTANT.
!
!   TAUCRT  :  ICE-CLOUD OPTICAL DEPTH
!   TAUH    :  DEFAULT NUMBER OF HOURS BETWEEN WRITING OUTPUT TO HISTORY
!                     TAPE.
!   TAUID   :  DEFAULT STARTING DAY FOR A GCM RUN.
!   TAUIH   :  DEFAULT STARTING HOUR FOR A GCM RUN.
!   TAUTOT  :  TOTAL GLOBAL OPTICAL DEPTH
!   TDINCR  :  MINIMUM AND INCREMENT VALUES
!   TDMIN   :  FOR SETS OF DUST OPT. DEPTHS.
!   THD     :  THERMAL DIFFUSIVITY
!   TIINCR  :  MINIMUM AND INCREMENT VALUES
!   TIMIN   :  FOR SET OF ICE OPT. DEPTHS.
!
!   end original bkdata.f comments

!     THE C ARRAY IS A SMALLER RECORD-TYPE STRUCTURE USED TO GROUP
!     TOGETHER A NUMBER OF GLOBALLY USED CONSTANTS.

      real(f8) :: RSDIST

!%      real(f8), parameter :: dsig(L_LAYERS) = [                          &
!%                          3.2718D-07, 5.3944D-07, 8.8940D-07,          &
!%                          1.4663D-06, 2.4175D-06, 3.9859D-06,          &
!%                          6.5718D-06, 1.0835D-05, 1.7863D-05,          &
!%                          2.9452D-05, 4.8556D-05, 8.0060D-05,          &
!%                          1.3199D-04, 2.1763D-04, 3.5879D-04,          &
!%                          5.9155D-04, 9.7530D-04, 1.6080D-03,          &
!%                          2.6512D-03, 4.3714D-03, 7.2060D-03,          &
!%                          1.1882D-02, 1.9590D-02, 3.2298D-02,          &
!%                          5.3245D-02, 8.7800D-02, 1.4475D-01,          &
!%                          2.3865D-01, 3.9347017648D-01 ]

      real(f8) :: MWCO2 = 4.41e-2_f8
      real(f8) :: MWH2O = 1.80153e-2_f8
      real(f8) :: MWRATIO

      integer :: NPDST     = L_NPDST
      integer :: LAYERS   ! = L_LAYERS

      integer :: NLAY    ! = L_LAYERS
      real(f8) :: PSF
      real(f8) :: PTROP

!  Standard value = 0.03
!     DATA  CONRNU / 0.03  /   ! Standard value  ~25km half-height
!     DATA  CONRNU / 0.003 /   ! ~50 km half-height
!     DATA  CONRNU / 0.5   /   ! ~10 km half-height

      real(f8) :: CONRNU

      end module ARCnRT_standard_h  !TIM
      module ARCnRT_comp3cmn_h  !TIM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            comp3cmn_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use univ_kind_defs_mod, only: f8
      use ARCnRT_grid_h, only: L_LAYERS, L_LEVELS  !TIM
      implicit none

!%      real(f8) :: PL(L_LEVELS)
!%      real(f8) :: TL(L_LEVELS)

      end module ARCnRT_comp3cmn_h  !TIM
      module ARCnRT_cldcommon_h  !TIM
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!                            cldcommon_h
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      use ARCnRT_grid_h, only: L_LAYERS  !TIM
      use ARCnRT_radinc_h, only: L_NSPECTI, L_NSPECTV  !TIM
      implicit none

      integer, parameter :: nlonv   = L_NSPECTV
      integer, parameter :: nloni   = L_NSPECTI

      end module ARCnRT_cldcommon_h  !TIM
