!```````````````````````````````````````````````````````````````````
module aerad_aeropd_mod

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine calc_abp_opacities(ext_n1, ext_n2, ext_n3, ext_naersclr,  &
     ext_aersclp, ext_kz)

!+ Computes opacities of all model layers due to airborne particles (per species
!   and spectral interval)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4, ONE_f4
use univ_convert_units_mod, only: cm3_to_m3_f4
use univ_num_search_mod, only: locidx_1D_mono_type02
use aertxrad_mem_mod, only: nz1_rad, naerspc_rad, spc_begiv, dzc, spc_endiv,  &
     i_curr, j_curr, spc_xsarea_fact, tau_aer, singscat_albd, asym_fact,  &
     spc_radtx, lw_iwend, lw_iwbeg, sw_iwend, sw_iwbeg, sw_on, spc_nbins,  &
     bdst_Nconc, sw_abp_optprop, lw_abp_optprop

implicit none

!=== Argument declarations:

integer, intent(IN) :: ext_n1, ext_n2, ext_n3, ext_naersclr, ext_kz
real(f4), dimension(ext_n1,ext_n2,ext_n3,ext_naersclr), intent(IN) ::   &
     ext_aersclp                 ! In CGS: [g/cm^3 OR #/cm^3]

!== Local declarations:

integer :: k, ik, iw, iwa, ip, ib, ifr, ifb, iv
real(f4) :: c1, vtmp1, fint_r, fint_b, xsnorm_fact, vtmp2, vtmp, tgt_fvws_r,  &
     tgt_fvws_b
real(f4), dimension(3) :: opp_sum

!=== Executable statements:

ik = ext_kz
do k=1,nz1_rad

! "Longwave":
   
   iw = 0
   do iwa=lw_iwbeg,lw_iwend   ! Only calculate for relevant wavelength bands
      iw = iw+1

      ! Determine total (up+down) flux vs. wavelength slopes:
      tgt_fvws_r = 0._f4    ! For now, just a placeholder
      tgt_fvws_b = 0._f4    ! For now, just a placeholder

      do ip=1,naerspc_rad

         if (spc_radtx(ip)) then  ! Only perform if specie currently RT-active

            ! Determine total (up+down) flux vs. wavelength slope interpolation
            !  parameters:
               !## fvws_r:
            if (tgt_fvws_r >= lw_abp_optprop(ip)%max_fvws_r) then
               ifr = lw_abp_optprop(ip)%n_fvws_r-1      ! Clip to max
            else if (tgt_fvws_r <= lw_abp_optprop(ip)%min_fvws_r) then
               ifr = 1     ! Clip to min
            else     ! Search for appropriate index
               ifr = locidx_1D_mono_type02(lw_abp_optprop(ip)%fvws_r,  &
                    tgt_fvws_r, lw_abp_optprop(ip)%search_beg_idx_fvws_r)
            end if
            fint_r = (tgt_fvws_r-lw_abp_optprop(ip)%fvws_r(ifr))/  &
                 lw_abp_optprop(ip)%d_fvws_r(ifr)
               !## fvws_b:
            if (tgt_fvws_b >= lw_abp_optprop(ip)%max_fvws_b) then
               ifb = lw_abp_optprop(ip)%n_fvws_b-1      ! Clip to max
            else if (tgt_fvws_b <= lw_abp_optprop(ip)%min_fvws_b) then
               ifb = 1     ! Clip to min
            else     ! Search for appropriate index
               ifb = locidx_1D_mono_type02(lw_abp_optprop(ip)%fvws_b,  &
                    tgt_fvws_b, lw_abp_optprop(ip)%search_beg_idx_fvws_b)
            end if
            fint_b = (tgt_fvws_b-lw_abp_optprop(ip)%fvws_b(ifb))/  &
                 lw_abp_optprop(ip)%d_fvws_b(ifb)

            ib = 0; xsnorm_fact = ZERO_f4
            opp_sum = ZERO_f4    ! Entire array at once

            if (spc_begiv(ip) /= -999) then
               do iv=spc_begiv(ip),spc_endiv(ip)
                  ib = ib+1

                  ! (#/m^3)*(pi*r^2)
                  c1 = ext_aersclp(ik,i_curr,j_curr,iv)/cm3_to_m3_f4*  &
                       spc_xsarea_fact(ib,ip)

                  vtmp1 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_ext(ifr,ifb,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_ext(ifr+1,ifb,ib,iw)
                  vtmp2 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_ext(ifr,ifb+1,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_ext(ifr+1,ifb+1,ib,iw)
                  vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                  opp_sum(1) = opp_sum(1)+vtmp*c1  ! Cross-sect weighted Qext

                  vtmp1 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_sca(ifr,ifb,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_sca(ifr+1,ifb,ib,iw)
                  vtmp2 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_sca(ifr,ifb+1,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_sca(ifr+1,ifb+1,ib,iw)
                  vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                  opp_sum(2) = opp_sum(2)+vtmp*c1  ! Cross-sect weighted Qsca

                  vtmp1 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%gQsca(ifr,ifb,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%gQsca(ifr+1,ifb,ib,iw)
                  vtmp2 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%gQsca(ifr,ifb+1,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%gQsca(ifr+1,ifb+1,ib,iw)
                  vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                  opp_sum(3) = opp_sum(3)+vtmp*c1  ! Cross-sect weighted gQsca

                  xsnorm_fact = xsnorm_fact+c1
               end do
            else
               do ib=1,spc_nbins(ip)
                  ! (#/m^3)*(pi*r^2)
                  c1 = bdst_Nconc(k,ib)/cm3_to_m3_f4*  &
                       spc_xsarea_fact(ib,ip)

                  vtmp1 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_ext(ifr,ifb,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_ext(ifr+1,ifb,ib,iw)
                  vtmp2 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_ext(ifr,ifb+1,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_ext(ifr+1,ifb+1,ib,iw)
                  vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                  opp_sum(1) = opp_sum(1)+vtmp*c1  ! Cross-sect weighted Qext

                  vtmp1 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_sca(ifr,ifb,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_sca(ifr+1,ifb,ib,iw)
                  vtmp2 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%Q_sca(ifr,ifb+1,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%Q_sca(ifr+1,ifb+1,ib,iw)
                  vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                  opp_sum(2) = opp_sum(2)+vtmp*c1  ! Cross-sect weighted Qsca

                  vtmp1 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%gQsca(ifr,ifb,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%gQsca(ifr+1,ifb,ib,iw)
                  vtmp2 = (ONE_f4-fint_r)*  &
                       lw_abp_optprop(ip)%gQsca(ifr,ifb+1,ib,iw)+  &
                       fint_r*lw_abp_optprop(ip)%gQsca(ifr+1,ifb+1,ib,iw)
                  vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                  opp_sum(3) = opp_sum(3)+vtmp*c1  ! Cross-sect weighted gQsca

                  xsnorm_fact = xsnorm_fact+c1
               end do
            end if

            ! Airborne particle optical depth:
            tau_aer(ip,iwa,k) = opp_sum(1)*dzc(k)

            ! Particle cross-sectional area weighted single-scattering albedo
            !  and asymmetry factor:

            if (xsnorm_fact >= 1.e-18_f4) then
               singscat_albd(ip,iwa,k) = opp_sum(2)/opp_sum(1)
               asym_fact(ip,iwa,k) = opp_sum(3)/opp_sum(2)
            else ! Aerosol opacity near zero; set to some representative values
               singscat_albd(ip,iwa,k) = ONE_f4
               asym_fact(ip,iwa,k) = 0.5_f4
            end if

         end if

      end do
   end do

! "Shortwave":

   if (sw_on) then
      iw = 0
      do iwa=sw_iwbeg,sw_iwend   ! Only calculate for relevant wavelength bands
         iw = iw+1

         ! Determine total (up+down) flux vs. wavelength slopes:
         tgt_fvws_r = 0._f4    ! For now, just a placeholder
         tgt_fvws_b = 0._f4    ! For now, just a placeholder

         do ip=1,naerspc_rad 

            if (spc_radtx(ip)) then   ! Only perform if specie is RT-active now
 
               ! Determine total (up+down) flux vs. wavelength slope
               !  interpolation parameters:
                  !## fvws_r:
               if (tgt_fvws_r >= sw_abp_optprop(ip)%max_fvws_r) then
                  ifr = sw_abp_optprop(ip)%n_fvws_r-1      ! Clip to max
               else if (tgt_fvws_r <= sw_abp_optprop(ip)%min_fvws_r) then
                  ifr = 1     ! Clip to min
               else     ! Search for appropriate index
                  ifr = locidx_1D_mono_type02(sw_abp_optprop(ip)%fvws_r,  &
                       tgt_fvws_r, sw_abp_optprop(ip)%search_beg_idx_fvws_r)
               end if
               fint_r = (tgt_fvws_r-sw_abp_optprop(ip)%fvws_r(ifr))/  &
                    sw_abp_optprop(ip)%d_fvws_r(ifr)
                  !## fvws_b:
               if (tgt_fvws_b >= sw_abp_optprop(ip)%max_fvws_b) then
                  ifb = sw_abp_optprop(ip)%n_fvws_b-1      ! Clip to max
               else if (tgt_fvws_b <= sw_abp_optprop(ip)%min_fvws_b) then
                  ifb = 1     ! Clip to min
               else     ! Search for appropriate index
                  ifb = locidx_1D_mono_type02(sw_abp_optprop(ip)%fvws_b,  &
                       tgt_fvws_b, sw_abp_optprop(ip)%search_beg_idx_fvws_b)
               end if
               fint_b = (tgt_fvws_b-sw_abp_optprop(ip)%fvws_b(ifb))/  &
                    sw_abp_optprop(ip)%d_fvws_b(ifb)

               ib = 0; xsnorm_fact = ZERO_f4
               opp_sum = ZERO_f4    ! Entire array at once

               if (spc_begiv(ip) /= -999) then
                  do iv=spc_begiv(ip),spc_endiv(ip)
                     ib = ib+1

                     ! (#/m^3)*(pi*r^2)
                     c1 = ext_aersclp(ik,i_curr,j_curr,iv)/cm3_to_m3_f4*  &
                          spc_xsarea_fact(ib,ip)
                     
                     vtmp1 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_ext(ifr,ifb,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_ext(ifr+1,ifb,ib,iw)
                     vtmp2 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_ext(ifr,ifb+1,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_ext(ifr+1,ifb+1,ib,iw)
                     vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                     opp_sum(1) = opp_sum(1)+vtmp*c1 ! Cross-sect weighted Qext
                     
                     vtmp1 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_sca(ifr,ifb,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_sca(ifr+1,ifb,ib,iw)
                     vtmp2 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_sca(ifr,ifb+1,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_sca(ifr+1,ifb+1,ib,iw)
                     vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                     opp_sum(2) = opp_sum(2)+vtmp*c1 ! Cross-sect weighted Qsca

                     vtmp1 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%gQsca(ifr,ifb,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%gQsca(ifr+1,ifb,ib,iw)
                     vtmp2 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%gQsca(ifr,ifb+1,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%gQsca(ifr+1,ifb+1,ib,iw)
                     vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                     opp_sum(3) = opp_sum(3)+vtmp*c1 ! Cross-sect weighted gQsca
                     
                     xsnorm_fact = xsnorm_fact+c1
                  end do
               else
                  do ib=1,spc_nbins(ip)

                     ! (#/m^3)*(pi*r^2)
                     c1 = bdst_Nconc(k,ib)/cm3_to_m3_f4*  &
                          spc_xsarea_fact(ib,ip)
                     
                     vtmp1 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_ext(ifr,ifb,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_ext(ifr+1,ifb,ib,iw)
                     vtmp2 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_ext(ifr,ifb+1,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_ext(ifr+1,ifb+1,ib,iw)
                     vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                     opp_sum(1) = opp_sum(1)+vtmp*c1 ! Cross-sect weighted Qext
                     
                     vtmp1 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_sca(ifr,ifb,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_sca(ifr+1,ifb,ib,iw)
                     vtmp2 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%Q_sca(ifr,ifb+1,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%Q_sca(ifr+1,ifb+1,ib,iw)
                     vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                     opp_sum(2) = opp_sum(2)+vtmp*c1 ! Cross-sect weighted Qsca

                     vtmp1 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%gQsca(ifr,ifb,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%gQsca(ifr+1,ifb,ib,iw)
                     vtmp2 = (ONE_f4-fint_r)*  &
                          sw_abp_optprop(ip)%gQsca(ifr,ifb+1,ib,iw)+  &
                          fint_r*sw_abp_optprop(ip)%gQsca(ifr+1,ifb+1,ib,iw)
                     vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
                     opp_sum(3) = opp_sum(3)+vtmp*c1 ! Cross-sect weighted gQsca
                     
                     xsnorm_fact = xsnorm_fact+c1
                  end do
               end if

               ! Airborne particle optical depth:
               tau_aer(ip,iwa,k) = opp_sum(1)*dzc(k)

               ! Particle cross-sectional area weighted single-scattering
               !  albedo and asymmetry factor:
    
               if (xsnorm_fact >= 1.e-18_f4) then
                  singscat_albd(ip,iwa,k) = opp_sum(2)/opp_sum(1)
                  asym_fact(ip,iwa,k) = opp_sum(3)/opp_sum(2)
               else ! Aerosol opacity near zero; set to representative values
                  singscat_albd(ip,iwa,k) = ONE_f4
                  asym_fact(ip,iwa,k) = 0.5_f4
               end if
    
            end if

         end do
      end do
   end if

   ik = ik-1
end do

end subroutine calc_abp_opacities


!```````````````````````````````````````````````````````````````````
subroutine calc_abp_opd_specific(m1a, m1z, spc_nb, case_str, ka, kz,  &
     spc_Nconc, dzt, N_fact, tau_sum)

!+ Computes total extinction optical depth of model column at the given
!   spectral interval due to specified airborne particles

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4, ONE_f4
use univ_convert_units_mod, only: cm3_to_m3_f4
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_FILE_PATH_LEN, MAX_PROC_NAME_LEN
use univ_num_search_mod, only: locidx_1D_mono_type02
use aertxrad_mem_mod, only: spc_xsarea_fact,  &
     sw_abp_optprop, nlw_wavlnrng, lw_abp_optprop, aeradID_dust,  &
     aeradID_H2Oi, aeradID_CO2i, spc_ctype, naerspc_rad

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1a, m1z, ka, kz, spc_nb
real(f4), dimension(m1a:m1z,spc_nb), intent(IN) :: spc_Nconc  ! In CGS: [#/cm^3]
real(f4), dimension(m1a:m1z), intent(IN) :: dzt    ! [m]
real(f4), intent(IN) :: N_fact
character(len=*), intent(IN) :: case_str

real(f4), intent(OUT) :: tau_sum

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'calc_abp_opd_specific'

integer, parameter :: n_specific_cases = 6
character(len=8), parameter, dimension(n_specific_cases) :: spec_case_str =  &
     (/ 'bdst_9um', 'bdst_vis', 'dust_9um', 'dust_vis', 'H2Oi_vis',  &
        'CO2i_vis' /)
integer, parameter, dimension(n_specific_cases) :: spec_case_iwa =  &
     (/ 4, 11, 4, 11, 11, 11 /)
integer, parameter, dimension(n_specific_cases) :: spec_case_iptog =  &
     (/ -1, -1, aeradID_dust, aeradID_dust, aeradID_H2Oi, aeradID_CO2i /)
real(f4), parameter, dimension(n_specific_cases) ::  &
     spec_case_fvws_r = (/ 0._f4, 0._f4, 0._f4, 0._f4, 0._f4, 0._f4 /),  &
     spec_case_fvws_b = (/ 0._f4, 0._f4, 0._f4, 0._f4, 0._f4, 0._f4 /)

integer :: k, iw, ib, kinc, ifr, ifb, ic, ia, ip
real(f4) :: c1, vtmp, vtmp1, vtmp2, fint_r, fint_b, tgt_fvws_r, tgt_fvws_b
real(f4), dimension(3) :: opp_sum
logical :: use_sw_optprop
character(len=MAX_FILE_PATH_LEN) :: tmp_str

!=== Executable statements:

! Determine specific case index ('ia'):
ic = len_trim(case_str)
do ia=1,n_specific_cases
   if (trim(spec_case_str(ia)) == case_str(1:ic)) exit  ! Found
end do
if (ia > n_specific_cases) then   ! Failed to match anything
   tmp_str = 'Specific case ('//case_str(1:ic)//' was not found.'
   call exception(procedure_name, FATAL, tmp_str)
end if

! Determine RT-active specie index ('ip') of specific case:
if (spec_case_iptog(ia) < 0) then      ! "Background"
   ip = naerspc_rad
else                     ! Must search
   do ip=1,naerspc_rad
      if (spc_ctype(ip) == spec_case_iptog(ia)) exit
   end do
   if (ip > naerspc_rad) then
      tmp_str = 'RT-active specie index for specific case ('//  &
           case_str(1:ic)//' was not found.'
      call exception(procedure_name, FATAL, tmp_str)
   end if
end if

! Use SW optical properties?  Which relative spectra interval ('iw')?
if (spec_case_iwa(ia) > nlw_wavlnrng) then
   use_sw_optprop = .TRUE.; iw = spec_case_iwa(ia)-nlw_wavlnrng
else
   use_sw_optprop = .FALSE.; iw = spec_case_iwa(ia)
end if

! Determine total (up+down) flux vs. wavelength slope
!  interpolation parameters:
tgt_fvws_r = spec_case_fvws_r(ia); tgt_fvws_b = spec_case_fvws_b(ia)
if (use_sw_optprop) then    ! Shortwave
      !## fvws_r:
   if (tgt_fvws_r >= sw_abp_optprop(ip)%max_fvws_r) then
      ifr = sw_abp_optprop(ip)%n_fvws_r-1      ! Clip to max
   else if (tgt_fvws_r <= sw_abp_optprop(ip)%min_fvws_r) then
      ifr = 1     ! Clip to min
   else     ! Search for appropriate index
      ifr = locidx_1D_mono_type02(sw_abp_optprop(ip)%fvws_r,  &
           tgt_fvws_r, sw_abp_optprop(ip)%search_beg_idx_fvws_r)
   end if
   fint_r = (tgt_fvws_r-sw_abp_optprop(ip)%fvws_r(ifr))/  &
        sw_abp_optprop(ip)%d_fvws_r(ifr)
      !## fvws_b:
   if (tgt_fvws_b >= sw_abp_optprop(ip)%max_fvws_b) then
      ifb = sw_abp_optprop(ip)%n_fvws_b-1      ! Clip to max
   else if (tgt_fvws_b <= sw_abp_optprop(ip)%min_fvws_b) then
      ifb = 1     ! Clip to min
   else     ! Search for appropriate index
      ifb = locidx_1D_mono_type02(sw_abp_optprop(ip)%fvws_b,  &
           tgt_fvws_b, sw_abp_optprop(ip)%search_beg_idx_fvws_b)
   end if
   fint_b = (tgt_fvws_b-sw_abp_optprop(ip)%fvws_b(ifb))/  &
        sw_abp_optprop(ip)%d_fvws_b(ifb)
else    ! Longwave
         !## fvws_r:
   if (tgt_fvws_r >= lw_abp_optprop(ip)%max_fvws_r) then
      ifr = lw_abp_optprop(ip)%n_fvws_r-1      ! Clip to max
   else if (tgt_fvws_r <= lw_abp_optprop(ip)%min_fvws_r) then
      ifr = 1     ! Clip to min
   else     ! Search for appropriate index
      ifr = locidx_1D_mono_type02(lw_abp_optprop(ip)%fvws_r,  &
           tgt_fvws_r, lw_abp_optprop(ip)%search_beg_idx_fvws_r)
   end if
   fint_r = (tgt_fvws_r-lw_abp_optprop(ip)%fvws_r(ifr))/  &
        lw_abp_optprop(ip)%d_fvws_r(ifr)
      !## fvws_b:
   if (tgt_fvws_b >= lw_abp_optprop(ip)%max_fvws_b) then
      ifb = lw_abp_optprop(ip)%n_fvws_b-1      ! Clip to max
   else if (tgt_fvws_b <= lw_abp_optprop(ip)%min_fvws_b) then
      ifb = 1     ! Clip to min
   else     ! Search for appropriate index
      ifb = locidx_1D_mono_type02(lw_abp_optprop(ip)%fvws_b,  &
           tgt_fvws_b, lw_abp_optprop(ip)%search_beg_idx_fvws_b)
   end if
   fint_b = (tgt_fvws_b-lw_abp_optprop(ip)%fvws_b(ifb))/  &
        lw_abp_optprop(ip)%d_fvws_b(ifb)
end if

! Determine k-loop increment:
if (kz < ka) then 
   kinc = -1
else
   kinc = 1
end if 

tau_sum = ZERO_f4   ! Initialize sum
do k=ka,kz,kinc
   opp_sum = ZERO_f4    ! Entire array at once

   if (use_sw_optprop) then    ! Shortwave
      do ib=1,spc_nb
         c1 = N_fact*spc_nconc(k,ib)/cm3_to_m3_f4*spc_xsarea_fact(ib,ip)

         vtmp1 = (ONE_f4-fint_r)*  &
              sw_abp_optprop(ip)%Q_ext(ifr,ifb,ib,iw)+  &
              fint_r*sw_abp_optprop(ip)%Q_ext(ifr+1,ifb,ib,iw)
         vtmp2 = (ONE_f4-fint_r)*  &
              sw_abp_optprop(ip)%Q_ext(ifr,ifb+1,ib,iw)+  &
              fint_r*sw_abp_optprop(ip)%Q_ext(ifr+1,ifb+1,ib,iw)
         vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
         opp_sum(1) = opp_sum(1)+vtmp*c1 ! Cross-section weighted Qext
      end do
   else    ! Longwave
      do ib=1,spc_nb
         c1 = N_fact*spc_nconc(k,ib)/cm3_to_m3_f4*spc_xsarea_fact(ib,ip)

         vtmp1 = (ONE_f4-fint_r)*  &
              lw_abp_optprop(ip)%Q_ext(ifr,ifb,ib,iw)+  &
              fint_r*lw_abp_optprop(ip)%Q_ext(ifr+1,ifb,ib,iw)
         vtmp2 = (ONE_f4-fint_r)*  &
              lw_abp_optprop(ip)%Q_ext(ifr,ifb+1,ib,iw)+  &
              fint_r*lw_abp_optprop(ip)%Q_ext(ifr+1,ifb+1,ib,iw)
         vtmp = (ONE_f4-fint_b)*vtmp1+fint_b*vtmp2
         opp_sum(1) = opp_sum(1)+vtmp*c1 ! Cross-section weighted Qext
      end do
   end if

   tau_sum = tau_sum+opp_sum(1)*dzt(k)    ! Cumulative aerosol optical depth

end do

end subroutine calc_abp_opd_specific


!```````````````````````````````````````````````````````````````````
subroutine bdst_calc(ext_vtau_pt,ext_crnu_pt,ext_dust_nb)

!+ Computes "background" dust profile number concentrations

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use aertxrad_mem_mod, only: nz1_rad,dzc,p_lev,p,t,bdst_dz_scr,bdst_N_scr  &
     ,bdst_ndist_bfrac,bdst_Nconc
use body_specific_constants_mod, only: R_air, g_Xoid
use main_memory_bspec_mod, only: dust_ctau_refp

implicit none

!=== Argument declarations:

real(f4), intent(IN) :: ext_vtau_pt,ext_crnu_pt
integer, intent(IN) :: ext_dust_nb

!== Local declarations:

integer :: k_norm,k,ib
real(f4) :: conrath_nu,q_q0,air_density,N_raw,p_bot,p_top,ftmp,tau_tgt  &
     ,column_tau,N_fact,lyr_press

!=== Executable statements:

! Construct vertical profiles of dust:

k_norm = 0
      
conrath_nu = ext_crnu_pt

do k=1,nz1_rad
   lyr_press = p(k)

   if (lyr_press < dust_ctau_refp) then
      q_q0 = max(1.e-30_f4,exp(conrath_nu*(ONE_f4-dust_ctau_refp/lyr_press)))
   else
      q_q0 = ONE_f4
   end if

   air_density = p(k)/(R_air*t(k))
   N_raw = q_q0*air_density   ! dust N proportional to mass
            
   do ib=1,ext_dust_nb            ! Initialize all dust bins
      bdst_N_scr(k,ib) = bdst_ndist_bfrac(ib)*N_raw  
   enddo

   bdst_dz_scr(k) = dzc(k)
enddo

! Attempt to determine layer that includes 'dust_ctau_refp' value:

k_norm = 0
if (dust_ctau_refp < p_lev(nz1_rad)) then   ! Only do if value is in column
   do k=nz1_rad,2,-1
      p_bot = p_lev(k)
      p_top = p_lev(k-1)
      if(p_bot >= dust_ctau_refp .and. p_top < dust_ctau_refp) then
         k_norm = k
         ftmp = R_air*t(k)/g_Xoid*log(dust_ctau_refp/p_top)
         bdst_dz_scr(k) = min(dzc(k),ftmp)
         exit
      endif
   enddo
endif

if(k_norm == 0) then   ! Need to invent some atmosphere
            
   p_bot = dust_ctau_refp
   ftmp = (dust_ctau_refp-p_lev(nz1_rad))/10._f4
   k_norm = nz1_rad+10

   do k=nz1_rad+10,nz1_rad+1,-1
      p_top = p_bot-ftmp
      lyr_press = 0.5_f4*(p_bot+p_top)

      if (lyr_press < dust_ctau_refp) then
         q_q0 = exp(conrath_nu*(ONE_f4-dust_ctau_refp/lyr_press))
      else
         q_q0 = ONE_f4
      end if

      bdst_dz_scr(k) = R_air*t(nz1_rad)/g_Xoid*log(p_bot/p_top)

      air_density = lyr_press/(R_air*t(nz1_rad))
      N_raw = q_q0*air_density   ! dust N proportional to mass
   
      do ib=1,ext_dust_nb
         bdst_N_scr(k,ib) = bdst_ndist_bfrac(ib)*N_raw 
      enddo

      p_bot = p_top
   enddo
           
endif

tau_tgt = ext_vtau_pt   ! Target opacity in column above 'k_norm'

! Initial guess at value of 'N_fact' resulting in 'tau_tgt' opacity (at
!  "visible" wavelengths):
N_fact = ONE_f4
call calc_abp_opd_specific(1, nz1_rad+10, ext_dust_nb, 'bdst_vis', k_norm, 1,  &
     bdst_N_scr, bdst_dz_scr, N_fact, column_tau)

N_fact = tau_tgt/column_tau   ! Final normalization value

! Scale dust number concentrations:

do ib=1,ext_dust_nb
   bdst_Nconc(1:nz1_rad,ib) = bdst_N_scr(1:nz1_rad,ib)*N_fact   ! [#/cm^3]
enddo

end subroutine bdst_calc


!```````````````````````````````````````````````````````````````````
subroutine bdst_calc_htop(ext_vtau_pt,ext_htop_pt,ext_dust_nb)

!+ Computes "background" dust profile number concentrations ("vtau"/haze_top)
!   prescription.

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use aertxrad_mem_mod, only: nz1_rad,dzc,p_lev,p,t,bdst_dz_scr,bdst_N_scr  &
     ,bdst_ndist_bfrac,bdst_Nconc
use body_specific_constants_mod, only: R_air, g_Xoid
use main_memory_bspec_mod, only: dust_ctau_refp

implicit none

!=== Argument declarations:

real(f4), intent(IN) :: ext_vtau_pt,ext_htop_pt
integer, intent(IN) :: ext_dust_nb

!== Local declarations:

integer :: k_norm,k,ib
real(f4) :: haze_top,q_q0,air_density,N_raw,p_bot,p_top,ftmp,tau_tgt  &
     ,column_tau,N_fact,lyr_press

!=== Executable statements:

! Construct vertical profiles of dust:

k_norm = 0
      
haze_top = ext_htop_pt

do k=1,nz1_rad
   lyr_press = p(k)

   if (lyr_press < dust_ctau_refp) then
      q_q0 = max(1.e-30_f4,exp(0.007_f4*(ONE_f4-(dust_ctau_refp/lyr_press)**  &
           (70._f4/haze_top))))
   else
      q_q0 = ONE_f4
   end if

   air_density = p(k)/(R_air*t(k))
   N_raw = q_q0*air_density   ! dust N proportional to mass
            
   do ib=1,ext_dust_nb            ! Initialize all dust bins
      bdst_N_scr(k,ib) = bdst_ndist_bfrac(ib)*N_raw  
   enddo

   bdst_dz_scr(k) = dzc(k)
enddo

! Attempt to determine layer that includes 'dust_ctau_refp' value:

k_norm = 0
if (dust_ctau_refp < p_lev(nz1_rad)) then   ! Only do if value is in column
   do k=nz1_rad,2,-1
      p_bot = p_lev(k)
      p_top = p_lev(k-1)
      if(p_bot >= dust_ctau_refp .and. p_top < dust_ctau_refp) then
         k_norm = k
         ftmp = R_air*t(k)/g_Xoid*log(dust_ctau_refp/p_top)
         bdst_dz_scr(k) = min(dzc(k),ftmp)
         exit
      endif
   enddo
endif

if(k_norm == 0) then   ! Need to invent some atmosphere
            
   p_bot = dust_ctau_refp
   ftmp = (dust_ctau_refp-p_lev(nz1_rad))/10._f4
   k_norm = nz1_rad+10

   do k=nz1_rad+10,nz1_rad+1,-1
      p_top = p_bot-ftmp
      lyr_press = 0.5_f4*(p_bot+p_top)

      if (lyr_press < dust_ctau_refp) then
         q_q0 = exp(0.007_f4*(ONE_f4-(dust_ctau_refp/lyr_press)**  &
              (70._f4/haze_top)))
      else
         q_q0 = ONE_f4
      end if

      bdst_dz_scr(k) = R_air*t(nz1_rad)/g_Xoid*log(p_bot/p_top)

      air_density = lyr_press/(R_air*t(nz1_rad))
      N_raw = q_q0*air_density   ! dust N proportional to mass
   
      do ib=1,ext_dust_nb
         bdst_N_scr(k,ib) = bdst_ndist_bfrac(ib)*N_raw 
      enddo

      p_bot = p_top
   enddo
           
endif

tau_tgt = ext_vtau_pt   ! Target opacity in column above 'k_norm'

! Initial guess at value of 'N_fact' resulting in 'tau_tgt' opacity (at "9um"):
N_fact = ONE_f4
call calc_abp_opd_specific(1, nz1_rad+10, ext_dust_nb, 'bdst_9um', k_norm, 1,  &
     bdst_N_scr, bdst_dz_scr, N_fact, column_tau)

N_fact = tau_tgt/column_tau   ! Final normalization value

! Scale dust number concentrations:

do ib=1,ext_dust_nb
   bdst_Nconc(1:nz1_rad,ib) = bdst_N_scr(1:nz1_rad,ib)*N_fact   ! [#/cm^3]
enddo

end subroutine bdst_calc_htop


end module aerad_aeropd_mod
