!```````````````````````````````````````````````````````````````````
module aerad_driv_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine aerad_driver(ext_n1, ext_n2, ext_n3, ext_n5, ext_ia, ext_iz,  &
     ext_ja, ext_jz, ext_ka, ext_kz, ext_naersclr, ext_maxnbins, ext_ngroups,  &
     ext_aer_ngrp, ext_aer_grpbins, ext_press, ext_tempk, ext_rv, ext_sfcT,  &
     ext_tautot, ext_cosZ, ext_msdist, ext_albedo, ext_emissivity,  &
     ext_use_gas_opacities, DUST_tau, DUST_tau_pr, DIM_dust_levs, ext_rlong,  &
     ext_fthrd, ext_dzc_static, ext_rtgt, ext_solar_azm_ang, ext_tazm_ang,  &
     ext_tslope_ang, ext_tslas_tog, ext_tshadow_tog, ext_nazm_tshadow,  &
     ext_cosz_horizon, ext_TCx_obstruct, ext_TCz_obstruct, ext_zc_static,  &
     init_rad, ext_haveH2Ov, ext_vtau, ext_crnu, ext_vtau_tog,  &
     ext_flxout_imask, ext_bottom_other_sflux_down, ext_bottom_SWflux_down,  &
     ext_topo_factor_sflux, ext_fluxo_np,  &
     ext_up_fluxes, ext_dn_fluxes, ext_dT_dt, ext_solar_sw,  &
     ext_aerradius, ext_aer_group, ext_aer_subgrp, ext_aer_ig, ext_XOrad,  &
     ext_XOrad_mask, ext_aersclp)

!+ Modular radiative transfer code driver

use univ_kind_defs_mod, only: f8, f4
use univ_math_constants_mod, only: twopi_f4, ZERO_f4, ONE_f4
use univ_run_timing_mod, only: run_timing_obj, record_wall_time, START, END,  &
     record_CPU_time, calc_elapsed_time, display_elapsed_time,  &
     initialize_elapsed_time
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use body_specific_constants_mod, only: R_air, solar_mean_broadband
use aertxrad_mem_mod, only: nz_rad, nz1_rad, max_aerbins, naerspc_rad,  &
     i_curr, j_curr, two_nz1_rad, p, t, qH2O, t_lev, p_lev, qH2O_lev,  &
     sfc_tempk, sflux_frac, cos_mu, sw_iwbeg, sw_iwend, sfc_albedo,  &
     sfc_emiss, scr_dim, sw_sfc_dnflux, lw_sfc_dnflux, net_dTdt, rho_air,  &
     dzc, aertxrad_AD, sw_on, ip_ibeg, ip_iend, ntot_gpt, lw_ipbeg,  &
     lw_ipend, iw_iend, iw_ibeg, lw_iwbeg, lw_iwend, ntot_wavlnrng,  &
     spc_radtx, spc_begiv, part_in_tshadow, swcut, ntot_wavlnrng,  &
     want_rtflux_output, up_fluxes, dn_fluxes, tau_gas, refine_lw_fluxes,  &
     solar_sw
use aermic_rad_mem_mod, only: faux_dust_rtxinfo_needed
use aerad_aeropd_mod, only: bdst_calc, bdst_calc_htop, calc_abp_opacities
use aerad_txrad_mod, only: two_stream, add_txrad, refine_lwflux
use aerad_precalc_mod, only: rad_precalc
use aerad_gasopd_mod, only: calc_gasopd
use aerad_postcalc_mod, only: rad_postcalc

implicit none

!=== Argument declarations:

integer, intent(IN) :: ext_n1,ext_n2,ext_n3,ext_n5,ext_ia,ext_iz,ext_ja,ext_jz  &
     ,ext_ka,ext_kz,DIM_dust_levs,ext_naersclr,ext_maxnbins,ext_ngroups  &
     ,ext_aer_ngrp,ext_tslas_tog,ext_tshadow_tog,ext_nazm_tshadow,ext_vtau_tog
real(f4), intent(IN) :: ext_tautot,ext_msdist
real(f4), dimension(DIM_dust_levs), intent(IN) :: DUST_tau,DUST_tau_pr
real(f4), dimension(ext_n2,ext_n3,ext_n5), intent(IN) :: ext_sfcT
real(f4), dimension(ext_n2,ext_n3), intent(IN) :: ext_albedo,ext_cosZ,ext_rtgt  &
     ,ext_solar_azm_ang,ext_tazm_ang,ext_tslope_ang,ext_vtau,ext_crnu,  &
     ext_emissivity
real(f4), dimension(ext_n2,ext_n3,ext_nazm_tshadow), intent(IN) ::  &
     ext_cosz_horizon,ext_TCx_obstruct,ext_TCz_obstruct
real(f4), dimension(ext_n1), intent(IN) :: ext_dzc_static,ext_zc_static
real(f4), dimension(ext_n1,ext_n2,ext_n3), intent(IN) :: ext_press,ext_tempk
integer, dimension(ext_aer_ngrp), intent(IN) :: ext_aer_grpbins
logical, intent(IN) :: init_rad, ext_haveH2Ov, ext_use_gas_opacities
! ***NOTE: the following have CGS units:
real(f4), dimension(ext_n1,ext_n2,ext_n3), intent(IN) :: ext_rv
integer, dimension(ext_n2,ext_n3), intent(IN) :: ext_flxout_imask
integer, intent(IN) :: ext_fluxo_np

real(f4), dimension(ext_n2,ext_n3), intent(INOUT) :: ext_rlong,  &
     ext_bottom_other_sflux_down, ext_bottom_SWflux_down, ext_topo_factor_sflux
real(f4), dimension(ext_n1,ext_n2,ext_n3), intent(INOUT) :: ext_fthrd

real(f4), dimension(ext_n1,ntot_wavlnrng,ext_fluxo_np), optional,  &
     intent(INOUT) :: ext_up_fluxes, ext_dn_fluxes
real(f4), dimension(ext_n1,ext_fluxo_np), optional, intent(INOUT) :: ext_dT_dt
real(f4), optional, intent(OUT) :: ext_solar_sw

logical, dimension(ext_naersclr), optional, intent(IN) :: ext_XOrad,  &
     ext_XOrad_mask
integer, dimension(ext_naersclr), optional, intent(IN) :: ext_aer_group,  &
     ext_aer_subgrp, ext_aer_ig
real(f4), dimension(ext_n1,ext_n2,ext_n3,ext_naersclr), optional, intent(IN) ::&
     ext_aersclp
real(f4), dimension(ext_maxnbins,ext_ngroups), optional, intent(IN) ::  &
     ext_aerradius

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'aerad_driver'

   ! Time selected RT driver elements (and display the results)?
logical, parameter :: timing_is_desired = .FALSE.

real(f4) :: tmp,sinz,cosai,atmp,aint,cosz_h,x_obst,z_obst,sw_cutoff
integer :: iv,ig,ik,k,i,j,iw,ip,ia0,ia1, ix_fo, iy_fo
logical :: horizon_extension,hextn_active
!%logical :: found

type(run_timing_obj), dimension(8) :: CPU_timing

!=== Executable statements:

if (timing_is_desired) then
   CPU_timing(1)%description = 'radinit'
   CPU_timing(2)%description = 'predrv'
   CPU_timing(3)%description = 'gasopd'
   CPU_timing(4)%description = 'abp_opacities'
   CPU_timing(5)%description = 'precalc'
   CPU_timing(6)%description = 'radtx'
   CPU_timing(7)%description = 'postcalc'
   CPU_timing(8)%description = 'pstdrv'
   call initialize_elapsed_time(CPU_timing)  ! Entire array at once
end if

if (timing_is_desired) call record_CPU_time(CPU_timing(1), START)
!#### timing

if (init_rad) then     ! Initialize radiative transfer model (do only once)

   ! Check if this routine has been called correctly:
   if (ext_naersclr > 0) then
      if (.not.(present(ext_aer_group) .and. present(ext_aer_subgrp) .and.  &
           present(ext_aer_ig) .and. present(ext_XOrad) .and.  &
           present(ext_XOrad_mask) .and. present(ext_aersclp) .and.  &
           present(ext_aerradius))) call exception(  &
           procedure_name, FATAL,  &
           'In this use case, all optional arguments must be provided.')
   end if

!%%%%%% These are now called and/or done by PRAMS during model initialization:
!%
!%   nz1_rad = ext_kz-ext_ka+1  ! Set vertical dims for radiation code use (NOTE:
!%   nz_rad = nz1_rad-1         !  may not be used with vertical grid nesting)
!%   two_nz1_rad = 2*nz1_rad    !
!%
!%   scr_dim = max(nz1_rad*nz1_rad,nz1_rad*4)   ! Dim for scratch arrays
!%   max_aerbins = ext_maxnbins  ! Max # of aerosol bins used
!%
!%   ! Find # of possibly radiatively active aerosol species:
!%
!%   naerspc_rad = 0
!%   if (ext_naersclr > 0) then
!%      ig = ext_aer_ig(1)
!%      do iv=1,ext_naersclr  ! Loop over all aerosol scalars   
!%         if(ext_XOrad_mask(iv)) then  ! May have to calc RT using this scalar
!%            if(ig /= ext_aer_ig(iv)) then   ! Changing groups
!%               ig = ext_aer_ig(iv)   ! relative group ID #
!%               found = .FALSE.
!%            endif
!%            if(ext_aer_subgrp(iv) == 0 .and. .NOT.found) then ! Uncounted # conc
!%               naerspc_rad = naerspc_rad+1
!%               found = .TRUE.
!%            endif
!%         endif 
!%      enddo
!%   end if
!%
!%   call aertxrad_AD(.TRUE.)    ! Allocate needed memory
!%
!%   if (ext_naersclr > 0) then
!%      ! Fill some general reference arrays:
!%      call init_ref(ext_naersclr, ext_aer_ngrp, ext_aer_grpbins,  &
!%           ext_aer_group, ext_aer_subgrp, ext_aer_ig, ext_XOrad_mask)
!%      ! Fill optical property reference arrays:
!%      call init_optprop(ext_maxnbins,ext_ngroups,ext_aerradius)
!%   else
!%      ! Fill some general reference arrays:
!%      call init_ref(ext_naersclr, ext_aer_ngrp, ext_aer_grpbins)
!%   end if
!%

end if

! Check whether optional arguments are required:
if (maxval(ext_flxout_imask) > 0) then
   if (.not. (present(ext_up_fluxes) .and. present(ext_dn_fluxes) .and.  &
        present(ext_dT_dt) .and. present(ext_solar_sw)))  &
        call exception(procedure_name, FATAL,  &
        'Optional arguments not provided, but are required.')
end if

!#### timing
if (timing_is_desired) then
   call record_CPU_time(CPU_timing(1), END)
   call calc_elapsed_time(CPU_timing(1), accumulate_elapsed_time=.TRUE.)
end if

! Find and flag currently radiatively active aerosol species:

ig = naerspc_rad
if(faux_dust_rtxinfo_needed) then
   ig = ig-1
   spc_radtx(naerspc_rad) = .TRUE.
endif

do ip=1,ig     ! Loop over all *possible* real rad_active species
   iv = spc_begiv(ip)
   spc_radtx(ip) = ext_XOrad(iv)   ! .TRUE. if rad_active now, else .FALSE.
enddo

! Fraction of the interplanetary solar flux (at 1 AU) at top of
!  atmosphere:

sflux_frac = real(ONE_f4/ext_msdist, kind=f8)    ! [1/AU^2]

ix_fo = 0; iy_fo = 0

do j=ext_ja,ext_jz        !
   j_curr = j             ! Loop over all *computational* horizontal grid
   do i=ext_ia,ext_iz     !  points of external model
      i_curr = i          !

      if (timing_is_desired) call record_CPU_time(CPU_timing(2), START)
      !#### timing

      ! Set temperature, pressure, and water vapor mixing ratio of layer
      !  MIDPOINTS [with ik=nz1_rad being the lowest layer (in z)]:

      ik = nz1_rad
      if(ext_haveH2Ov) then   ! Water vapor present in external model
         do k=ext_ka,ext_kz     
            dzc(ik) = ext_rtgt(i,j)/ext_dzc_static(k) ! dz (real-world) of lyr
            t(ik) = ext_tempk(k,i,j)     ! [K]
            p(ik) = ext_press(k,i,j)      ! [Pa]
            rho_air(ik) = p(ik)/(t(ik)*R_air)   ! air density [kg/m^3]
            tmp = ext_rv(k,i,j)   ! H2O vapor mass conc [g/cm^3] 
            qH2O(ik) = tmp/(rho_air(ik)*1.e-3_f4)  ! H2Ov mix ratio [g/g = kg/kg]
            ik = ik-1
         enddo
      else   ! No water vapor, and ext_rv() likely holds undefined values
         do k=ext_ka,ext_kz     
            dzc(ik) = ext_rtgt(i,j)/ext_dzc_static(k) ! dz (real-world) of lyr
            t(ik) = ext_tempk(k,i,j)     ! [K]
            p(ik) = ext_press(k,i,j)      ! [Pa]
            rho_air(ik) = p(ik)/(t(ik)*R_air)   ! air density [kg/m^3]
            qH2O(ik) = ZERO_f4  ! H2Ov mix ratio [g/g = kg/kg]
            ik = ik-1
         enddo
      endif

      ! Set ground (surface) temperature:
      
      sfc_tempk = ext_sfcT(i,j,1)   ! [K]

      ! Set temperature, pressure, and water vapor mixing ratio of layer
      !  BOUNDARIES:

      do k=2,nz_rad
         t_lev(k) = (t(k)+t(k+1))*0.5_f4
         p_lev(k) = exp(0.5_f4*log(p(k+1)/p(k))+log(p(k)))  ! log(p) interp
         qH2O_lev(k) = (qH2O(k)+qH2O(k+1))*0.5_f4
      enddo

      k = 1   ! Uppermost layer BOUNDARY (top bound of layer k=2)
      t_lev(k) = (t(k)+t(k+1))*0.5_f4
      p_lev(k) = exp(0.5_f4*log(p(k+1)/p(k))+log(p(k)))   ! log(p) interp
      qH2O_lev(k) = (qH2O(k)+qH2O(k+1))*0.5_f4

      k = nz1_rad   ! Bottom layer BOUNDARY (lower bound of layer k=nz1_rad)
      t_lev(k) = sfc_tempk
!#      p_lev(k) = exp(log(p(k)/p_lev(k-1))+log(p(k)))
      p_lev(k) = exp(0.5_f4*log(ext_press(ext_ka-1,i,j)/p(k))+  &
           log(p(k)))                             ! log(p) interp
      qH2O_lev(k) = qH2O(k)

      ! Calculate "background" dust profile ('bdst_Nconc') for this column:

      if (ext_vtau_tog == 3 .or. ext_vtau_tog == 4) then
         ! Use "vtau" "background" dust loading:
         call bdst_calc_htop(ext_vtau(i,j),ext_crnu(i,j),ext_aer_grpbins(1))
      else if (ext_vtau_tog == 1 .or. ext_vtau_tog == 2) then
         ! Use "background" dust loading:
         call bdst_calc(ext_vtau(i,j),ext_crnu(i,j),ext_aer_grpbins(1))
      end if

      ! Set "longwave" surface properties:
      do iw=lw_iwbeg,lw_iwend  ! Loop over relevant wavelength intervals
         sfc_emiss(iw) = ext_emissivity(i,j)    ! [none]
         sfc_albedo(iw) = ONE_f4-sfc_emiss(iw)     ! [none]  
      end do

      !*** Set cosine of incident solar angle (zenith angle); take into account
      !     (crudely) any horizon extension:

      cos_mu = real(ext_cosZ(i,j), kind=f8)     ! [none]

      hextn_active = .FALSE.
      if(ext_tshadow_tog == 1) then   ! Need to check for horizon extension
      
         ! Calculate cosz_horizon array location/interpolative indices (using
         !  solar azimuth values):
         atmp = ext_solar_azm_ang(i,j)/(twopi_f4/real(ext_nazm_tshadow))
         ia0 = int(atmp)
         aint = atmp-real(ia0)   ! Location crucial to avoid ia0 complications
         
         if(ia0 == 0) then
            ia0 = ext_nazm_tshadow
            ia1 = 1
         elseif(ia0 == ext_nazm_tshadow) then
            ia1 = 1
         else
            ia1 = ia0+1
         endif

         ! Calc horizon cosz, etc. for current solar azimuth via interpolation:
         z_obst = (ONE_f4-aint)*ext_TCz_obstruct(i,j,ia0)+  &
              aint*ext_TCz_obstruct(i,j,ia1)
         if(z_obst < epsilon(z_obst)) then   ! Horizon extension
            horizon_extension = .TRUE.
            cosz_h = (ONE_f4-aint)*ext_cosz_horizon(i,j,ia0)+  &
                 aint*ext_cosz_horizon(i,j,ia1)
            if(real(cos_mu) > cosz_h .and. cos_mu < 4.e-2_f8) then
               ! Sun is above extended horizon, but still low in sky;
               !  cosZ=0.04 =~ 87.71deg
               cos_mu = min(cos_mu-real(cosz_h, kind=f8),4.e-2_f8)
               hextn_active = .TRUE.
            endif
         else
            horizon_extension = .FALSE.
         endif
      endif

      ! Only calculate "shortwave" stuff if sun is sufficiently above horizon;
      !  set probability point interval indices for use in radiative transfer
      !  calcs:

      if(cos_mu > 1.e-6_f8) then
         sw_on = .TRUE.
         ip_ibeg = 1          ! Loop over ALL probability points
         ip_iend = ntot_gpt   !  
         iw_ibeg = 1               ! Loop over ALL wavelength bands
         iw_iend = ntot_wavlnrng   !

         ! Set "shortwave" wavelength group sfc optical properties:      
         do iw=sw_iwbeg,sw_iwend    ! Loop over relevant wavelength intervals
            sfc_albedo(iw) = ext_albedo(i,j)       ! [none]
            sfc_emiss(iw) = ONE_f4-sfc_albedo(iw)         ! [none]   
         enddo

         ! Topographic slope/aspect correction:

         if (ext_tslas_tog > 1) then
            sinz = real(sin(acos(cos_mu)), kind=f4)

            ! If cosai < 0., point not sunlit
            cosai = (real(cos_mu, kind=f4)*cos(ext_tslope_ang(i,j)))+  &
                 (sinz*sin(ext_tslope_ang(i,j))*  &
                 cos(ext_solar_azm_ang(i,j)-ext_tazm_ang(i,j)))
         end if

         ! Topographic shadowing:

         part_in_tshadow  = .FALSE.   ! Initialize
         if(ext_tshadow_tog == 1 .and.  &     ! Possible shadowing
              .NOT.horizon_extension) then
            cosz_h = (ONE_f4-aint)*ext_cosz_horizon(i,j,ia0)+  &
                 aint*ext_cosz_horizon(i,j,ia1)
            x_obst = (ONE_f4-aint)*ext_TCx_obstruct(i,j,ia0)+  &
                 aint*ext_TCx_obstruct(i,j,ia1)
!Calc'd          z_obst = (1.-aint)*ext_TCz_obstruct(i,j,ia0)+  &
!above                 aint*ext_TCz_obstruct(i,j,ia1)

            sw_cutoff = z_obst-(x_obst*tan(asin(real(cos_mu)))) 

            if(sw_cutoff > ZERO_f4 .and.  &  ! Part of column shadowed
                 sw_cutoff < ext_zc_static(ext_kz)*ext_rtgt(i,j)) then 

               part_in_tshadow = .TRUE.

               swcut = 0
               atmp = ZERO_f4
               shadow_top: do ik=ext_ka,ext_kz        ! Find top of shadow
                  aint = ext_zc_static(ik)*ext_rtgt(i,j)
                  if(sw_cutoff < aint) then
                     if(ik /= ext_ka) then
                        if(abs(aint-sw_cutoff) < abs(atmp-sw_cutoff)) then
                           swcut = ext_kz-ik+1
                        else
                           swcut = ext_kz-ik+2
                        endif
                     else
                        swcut = ext_kz-ik+1
                     endif
                     exit shadow_top
                  endif
                  atmp = aint
               enddo shadow_top   ! Column sunlit above layer 'swcut'
            
               if(swcut == 0) then
                  write(*,*) 'ERROR: unable to find shadow top.'
                  stop '*** ERROR: aerad_driver01 ***'
               endif
               
            endif
         endif

      else
         sw_on = .FALSE.
         ip_ibeg = lw_ipbeg    ! Only loop over "longwave" pts if
         ip_iend = lw_ipend    !  "shortwave" is off
         iw_ibeg = lw_iwbeg    !
         iw_iend = lw_iwend    !
      endif

      !#### timing
      if (timing_is_desired) then
         call record_CPU_time(CPU_timing(2), END)
         call calc_elapsed_time(CPU_timing(2), accumulate_elapsed_time=.TRUE.)
      end if

      ! Compute "air" (mixed CO2 *and* H2O_vapor) opacities at layer MIDPOINTS
      !  [results placed in tau_air(ig,iw,k)]:

      if (timing_is_desired) call record_CPU_time(CPU_timing(3), START)
      !#### timing
      if (ext_use_gas_opacities) then
         call calc_gasopd
      else
         tau_gas = ZERO_f4
      end if
      !#### timing
      if (timing_is_desired) then
         call record_CPU_time(CPU_timing(3), END)
         call calc_elapsed_time(CPU_timing(3), accumulate_elapsed_time=.TRUE.)
      end if

      if (timing_is_desired) call record_CPU_time(CPU_timing(4), START)
      !#### timing
      if (ext_naersclr > 0) call calc_abp_opacities(ext_n1, ext_n2, ext_n3,  &
           ext_naersclr, ext_aersclp, ext_kz)
      !#### timing
      if (timing_is_desired) then
         call record_CPU_time(CPU_timing(4), END)
         call calc_elapsed_time(CPU_timing(4), accumulate_elapsed_time=.TRUE.)
      end if

      !================================================================

      if (timing_is_desired) call record_CPU_time(CPU_timing(5), START)
      !#### timing
      call rad_precalc      ! Do pre-radtx calcs
      !#### timing
      if (timing_is_desired) then
         call record_CPU_time(CPU_timing(5), END)
         call calc_elapsed_time(CPU_timing(5), accumulate_elapsed_time=.TRUE.)
      end if

      if (timing_is_desired) call record_CPU_time(CPU_timing(6), START)
      !#### timing
      call two_stream      !
      call add_txrad       ! Do radiative transfer calcs
      if (refine_lw_fluxes) call refine_lwflux   !
      !#### timing
      if (timing_is_desired) then
         call record_CPU_time(CPU_timing(6), END)
         call calc_elapsed_time(CPU_timing(6), accumulate_elapsed_time=.TRUE.)
      end if

      if (ext_flxout_imask(i,j) > 0) then
         want_rtflux_output = .TRUE.      ! Calc/save 2-stream fluxes
      else
         want_rtflux_output = .FALSE.
      end if

      if (timing_is_desired) call record_CPU_time(CPU_timing(7), START)
      !#### timing
      call rad_postcalc   ! Calculate final fluxes / heating rates
      !#### timing
      if (timing_is_desired) then
         call record_CPU_time(CPU_timing(7), END)
         call calc_elapsed_time(CPU_timing(7), accumulate_elapsed_time=.TRUE.)
      end if  

      !================================================================

      if (timing_is_desired) call record_CPU_time(CPU_timing(8), START)
      !#### timing

      ! *Incident* surface "shortwave" flux [W/m^2]:
      if (sw_on) then
         ! The portion of the stellar flux (at the body's surface) that *IS*
         !  treated by the "shortwave" atmospheric RT calculations:
         ext_bottom_SWflux_down(i,j) = sw_sfc_dnflux

         ! The portion of the stellar flux (at the body's surface) that *IS NOT*
         !  treated by the "shortwave" atmospheric RT calculations:
         ext_bottom_other_sflux_down(i,j) = (solar_mean_broadband-solar_sw)/  &
              ext_msdist*real(cos_mu, kind=f4)

         if (ext_tslas_tog > 1 .and. .NOT.hextn_active) then
            ! Multiply by cosai to take both topography orientation AND stellar
            !  incidence angle into account (after dividing by cos_mu to obtain
            !  the non-angle-corrected incident stellar flux):
            ext_topo_factor_sflux(i,j) = cosai/real(cos_mu, kind=f4)
         else
            ! Take ONLY cos(Z) into account:
            ext_topo_factor_sflux(i,j) = ONE_f4
         end if
      else
         ext_bottom_SWflux_down(i,j) = ZERO_f4
         ext_bottom_other_sflux_down(i,j) = ZERO_f4
         ext_topo_factor_sflux(i,j) = ZERO_f4
      end if

      ext_rlong(i,j) = lw_sfc_dnflux  ! Surface "longwave" down flux [W/m^2]

      if (want_rtflux_output) then        ! Calc/save 2-stream fluxes
         ip = ext_flxout_imask(i,j)
         do iw=1,ntot_wavlnrng
            ik = nz1_rad
            do k=ext_ka-1,ext_kz-1
               ext_up_fluxes(k,iw,ip) = real(up_fluxes(ik,iw), kind=f4)
               ext_dn_fluxes(k,iw,ip) = real(dn_fluxes(ik,iw), kind=f4)
               ik = ik-1
            end do
            ext_up_fluxes(ext_kz,iw,ip) = ext_up_fluxes(ext_kz-1,iw,ip)
            ext_dn_fluxes(ext_kz,iw,ip) = ext_dn_fluxes(ext_kz-1,iw,ip)
         end do

         ik = nz1_rad
         do k=ext_ka,ext_kz
            ext_dT_dt(k,ip) = net_dTdt(ik)  ! Net layer heating rate
            ik = ik-1
         end do

         ! The portion of the stellar flux @1 AU that *IS* treated by the
         !  "shortwave" atmospheric RT calculations:
         ext_solar_sw = solar_sw
      end if
 
      ik = nz1_rad
      do k=ext_ka,ext_kz     
         ext_fthrd(k,i,j) = net_dTdt(ik)  ! Net layer heating rate
         ik = ik-1
      enddo

      !#### timing
      if (timing_is_desired) then
         call record_CPU_time(CPU_timing(8), END)
         call calc_elapsed_time(CPU_timing(8), accumulate_elapsed_time=.TRUE.)
      end if

   enddo
enddo

if (timing_is_desired)  &
     call display_elapsed_time(CPU_timing, 'CARMA_based RT driver CPU timings:')

end subroutine aerad_driver


end module aerad_driv_mod
