!```````````````````````````````````````````````````````````````````
module aerad_postcalc_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine rad_postcalc

!+ Calculate total radiative fluxes; put in a form suitable for output

use univ_kind_defs_mod, only: f4, f8
use univ_math_constants_mod, only: ZERO_f4, ZERO_f8
use aertxrad_mem_mod, only: sw_upflux,sw_dnflux,lw_upflux,lw_dnflux,lw_ipend  &
     ,lw_ipbeg,sw_on,nz1_rad,direc,directu,ck1,ck2,el1,em1,cpb,cmb,el2  &
     ,sw_sfc_dnflux,lw_sfc_upflux,lw_sfc_dnflux,sw_ipbeg,sw_ipend,em2  &
     ,direct,i_curr,j_curr,cp_air,net_dTdt,rho_air,dzc, swcut  &
     ,part_in_tshadow, up_fluxes, dn_fluxes, want_rtflux_output,  &
     ngauss_pts, lw_iwbeg, sw_iwbeg, refine_lw_fluxes

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: k, ip, k_1, iw
real(f4) :: lyr_mass_fact, sw_dTdt, lw_dTdt
real(f8) :: sum_A, sum_B, fdiv_sw, fdiv_lw

!=== Executable statements:

!  sw_upflux = Net "shortwave" flux up [W/m^2]
!  sw_dnflux = Net "shortwave" flux down [W/m^2]
!  lw_upflux = Net "longwave" flux up [W/m^2]
!  lw_dnflux = Net "longwave" flux down [W/m^2]

if (want_rtflux_output) then  ! Calc/save 2-stream fluxes [f(wavelength)]
   up_fluxes = ZERO_f4; dn_fluxes = ZERO_f4   ! Initialize for sums below
end if

if (sw_on) then      !***** BOTH "longwave" and "shortwave" *****

   ! Finalize fluxes:

   do k=1,nz1_rad    ! Loop over all layer BOUNDARIES

      sum_A = ZERO_f8; sum_B = zero_f8
      if (refine_lw_fluxes) then
         do ip=lw_ipbeg,lw_ipend ! Loop over all "longwave" wavelength intervals
            sum_A = sum_A+DIREC(ip,k)  ! Down
            sum_B = sum_B+DIRECTU(ip,k)   ! Up
         end do
      else
         do ip=lw_ipbeg,lw_ipend ! Loop over all "longwave" wavelength intervals
            sum_A = sum_A+CK1(ip,k)*EL2(ip,k)+  &   ! Down
                 CK2(ip,k)*EM2(ip,k)+CMB(ip,k)+DIRECT(ip,k)
            sum_B = sum_B+CK1(ip,k)*EL1(ip,k)+CK2(ip,k)*EM1(ip,k)+CPB(ip,k) ! Up
         end do
      end if
      lw_dnflux(k) = sum_A; lw_upflux(k) = sum_B

      sum_A = ZERO_f8; sum_B = zero_f8
      do ip=sw_ipbeg,sw_ipend  ! Loop over all "shortwave" wavelength intervals
         sum_A = sum_A+CK1(ip,k)*EL2(ip,k)+  &   ! Down
              CK2(ip,k)*EM2(ip,k)+CMB(ip,k)+DIRECT(ip,k)
         sum_B = sum_B+CK1(ip,k)*EL1(ip,k)+CK2(ip,k)*EM1(ip,k)+CPB(ip,k)  ! Up
      end do
      sw_dnflux(k) = sum_A; sw_upflux(k) = sum_B

   end do

   if (want_rtflux_output) then  ! Calc/save 2-stream fluxes [f(wavelength)]
      do k=1,nz1_rad    ! Loop over all layer BOUNDARIES
         if (refine_lw_fluxes) then
            do ip=lw_ipbeg,lw_ipend ! Loop over "longwave" wavelength intervals
               iw = (ip-lw_ipbeg)/ngauss_pts+lw_iwbeg
               dn_fluxes(k,iw) = dn_fluxes(k,iw)+DIREC(ip,k)
               up_fluxes(k,iw) = up_fluxes(k,iw)+DIRECTU(ip,k)
            end do
         else
            do ip=lw_ipbeg,lw_ipend ! Loop over "longwave" wavelength intervals
               iw = (ip-lw_ipbeg)/ngauss_pts+lw_iwbeg
               dn_fluxes(k,iw) = dn_fluxes(k,iw)+CK1(ip,k)*EL2(ip,k)+  &  ! Down
                    CK2(ip,k)*EM2(ip,k)+CMB(ip,k)+DIRECT(ip,k)
               up_fluxes(k,iw) = up_fluxes(k,iw)+CK1(ip,k)*EL1(ip,k)+  &
                    CK2(ip,k)*EM1(ip,k)+CPB(ip,k)  ! Up
            end do
         end if
         
         do ip=sw_ipbeg,sw_ipend ! Loop over "shortwave" wavelength intervals
            iw = (ip-sw_ipbeg)/ngauss_pts+sw_iwbeg
            dn_fluxes(k,iw) = dn_fluxes(k,iw)+real(CK1(ip,k)*EL2(ip,k)+  &
                 CK2(ip,k)*EM2(ip,k)+CMB(ip,k)+DIRECT(ip,k), kind=f4)
            up_fluxes(k,iw) = up_fluxes(k,iw)+real(CK1(ip,k)*EL1(ip,k)+  &
                 CK2(ip,k)*EM1(ip,k)+CPB(ip,k), kind=f4)
         end do
      end do
   end if

   ! Set surface "down" fluxes ("shortwave" and "longwave"):
   sw_sfc_dnflux = real(sw_dnflux(nz1_rad), kind=f4)
   lw_sfc_dnflux = real(lw_dnflux(nz1_rad), kind=f4)
!%not currently used%     lw_sfc_upflux = real(lw_upflux(nz1_rad), kind=f4)

   ! Calculate atmospheric heating rates (dT/dt) [K/s]:

   net_dTdt(1) = ZERO_f4  ! Set heating rate for "fake layer"
      
   if (part_in_tshadow) then

      do k=2,swcut-1  ! Above shadow
         k_1 = k-1

         lyr_mass_fact = dzc(k)*rho_air(k)*cp_air
   
         fdiv_sw = (sw_upflux(k)-sw_dnflux(k))-(sw_upflux(k_1)-sw_dnflux(k_1))
         sw_dTdt = real(fdiv_sw, kind=f4)/lyr_mass_fact ! "shortwave" heating rate [K/s]

         fdiv_lw = (lw_upflux(k)-lw_dnflux(k))-(lw_upflux(k_1)-lw_dnflux(k_1))
         lw_dTdt = real(fdiv_lw, kind=f4)/lyr_mass_fact ! "longwave" heating rate [K/s]

         net_dTdt(k) = sw_dTdt+lw_dTdt       !  Net heating rate [K/s]
      end do
      if (swcut < nz1_rad) then
         do k=nz1_rad,swcut+1,-1  ! Within shadow
            k_1 = k-1

            lyr_mass_fact = dzc(k)*rho_air(k)*cp_air
   
            fdiv_sw = (sw_upflux(k)-sw_dnflux(k))-  &
                 (sw_upflux(k_1)-sw_dnflux(k_1))
            sw_dTdt = real(fdiv_sw, kind=f4)/lyr_mass_fact ! "shortwave" heating rate [K/s]

            fdiv_lw = (lw_upflux(k)-lw_dnflux(k))-  &
                 (lw_upflux(k_1)-lw_dnflux(k_1))
            lw_dTdt = real(fdiv_lw, kind=f4)/lyr_mass_fact ! "longwave" heating rate [K/s]

            net_dTdt(k) = sw_dTdt+lw_dTdt       !  Net heating rate [K/s]
         end do
         k = swcut     ! Top of shadow [use fdiv_sw(swcut+1)]
         k_1 = k-1
         lyr_mass_fact = dzc(k)*rho_air(k)*cp_air
         sw_dTdt = real(fdiv_sw, kind=f4)/lyr_mass_fact ! "shortwave" heating rate [K/s]
         fdiv_lw = (lw_upflux(k)-lw_dnflux(k))-  &
              (lw_upflux(k_1)-lw_dnflux(k_1))
         lw_dTdt = real(fdiv_lw, kind=f4)/lyr_mass_fact ! "longwave" heating rate [K/s]
         net_dTdt(k) = sw_dTdt+lw_dTdt       !  Net heating rate [K/s]
      else
         k = swcut     ! Top of shadow [use fdiv_sw = 0]
         k_1 = k-1
         lyr_mass_fact = dzc(k)*rho_air(k)*cp_air
         sw_dTdt = ZERO_f4   ! "shortwave" heating rate [K/s]
         fdiv_lw = (lw_upflux(k)-lw_dnflux(k))-  &
              (lw_upflux(k_1)-lw_dnflux(k_1))
         lw_dTdt = real(fdiv_lw, kind=f4)/lyr_mass_fact ! "longwave" heating rate [K/s]
         net_dTdt(k) = sw_dTdt+lw_dTdt       !  Net heating rate [K/s]
      end if

   else
   
      do k=2,nz1_rad
         k_1 = k-1

         lyr_mass_fact = dzc(k)*rho_air(k)*cp_air
   
         fdiv_sw = (sw_upflux(k)-sw_dnflux(k))-(sw_upflux(k_1)-sw_dnflux(k_1))
         sw_dTdt = real(fdiv_sw, kind=f4)/lyr_mass_fact ! SW heating rate [K/s]

         fdiv_lw = (lw_upflux(k)-lw_dnflux(k))-(lw_upflux(k_1)-lw_dnflux(k_1))
         lw_dTdt = real(fdiv_lw, kind=f4)/lyr_mass_fact ! LW heating rate [K/s]
      
         net_dTdt(k) = sw_dTdt+lw_dTdt       !  Net heating rate [K/s]
      end do

   end if

else     !***** ONLY "longwave" *****

   ! Finalize fluxes:

   do k=1,nz1_rad    ! Loop over all layer BOUNDARIES
      sum_A = ZERO_f8; sum_B = ZERO_f8
      if (refine_lw_fluxes) then
         do ip=lw_ipbeg,lw_ipend ! Loop over all "longwave" wavelength intervals
            sum_A = sum_A+DIREC(ip,k)
            sum_B = sum_B+DIRECTU(ip,k)
         end do
      else
         do ip=lw_ipbeg,lw_ipend ! Loop over all "longwave" wavelength intervals
            sum_A = sum_A+CK1(ip,k)*EL2(ip,k)+  &   ! Down
                 CK2(ip,k)*EM2(ip,k)+CMB(ip,k)+DIRECT(ip,k)
            sum_B = sum_B+CK1(ip,k)*EL1(ip,k)+CK2(ip,k)*EM1(ip,k)+CPB(ip,k) ! Up
         end do
      end if
      lw_dnflux(k) = sum_A; lw_upflux(k) = sum_B
   end do

   ! Set RT fluxes at surface ("longwave") [W/m^2]:
   lw_sfc_dnflux = real(lw_dnflux(nz1_rad), kind=f4)
!%not currently used%     lw_sfc_upflux = real(lw_upflux(nz1_rad), kind=f4)

   if (want_rtflux_output) then  ! Calc/save 2-stream fluxes [f(wavelength)]
      if (refine_lw_fluxes) then
         do k=1,nz1_rad    ! Loop over all layer BOUNDARIES
            do ip=lw_ipbeg,lw_ipend ! Loop over "longwave" wavelength intervals
               iw = (ip-lw_ipbeg)/ngauss_pts+lw_iwbeg
               dn_fluxes(k,iw) = dn_fluxes(k,iw)+DIREC(ip,k)
               up_fluxes(k,iw) = up_fluxes(k,iw)+DIRECTU(ip,k)
            end do
         end do
      else
         do k=1,nz1_rad    ! Loop over all layer BOUNDARIES
            do ip=lw_ipbeg,lw_ipend ! Loop over "longwave" wavelength intervals
               iw = (ip-lw_ipbeg)/ngauss_pts+lw_iwbeg
               dn_fluxes(k,iw) = dn_fluxes(k,iw)+CK1(ip,k)*EL2(ip,k)+  &  ! Down
                    CK2(ip,k)*EM2(ip,k)+CMB(ip,k)+DIRECT(ip,k)
               up_fluxes(k,iw) = up_fluxes(k,iw)+CK1(ip,k)*EL1(ip,k)+  &
                    CK2(ip,k)*EM1(ip,k)+CPB(ip,k)
            end do
         end do
      end if

   end if

   ! Calculate atmospheric heating rates (dT/dt) [K/s]:

   net_dTdt(1) = ZERO_f4  ! Set heating rate for "fake layer"
      
   do k=2,nz1_rad
      k_1 = k-1

      lyr_mass_fact = dzc(k)*rho_air(k)*cp_air
   
      fdiv_lw = (lw_upflux(k)-lw_dnflux(k))-(lw_upflux(k_1)-lw_dnflux(k_1))

         ! "longwave" heating rate [K/s]:
      net_dTdt(k) = real(fdiv_lw, kind=f4)/lyr_mass_fact
   end do

end if

end subroutine rad_postcalc


end module aerad_postcalc_mod
