!```````````````````````````````````````````````````````````````````
module aerad_txrad_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine two_stream     

!+  Defines matrix properties and sets up matrix coefficients that do not
!    depend on solar zenith angle or atmospheric temperature
                                    
use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ONE_f8, ZERO_f8
use aertxrad_mem_mod, only: b1,b2,ak,ee1,el1,ip_ibeg,ip_iend,u1i_2  &
     ,nz1_rad,w0,g0,gami,taul,em1,el2,em2,two_nz1_rad,af,bf,ef,rsfx

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:
                                                                     
real(f8) :: x1
integer :: ip,k,kd
   
!=== Executable statements:
                                                                  
! HERE WE DEFINE LAYER PROPERTIES FOLLOWING GENERAL SCHEME         
!  OF MEADOR AND WEAVOR. THEN WE SET UP LAYER PROPERTIES            
!  NEEDED FOR MATRIX:
                                                                       
do k=1,nz1_rad                                     
   do ip=ip_ibeg,ip_iend

      ! THESE ARE FOR TWOSTREAM AND HEMISPHERIC MEANS
      B1(ip,k) = u1i_2(ip)*(2._f8-W0(ip,k)*(ONE_f8+G0(ip,k)))
      B2(ip,k) = u1i_2(ip)*W0(ip,k)*(ONE_f8-G0(ip,k))
      AK(ip,k) = SQRT(ABS(B1(ip,k)**2-B2(ip,k)**2))
      GAMI(ip,k) = B2(ip,k)/(B1(ip,k)+AK(ip,k))
      x1 = AK(ip,k)*TAUL(ip,k)         !                      
      if(x1 <= 400._f8) then            !
         ee1(ip,k) = exp(-x1)          !
      else                             ! TIM add mod
         ee1(ip,k) = ZERO_f8              !
      endif                            !
      EL1(ip,k) = ONE_f8+GAMI(ip,k)*EE1(ip,k)
      EM1(ip,k) = ONE_f8-GAMI(ip,k)*EE1(ip,k)
      EL2(ip,k) = GAMI(ip,k)+EE1(ip,k)
      EM2(ip,k) = GAMI(ip,k)-EE1(ip,k)
      
   enddo
enddo
                                                            
! WE SEEK TO SOLVE AX(L-1)+BX(L)+EX(L+1) = D.                       
!  L=2N FOR EVEN L, L=N+1 FOR ODD L. THE MEAN INTENSITY (TMI/4PI)    
!  AND THE NET FLUX (FNET) ARE RELATED TO X'S AS NOTED IN ADD.       
!  FIRST WE SET UP THE COEFFICIENTS THAT ARE INDEPENDENT OF SOLAR    
!  ANGLE OR TEMPERATURE: A(I),B(I),E(I). D(I) IS DEFINED IN ADD: 
                                                                       
k = 0
do kd=2,two_nz1_rad-1,2
   k = k+1
   do ip=ip_ibeg,ip_iend
      !           HERE ARE THE EVEN MATRIX ELEMENTS                           
      AF(ip,kd) = EM1(ip,k+1)*EL1(ip,k)-EM2(ip,k+1)*EL2(ip,k)
      BF(ip,kd) = EM1(ip,k+1)*EM1(ip,k)-EM2(ip,k+1)*EM2(ip,k)
      EF(ip,kd) = EL1(ip,k+1)*EM2(ip,k+1)-EL2(ip,k+1)*EM1(ip,k+1)
      !           HERE ARE THE ODD MATRIX ELEMENTS EXCEPT FOR THE TOP.        
      AF(ip,kd+1) = EM1(ip,k)*EL2(ip,k)-EL1(ip,k)*EM2(ip,k)
      BF(ip,kd+1) = EL1(ip,k+1)*EL1(ip,k)-EL2(ip,k+1)*EL2(ip,k)
      EF(ip,kd+1) = EL2(ip,k)*EM2(ip,k+1)-EL1(ip,k)*EM1(ip,k+1)
   enddo
enddo
                                                                  
! HERE ARE THE TOP AND BOTTOM BOUNDARY CONDITIONS AS WELL AS THE    
!  BEGINNING OF THE TRIDIAGONAL SOLUTION DEFINITIONS. I ASSUME       
!  NO DIFFUSE RADIATION IS INCIDENT AT UPPER BOUNDARY:
                                                                       
do ip=ip_ibeg,ip_iend                                          
   AF(ip,1) = ZERO_f8                                              
   BF(ip,1) = EL1(ip,1)                                                 
   EF(ip,1) = -EM1(ip,1)                                                
   AF(ip,two_nz1_rad) = EL1(ip,nz1_rad)-RSFX(ip)*EL2(ip,nz1_rad)              
   BF(ip,two_nz1_rad) = EM1(ip,nz1_rad)-RSFX(ip)*EM2(ip,nz1_rad)              
   EF(ip,two_nz1_rad) = ZERO_f8                                              
enddo

end subroutine two_stream


!```````````````````````````````````````````````````````````````````
subroutine add_txrad

!+ Defines source terms, forms matrix for multiple layers and solves the
!   tri-diagonal equations to obtain mean intensity and net flux

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ONE_f8, ZERO_f8
use aertxrad_mem_mod, only: ntot_gpt,df,ds,as,xk,ck1,ck2,fnet,tmi,cos_mu  &
     ,sw_ipbeg,sw_ipend,nz1_rad,b3,sqrt3,g0,taul,ee3,opd,el3,sol,direct,b1  &
     ,ak,epsilon,w0,b2,cpb,cpp,cmb,cm,rsfx,lw_ipbeg,lw_ipend,ptemp,slope,u1s  &
     ,emis,ptempg,em1,em2,el1,el2,bf,af,ef,u1i,two_nz1_rad,pi,ip_ibeg,sw_on  &
     ,ip_iend,i_curr,j_curr

implicit none       

!=== Argument declarations:

! NONE

!== Local declarations:

!     *  Input               :  G0, Uo, RSFX, TAUL, OPD            *   
!     *  Output              :  B3, EE3, DIRECT, SFCS, CPB, CMB, DS *   

real(f8), dimension(ntot_gpt) :: sfcs
real(f8) :: DUo,B4,X2,X3,X4,C1,C2,CP1,CM1,X,x_lim
integer :: KINDEX,k,ip,k_1,kd

!=== Executable statements:
                                                                       
! THIS SUBROUTINE FORMS THE MATRIX FOR THE MULTIPLE LAYERS AND      
!  USES A TRIDIAGONAL ROUTINE TO FIND RADIATION IN THE ENTIRE        
!  ATMOSPHERE.                                                       
!                                                                       
! ******************************                                    
! *   CALCULATIONS FOR SOLAR   *                                    
! ******************************           
                      
if(sw_on) then

   DUo = ONE_f8/cos_mu

   do k=1,nz1_rad                                  
      k_1 = max(1,k-1)

      do ip=sw_ipbeg,sw_ipend
         B3(ip,k) = 0.5_f8*(ONE_f8-sqrt3*G0(ip,k)*cos_mu)     ! TIM mod            
         B4 = ONE_f8-B3(ip,k)                              
         X2 = TAUL(ip,k)*DUo         
         if(x2 > 400._f8) then   !
            x2 = 400._f8        ! TIM add
         endif                  !
         EE3(ip,k) = exp(-x2)
         X3 = OPD(ip,k)*DUo
         if(x3 > 400._f8) then   !
            x3 = 400._f8         ! TIM add
         endif                  !
         EL3(ip,k) = exp(-x3)*SOL(ip,k)    ! TIM mod
         DIRECT(ip,k) = cos_mu*EL3(ip,k)
         C1 = B1(ip,k)-DUo
         if(abs(C1) < epsilon) then
            c1 = sign(epsilon,C1)
         endif
         C2 = AK(ip,k)*AK(ip,k)-DUo*DUo                 
         if(abs(c2) <= epsilon) then
            c2 = epsilon
         endif
         CP1 = W0(ip,k)*(B3(ip,k)*C1+B4*B2(ip,k))/C2
         CPB(ip,k) = CP1*EL3(ip,k)
         if(k /= 1) then
            x4 = el3(ip,k_1)
         else
            x4 = sol(ip,k)        ! TIM mod
         endif
         CPP(ip,k) = CP1*X4     ! TIM mod
         CM1 = (CP1*B2(ip,k)+W0(ip,k)*B4)/C1
         CMB(ip,k) = CM1*EL3(ip,k)
         CM(ip,k) = CM1*X4
      enddo
   enddo
                                                            
   ! CALCULATE SFCS, THE SOURCE AT THE BOTTOM: 
                                                                       
   do ip=sw_ipbeg,sw_ipend
      sfcs(ip) = direct(ip,nz1_rad)*rsfx(ip)
   enddo
                                     
endif
                        
! ******************************                                    
! * CALCULATIONS FOR INFRARED. *                                    
! ******************************                                    
                                                                       
do k=1,nz1_rad                                 
   kindex = max(1,k-1)
   do ip=lw_ipbeg,lw_ipend
      B3(ip,k) = ONE_f8/(B1(ip,k)+B2(ip,k))
      CPP(ip,k) = (PTEMP(ip,KINDEX)+SLOPE(ip,k)*B3(ip,k))*U1S(ip)  ! TIM mod
      CPB(ip,k) = CPP(ip,k)+SLOPE(ip,k)*TAUL(ip,k)*U1S(ip)         !        
      CM(ip,k) = (PTEMP(ip,KINDEX)-SLOPE(ip,k)*B3(ip,k))*U1S(ip) 
      CMB(ip,k) = CM(ip,k)+SLOPE(ip,k)*TAUL(ip,k)*U1S(ip)
      EL3(ip,k)    = ZERO_f8        
      DIRECT(ip,k) = ZERO_f8
      EE3(ip,k)    = ZERO_f8
   enddo
enddo

do ip=lw_ipbeg,lw_ipend                        
   sfcs(ip) = emis(ip)*ptempg(ip)*pi
enddo
                                                                     
k = 0                                             
do kd=2,two_nz1_rad-1,2                                        
   k = k+1    
   do ip=ip_ibeg,ip_iend                                       
      ! HERE ARE THE EVEN MATRIX ELEMENTS                           
      DF(ip,kd) = (CPP(ip,k+1)-CPB(ip,k))*EM1(ip,k+1)-  &   ! TIM mod
           (CM(ip,k+1)-CMB(ip,k))*EM2(ip,k+1)
      ! HERE ARE THE ODD MATRIX ELEMENTS EXCEPT FOR THE TOP.        
      DF(ip,kd+1) = EL2(ip,k)*(CPP(ip,k+1)-CPB(ip,k))+  &   ! TIM mod
           EL1(ip,k)*(CMB(ip,k)-CM(ip,k+1))
   enddo
enddo
                                                 
! HERE ARE THE TOP AND BOTTOM BOUNDARY CONDITIONS AS WELL AS THE    
!  BEGINNING OF THE TRIDIAGONAL SOLUTION DEFINITIONS. I ASSUME NO    
!  DIFFUSE RADIATION IS INCIDENT AT THE TOP: 
                                                                       
do ip=ip_ibeg,ip_iend
   DF(ip,1) = -CM(ip,1)
   DF(ip,two_nz1_rad) = SFCS(ip)+RSFX(ip)*CMB(ip,nz1_rad)-CPB(ip,nz1_rad)
   DS(ip,two_nz1_rad) = DF(ip,two_nz1_rad)/BF(ip,two_nz1_rad)    
   AS(ip,two_nz1_rad) = AF(ip,two_nz1_rad)/BF(ip,two_nz1_rad)
enddo
                                                                       
! ********************************************                      
! *     WE SOLVE THE TRIDIAGONAL EQUATIONS   *                      
! ********************************************                      

x_lim = tiny(x)*1.e100_f8   ! Lower limit for 'x' below        ! TIM add
do kd=2,two_nz1_rad
   k = two_nz1_rad+1-kd     ! TIM add
   do ip=ip_ibeg,ip_iend
      x = (BF(ip,k)-EF(ip,k)*AS(ip,k+1))   ! TIM mod
      if(abs(x) < x_lim) then     !
         x = sign(ONE_f8,x)*x_lim   ! TIM add
      endif                       !
      X = ONE_f8/x                  !

      AS(ip,k) = AF(ip,k)*X
      DS(ip,k) = (DF(ip,k)-EF(ip,k)*DS(ip,k+1))*X
   enddo
enddo
                                            
do ip=ip_ibeg,ip_iend
   XK(ip,1) = DS(ip,1)
enddo
                                                                     
do kd=2,two_nz1_rad
   do ip=ip_ibeg,ip_iend
      XK(ip,kd) = DS(ip,kd)-AS(ip,kd)*XK(ip,kd-1)
   enddo
enddo

! ***************************************************************      
!    CALCULATE LAYER COEFFICIENTS, NET FLUX AND MEAN INTENSITY         
! ***************************************************************      
                                                                       
do k=1,nz1_rad
   do ip=ip_ibeg,ip_iend                                   
      CK1(ip,k) = XK(ip,2*k-1)
      CK2(ip,k) = XK(ip,2*k)
!### FNET and TMI not used in this application (similar quantities are instead
!     computed in `postcalc`):
!      FNET(ip,k)  = CK1(ip,k)  *( EL1(ip,k) -EL2(ip,k))   +  &               
!           CK2(ip,k) *( EM1(ip,k)-EM2(ip,k) ) + CPB(ip,k) -  & 
!           CMB(ip,k) - DIRECT(ip,k)
!                                                                       
!      TMI(ip,k)   =  EL3(ip,k) + U1I(ip) *( CK1(ip,k)  *  &                 
!           ( EL1(ip,k) + EL2(ip,k))   +  &                  
!           CK2(ip,k) *( EM1(ip,k)+EM2(ip,k) ) +  &            
!           CPB(ip,k) + CMB(ip,k) )                        
   enddo
enddo

end subroutine add_txrad


!```````````````````````````````````````````````````````````````````
subroutine refine_lwflux

!+ Calculate upward and downward intensities and fluxes using Gauss
!   Quadrature angles and weights

use univ_kind_defs_mod, only: f8
use univ_math_constants_mod, only: ONE_f8, ZERO_f8
use aertxrad_mem_mod, only: nz1_rad,direc,dintent,directu,lw_ipend,lw_ipbeg,a3  &
     ,a4,y5,a7,twopi,ptemp,slope,y1,y2,y4,y8,g_angle,u1i,u1s,taul,a1,a2,gami  &
     ,ngangles,b3,ak,y3,ee1,ck1,ck2,g_ang_weight,uintent,ptempg,emis,rsfx

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:
  
!     *  Input               :  PTEMP, SLOPE, Y3, B3, EE1, EE2     *    
!     *  Output              :  DINTENT, UINTENT, DIREC, DIRECTU   *    
    
real(f8) :: X4,YA,YB,CKP,CKM
integer :: KINDEX,k,kd,ip,ia

!=== Executable statements:
                                                                      
do k=1,nz1_rad                                    
   kindex = max(1,k-1)             
   do ip=lw_ipbeg,lw_ipend
      ! HERE WE DO NO SCATTERING COEFFICIENTS                      
      A3(ip,k) = PTEMP(ip,KINDEX)*twopi                         
      A4(ip,k) = twopi*SLOPE(ip,k)                              
      A7(ip,k) = A3(ip,k)                                     
      Y5(ip,k) = A4(ip,k)*TAUL(ip,k)                   
   enddo
                                                                  
   ! HERE WE DO SCATTERING                                         
!?not currently supported?   IF(IRS /= 0) THEN
   do ip=lw_ipbeg,lw_ipend
      X4 = SLOPE(ip,k)*(twopi*B3(ip,k)-U1S(ip))             
      A1(ip,k) = U1I(ip)-AK(ip,k)                            
      A2(ip,k) = GAMI(ip,k)*(AK(ip,k)+U1I(ip))                  
      A3(ip,k) = A3(ip,k)+X4                                  
      A7(ip,k) = A7(ip,k)-X4                                  
   enddo
!?not currently supported?   ENDIF
enddo
                                                                      
! CALCULATIONS FOR ALL GAUSS POINTS:
                                                                       
do k=1,nz1_rad                                
   do ia=1,ngangles

      ! HERE WE DO NO SCATTERING COEFFS
      do ip=lw_ipbeg,lw_ipend
!%waste CPU%         Y1(ip,ia,k)  =  0.d0                                    
!%         Y2(ip,ia,k)  =  0.d0                                        
         Y4(ip,ia,k) = A7(ip,k)-A4(ip,k)*g_angle(ia)
         Y8(ip,ia,k) = A3(ip,k)+A4(ip,k)*g_angle(ia)
!%      enddo
                                                                 
      ! HERE WE DO SCATTERING
                                      
!%      do ip=lw_ipbeg,lw_ipend
         YA = A1(ip,k)*(Y3(ip,ia,k)-EE1(ip,k))/(AK(ip,k)*g_angle(ia)-ONE_f8)
         YB = A2(ip,k)*(ONE_f8- EE1(ip,k)*Y3(ip,ia,k))/  &
              (AK(ip,k)*g_angle(ia)+ONE_f8)
         CKP = CK1(ip,k)+CK2(ip,k)
         CKM = CK1(ip,k)-CK2(ip,k)
         Y1(ip,ia,k) = CKP*YB+CKM*YA                                      
         Y2(ip,ia,k) = CKP*YA+CKM*YB                                     
      enddo

   enddo
enddo
                                                                   
do k=1,nz1_rad                                  
   do ip=lw_ipbeg,lw_ipend
      DIREC(ip,k) = ZERO_f8                     
      DIRECTU(ip,k) = ZERO_f8                                      
   enddo
enddo
                                                                  
! DIREC IS DOWNWARD FLUX. DIRECTU IS UPWARD FLUX.                   
!  CALCULATE DINTENT THE DOWNWARD INTENSITY AND DIREC THE DOWNWARD FL
                                                                       
do ia=1,ngangles                               
   do ip=lw_ipbeg,lw_ipend
      DINTENT(ip,ia,1) = (ONE_f8-Y3(ip,ia,1))*Y4(ip,ia,1)+Y1(ip,ia,1)
      DIREC(ip,1) = DIREC(ip,1)+DINTENT(ip,ia,1)*g_ang_weight(ia)
   enddo
enddo
                                                     
! DINTENT IS DOWNWARD INTENSITY * TwoPI. DIREC IS THE DOWNWARD FLUX. 
                                                                       
do k=2,nz1_rad                             
   do ia=1,ngangles                             
      do ip=lw_ipbeg,lw_ipend
         
         DINTENT(ip,ia,k) = DINTENT(ip,ia,k-1)*Y3(ip,ia,k)  &         
              +Y1(ip,ia,k)+Y5(ip,k)+(ONE_f8-Y3(ip,ia,k))*Y4(ip,ia,k)

         DIREC(ip,k) = DIREC(ip,k)+DINTENT(ip,ia,k)*g_ang_weight(ia)

      enddo
   enddo
enddo
                                                                       
! UINTENT IS THE UPWARD INTENSITY * TwoPI. DIRECTU IS THE UPWARD FLUX.
!  ASSUME THAT THE REFLECTIVITY IS LAMBERT.                          
                                            
do ia=1,ngangles                           
   do ip=lw_ipbeg,lw_ipend
      UINTENT(ip,ia,nz1_rad) = PTEMPG(ip)*EMIS(ip)*twopi+RSFX(ip)*  &
           DIREC(ip,nz1_rad)*2._f8
      DIRECTU(ip,nz1_rad) = DIRECTU(ip,nz1_rad)+UINTENT(ip,ia,nz1_rad)*  &
           g_ang_weight(ia)
   enddo
enddo

kd = nz1_rad                                                       
do k=2,nz1_rad                           
   kd = kd-1                         
   do ia=1,ngangles                           
      do ip=lw_ipbeg,lw_ipend

         UINTENT(ip,ia,kd) = (UINTENT(ip,ia,kd+1)-Y5(ip,kd+1))  &     
              *Y3(ip,ia,kd+1)+Y2(ip,ia,kd+1)+  &        
              (ONE_f8-Y3(ip,ia,kd+1))*Y8(ip,ia,kd+1)       

         DIRECTU(ip,kd) = DIRECTU(ip,kd)+g_ang_weight(ia)*UINTENT(ip,ia,kd)

      enddo
   enddo
enddo

end subroutine refine_lwflux


end module aerad_txrad_mod
