!```````````````````````````````````````````````````````````````````
module radn_CO2ea_mem_mod

!~ 18.11.2008 (last modified; TIM)

!+ Memory and parameters specific to CO2 gas absorptivity/emissivity calcs

use univ_kind_defs_mod, only: f4, f8

implicit none

private

integer, public, parameter :: DIM_mubar = 19,  &  ! # OF COSZ POINTS FOR CO2 CALCS
                        DIM_CO2_data_pr = 9,  &  ! # OF PRESS. FOR CO2 CALCS
                        DIM_CO2_data_tm = 3,  &  ! # OF TEMPS FOR CO2 CALCS
                        DIM_CO2_data_cn = 25,  & ! # OF GAS CONC FOR CO2 CALCS
                        DIM_CO2_emiss_pr = 3000  !

real(f4), public, dimension(DIM_mubar) :: MUBAR            ! CO2FIT DIMENSIONS
real(f8), public, dimension(DIM_CO2_data_pr) :: PRLN
real(f4), public, dimension(DIM_CO2_data_tm) :: TR
real(f4), public, dimension(DIM_CO2_data_cn) :: WR

real(f4), public, allocatable, dimension(:,:,:) ::   &
     sw_CO2_emiss  &  ! [3,DIM_mubar,0:DIM_CO2_emiss_pr]
     ,lw_CO2_emiss    ! [3,0:DIM_CO2_emiss_pr,0:DIM_CO2_emiss_pr]
real(f4), public, allocatable, dimension(:) ::  &     ! [0:DIM_CO2_emiss_pr]
     CO2_emiss_pr

real(f4), public, dimension(DIM_CO2_data_pr,DIM_CO2_data_cn,DIM_CO2_data_tm) :: &
     abs15,abssol

real(f4), public :: CO2_emiss_pr_incr,cem
integer, public :: ncon


data CO2_emiss_pr_incr /0.005_f4/

data ncon / DIM_CO2_data_cn /
data cem / 7.362e22_f4 /

data mubar / 0.02_f4, 0.03_f4, 0.04_f4, 0.05_f4, 0.06_f4, 0.07_f4, 0.085_f4,  &
     0.1_f4 , 0.15_f4, 0.2_f4 , 0.25_f4, 0.3_f4 , 0.4_f4 , 0.5_f4  ,  &
     0.6_f4 , 0.7_f4 , 0.8_f4 , 0.9_f4 , 1._f4 /

! CO2 interpolation pressure grid points; ln(pressure in mb):

data prln / -13.81551056_f8, -11.51292547_f8, -9.210340372_f8,  &
     -6.907755279_f8, -4.605170186_f8, -2.302585093_f8,  &
     0._f8,  2.302585093_f8,  4.605170186_f8  /
  
data tr / 100._f4, 200._f4, 300._f4 /

data wr / 1.0e18_f4, 2.0e18_f4, 5.0e18_f4, 1.0e19_f4, 2.0e19_f4, 5.0e19_f4,  &
     1.0e20_f4, 2.0e20_f4, 5.0e20_f4, 1.0e21_f4, 2.0e21_f4, 5.0e21_f4,  &
     1.0e22_f4, 2.0e22_f4, 5.0e22_f4, 1.0e23_f4, 2.0e23_f4, 5.0e23_f4,  &
     1.0e24_f4, 2.0e24_f4, 5.0e24_f4, 1.0e25_f4, 2.0e25_f4, 5.0e25_f4,  &
     1.0e26_f4 /

include 'radn_CO2data.f90'    ! Fill abssol, abs15 arrays

end module radn_CO2ea_mem_mod
