!```````````````````````````````````````````````````````````````````
module radn_calc_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine radn_calc(n1, n2, n3, ia, iz, ja, jz, cosZ, msdist, PRAMS_T,  &
     PRAMS_p, radheat_rate, rlong, PRAMS_sfcT, sw_g_flux,  &
     lw_15_dn_flux, lw_15_up_flux, lw_out_dn_flux, lw_out_up_flux,  &
     sw_d_flux, albedt, nazm, solar_azm, cosz_horizon, tazm_angle,  &
     tslope_angle, shadow_tog, slpasp_tog, TCx_obstruct, TCz_obstruct, zt,  &
     rtgt, tautot_ij, conrath_nu, emissivity, idx_mask_xy,  &
     ext_bottom_other_sflux_down, ext_bottom_SWflux_down, ext_topo_factor_sflux)

!+ Main driver for radiative calculations

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use univ_math_constants_mod, only: twopi_f4, pi_f4
use radn_main_mem_mod, only: fo,po,fmcut,rdlev_f,rdlev_t,xo,idx02,xs,tg,rm1,rm2  &
     ,rm3,acosz,scosz,ym,pl,fluxes,ir15dn,iroutdn,ir15net,ironet,sco2p,fmcnt  &
     ,pst,psp,x,yp,y2,ir15up,iroutup
use radn_BBf_mem_mod, only: deltanu, egognd, eg15gnd
use body_specific_constants_mod, only: cp, g_Xoid, solar_mean_broadband
use spline_mod, only: calc_spline
use radn_IRout_mod, only: radn_IRout
use radn_IRin_mod, only: radn_IRin
use radn_sdust_mod, only: radn_sdust
use radn_sCO2_mod, only: radn_sCO2
use radn_pt_mod, only: radn_pt

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1,n2,n3,nazm,shadow_tog,slpasp_tog,ia,iz,ja,jz
integer, dimension(n2,n3), intent(IN) :: idx_mask_xy
real(f4), dimension(n1,n2,n3), intent(IN) :: PRAMS_T, PRAMS_p
real(f4), dimension(n1,n2,n3), intent(INOUT) :: sw_g_flux, lw_15_dn_flux,  &
     lw_15_up_flux, lw_out_dn_flux, lw_out_up_flux, sw_d_flux
real(f4), dimension(n2,n3), intent(IN) :: cosZ,PRAMS_sfcT,rtgt  &
     ,albedt,solar_azm,tazm_angle,tslope_angle, emissivity
real(f4), dimension(n2,n3,nazm), intent(IN) :: cosz_horizon,TCx_obstruct,TCz_obstruct
real(f4), dimension(n1), intent(IN) :: zt
real(f4), intent(IN) :: msdist

real(f4), dimension(n2,n3), intent(INOUT) :: rlong,  &
     ext_bottom_other_sflux_down, ext_bottom_SWflux_down, ext_topo_factor_sflux
real(f4), dimension(n1,n2,n3), intent(INOUT) :: radheat_rate
real(f4), dimension(n2,n3), intent(INOUT) :: tautot_ij,conrath_nu

!== Local declarations:

integer :: iii,jjj,ik,lrz,ia0,ia1,LB_end,LM_end,topOL,swcut
real(f4) :: flxc,cazm,atmp,aint,cosz_h,fcosz,sinz,cosai,z_obst  &
     ,x_obst,sw_cutoff
logical :: want_rtflux_output

!=== Executable statements:

!*** Set some constants:

flxc = pi_f4*deltanu
LM_end = rdlev_f/2
LB_end = LM_end+1
topOL = rdlev_t/2

!*** Fill some constant interpolative arrays (for dealing with fine mesh)

do ik=1,fmcut
   xo(ik) = real(n1-ik)+0.5_f4
enddo

!*** MAIN COMPUTATIONAL LOOP; LOOP OVER ALL I AND J:
                 
do jjj=ja,jz
   do iii=ia,iz

      !*** Fill sun angle quantities:

      cazm = solar_azm(iii,jjj)     ! Current solar azimuth for this gridpoint
      acosz = cosZ(iii,jjj)     ! Current cos(solar zenith angle) " "

      !*** Calculate cosz_horizon array location/interpolative indices (using
      !     solar azimuth values):

      atmp = cazm/(twopi_f4/real(nazm))
      ia0 = int(atmp)
      aint = atmp-real(ia0)   ! Location crucial to avoid ia0 complications

      if(ia0 == 0) then
         ia0 = nazm
         ia1 = 1
      elseif(ia0 == nazm) then
         ia1 = 1
      else
         ia1 = ia0+1
      endif

      !*** Obtain horizon cosz for current solar azimuth via interpolation:

      if(shadow_tog == 0) then
         cosz_h = ZERO_f4
      else
         cosz_h = (ONE_f4-aint)*cosz_horizon(iii,jjj,ia0)+  &
              aint*cosz_horizon(iii,jjj,ia1)
         x_obst = (ONE_f4-aint)*TCx_obstruct(iii,jjj,ia0)+  &
              aint*TCx_obstruct(iii,jjj,ia1)
         z_obst = (ONE_f4-aint)*TCz_obstruct(iii,jjj,ia0)+  &
              aint*TCz_obstruct(iii,jjj,ia1)

         if(z_obst < epsilon(z_obst) .or. acosz < 1.e-6_f4) then ! Extension
                                                              !  or night
            swcut = 0

         else        ! Shadowing
            sw_cutoff = z_obst-(x_obst*tan(asin(acosz))) 

            if(sw_cutoff < zt(n1-1)*rtgt(iii,jjj)) then
               swcut = 0
               atmp = ZERO_f4
               shadow_top: do ik=2,n1        ! Find top of shadow
                  aint = zt(ik)*rtgt(iii,jjj)
                  if(sw_cutoff < aint) then
                     if(ik /= 2) then
                        if(abs(aint-sw_cutoff) < abs(atmp-sw_cutoff)) then
                           swcut = n1-ik+1
                        else
                           swcut = n1-ik+2
                        endif
                     else
                        swcut = n1-ik+1
                     endif
                     exit shadow_top
                  endif
                  atmp = aint
               enddo shadow_top

               if(swcut == 0) then
                  write(*,*) 'ERROR: unable to find shadow top.'
                  stop '*** ERROR: radn_calc ***'
               endif
               
               shdwtop_lev: do ik=idx02,1,-1        ! Find level of top of shadow
                  if(xs(ik) < real(swcut)) then
                     swcut = ik
                     exit shdwtop_lev
                  endif
               enddo shdwtop_lev

               swcut = ik/2+1+rdlev_t/2  ! Column is still sunlit above this level
            else  ! tan(elev angle) asymptotes at EA=pi/2.; need to be careful
               swcut = LB_end+1  ! Column sunlit completely
            endif

         endif
      endif

      if (slpasp_tog > 0) then
         sinz = sin(acos(acosz))

         ! If cosai < 0., point not sunlit
         cosai = (acosz*cos(tslope_angle(iii,jjj)))+  &
              (sinz*sin(tslope_angle(iii,jjj))*  &
              cos(solar_azm(iii,jjj)-tazm_angle(iii,jjj)))
      else
         cosai = 2._f4  ! Sentinel value (avoids uninitialized use in `sdust`)
      end if

      !*** Calculate solar flux at top of column:

     !write (*,'(//a//)') '***** WARNING: using Ames MGCM solar constant *****'

      if(acosz < 1.e-6_f4) then
         if(acosz >= cosz_h) then    ! Horizon extension
!            scosz = 1.353e3*0.04/xsdist   ! Ames MGCM value
!@            scosz = 1.3676e3*0.04/xsdist! Solar const (at 1AU) from MARS,1992
            scosz = solar_mean_broadband*0.04_f4/msdist
         else    ! Sun is below horizon 
            scosz = ZERO_f4
         endif
      else   ! Regular case
!         scosz = 1.353e3*acosz/xsdist    ! Ames MGCM value
!@         scosz = 1.3676e3*acosz/xsdist  ! Solar const (at 1AU) from MARS,1992
         scosz = solar_mean_broadband*acosz/msdist
      endif

      !*** Fill ground temperature:

      tg = PRAMS_sfcT(iii,jjj)

      !*** Create temperature and pressure profiles for this grid point:

      call radn_pt(iii,jjj,rm1,rm2,rm3,PRAMS_p,PRAMS_T,LM_end)

      !*** Calculate mass heat capacity per unit area values for each layer:
      !    (Pressures converted to Pa from mb)

      do ik=topOL,n1-fmcut+topOL  !LM_end     ! Loop over coarse mesh OL
         ym(ik) = (pl(ik+1)-pl(ik))*100._f4*cp/g_Xoid
      enddo

      if (idx_mask_xy(iii,jjj) > 0) then
         want_rtflux_output = .TRUE.      ! Calc/save 2-stream fluxes
      else
         want_rtflux_output = .FALSE.
      end if

      !*** BEGIN RADIATIVE CALCULATIONS:

      ! Set surface emissivity:
      egognd = emissivity(iii,jjj)
      eg15gnd = emissivity(iii,jjj)

      !*** CALCULATE SOLAR ABSORPTION BY ATMOSPHERIC CO2:

      call radn_sCO2(cosz_h,fcosz,LB_end,LM_end,swcut)

      !*** CALCULATE SOLAR ABSORPTION BY ATMOSPHERIC DUST:

      call radn_sdust(albedt(iii,jjj),cosz_h,fcosz,LB_end,LM_end,swcut,cosai  &
           ,tautot_ij(iii,jjj),conrath_nu(iii,jjj))

      ! *Incident* surface "shortwave" flux [W/m^2]:

      ! The portion of the stellar flux (at the body's surface) that *IS*
      !  treated by the "shortwave" atmospheric RT calculations; divide by
      !  (1.-sfc_albedo) to obtain direct SW flux from net SW flux:
      ext_bottom_SWflux_down(iii,jjj) = fluxes(LB_end)/(ONE_f4-albedt(iii,jjj))

      ! The portion of the stellar flux (at the body's surface) that *IS NOT*
      !  treated by the "shortwave" atmospheric RT calculations:
      ext_bottom_other_sflux_down(iii,jjj) = ZERO_f4

      if (slpasp_tog > 0) then
         ! Multiply by cosai to take both topography orientation AND stellar
         !  incidence angle into account (after dividing by cos_mu to obtain
         !  the non-angle-corrected incident stellar flux):
         ext_topo_factor_sflux(iii,jjj) = cosai/acosz
      else
         ! Take ONLY cos(Z) into account:
         ext_topo_factor_sflux(iii,jjj) = ONE_f4
      end if

      !*** CALCULATE THE INFRARED ABSORPTION INSIDE THE 15 MICRON BAND:

      call radn_IRin(flxc,LB_end,LM_end)

      !*** CALCULATE THE INFRARED ABSORPTION OUTSIDE THE 15 MICRON BAND:

      call radn_IRout(LB_end, LM_end)

      !*** Fill rlong with surface IR flux value:

      rlong(iii,jjj) = ir15dn(LB_end)+iroutdn(LB_end)

      !*** Prepare for output to PRAMS:

!*** Fill PRAMS total radiative heating rate array:
      
! Coarse mesh region:

      lrz = n1
      do ik=topOL+1,n1-fmcut+topOL
         radheat_rate(lrz,iii,jjj) =  &         ! Total radiative heating rate
              (((ir15net(ik+1)+ironet(ik+1)+sco2p(ik+1))-  &  ! at PRAMS level
              (ir15net(ik)+ironet(ik)+sco2p(ik)))+  &
              (fluxes(ik)-fluxes(ik+1)))/ym(ik)
         lrz = lrz-1
      enddo

      if (want_rtflux_output) then
         lrz = n1
         do ik=topOL+1,n1-fmcut+topOL
            sw_g_flux(lrz,iii,jjj) = sco2p(ik)
            lw_15_dn_flux(lrz,iii,jjj) = ir15dn(ik)
            lw_15_up_flux(lrz,iii,jjj) = ir15up(ik)
            lw_out_dn_flux(lrz,iii,jjj) = iroutdn(ik)
            lw_out_up_flux(lrz,iii,jjj) = iroutup(ik)
            sw_d_flux(lrz,iii,jjj) = fluxes(ik)
            lrz = lrz-1
         end do
      end if

! Fine mesh region:

      lrz = 0
      do ik=idx02-fmcnt-3,idx02,2
         lrz = lrz+1
         pst(lrz) = xs(ik)
      enddo

      !call spline4(x,yp,lrz,u,y2)  ! Not needed since filled in last call to spline4
      call speval4(x,yp,y2,n1,xo,po,fmcut)    ! Layer boundary pressures (FM)

      if (want_rtflux_output) then
         lrz = 0
         do ik=n1-fmcut+topOL,LB_end
            lrz = lrz+1  ! Flux at each level
            psp(lrz) = ir15dn(ik)
         end do
         call calc_spline(pst,psp,lrz,y2)
         call speval4(pst,psp,y2,lrz,xo,fo,fmcut)
         lrz = fmcut
         do ik=fmcut,1,-1
            lw_15_dn_flux(lrz,iii,jjj) = fo(ik)
            lrz = lrz-1
         end do

         lrz = 0
         do ik=n1-fmcut+topOL,LB_end
            lrz = lrz+1  ! Flux at each level
            psp(lrz) = ir15up(ik)
         end do
         call calc_spline(pst,psp,lrz,y2)
         call speval4(pst,psp,y2,lrz,xo,fo,fmcut)
         lrz = fmcut
         do ik=fmcut,1,-1
            lw_15_up_flux(lrz,iii,jjj) = fo(ik)
            lrz = lrz-1
         end do

         lrz = 0
         do ik=n1-fmcut+topOL,LB_end
            lrz = lrz+1  ! Flux at each level
            psp(lrz) = iroutdn(ik)
         end do
         call calc_spline(pst,psp,lrz,y2)
         call speval4(pst,psp,y2,lrz,xo,fo,fmcut)
         lrz = fmcut
         do ik=fmcut,1,-1
            lw_out_dn_flux(lrz,iii,jjj) = fo(ik)
            lrz = lrz-1
         end do

         lrz = 0
         do ik=n1-fmcut+topOL,LB_end
            lrz = lrz+1  ! Flux at each level
            psp(lrz) = iroutup(ik)
         end do
         call calc_spline(pst,psp,lrz,y2)
         call speval4(pst,psp,y2,lrz,xo,fo,fmcut)
         lrz = fmcut
         do ik=fmcut,1,-1
            lw_out_up_flux(lrz,iii,jjj) = fo(ik)
            lrz = lrz-1
         end do
      end if

      lrz = 0
      do ik=n1-fmcut+topOL,LB_end
         lrz = lrz+1  ! Net flux at each level (excluding SW dust contribution)
         psp(lrz) = ir15net(ik)+ironet(ik)+sco2p(ik)   
      enddo

      call calc_spline(pst,psp,lrz,y2)
!#      call spline4(pst,psp,lrz,u,y2)                   ! For fluxes (FM)
      call speval4(pst,psp,y2,lrz,xo,fo,fmcut)         !

      lrz = fmcut
      do ik=fmcut,2,-1  ! Radiative heating rate at PRAMS level (minus SW dust)
         radheat_rate(lrz,iii,jjj) = (fo(ik-1)-fo(ik))/  &
              ((po(ik-1)-po(ik))*100._f4*cp/g_Xoid)
         lrz = lrz-1
      enddo

      ! If Sun is up, calculate SW dust contribution; otherwise its
      !   contribution to the total radiative heating rate is zero, and we
      !   don't waste CPU time computing it

      ! do compute if shadowed

      if (fcosz > 1.e-6_f4) then

         if (want_rtflux_output) then
            lrz = 0
            do ik=n1-fmcut+topOL,LB_end
               lrz = lrz+1  ! Flux at each level
               psp(lrz) = sco2p(ik)
            end do
            call calc_spline(pst,psp,lrz,y2)
            call speval4(pst,psp,y2,lrz,xo,fo,fmcut)
            lrz = fmcut
            do ik=fmcut,1,-1
               sw_g_flux(lrz,iii,jjj) = fo(ik)
               lrz = lrz-1
            end do

            lrz = 0
            do ik=n1-fmcut+topOL,LB_end
               lrz = lrz+1  ! Flux at each level
               psp(lrz) = fluxes(ik)
            end do
            call calc_spline(pst,psp,lrz,y2)
            call speval4(pst,psp,y2,lrz,xo,fo,fmcut)
            lrz = fmcut
            do ik=fmcut,1,-1
               sw_d_flux(lrz,iii,jjj) = fo(ik)
               lrz = lrz-1
            end do
         end if

         lrz = 0
         do ik=n1-fmcut+topOL,LB_end
            lrz = lrz+1
            psp(lrz) = fluxes(ik)    ! Fill intermediate array
         enddo

         call calc_spline(pst,psp,lrz,y2) 
!#         call spline4(pst,psp,lrz,u,y2)              ! Fit SW dust fluxes
         call speval4(pst,psp,y2,lrz,xo,fo,fmcut)    !

         lrz = fmcut
         do ik=fmcut,2,-1   ! Add SW dust DT/dt to total
            radheat_rate(lrz,iii,jjj) = radheat_rate(lrz,iii,jjj)+  &
                 (-fo(ik-1)+fo(ik))/((po(ik-1)-po(ik))*100._f4*cp/g_Xoid)
            lrz = lrz-1
         enddo
         
      endif

! Lowest PRAMS level ("fake"):

      radheat_rate(1,iii,jjj) = radheat_rate(2,iii,jjj) ! Last PRAMS sigZ level

   enddo        ! End iii and jjj loops
enddo           !

end subroutine radn_calc


end module radn_calc_mod
