!```````````````````````````````````````````````````````````````````
module radn_init1_mod

!~ 23.09.2008 (last modified; TIM)

!+

implicit none

private

public :: radn_init1

contains


!```````````````````````````````````````````````````````````````````
subroutine radn_init1(pmin,pmax)

!~ 18.11.2008 (last modified; TIM)

!+ Radiation initialization routine (1 of 2)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4
use radn_CO2ea_mem_mod, only: DIM_CO2_emiss_pr,CO2_emiss_pr_incr,CO2_emiss_pr  &
     ,DIM_mubar,mubar,cem,abssol,wr,tr,prln,ncon,DIM_CO2_data_pr  &
     ,DIM_CO2_data_cn,DIM_CO2_data_tm,sw_CO2_emiss,lw_CO2_emiss,abs15
use radn_BBf_mem_mod, only: deltanu,rdelnu
use radn_CO2fit_mod, only: co2fit4

implicit none

!=== Argument declarations:

real(f4), intent(IN) :: pmin,pmax

!== Local declarations:

!real(f_normal), dimension(3,DIM_mubar,0:DIM_CO2_emiss_pr) :: sw_CO2_emiss
!real(f_normal), dimension(3,0:DIM_CO2_emiss_pr,0:DIM_CO2_emiss_pr) ::  &
!     lw_CO2_emiss
!real(f_normal), dimension(0:DIM_CO2_emiss_pr) :: CO2_emiss_pr
!common /CO2_emissivity/ sw_CO2_emiss,lw_CO2_emiss,CO2_emiss_pr

integer :: iq,ic,ib
real(f4) :: w,pavg
real(f4), dimension(3) :: outcoef

!=== Executable statements:

!*** Calculate a constant for use elsewhere:

rdelnu  = (2.55e4_f4)/deltanu

!*** Calculate pressure interval from highest initial PRAMS pressure + 0.1mb
!     (Changed to +0.2mb to combat pressure change problem during runtime)

CO2_emiss_pr_incr = (pmax*0.01_f4+0.2_f4)/real(DIM_CO2_emiss_pr, kind=f4)

!*** Evenly spaced pressure levels to interpolate CO2 emissivities from:

CO2_emiss_pr(0) = ZERO_f4
do iq=1,DIM_CO2_emiss_pr
   CO2_emiss_pr(iq) = CO2_emiss_pr(iq-1)+CO2_emiss_pr_incr
enddo

!*** CO2 shortwave absorption coefficients between level `iq` and the top of
!     the atmosphere; function of cos(solar zenith angle):

do iq=1,DIM_CO2_emiss_pr
   do ic=1,DIM_mubar
      pavg = 0.5_f4*CO2_emiss_pr(iq)   ! Actually (CO2_emiss_pr(iq)+0.)
      w = cem*pavg/mubar(ic)
      call co2fit4(abssol,w,pavg,outcoef,DIM_CO2_data_pr,DIM_CO2_data_cn  &
           ,DIM_CO2_data_tm,wr,tr,real(prln, kind=f4),ncon)
      sw_CO2_emiss(1,ic,iq) = outcoef(1)
      sw_CO2_emiss(2,ic,iq) = outcoef(2)
      sw_CO2_emiss(3,ic,iq) = outcoef(3)
   enddo
enddo

do ic=1,DIM_mubar
   sw_CO2_emiss(1,ic,0) = ZERO_f4     ! No SW absorption if no atmospheric mass
   sw_CO2_emiss(2,ic,0) = ZERO_f4     !
   sw_CO2_emiss(3,ic,0) = ZERO_f4     !
enddo

!*** CO2 15um emissivity coefficients between levels `iq` and `ib` for the
!     entire range of pressures in `CO2_emiss_pr`:

do iq=0,DIM_CO2_emiss_pr-1
   do ib=iq+1,DIM_CO2_emiss_pr
      w = cem*abs(CO2_emiss_pr(iq)-CO2_emiss_pr(ib))
      pavg = 0.5_f4*(CO2_emiss_pr(iq)+CO2_emiss_pr(ib))
      call co2fit4(abs15,w,pavg,outcoef,DIM_CO2_data_pr,DIM_CO2_data_cn  &
           ,DIM_CO2_data_tm,wr,tr,real(prln, kind=f4),ncon)
      lw_CO2_emiss(1,iq,ib) = outcoef(1)   !
      lw_CO2_emiss(2,iq,ib) = outcoef(2)   !
      lw_CO2_emiss(3,iq,ib) = outcoef(3)   ! Fill CO2 emiss array symmetrically
      lw_CO2_emiss(1,ib,iq) = outcoef(1)   !
      lw_CO2_emiss(2,ib,iq) = outcoef(2)   !
      lw_CO2_emiss(3,ib,iq) = outcoef(3)   !
   enddo
enddo

do iq=0,DIM_CO2_emiss_pr
   lw_CO2_emiss(1,iq,iq) = ZERO_f4   !
   lw_CO2_emiss(2,iq,iq) = ZERO_f4   !  Zero out diagonal terms of CO2 emiss array
   lw_CO2_emiss(3,iq,iq) = ZERO_f4   !     ************ NEED?????? ********
enddo

end subroutine radn_init1


end module radn_init1_mod
