!```````````````````````````````````````````````````````````````````
module radn_mem_mod

!~ 24.09.2008 (last modified; TIM)

!+

implicit none

private

public :: radn_mem_AD

contains


!```````````````````````````````````````````````````````````````````
subroutine radn_mem_AD(on_or_off)

!+ Allocates/deallocates radiation routine work arrays

!~ 14.01.2011 (last modified; TIM)

use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use radn_main_mem_mod
use radn_netflux_mem_mod
use radn_dust_mem_mod
use radn_CO2ea_mem_mod
use radn_BBf_mem_mod

implicit none

!=== Argument declarations:

logical, intent(IN) :: on_or_off

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'radn_mem_AD'

integer :: LB_end,aerr

!=== Executable statements:

if(on_or_off) then

   !*** Allocate GLOBAL work arrays:

   LB_end = rdlev_f/2+1
   
   ! In module -- radn_dust_mem:
   
   allocate(DUST_taucum(LB_end), stat=aerr)   ! Allocate `DUST_taucum` array
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit01')
   allocate(emd(LB_end,LB_end),emisdst(LB_end,LB_end)  &
        ,mdst(LB_end,LB_end),nmdst(LB_end,LB_end), stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit02')
   
   ! In module -- radn_BBf_mem:

   allocate(bbp(LB_end), stat=aerr)        ! Allocate black body value array
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit03')
   
   ! In module -- radn_main_mem:
   
   allocate(u(idx02),x(idx02),yp(idx02),yt(idx02),y2(idx02)  &
        ,xs(idx02),psp(idx02),pst(idx02),xo(idx03),po(idx03),fo(idx03)  &
        , stat=aerr)
   !     ,xs(idx02+fmpc-1),psp(idx02),pst(idx02))
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit04')
   
   allocate(obft(idx03),obfb(idx03),obe(idx03),oblt(idx03),oblb(idx03)  &
        , stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit05')
   
   allocate(irtotal(LB_end),ir15dn(LB_end),ir15up(LB_end),ir15net(LB_end)  &
        ,ir15(LB_end),iroutdn(LB_end),iroutup(LB_end),ironet(LB_end)  &
        ,irout(LB_end),irnet(LB_end), stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit06')
   
   allocate(suntot(LB_end),tl(LB_end),pl(LB_end),pl2(LB_end),tl2(LB_end)  &
        ,ym(LB_end),sco2p(LB_end),sco2(LB_end),fluxes(LB_end),sdust(LB_end)  &
        , stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit07')

   ! In module -- radn_CO2ea_mem:

   allocate(sw_CO2_emiss(3,DIM_mubar,0:DIM_CO2_emiss_pr)  &
        ,lw_CO2_emiss(3,0:DIM_CO2_emiss_pr,0:DIM_CO2_emiss_pr)  &
        ,CO2_emiss_pr(0:DIM_CO2_emiss_pr), stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
        'radn_meminit08')

else

   !*** Deallocate GLOBAL work arrays:
 
   ! In module -- radn_dust_mem:
  
   deallocate(DUST_taucum, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit01')
   deallocate(emd,emisdst,mdst,nmdst, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit02')
   
   ! In module -- radn_BBf_mem:

   deallocate(bbp, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit03')
   
   ! In module -- radn_main_mem:
   
   deallocate(u,x,yp,yt,y2,xs,psp,pst,xo,po,fo, stat=aerr)
   !     ,xs(idx02+fmpc-1),psp(idx02),pst(idx02))
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit04')
   
   deallocate(obft,obfb,obe,oblt,oblb, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit05')
   
   deallocate(irtotal,ir15dn,ir15up,ir15net,ir15,iroutdn,iroutup,ironet  &
        ,irout,irnet, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit06')

   deallocate(suntot,tl,pl,pl2,tl2,ym,sco2p,sco2,fluxes,sdust, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit07')

   ! In module -- radn_CO2ea_mem:

   deallocate(sw_CO2_emiss,lw_CO2_emiss,CO2_emiss_pr, stat=aerr)
   if (aerr /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
        'radn_meminit08')

endif

end subroutine radn_mem_AD


end module radn_mem_mod
