!```````````````````````````````````````````````````````````````````
module radn_netflux_mem_mod

!~ 17.11.2008 (last modified; TIM)

!+ Memory and parameters specific to the `NETFLUX` data array

use univ_kind_defs_mod, only: f4, f8

implicit none

integer, parameter :: MMAXTOT = 35,  &   ! TAU-TOTAL DIMENSION
                       MMAXTAU = 17,  &   ! TAU-CUMULATIVE DIMENSION
                        MTAMTO = 595,  &   ! PRODUCT OF ABOVE TWO DIMENSIONS
                        MMAXU0 = 17,  &    ! MU0 DIMENSION
                        MMAXALB = 3        ! ALBEDO DIMENSION

real(f8), dimension(MTAMTO,MMAXU0,MMAXALB) :: fluxnet

real(f4), dimension(MMAXTOT) :: tautots     ! NETFLUX DIMENSIONS
real(f4), dimension(MMAXTAU) :: taucum0
real(f4), dimension(MMAXU0) :: u0vals
real(f4), dimension(MMAXALB) :: albedos

!*** FILL ABOVE ARRAYS:

! Total optical depth `netflux` grid points:

!data tautots / 1.00e-99, 0.01000, 0.01292, 0.01668, 0.02154,  &  ! Undeflow
data tautots / 1.e-36_f4, 0.01000_f4, 0.01292_f4, 0.01668_f4, 0.02154_f4,  &
     0.02783_f4 ,  0.03594_f4, 0.04642_f4, 0.05995_f4, 0.07743_f4,  &
     0.10000_f4 ,  0.1292_f4 , 0.1668_f4 , 0.2154_f4 , 0.2783_f4 ,  &
     0.3594_f4  ,  0.4642_f4 , 0.5995_f4 , 0.7743_f4 , 1.0000_f4 ,  &
     1.292_f4   ,  1.668_f4  , 2.154_f4  , 2.783_f4  , 3.594_f4  ,  &
     4.642_f4   ,  5.995_f4  , 7.743_f4  , 10.00_f4  , 12.92_f4  ,  &
     16.68_f4   ,  21.54_f4  , 27.83_f4  , 35.94_f4  , 46.42_f4   /

! Cumulative optical depth `netflux` grid points:

data taucum0 / 0.0000_f4 , 0.0010_f4, 0.0035_f4, 0.0100_f4, 0.0200_f4,  &
     0.0400_f4  , 0.0800_f4, 0.1500_f4, 0.2500_f4, 0.3500_f4,  &
     0.4500_f4  , 0.5500_f4, 0.6500_f4, 0.7500_f4, 0.8500_f4,  &
     0.9500_f4  , 1.0000_f4 /

! Cos(Z) `netflux` grid points:

data u0vals / 0.01_f4   , 0.013335_f4, 0.017783_f4, 0.023714_f4, 0.031623_f4,  &
     0.04217_f4, 0.056234_f4, 0.074989_f4, 0.10000_f4 , 0.13335_f4 ,  &
     0.17783_f4, 0.23714_f4 , 0.31623_f4 , 0.42170_f4 , 0.56234_f4 ,  &
     0.74989_f4, 1.0_f4 /

! Albedo `netflux` grid points:

data albedos / 0.1_f4 , 0.4_f4 , 0.7_f4 /

end module radn_netflux_mem_mod
