!```````````````````````````````````````````````````````````````````
module radn_start_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine radn_start(ng, n1, n2, n3, ia, iz, ja, jz, cosZ, msdist, PRAMS_T,  &
     PRAMS_p, radheat_rate, rlong, PRAMS_sfcT, sw_g_flux,  &
     lw_15_dn_flux, lw_15_up_flux, lw_out_dn_flux, lw_out_up_flux, sw_d_flux,  &
     mrad, albedt, pmin, pmax, nazm, cosz_horizon, solar_azm, tazm_angle,  &
     tslope_angle, shadow_tog, slpasp_tog, TCx_obstruct, TCz_obstruct, zt,  &
     rtgt, tautot_ij, conrath_nu, emissivity, idx_mask_xy,  &
     ext_bottom_other_sflux_down, ext_bottom_SWflux_down, ext_topo_factor_sflux)

!+ Radiation driver

use univ_kind_defs_mod, only: f4
use univ_IO_mem_status_mod, only: for_allocation
use radn_main_mem_mod, only: rdlyr_m, idx01, fmcut, rdlev_t, rm1, rm2, rm3,  &
     idx02, idx03, rdlev_f, fmcnt
use radn_mem_mod, only: radn_mem_AD
use radn_init1_mod, only: radn_init1
use radn_init2_mod, only: radn_init2
use radn_calc_mod, only: radn_calc
use radn_pt_mod, only: column_init, pt_init

implicit none

!=== Argument declarations:

integer, intent(IN) :: ng,n1,n2,n3,nazm,shadow_tog,slpasp_tog,ia,iz,ja,jz
integer, dimension(n2,n3), intent(IN) :: idx_mask_xy
real(f4), dimension(n1,n2,n3), intent(IN) :: PRAMS_T, PRAMS_p
real(f4), dimension(n1,n2,n3), intent(INOUT) :: sw_g_flux, lw_15_dn_flux,  &
     lw_15_up_flux, lw_out_dn_flux, lw_out_up_flux, sw_d_flux
real(f4), dimension(n2,n3), intent(IN) :: cosZ,PRAMS_sfcT,rtgt,albedt  &
     ,solar_azm,tazm_angle,tslope_angle, emissivity
real(f4), dimension(n2,n3,nazm), intent(IN) :: cosz_horizon,TCx_obstruct,TCz_obstruct
real(f4), intent(IN) :: msdist,mrad,pmin,pmax
real(f4), dimension(n1), intent(IN) :: zt

real(f4), dimension(n1,n2,n3), intent(OUT) :: radheat_rate
real(f4), dimension(n2,n3), intent(OUT) :: rlong,  &
     ext_bottom_other_sflux_down, ext_bottom_SWflux_down, ext_topo_factor_sflux
real(f4), dimension(n2,n3), intent(INOUT) :: tautot_ij,conrath_nu

!== Local declarations:

real(f4) :: qfact,inidel
integer, save :: ntimes

data ntimes/0/

!=== Executable statements:

!*** Set some global dimensions:

rdlyr_m = n1-1
!rdlev_t = 4     ! Now filled in RAMSIN
!fmcut = 18   ! Now filled in RAMSIN   ! Fine mesh ends at PRAMS level fmcut+1
idx01 = 2*(rdlyr_m-fmcut)+rdlev_t

rm1 = n1              ! PRAMS nzp
rm2 = n2              ! PRAMS nxp
rm3 = n3              ! PRAMS nyp

! These should be done only once (initialization):

if(ntimes == 0) then
   write (*,'(/a)') 'Beginning radiation initialization ...'

!**** NOW done during main model initialization (for parallel reasons)
!   call radn_input      ! Perform any needed input

   inidel = 0.5_f4       ! Coarse to fine mesh initial offset (in gridpoint space)
   call column_init(n1,qfact,inidel)   ! Calculate column size

   idx02 = (n1-fmcut+1)*2+fmcnt  ! Total coarse mesh and fine mesh levels
   idx03 = n1
   rdlev_f = rdlev_t+(n1-fmcut+1)*2+fmcnt    ! Partial variable fine mesh
   call radn_mem_AD(for_allocation)        ! Allocate global column memory

   call pt_init(n1,qfact,inidel)  ! Set up column level values

   call radn_init1(pmin,pmax)   ! Calculate CO2 emissivity/absorption tables
   call radn_init2      ! Calculate dust optical depth table
   ntimes = 1

endif

!*** Calculate radiative transfer:

write (*,'(/a,i2,a)') 'Computing radiative fluxes on grid #',ng,' ...'

call radn_calc(rm1, rm2, rm3, ia, iz, ja, jz, cosZ, msdist, PRAMS_T, PRAMS_p,  &
     radheat_rate, rlong, PRAMS_sfcT, sw_g_flux, lw_15_dn_flux,  &
     lw_15_up_flux, lw_out_dn_flux, lw_out_up_flux, sw_d_flux, albedt, nazm,  &
     solar_azm, cosz_horizon, tazm_angle, tslope_angle, shadow_tog,  &
     slpasp_tog, TCx_obstruct, TCz_obstruct, zt, rtgt, tautot_ij, conrath_nu,  &
     emissivity, idx_mask_xy, ext_bottom_other_sflux_down,  &
     ext_bottom_SWflux_down, ext_topo_factor_sflux)

end subroutine radn_start


end module radn_start_mod
