!```````````````````````````````````````````````````````````````````
module PRAMS_turbdiff_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine tridiff1(m1,m2,m3,ia,iz,ja,jz,kz,cim1,ci,cip1,rhs,cj,cjp1)

!+ Solve diffusion matrix????

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,kz
real(f4), dimension(m1,m2,m3), intent(IN) :: cim1,ci,cip1

real(f4), dimension(m1,m2,m3), intent(INOUT) :: rhs,cj,cjp1

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

do j = ja,jz
   do i = ia,iz

      cjp1(2,i,j) = cip1(2,i,j) / ci(2,i,j)
      rhs(2,i,j) = rhs(2,i,j) / ci(2,i,j)

      do k = 3,kz
         cj(k,i,j) = ci(k,i,j) - cim1(k,i,j) * cjp1(k-1,i,j)
         cjp1(k,i,j) = cip1(k,i,j) / cj(k,i,j)
         rhs(k,i,j) = (rhs(k,i,j) - cim1(k,i,j) * rhs(k-1,i,j)) / cj(k,i,j)
      enddo

      cj(kz,i,j) = rhs(kz,i,j)

      do k = kz-1,2,-1
         cj(k,i,j) = rhs(k,i,j) - cjp1(k,i,j) * cj(k+1,i,j)
      enddo

   enddo
enddo

end subroutine tridiff1


!```````````````````````````````````````````````````````````````````
subroutine truhor(m1,m2,m3,ia,iz,ja,jz,vc3da,vc3db,dir,gpnt,dxy,topo,rtg  &
     ,z,vctr1,vctr2,vctr3,zintl,zintr,dxynu,dxytem,jd,dn0hkh,dn0,dtl,grid)

!+ Get Cartesian horizontal gradient

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use univ_math_constants_mod, only: ONE_f4
use univ_interpolation_mod, only: inextrp_1D_arr_type01a

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd,grid
real(f4), dimension(m1,m2,m3), intent(IN) :: vc3da,dn0hkh,dn0
real(f4), dimension(m2,m3), intent(IN) :: dxy,topo,rtg
real(f4), dimension(m1), intent(IN) :: z
character(len=*), intent(IN) :: dir,gpnt

real(f4), dimension(m1), intent(INOUT) :: zintl,zintr
real(f4), dimension(m1,m2,m3), intent(INOUT) :: vc3db
real(f4), dimension(m1), intent(INOUT) :: vctr1,vctr2,vctr3,dxynu,dxytem

!== Local declarations:

real(f4) :: delz1,delz3,distnu,distold,scale,aknu,akdiff,dtl
integer :: nz,jaa,jzz,nclip,je,jf,i,j,k

!=== Executable statements:

nz = m1

if(jd /= 0) then
   jaa = ja
   jzz = jz
else
   jaa = 1
   jzz = 1
endif

nclip = 0
if(dir == 'xdir') then
   if(gpnt == 'dyt' .or. gpnt == 'dxt') then
      je = 0
      jf = 1
   else
      je = -1
      jf = 0
   endif
   do j=jaa,jzz
      do i=ia,iz
         do k=1,m1
            vctr1(k) = topo(i-1,j)+z(k)*rtg(i-1,j)
            vctr2(k) = topo(i  ,j)+z(k)*rtg(i ,j)
            vctr3(k) = topo(i+1,j)+z(k)*rtg(i+1,j)
            dxytem(k) = vc3da(k,i-1,j)
            dxynu(k) = vc3da(k,i+1,j)
         enddo

         ! extrapolate the gradient from 3--->2 to 2--->1.
         delz1 = (vctr1(2)-vctr1(1))/(vctr1(3)-vctr1(2))
         delz3 = (vctr3(2)-vctr3(1))/(vctr3(3)-vctr3(2))
         dxytem(1) = dxytem(2)-(dxytem(3)-dxytem(2))*delz1
         dxynu(1) = dxynu(2) -(dxynu(3) -dxynu(2)) *delz3

         zintl(1:nz) = inextrp_1D_arr_type01a(dxytem(1:nz), vctr1(1:nz),  &
              vctr2(1:nz))
         zintr(1:nz) = inextrp_1D_arr_type01a(dxynu(1:nz), vctr3(1:nz),  &
              vctr2(1:nz))

         ! check intersection of cartesian sfc. on left side of mtn.
         call topobnd(m1,m2,m3,i,j, 1,0,'l',vc3da,dxy(i+je,j)  &
              ,z,vctr1,vctr2,vctr3,zintr,dxytem,jd)
        ! check intersection of cartesian sfc. on right side of mtn.
         call topobnd(m1,m2,m3,i,j,-1,0,'r',vc3da,dxy(i+jf,j)  &
              ,z,vctr1,vctr2,vctr3,zintl,dxynu,jd)
         do k=2,m1
            distnu = ONE_f4/dxytem(k)+ONE_f4/dxynu(k)
            distold = ONE_f4/dxy(i+je,j)+ONE_f4/dxy(i+jf,j)
            scale = distnu**2/distold**2
!                avgden=0.5*(dn0(k,i,j)+dn0(k,i+je+jf,j))
!cfast                if(scale.gt.1)stop' scale incorrect-xdir--'
!c            clp=min(dn0hkh(k,i,j)*scale,0.124/16.*avgden*distnu**2/dtl)
!                clp=min(dn0hkh(k,i,j),0.124*avgden*distold**2/dtl)
!cfast                if(dn0hkh(k,i,j).ne.clp)nclip=nclip+1
!               Need to multiply hkh by 2 since going over 2dx
            aknu = dn0hkh(k,i,j)*scale*2._f4
!                aknu=clp*scale
            akdiff = -aknu/distnu

            vc3db(k,i,j) = akdiff*( (zintr(k)-vc3da(k,i,j))*dxytem(k)  &
                 -(vc3da(k,i,j)-zintl(k))*dxynu(k) )

!                print*,'distnu,old,scale,tend=',k,i,j,distnu
!     &,                 distold,scale,vc3db(k,i,j)
!               if(i.eq.ip.and.j.eq.jp.and.k.le.3
!     &            .and.vc3da(k,i,j).gt.100.)
!     &        print*,'aknu,scal,lmr=',k,i,j,akdiff,scale,zintl(k)
!     &,                            vc3da(k,i,j),zintr(k),vc3db(k,i,j)
         enddo
!              vc3db(2,i,j)=0.
         vc3db(1,i,j) = vc3db(2,i,j)
         if(gpnt == 'wpnt') then
            vc3db(m1-1,i,j) = vc3db(m1-2,i,j)
         else
            vc3db(m1,i,j) = vc3db(m1-1,i,j)
         endif
      enddo
   enddo

elseif(dir == 'ydir') then
   if(gpnt == 'dyt' .or. gpnt == 'dxt') then
      je = 0
      jf = jd
   else
      je = -jd
      jf = 0
   endif
   do j=jaa,jzz
      do i=ia,iz
         do k=1,m1
            vctr1(k) = topo(i,j-jd)+z(k)*rtg(i,j-jd)
            vctr2(k) = topo(i,j)   +z(k)*rtg(i,j)
            vctr3(k) = topo(i,j+jd)+z(k)*rtg(i,j+jd)
            dxytem(k) = vc3da(k,i,j-jd)
            dxynu(k) = vc3da(k,i,j+jd)
         enddo

!             extrapolate the gradient from 3--->2 to 2--->1.
         delz1 = (vctr1(2)-vctr1(1))/(vctr1(3)-vctr1(2))
         delz3 = (vctr3(2)-vctr3(1))/(vctr3(3)-vctr3(2))
         dxytem(1) = dxytem(2)-(dxytem(3)-dxytem(2))*delz1
         dxynu(1) = dxynu(2) -(dxynu(3) -dxynu(2)) *delz3

         zintl(1:nz) = inextrp_1D_arr_type01a(dxytem(1:nz), vctr1(1:nz),  &
              vctr2(1:nz))
         zintr(1:nz) = inextrp_1D_arr_type01a(dxynu(1:nz), vctr3(1:nz),  &
              vctr2(1:nz))

!             check intersection of cartesian sfc. on left side of mtn.
         call topobnd(m1,m2,m3,i,j,0, 1,'l',vc3da,dxy(i,j+je)  &
              ,z,vctr1,vctr2,vctr3,zintr,dxytem,jd)
!             check intersection of cartesian sfc. on right side of mtn.
         call topobnd(m1,m2,m3,i,j,0,-1,'r',vc3da,dxy(i,j+jf)  &
              ,z,vctr1,vctr2,vctr3,zintl,dxynu,jd)
         do k=2,m1
            distnu = ONE_f4/dxytem(k)+ONE_f4/dxynu(k)
            distold = ONE_f4/dxy(i,j+je)+ONE_f4/dxy(i,j+jf)
            scale = distnu**2/distold**2
!cfast                if(scale.gt.1)stop' scale incorrect-ydir--'
!                avgden=0.5*(dn0(k,i,j)+dn0(k,i,j+je+jf))
!c            clp=min(dn0hkh(k,i,j)*scale,0.124/16.*avgden*distnu**2/dtl)
!                clp=min(dn0hkh(k,i,j),0.124*avgden*distold**2/dtl)
!cfast                if(dn0hkh(k,i,j).ne.clp)nclip=nclip+1

!               Need to multiply hkh by 2 since going over 2dx


            aknu = dn0hkh(k,i,j)*scale*2._f4
!                aknu=clp*scale*2.

            akdiff = -aknu/distnu
            vc3db(k,i,j) = akdiff*( (zintr(k)-vc3da(k,i,j))*dxytem(k)  &
                 -(vc3da(k,i,j)-zintl(k))*dxynu(k) )
         enddo
!              vc3db(2,i,j)=0.
         vc3db(1,i,j) = vc3db(2,i,j)
         if(gpnt == 'npnt') then
            vc3db(m1-1,i,j) = vc3db(m1-2,i,j)
         else
            vc3db(m1,i,j) = vc3db(m1-1,i,j)
         endif
      enddo
   enddo

endif
if(nclip >= 1 .and. gpnt == 'dxt') then
   print*,'clip on grd',grid,' dir=',dir(1:1),' ',gpnt(1:3),nclip
endif
!      endif

end subroutine truhor


!```````````````````````````````````````````````````````````````````
subroutine topobnd(m1,m2,m3,i,j,ipm1,jpm1,sid,vc3da,dxy,z,vctr1,vctr2,vctr3  &
     ,zint,dxynu,jd)

!+ Check intersection of Cartesian plane with mountain???

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,i,j,ipm1,jpm1,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: vc3da
real(f4), dimension(m1), intent(IN) :: z
character(len=*), intent(IN) :: sid
real(f4), intent(IN) :: dxy

real(f4), dimension(m1), intent(INOUT) :: vctr1,vctr2,zint,dxynu,vctr3

!== Local declarations:

integer :: ip,jp,ij,k
real(f4) :: tanth,pct,delz1,delz2,delz3,vc1,vc3,vc2

!=== Executable statements:

ip = 6
jp = 2

ij = ipm1+jpm1

!     keep track of the distance between scalar points.
dxynu(1:m1) = dxy

!                        |
!                    ----|----
!                   /    |    \
!                  /     |     \
!                 /      |      \
!                /       |       \
!               /        |        \
!              /         |         \
!             /          |          \
!            /           |           \
!      ------            |            -------
!     |      |      |    |   |     |     |
!     vctr1  |   vctr3   | vctr1   |     vctr3
!           vctr2        |       vctr2
!     t(i-1) t(i) t(i+1) | t(i-1) t(i)   t(i+1)
!
!     if the cartesian slice intersects topo.,
!       find the slope of the topography (tanth), then calculate the
!       horizontal cartesian distance to sigma=2 level (1/dxynu).
!       then interpolate along the sigma surface to find the value of
!       the variable at this intersection.

!     check left side of hill for rhs intersection.
if(sid == 'l') then

   do k=3,m1
      if(vctr3(2) <= vctr2(k)) then
         exit    ! exit do-loop
      endif
      tanth = (vctr3(2)-vctr2(2))*dxy
      dxynu(k) = tanth/(vctr2(k)-vctr2(2))
      pct = dxy/dxynu(k)
      zint(k) = vc3da(2,i+ipm1,j+jpm1)*pct+vc3da(2,i,j)*(ONE_f4-pct)
   enddo

   if(vctr3(1) <= vctr2(2)) then
      return
   endif

   tanth = (vctr3(1)-vctr2(1))*dxy
   dxynu(2) = tanth/(vctr2(2)-vctr2(1))
   pct = dxy/dxynu(2)
   
!       extrapolate the gradient from 3--->2 to 2--->1.
   delz1 = (vctr1(2)-vctr1(1))/(vctr1(3)-vctr1(2))
   delz2 = (vctr2(2)-vctr2(1))/(vctr2(3)-vctr2(2))
   delz3 = (vctr3(2)-vctr3(1))/(vctr3(3)-vctr3(2))
   vc1 = vc3da(2,i-ipm1,j-jpm1)-(vc3da(3,i-ipm1,j-jpm1)  &
        -vc3da(2,i-ipm1,j-jpm1))*delz1
   vc3 = vc3da(2,i+ipm1,j+jpm1)-(vc3da(3,i+ipm1,j+jpm1)  &
        -vc3da(2,i+ipm1,j+jpm1))*delz3
!-----------------------------------------------------------------------
!       use the first vc2 formulation for strictly hh runs, as this is
!         the 'correct' value to use as an endpoint when interpolating
!         along a sigma surface. the second formulation for vc2 is a
!         'fix' around the excessive gridpoint cooling caused by
!         extrapolating a locally cold gridpoint to k=1, and then using
!         this excessively cold value as an endpoint. this then forces
!         the gridpoint cooler at k=2 since the interpolation along the
!         k=1 sigma surface will results in a value colder than the k=2
!         gridpoint value. while a hh run with radiate and sfclyr
!         commented out, lsflg=3, nudlat=0, initial=1 and us,vs=0 will
!         not stay hh, it appears that simulations with the standard
!         rturb.f and this version produce very similar results. this
!         gives confidence to the solution when using the substantially
!         higher resolution allowed by this formulation.
!       note that this scheme does not conserve energy in its present
!         formulation since it diffuses according to a simple 1:2:1
!         scheme.
!       it also appears that while vertical resolution can be as high
!         as 20m with no winds, experiments with the real time
!         forecasting system indicate a resolution nearer 70m is needed
!         since problems with horizontal gradients in a sigma coordinate
!         system near steep topography exist in other routines as well.
!-----------------------------------------------------------------------
!        vc2=vc3da(2,i,j)-(vc3da(3,i,j)-vc3da(2,i,j))*delz2
   vc2 = 0.5_f4*(vc3+vc1)
   zint(2) = vc3*pct+vc2*(ONE_f4-pct)

!     check right side of hill for lhs intersection.
elseif(sid == 'r') then

   do k=3,m1
      if(vctr1(2) <= vctr2(k)) then
         exit  ! exit do-loop
      endif
      tanth = (vctr1(2)-vctr2(2))*dxy
      dxynu(k) = tanth/(vctr2(k)-vctr2(2))
      pct = dxy/dxynu(k)
      zint(k) = vc3da(2,i+ipm1,j+jpm1)*pct+vc3da(2,i,j)*(ONE_f4-pct)
   enddo
   
   if(vctr1(1) <= vctr2(2)) then
      return
   endif
   
   tanth = (vctr1(1)-vctr2(1))*dxy
   dxynu(2) = tanth/(vctr2(2)-vctr2(1))
   pct = dxy/dxynu(2)

!       extrapolate the gradient from 3--->2 to 2--->1.
   delz1 = (vctr1(2)-vctr1(1))/(vctr1(3)-vctr1(2))
   delz2 = (vctr2(2)-vctr2(1))/(vctr2(3)-vctr2(2))
   delz3 = (vctr3(2)-vctr3(1))/(vctr3(3)-vctr3(2))
   vc1 = vc3da(2,i+ipm1,j+jpm1)-(vc3da(3,i+ipm1,j+jpm1)  &
        -vc3da(2,i+ipm1,j+jpm1))*delz1
   vc3 = vc3da(2,i-ipm1,j-jpm1)-(vc3da(3,i-ipm1,j-jpm1)  &
        -vc3da(2,i-ipm1,j-jpm1))*delz3
!        vc2=vc3da(2,i,j)-(vc3da(3,i,j)-vc3da(2,i,j))*delz2
   vc2 = 0.5_f4*(vc3+vc1)
   zint(2) = vc1*pct+vc2*(ONE_f4-pct)

endif

end subroutine topobnd


!```````````````````````````````````````````````````````````````````
subroutine true_horiz_gradient(m1,m2,m3,ib,ie,jb,je,var,gradient,grad_dir  &
     ,var_spc,ds,comp_topo,grad_topo,comp_rtg,grad_rtg,z,jd,grid  &
     ,comp_wo_vertgrad,z_gradient)

!+ Calculate Cartesian (on OPS grid) horizontal gradient

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use univ_exception_mod, only: exception, FATAL
use main_memory_mod, only: az_pt1 => vctr1, az_pt2 => vctr2, az_tgt => vctr3  &
     ,openside_var => vctr4, slopside_var => vctr5, grad_invdist => vctr6  &
     ,var_wo_zgrad1 => vctr7, var_wo_zgrad2 => vctr8, c_gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ib,ie,jb,je,jd,grid
real(f4), dimension(m1,m2,m3), intent(IN) :: var  &  ! values to take gradient of 
                                      ,z_gradient  ! "real-world" var vert_grad
real(f4), dimension(m2,m3), intent(IN) :: ds  &  ! 1/"real-world"_distance over which
                                   ,    &  !  ideal gradient would be computed
                                    comp_topo  & ! topography of compute points
                                   ,grad_topo  &   ! topography at gradient pt
                                   ,comp_rtg  & ! vert-cmprsn of compute points
                                   ,grad_rtg      ! vert-cmprsn of gradient pt
real(f4), dimension(m1), intent(IN) :: z   ! Basis vertical grid heights
character(len=4), intent(IN) :: grad_dir  &   ! Direction of gradient?  'XDIR' or 'YDIR'
                   ,var_spc   ! On what type of grid point is gradient valid?
     ! If grad_dir='XDIR':
     !    'DXT' = gradient valid at 'T'-point (dU/dx)
     !    'DXU' = gradient valid at 'U'-point (dq/dx)
     !    'WDXU' = gradient valid at 'U'-point (vertically displaced) (dw/dx)
     !    'WDXT' = gradient valid at 'T'-point (vertically displaced)
     !    'DXM' = gradient valid between 'V'-points (dV/dx)
     ! If grad_dir='YDIR':
     !    'DYT' = gradient valid at 'T'-point (dV/dy)
     !    'DYV' = gradient valid at 'V'-point (dq/dy)
     !    'WDYV' = gradient valid at 'V'-point (vertically displaced) (dw/dy)
     !    'WDYT' = gradient valid at 'T'-point (vertically displaced) 
     !    'DYM' = gradient valid between 'U'-points (dU/dy)
logical, intent(IN) :: comp_wo_vertgrad

real(f4), dimension(m1,m2,m3), intent(INOUT) :: gradient    ! d(var)/d(?)

!== Local declarations:

real(f4) :: tgt_topo,tgt_rtg,dq,var_q2,var_q3,az_q2,az_q3,tan_thg
integer :: i,j,k,endpt_slope1,endpt_slope2,icoord_offset  &
     ,icoord_offset_1,jcoord_offset,jcoord_offset_1,iq_1,iq,jq_jd,jq

!=== Executable statements:

if(grad_dir(1:4) == 'XDIR') then   ! Compute gradient in X-direction

   if(var_spc(1:3) == 'DXT' .or. var_spc(2:4) == 'DXT') then
      icoord_offset_1 = 0
      icoord_offset = 0
   else
      icoord_offset_1 = 1
      icoord_offset = 1
   endif

   jcoord_offset_1 = jd
   jcoord_offset = 0

elseif(grad_dir(1:4) == 'YDIR') then   ! Compute gradient in Y-direction

   if(var_spc(1:3) == 'DYT' .or. var_spc(2:4) == 'DYT') then
      jcoord_offset_1 = 0
      jcoord_offset = 0
   else
      jcoord_offset_1 = jd
      jcoord_offset = jd
   endif

   icoord_offset_1 = 1
   icoord_offset = 0

else    ! Unsupported direction
  
   call exception('true_horizontal_gradient', FATAL,  &
        'Can ONLY calculate horizontal gradients.')

endif

do j=jb,je
   jq_jd = j-jd+jcoord_offset_1
   jq = j+jcoord_offset
   do i=ib,ie
      iq_1 = i-1+icoord_offset_1 
      iq = i+icoord_offset

      tgt_topo = grad_topo(i,j)
      tgt_rtg = grad_rtg(i,j)  
      do k=1,m1
         ! Bodypotential heights [m] of each compute point level:
         az_pt1(k) = comp_topo(iq_1,jq_jd)+z(k)*comp_rtg(iq_1,jq_jd)
         az_pt2(k) = comp_topo(iq,jq)+z(k)*comp_rtg(iq,jq)
         az_tgt(k) = tgt_topo+z(k)*tgt_rtg   ! az-levs at tgt (grad) point
      enddo

      ! Subtract out vertical gradient if desired:
      
      if(comp_wo_vertgrad) then     ! operation is desired

         if(var_spc(1:1) /= 'W') then
            var_wo_zgrad1(1:2) = var(1:2,iq_1,jq_jd)
            do k=2,m1-1
               var_wo_zgrad1(k+1) = var(k+1,iq_1,jq_jd)-  &
                    z_gradient(k,iq_1,jq_jd)*comp_rtg(iq_1,jq_jd)/c_gd%dzm(k)
            enddo
            var_wo_zgrad2(1:2) = var(1:2,iq,jq)
            do k=2,m1-1
               var_wo_zgrad2(k+1) = var(k+1,iq,jq)-z_gradient(k,iq,jq)*  &
                    comp_rtg(iq,jq)/c_gd%dzm(k)
            enddo
         else
!@            var_wo_zgrad1(:) = var(:,i-1,j)
!@            var_wo_zgrad2(:) = var(:,i,j)
            var_wo_zgrad1(1) = var(1,iq_1,jq_jd)
            do k=2,m1
               var_wo_zgrad1(k) = var(k,iq_1,jq_jd)-  &
                    z_gradient(k,iq_1,jq_jd)*comp_rtg(iq_1,jq_jd)/c_gd%dzt(k)
            enddo
            var_wo_zgrad2(1) = var(1,iq,jq)
            do k=2,m1
               var_wo_zgrad2(k) = var(k,iq,jq)-z_gradient(k,iq,jq)*  &
                    comp_rtg(iq,jq)/c_gd%dzt(k)
            enddo
         endif

      else
         
         var_wo_zgrad1(1:m1) = var(1:m1,iq_1,jq_jd)
         var_wo_zgrad2(1:m1) = var(1:m1,iq,jq)

      endif

      if(az_pt2(2) > az_tgt(2) .and.  &   ! Gradient pt is on "valley floor"
           az_pt1(2) > az_tgt(2)) then 

!^         print *,'valleybot',i,j,grid,grad_dir(1:4)

         ! Find bounds of vertical regions that will be treated specially:

         k = 2
         do while(az_tgt(k) < az_pt1(2))
            k = k+1
         enddo
         endpt_slope1 = k-1   ! Special treatment at or below this level  

         k = 2
         do while(az_tgt(k) < az_pt2(2))
            k = k+1
         enddo
         endpt_slope2 = k-1   ! Special treatment at or below this level   
         
         ! Interpolate to find "iq" values at tgt "normal" levels:

         call linear_ietp_xrng(m1,az_pt2,var_wo_zgrad2,m1,az_tgt  &
              ,openside_var,endpt_slope2+1,m1)
!#         call htint(m1,var_wo_zgrad2,az_pt2,m1,openside_var,az_tgt)

         ! Interpolate to find "iq_1" values at tgt "normal" levels:

         call linear_ietp_xrng(m1,az_pt1,var_wo_zgrad1,m1,az_tgt  &
              ,slopside_var,endpt_slope1+1,m1)            
!#         call htint(m1,var_wo_zgrad1,az_pt1,m1,slopside_var,az_tgt)

         ! "Normal" denominator:
         grad_invdist(max(endpt_slope1,endpt_slope2)+1:m1) = ds(i,j)

         ! (Possibly) treat some levels specially:

         grad_invdist(2:max(endpt_slope1,endpt_slope2)) = 1.e+12_f4  ! delta_s~0.

         if(endpt_slope2 > endpt_slope1) then
            do k=endpt_slope1+1,endpt_slope2
               grad_invdist(k) = 2._f4*ds(i,j)  ! 0.5*delta_s
            enddo
         elseif(endpt_slope1 > endpt_slope2) then
            do k=endpt_slope2+1,endpt_slope1
               grad_invdist(k) = 2._f4*ds(i,j)  ! 0.5*delta_s
            enddo
         endif

         ! "Right" side slope:
         
         tan_thg = (comp_topo(iq,jq)-grad_topo(i,j))*2._f4*ds(i,j)
         do k=2,endpt_slope2
            
            dq = (az_tgt(k)-grad_topo(i,j))/tan_thg
!            if(dq > 0.5/ds(i,j)) then
!               print *,'rrr',i,j,dq,0.5/ds(i,j),dq-0.5/ds(i,j)
!            endif
            dq = min(0.5_f4/ds(i,j),dq)   ! No larger than 0.5*delta_s

            var_q2 = (var_wo_zgrad2(2)-var_wo_zgrad1(2))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(2)+var_wo_zgrad2(2))
            var_q3 = (var_wo_zgrad2(3)-var_wo_zgrad1(3))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(3)+var_wo_zgrad2(3))
            az_q2 = (az_pt2(2)-az_tgt(2))*2._f4*ds(i,j)*dq+az_tgt(2)
            az_q3 = (az_pt2(3)-az_tgt(3))*2._f4*ds(i,j)*dq+az_tgt(3)
            
            openside_var(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_tgt(k))
            grad_invdist(k) = ONE_f4/(dq+ONE_f4/grad_invdist(k))

!print *,'rre',k,i,j,comp_topo(iq_1,j),grad_topo(i,j),comp_topo(iq,j)
!print *,'-1-',dq,1./ds(i,j),1./grad_invdist(k)
!print *,'lv2',var_wo_zgrad1(2),0.5*(var_wo_zgrad1(2)+var_wo_zgrad2(2))  &
!     ,var_q2,var_wo_zgrad2(2)
!print *,'lv3',var_wo_zgrad1(3),0.5*(var_wo_zgrad1(3)+var_wo_zgrad2(3))  &
!     ,var_q3,var_wo_zgrad2(3)
!print *,'fin',var_q3,var_q2,slopside_var(k),az_tgt(k)

         enddo
!pause

         ! "Left" side slope:

         tan_thg = (grad_topo(i,j)-comp_topo(iq_1,jq_jd))*2._f4*ds(i,j)
         do k=2,endpt_slope1

            dq = (az_tgt(k)-grad_topo(i,j))/tan_thg
!            if(dq < -0.5/ds(i,j)) then
!               print *,'rrr',i,j,dq,0.5/ds(i,j),abs(dq)-0.5/ds(i,j)
!            endif
            dq = max(-0.5_f4/ds(i,j),dq)   ! No less than -0.5*delta_s

            var_q2 = (var_wo_zgrad2(2)-var_wo_zgrad1(2))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(2)+var_wo_zgrad2(2))
            var_q3 = (var_wo_zgrad2(3)-var_wo_zgrad1(3))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(3)+var_wo_zgrad2(3))
            az_q2 = (az_tgt(2)-az_pt1(2))*2._f4*ds(i,j)*dq+az_tgt(2)
            az_q3 = (az_tgt(3)-az_pt1(3))*2._f4*ds(i,j)*dq+az_tgt(3)

            slopside_var(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_tgt(k))
            grad_invdist(k) = ONE_f4/(abs(dq)+ONE_f4/grad_invdist(k))

!print *,'rre',k,i,j,comp_topo(iq_1,j),grad_topo(i,j),comp_topo(iq,j)
!print *,'-1-',dq,1./ds(i,j),1./grad_invdist(k)
!print *,'lv2',var_wo_zgrad1(2),var_q2  &
!     ,0.5*(var_wo_zgrad1(2)+var_wo_zgrad2(2)),var_wo_zgrad2(2)
!print *,'lv3',var_wo_zgrad1(3),var_q3  &
!     ,0.5*(var_wo_zgrad1(3)+var_wo_zgrad2(3)),var_wo_zgrad2(3)
!print *,'fin',var_q3,var_q2,slopside_var(k),az_tgt(k)

         enddo
!pause

         do k=2,m1
            gradient(k,i,j) =  &
                 (openside_var(k)-slopside_var(k))*grad_invdist(k)
if(var_spc == 'WDYV' .and.  gradient(k,i,j) > 1.e6_f4) print *,'graddist',k,1./grad_invdist(k),endpt_slope1,endpt_slope2
         enddo

      elseif(az_pt2(2) > az_tgt(2)) then  ! Slope present on the "right" side

!^         print *,'slope_on_right',i,j,grid,grad_dir(1:4)

         ! Find bound of vertical region that will be treated specially:

         k = 2
         do while(az_tgt(k) < az_pt2(2))
            k = k+1
         enddo
         endpt_slope2 = k-1   ! Special treatment at or below this level    

         ! Interpolate to find "iq_1" values at tgt "normal" levels:
            
         call linear_ietp_xrng(m1,az_pt1,var_wo_zgrad1,m1,az_tgt  &
              ,openside_var,2,m1)
!#         call htint(m1,var_wo_zgrad1,az_pt1,m1,openside_var,az_tgt)
           
         ! Interpolate to find "iq" values at tgt "normal" levels:

         call linear_ietp_xrng(m1,az_pt2,var_wo_zgrad2,m1,az_tgt  &
              ,slopside_var,endpt_slope2+1,m1)
!#         call htint(m1,var_wo_zgrad2,az_pt2,m1,slopside_var,az_tgt)

         ! (Possibly) treat some levels specially:
         
         tan_thg = (comp_topo(iq,jq)-grad_topo(i,j))*2._f4*ds(i,j)

         do k=2,endpt_slope2
            
            dq = (az_tgt(k)-grad_topo(i,j))/tan_thg
!            if(dq > 0.5/ds(i,j)) then
!               print *,'rrr',i,j,dq,0.5/ds(i,j),dq-0.5/ds(i,j)
!            endif
            dq = min(0.5_f4/ds(i,j),dq)   ! No larger than 0.5*delta_s

            var_q2 = (var_wo_zgrad2(2)-var_wo_zgrad1(2))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(2)+var_wo_zgrad2(2))
            var_q3 = (var_wo_zgrad2(3)-var_wo_zgrad1(3))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(3)+var_wo_zgrad2(3))
            az_q2 = (az_pt2(2)-az_tgt(2))*2._f4*ds(i,j)*dq+az_tgt(2)
            az_q3 = (az_pt2(3)-az_tgt(3))*2._f4*ds(i,j)*dq+az_tgt(3)
               
            slopside_var(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_tgt(k))
            grad_invdist(k) = ONE_f4/(dq+0.5_f4/ds(i,j))

!print *,'rre',k,i,j,comp_topo(iq_1,j),grad_topo(i,j),comp_topo(iq,j)
!print *,'-1-',dq,1./ds(i,j),1./grad_invdist(k)
!print *,'lv2',var_wo_zgrad1(2),0.5*(var_wo_zgrad1(2)+var_wo_zgrad2(2))  &
!     ,var_q2,var_wo_zgrad2(2)
!print *,'lv3',var_wo_zgrad1(3),0.5*(var_wo_zgrad1(3)+var_wo_zgrad2(3))  &
!     ,var_q3,var_wo_zgrad2(3)
!print *,'fin',var_q3,var_q2,slopside_var(k),az_tgt(k)

         enddo
!pause       
         grad_invdist(endpt_slope2+1:m1) = ds(i,j)  ! "normal" denominator

!if(i == 27 .and. j == 65) then
!   k = 3
!   print *,'yyyR',grid,var_spc(1:4),i,j,1./grad_invdist(k),slopside_var(k)  &
!        ,openside_var(k),(slopside_var(k)-openside_var(k))*grad_invdist(k)  &
!        ,1./ds(i,j),iq_1,jq_jd,iq,jq
!endif
         
         do k=2,m1
            gradient(k,i,j) =  &
                 (slopside_var(k)-openside_var(k))*grad_invdist(k)
if(var_spc == 'WDYV' .and.  gradient(k,i,j) > 1.e6_f4) print *,'feefe',k,slopside_var(k),openside_var(k),grad_invdist(k)
         enddo

      elseif(az_pt1(2) > az_tgt(2)) then  ! Slope present on the "left" side

!^         print *,'slope_on_left',i,j,grid,grad_dir(1:4)

         ! Find bound of vertical region that will be treated specially:

         k = 2
         do while(az_tgt(k) < az_pt1(2))
            k = k+1
         enddo
         endpt_slope1 = k-1   ! Special treatment at or below this level 

         ! Interpolate to find "iq" values at tgt "normal" levels:

         call linear_ietp_xrng(m1,az_pt2,var_wo_zgrad2,m1,az_tgt  &
              ,openside_var,2,m1)
!#         call htint(m1,var_wo_zgrad2,az_pt2,m1,openside_var,az_tgt)  
            
         ! Interpolate to find "iq_1" values at tgt "normal" levels:
            
         call linear_ietp_xrng(m1,az_pt1,var_wo_zgrad1,m1,az_tgt  &
              ,slopside_var,endpt_slope1+1,m1)
!#         call htint(m1,var_wo_zgrad1,az_pt1,m1,slopside_var,az_tgt)

         ! (Possibly) treat some levels specially:

         tan_thg = (grad_topo(i,j)-comp_topo(iq_1,jq_jd))*2._f4*ds(i,j)
         
         do k=2,endpt_slope1

            dq = (az_tgt(k)-grad_topo(i,j))/tan_thg
!            if(dq < -0.5/ds(i,j)) then
!               print *,'rrr',i,j,dq,0.5/ds(i,j),abs(dq)-0.5/ds(i,j)
!            endif
            dq = max(-0.5_f4/ds(i,j),dq)   ! No less than -0.5*delta_s

            var_q2 = (var_wo_zgrad2(2)-var_wo_zgrad1(2))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(2)+var_wo_zgrad2(2))
            var_q3 = (var_wo_zgrad2(3)-var_wo_zgrad1(3))*ds(i,j)*dq+  &
                 0.5_f4*(var_wo_zgrad1(3)+var_wo_zgrad2(3))
            az_q2 = (az_tgt(2)-az_pt1(2))*2._f4*ds(i,j)*dq+az_tgt(2)
            az_q3 = (az_tgt(3)-az_pt1(3))*2._f4*ds(i,j)*dq+az_tgt(3)

            slopside_var(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_tgt(k))
            grad_invdist(k) = ONE_f4/(abs(dq)+0.5_f4/ds(i,j))

!print *,'rre',k,i,j,comp_topo(iq_1,j),grad_topo(i,j),comp_topo(iq,j)
!print *,'-1-',dq,1./ds(i,j),1./grad_invdist(k)
!print *,'lv2',var_wo_zgrad1(2),var_q2  &
!     ,0.5*(var_wo_zgrad1(2)+var_wo_zgrad2(2)),var_wo_zgrad2(2)
!print *,'lv3',var_wo_zgrad1(3),var_q3  &
!     ,0.5*(var_wo_zgrad1(3)+var_wo_zgrad2(3)),var_wo_zgrad2(3)
!print *,'fin',var_q3,var_q2,slopside_var(k),az_tgt(k)

      enddo

!pause
                   
         grad_invdist(endpt_slope1+1:m1) = ds(i,j)  ! "normal" denominator

!if(i == 27 .and. j == 65) then
!   k = 3
!   print *,'yyyL',grid,var_spc(1:4),i,j,1./grad_invdist(k),openside_var(k)  &
!        ,slopside_var(k),(openside_var(k)-slopside_var(k))*grad_invdist(k)  &
!        ,1./ds(i,j),iq_1,jq_jd,iq,jq
!endif
            
         do k=2,m1
            gradient(k,i,j) =  &
                 (openside_var(k)-slopside_var(k))*grad_invdist(k)
if(var_spc == 'WDYV' .and.  gradient(k,i,j) > 1.e6_f4) print *,'qqqq',k,slopside_var(k),openside_var(k),grad_invdist(k)
         enddo
            
      elseif(az_pt2(2) < az_tgt(2) .and.  &   ! Gradient pt is on "ridgetop"
           az_pt1(2) < az_tgt(2)) then

!^         print *,'ridgetop',i,j,grid,grad_dir(1:4)

         ! Interpolate to find "iq" values at tgt levels:

         call linear_ietp_xrng(m1,az_pt2,var_wo_zgrad2,m1,az_tgt  &
              ,openside_var,2,m1)
!#         call htint(m1,var_wo_zgrad2,az_pt2,m1,openside_var,az_tgt)  
            
         ! Interpolate to find "iq_1" values at tgt levels:

         call linear_ietp_xrng(m1,az_pt1,var_wo_zgrad1,m1,az_tgt  &
              ,slopside_var,2,m1)
!#         call htint(m1,var_wo_zgrad1,az_pt1,m1,slopside_var,az_tgt)

         do k=2,m1
            gradient(k,i,j) =  &   ! Delta-s in this case same for all levels
                 (openside_var(k)-slopside_var(k))*ds(i,j)
if(var_spc == 'WDYV' .and.  gradient(k,i,j) > 1.e6_f4) print *,'wwww',k,slopside_var(k),openside_var(k),grad_invdist(k)
         enddo

      else     ! NO topographic slope

!^         print *,'none',i,j,grid,grad_dir(1:4)
            
         do k=2,m1
            gradient(k,i,j) = (var(k,iq,jq)-var(k,iq_1,jq_jd))*ds(i,j)
if(var_spc == 'WDYV' .and.  gradient(k,i,j) > 1.e6_f4) print *,'gggg',k,slopside_var(k),openside_var(k),grad_invdist(k)
         enddo

      endif

      ! Impose top/bottom boundary conditions on the gradient:
         
      gradient(1,i,j) = gradient(2,i,j)       ! Bottom BC

      if(var_spc(1:1) == 'W') then                !
         gradient(m1-1,i,j) = gradient(m1-2,i,j)  ! Top BC
      else                                        !
         gradient(m1,i,j) = gradient(m1-1,i,j)    !
      endif                                       !

   enddo
enddo

! Impose lateral boundary conditions on the gradient:

if (jb > 1) then
   do j=1,jb-jd
      do i=ib,ie
         gradient(:,i,j) = gradient(:,i,jb)
      end do
   end do
end if
if (je < m3) then
   do j=je+jd,m3
      do i=ib,ie
         gradient(:,i,j) = gradient(:,i,je)
      end do
   end do
end if
if (ib > 1) then
   do j=1,m3
      do i=1,ib
         gradient(:,i,j) = gradient(:,ib,j)
      end do
   end do
end if
if (ie < m2) then
   do j=1,m3
      do i=ie+1,m2
         gradient(:,i,j) = gradient(:,ie,j)
      end do
   end do
end if

end subroutine true_horiz_gradient


!```````````````````````````````````````````````````````````````````
subroutine linear_ietp_xrng(dim_A,x_A,y_A,dim_B,x_B,y_B,ix_beg,ix_end)

!+ Performs *linear* interpolation/extrapolation on "y" values [for given
!   arrays of "old" and "new" "x" values (with corresponding "new" array index
!   range)]

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: dim_A,dim_B,ix_beg,ix_end
real(f4), dimension(dim_A), intent(IN) :: y_A,x_A
real(f4), dimension(dim_B), intent(IN) :: x_B

real(f4), dimension(dim_B), intent(INOUT) :: y_B

!== Local declarations:

integer :: ix,iq
real(f4) :: xweight

!=== Executable statements:

! Assumes that 'x_A' and 'x_B' each have values in ascending order
!  (small -> large):

iq = 1
do ix=ix_beg,ix_end   ! Loop over all points to (inter/extra)-polate *to*

   do while(iq < dim_A)   ! usage ERROR if iq >= dim_A
      
      if(x_B(ix) >= x_A(iq) .and. x_B(ix) <= x_A(iq+1)) then   ! Interpolation
         xweight = (x_B(ix)-x_A(iq))/(x_A(iq+1)-x_A(iq))
         y_B(ix) = y_A(iq)+(y_A(iq+1)-y_A(iq))*xweight        
         exit
      elseif(x_B(ix) < x_A(1)) then    ! "Downward" extrapolation
         xweight = (x_B(ix)-x_A(iq))/(x_A(iq+1)-x_A(iq))
         y_B(ix) = y_A(iq)+(y_A(iq+1)-y_A(iq))*xweight        
         exit
      elseif(x_B(ix) > x_A(dim_A)) then   ! "Upward" extrapolation
         xweight = (x_B(ix)-x_A(dim_A))/(x_A(dim_A-1)-x_A(dim_A))
         y_B(ix) = y_A(dim_A)+(y_A(dim_A-1)-y_A(dim_A))*xweight
         exit
      endif
         
      iq = iq+1
   enddo

enddo

end subroutine linear_ietp_xrng


!```````````````````````````````````````````````````````````````````
subroutine avg2velpt(m1,m2,m3,ia,iz,ja,jz,avg_dir,jd,var_origpt,var_avgpt)

!+ Average field (valid at T-pts) in the given direction (to U, V, W, or
!   "M" points; e.g., avg in 'XDIR' to obtain field values at U-pts)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: var_origpt
character(len=4), intent(IN) :: avg_dir   ! 'XDIR','YDIR','ZDIR'

real(f4), dimension(m1,m2,m3), intent(INOUT) :: var_avgpt

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

if(avg_dir(1:1) == 'X') then   ! Average in X-direction

   do j=1,m3
      do i=1,m2-1
         do k=1,m1
            var_avgpt(k,i,j) = 0.5_f4*(var_origpt(k,i,j)+var_origpt(k,i+1,j))
         enddo
      enddo
   enddo

   i = m2
   do j=1,m3
      do k=1,m1
         var_avgpt(k,i,j) = 2._f4*var_origpt(k,i,j)-var_avgpt(k,i-1,j)
      enddo
   enddo

elseif(avg_dir(1:1) == 'Y') then   ! Average in Y-direction

   if(jd /= 0) then

      do j=1,m3-jd
         do i=1,m2
            do k=1,m1
               var_avgpt(k,i,j) = 0.5_f4*(var_origpt(k,i,j)+var_origpt(k,i,j+jd))
            enddo
         enddo
      enddo

      j = m3
      do i=1,m2
         do k=1,m1
            var_avgpt(k,i,j) = 2._f4*var_origpt(k,i,j)-var_avgpt(k,i,j-jd)
         enddo
      enddo

   else

      do j=1,m3
         do i=1,m2
            do k=1,m1
               var_avgpt(k,i,j) = var_origpt(k,i,j)
            enddo
         enddo
      enddo

   endif

elseif(avg_dir(1:1) == 'Z') then   ! Average in Z-direction

   do j=1,m3
      do i=1,m2
         do k=1,m1-1
            var_avgpt(k,i,j) = 0.5_f4*(var_origpt(k,i,j)+var_origpt(k+1,i,j))
         enddo
      enddo
   enddo

   k = m1
   do j=1,m3
      do i=1,m2
         var_avgpt(k,i,j) = 2._f4*var_origpt(k,i,j)-var_avgpt(k-1,i,j)
      enddo
   enddo

endif

end subroutine avg2velpt


!```````````````````````````````````````````````````````````````````
subroutine true_horiz_difftf(m1,m2,m3,ib,ie,jb,je,var,result,grad_dir  &
     ,resvar_spc,resvar_ds,comp_ds,resvar_topo,resvar_rtg,z,jd,grid,K_fact)

!+ Calculate d/dx*[dn0*K*(dphi/dx)] -- proportional to Cartesian (on OPS grid)
!   horizontal second derivative of variable.  Result is divided by -dn0
!   elsewhere to produce a diffusive tendency.

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use univ_exception_mod, only: exception, FATAL
use main_memory_mod, only: az_resvar1 => vctr1, az_resvar2 => vctr2  &
     ,az_resvar3 => vctr3, var_rvpt1_raw => vctr4, var_rvpt2_raw => vctr5  &
     ,var_rvpt3_raw => vctr6, var_rvpt1_itp => vctr7  &
     ,var_rvpt3_itp => vctr8, grad_dist1 => vctr9, grad_dist3 => vctr10  &
     ,Kf1_raw => vctr11, Kf2_raw => vctr12, Kf3_raw => vctr13  &
     ,Kf1_itp => vctr14, Kf3_itp => vctr15
 
implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ib,ie,jb,je,grid,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: var  &   ! values to take gradient of 
                                      ,K_fact  ! K_coeff*optional factor (must
                                               !  be valid at result/var pts)
  ! 1/"real-world"_distance over which ideal gradient would be computed:
real(f4), dimension(m2,m3), intent(IN) :: resvar_ds  &  ! across result/var points
                                   ,comp_ds    ! across comp points

real(f4), dimension(m2,m3), intent(IN) :: resvar_topo  &  ! topo at result/var points
                                   ,resvar_rtg  ! res/var pt vert_compress_fact

real(f4), dimension(m1), intent(IN) :: z   ! Basis vertical grid heights
character(len=1), intent(IN) :: grad_dir     ! Direction of gradient(s)?  'X' or 'Y'

character(len=3), intent(IN) :: resvar_spc   ! What type of gridpoint are the result/var
                                 !  values valid?
     ! If grad_dir='XDIR', resvar_spc =
     !    'XT' = if result/var points valid at 'T'-points
     !    'XU' = if result/var points valid at 'U'-points
     !    'WTX' = if result/var points valid at 'T'-points (vert displaced)
     !    'XV' = if result/var points valid at 'V'-points
     ! If grad_dir='YDIR', resvar_spc =
     !    'YT' = if result/var points valid at 'T'-points
     !    'YV' = if result/var points valid at 'V'-points
     !    'WTY' = if result/var points valid at 'T'-points (vert displaced)
     !    'YU' = if result/var points valid at 'U'-points

real(f4), dimension(m1,m2,m3), intent(INOUT) :: result

!== Local declarations:

real(f4) :: tan_thg,dq,var_q2,var_q3,az_q2,az_q3,gradA,gradB,fB,fA  &
     ,gdrat,Kf_q2,Kf_q3,gdratK
integer :: i,j,k,endpt_slope1,endpt_slope3,icomp_offset  &
     ,icomp_offset_1,jcomp_offset,jcomp_offset_1,iq_1,iq1,jq_jd,jqjd  &
     ,cgjq_jd,cgjq,cgiq,cgiq_1,irvpt_offset_1,irvpt_offset1,jrvpt_offset_1  &
     ,jrvpt_offset1

!=== Executable statements:

if(grad_dir == 'X') then   ! Compute gradient(s) in X-direction

   if(resvar_spc(2:2) == 'T') then
      icomp_offset_1 = 0
      icomp_offset = 0
      irvpt_offset_1 = 0
      irvpt_offset1 = 0
   else
      icomp_offset_1 = 1
      icomp_offset = 1
      irvpt_offset_1 = 0
      irvpt_offset1 = 0
   endif

   jcomp_offset_1 = jd
   jcomp_offset = 0
   jrvpt_offset_1 = jd
   jrvpt_offset1 = -jd

elseif(grad_dir == 'Y') then   ! Compute gradient(s) in Y-direction

   if(resvar_spc(2:2) == 'T') then
      jcomp_offset_1 = 0
      jcomp_offset = 0
      jrvpt_offset_1 = 0
      jrvpt_offset1 = 0
   else
      jcomp_offset_1 = jd
      jcomp_offset = jd
      jrvpt_offset_1 = 0
      jrvpt_offset1 = 0
   endif

   icomp_offset_1 = 1
   icomp_offset = 0
   irvpt_offset_1 = 1
   irvpt_offset1 = -1

else    ! Unsupported direction
  
   call exception('true_horiz_difftf', FATAL,  &
        'Can ONLY calculate horizontal gradients.')

endif

do j=jb,je
   cgjq_jd = j-jd+jcomp_offset_1  ! Indices for gradients at comp points
   cgjq = j+jcomp_offset          ! 
   jq_jd = j-jd+jrvpt_offset_1
   jqjd = j+jd+jrvpt_offset1

   do i=ib,ie
      cgiq_1 = i-1+icomp_offset_1   ! Indices for gradients at comp points
      cgiq = i+icomp_offset         !
      iq_1 = i-1+irvpt_offset_1 
      iq1 = i+1+irvpt_offset1

      do k=1,m1
         ! Bodypotential heights [m] of each result/var point level:
         az_resvar1(k) = resvar_topo(iq_1,jq_jd)+z(k)*resvar_rtg(iq_1,jq_jd)
         az_resvar2(k) = resvar_topo(i,j)+z(k)*resvar_rtg(i,j)
         az_resvar3(k) = resvar_topo(iq1,jqjd)+z(k)*resvar_rtg(iq1,jqjd)
      enddo

      ! Load data arrays at result/var points:

      var_rvpt1_raw(1:m1) = var(1:m1,iq_1,jq_jd)
      var_rvpt2_raw(1:m1) = var(1:m1,i,j)    
      var_rvpt3_raw(1:m1) = var(1:m1,iq1,jqjd)
      Kf1_raw(1:m1) = K_fact(1:m1,iq_1,jq_jd)
      Kf2_raw(1:m1) = K_fact(1:m1,i,j)
      Kf3_raw(1:m1) = K_fact(1:m1,iq1,jqjd)

      if(az_resvar3(2) > az_resvar2(2) .and.  &   ! Point is on "valley floor"
           az_resvar1(2) > az_resvar2(2)) then    

!^         print *,'valleybot',i,j,grid,grad_dir,resvar_spc(1:3)

         ! Find bounds of vertical regions that will be treated specially:

           ! "Left" side
         k = 2
         do while(az_resvar2(k) < az_resvar1(2))
            k = k+1
         enddo
         endpt_slope1 = k-1

           ! "Right" side
         k = 2
         do while(az_resvar2(k) < az_resvar3(2))
            k = k+1
         enddo
         endpt_slope3 = k-1
        
        ! Interpolate to find "iq1" values at tgt "normal" levels:

         call linear_ietp_xrng(m1,az_resvar3,var_rvpt3_raw,m1,az_resvar2  &
              ,var_rvpt3_itp,endpt_slope3+1,m1)

         call linear_ietp_xrng(m1,az_resvar3,Kf3_raw,m1,az_resvar2  &
              ,Kf3_itp,endpt_slope3+1,m1)  ! K factor

         ! Interpolate to find "iq_1" values at tgt "normal" levels:
         
         call linear_ietp_xrng(m1,az_resvar1,var_rvpt1_raw,m1,az_resvar2  &
              ,var_rvpt1_itp,endpt_slope1+1,m1)            

         call linear_ietp_xrng(m1,az_resvar1,Kf1_raw,m1,az_resvar2  &
              ,Kf1_itp,endpt_slope1+1,m1)  ! K factor

         ! "Normal" denominators (delta_s):
         grad_dist1(endpt_slope1+1:m1) = ONE_f4/comp_ds(cgiq_1,cgjq_jd)
         grad_dist3(endpt_slope3+1:m1) = ONE_f4/comp_ds(cgiq,cgjq)

         ! "Right" side (intersects ground):

         tan_thg = (resvar_topo(iq1,jqjd)-resvar_topo(i,j))*comp_ds(cgiq,cgjq)
         do k=2,endpt_slope3
               
            dq = (az_resvar2(k)-resvar_topo(i,j))/tan_thg
!            if(dq > 1./comp_ds(cgiq,cgjq)) then
!               print *,'rrr',i,j,dq,1./comp_ds(cgiq,cgjq)  &
!                    ,dq-1./comp_ds(cgiq,cgjq)
!            endif
            dq = min(1./comp_ds(cgiq,cgjq),dq)   ! No larger than delta_s
            
            var_q2 = (var_rvpt3_raw(2)-var_rvpt2_raw(2))*  &
                 comp_ds(cgiq,cgjq)*dq+var_rvpt2_raw(2)
            var_q3 = (var_rvpt3_raw(3)-var_rvpt2_raw(3))*  &
                 comp_ds(cgiq,cgjq)*dq+var_rvpt2_raw(3)
            az_q2 = (az_resvar3(2)-az_resvar2(2))*comp_ds(cgiq,cgjq)*dq  &
                 +az_resvar2(2)
            az_q3 = (az_resvar3(3)-az_resvar2(3))*comp_ds(cgiq,cgjq)*dq  &
                 +az_resvar2(3)
           
            var_rvpt3_itp(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            Kf_q2 = (Kf3_raw(2)-Kf2_raw(2))*comp_ds(cgiq,cgjq)*dq+Kf2_raw(2)
            Kf_q3 = (Kf3_raw(3)-Kf2_raw(3))*comp_ds(cgiq,cgjq)*dq+Kf2_raw(3)

            Kf3_itp(k) = Kf_q2-((Kf_q3-Kf_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            grad_dist3(k) = dq

!print *,'rre',k,i,j,resvar_topo(iq_1,jq_jd),resvar_topo(i,j),resvar_topo(iq1,jqjd)
!print *,'-1-',dq,1./comp_ds(cgiq,cgjq),grad_dist3(k)
!print *,'lv2',var_rvpt1_raw(2),var_rvpt2_raw(2),var_q2,var_rvpt3_raw(2)
!print *,'lv3',var_rvpt1_raw(3),var_rvpt2_raw(3),var_q3,var_rvpt3_raw(3)
!print *,'fin',var_q3,var_q2,var_rvpt3_itp(k),az_resvar2(k)
            
         enddo
!pause
         ! "Left" side (intersects ground):

         tan_thg = (resvar_topo(i,j)-resvar_topo(iq_1,jq_jd))*  &
              comp_ds(cgiq_1,cgjq_jd)
!print *,endpt_slope1
         do k=2,endpt_slope1
               
            dq = (az_resvar2(k)-resvar_topo(i,j))/tan_thg
!            if(dq < -1./comp_ds(cgiq_1,cgjq_jd)) then
!               print *,'rrr',i,j,dq,1./comp_ds(cgiq_1,cgjq_jd)  &
!                    ,dq-1./comp_ds(cgiq_1,cgjq_jd)
!            endif
            dq = max(-ONE_f4/comp_ds(cgiq_1,cgjq_jd),dq)  ! No less than -delta_s
            
            var_q2 = (var_rvpt2_raw(2)-var_rvpt1_raw(2))*  &
                 comp_ds(cgiq_1,cgjq_jd)*dq+var_rvpt2_raw(2)
            var_q3 = (var_rvpt2_raw(3)-var_rvpt1_raw(3))*  &
                 comp_ds(cgiq_1,cgjq_jd)*dq+var_rvpt2_raw(3)
            az_q2 = (az_resvar2(2)-az_resvar1(2))*comp_ds(cgiq_1,cgjq_jd)*dq  &
                 +az_resvar2(2)
            az_q3 = (az_resvar2(3)-az_resvar1(3))*comp_ds(cgiq_1,cgjq_jd)*dq  &
                 +az_resvar2(3)
           
            var_rvpt1_itp(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            Kf_q2 = (Kf2_raw(2)-Kf1_raw(2))*comp_ds(cgiq_1,cgjq_jd)*dq+  &
                 Kf2_raw(2) 
            Kf_q3 = (Kf2_raw(3)-Kf1_raw(3))*comp_ds(cgiq_1,cgjq_jd)*dq+  &
                 Kf2_raw(3) 

            Kf1_itp(k) = Kf_q2-((Kf_q3-Kf_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            grad_dist1(k) = abs(dq)

!if(k == 8 .and. i == 5 .and. j == 2) then
!print *,'rre',k,i,j,resvar_topo(iq_1,jq_jd),resvar_topo(i,j),resvar_topo(iq1,jqjd)
!print *,'-1-',dq,1./comp_ds(cgiq_1,cgjq_jd),grad_dist1(k)
!print *,'lv2',var_rvpt1_raw(2),var_q2,var_rvpt2_raw(2),var_rvpt3_raw(2)
!print *,'lv3',var_rvpt1_raw(3),var_q3,var_rvpt2_raw(3),var_rvpt3_raw(3)
!print *,'fin',var_q3,var_q2,var_rvpt1_itp(k),az_resvar2(k)            
!endif
         enddo
!pause
      elseif(az_resvar3(2) > az_resvar2(2)) then  ! Slope on the "right" side

!^         print *,'slope_on_right',i,j,grid,grad_dir,resvar_spc(1:3)

         ! Find bound of vertical region that will be treated specially:

         k = 2
         do while(az_resvar2(k) < az_resvar3(2))
            k = k+1
         enddo
         endpt_slope3 = k-1   ! Special treatment at or below this level    

         ! Interpolate to find "iq_1" values at tgt "normal" levels:
            
         call linear_ietp_xrng(m1,az_resvar1,var_rvpt1_raw,m1,az_resvar2  &
              ,var_rvpt1_itp,2,m1)
           
         call linear_ietp_xrng(m1,az_resvar1,Kf1_raw,m1,az_resvar2  &
              ,Kf1_itp,2,m1)  ! K factor

         ! Interpolate to find "iq1" values at tgt "normal" levels:

         call linear_ietp_xrng(m1,az_resvar3,var_rvpt3_raw,m1,az_resvar2  &
              ,var_rvpt3_itp,endpt_slope3+1,m1)

         call linear_ietp_xrng(m1,az_resvar3,Kf3_raw,m1,az_resvar2  &
              ,Kf3_itp,endpt_slope3+1,m1)  ! K factor

         ! "Normal" denominators (delta_s):
         grad_dist1(2:m1) = ONE_f4/comp_ds(cgiq_1,cgjq_jd)
         grad_dist3(endpt_slope3+1:m1) = ONE_f4/comp_ds(cgiq,cgjq)

         ! (Possibly) treat some levels specially:
         
         tan_thg = (resvar_topo(iq1,jqjd)-resvar_topo(i,j))*comp_ds(cgiq,cgjq)
         do k=2,endpt_slope3
            
            dq = (az_resvar2(k)-resvar_topo(i,j))/tan_thg
!            if(dq > 1./comp_ds(cgiq,cgjq)) then
!               print *,'rrr',i,j,dq,1./comp_ds(cgiq,cgjq)  &
!                    ,dq-1./comp_ds(cgiq,cgjq)
!            endif
            dq = min(ONE_f4/comp_ds(cgiq,cgjq),dq)   ! No larger than delta_s

            var_q2 = (var_rvpt3_raw(2)-var_rvpt2_raw(2))*  &
                 comp_ds(cgiq,cgjq)*dq+var_rvpt2_raw(2)
            var_q3 = (var_rvpt3_raw(3)-var_rvpt2_raw(3))*  &
                 comp_ds(cgiq,cgjq)*dq+var_rvpt2_raw(3)
            az_q2 = (az_resvar3(2)-az_resvar2(2))*comp_ds(cgiq,cgjq)*dq  &
                 +az_resvar2(2)
            az_q3 = (az_resvar3(3)-az_resvar2(3))*comp_ds(cgiq,cgjq)*dq  &
                 +az_resvar2(3)
           
            var_rvpt3_itp(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            Kf_q2 = (Kf3_raw(2)-Kf2_raw(2))*comp_ds(cgiq,cgjq)*dq+Kf2_raw(2)
            Kf_q3 = (Kf3_raw(3)-Kf2_raw(3))*comp_ds(cgiq,cgjq)*dq+Kf2_raw(3)

            Kf3_itp(k) = Kf_q2-((Kf_q3-Kf_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            grad_dist3(k) = dq

!print *,'rre',k,i,j,resvar_topo(iq_1,jq_jd),resvar_topo(i,j),resvar_topo(iq1,jqjd)
!print *,'-1-',dq,1./comp_ds(cgiq,cgjq),grad_dist3(k)
!print *,'lv2',var_rvpt1_raw(2),var_rvpt2_raw(2),var_q2,var_rvpt3_raw(2)
!print *,'lv3',var_rvpt1_raw(3),var_rvpt2_raw(3),var_q3,var_rvpt3_raw(3)
!print *,'fin',var_q3,var_q2,var_rvpt3_itp(k),az_resvar2(k)

         enddo
!pause
      elseif(az_resvar1(2) > az_resvar2(2)) then  ! Slope on the "left" side

!^         print *,'slope_on_left',i,j,grid,grad_dir,resvar_spc(1:3)

         ! Find bound of vertical region that will be treated specially:

         k = 2
         do while(az_resvar2(k) < az_resvar1(2))
            k = k+1
         enddo
         endpt_slope1 = k-1   ! Special treatment at or below this level 

         ! Interpolate to find "iq1" values at tgt "normal" levels:

         call linear_ietp_xrng(m1,az_resvar3,var_rvpt3_raw,m1,az_resvar2  &
              ,var_rvpt3_itp,2,m1)

         call linear_ietp_xrng(m1,az_resvar3,Kf3_raw,m1,az_resvar2  &
              ,Kf3_itp,2,m1)
            
         ! Interpolate to find "iq_1" values at tgt "normal" levels:
            
         call linear_ietp_xrng(m1,az_resvar1,var_rvpt1_raw,m1,az_resvar2  &
              ,var_rvpt1_itp,endpt_slope1+1,m1)
         call linear_ietp_xrng(m1,az_resvar1,Kf1_raw,m1,az_resvar2  &
              ,Kf1_itp,endpt_slope1+1,m1)

         ! "Normal" denominators (delta_s):
         grad_dist3(2:m1) = ONE_f4/comp_ds(cgiq,cgjq)
         grad_dist1(endpt_slope1+1:m1) = ONE_f4/comp_ds(cgiq_1,cgjq_jd)

         ! (Possibly) treat some levels specially:

         tan_thg = (resvar_topo(i,j)-resvar_topo(iq_1,jq_jd))*  &
              comp_ds(cgiq_1,cgjq_jd)
         do k=2,endpt_slope1

            dq = (az_resvar2(k)-resvar_topo(i,j))/tan_thg
!            if(dq < -1./comp_ds(cgiq_1,cgjq_jd)) then
!               print *,'rrr',i,j,dq,1./comp_ds(cgiq_1,cgjq_jd)  &
!                    ,dq-1./comp_ds(cgiq_1,cgjq_jd)
!            endif
            dq = max(-ONE_f4/comp_ds(cgiq_1,cgjq_jd),dq)  ! No less than -delta_s
            
            var_q2 = (var_rvpt2_raw(2)-var_rvpt1_raw(2))*  &
                 comp_ds(cgiq_1,cgjq_jd)*dq+var_rvpt2_raw(2)
            var_q3 = (var_rvpt2_raw(3)-var_rvpt1_raw(3))*  &
                 comp_ds(cgiq_1,cgjq_jd)*dq+var_rvpt2_raw(3)
            az_q2 = (az_resvar2(2)-az_resvar1(2))*comp_ds(cgiq_1,cgjq_jd)*dq  &
                 +az_resvar2(2)
            az_q3 = (az_resvar2(3)-az_resvar1(3))*comp_ds(cgiq_1,cgjq_jd)*dq  &
                 +az_resvar2(3)
           
            var_rvpt1_itp(k) = var_q2-((var_q3-var_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            Kf_q2 = (Kf2_raw(2)-Kf1_raw(2))*comp_ds(cgiq_1,cgjq_jd)*dq+  &
                 Kf2_raw(2) 
            Kf_q3 = (Kf2_raw(3)-Kf1_raw(3))*comp_ds(cgiq_1,cgjq_jd)*dq+  &
                 Kf2_raw(3) 

            Kf1_itp(k) = Kf_q2-((Kf_q3-Kf_q2)/(az_q3-az_q2))*  &
                 (az_q2-az_resvar2(k))

            grad_dist1(k) = abs(dq)

!print *,'rre',k,i,j,resvar_topo(iq_1,jq_jd),resvar_topo(i,j),resvar_topo(iq1,jqjd)
!print *,'-1-',dq,1./comp_ds(cgiq_1,cgjq_jd),grad_dist1(k)
!print *,'lv2',var_rvpt1_raw(2),var_q2,var_rvpt2_raw(2),var_rvpt3_raw(2)
!print *,'lv3',var_rvpt1_raw(3),var_q3,var_rvpt2_raw(3),var_rvpt3_raw(3)
!print *,'fin',var_q3,var_q2,var_rvpt1_itp(k),az_resvar2(k)

         enddo
!pause

      elseif(az_resvar3(2) < az_resvar2(2) .and.  &   ! Point is on "ridgetop"
           az_resvar1(2) < az_resvar2(2)) then

!^         print *,'ridgetop',i,j,grid,grad_dir,resvar_spc(1:3)

         ! Interpolate to find "iq1" values at tgt levels:

         call linear_ietp_xrng(m1,az_resvar3,var_rvpt3_raw,m1,az_resvar2  &
              ,var_rvpt3_itp,2,m1)

         call linear_ietp_xrng(m1,az_resvar3,Kf3_raw,m1,az_resvar2  &
              ,Kf3_itp,2,m1)
            
         ! Interpolate to find "iq_1" values at tgt levels:

         call linear_ietp_xrng(m1,az_resvar1,var_rvpt1_raw,m1,az_resvar2  &
              ,var_rvpt1_itp,2,m1)

         call linear_ietp_xrng(m1,az_resvar1,Kf1_raw,m1,az_resvar2  &
              ,Kf1_itp,2,m1)

         grad_dist1(2:m1) = ONE_f4/comp_ds(cgiq_1,cgjq_jd)
         grad_dist3(2:m1) = ONE_f4/comp_ds(cgiq,cgjq)

      else     ! NO topographic slope

!^         print *,'none',i,j,grid,grad_dir,resvar_spc(1:3)
           
         var_rvpt1_itp(2:m1) = var_rvpt1_raw(2:m1)
         var_rvpt3_itp(2:m1) = var_rvpt3_raw(2:m1)

         Kf1_itp(2:m1) = Kf1_raw(2:m1)
         Kf3_itp(2:m1) = Kf3_raw(2:m1)

         grad_dist1(2:m1) = ONE_f4/comp_ds(cgiq_1,cgjq_jd)
         grad_dist3(2:m1) = ONE_f4/comp_ds(cgiq,cgjq)

      endif

      ! Calculate final result:

      do k=2,m1

         gradA = (var_rvpt2_raw(k)-var_rvpt1_itp(k))/grad_dist1(k)
         gdratK = (grad_dist1(k)**2)*(comp_ds(cgiq_1,cgjq_jd)**2)   ! K scaling
         fA = -0.5_f4*(Kf1_itp(k)+Kf2_raw(k))*gdratK*gradA  ! K avgd to grad midpt

         gradB = (var_rvpt3_itp(k)-var_rvpt2_raw(k))/grad_dist3(k)
         gdratK = (grad_dist3(k)**2)*(comp_ds(cgiq,cgjq)**2)   ! K scaling
         fB = -0.5_f4*(Kf3_itp(k)+Kf2_raw(k))*gdratK*gradB  ! K avgd to grad midpt

         gdrat = min(ONE_f4,0.5_f4*(grad_dist1(k)*comp_ds(cgiq_1,cgjq_jd)+  &
              grad_dist3(k)*comp_ds(cgiq,cgjq)))

         result(k,i,j) = (fB-fA)*resvar_ds(i,j)/gdrat

!         if(k<=3 .and. grid == 5 .and. i == 20 .and. j == 20 .and.  &
!                 var_rvpt2_raw(k) > 100.) then
!print *,'b',cgiq_1,cgjq_jd
!print *,'-',cgiq,cgjq
!print *,'-',iq_1,jq_jd 
!print *,'-',iq1,jqjd
!print *,'d',i,j
!            print *,'var',k,var_rvpt1_itp(k),var_rvpt2_raw(k),var_rvpt3_itp(k)
!            print *,'dss',grad_dist1(k),1./resvar_ds(i,j),grad_dist3(k)
!            print *,'gd1',gradA,gradB
!            print *,'dfn',result(k,i,j)
!            pause
!         endif

      enddo

      ! Impose boundary conditions on result:
         
      result(1,i,j) = result(2,i,j)       ! Bottom BC

      if(resvar_spc(1:1) == 'W') then             !
         result(m1-1,i,j) = result(m1-2,i,j)      ! Top BC
      else                                        !
         result(m1,i,j) = result(m1-1,i,j)        !
      endif                                       !

   enddo
enddo

end subroutine true_horiz_difftf


!```````````````````````````````````````````````````````````````````
subroutine avgvel(m1,m2,m3,ia,iz,ja,jz,dir,jd,a,b)

!+ Average velocities???

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd
real(f4), dimension(m1,m2,m3), intent(IN) :: b
character(len=*), intent(IN) :: dir

real(f4), dimension(m1,m2,m3), intent(INOUT) :: a

!== Local declarations:

integer :: jaa,jzz,i,j,k

!=== Executable statements:

if(jd /= 0) then
   jaa = ja
   jzz = jz
else
   jaa = 1
   jzz = 1
endif

if(dir == 'xdir') then
   do j=jaa,jzz
      do i=ia,iz
         do k=1,m1
            a(k,i,j) = 0.5_f4*(b(k,i,j)+b(k,i+1,j))
         enddo
      enddo
   enddo
elseif(dir == 'ydir') then
   do j=jaa,jzz
      do i=ia,iz
         do k=1,m1
            a(k,i,j) = 0.5_f4*(b(k,i,j)+b(k,i,j+jd))
         enddo
      enddo
   enddo
endif

end subroutine avgvel


end module PRAMS_turbdiff_mod


!```````````````````````````````````````````````````````````````````
subroutine diffvel(m1,m2,m3,ia,iz,ja,jz,jd,ia_1,ja_1,ia1,ja1,iz_1,jz_1  &
     ,iz1,jz1,izu,jzv,idiffkk,up,vp,wp,ut,vt,wt,vt3da,vt3db,vt3dc,vt3dd  &
     ,vt3de,vt3df,vt3dg,vt3dj,vt3dk,vt3dl,vt3dm,vt3dn,vt3do,rtgu,rtgv,rtgt  &
     ,uw,vw,wfz,dn0,dn0u,dn0v,scr1,scr2,ibcon,mynum)

!+ Diffuse velocities

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use main_memory_mod, only: ihorgrad, vctr1, vctr2, dtlv, vctr3, vctr4,  &
     i_grid, a, vctr5, vctr6, vctr7, c_gd
use vtables_mod, only: idxt,itopu,irtgu,idym,idyt,itopv,irtgv,idxm,itopt,irtgt  &
     ,idyv,idxu,idyu,idyv,idxv
use PRAMS_turbdiff_mod, only: true_horiz_difftf, tridiff1, avg2velpt, avgvel,  &
     truhor
use PRAMS_graddiv_mod, only: divcart

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd,ia_1,ja_1,ia1,ja1,iz_1,jz_1,iz1  &
     ,jz1,izu,jzv,idiffkk,ibcon,mynum
real(f4), dimension(m1,m2,m3), intent(IN) :: up,vp,wp,scr1,scr2  &
     ,dn0,dn0u,dn0v
real(f4), dimension(m2,m3), intent(IN) :: uw,vw,wfz,rtgt,rtgu,rtgv

real(f4), dimension(m1,m2,m3), intent(INOUT) :: vt3da,vt3db,vt3dc  &
     ,vt3dd,vt3de,vt3df,vt3dg,vt3dj,vt3dk,vt3dl,vt3dm,vt3dn,vt3do,ut,vt,wt

!== Local declarations:

logical, parameter :: use_Weissbluth_truhor = .FALSE.

real(f4) :: akn,ako,akp,cross,c1,c2,dtlvi
integer :: i,j,k

!=== Executable statements:

! Compute fluxes with k's and previous strains:

if(ihorgrad == 1) then
   
   do j = 1,jz1
      do i = 1,iz1
         do k = 1,m1-1
            vt3da(k,i,j) = -vt3da(k,i,j) * scr2(k,i,j)
            vt3dc(k,i,j) = -vt3dc(k,i,j) * scr2(k,i,j)
         enddo
      enddo
   enddo

   do j = 1,jz
      do i = 1,iz
         do k = 1,m1-1
            akn = 0.25_f4 * (scr2(k,i  ,j ) + scr2(k+1,i  ,j )  &
                 + scr2(k,i  ,j+jd) + scr2(k+1,i  ,j+jd))
            ako = 0.25_f4 * (scr2(k,i  ,j ) + scr2(k+1,i  ,j )  &
                 + scr2(k,i+1,j ) + scr2(k+1,i+1,j ))
            akp = 0.25_f4 * (scr2(k,i  ,j ) + scr2(k  ,i+1,j )  &
                 + scr2(k,i  ,j+jd) + scr2(k  ,i+1,j+jd))

            vt3db(k,i,j) = -vt3db(k,i,j) * akp
            vt3dn(k,i,j) = -vt3dn(k,i,j) * akp
            vt3dd(k,i,j) = -vt3dd(k,i,j) * ako
            vt3df(k,i,j) = -vt3df(k,i,j) * ako
            vt3de(k,i,j) = -vt3de(k,i,j) * akn
            vt3dg(k,i,j) = -vt3dg(k,i,j) * akn
         enddo
      enddo
   enddo
   
elseif(ihorgrad == 2) then

   do j = 1,jz
      do i = 1,iz
         do k = 1,m1-1
            akn = 0.25_f4 * (scr2(k,i  ,j ) + scr2(k+1,i  ,j )  &
                 + scr2(k,i  ,j+jd) + scr2(k+1,i  ,j+jd))
            ako = 0.25_f4 * (scr2(k,i  ,j ) + scr2(k+1,i  ,j )  &
                 + scr2(k,i+1,j ) + scr2(k+1,i+1,j ))

            vt3df(k,i,j) = -vt3df(k,i,j) * ako
            vt3dg(k,i,j) = -vt3dg(k,i,j) * akn
         enddo
      enddo
   enddo

endif

cross = 0.
if (idiffkk == 3 .or. idiffkk == 4 .or. idiffkk == 5) then
   cross = ONE_f4
end if

! Vertical u-component diffusion:

do j = ja,jz
   do i = ia,izu
!print *,'here3'
      do k = 1,m1-1
         vctr1(k) = cross * vt3df(k,i,j)
         vctr2(k) = c_gd%dzm(k) * (scr1(k,i,j) + scr1(k+1,i,j)  &
              + scr1(k,i+1,j) + scr1(k+1,i+1,j))
      enddo

!print *,'here4'
      if(c_gd%nstbot == 1) then
         vctr1(1) = 0.25_f4 * (uw(i,j)* (dn0(1,i,j) + dn0(2,i,j))  &
              + uw(i+1,j)* (dn0(1,i+1,j) + dn0(2,i+1,j)))
      else
         vctr1(1) = vctr1(1) + up(1,i,j)  &
              * 0.25_f4 * (scr1(1,i,j) + scr1(2,i,j) + scr1(1,i+1,j)  &
              + scr1(2,i+1,j)) * c_gd%dzm(1) / rtgu(i,j)
      endif

!print *,'here5'      
      if(c_gd%nsttop == 1) then
         vctr1(m1-1) = ZERO_f4
      else
         vctr1(m1-1) = vctr1(m1-1) - up(m1-1,i,j)  &
              * 0.25_f4 * (scr1(m1,i,j) + scr1(m1-1,i,j)  &
              + scr1(m1,i+1,j) + scr1(m1-1,i+1,j)) * c_gd%dzm(m1-1) / rtgu(i,j)
      endif

!print *,'here6'
      c1 = dtlv / rtgu(i,j)
      c2 = 0.25_f4 * dtlv / (rtgu(i,j) * rtgu(i,j))
      do k = 2,m1-1
         vt3dl(k,i,j) = up(k,i,j) * dn0u(k,i,j)  &
              + c1 * c_gd%dzt(k) * (vctr1(k-1) - vctr1(k))
         vt3dj(k,i,j) = - c2 * c_gd%dzt(k) * vctr2(k-1)
         vt3dk(k,i,j) = - c2 * c_gd%dzt(k) * vctr2(k)
         vt3do(k,i,j) = dn0u(k,i,j) - vt3dj(k,i,j) - vt3dk(k,i,j)
      enddo
      vt3dj(2,i,j) = ZERO_f4
      if(c_gd%nstbot == 1) vt3do(2,i,j) = dn0u(2,i,j) - vt3dk(2,i,j)
      vt3dk(m1-1,i,j) = ZERO_f4
      if (c_gd%nsttop == 1) vt3do(m1-1,i,j) = dn0u(m1-1,i,j) - vt3dj(m1-1,i,j)

!print *,'here7'
   enddo
enddo

call tridiff1(m1,m2,m3,ia,izu,ja,jz,m1-1,vt3dj,vt3do,vt3dk,vt3dl,vt3do,vt3dk)
!print *,'here8'

! Horizontal u-component diffusion:

if(ihorgrad == 1) then

   call divcart(m1,m2,m3,ia,izu,ja,jz,vt3da,vt3dj,'XDIR','TPNT')
   call divcart(m1,m2,m3,ia,izu,ja,jz,vt3dn,vt3dk,'YDIR','PPNT')

elseif(ihorgrad == 2) then
   
   if(use_Weissbluth_truhor) then

      ! average the dn0*hkh to the velocity points:
      call avgvel(m1,m2,m3,ia,izu,ja,jz,'xdir',jd,vt3da,scr2)
      call truhor(m1,m2,m3,ia,izu,ja,jz,up,vt3dj,'xdir','dxt',a(idxt)  &
           ,a(itopu),a(irtgu),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5,vctr6,vctr7  &
           ,jd,vt3da,dn0,dtlv,i_grid)

      call avgvel(m1,m2,m3,ia,izu,ja,jz,'ydir',jd,vt3da,scr2)
      call truhor(m1,m2,m3,ia,izu,ja,jz,up,vt3dk,'ydir','dym',a(idym)  &
           ,a(itopu),a(irtgu),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5,vctr6,vctr7  &
           ,jd,vt3da,dn0,dtlv,i_grid)

   else   ! Use TIM's scheme
!print *,'here9'
      ! Average scr2 (hkm*dn0; on T-pts) to U-pts {vt3da}:
      call avg2velpt(m1,m2,m3,ia,izu,ja,jz,'XDIR',jd,scr2,vt3da)

      ! Calculate Cartesian horizontal diffusion for U in x-dir; one step:
      !  (1) Calculate d/dx[dn0*Km*(dU/dx)] {vt3dj}
      call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,up,vt3dj,'X','XU '  &
           ,a(idxu),a(idxt),a(itopu),a(irtgu),c_gd%zt,jd,i_grid,vt3da)

      ! Calculate Cartesian horizontal diffusion for U in y-dir; one step:
      !  (1) Calculate d/dy[dn0*Km*(dU/dy)] {vt3dk}
      call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,up,vt3dk,'Y','YU '  &
           ,a(idyu),a(idym),a(itopu),a(irtgu),c_gd%zt,jd,i_grid,vt3da)
!print *,'here10'
   endif

endif

dtlvi = ONE_f4 / dtlv

!print *,'here11'
if(jd == 1) then
   do j = ja,jz
      do i = ia,izu
         do k = 2,m1-1
            ut(k,i,j) = ut(k,i,j) + dtlvi * (vt3do(k,i,j) - up(k,i,j))  &
                 - (vt3dj(k,i,j) + vt3dk(k,i,j)) / dn0u(k,i,j)
         enddo
      enddo
   enddo
else
   do j = ja,jz
      do i = ia,izu
         do k = 2,m1-1
            ut(k,i,j) = ut(k,i,j) + dtlvi * (vt3do(k,i,j)-up(k,i,j))  &
                 - vt3dj(k,i,j) / dn0u(k,i,j)
         enddo
      enddo
   enddo
endif

! Vertical v-component diffusion:

!print *,'here12'
if(jd /= 0) then

   do j = ja,jzv
      do i = ia,iz
         do k = 1,m1-1
            vctr1(k) = cross * vt3dg(k,i,j)
            vctr2(k) = c_gd%dzm(k) * (scr1(k,i,j) + scr1(k+1,i,j)  &
                 + scr1(k,i,j+jd) + scr1(k+1,i,j+jd))
         enddo
         
         if(c_gd%nstbot == 1) then
            vctr1(1) = 0.25_f4 * (vw(i,j)  &
                 * (dn0(1,i,j) + dn0(2,i,j))+ vw(i,j+jd)  &
                 * (dn0(1,i,j+jd) + dn0(2,i,j+jd)))
         else
            vctr1(1) = vctr1(1) + vp(1,i,j)  &
                 * 0.25_f4 * (scr1(1,i,j) + scr1(2,i,j) + scr1(1,i,j+jd)  &
                 + scr1(2,i,j+jd)) * c_gd%dzm(1) / rtgv(i,j)
         endif
         if(c_gd%nsttop == 1) then
            vctr1(m1-1) = ZERO_f4
         else
            vctr1(m1-1) = vctr1(m1-1) - vp(m1-1,i,j)  &
                 * 0.25_f4 * (scr1(m1,i,j) + scr1(m1-1,i,j)  &
                 + scr1(m1,i,j+jd) + scr1(m1-1,i,j+jd)) * c_gd%dzm(m1-1)  &
                 / rtgv(i,j)
         endif

         c1 = dtlv / rtgv(i,j)
         c2 = 0.25_f4 * dtlv / (rtgv(i,j) * rtgv(i,j))
         do k = 2,m1-1
            vt3dm(k,i,j) = vp(k,i,j) * dn0v(k,i,j)  &
                 + c1 * c_gd%dzt(k) * (vctr1(k-1) - vctr1(k))
            vt3dj(k,i,j) = -c2 * c_gd%dzt(k) * vctr2(k-1)
            vt3dk(k,i,j) = -c2 * c_gd%dzt(k) * vctr2(k)
            vt3do(k,i,j) = dn0v(k,i,j) - vt3dj(k,i,j) - vt3dk(k,i,j)
         enddo
         vt3dj(2,i,j) = ZERO_f4
         if(c_gd%nstbot == 1) then
            vt3do(2,i,j) = dn0v(2,i,j) - vt3dk(2,i,j)
         endif
         
         vt3dk(m1-1,i,j) = ZERO_f4
         if(c_gd%nsttop == 1)  then
            vt3do(m1-1,i,j) = dn0v(m1-1,i,j) - vt3dj(m1-1,i,j)
         endif
      enddo
   enddo
   
   call tridiff1(m1,m2,m3,ia,iz,ja,jzv,m1-1,vt3dj,vt3do,vt3dk,vt3dm,vt3do  &
        ,vt3dk)

! Horizontal v-component diffusion:
   
   if(ihorgrad == 1) then

      call divcart(m1,m2,m3,ia,iz,ja,jzv,vt3dc,vt3dj,'YDIR','TPNT')
      call divcart(m1,m2,m3,ia,iz,ja,jzv,vt3db,vt3dk,'XDIR','PPNT')
      
   elseif(ihorgrad == 2) then

      if(use_Weissbluth_truhor) then

         ! average the dn0*hkh to the velocity points
         call avgvel(m1,m2,m3,ia,iz,ja,jzv,'ydir',jd,vt3da,scr2)
         call truhor(m1,m2,m3,ia,iz,ja,jzv ,vp,vt3dj,'ydir','dyt',a(idyt)  &
              ,a(itopv),a(irtgv),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5  &
              ,vctr6,vctr7,jd,vt3da,dn0,dtlv,i_grid)
         
         call avgvel(m1,m2,m3,ia,iz,ja,jzv,'xdir',jd,vt3da,scr2)
         call truhor(m1,m2,m3,ia,iz,ja,jzv,vp,vt3dk,'xdir','dxm',a(idxm)  &
              ,a(itopv),a(irtgv),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5  &
              ,vctr6,vctr7,jd,vt3da,dn0,dtlv,i_grid)

      else   ! Use TIM's scheme

         ! Average scr2 (hkm*dn0; on T-pts) to V-pts {vt3da}:
         call avg2velpt(m1,m2,m3,ia,iz,ja,jzv,'YDIR',jd,scr2,vt3da)

         ! Calculate Cartesian horizontal diffusion for V in x-dir; one step:
         !  (1) Calculate d/dx[dn0*Km*(dV/dx)] {vt3dj}
         call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,vp,vt3dj,'X','XV '  &
              ,a(idxv),a(idxm),a(itopv),a(irtgv),c_gd%zt,jd,i_grid,vt3da)

         ! Calculate Cartesian horizontal diffusion for V in y-dir; one step:
         !  (1) Calculate d/dy[dn0*Km*(dV/dy)] {vt3dk}
         call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,vp,vt3dk,'Y','YV '  &
              ,a(idyv),a(idyt),a(itopv),a(irtgv),c_gd%zt,jd,i_grid,vt3da)

      endif

   endif

   do j = ja,jzv
      do i = ia,iz
         do k = 2,m1-1
            vt(k,i,j) = vt(k,i,j) + dtlvi * (vt3do(k,i,j)-vp(k,i,j))  &
                 - (vt3dj(k,i,j) + vt3dk(k,i,j)) / dn0v(k,i,j)
         enddo
      enddo
   enddo
   
endif
!print *,'here13'

! Vertical w-component diffusion:

do j = ja,jz
   do i = ia,iz
      do k = 2,m1-1
         vctr1(k) = ZERO_f4
         vctr2(k) = c_gd%dzt(k) * scr1(k,i,j)
      enddo

      if(c_gd%nstbot == 1) then
         vctr1(2) = wfz(i,j) * dn0(2,i,j)
      else
         vctr1(2) = (ONE_f4 + cross) * wp(1,i,j)  &
              * scr1(2,i,j) * c_gd%dzt(2) / rtgt(i,j)
      endif
      if(c_gd%nsttop == 1) then
         vctr1(m1-1) = ZERO_f4
      else
         vctr1(m1-1) = -(ONE_f4 + cross) * wp(m1-1,i,j)  &
              * scr1(m1-1,i,j) * c_gd%dzt(m1-1) / rtgt(i,j)
      endif

      c1 = dtlv / rtgt(i,j)
      c2 = 2._f4 * dtlv / (rtgt(i,j) * rtgt(i,j))
      do k = 2,m1-2
         vt3dn(k,i,j) = wp(k,i,j)  &
              * (dn0(k,i,j) + dn0(k+1,i,j)) * 0.5_f4  &
              + c1 * c_gd%dzm(k) * (vctr1(k) - vctr1(k+1))
         vt3dj(k,i,j) = - c2 * c_gd%dzm(k) * vctr2(k)
         vt3dk(k,i,j) = - c2 * c_gd%dzm(k) * vctr2(k+1)
         vt3do(k,i,j) = (dn0(k,i,j) + dn0(k+1,i,j)) * 0.5_f4  &
              - vt3dj(k,i,j) - vt3dk(k,i,j)
      enddo
      vt3dj(2,i,j) = ZERO_f4
      if(c_gd%nstbot == 1) then
         vt3do(2,i,j) = (dn0(2,i,j) + dn0(3,i,j)) * 0.5_f4 - vt3dk(2,i,j)
      endif
      vt3dk(m1-2,i,j) = ZERO_f4
      if(c_gd%nsttop == 1) then
         vt3do(m1-2,i,j) = (dn0(m1-2,i,j) + dn0(m1-1,i,j)) * 0.5_f4  &
              -vt3dj(m1-2,i,j)
      endif
   enddo
enddo

call tridiff1(m1,m2,m3,ia,iz,ja,jz,m1-2,vt3dj,vt3do,vt3dk,vt3dn,vt3do,vt3dk)

!print *,'here14'

! Horizontal w-component diffusion:

if (idiffkk == 3 .or. idiffkk == 4 .or. idiffkk == 5) then

   if(ihorgrad == 1) then

      call divcart(m1,m2,m3,ia,iz,ja,jz,vt3dd,vt3dj,'XDIR','OPNT')
      call divcart(m1,m2,m3,ia,iz,ja,jz,vt3de,vt3dk,'YDIR','NPNT')

   elseif(ihorgrad == 2) then

      if(use_Weissbluth_truhor) then

         call truhor(m1,m2,m3,ia,iz,ja,jz,wp,vt3dj,'xdir','dxu',a(idxu)  &
              ,a(itopt),a(irtgt),c_gd%zm,vctr1,vctr2,vctr3,vctr4,vctr5  &
              ,vctr6,vctr7,jd,scr2,dn0,dtlv,i_grid)

         call truhor(m1,m2,m3,ia,iz,ja,jz,wp,vt3dk,'ydir','dyv',a(idyv)  &
              ,a(itopt),a(irtgt),c_gd%zm,vctr1,vctr2,vctr3,vctr4,vctr5  &
              ,vctr6,vctr7,jd,scr2,dn0,dtlv,i_grid)

      else   ! Use TIM's scheme

         ! Average scr2 (hkm*dn0; on T-pts) to W-pts {vt3da}:
         call avg2velpt(m1,m2,m3,ia,iz,ja,jz,'ZDIR',jd,scr2,vt3da)

         ! Calculate Cartesian horizontal diffusion for W in x-dir; one step:
         !  (1) Calculate d/dx[dn0*Km*(dW/dx)] {vt3dj}
         call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,wp,vt3dj,'X','WXT'  &
              ,a(idxt),a(idxu),a(itopt),a(irtgt),c_gd%zm,jd,i_grid,vt3da)

         ! Calculate Cartesian horizontal diffusion for W in y-dir; one step:
         !  (1) Calculate d/dy[dn0*Km*(dW/dy)] {vt3dk}
         call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,wp,vt3dk,'Y','WYT'  &
              ,a(idyt),a(idyv),a(itopt),a(irtgt),c_gd%zm,jd,i_grid,vt3da)

      endif

   endif

else

   if(ihorgrad == 1) then

      call divcart(m1,m2,m3,ia,iz,ja,jz,vt3df,vt3dj,'XDIR','OPNT')
      call divcart(m1,m2,m3,ia,iz,ja,jz,vt3dg,vt3dk,'YDIR','NPNT')

   elseif(ihorgrad == 2) then

      if(use_Weissbluth_truhor) then

         call truhor(m1,m2,m3,ia,iz,ja,jz,wp,vt3dj,'xdir','dxu',a(idxu)  &
              ,a(itopt),a(irtgt),c_gd%zm,vctr1,vctr2,vctr3,vctr4,vctr5  &
              ,vctr6,vctr7,jd,scr2,dn0,dtlv,i_grid)
         
         call truhor(m1,m2,m3,ia,iz,ja,jz,wp,vt3dk,'ydir','dyv',a(idyv)  &
              ,a(itopt),a(irtgt),c_gd%zm,vctr1,vctr2,vctr3,vctr4,vctr5,vctr6  &
              ,vctr7,jd,scr2,dn0,dtlv,i_grid)

      else   ! Use TIM's scheme

         ! Average scr2 (hkm*dn0; on T-pts) to W-pts {vt3da}:
         call avg2velpt(m1,m2,m3,ia,iz,ja,jz,'ZDIR',jd,scr2,vt3da)

         ! Calculate Cartesian horizontal diffusion for W in x-dir; one step:
         !  (1) Calculate d/dx[dn0*Km*(dW/dx)] {vt3dj}
         call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,wp,vt3dj,'X','WXT'  &
              ,a(idxt),a(idxu),a(itopt),a(irtgt),c_gd%zm,jd,i_grid,vt3da)
         
         ! Calculate Cartesian horizontal diffusion for W in y-dir; one step:
         !  (1) Calculate d/dy[dn0*Km*(dW/dy)] {vt3dk}
         call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,wp,vt3dk,'Y','WYT'  &
              ,a(idyt),a(idyv),a(itopt),a(irtgt),c_gd%zm,jd,i_grid,vt3da)
      
      endif

   endif

endif

!print *,'here15'

if(jd == 1) then
   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-2
            wt(k,i,j) = wt(k,i,j)  &
                 + dtlvi * (vt3do(k,i,j) - wp(k,i,j))  &
                 - (vt3dj(k,i,j) + vt3dk(k,i,j))  &
                 / ((dn0(k,i,j) + dn0(k+1,i,j)) * 0.5_f4)
         enddo
      enddo
   enddo
else
   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-2
            wt(k,i,j) = wt(k,i,j)  &
                 + dtlvi * (vt3do(k,i,j) - wp(k,i,j))  &
                 - vt3dj(k,i,j)  &
                 / ((dn0(k,i,j) + dn0(k+1,i,j)) * 0.5_f4)
         enddo
      enddo
   enddo
endif

!print *,'here16'

end subroutine diffvel


!```````````````````````````````````````````````````````````````````
subroutine diffsclr(m1,m2,m3,ia,iz,ja,jz,jd,ia_1,ja_1,ia1,ja1,iz_1,jz_1,iz1  &
     ,jz1,scp,sct,vt3da,vt3db,vt3df,vt3dg,vt3dj,vt3dk,vt3do,vt3dc,rtgt  &
     ,sfcflx,dn0,vkkh,hkkh,reinit, id_idx)

!+ Diffuse scalar

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use main_memory_mod, only: vctr1, a, vctr2, vctr3, vctr4, vctr5, vctr6,  &
     vctr7, i_grid, c_gd, ihorgrad
use vtables_mod, only: idxu, itopt, irtgt, idyv, idxt, idyt, TKEP_vidx
use PRAMS_turbdiff_mod, only: true_horiz_difftf, tridiff1, truhor
use PRAMS_graddiv_mod, only: divcart, gradient

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd,ia_1,ja_1,ia1,ja1,iz_1,jz_1  &
     ,iz1,jz1, id_idx
real(f4), dimension(m1,m2,m3), intent(IN) :: scp,vkkh,hkkh,dn0
real(f4), dimension(m2,m3), intent(IN) :: sfcflx,rtgt
logical, intent(IN) :: reinit

real(f4), dimension(m1,m2,m3), intent(INOUT) :: sct,vt3da,vt3db,vt3dc,vt3df,  &
     vt3dg,vt3dj,vt3dk,vt3do

!== Local declarations:

logical, parameter :: use_Weissbluth_truhor = .FALSE.

real(f4) :: c1, dt_dyn_i
integer :: i,j,k

!=== Executable statements:

! Compute vertical diffusion matrix coefficients for scalars:

if(reinit) then   ! If it is known that a var below has changed, perform reinit

   do j = ja,jz
      do i = ia,iz
         do k = 1,m1-1
            vctr1(k) = c_gd%dzm(k) * (vkkh(k,i,j) + vkkh(k+1,i,j))
         enddo
         c1 = 0.5_f4 * c_gd%dt_dyn / (rtgt(i,j) * rtgt(i,j))
         do k = 2,m1-1
            vt3dj(k,i,j) = -c1 * c_gd%dzt(k) * vctr1(k-1)
            vt3dk(k,i,j) = -c1 * c_gd%dzt(k) * vctr1(k)
            vt3do(k,i,j) = dn0(k,i,j) - vt3dj(k,i,j) - vt3dk(k,i,j)
         enddo
         vt3dj(2,i,j) = ZERO_f4
         if(c_gd%nstbot == 1) then
            vt3do(2,i,j) = dn0(2,i,j) - vt3dk(2,i,j)
         endif
         vt3dk(m1-1,i,j) = ZERO_f4
         if(c_gd%nsttop == 1) then
            vt3do(m1-1,i,j) = dn0(m1-1,i,j) - vt3dj(m1-1,i,j)
         endif
      enddo
   enddo

endif

! compute 2 horizontal scalar gradients needed for dscp/dt:

if(ihorgrad == 1) then
   call gradient(m1,m2,m3,1,iz,ja,jz,scp,vt3df,'XDIR','TPNT')
   call gradient(m1,m2,m3,ia,iz,1,jz,scp,vt3dg,'YDIR','TPNT')

   do j = ja,jz
      do i = 1,iz
         do k = 1,m1-1
            vt3df(k,i,j) = -vt3df(k,i,j) * 0.5_f4 * (hkkh(k,i,j) + hkkh(k,i+1,j))
         enddo
      enddo
   enddo
   
   do j = 1,jz
      do i = ia,iz
         do k = 1,m1-1
            vt3dg(k,i,j) = -vt3dg(k,i,j) * 0.5_f4 * (hkkh(k,i,j) + hkkh(k,i,j+jd))
         enddo
      enddo
   enddo
endif

! horizontal flux divergence for scalars:

if(ihorgrad == 1) then

   call divcart(m1,m2,m3,ia,iz,ja,jz,vt3df,vt3da,'XDIR','UPNT')
   call divcart(m1,m2,m3,ia,iz,ja,jz,vt3dg,vt3db,'YDIR','VPNT')

elseif(ihorgrad == 2) then

   if(use_Weissbluth_truhor) then

      call truhor(m1,m2,m3,ia,iz,ja,jz,scp,vt3da,'xdir','dxu',a(idxu)  &
           ,a(itopt),a(irtgt),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5,vctr6,vctr7  &
           ,jd,hkkh,dn0,c_gd%dt_dyn,i_grid)

      call truhor(m1,m2,m3,ia,iz,ja,jz,scp,vt3db,'ydir','dyv',a(idyv)  &
           ,a(itopt),a(irtgt),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5,vctr6,vctr7  &
           ,jd,hkkh,dn0,c_gd%dt_dyn,i_grid)

   else   ! Use TIM's scheme

      ! Calculate Cartesian horizontal diffusion for Scalar in x-dir; one step:
      !  (1) Calculate d/dx[dn0*Kh*(dScalar/dx)] {vt3da}
      call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,scp,vt3da,'X','XT '  &
           ,a(idxt),a(idxu),a(itopt),a(irtgt),c_gd%zt,jd,i_grid,hkkh)

      ! Calculate Cartesian horizontal diffusion for Scalar in y-dir; one step:
      !  (1) Calculate d/dy[dn0*Kh*(dScalar/dy)] {vt3db}
      call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,scp,vt3db,'Y','YT '  &
           ,a(idyt),a(idyv),a(itopt),a(irtgt),c_gd%zt,jd,i_grid,hkkh)

   endif

endif

! Finish matrix coefficients:

do j=ja,jz
   do i=ia,iz
      do k=2,m1-1
         vt3dc(k,i,j) = scp(k,i,j) * dn0(k,i,j)
      enddo
      
      if(c_gd%nstbot == 1) then
         vt3dc(2,i,j) = scp(2,i,j) * dn0(2,i,j)  &
              + sfcflx(i,j) * (dn0(1,i,j) + dn0(2,i,j)) * 0.5_f4  &
              * c_gd%dt_dyn * c_gd%dzt(2) / rtgt(i,j)
      else
         vt3dc(2,i,j) = scp(2,i,j) * dn0(2,i,j)  &
              + 0.5_f4 * c_gd%dt_dyn * (vkkh(1,i,j) + vkkh(2,i,j))  &
              * scp(1,i,j) * c_gd%dzm(2) * c_gd%dzt(2) / rtgt(i,j) ** 2
      endif

      if(c_gd%nsttop == 0) then
         vt3dc(m1-1,i,j) = scp(m1-1,i,j) * dn0(m1-1,i,j)  &
              - 0.5_f4 * c_gd%dt_dyn * (vkkh(m1-1,i,j) + vkkh(m1,i,j))  &
              * scp(m1,i,j) * c_gd%dzm(m1-1) * c_gd%dzt(m1) / rtgt(i,j) ** 2
      endif
   enddo
enddo

call tridiff1(m1,m2,m3,ia,iz,ja,jz,m1-1,vt3dj,vt3do,vt3dk,vt3dc,vt3df,vt3dg)

dt_dyn_i = ONE_f4/c_gd%dt_dyn

! Only for SGS TKE field(s). To avoid numerical instability at larger timesteps,
!  do not allow the combined (parameterized) source/sink tendencies to remove
!  more than 80% of the current TKE value:
if (id_idx == TKEP_vidx) then
   do j=ja,jz
      do i=ia,iz
         do k=2,m1-1
            if (sct(k,i,j)*c_gd%dt_dyn < -0.80_f4*scp(k,i,j))  &
                 sct(k,i,j) = -0.80_f4*scp(k,i,j)/c_gd%dt_dyn
         end do
      end do
   end do
end if

if(jd == 1) then
   do j=ja,jz
      do i=ia,iz
         do k=2,m1-1
            sct(k,i,j) = sct(k,i,j)  &
                 + dt_dyn_i * (vt3df(k,i,j) - scp(k,i,j))  &
                 - (vt3da(k,i,j) + vt3db(k,i,j)) / dn0(k,i,j)
         enddo
      enddo
   enddo
else
   do j=ja,jz
      do i=ia,iz
         do k=2,m1-1
            sct(k,i,j) = sct(k,i,j)  &
                 + dt_dyn_i * (vt3df(k,i,j) - scp(k,i,j))  &
                 - vt3da(k,i,j) / dn0(k,i,j)
         enddo
      enddo
   enddo
endif
 
end subroutine diffsclr


!```````````````````````````````````````````````````````````````````
subroutine diffsclr_A(m1,m2,m3,ia,iz,ja,jz,jd,ia_1,ja_1,ia1,ja1,iz_1,jz_1  &
     ,iz1,jz1,scp,sct,vt3da,vt3db,vt3df,vt3dg,vt3dj,vt3dk,vt3do,vt3dc,rtgt  &
     ,sfcflx,dn0,vkkh,hkkh,reinit)

!+ Diffuse scalar

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use main_memory_mod, only: vctr1, a, vctr2, vctr3, vctr4, vctr5, vctr6,  &
     vctr7, i_grid, c_gd, ihorgrad
use vtables_mod, only: idxu, itopt, irtgt, idyv, idxt, idyt
use PRAMS_turbdiff_mod, only: true_horiz_difftf, tridiff1, truhor
use PRAMS_graddiv_mod, only: divcart, gradient

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1,m2,m3,ia,iz,ja,jz,jd,ia_1,ja_1,ia1,ja1,iz_1,jz_1  &
     ,iz1,jz1
real(f4), dimension(m1,m2,m3), intent(IN) :: scp,vkkh,hkkh,dn0
real(f4), dimension(m2,m3), intent(IN) :: sfcflx,rtgt
logical, intent(IN) :: reinit

real(f4), dimension(m1,m2,m3), intent(INOUT) :: sct,vt3da,vt3db,vt3dc,vt3df,  &
     vt3dg,vt3dj,vt3dk,vt3do

!== Local declarations:

logical, parameter :: use_Weissbluth_truhor = .FALSE.

real(f4) :: c1, dt_dyn_i
integer :: i, j, k

!=== Executable statements:

! Compute vertical diffusion matrix coefficients for scalars:

if(reinit) then   ! If it is known that a var below has changed, perform reinit

   do j = ja,jz
      do i = ia,iz
         do k = 1,m1-1
            vctr1(k) = c_gd%dzm(k) * (vkkh(k,i,j) + vkkh(k+1,i,j))
         enddo
         c1 = 0.5_f4 * c_gd%dt_dyn / (rtgt(i,j) * rtgt(i,j))
         do k = 2,m1-1
            vt3dj(k,i,j) = -c1 * c_gd%dzt(k) * vctr1(k-1)
            vt3dk(k,i,j) = -c1 * c_gd%dzt(k) * vctr1(k)
            vt3do(k,i,j) = dn0(k,i,j) - vt3dj(k,i,j) - vt3dk(k,i,j)
         enddo
         vt3dj(2,i,j) = ZERO_f4
         if(c_gd%nstbot == 1) then
            vt3do(2,i,j) = dn0(2,i,j) - vt3dk(2,i,j)
         endif
         vt3dk(m1-1,i,j) = ZERO_f4
         if(c_gd%nsttop == 1) then
            vt3do(m1-1,i,j) = dn0(m1-1,i,j) - vt3dj(m1-1,i,j)
         endif
      enddo
   enddo

endif

! compute 2 horizontal scalar gradients needed for dscp/dt:

if(ihorgrad == 1) then
   call gradient(m1,m2,m3,1,iz,ja,jz,scp,vt3df,'XDIR','TPNT')
   call gradient(m1,m2,m3,ia,iz,1,jz,scp,vt3dg,'YDIR','TPNT')

   do j = ja,jz
      do i = 1,iz
         do k = 1,m1-1
            vt3df(k,i,j) = -vt3df(k,i,j) * 0.5_f4 * (hkkh(k,i,j) + hkkh(k,i+1,j))
         enddo
      enddo
   enddo
   
   do j = 1,jz
      do i = ia,iz
         do k = 1,m1-1
            vt3dg(k,i,j) = -vt3dg(k,i,j) * 0.5_f4 * (hkkh(k,i,j) + hkkh(k,i,j+jd))
         enddo
      enddo
   enddo
endif

! horizontal flux divergence for scalars:

if(ihorgrad == 1) then

   call divcart(m1,m2,m3,ia,iz,ja,jz,vt3df,vt3da,'XDIR','UPNT')
   call divcart(m1,m2,m3,ia,iz,ja,jz,vt3dg,vt3db,'YDIR','VPNT')

elseif(ihorgrad == 2) then

   if(use_Weissbluth_truhor) then

      call truhor(m1,m2,m3,ia,iz,ja,jz,scp,vt3da,'xdir','dxu',a(idxu)  &
           ,a(itopt),a(irtgt),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5,vctr6,vctr7  &
           ,jd,hkkh,dn0,c_gd%dt_dyn,i_grid)

      call truhor(m1,m2,m3,ia,iz,ja,jz,scp,vt3db,'ydir','dyv',a(idyv)  &
           ,a(itopt),a(irtgt),c_gd%zt,vctr1,vctr2,vctr3,vctr4,vctr5,vctr6,vctr7  &
           ,jd,hkkh,dn0,c_gd%dt_dyn,i_grid)

   else   ! Use TIM's scheme

      ! Calculate Cartesian horizontal diffusion for Scalar in x-dir; one step:
      !  (1) Calculate d/dx[dn0*Kh*(dScalar/dx)] {vt3da}
      call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,scp,vt3da,'X','XT '  &
           ,a(idxt),a(idxu),a(itopt),a(irtgt),c_gd%zt,jd,i_grid,hkkh)

      ! Calculate Cartesian horizontal diffusion for Scalar in y-dir; one step:
      !  (1) Calculate d/dy[dn0*Kh*(dScalar/dy)] {vt3db}
      call true_horiz_difftf(m1,m2,m3,ia,iz,ja,jz,scp,vt3db,'Y','YT '  &
           ,a(idyt),a(idyv),a(itopt),a(irtgt),c_gd%zt,jd,i_grid,hkkh)

   endif

endif

! Finish matrix coefficients:

do j=ja,jz
   do i=ia,iz
      do k=2,m1-1
         vt3dc(k,i,j) = scp(k,i,j) * dn0(k,i,j)
      enddo
      
      if(c_gd%nstbot == 1) then
         vt3dc(2,i,j) = scp(2,i,j) * dn0(2,i,j)  &
              + sfcflx(i,j) * (dn0(1,i,j) + dn0(2,i,j)) * 0.5_f4  &
              * c_gd%dt_dyn * c_gd%dzt(2) / rtgt(i,j)
      else
         vt3dc(2,i,j) = scp(2,i,j) * dn0(2,i,j)  &
              + 0.5_f4 * c_gd%dt_dyn * (vkkh(1,i,j) + vkkh(2,i,j))  &
              * scp(1,i,j) * c_gd%dzm(2) * c_gd%dzt(2) / rtgt(i,j) ** 2
      endif

      if(c_gd%nsttop == 0) then
         vt3dc(m1-1,i,j) = scp(m1-1,i,j) * dn0(m1-1,i,j)  &
              - 0.5_f4 * c_gd%dt_dyn * (vkkh(m1-1,i,j) + vkkh(m1,i,j))  &
              * scp(m1,i,j) * c_gd%dzm(m1-1) * c_gd%dzt(m1) / rtgt(i,j) ** 2
      endif
   enddo
enddo

call tridiff1(m1,m2,m3,ia,iz,ja,jz,m1-1,vt3dj,vt3do,vt3dk,vt3dc,vt3df,vt3dg)

dt_dyn_i = ONE_f4/c_gd%dt_dyn

if(jd == 1) then
   do j=ja,jz
      do i=ia,iz
         do k=2,m1-1
            sct(k,i,j) = sct(k,i,j) + dn0(k,i,j)*  &
                 (dt_dyn_i * (vt3df(k,i,j) - scp(k,i,j))  &
                 - (vt3da(k,i,j) + vt3db(k,i,j)) / dn0(k,i,j))
         enddo
      enddo
   enddo
else
   do j=ja,jz
      do i=ia,iz
         do k=2,m1-1
            sct(k,i,j) = sct(k,i,j) + dn0(k,i,j)*  &
                 (dt_dyn_i * (vt3df(k,i,j) - scp(k,i,j))  &
                 - vt3da(k,i,j) / dn0(k,i,j))
         enddo
      enddo
   enddo
endif

end subroutine diffsclr_A
