!```````````````````````````````````````````````````````````````````
module turbulent_K_coeffs_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine brunt_vaisalla(m1, m2, m3, ia, iz, ja, jz, theta, rv, rcp, pp,  &
     pi0, en2, rtgt)

!+ Calculate Brunt-Vaisalla frequency (squared)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4
use body_specific_constants_mod, only: R_air, ep, cp, g_Xoid
use main_memory_mod, only: vctr11, vctr12, vctr1, vctr22, vctr32 => vctr3,  &
     c_gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz
real(f4), dimension(m1,m2,m3), intent(IN) :: theta, rcp, rv, pp, pi0
real(f4), dimension(m2,m3), intent(IN) :: rtgt

real(f4), dimension(m1,m2,m3), intent(INOUT) :: en2

!== Local declarations:

integer :: i, j, k, iweten, iwdiffk !% ,ki
!%real(f4) :: c1,c2,c3,ci1,ci2,ci3,rvlsi,rvii

!=== Executable statements:

! Calculate BV frq squared (= N^2 = `en2`):

iweten = 1

iwdiffk = 0
!%c1 = alvl / R_air
!%c2 = ep * alvl ** 2. / (cp * R_air)
!%c3 = alvl / cp
!%ci1 = alvi / R_air
!%ci2 = ep * alvi ** 2. / (cp * R_air)
!%ci3 = alvi / cp

! Calculate potential temperature profile:

do j = ja,jz
   do i = ia,iz

      do k = 1,m1
         vctr11(k) = theta(k,i,j)
         vctr12(k) = theta(k,i,j)
         vctr32(k) = ZERO_f4
      end do
!%      if(iaertog == 1) then
!%         do k = 1,m1
!%            vctr12(k) = vctr11(k) * (1. + 0.61 * rv(k,i,j))
!%            vctr32(k) = (rtp(k,i,j) - rv(k,i,j))
!%         enddo
!%      endif
      
!%      ! Check for saturation (if "wet BV freq" wanted):
!%
!%      if(iaertog > 1 .and. iweten == 1) then
!%         do k = 1,m1
!%            pi(k) = (pp(k,i,j) + pi0(k,i,j)) / cp
!%            temp(k) = theta(k,i,j) * pi(k)
!%            prt(k) = p00 * pi(k) ** cpor
!%         enddo
!%         call mrsl(m1,prt(1),temp(1),rvls(1))
!%         do k = 1,m1
!%            vctr2(k) = c1
!%            vctr3(k) = c2
!%            vctr4(k) = c3
!%         enddo
!%         ki = m1 + 1
!%
!%        ! if any ice phase microphysics are activated ....
!%
!%         if ((ipris .ge. 1 .or. isnow .ge. 1 .or.  &
!%              igraup .ge. 1 .or. iaggr .ge. 1 .or.  &
!%              ihail .ge. 1) .and. level .eq. 3) then
!%
!%            !   find level of -20 c.  assume ice saturation above this
!%            !           level.
!%
!%            do k = 1,m1
!%               if (temp(k) .le. -20.) then
!%                  ki = k
!%                  go to 10
!%               endif
!%            enddo
!%            ki = m1 + 1
!%10          continue
!%
!%            call mrsi(m1-ki+1,prt(ki),temp(ki),rvls(ki))
!%            call mrsi(1,prt(ki-1),temp(ki-1),rvlsi)
!%            do k = ki,m1
!%               vctr2(k) = ci1
!%               vctr3(k) = ci2
!%               vctr4(k) = ci3
!%            enddo
!%         endif
!%
!%         if (level .eq. 3) then
!%            do k = 1,m1
!%               rc(k) = rcp(k,i,j)
!%            enddo
!%         else
!%            do k = 1,m1
!%               rc(k) = max(rv(k,i,j) / rvls(k) - .999,0.)
!%            enddo
!%         endif
!%
!%      endif

      if (iwdiffk == 0) then

         do k = 2,m1-1   ! g/dz_2kt
            vctr1(k) = g_Xoid / ((c_gd%zt(k+1) - c_gd%zt(k-1)) * rtgt(i,j))
         end do
!%         if(iaertog > 1 .and. iweten == 1) then
!%            do k = 2,m1-1
!%               if (rc(k) .gt. 0.) then
!%                  rvii = rvls(k-1)
!%                  if (k .eq. ki) rvii = rvlsi
!%                  en2(k,i,j) = vctr1(k) * (  &
!%                     (1. + vctr2(k) * rvls(k) / temp(k))  &
!%                     / (1. + vctr3(k) * rvls(k) / temp(k) ** 2)  &
!%                     * ((vctr11(k+1) - vctr11(k-1)) / vctr11(k)  &
!%                     + vctr4(k) / temp(k) * (rvls(k+1) - rvii))  &
!%                     - (rtp(k+1,i,j) - rtp(k-1,i,j)))
!%               else
!%                  en2(k,i,j) = vctr1(k)*((vctr12(k+1)-vctr12(k-1))  &
!%                     / vctr12(k) - (vctr32(k+1) - vctr32(k-1)))
!%               endif
!%
!%            enddo
!%         else
            do k = 2,m1-1
                  !   g    [ dtheta_2kt    dvapterm_2kt ]
                  ! -----  [  -------   -   ----------  ]
                  ! dz_2kt [ theta_kt                   ]
               en2(k,i,j) = vctr1(k) * ((vctr12(k+1)-vctr12(k-1))  &
                    / vctr12(k) - (vctr32(k+1) - vctr32(k-1)))
            end do
!%         endif
         en2(1,i,j) = en2(2,i,j)

      else

         do k=1,m1-1   ! g/dz_km
            vctr1(k)=g_Xoid*c_gd%dzm(k) /rtgt(i,j)
         end do
!%         if(iaertog > 1 .and. iweten == 1) then
!%            do k=1,m1-1
!%               vctr20(k)=(temp(k)+temp(k+1))*.5
!%               vctr21(k)=(vctr11(k)+vctr11(k+1))*.5
!%               vctr22(k)=(vctr12(k)+vctr12(k+1))*.5
!%               vctr23(k)=(vctr23(k)+vctr23(k+1))*.5
!%            enddo
!%            do k=1,m1-1
!%               if(rc(k).gt.0.)then
!%                  en2(k,i,j)=vctr1(k)*(  &
!%                       (1.+vctr2(k)*vctr23(k)/vctr20(k))  &
!%                       /(1.+vctr3(k)*vctr23(k)/vctr20(k)**2)  &
!%                       *( (vctr11(k+1)-vctr11(k))/vctr21(k)  &
!%                       +  vctr4(k)/vctr20(k)*(rvls(k+1)-rvls(k)))  &
!%                       -(rtp(k+1,i,j)-rtp(k,i,j))   )
!%               else
!%                  en2(k,i,j)=vctr1(k)*((vctr12(k+1)-vctr12(k))  &
!%                       /vctr22(k)-(vctr32(k+1)-vctr32(k)))
!%               endif
!%            enddo
!%         else
            do k=1,m1-1
               vctr22(k) = (vctr12(k)+vctr12(k+1))*0.5_f4    ! theta_km
                  !   g   [ dtheta_km    dvapterm_km ]
                  ! ----- [  -------  -   ---------- ]
                  ! dz_km [ theta_km                 ]
               en2(k,i,j) = vctr1(k)*((vctr12(k+1)-vctr12(k))  &
                    /vctr22(k)-(vctr32(k+1)-vctr32(k))) 
            end do
!%         endif
       end if

      en2(c_gd%nzp,i,j) = en2(c_gd%nz,i,j)

   end do
end do

end subroutine brunt_vaisalla


!```````````````````````````````````````````````````````````````````
subroutine mxdefm(m1, m2, m3, ia, iz, ja, jz, ibndc, jd, vt3dh, vt3di,  &
     vt3dj, vt3dk, scr1, scr2, dn0, rtgt, dxt, dyt, my_rank)

!+ Calculate deformation K's (mixing coefficients)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use main_memory_mod, only: vctr1, vctr2, c_gd
use body_specific_constants_mod, only: vonk

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz, ibndc, jd, my_rank
real(f4), dimension(m1,m2,m3), intent(IN) :: vt3di,  & ! [s^-2]
                                             vt3dj,  & ! [s^-2]
                                             dn0 ! [kg/m^3]
real(f4), dimension(m2,m3), intent(IN) :: rtgt,  &
                                          dxt, dyt ! [m^-1]

real(f4), dimension(m1,m2,m3), intent(INOUT) ::  &
                                           vt3dh,  & ! IN:[s^-2], OUT:[]
                                           vt3dk, scr1, scr2 ! OUT:[]

!== Local declarations:

! This routine calculates the mixing coefficients with a   
!  Smagorinsky-type deformational based K with an optional 
!  unstable Brunt-Vaisala enhancement and an optional      
!  Richardson number modification.                         

integer :: i, j, k, irich, ienfl
real(f4) :: csx2, sq300, enfl, rchmax, c1, c2, c3, c4, akm, ambda, vkz2,  &
     rmin, rmax

!=== Executable statements:

irich = 1
ienfl = 1

csx2 = c_gd%csx*c_gd%csx
sq300 = 90000._f4
if (c_gd%idiffk == 2 .or. c_gd%idiffk == 3) then
   rmin = -100._f4
   rmax = ONE_f4 / c_gd%zkhkm
   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-1
            vt3dk(k,i,j) = max(min(vt3dj(k,i,j)  & ! [none]
                 / max(vt3di(k,i,j),1.e-15_f4), rmax), rmin)
         end do
      end do
   end do
   enfl = real(ienfl, kind=f4)
   rchmax = ONE_f4+ 9._f4*real(irich, kind=f4)
   do k = 2,m1
      vctr1(k) = c_gd%csz * (c_gd%zm(k) - c_gd%zm(k-1)) ! [m]
      vctr2(k) = vctr1(k) * vctr1(k) ! [m^2]
   end do
end if

select case (c_gd%idiffk)
case (1,6:9)
   do j = ja,jz
      do i = ia,iz
         c2 = ONE_f4 / (dxt(i,j) * dxt(i,j)) ! [m^2]
         c3 = csx2 * c2 ! [m^2]
         akm = c_gd%akmin * 0.075_f4 * c2 ** (0.666667_f4) ! [m^2]
         do k = 2,m1-1
            scr2(k,i,j) = dn0(k,i,j)  & ! [kg/(m.s)]
                 * max(akm, c3*sqrt(vt3dh(k,i,j)))
         end do
      end do
   end do

case (2)
   do j = ja,jz
      do i = ia,iz
         c1 = rtgt(i,j) * rtgt(i,j)
         c2 = ONE_f4 / (dxt(i,j) * dxt(i,j)) ! [m^2]
         c3 = csx2 * c2 ! [m^2]
         akm = c_gd%akmin * 0.075_f4 * c2 ** (0.666667_f4) ! [m^2]
         c4 = vonk * vonk * c1
         do k = 2,m1-1
! old csz*dz len  scr1(k,i,j) = dn0(k,i,j) * c1 * vctr2(k)

! asymptotic vertical scale length from bjorn with modifications:
! c3 is (csx * dx)^2, c1*vctr2(k) is (csz * dz)^2, sq300 is the square
! of 300 meters (used as a limit for horizontal grid spacing influence
! on vertical scale length), ambda is (asymptotic_vertical_length_scale)^2,
! and vkz2 is (vonk * height_above_surface)^2.

            ambda = max(c1 * vctr2(k), min(sq300, c3)) ! [m^2]
            vkz2 = c4 * c_gd%zt(k) * c_gd%zt(k) ! [m^2]
            scr1(k,i,j) = dn0(k,i,j) * vkz2 / (vkz2 / ambda + 1)  & ! [kg/(m.s)]
                 * (sqrt(vt3di(k,i,j))  &
                 + enfl * sqrt(max(ZERO_f4, -vt3dj(k,i,j))))*min(rchmax  &
                 , sqrt(max(ZERO_f4,(ONE_f4-c_gd%zkhkm*vt3dk(k,i,j)))))

            scr2(k,i,j) = dn0(k,i,j)  & ! [kg/(m.s)]
                 * max(akm, c3*sqrt(vt3dh(k,i,j)))
            vt3dh(k,i,j) = scr1(k,i,j) * c_gd%zkhkm ! [kg/(m.s)]

         end do
      end do
   end do
!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!     call friclyr(nzp,nxp,nyp,a(iscr1),a(iustarl),a(itstarl)
!    +    ,a(iustarw),a(itstarw),a(ipctlnd),a(itheta),a(irtgt))
!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

case (3)
   do j = ja,jz
      do i = ia,iz
         c1 = rtgt(i,j) * rtgt(i,j)
         do k = 2,m1-1
            scr1(k,i,j) = dn0(k,i,j) * c1 * vctr2(k)  & ! [kg/(m.s)]
                 * (sqrt(vt3dh(k,i,j))  &
                 + enfl * sqrt(max(ZERO_f4, -vt3dj(k,i,j))))*min(rchmax  &
                 , sqrt(max(ZERO_f4, (ONE_f4-c_gd%zkhkm*vt3dk(k,i,j)))))
            scr2(k,i,j) = scr1(k,i,j) ! [kg/(m.s)]
            vt3dh(k,i,j) = scr1(k,i,j) * c_gd%zkhkm ! [kg/(m.s)]
         end do
      end do
   end do
!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!     call friclyr(nzp,nxp,nyp,a(iscr1),a(iustarl),a(itstarl)
!    +    ,a(iustarw),a(itstarw),a(ipctlnd),a(itheta),a(irtgt))
!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

end select

end subroutine mxdefm


!```````````````````````````````````````````````````````````````````
subroutine mxtked(m1, m2, m3, ia, iz, ja, jz, ibndc, jd, tkep, tket, up, vp,  &
     wp, rv, theta, vt3da, vt3dc, vt3dh, vt3dj, scr1, scr2, uw, vw, wfz, tfz,  &
     dxt, rtgt, dn0)

!+ Calculates K's based on SGS turb. scheme of Deardorff (1980)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use main_memory_mod, only: vctr8, vctr5, vctr7, c_gd
use body_specific_constants_mod, only: g_Xoid

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz, ibndc, jd
real(f4), dimension(m1,m2,m3), intent(IN) :: tkep, up, vp, wp, rv, theta,  &
     vt3da, vt3dc, vt3dj, dn0
real(f4), dimension(m2,m3), intent(IN) :: dxt, uw, vw, wfz, tfz, rtgt

real(f4), dimension(m1,m2,m3), intent(INOUT) :: tket, vt3dh, scr1, scr2

!== Local declarations:

! This routine calculates the mixing coefficients based on the     
!   prognostic tke and diagnostic length scale, according to the   
!   deardorff scheme.                                              

real(f4) :: sqrttkep, tket2, c1v, scluh, scluv, sclh, sclv, scluh2, scla, ftke
integer :: i, j, k

!=== Executable statements:

do j=ja,jz
   do i=ia,iz
 
      tket2 = tket(2,i,j)    ! Save OLD value of SGS TKE tendency
                             !   for lowest PRAMS level

!      scluh = 0.43/dxt(i,j)      ! Horizontal length scale
!@      scluh = 0.17_f4/dxt(i,j)
      scluh = 0.43_f4/dxt(i,j)
      scluh2 = scluh*scluh !*scluh*scluh   ! Don't calculate this for every k below
!      c1v = 0.27*rtgt(i,j)             ! Vertical factor (for length scale)
!@      c1v = 0.15_f4*rtgt(i,j)
      c1v = 0.27_f4*rtgt(i,j)

      do k=2,m1-1     ! Loop over most model levels

!         scluv = (scluh2*c1v*(zm(k)-zm(k-1)))**0.2   ! Vertical length scale
         ! Don't let scluv get too small
         scluv = max(0.7_f4*scluh, (scluh2*c1v*(c_gd%zm(k)-c_gd%zm(k-1)))**0.3333333_f4)
         
         if (vt3dj(k,i,j) < 1.e-10_f4) then   ! If N^2 very small, then unstable
            sclh = scluh
            sclv = scluv
         else                         ! Stable case
            sclh = 0.76_f4*sqrt(tkep(k,i,j)/vt3dj(k,i,j)) ! Stable length scale

            ! If default vertical length scale is smaller than the 
            !   "stable case" length scale, use it instead:
            if (sclh > scluv) then
               sclv = scluv
            else
               sclv = sclh
            end if
            ! If default horizontal length scale is smaller than the 
            !   "stable case" length scale, use it instead:
            if (sclh > scluh) sclh = scluh
         end if

! now, vt3dh is strain rate and scr2 is dw/dz

         sqrttkep = sqrt(tkep(k,i,j))  ! Square root of previous SGS TKE value
         ftke = 0.1_f4*sqrttkep    ! TKE factor for K calcs

         scr1(k,i,j) = sclv*ftke        ! vkm - Vertical momentum K

         scla = (sclh*sclh*sclv)**0.333333_f4    ! "Average" 3D length scale

         vctr5(k) = scla*ftke*vt3dh(k,i,j)  &    ! Shear prod term
              -0.333333_f4*tkep(k,i,j)  &                 ! Turbulent transport
              *(vt3da(k,i,j)+vt3dc(k,i,j)+scr2(k,i,j)) !   term

         ! vkh = vkm * (turbulent Prandtl number)
         !   (vkh - Vertical thermodynamic K)
         vt3dh(k,i,j) = scr1(k,i,j)*(ONE_f4+2._f4*sclv/scluv)

         scr2(k,i,j) = sclh*ftke     ! hkm - Horizontal momentum K

! now, vt3dh is vkh and scr2 is hkm

         ! A "constant" in the dissipation term (here proportional to ratio of
         !   length scales)
         ! A recent (2000) ARPS paper suggests that this should be = 0.93
         !   everywhere except the lowest model level, where = 3.9 
!combined         vctr26(k) = 0.19+  &          ! Original (Deardorff 1980)  
!combined              0.51*(((sclh*sclh*sclv)/(scluh2*scluv))**0.333333)
        ! Dissipation term:
!         vctr8(k) = vctr26(k) * tkep(k,i,j) * sqrttkep / scl
!combined         vctr8(k) = vctr26(k)*tkep(k,i,j)*sqrttkep/(sclh*sclh*sclv)**0.333333
!         vctr8(k) = 0.93*tkep(k,i,j)*sqrttkep/(sclh*sclh*sclv)**0.333333

! Combined form of dissipation term
         vctr8(k) = tkep(k,i,j)*sqrttkep*(0.19_f4/scla+  &
              0.51_f4/(scluh2*scluv)**0.333333_f4)

         vctr7(k) = -vt3dh(k,i,j)*vt3dj(k,i,j)   ! Buoyancy term?
 
         tket(k,i,j) = tket(k,i,j)+vctr5(k)+vctr7(k)-vctr8(k)  ! TKE eq.

         vt3dh(k,i,j) = vt3dh(k,i,j)*dn0(k,i,j)
         scr1(k,i,j) = scr1(k,i,j)*dn0(k,i,j)
         scr2(k,i,j) = scr2(k,i,j)*dn0(k,i,j)

      end do

      ! TKE at lowest level (note dissipation term factor change):

!no surface linkage  
!      tket(2,i,j) = tket2+vctr5(3)+vctr7(2)-(vctr8(2)*3.9)
     
      tket(2,i,j) = tket2  &     ! Past TKE tendency
           +vctr5(3)  &    ! Shear production/turb transport
           +(tfz(i,j)*g_Xoid/theta(2,i,j))  &   ! Buoyancy term
           -(vctr8(2)*3.9_f4)      ! Dissipation term
     
   end do
end do

end subroutine mxtked


!```````````````````````````````````````````````````````````````````
subroutine klbnd(m1, m2, m3, ibndc, jd, akay, dn0)

!+ Mixing coefficient boundary conditions

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ibndc, jd
real(f4), dimension(m1,m2,m3), intent(IN) :: dn0

real(f4), dimension(m1,m2,m3), intent(INOUT) :: akay

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

do j = 1,m3
   do i = 1,m2
      akay(1,i,j) = akay(2,i,j) * dn0(1,i,j) / dn0(2,i,j)
      akay(m1,i,j) = akay(m1-1,i,j) * dn0(m1,i,j) / dn0(m1-1,i,j)
   end do
end do

if (iand(ibndc, 1) /= 0) then
   do j = 1,m3
      do k = 1,m1
         akay(k,1,j) = akay(k,2,j)
      end do
   end do
endif

if (iand(ibndc, 2) /= 0) then
   do j = 1,m3
      do k = 1,m1
         akay(k,m2,j) = akay(k,m2-1,j)
      end do
   end do
end if

if (jd == 1) then
   if (iand(ibndc, 4) /= 0) then
      do i = 1,m2
         do k = 1,m1
            akay(k,i,1) = akay(k,i,2)
         end do
      end do
   end if

   if (iand(ibndc, 8) /= 0) then
      do i = 1,m2
         do k = 1,m1
            akay(k,i,m3) = akay(k,i,m3-1)
         end do
      end do
   end if
end if

end subroutine klbnd


!```````````````````````````````````````````````````````````````````
subroutine tkeinit(n1, n2, n3, tkep)

!+ Initialize and/or keep TKE values >= tkmin

use univ_kind_defs_mod, only: f4
use body_specific_constants_mod, only: tkmin

implicit none

!=== Argument declarations:

integer, intent(IN) :: n1, n2, n3

real(f4), dimension(n1,n2,n3), intent(INOUT) :: tkep

!== Local declarations:

integer :: i, j, k

!=== Executable statements:

! Limit the values to the minimum:

do j=1,n3
   do i=1,n2
      do k=1,n1
         tkep(k,i,j) = max(tkmin, tkep(k,i,j))
      end do
   end do
end do

end subroutine tkeinit


!```````````````````````````````````````````````````````````````````
subroutine filter_avg9pt(m1, m2, m3, ia, iz, ja, jz, var, scr, rtgt)

!+ Apply "nine-point" spatial filter (mean) to array values

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_NINTH_f4, ZERO_f4
use main_memory_mod, only: jdim, c_gd

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz
real(f4), dimension(m2,m3), intent(IN) :: rtgt

real(f4), dimension(m1,m2,m3), intent(INOUT) :: var, scr

!== Local declarations:

integer :: i, j, k
real(f4) :: sum, z

!=== Executable statements:

scr = var   ! Copy contents of entire array

do j=ja,jz
   do i=ia,iz

      k = 1
      z = c_gd%zt(k)*rtgt(i,j)
      do while (z <= 500._f4)    ! Only perform for lowest levels   

         if (scr(k,i,j) > ZERO_f4) then  ! BVfreq only: if stable
            sum = scr(k,i-1,j+jdim)+scr(k,i,j+jdim)+scr(k,i+1,j+jdim)+  &
                 scr(k,i-1,j)+scr(k,i,j)+scr(k,i+1,j)+  &
                 scr(k,i-1,j-jdim)+scr(k,i,j-jdim)+scr(k,i+1,j-jdim)
            
            var(k,i,j) = sum*ONE_NINTH_f4
         else   ! BVfreq only: if neutral or unstable
            var(k,i,j) = scr(k,i,j)
         end if
    
         k = k+1
         z = c_gd%zt(k)*rtgt(i,j)
      end do

      var(k:m1,i,j) = scr(k:m1,i,j)  ! Fill remainder of column

   end do
end do

end subroutine filter_avg9pt


!```````````````````````````````````````````````````````````````````
subroutine SGS_Sullivan_etal_1994(m1, m2, m3, ia, iz, ja, jz, ibndc, jd,  &
     tkep, tket, up, vp, wp, rv, theta, vt3da, vt3dc, vt3dh, vt3dj, scr1,  &
     scr2, uw, vw, wfz, tfz, dxt, rtgt, dn0)

!+ Calculates K-coeffs based on SGS turbulence scheme of Sullivan et al. (1994)
!
!    ***************** NOT YET FINISHED **********************

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use main_memory_mod, only: vctr8, vctr5, vctr7, c_gd
use body_specific_constants_mod, only: g_Xoid, vonk

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz, ibndc, jd
real(f4), dimension(m1,m2,m3), intent(IN) :: tkep, up, vp, wp, rv, theta,  &
     vt3da, vt3dc, vt3dj, dn0
real(f4), dimension(m2,m3), intent(IN) :: dxt, uw, vw, wfz, tfz, rtgt

real(f4), dimension(m1,m2,m3), intent(INOUT) :: tket, vt3dh, scr1, scr2

!== Local declarations:

! This routine calculates the mixing coefficients based on the     
!   prognostic tke and diagnostic length scale, according to the   
!   deardorff scheme.                                              

real(f4), parameter :: eps_f4 = epsilon(3._f4)

real(f4) :: tket2, c1v, scluh, scluv, sclh, sclv, scluh2, scla, ftke
integer :: i, j, k

real(f4) :: knob_h, knob_v, scale_h, scale_h_sq, factor_v, scale_v,  &
     mixing_scale, nu_m, nu_s, mag_rate_of_strain_tensor, tau_m,  &
     avg_mesh_spacing, sqrt_tkep

!=== Executable statements:

do j=ja,jz
   do i=ia,iz
 
      tket2 = tket(2,i,j)    ! Save OLD value of SGS TKE tendency
                             !   for lowest PRAMS level

      ! "Knobs" for modulating the mesh length scales used in the following
      !  calculations:
      knob_h = ONE_f4; knob_v = ONE_f4

      scale_h = knob_h/dxt(i,j)      ! Horizontal mesh length scale
      scale_h_sq = scale_h*scale_h
 
      factor_v = knob_v*rtgt(i,j)   ! Factor for vertical mesh length scale

      do k=2,m1-1     ! Loop over most model levels

!         scluv = (scluh2*c1v*(zm(k)-zm(k-1)))**0.2   ! Vertical length scale
!         ! Don't let scluv get too small
!         scluv = max(0.7_f4*scluh,(scluh2*c1v*(zm(k)-zm(k-1)))**0.3333333_f4)
         
         scale_v = factor_v*(c_gd%zm(k)-c_gd%zm(k-1))
     
!         else                         ! Stable case
!            sclh = 0.76_f4*sqrt(tkep(k,i,j)/vt3dj(k,i,j)) ! Stable length scale!
!
!            ! If default vertical length scale is smaller than the 
!            !   "stable case" length scale, use it instead:
!            if(sclh > scluv) then
!               sclv = scluv
!            else
!               sclv = sclh
!            endif
!            ! If default horizontal length scale is smaller than the 
!            !   "stable case" length scale, use it instead:
!            if(sclh > scluh) then
!               sclh = scluh
!            endif
!         endif

         sqrt_tkep = sqrt(tkep(k,i,j))  ! Square root of previous SGS TKE value

         if (vt3dj(k,i,j) < eps_f4) then         ! neutral or unstable
            mixing_scale = (scale_h_sq*scale_v)**0.3333333_f4
         else                                    ! stable
            ! As suggested by Deardorff (1980):
            mixing_scale = 0.76_f4*sqrt_tkep/sqrt(vt3dj(k,i,j))
         end if
 
         ! At k=2, ensure that the mixing_scale remains less than or equal to
         !  the value from M-O similarity theory: 
         if (k == 2) mixing_scale = min(mixing_scale, vonk*c_gd%zt(k)*rtgt(i,j))

         ! Turbulent eddy viscosity of momentum ("nu_t"):
         nu_m = 0.1_f4*mixing_scale*sqrt_tkep
         ! Turbulent eddy viscosity of scalars ("nu_theta"):
         nu_s = (ONE_f4+2._f4*mixing_scale/avg_mesh_spacing)*nu_m

         mag_rate_of_strain_tensor = sqrt(vt3dh(k,i,j))

         ! SGS 3D stress [rho*(ui'uj')_bar] for momentum:
         tau_m = -2._f4*nu_m*mag_rate_of_strain_tensor

         
!         shear_production_term = 




         ftke = 0.1_f4*sqrt_tkep    ! TKE factor for K calcs

         scr1(k,i,j) = sclv*ftke        ! vkm - Vertical momentum K

         scla = (sclh*sclh*sclv)**0.333333_f4    ! "Average" 3D length scale

!         vctr5(k) = 
!         vctr5(k) = scla*ftke*vt3dh(k,i,j)  &    ! Shear prod term
!              -0.333333_f4*tkep(k,i,j)  &                 ! Turbulent transport
!              *(vt3da(k,i,j)+vt3dc(k,i,j)+scr2(k,i,j)) !   term

         ! vkh = vkm * (turbulent Prandtl number)
         !   (vkh - Vertical thermodynamic K)
         vt3dh(k,i,j) = scr1(k,i,j)*(ONE_f4+2._f4*sclv/scluv)

         scr2(k,i,j) = sclh*ftke     ! hkm - Horizontal momentum K

! now, vt3dh is vkh and scr2 is hkm

         ! A "constant" in the dissipation term (here proportional to ratio of
         !   length scales)
         ! A recent (2000) ARPS paper suggests that this should be = 0.93
         !   everywhere except the lowest model level, where = 3.9 
!combined         vctr26(k) = 0.19+  &          ! Original (Deardorff 1980)  
!combined              0.51*(((sclh*sclh*sclv)/(scluh2*scluv))**0.333333)
        ! Dissipation term:
!         vctr8(k) = vctr26(k) * tkep(k,i,j) * sqrt_tkep / scl
!combined         vctr8(k) = vctr26(k)*tkep(k,i,j)*sqrt_tkep/(sclh*sclh*sclv)**0.333333
!         vctr8(k) = 0.93*tkep(k,i,j)*sqrt_tkep/(sclh*sclh*sclv)**0.333333

! Combined form of dissipation term
         vctr8(k) = tkep(k,i,j)*sqrt_tkep*(0.19_f4/scla+  &
              0.51_f4/(scluh2*scluv)**0.333333_f4)

         vctr7(k) = -vt3dh(k,i,j)*vt3dj(k,i,j)   ! Buoyancy term?
 
         tket(k,i,j) = tket(k,i,j)+vctr5(k)+vctr7(k)-vctr8(k)  ! TKE eq.

         vt3dh(k,i,j) = vt3dh(k,i,j)*dn0(k,i,j)
         scr1(k,i,j) = scr1(k,i,j)*dn0(k,i,j)
         scr2(k,i,j) = scr2(k,i,j)*dn0(k,i,j)

      enddo

      ! TKE at lowest level (note dissipation term factor change):

!no surface linkage  
!      tket(2,i,j) = tket2+vctr5(3)+vctr7(2)-(vctr8(2)*3.9)
     
      tket(2,i,j) = tket2  &     ! Past TKE tendency
           +vctr5(3)  &    ! Shear production/turb transport
           +(tfz(i,j)*g_Xoid/theta(2,i,j))  &   ! Buoyancy term
           -(vctr8(2)*3.9_f4)      ! Dissipation term
     
   enddo
enddo

end subroutine SGS_Sullivan_etal_1994


!```````````````````````````````````````````````````````````````````
subroutine MYNN2009(m1, m2, m3, m5, ia, iz, ja, jz, ibndc, jd, tkep,  &
     tket, vt3dh, vt3di, vt3dj, scr1, rtgt, theta, dn0, up, vp, wp,  &
     upwp_sfc, vpwp_sfc, wpwp_sfc, wptp_sfc, vt3dp, ustar, tempk_sfc, pi0,  &
     pp, MY_level, use_K2010_mods)

!+ Calculates turblent eddy viscosities (K factors) based on the level 2 or
!   level 2.5 scheme of Mellor-Yamada (1982), as modified by Nakanishi and
!   Niino (2009), with additional modifications by Kitamura (2010).

use univ_kind_defs_mod, only: f4, f8
use univ_math_constants_mod, only: ONE_f4, ONE_THIRD_f4, ZERO_f4, ONE_HALF_f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL
use main_memory_mod, only: tstep, c_gd, i_grid
use body_specific_constants_mod, only: tkmin, vonk, g_Xoid, cpi

implicit none

!=== Argument declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'MYNN2009'

integer, intent(IN) :: m1, m2, m3, m5, ia, iz, ja, jz, ibndc, jd, MY_level
real(f4), dimension(m1,m2,m3), intent(IN) :: tkep,  & ! [m^2/s^2]
                                             theta,  & ! [K]
                                             dn0,  & ! [kg/m^3]
                                             up, vp, wp,  & ! [m/s]
                                             pp, pi0 ! [J/(kg.K)]
real(f4), dimension(m2,m3), intent(IN) :: rtgt,  &
                                    upwp_sfc, vpwp_sfc, wpwp_sfc,  & ! [m^2/s^2]
                                          wptp_sfc ! [(K.m)/s]
real(f4), dimension(m2,m3,m5), intent(IN) :: ustar,  & ! [m/s]
                                             tempk_sfc ! [K]
logical, intent(IN) :: use_K2010_mods

real(f4), dimension(m1,m2,m3), intent(INOUT) :: vt3dj,  & ! IN:[s^-2]
                                                scr1, vt3dh,  & ! OUT:[kg/(m.s)]
                                                tket,  & ! INOUT[m^2/s^3]
                                           vt3di,  & ! IN:[s^-2], OUT:[kg/(m.s)]
                                                vt3dp

!== Local declarations:

! From Mellor and Yamada (1982); N2001 Eq(9):
!real(f4), parameter :: A1 = 0.92_f4, A2 = 0.74_f4, B1 = 16.6_f4,  &
!     B2 = 10.1_f4, C1 = 0.08_f4, C2 = ZERO_f4, C3 = ZERO_f4, C4 = ZERO_f4,  &
!     C5 = ZERO_f4

! From Nakanishi (2001) Eq(37):
!real(f4), parameter :: A1 = 1.18_f4, A2 = 0.665_f4, B1 = 24.0_f4,  &
!     B2 = 15._f4, C1 = 0.137, C2 = 0.65_f4, C3 = 0.294_f4, C4 = ZERO_f4,  &
!     C5 = 0.2_f4

! From Nakanishi and Niino (2004) Eq(21a,21b):
!real(f4), parameter :: A1 = 1.18_f4, A2 = 0.665_f4, B1 = 24.0_f4,  &
!     B2 = 15._f4, C1 = 0.137, C2 = 0.7_f4, C3 = 0.323_f4, C4 = ZERO_f4,  &
!     C5 = 0.2_f4 

! From Nakanishi and Niino (2009) Eq(66):
real(f4), parameter :: A1 = 1.18_f4, A2 = 0.665_f4, B1 = 24.0_f4,  &
     B2 = 15._f4, C1 = 0.137, C2 = 0.75_f4, C3 = 0.352_f4, C4 = ZERO_f4,  &
     C5 = 0.2_f4 

real(f4), parameter :: gamma1 = ONE_THIRD_f4-2._f4*A1/B1
real(f4), parameter ::  &
     gamma2 = (2._f4*A1*(3._f4-2._f4*C2)+B2*(ONE_f4-C3))/B1,  & ! NN2009 Eq(A5)
     F1 = B1*(gamma1-C1)+2._f4*A1*(3._f4-2._f4*C2)+  &          ! NN2009 Eq(A6)
          3._f4*A2*(ONE_f4-C2)*(ONE_f4-C5),  &                  !
     F2 = B1*(gamma1+gamma2)-3._f4*A1*(ONE_f4-C2),  &           ! NN2009 Eq(A7)
     Rf1 = B1*(gamma1-C1)/F1,  &                                ! NN2009 Eq(A8)
     Rf2 = B1*gamma1/F2                                         ! NN2009 Eq(A9)
real(f4), parameter ::  &
     Ri1 = ONE_HALF_f4*A2*F2/(A1*F1),  &   ! N2009 Eq(A12)
     Ri2 = ONE_HALF_f4*Rf1/Ri1,  &         ! N2009 Eq(A13)
     Ri3 = (2._f4*Rf2-Rf1)/Ri1             ! N2009 Eq(A14)
real(f4), parameter ::  &
     Rfc = gamma1/(gamma1+gamma2)     ! Critical flux Ri [NN2009 Eq(A10)]

! Small (but not too small) value to use in clipping variable values to avoid
!  underflow/overflow errors:
real(f4), parameter :: clip_minv = 1.e-9_f4

integer :: i, j, k
real(f4) :: deltaz_t, dissipation_TKE, dtheta_dz, D_L25, P1, P2, P3, P4, P5,  &
     factor, G_h, G_m, L_Obukhov, L_s, L_t, L_b, L_t_inv, N, P_b, P_s,  &
     qc, qdz, qq_L2, Rf, Ri, S_h_L2, S_m_L2, S_h_L25, S_m_L25, alpha_c,  &
     tket_input, zeta, beta_theta, alpha_c_sq, S_q, beta1, beta2, beta3,  &
     beta4, E1, E2, E3, E4, R1, R2, theta_sfc, denom, ONE_plus_Ri
real(f8) :: sum1, sum2     

real(f4), dimension(m1) :: z_t, q, en_sq, L

!^^^^^^^^^^
!integer, parameter :: i_out = 3, j_out = 1, grid_out = 1, unit_out = 34
!character(len=256) :: fn_out
!real(f4) :: dbg_vt3di
!^^^^^^^^^^

!=== Executable statements:

! The algorithm below uses several surface layer quantities, so enforcing
!  NSTBOT=1 is the safest way to proceed:
if (c_gd%nstbot /= 1) call exception(procedure_name, FATAL,  &
     'Will only work correctly with NSTBOT=1')

! Definition(s) in this context:
!   SGS_TKE = 0.5*(q^2) = 'tkep'

do j=ja,jz 
   do i=ia,iz
      tket_input = tket(2,i,j)  ! [m^2/s^3]; save value of k=2 SGS_TKE tendency

      z_t(:) = c_gd%zt(:)*rtgt(i,j) ! [m]

      denom = vonk*g_Xoid*wptp_sfc(i,j)                     ! [(K.m^2)/s^3];
      denom = sign(ONE_f4, denom)*max(abs(denom), clip_minv)  !  avoid zero
      L_Obukhov = -theta(2,i,j)*(ustar(i,j,1)**3)/denom     ! [m]; N2001 Eq(4)

      sum1 = ZERO_f4; sum2 = ZERO_f4
      do k=2,m1-1
         q(k) = sqrt(2._f4*max(tkep(k,i,j), tkmin)) ! [m/s]
         deltaz_t = (c_gd%zm(k)-c_gd%zm(k-1))*rtgt(i,j) ! [m]
         qdz = q(k)*deltaz_t ! [m^2/s]
         sum1 = sum1+real(qdz*z_t(k), kind=f8) ! [m^3/s]
         sum2 = sum2+real(qdz, kind=f8) ! [m^2/s]
         en_sq(k) = vt3dj(k,i,j) ! [s^-2]
      end do

!%      en_sq(2) = en_sq(3) ! Better approx than k=3 to k=1 gradients
      ! Using ~dtheta/dz between z=3 and z="1.55":
      theta_sfc = tempk_sfc(i,j,1)*cpi*  & ! [K]
           ONE_HALF_f4*(pp(1,i,j)+pp(2,i,j)+pi0(1,i,j)+pi0(2,i,j))
      en_sq(2) = g_Xoid/theta(2,i,j)*  & ! [s^-2]
           (theta(3,i,j)-(0.9_f4*theta(2,i,j)+0.1_f4*theta_sfc))/  &
           (z_t(3)-0.1_f4*z_t(2))

      L_t = 0.23_f4*real(sum1/sum2, kind=f4)           ! [m]; NN2009 Eq(54)
   
      if (L_Obukhov < ZERO_f4) then  ! Only needed when this condition is true
         qc = (g_Xoid/theta(2,i,j)*  &                 ! [m/s]; NN2004 pg(26),
              max(wptp_sfc(i,j), clip_minv)*L_t)**ONE_THIRD_f4 ! keep positive
      else
         qc = clip_minv   ! Something near zero, for completeness
      end if

      L_t_inv = ONE_f4/L_t ! [m^-1]

!^^^^^^^^^^
!      if (i_out == i .and. j_out == j .and. grid_out == i_grid .and.  &
!           mod(tstep, 100) == 0) then
!         write (fn_out, '(a,i6.6,a4)') 'MYNN-j-', tstep, '.hdr'
!         open (unit_out, file=fn_out(1:len_trim(fn_out)), status='unknown',  &
!              form='formatted')
!         write (unit_out, '(a)') 'th_k2  u*  wptp  L_O  sum1  sum2  L_t  qc'
!         write (unit_out, *) theta(2,i,j), ustar(i,j,1), wptp_sfc(i,j),  &
!              L_Obukhov, sum1, sum2, L_t, qc 
!         write (unit_out, '(3a)')  &
!              'lev  z_t  tkep  theta  q  zeta  L_s  N^2  dth_dz  L_b  L  ',  &
!              'Ri  Rf  S_h_L2  S_m_L2  qq_L2  alpha_c  G_m  G_h  S_h_L25  ',  &
!              'S_m_L25  dn0  P_s  P_b  diss_TKE  tket  tau_q  tau_m  tau_h dbg_vt3di'
!         close (unit_out)
!
!         write (fn_out, '(a,i6.6,a4)') 'MYNN-j-', tstep, '.dat'
!         open (unit_out, file=fn_out(1:len_trim(fn_out)), status='unknown',  &
!              form='formatted')
!      end if
!^^^^^^^^^^

      do k=2,m1-1
         beta_theta = ONE_f4   ! No vapor; cf. NN2009 Eq(B9)

         zeta = z_t(k)/L_Obukhov ! [none]

      ! Determine "master" mixing length using the scheme of Nakanishi and Niino
      !  (2009):
         if (zeta >= ONE_f4) then                           !
            L_s = vonk*z_t(k)/3.7_f4                        !
         else if (zeta < ZERO_f4) then                      !
            L_s = vonk*z_t(k)*(ONE_f4-100._f4*zeta)**0.2_f4 ! [m]; NN2009 Eq(53)
         else   ! 0. <= zeta < 1.                           !
            L_s = vonk*z_t(k)/(ONE_f4+2.7*zeta)             !
         end if                                             !

         dtheta_dz = en_sq(k)*theta(k,i,j)/g_Xoid ! [K/m]
         if (dtheta_dz > ZERO_f4) then                        !
            N = sqrt(max(clip_minv, en_sq(k))) ! [s^-1]; No sqrt(-)   !
            if (zeta >= ZERO_f4) then                         !
               L_b = q(k)/N                                   !
            else                                              ! [m];
               L_b = q(k)*(ONE_f4+5._f4*sqrt(qc*L_t_inv/N))/N !  NN2009 Eq(55)
            end if                                            !       
         else                                                 !
            L_b = ONE_f4/clip_minv     ! "Infinity"           !
         end if                                               !

         L(k) = ONE_f4/(ONE_f4/L_s+L_t_inv+ONE_f4/L_b)      ! [m]; NN2009 Eq(52)

         ! Gradient Richardson number (with evasion of divide-by-zero):
         Ri = en_sq(k)/max(vt3di(k,i,j), clip_minv) ! [none]; cf. NN2009 pg(910)

!^^^^^^^^^^
!         if (i_out == i .and. j_out == j .and. grid_out == i_grid .and.  &
!              mod(tstep, 100) == 0) then
!            dbg_vt3di = vt3di(k,i,j)
!         end if
!^^^^^^^^^^

         ! Coefficient used in Kitamura (2010) modifications:
         if (use_K2010_mods) then
            ONE_plus_Ri = ONE_f4+Ri ! [none]
         else
            ONE_plus_Ri = ONE_f4 ! [none]
         end if

      ! Compute "Level 2" mixing scales in order to diagnose the "Level 2" TKE,
      !  which is needed to implement the Helfand and Labraga (1988)
      !  modification for growing turbulence:
         if (Ri > ZERO_f4) then   ! "stable"; use Kitamura (2010) modifications
            beta1 = A1*B1*ONE_plus_Ri*(gamma1-C1)            ! K2010 Eq(26)
            beta2 = A1*(B1*ONE_plus_Ri*(gamma1-C1)+  &       !
                 2._f4*A1*ONE_plus_Ri*(3._f4-2._f4*C2)+  &   ! K2010 Eq(27)
                 3._f4*A2*(ONE_f4-C2)*(ONE_f4-C5))           !
            beta3 = A2*B1*gamma1                        ! K2010 Eq(28)
            beta4 = A2*(B1*(gamma1+gamma2)-3._f4*A1*(ONE_f4-C2))  ! K2010 Eq(29)

            Rf = ONE_HALF_f4/beta2*((beta4*Ri+beta1)-  &       ! K2010 Eq(31)
                 sqrt(max(clip_minv,  &                        !
                 ((beta4*Ri+beta1)**2)-4._f4*beta2*beta3*Ri))) !

            denom = ONE_plus_Ri*(ONE_f4-Rf)                       !
            denom = sign(ONE_f4, denom)*max(abs(denom), clip_minv)!  avoid zero
            S_h_L2 = 3._f4*A2*(gamma1+gamma2)*(Rfc-Rf)/denom      ! K2010 Eq(21)
            S_h_L2 = max(S_h_L2, clip_minv)
            denom = (beta3-beta4*Rf)                              !
            denom = sign(ONE_f4, denom)*max(abs(denom), clip_minv)!  avoid zero
            S_m_L2 = (beta1-beta2*Rf)*S_h_L2/denom                ! K2010 Eq(22)
            S_m_L2 = max(S_m_L2, clip_minv)
         else                      ! "unstable"; use original NN2009 scheme
            ! Flux Ri [NN2009 Eq(A11)]:
            Rf = Ri1*(Ri+Ri2-sqrt(max(Ri*Ri-Ri3*Ri+Ri2*Ri2, clip_minv)))

            denom = ONE_f4-Rf                                    !
            denom = sign(ONE_f4, denom)*max(abs(denom), clip_minv) ! avoid zero
            S_h_L2 = 3._f4*A2*(gamma1+gamma2)*(Rfc-Rf)/denom     ! NN2009 Eq(A4)
            S_h_L2 = max(S_h_L2, clip_minv)
            denom = A2*F2*(Rf2-Rf)                               !
            denom = sign(ONE_f4, denom)*max(abs(denom), clip_minv) ! avoid zero
            S_m_L2 = A1*F1*(Rf1-Rf)*S_h_L2/denom                 ! NN2009 Eq(A3)
            S_m_L2 = max(S_m_L2, clip_minv)                      
         end if

         ! NN2009 Eq(A2):
         !  NOTE: The modified algorithm here supplements the vt3di value
         !        { ((d(Vh)/dz)^2 } if it is positive and very small or zero
         !        (equivalent to d(Vh)/dz ~= 0.032 m/s per km).  This seems to
         !        fix an issue where reasonable amounts of parameterized
         !        turbulence were not realized when/where the vertical change of
         !        the spatial-mean horizontal wind was zero or ver small (even
         !        under very statically unstable conditions!).        
         denom = vt3di(k,i,j) ! [s^-2]
         if (denom >= ZERO_f4) denom = max(denom, 1.e-9_f4) ! [s^-2]
         qq_L2 = max(B1*L(k)*L(k)*S_m_L2*(ONE_f4-Rf)*denom,  & ! [m^2/s^2]
              2._f4*tkmin)

         if (MY_level == 2) then   ! Finalize MY (Level 2) K values, if desired
            factor = L(k)*sqrt(qq_L2) ! [m^2/s]
            scr1(k,i,j) = factor*S_m_L2 * dn0(k,i,j)  ! Convert K's to stresses
            vt3dh(k,i,j) = factor*S_h_L2 * dn0(k,i,j) !  [kg/(m.s)]
            cycle  ! No need to do more for this layer, skip to the next layer
         end if
       
         ! Helfand and Labraga (1988) adjustment for MY L2.5(+) [NN2009 Eq(42)]:
         if (tkep(k,i,j) < ONE_HALF_f4*qq_L2) then
            alpha_c = q(k)/sqrt(qq_L2)   ! [none]; Growing turbulence
         else
            alpha_c = ONE_f4             ! [none]; Static/decaying turbulence
         end if
         alpha_c_sq = alpha_c*alpha_c ! [none]

      ! Compute "Level 2.5" mixing scales, possibly as a part of computing the
      !  "Level 3" mixing scales:

         G_m = L(k)*L(k)/(q(k)*q(k))*vt3di(k,i,j)    ! [none]; NN2009 Eq(39)

         ! cf. NN2009 Eq(40):
         G_h = -L(k)*L(k)/(q(k)*q(k))*(beta_theta*en_sq(k))  ! [none]; No vapor

         if (Ri > ZERO_f4) then  ! "stable"; use Kitamura (2010) modifications
            E1 = ONE_plus_Ri+alpha_c_sq*3._f4*A1*  &               ! K2010 Eq(5)
                 (2._f4*A1*ONE_plus_Ri*G_m-3._f4*A2*(ONE_f4-C2)*G_h) !
            E2 = -alpha_c_sq*3._f4*A1*  &                          ! K2010 Eq(6)
              (4._f4*A1*ONE_plus_Ri+3._f4*A2*(ONE_f4-C5))*(ONE_f4-C2)*G_h !
            E3 = alpha_c_sq*6._f4*A1*A2*G_m              ! K2010 Eq(7)
            E4 = ONE_plus_Ri-alpha_c_sq*12._f4*A1*A2*(ONE_f4-C2)*G_h-  & !
                 alpha_c_sq*3._f4*A2*B2*(ONE_f4-C3)*G_h            ! K2010 Eq(8)
            R1 = alpha_c*A1*(ONE_f4-3._f4*C1)*ONE_plus_Ri  ! K2010 Eq(9)
            R2 = alpha_c*A2                   ! K2010 Eq(10)

            D_L25 = E1*E4-E2*E3
            D_L25 = sign(ONE_f4, D_L25)*max(abs(D_L25), clip_minv) ! avoid zero

            S_m_L25 = (R1*E4-R2*E2)/D_L25    ! K2010 Eq(3)
            S_m_L25 = max(S_m_L25, clip_minv)
            S_h_L25 = (R2*E1-R1*E3)/D_L25    ! K2010 Eq(4)
            S_h_L25 = max(S_h_L25, clip_minv)
         else                          ! "unstable"; use original NN2009 scheme
            P1 = ONE_f4-3._f4*alpha_c_sq*A2*B2*(ONE_f4-C3)*G_h   ! NN2009 Eq(33)
            P2 = ONE_f4-9._f4*alpha_c_sq*A1*A2*(ONE_f4-C2)*G_h   ! NN2009 Eq(34)
            P3 = P1+9._f4*alpha_c_sq*A2*A2*  &     ! NN2009 Eq(35)
                 (ONE_f4-C2)*(ONE_f4-C5)*G_h       !
            P4 = P1-12._f4*alpha_c_sq*A1*A2*(ONE_f4-C2)*G_h  ! NN2009 Eq(36)
            P5 = 6._f4*alpha_c_sq*A1*A1*G_m      ! NN2009 Eq(37)

            D_L25 = P2*P4+P5*P3            ! NN2009 Eq(31)
            D_L25 = sign(ONE_f4, D_L25)*max(abs(D_L25), clip_minv) ! avoid zero
     
            S_m_L25 = alpha_c*A1*(P3-3._f4*C1*P4)/D_L25       ! NN2009 Eq(27)
            S_m_L25 = max(S_m_L25, clip_minv)   ! see HL1988 appendix
            S_h_L25 = alpha_c*A2*(P2+3._f4*C1*P5)/D_L25       ! NN2009 Eq(28)
            S_h_L25 = max(S_h_L25, clip_minv)   ! see HL1988 appendix
         end if

         if (MY_level == 25) then  ! Finalize MY (Level 2.5) K values, TKE/dt
            factor = L(k)*q(k) ! [m^2/s]
            scr1(k,i,j) = factor*S_m_L25    ! No air density factor, YET
            vt3dh(k,i,j) = factor*S_h_L25   !  [m^2/s]

            P_s = scr1(k,i,j)*vt3di(k,i,j)   ! [m^2/s^3]; MY1982 Eq(25a)

            ! MY1982 Eq(25b), with beta=1; also, implicitly divided by theta
            ! for unit consistency with other terms:
            P_b = -vt3dh(k,i,j)*en_sq(k)  ! [m^2/s^3]

            dissipation_TKE = (q(k)**3)/  &  ! MY1982 Eq(25c), NN2004 Eq(9a)
                 (B1*L(k))                   !  [m^2/s^3]

            ! MY1982 Eq(24), excluding diffusion/advection:
            tket(k,i,j) = tket(k,i,j)+P_s+P_b-dissipation_TKE ! [m^2/s^3]
            
            ! From MY1982 Eq(24), TKE diffusion portion:
!%            S_q = 0.2_f4    ! MY1982 value
            S_q = 3._f4*S_m_L25    ! [none]; cf. NN2009 Eq(67)
            vt3di(k,i,j) = factor*S_q*dn0(k,i,j) ! [kg/(m.s)]; Conv. K to stress
            
            scr1(k,i,j) = scr1(k,i,j)*dn0(k,i,j)   ! [kg/(m.s)]; Conv. to stress
            vt3dh(k,i,j) = vt3dh(k,i,j)*dn0(k,i,j) !

!^^^^^^^^^^^^ Comment out if writing scheme calculation info to disk:
            cycle
!^^^^^^^^^^^^
         end if

!^^^^^^^^^^^^
!         if (i_out == i .and. j_out == j .and. grid_out == i_grid .and.  &
!              mod(tstep, 100) == 0) then
!            write (unit_out, *) k, z_t(k), tkep(k,i,j), theta(k,i,j), q(k),  &
!                 zeta, L_s, en_sq(k), dtheta_dz, L_b, L(k), Ri, Rf, S_h_L2,  &
!                 S_m_L2, qq_L2, alpha_c, G_m, G_h, S_h_L25, S_m_L25,  &
!                 dn0(k,i,j), P_s, P_b, dissipation_TKE, tket(k,i,j),  &
!                 vt3di(k,i,j), scr1(k,i,j), vt3dh(k,i,j), dbg_vt3di
!         end if
!^^^^^^^^^^^^
      end do

   ! Directly use surface layer fields to set tket(2):
   ! **WARNING/NOTE**: This seems to create significantly more noise in the
   !                   SGS TKE fields at k=2,3; unsure if this ought to be
   !                   here, in the diffusion routine (like wptp is), or is
   !                   somehow not appropriate at all.

!%      P_s = -(upwp_sfc(i,j)*(up(2,i,j)+up(2,i-1,j))+  &    ! MY1982 Eq(25a)
!%           vpwp_sfc(i,j)*(vp(2,i,j)+vp(2,i,j-jd))+  &      !
!%           wpwp_sfc(i,j)*wp(2,i,j))*c_gd%dzt(2)/rtgt(i,j)       !

      ! MY1982 Eq(25b), with beta=1; also, explicitly divided by theta for
      ! unit consistency with other terms:
!%      P_b = wptp_sfc(i,j)*g_Xoid/theta(2,i,j)

!%      dissipation_TKE = (q(2)**3)/(B1*L(2))  ! MY1982 Eq(25c), NN2004 Eq(9a)

      ! MY1982 Eq(24), excluding diffusion/advection:
!%      tket(2,i,j) = tket_input+P_s+P_b-dissipation_TKE

!^^^^^^^^^^^^   
!      if (i_out == i .and. j_out == j .and. grid_out == i_grid .and.  &
!           mod(tstep, 100) == 0) close (unit_out)
!^^^^^^^^^^^^
   end do
end do

end subroutine MYNN2009


end module turbulent_K_coeffs_mod


!```````````````````````````````````````````````````````````````````
subroutine diffuse

!+ Driver to compute tendencies due to subgrid-scale turbulence

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4
use main_memory_mod, only: a, jdim, iscr1, iscr2, nscl, vctr14, naerscl,  &
     npatch, c_gd, i_grid, enable_turb_diffusion, mxp, myp, mzp,  &
     ibndc, ia, iz, ja, jz, ia_1, iz1, jz1, ja_1, my_rank, ia1, ja1, iz_1,  &
     jz_1, izu, jzv
use main_memory_bspec_mod, only: H2Ov_tog
use vtables_mod, only: idn0, ivt3dh, ihkm, ivkh, ivkm, ivt3di, ivt2da,  &
     isclrp, itfz, iqfz, iup, ivp, iwp, ivt3da, ivt3db, ivt3dc, ivt3dd,  &
     ivt3de, ivt3df, ivt3dg, ivt3dn, itheta, irv, ipp, ipi0, ivt3dj, irtgt,  &
     ivt3dk, idxt, idyt, itkep, itket, iuw, ivw, iwfz, iut, ivt, iwt, ivt3dl,  &
     ivt3dm, ivt3do, irtgu, irtgv, idn0v, idn0u, iscp, isct,  &
     iaerscp, iaersct, iaerfi, ivt2db, ivt3dp, iustar, isfctempk, RV_vidx,  &
     THP_vidx, TKEP_vidx
use aermic_rad_mem_mod, only: aer_XOdiffuse, aer_tyet
use turbulent_K_coeffs_mod, only: brunt_vaisalla, mxdefm, mxtked, klbnd,  &
     MYNN2009
use grid_mem_mod, only: switch_to_Ascalar, switch_to_Nscalar, a2e3

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

integer :: mxyzp1, ind, n, ivkkh, ihkkh, idx_r
real(f4) :: s1, s2, s3
logical :: need_to_init

!=== Executable statements:

if (.not. enable_turb_diffusion) return

mxyzp1 = mxp*myp*mzp-1

call strain(mzp, mxp, myp, ia, iz, ja, jz, ia_1, ja_1, iz1, jz1, jdim,  &
     a(iup), a(ivp), a(iwp), a(ivt3da), a(ivt3db), a(ivt3dc), a(ivt3dd),  &
     a(ivt3de), a(ivt3df), a(ivt3dg), a(ivt3dh), a(ivt3di), a(ivt3dn),  &
     a(iscr2), c_gd%idiffk)
!== DIFFK = 1,2,6:9:
! vt3da: du/dx  |  vt3dn: du/dy  |  vt3dd: du/dz     [s^-1]
! vt3db: dv/dx  |  vt3dc: dv/dy  |  vt3de: dv/dz     [s^-1]
! vt3df: dw/dx  |  vt3dg: dw/dy  |  scr2: SCRATCH    [s^-1]
! vt3dh: 2*[(du/dx)^2+(dv/dy)^2]+[(dv/dx)_T+(du/dy)_T]^2, [s^-2]
! vt3di: [(du/dz)_T+(dv/dz)_T]^2, [s^-2]
!== DIFFK = 3,4 :
! vt3da: 2*(du/dx)  |  vt3dn: dv/dx  |  vt3dd: (du/dz)+(dw/dx)    [s^-1]
! vt3db: (dv/dx)+(du/dy)  |  vt3dc: 2*(dv/dy)  |  vt3de: (dv/dz)+(dw/dy)  [s^-1]
! vt3df: dw/dx  |  vt3dg: dw/dy  |  scr2: 2*(dw/dz)     [s^-1]
! vt3dh: 0.5*[[2*(du/dx)-2*delV]^2+[2*(dv/dy)-2*delV]^2+[2*(dw/dz)-2*delV]^2]+
!        [(dv/dx+du/dy)_T]^2+[(du/dz+dw/dx)_T]^2+[(dv/dz+dw/dy)_T]^2
!        = (full 3D rate-of-strain tensor MAGNITUDE)**2, [s^-2]
! vt3di: vt3dh, [s^-2]
!== DIFFK = 5 :
! *NOT YET IMPLEMENTED*

if (H2Ov_tog > 0) then
   idx_r = irv
else
   idx_r = iscr1   ! "Dummy"
end if

! ivt3dj: N^2, [s^-2]
call brunt_vaisalla(mzp, mxp, myp, ia, iz, ja, jz, a(itheta), a(idx_r),  &
     a(iscr1), a(ipp), a(ipi0), a(ivt3dj), a(irtgt))

! Compute simple deformation-based horizontal SGS turbulent stresses, and for
!  some cases (IDIFFK=2,3), vertical SGS turbulent stresses:
if (c_gd%idiffk <= 3 .or. (c_gd%idiffk >= 6 .and. c_gd%idiffk <= 9)) then
   ! Compute deformation K coeffs:
   call mxdefm(mzp, mxp, myp, ia, iz, ja, jz, ibndc, jdim, a(ivt3dh),  &
        a(ivt3di), a(ivt3dj), a(ivt3dk), a(iscr1), a(iscr2), a(idn0),  &
        a(irtgt), a(idxt), a(idyt), my_rank)
   ! vt3dh: (IDIFFK=1,6:9) SCRATCH
   !        (IDIFFK=2,3) (vertical heat K) * dn0, [kg/(m.s)]
   ! vt3dk: (IDIFFK=1,6:9) SCRATCH
   !        (IDIFFK=2) Ri, [none]
   !        (IDIFFK=3) (N^2)/vt3di, [none]
   ! scr1: (IDIFFK=1,6:9) SCRATCH
   !       (IDIFFK=2,3) (vertical momentum K) * dn0, [kg/(m.s)]
   ! scr2: (horizontal momentum K) * dn0, [kg/(m.s)]
end if

! Compute vertical SGS turbulent stresses (not simple deformation-based):
select case (c_gd%idiffk)
case (1)   ! "Original" Mellor-Yamada Level 2.5 scheme (thought to be BROKEN)
   call tkemy(mzp, mxp, myp, ia, iz, ja, jz, ibndc, jdim, a(itkep), a(itket),  &
        a(ivt3dh), a(ivt3di), a(ivt3dj), a(iscr1), a(irtgt), a(itheta),  &
        a(idn0), a(iup), a(ivp), a(iwp), a(iuw), a(ivw), a(iwfz), a(itfz),  &
        vctr14)
   ! vt3dh: (vertical heat K) * dn0, [kg/(m.s)]
   ! vt3di: (vertical TKE K) * dn0, [kg/(m.s)]
   ! scr1: (vertical momentum K) * dn0, [kg/(m.s)]
case (4)  ! "Original"  LES SGS (cf. Deardorff, 1980) (thought to be BROKEN)
   call mxtked(mzp, mxp, myp, ia, iz, ja, jz, ibndc, jdim, a(itkep),  &
        a(itket), a(iup), a(ivp), a(iwp), a(idx_r), a(itheta), a(ivt3da),  &
        a(ivt3dc), a(ivt3dh), a(ivt3dj), a(iscr1), a(iscr2), a(iuw), a(ivw),  &
        a(iwfz), a(itfz), a(idxt), a(irtgt), a(idn0))
   ! vt3dh: (vertical heat K)   [* dn0]????
   ! vt3di: unchanged since `strain` calculation above   [* dn0]????
   ! scr1: (vertical momentum K) [* dn0]????
   ! scr2: (horizontal momentum K) [* dn0]????
case (5)  ! Sullivan et al. (1994) LES SGS scheme (UNFINISHED)
   write (*, '(a)') 'NOT YET FINISHED'
!   call mxtked(mzp,mxp,myp,ia,iz,ja,jz,ibndc,jdim,a(itkep),a(itket),a(iup)  &
!        ,a(ivp),a(iwp),a(idx_r),a(itheta),a(ivt3da),a(ivt3dc)  &
!        ,a(ivt3dh),a(ivt3dj),a(iscr1),a(iscr2),a(iuw),a(ivw),a(iwfz)  &
!        ,a(itfz),a(idxt),a(irtgt), a(idn0))
   ! vt3dh: (vertical heat K)   [* dn0]????
   ! vt3di: unchanged since `strain` calculation above   [* dn0]????
   ! scr1: (vertical momentum K) [* dn0]????
   ! scr2: (horizontal momentum K) [* dn0]????
case (6)      ! Mellor-Yamada Level 2 scheme, MYNN2009 with K2010 mods
   call MYNN2009(mzp, mxp, myp, npatch, ia, iz, ja, jz, ibndc, jdim,  &
        a(itkep), a(itket), a(ivt3dh), a(ivt3di), a(ivt3dj), a(iscr1),  &
        a(irtgt), a(itheta), a(idn0), a(iup), a(ivp), a(iwp), a(iuw),  &
        a(ivw), a(iwfz), a(itfz), a(ivt3dp), a(iustar), a(isfctempk),  &
        a(ipi0), a(ipp), 2, .TRUE.)
   ! vt3dh: (vertical heat K) * dn0, [kg/(m.s)]
   ! vt3di: SCRATCH  
   ! scr1: (vertical momentum K) * dn0, [kg/(m.s)]
case (7)      ! Mellor-Yamada Level 2.5 scheme, MYNN2009 with K2010 mods
   call MYNN2009(mzp, mxp, myp, npatch, ia, iz, ja, jz, ibndc, jdim,  &
        a(itkep), a(itket), a(ivt3dh), a(ivt3di), a(ivt3dj), a(iscr1),  &
        a(irtgt), a(itheta), a(idn0), a(iup), a(ivp), a(iwp), a(iuw),  &
        a(ivw), a(iwfz), a(itfz), a(ivt3dp), a(iustar), a(isfctempk),  &
        a(ipi0), a(ipp), 25, .TRUE.)  
   ! vt3dh: (vertical heat K) * dn0, [kg/(m.s)]
   ! vt3di: (vertical TKE K) * dn0, [kg/(m.s)]
   ! scr1: (vertical momentum K) * dn0, [kg/(m.s)]
case (8)      ! Mellor-Yamada Level 2 scheme, MYNN2009
   call MYNN2009(mzp, mxp, myp, npatch, ia, iz, ja, jz, ibndc, jdim,  &
        a(itkep), a(itket), a(ivt3dh), a(ivt3di), a(ivt3dj), a(iscr1),  &
        a(irtgt), a(itheta), a(idn0), a(iup), a(ivp), a(iwp), a(iuw),  &
        a(ivw), a(iwfz), a(itfz), a(ivt3dp), a(iustar), a(isfctempk),  &
        a(ipi0), a(ipp), 2, .FALSE.)
   ! vt3dh: (vertical heat K) * dn0, [kg/(m.s)]
   ! vt3di: SCRATCH  
   ! scr1: (vertical momentum K) * dn0, [kg/(m.s)]
case (9)      ! Mellor-Yamada Level 2.5 scheme, MYNN2009
   call MYNN2009(mzp, mxp, myp, npatch, ia, iz, ja, jz, ibndc, jdim,  &
        a(itkep), a(itket), a(ivt3dh), a(ivt3di), a(ivt3dj), a(iscr1),  &
        a(irtgt), a(itheta), a(idn0), a(iup), a(ivp), a(iwp), a(iuw),  &
        a(ivw), a(iwfz), a(itfz), a(ivt3dp), a(iustar), a(isfctempk),  &
        a(ipi0), a(ipp), 25, .FALSE.)  
   ! vt3dh: (vertical heat K) * dn0, [kg/(m.s)]
   ! vt3di: (vertical TKE K) * dn0, [kg/(m.s)]
   ! scr1: (vertical momentum K) * dn0, [kg/(m.s)]
end select

call klbnd(mzp, mxp, myp, ibndc, jdim, a(iscr1), a(idn0))
call klbnd(mzp, mxp, myp, ibndc, jdim, a(iscr2), a(idn0))
call klbnd(mzp, mxp, myp, ibndc, jdim, a(ivt3dh), a(idn0))
if (c_gd%idiffk == 1 .or. c_gd%idiffk == 4 .or. c_gd%idiffk == 5 .or.  &
     c_gd%idiffk == 7 .or. c_gd%idiffk == 9)  &
     call klbnd(mzp, mxp, myp, ibndc, jdim, a(ivt3di), a(idn0)) ! Vert TKE BCs

!bob  swap new hkm, vkm, and vkh with past time level:  lagged K's have
!bob  internal lateral boundary values from neighboring nodes

do ind = 0,mxyzp1
   s1 = a(iscr2 +ind)
   s2 = a(iscr1 +ind)
   s3 = a(ivt3dh+ind)
   a(iscr2 +ind) = a(ihkm+ind)
   a(iscr1 +ind) = a(ivkm+ind)
   a(ivt3dh+ind) = a(ivkh+ind)
   a(ihkm+ind) = s1
   a(ivkm+ind) = s2
   a(ivkh+ind) = s3
end do

! Diffuse velocities:

call diffvel(mzp, mxp, myp, ia, iz, ja, jz, jdim, ia_1, ja_1, ia1, ja1,  &
     iz_1, jz_1, iz1, jz1, izu, jzv, c_gd%idiffk, a(iup), a(ivp), a(iwp),  &
     a(iut), a(ivt), a(iwt), a(ivt3da), a(ivt3db), a(ivt3dc), a(ivt3dd),  &
     a(ivt3de), a(ivt3df), a(ivt3dg), a(ivt3dj), a(ivt3dk), a(ivt3dl),  &
     a(ivt3dm), a(ivt3dn), a(ivt3do), a(irtgu), a(irtgv), a(irtgt), a(iuw),  &
     a(ivw), a(iwfz), a(idn0), a(idn0u), a(idn0v), a(iscr1), a(iscr2), ibndc,  &
     my_rank)

select case (c_gd%idiffk)
case (1:3,6:9)
   do ind = 0,mxyzp1
      a(iscr2+ind) = a(iscr2+ind) * c_gd%xkhkm   ! Compute HKH from HKM
   ! vt3dh: (vertical heat K) * dn0, [kg/(m.s)]
   ! vt3di: (vertical TKE K) * dn0 (for IDIFFK = 1,7,9 only), [kg/(m.s)]
   ! scr2: (horizontal heat K) * dn0 , [kg/(m.s)]
   end do
case (4)
   do ind = 0,mxyzp1
      a(iscr2+ind) = a(ivt3dh+ind)
      a(ivt3di+ind) = 2._f4 * a(iscr1+ind)
   ! vt3dh: (vertical heat K)   [* dn0]????
   ! vt3di: 2*(vertical momentum K)  [* dn0]????
   ! scr2: horizontal heat K = vertical heat K   [* dn0]????
   end do
case (5)    ! UNFINISHED
   do ind = 0,mxyzp1
      a(iscr2+ind) = a(ivt3dh+ind)
      a(ivt3di+ind) = 2._f4 * a(iscr1+ind)
   ! vt3dh: (vertical heat K)   [* dn0]????
   ! vt3di: 2*(vertical momentum K)  [* dn0]????
   ! scr2: horizontal heat K = vertical heat K   [* dn0]????
   end do
end select

! Diffuse "normal" scalars:

do n=1,nscl
   call switch_to_Nscalar(i_grid, n)

   a(ivt2da:ivt2da+mxp*myp-1) = ZERO_f4
   if (c_gd%nstbot == 1) then
      if (isclrp(n,i_grid) == THP_vidx) then   ! for THP
         a(ivt2da:ivt2da+mxp*myp-1) = a(itfz:itfz+mxp*myp-1)
      else if (isclrp(n,i_grid) == RV_vidx) then   ! for RV
         a(ivt2da:ivt2da+mxp*myp-1) = a(iqfz:iqfz+mxp*myp-1)
      end if
   end if

   if (isclrp(n,i_grid) /= TKEP_vidx) then      ! for most scalars
      ivkkh = ivt3dh   ! (vertical heat K) * dn0
      ihkkh = iscr2   ! (horizontal heat K) * dn0 
   else                 ! for TKEP  (implicitly IDIFFK=1,4,5,7,9)
      ivkkh = ivt3di   ! (vertical TKE K) * dn0
      select case (c_gd%idiffk)
      case (1,7,9)
         ihkkh = iscr2    ! (horizontal heat K) * dn0
      case (4)
         ihkkh = ivt3di   ! ???? * dn0
      case (5)
         ihkkh = ivt3di   ! ???? * dn0    ! UNFINISHED
      end select
   end if

   ! Need to (re)initialize diffusion routine? 
   if (n == 1) then
      need_to_init = .TRUE.
   else if (isclrp(n,i_grid) == TKEP_vidx .or.  &
        isclrp(max(1,n-1),i_grid) == TKEP_vidx) then
      need_to_init = .TRUE.  ! TKE uses different VKH (switch back after TKE)
   else    ! Most "normal" scalars
      need_to_init = .FALSE.
   end if

   ! Diffuse scalar:
   if (isclrp(n,i_grid) /= RV_vidx) then   ! not RV 
      call diffsclr(mzp, mxp, myp, ia, iz, ja, jz, jdim, ia_1, ja_1, ia1,  &
           ja1, iz_1, jz_1, iz1, jz1, a(iscp), a(isct), a(ivt3da), a(ivt3db),  &
           a(ivt3df), a(ivt3dg), a(ivt3dj), a(ivt3dk), a(ivt3do), a(ivt3dc),  &
           a(irtgt), a(ivt2da), a(idn0), a(ivkkh), a(ihkkh), need_to_init,  &
           isclrp(n,i_grid))
   else   ! RV; convert to "mixing ratio" [Mg/kg]
      if (aer_tyet(0)) then
      a(iscp:iscp+mxyzp1) = a(iscp:iscp+mxyzp1)/a(idn0:idn0+mxyzp1)  ! [Mg/kg]
      call diffsclr_A(mzp, mxp, myp, ia, iz, ja, jz, jdim, ia_1, ja_1, ia1,  &
           ja1, iz_1, jz_1, iz1, jz1, a(iscp), a(isct), a(ivt3da), a(ivt3db),  &
           a(ivt3df), a(ivt3dg), a(ivt3dj), a(ivt3dk), a(ivt3do), a(ivt3dc),  &
           a(irtgt), a(ivt2da), a(idn0), a(ivkkh), a(ihkkh), need_to_init)
      a(iscp:iscp+mxyzp1) = a(iscp:iscp+mxyzp1)*a(idn0:idn0+mxyzp1) 
      end if
   end if
      
end do

! Diffuse "aerosol" scalars:

do n=1,naerscl

   !@@@@@
!   aer_XOdiffuse(n,i_grid) = .FALSE.
   !@@@@@

   if (aer_XOdiffuse(n,i_grid)) then   ! Only do if requested/required

      call switch_to_Ascalar(i_grid, n)

      a(ivt2da:ivt2da+mxp*myp-1) = ZERO_f4
      if (c_gd%nstbot == 1) then
         a(ivt2da:ivt2da+mxp*myp-1) = a(iaerfi:iaerfi+mxp*myp-1) ! Add inj. flux
         call a2e3(mzp, mxp, myp, a(idn0), a(ivt2db), 2)
         a(ivt2da:ivt2da+mxp*myp-1) = a(ivt2da:ivt2da+mxp*myp-1)/  &
              a(ivt2db:ivt2db+mxp*myp-1)  ! [Mg/kg]; use k=2 value
      end if
      
      ivkkh = ivt3dh   ! (vertical heat K) * dn0
      ihkkh = iscr2   ! (horizontal heat K) * dn0 

      ! Need to (re)initialize diffusion routine? 
      if (n == 1) then
         need_to_init = .TRUE.
      else    ! Most aerosol scalars
         need_to_init = .FALSE.
      end if

      ! Diffuse aerosol scalar:

      a(iaerscp:iaerscp+mxyzp1) = a(iaerscp:iaerscp+mxyzp1)/  &
           a(idn0:idn0+mxyzp1)
      call diffsclr_A(mzp, mxp, myp, ia, iz, ja, jz, jdim, ia_1, ja_1, ia1,  &
           ja1, iz_1, jz_1, iz1, jz1, a(iaerscp), a(iaersct), a(ivt3da),  &
           a(ivt3db), a(ivt3df), a(ivt3dg), a(ivt3dj), a(ivt3dk), a(ivt3do),  &
           a(ivt3dc), a(irtgt), a(ivt2da), a(idn0), a(ivkkh), a(ihkkh),  &
           need_to_init)
      a(iaerscp:iaerscp+mxyzp1) = a(iaerscp:iaerscp+mxyzp1)*  &
           a(idn0:idn0+mxyzp1)
   end if

end do

end subroutine diffuse


!```````````````````````````````````````````````````````````````````
subroutine strain(m1, m2, m3, ia, iz, ja, jz, ia_1, ja_1, iz1, jz1, jd, up,  &
     vp, wp, vt3da, vt3db, vt3dc, vt3dd, vt3de, vt3df, vt3dg, vt3dh, vt3di,  &
     vt3dn, scr2, idiffk)

!+ Compute fluid strain rates

use univ_kind_defs_mod, only: f4
use main_memory_mod, only: a, ihorgrad, c_gd, i_grid
use vtables_mod, only: irtgu, idxt, itopu, irtgv, itopv, idyt, idxm, idxu,  &
     itopt, irtgt, idym, idyv, itopm, irtgm
use PRAMS_turbdiff_mod, only: true_horiz_gradient
use PRAMS_graddiv_mod, only: gradient

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz, ia_1, ja_1, iz1, jz1, jd,  &
     idiffk
real(f4), dimension(m1,m2,m3), intent(IN) :: up, vp, wp

real(f4), dimension(m1,m2,m3), intent(INOUT) :: vt3da, vt3db, vt3dc, vt3dd,  &
     vt3de, vt3df, vt3dg, vt3dh, vt3di, vt3dn, scr2

!== Local declarations:

logical, parameter :: subtract_vert_gradient = .FALSE.

integer :: i, j, k

!=== Executable statements:

if (ihorgrad == 1) then

   call gradient(m1, m2, m3, ia, iz1, ja, jz, up, vt3da, 'XDIR', 'UPNT')    !
   call gradient(m1, m2, m3, ia_1, iz, ja_1, jz, vp, vt3db, 'XDIR', 'VPNT') ! X-direction
   call gradient(m1, m2, m3, ia_1, iz, ja, jz, wp, vt3df, 'XDIR', 'WPNT')   !

   call gradient(m1, m2, m3, ia_1, iz, ja_1, jz, up, vt3dn, 'YDIR', 'UPNT') !
   call gradient(m1, m2, m3, ia, iz, ja, jz1, vp, vt3dc, 'YDIR', 'VPNT')    ! Y-direction
   call gradient(m1, m2, m3, ia, iz, ja_1, jz, wp, vt3dg, 'YDIR', 'WPNT')   !
   
   call gradient(m1, m2, m3, ia_1, iz, ja, jz, up, vt3dd, 'ZDIR', 'UPNT')  !
   call gradient(m1, m2, m3, ia, iz, ja_1, jz, vp, vt3de, 'ZDIR', 'VPNT')  ! Z-direction
   if (idiffk == 3 .or. idiffk == 4 .or. idiffk == 5) then
      call gradient(m1, m2, m3, ia, iz, ja, jz, wp, scr2, 'ZDIR', 'WPNT')  !
   end if

else if (ihorgrad == 2) then

   ! Calculate vertical gradients (no significant sigma-z issues):

   call gradient(m1, m2, m3, ia_1, iz1, ja_1, jz1, up, vt3dd, 'ZDIR', 'UPNT')  !
   call gradient(m1, m2, m3, ia_1, iz1, ja_1, jz1, vp, vt3de, 'ZDIR', 'VPNT')  ! Z-dir
   if (subtract_vert_gradient .or. idiffk == 3 .or. idiffk == 4 .or.  &
        idiffk == 5) then
      call gradient(m1, m2, m3, ia_1, iz1, ja_1, jz1, wp, scr2, 'ZDIR', 'WPNT')   !
   end if

   ! Calculate x-dir gradients (sigma-z issues here, so don't use `gradient`):

   call true_horiz_gradient(m1, m2, m3, ia, iz1, ja, jz, up, vt3da, 'XDIR',  &
        'DXT ', a(idxt), a(itopu), a(itopt), a(irtgu), a(irtgt), c_gd%zt, jd,  &
        i_grid, subtract_vert_gradient, vt3dd)!       ,.TRUE.,vt3dd) ! dU/dx
   call true_horiz_gradient(m1, m2, m3, ia_1, iz, ja_1, jz, vp, vt3db,  &
        'XDIR', 'DXM ', a(idxm), a(itopv), a(itopm), a(irtgv), a(irtgm), c_gd%zt,  &
        jd, i_grid, subtract_vert_gradient, vt3de)!        ,.TRUE.,vt3de) ! dV/dx
   call true_horiz_gradient(m1, m2, m3, ia_1, iz, ja, jz, wp, vt3df, 'XDIR',  &
        'WDXU', a(idxu), a(itopt), a(itopu), a(irtgt), a(irtgu), c_gd%zt, jd,  &
        i_grid, subtract_vert_gradient, scr2)!        ,.TRUE.,scr2)  ! dW/dx

   ! Calculate y-dir gradients (sigma-z issues here, so don't use `gradient`):

   call true_horiz_gradient(m1, m2, m3, ia_1, iz, ja_1, jz, up, vt3dn,  &
        'YDIR', 'DYM ', a(idym), a(itopu), a(itopm), a(irtgu), a(irtgm), c_gd%zt,  &
        jd, i_grid, subtract_vert_gradient, vt3dd)!        ,.TRUE.,vt3dd) ! dU/dy
   call true_horiz_gradient(m1, m2, m3, ia, iz, ja, jz1, vp, vt3dc, 'YDIR',  &
        'DYT ', a(idyt), a(itopv), a(itopt), a(irtgv), a(irtgt), c_gd%zt, jd,  &
        i_grid, subtract_vert_gradient, vt3de)!        ,.TRUE.,vt3de) ! dV/dy
   call true_horiz_gradient(m1, m2, m3, ia, iz, ja_1, jz, wp, vt3dg, 'YDIR',  &
        'WDYV', a(idyv), a(itopt), a(itopv), a(irtgt), a(irtgv), c_gd%zt, jd,  &
        i_grid, subtract_vert_gradient, scr2)!        ,.TRUE.,scr2)  ! dW/dy

end if   ! End ihorgrad test
   
select case (idiffk)
case (1:2,6:9)

   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-1
            vt3dh(k,i,j) = 2._f4 * (vt3da(k,i,j) * vt3da(k,i,j)  &
                 + vt3dc(k,i,j) * vt3dc(k,i,j))  &
                 + 0.0625_f4 * (vt3db(k,i,j) + vt3db(k,i-1,j)  &
                 + vt3db(k,i,j-jd) + vt3db(k,i-1,j-jd)  &
                 + vt3dn(k,i,j) + vt3dn(k,i-1,j)  &
                 + vt3dn(k,i,j-jd) + vt3dn(k,i-1,j-jd)) ** 2
            vt3di(k,i,j) = 0.0625_f4 * ((vt3dd(k,i,j) + vt3dd(k-1,i,j)  &
                 + vt3dd(k,i-1,j) + vt3dd(k-1,i-1,j)) ** 2  &
                 + (vt3de(k,i,j) + vt3de(k-1,i,j)  &
                 + vt3de(k,i,j-jd) + vt3de(k-1,i,j-jd)) ** 2)
         end do
      end do
   end do

case (5)

   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-1
            vt3da(k,i,j) = 2._f4 * vt3da(k,i,j)
            vt3dc(k,i,j) = 2._f4 * vt3dc(k,i,j)
            scr2(k,i,j) = 2._f4 * scr2(k,i,j)
            vt3db(k,i,j) = vt3db(k,i,j) + vt3dn(k,i,j)
            vt3dn(k,i,j) = vt3db(k,i,j)
            vt3dd(k,i,j) = vt3dd(k,i,j) + vt3df(k,i,j)
            vt3de(k,i,j) = vt3de(k,i,j) + vt3dg(k,i,j)
            vt3di(k,i,j) = 0.333333_f4*(vt3da(k,i,j)+vt3dc(k,i,j)+scr2(k,i,j))
         end do
      end do

      do k = 2,m1-1
         vt3da(k,iz1,j) = 2._f4 * vt3da(k,iz1,j)
         vt3db(k,ia_1,j) = vt3db(k,ia_1,j) + vt3dn(k,ia_1,j)
         vt3dn(k,ia_1,j) = vt3db(k,ia_1,j)
         vt3dd(k,ia_1,j) = vt3dd(k,ia_1,j) + vt3df(k,ia_1,j)
      end do
   end do

   do i = ia_1,iz
      do k = 2,m1-1
         vt3dc(k,i,jz1) = 2._f4 * vt3dc(k,i,jz1)
         vt3db(k,i,ja_1) = vt3db(k,i,ja_1) + vt3dn(k,i,ja_1)
         vt3dn(k,i,ja_1) = vt3db(k,i,ja_1)
         vt3de(k,i,ja_1) = vt3de(k,i,ja_1) + vt3dg(k,i,ja_1)
      end do
   end do

   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-1
            vt3dh(k,i,j) = 0.5_f4 * (  &
                 (vt3da(k,i,j) - vt3di(k,i,j)) ** 2  &
                 + (vt3dc(k,i,j) - vt3di(k,i,j)) ** 2  &
                 + ( scr2(k,i,j) - vt3di(k,i,j)) ** 2)  &
                 + 0.0625_f4 * ((vt3db(k,i,j) + vt3db(k,i-1,j)  &
                 + vt3db(k,i,j-jd) + vt3db(k,i-1,j-jd)) ** 2  &
                 + (vt3dd(k,i,j) + vt3dd(k,i-1,j)  &
                 + vt3dd(k-1,i,j) + vt3dd(k-1,i-1,j)) ** 2  &
                 + (vt3de(k,i,j) + vt3de(k-1,i,j)  &
                 + vt3de(k,i,j-jd) + vt3de(k-1,i,j-jd)) ** 2)
            vt3di(k,i,j) = vt3dh(k,i,j)
         end do
      end do
   end do

case (3:4)

   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-1
            vt3da(k,i,j) = 2._f4 * vt3da(k,i,j)
            vt3dc(k,i,j) = 2._f4 * vt3dc(k,i,j)
            scr2(k,i,j) = 2._f4 * scr2(k,i,j)
            vt3db(k,i,j) = vt3db(k,i,j) + vt3dn(k,i,j)
            vt3dn(k,i,j) = vt3db(k,i,j)
            vt3dd(k,i,j) = vt3dd(k,i,j) + vt3df(k,i,j)
            vt3de(k,i,j) = vt3de(k,i,j) + vt3dg(k,i,j)
            vt3di(k,i,j) = 0.333333_f4*(vt3da(k,i,j)+vt3dc(k,i,j)+scr2(k,i,j))
         end do
      end do

      do k = 2,m1-1
         vt3da(k,iz1,j) = 2._f4 * vt3da(k,iz1,j)
         vt3db(k,ia_1,j) = vt3db(k,ia_1,j) + vt3dn(k,ia_1,j)
         vt3dn(k,ia_1,j) = vt3db(k,ia_1,j)
         vt3dd(k,ia_1,j) = vt3dd(k,ia_1,j) + vt3df(k,ia_1,j)
      end do
   end do

   do i = ia_1,iz
      do k = 2,m1-1
         vt3dc(k,i,jz1) = 2._f4 * vt3dc(k,i,jz1)
         vt3db(k,i,ja_1) = vt3db(k,i,ja_1) + vt3dn(k,i,ja_1)
         vt3dn(k,i,ja_1) = vt3db(k,i,ja_1)
         vt3de(k,i,ja_1) = vt3de(k,i,ja_1) + vt3dg(k,i,ja_1)
      end do
   end do

   do j = ja,jz
      do i = ia,iz
         do k = 2,m1-1
            vt3dh(k,i,j) = 0.5_f4 * (  &
                 (vt3da(k,i,j) - vt3di(k,i,j)) ** 2  &
                 + (vt3dc(k,i,j) - vt3di(k,i,j)) ** 2  &
                 + ( scr2(k,i,j) - vt3di(k,i,j)) ** 2)  &
                 + 0.0625_f4 * ((vt3db(k,i,j) + vt3db(k,i-1,j)  &
                 + vt3db(k,i,j-jd) + vt3db(k,i-1,j-jd)) ** 2  &
                 + (vt3dd(k,i,j) + vt3dd(k,i-1,j)  &
                 + vt3dd(k-1,i,j) + vt3dd(k-1,i-1,j)) ** 2  &
                 + (vt3de(k,i,j) + vt3de(k-1,i,j)  &
                 + vt3de(k,i,j-jd) + vt3de(k-1,i,j-jd)) ** 2)
            vt3di(k,i,j) = vt3dh(k,i,j)
         end do
      end do
   end do

end select

end subroutine strain


!```````````````````````````````````````````````````````````````````
subroutine tkemy(m1, m2, m3, ia, iz, ja, jz, ibndc, jd, tkep, tket, vt3dh,  &
     vt3di, vt3dj, scr1, rtgt, theta, dn0, up, vp, wp, uw, vw, wfz, tfz, tkep2)

!+ Calculates K's based on turb. scheme of Mellor-Yamada

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use main_memory_mod, only: vctr10, vctr11, vctr12, vctr13, vctr5, vctr9,  &
     vctr1, a, jdim, c_gd
use body_specific_constants_mod, only: tkmin, vonk, g_Xoid
use vtables_mod, only: ivt3dp, irtgt
use turbulent_K_coeffs_mod, only: filter_avg9pt

implicit none

!=== Argument declarations:

integer, intent(IN) :: m1, m2, m3, ia, iz, ja, jz, ibndc, jd
real(f4), dimension(m1,m2,m3), intent(IN) :: tkep,  & ! [m^2/s^2]
                                             theta,  & ! [K]
                                             dn0,  & ! [kg/m^3]
                                             up, vp, wp ! [m/s]
real(f4), dimension(m2,m3), intent(IN) :: rtgt,  &
                                          uw, vw, wfz,  & ! [m^2/s^2]
                                          tfz ! [(K.m)/s]

real(f4), dimension(m1), intent(INOUT) :: tkep2
real(f4), dimension(m1,m2,m3), intent(INOUT) :: vt3dj,  & ! IN:[s^-2]
                                           vt3di,  & ! IN:[s^-2], OUT:[kg/(m.s)]
                                               scr1, vt3dh,  & ! OUT:[kg/(m.s)] 
                                                tket ! INOUT[m^2/s^3]

!== Local declarations:

integer :: i, j, k
real(f4) :: wght1, wght3,  &
     sumtkz, sumtk, al0, tket2, ri, rf, shr, smr, tker, qq, ssmf, shf, sh0,  &
     ssm, aux, gm, gh, sm1, sm2, sh1, sh2, dzloc

real(f4), parameter :: a1 = 0.92_f4, a2 = 0.74_f4, b1 = 16.6_f4, b2 = 10.1_f4,  &
     c1 = 0.08_f4
real(f4), parameter :: aux1 = 0.758964199_f4, aux2 = 2.58286747_f4
real(f4), parameter :: rf1 = ONE_f4, rf2 = 0.191232309_f4, rf3 = 0.223117196_f4,  &
     rf4 = 0.234067819_f4

!=== Executable statements:

!@@! Populate boundary points of Brunt-Vaisalla frequency (squared) array; then 
!@@!  apply smoothing (9-pt) in the horizontal (sigma-z surface) -- done to
!@@!  significantly decrease noise in TKE field that does not respond well
!@@!  (hardly at all!) to increased diffusion (via increased K-coeffs):
!@@
!@@j = ja
!@@do i=ia,iz
!@@   vt3dj(1:m1,i,j-jdim) = vt3dj(1:m1,i,j)
!@@end do
!@@j = jz
!@@do i=ia,iz
!@@   vt3dj(1:m1,i,j+jdim) = vt3dj(1:m1,i,j)
!@@end do
!@@i = ia
!@@do j=ja,jz
!@@   vt3dj(1:m1,i-1,j) = vt3dj(1:m1,i,j)
!@@end do
!@@i = iz
!@@do j=ja,jz
!@@   vt3dj(1:m1,i+1,j) = vt3dj(1:m1,i,j)
!@@end do
!@@vt3dj(1:m1,ia-1,ja-jdim) =  &
!@@     (vt3dj(1:m1,ia-1,ja)+vt3dj(1:m1,ia,ja)+vt3dj(1:m1,ia,ja-jdim))/3.
!@@vt3dj(1:m1,iz+1,jz+jdim) =  &
!@@     (vt3dj(1:m1,iz,jz+jdim)+vt3dj(1:m1,iz,jz)+vt3dj(1:m1,iz+1,jz))/3.
!@@vt3dj(1:m1,ia-1,jz+jdim) =  &
!@@     (vt3dj(1:m1,ia-1,jz)+vt3dj(1:m1,ia,jz)+vt3dj(1:m1,ia,jz+jdim))/3.
!@@vt3dj(1:m1,iz+1,ja-jdim) =  &
!@@     (vt3dj(1:m1,iz,ja-jdim)+vt3dj(1:m1,iz,ja)+vt3dj(1:m1,iz+1,ja))/3.
!@@
!@@call filter_avg9pt(m1, m2, m3, ia, iz, ja, jz, vt3dj, a(ivt3dp), a(irtgt))

! 7 - mellor and yamada (after andre et al, 1978)

wght3 = ONE_f4
!c      wght3=c_gd%zt(2)/c_gd%zt(3)
wght1 = ONE_f4-wght3

do j=ja,jz 
   do i=ia,iz

     do k=2,m1-1
         tkep2(k) = 2._f4*tkep(k,i,j) ! [m^2/s^2]
         vctr10(k) = sqrt(tkep2(k)) ! [m/s]
         vctr11(k) = 0.75_f4*sqrt(tkep2(k)/max(1.e-20_f4, vt3dj(k,i,j))) ! [m]
         vctr1(k) = c_gd%zt(k)*rtgt(i,j) ! [m]
         dzloc = (c_gd%zm(k)-c_gd%zm(k-1))*rtgt(i,j) ! [m]
         vctr13(k) = vctr10(k)*dzloc ! [m^2/s]
         vctr12(k) = vctr13(k)*vctr1(k) ! [m^3/s]
      end do

      sumtkz = sum(vctr12(2:m1-1)) ! [m^3/s]
      sumtk = sum(vctr13(2:m1-1)) ! [m^2/s]
      al0 = 0.1_f4*sumtkz/sumtk ! [m]
      tket2 = tket(2,i,j) ! [m^2/s^3]

      do k=2,m1-1
         vctr9(k) = min(vonk*vctr1(k)/(ONE_f4+vonk*vctr1(k)/al0),  &
              vctr11(k)) ! [m]

! --- for growing turbulence use helfand and labraga's modified sm and sh (sh0)

         ri = min(vt3dj(k,i,j)/max(vt3di(k,i,j),1.e-11_f4), 0.190_f4) ! [none]
         rf = min(0.6588_f4*(ri+0.1776_f4-  &
              sqrt(ri*(ri-0.3221_f4)+0.03156_f4)), 0.16_f4) ! [none]
         shr = aux2*(rf-rf2)/(rf-rf1) ! [none]
         smr = aux1*(rf-rf4)/(rf-rf3)*shr ! [none]
         tker = max(16.6_f4*vctr9(k)*vctr9(k)*(smr*vt3di(k,i,j)  &
              -shr*vt3dj(k,i,j)), 2._f4*tkmin) ! [m^2/s^2]
         if (tker > tkep2(k) ) then
            qq = sqrt(tkep2(k)/tker) ! [none]
            ssmf = qq*smr ! [none]
            shf = qq*shr ! [none]
            sh0 = shf ! [none]
            ssm = ssmf ! [none]
         else

! --- for decaying turbulence use mellor and yamada's sm and sh (sh0)

            aux = vctr9(k)*vctr9(k)/tkep2(k) ! [s^2]
            gm = aux*vt3di(k,i,j) ! [none]
            gh = -aux*vt3dj(k,i,j) ! [none]
            sm1 = 0.6992_f4-9.33948672_f4*gh ! [none]
            sm2 = ONE_f4-(36.7188_f4-187.4408515_f4*gh+88.83949824_f4*gm)*gh+  &
                 5.0784_f4*gm ! [none]  
            ssm = sm1/max(sm2, 1.e-10_f4) ! [none]
            sh1 = 0.74_f4-4.0848_f4*ssm*gm ! [none]
            sh2 = ONE_f4-30.5916_f4*gh ! [none]
            sh0 = sh1/max(sh2, 1.e-10_f4) ! [none]
         end if

         scr1(k,i,j) = vctr9(k)*vctr10(k)*ssm ! [m^2/s]
         vt3dh(k,i,j) = vctr9(k)*vctr10(k)*sh0 ! [m^2/s]
         vctr5(k) = scr1(k,i,j)*vt3di(k,i,j)-  &
              vt3dh(k,i,j)*vt3dj(k,i,j) ! [m^2/s^3]

         tket(k,i,j) = tket(k,i,j) + 0.5_f4 * vctr5(k)  & ! [m^2/s^3]
              -tkep2(k) * sqrt(max(1.e-20_f4, tkep2(k)))/(vctr9(k)*16.6_f4)

         scr1(k,i,j) = scr1(k,i,j)*dn0(k,i,j) ! [kg/(m.s)]
         vt3dh(k,i,j) = vt3dh(k,i,j)*dn0(k,i,j) ! [kg/(m.s)]
         vt3di(k,i,j) = 0.2_f4*vctr9(k)*vctr10(k)*dn0(k,i,j) ! [kg/(m.s)]

      end do

      if (c_gd%nstbot == 1) then
!c
!c  experimental diagnostic equilibrium form - may try this someday
!c
!c             tkep(2,i,j) = 3.25 * sqrt(uw(i,j)**2 + vw(i,j)**2)
!c
!c      (then, zero out tket(2,i,j) just before calling advance)
!c
         tket(2,i,j) = tket2 + 0.5_f4 * wght3 * vctr5(3) + wght1  & ! [m^2/s^3]
              *(-(uw(i,j)*(up(2,i,j)+up(2,i-1,j))  &       
              +vw(i,j)*(vp(2,i,j)+vp(2,i,j-jd))  &         
              +wfz(i,j)*wp(2,i,j))*c_gd%dzt(2)/rtgt(i,j)  &      
              +tfz(i,j)*g_Xoid/theta(2,i,j)) - tkep2(2)  &
              *sqrt(max(1.e-20_f4,tkep2(2)))/(vctr9(2)*16.6_f4)

      end if

   end do
end do

!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
!     call friclyr(nzp,nxp,nyp,a(iscr1),a(iustarl),a(itstarl)
!    +    ,a(iustarw),a(itstarw),a(ipctlnd),a(itheta),a(irtgt))
!cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc

end subroutine tkemy
