!```````````````````````````````````````````````````````````````````
module postp_GrADS_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine postp_GrADS_driv

!+ Post-processing driver for GrADS format output

use univ_kind_defs_mod, only: f4, f8
use univ_GrADS_format_mod, only: open_GrADS_dfile, close_GrADS_dfile,  &
     write_GrADS_hfile, init_GrADS_format, exit_GrADS_format
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_integer_time_mod, only: time_obj, fp_seconds_eqv_Time
use univ_exception_mod, only: exception, FATAL
use univ_std_fileIO_mod, only: find_file_IOunit, mkdir_p, determine_IO_reclen
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN
use univ_namelist_utils_mod, only: Time_to_gen_time_strrep
use body_specific_constants_mod, only: mnsdy
use postp_main_memory_mod, only: cfg_f, cfg_f_inc, output_pp_file_pfx,  &
     ztrnsfm, cfg_n_ov, state_file_name, model_sim_ET, cfg_n_f, cfg_vname,  &
     nxp, nyp, nvar_read, what_vtype, cfg_vtype, zt, var_want_to_read,  &
     nvar_want, cfg_vlevs, cfg_vdesc, cfg_vunits, what_vlevs,  &
     what_vdesc, what_vunits, what_ofile, cfg_vofile, SBPS_Tlev,  &
     n_state_files, grid_idx_list, cfg_n_g, undef_value_f4, PRAMS_vdata,  &
     cfg_spatial, vctr1_sz, vctr1_az, use_full_az_scr, use_full_xy_scr,  &
     use_full_azxy_scr, use_full_sz_scr, full_az_scr_needed,  &
     full_xy_scr_needed, full_azxy_scr_needed, full_sz_scr_needed
use postp_varlib_mod, only: postp_varlib
use postp_driv_mod, only: manage_PRAMS_vdata, manage_cfg_v_memory
use postp_getdata_mod, only: count_vars_to_read, fill_prams_data, get_prams_data
use main_memory_mod, only: module_tps, start_UT, mnsdy_Time, morbpd_Time

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_GrADS_driv'

integer, parameter :: maxnGrADS_files = 2

integer :: igd, iv, ifn, lnoutp, lnstp, in_unit, iq, ic, rfcnt, nc, i, e,  &
     IO_reclen
character(len=MAX_FILE_PATH_LEN) :: s_dfn
character(len=MAX_FILE_PATH_LEN), dimension(maxnGrADS_files) :: g_dfn, g_hfn
character(len=8), dimension(maxnGrADS_files) :: file_handle
character(len=7) :: cgrid_a
character(len=10), dimension(maxnGrADS_files) :: cgrid_gd, cgrid_gh
character(len=40) :: gen_time_strrep
type(time_obj) :: tmp_Time
real(f4), allocatable, dimension(:,:) :: reclen_test
logical, dimension(maxnGrADS_files) :: file_used
logical, allocatable, dimension(:,:) :: var_in_file
real(f8) :: tinc

!=== Executable statements:

! Make initial pass through all grids, files, and variables; record what
!  variables need to be read in and where they may be found:

nvar_read = 0   ! Start out by assuming that no vars need read in
use_full_az_scr = .FALSE.; use_full_xy_scr = .FALSE.
use_full_azxy_scr = .FALSE.; use_full_sz_scr = .FALSE.

do i=1,cfg_n_g  ! Loop over all desired grids
   igd = grid_idx_list(i)
   do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         ! What state variable(s) does this variable require? 
         call postp_varlib(igd, i, ifn, cfg_vname(iv), .TRUE.)
         use_full_az_scr(i) = ( use_full_az_scr(i) .or. full_az_scr_needed )
         use_full_xy_scr(i) = ( use_full_xy_scr(i) .or. full_xy_scr_needed )
         use_full_azxy_scr(i) = (use_full_azxy_scr(i) .or. full_azxy_scr_needed)
         use_full_sz_scr(i) = ( use_full_sz_scr(i) .or. full_sz_scr_needed )

         ! Need to configure variable(s) for reading?
         call count_vars_to_read(igd, ifn)

      end do
      if (ztrnsfm == 2) then   ! If Cartesian interp desired, need topo values 
         nvar_want = 1
         var_want_to_read(1:nvar_want) = (/'TOPT'/)
         call count_vars_to_read(igd, ifn)
      end if
   end do
end do

! Allocate some data storage:
allocate(PRAMS_vdata(nvar_read), var_in_file(cfg_n_ov,maxnGrADS_files), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

var_in_file = .FALSE.   ! Initialize logical array

! Initialize GrADS format module:
file_handle(1) = 'atm'; file_handle(2) = 'sfc'
call init_GrADS_format(maxnGrADS_files, file_handle)

! Main pass; read in data, calculate and output POSTP variable values:

i = index(output_pp_file_pfx, '/', back=.TRUE.)-1  ! Determine dir name length
call mkdir_p(output_pp_file_pfx(1:i))  ! Create any needed directories

lnoutp = len_trim(output_pp_file_pfx)  ! Length of string

do i=1,cfg_n_g   ! Loop over all desired grids
   igd = grid_idx_list(i)

   call manage_PRAMS_vdata(igd)    ! Initialize PRAMS_vdata for this grid
   call manage_cfg_v_memory(igd, i)  ! Init cfg-dimensioned multi-use storage

   write (cgrid_a, '(a1,i2.2,a4)') 'g', igd, '.vfm'  ! State file grid suffix

   if (ztrnsfm == 1) then
      do ifn=1,maxnGrADS_files
         write (cgrid_gd(ifn), '(a4,i2.2,a4)') '-S-g', igd, '.dat' ! GrADS grid 
         write (cgrid_gh(ifn), '(a4,i2.2,a4)') '-S-g', igd, '.ctl' !  suffixes
      end do
   else if (ztrnsfm == 2) then
      do ifn=1,maxnGrADS_files
         write (cgrid_gd(ifn), '(a4,i2.2,a4)') '-C-g', igd, '.dat'  ! GrADS grid
         write (cgrid_gh(ifn), '(a4,i2.2,a4)') '-C-g', igd, '.ctl'  !  suffixes
      end do
   end if

   g_dfn(1) = output_pp_file_pfx(1:lnoutp)//'-atm'//cgrid_gd(1)  ! GrADS f_names
   g_hfn(1) = output_pp_file_pfx(1:lnoutp)//'-atm'//cgrid_gh(1)  ! 
   g_dfn(2) = output_pp_file_pfx(1:lnoutp)//'-sfc'//cgrid_gd(2)  !
   g_hfn(2) = output_pp_file_pfx(1:lnoutp)//'-sfc'//cgrid_gh(2)  ! 

   ! Open GrADS data file(s) for writing:
   call open_GrADS_dfile(g_dfn(1), cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
        'W', file_handle(1))
   call open_GrADS_dfile(g_dfn(2), cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
        'W', file_handle(2))

   rfcnt = 0
   do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
      rfcnt = rfcnt+1      ! File number (relative) counter

      tmp_Time = fp_seconds_eqv_Time(model_sim_ET(ifn), module_tps)
      call Time_to_gen_time_strrep(tmp_Time, gen_time_strrep,  &
           len(gen_time_strrep), mnsdy_Time, morbpd_Time,  &
           short_unit_style=.TRUE.)
      write (*, '(a,i5,a,i2,3a,f6.2,a)') 'Postprocessing < ', rfcnt,  &
           ' > grid #', igd, ', time = [', trim(gen_time_strrep), '] [',  &
           mod(real(model_sim_ET(ifn)/mnsdy*24._f8, kind=f4)+start_UT,  &
           24._f4), ' UT] ...'

      lnstp = len_trim(state_file_name(ifn))-7
      s_dfn = state_file_name(ifn)(1:lnstp)//cgrid_a    ! State file name

      in_unit = find_file_IOunit()

      allocate(reclen_test(nxp(igd,ifn),nyp(igd,ifn)), stat=e)
      if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
           'reclen')
      IO_reclen = determine_IO_reclen(reclen_test)
      deallocate(reclen_test, stat=e)
      if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
           'reclen')

      ! Open state data file:
      open(in_unit, file=s_dfn(1:lnstp+7), status='old', form='unformatted',  &
           access='direct', recl=IO_reclen)

      ! Read in all *NEEDED* PRAMS data for this grid/file; fill data array:
      call fill_PRAMS_data(in_unit, ifn, igd, nxp(igd,ifn)*nyp(igd,ifn))

      close(in_unit)   ! Close state file

      ! Compute values of configured variables; output data in desired manner:

      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         ! Read/compute var data:
         call postp_varlib(igd, i, ifn, cfg_vname(iv), .FALSE.)

         ! Store/calculate variable descriptors:
         if (ifn == cfg_f(1)) then
            cfg_vtype(iv) = what_vtype    ! Store var types, etc.
            cfg_vlevs(iv) = what_vlevs    !
            cfg_vdesc(iv) = what_vdesc    !
            cfg_vunits(iv) = what_vunits  !
            
            iq = 1
            do while(file_handle(iq) /= what_ofile)
               iq = iq+1
               if (iq > maxnGrADS_files) call exception(procedure_name,  &
                    FATAL, 'Incorrect ofile name in varlib')
            end do
            cfg_vofile(iv) = iq  ! Output file# this variable was written to

            var_in_file(iv,iq) = .TRUE.
         end if

         ! Output variable:
         call postp_GrADS_varout(i, file_handle(cfg_vofile(iv))) 

      end do

      if (ifn == cfg_f(1)) then  ! Write GrADS file descriptors after first
                                 !  file processed

         ! Ascertain which GrADS datafile(s) was(were) actually written to:
         file_used = .FALSE.
         do iv=1,cfg_n_ov    ! Loop over # of variables processed
            file_used(cfg_vofile(iv)) = .TRUE.
         end do
         
         ! Write GrADS descriptor file(s):
         
         write (*, '(/)')
         do iv=1,maxnGrADS_files

            if (file_used(iv)) then    ! Write GrADS header file
               if (n_state_files > cfg_f_inc) then   ! If enough files
                  tinc = model_sim_ET(cfg_f(1)+cfg_f_inc)-  &
                       model_sim_ET(cfg_f(1)) 
               else
                  tinc = model_sim_ET(cfg_f(1))   ! Best possible right now
               end if
               if (iv == 1) then
                  do ic=1,cfg_spatial(i)%n_AZ
                     vctr1_az%p1D(ic) = zt(cfg_spatial(i)%i_AZ(ic),igd,cfg_f(1))
                  end do
                  call write_GrADS_hfile(g_dfn(iv), g_hfn(iv),  &
                       cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y, cfg_n_f,  &
                       undef_value_f4, 0, 'green', vctr1_az%p1D,  &
                       model_sim_ET(cfg_f(1)), tinc, start_UT, cfg_vname,  &
                       cfg_vlevs, cfg_vdesc, cfg_vunits,  &
                       var_in_file(1:cfg_n_ov,iv), mnsdy)
               else if (iv == 2) then
                  do ic=1,cfg_spatial(i)%n_SZ
                     vctr1_sz%p1D(ic) =  &
                          SBPS_Tlev(cfg_spatial(i)%i_SZ(ic),cfg_f(1))
                  end do
                  call write_GrADS_hfile(g_dfn(iv), g_hfn(iv),  &
                       cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y, cfg_n_f,  &
                       undef_value_f4, 0, 'green', vctr1_sz%p1D,  &
                       model_sim_ET(cfg_f(1)), tinc, start_UT, cfg_vname,  &
                       cfg_vlevs, cfg_vdesc, cfg_vunits,  &
                       var_in_file(1:cfg_n_ov,iv), mnsdy)
               end if
               write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
                    g_hfn(iv)(1:len_trim(g_hfn(iv)))
            end if
            
         end do
         write (*, '(/)')

      end if

   end do

   ! Conclude GrADS file writes:
         
   write (*, '(/)')
   do iv=1,maxnGrADS_files
      
      ! Close GrADS data file:
      call close_GrADS_dfile(file_handle(iv))
            
      if (file_used(iv)) then    ! Display GrADS header file name
         write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
              g_hfn(iv)(1:len_trim(g_hfn(iv)))
      else  ! Remove unused GrADS datafiles
         nc = len_trim(g_dfn(iv))
         call univ_remove_f(nc, g_dfn(iv), e)
      end if

   end do
   write (*, '(/)')

end do

! Exit GrADS format module:
call exit_GrADS_format

! Recycle data structures:
call manage_PRAMS_vdata(-1)
call manage_cfg_v_memory(-1, -1) 

! Deallocate work array:
deallocate(PRAMS_vdata, var_in_file, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

end subroutine postp_GrADS_driv


!```````````````````````````````````````````````````````````````````
subroutine postp_GrADS_varout(cfg_igd, file_handle)

!+ Writes postprocessed variable data to GrADS-format file

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_GrADS_format_mod, only: write_GrADS_dfile
use univ_exception_mod, only: exception, FATAL
use postp_main_memory_mod, only: what_vtype, vl_reorder_t3,  &
     vl_reorder_t5, vl_result_t2, vl_result_t3, vl_result_t4, vl_result_t5,  &
     cfg_spatial

implicit none

!=== Argument declarations:

integer, intent(IN) :: cfg_igd
character(len=8), intent(IN) :: file_handle

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_GrADS_varout'

integer :: nx, ny, naz, nsz, ip, np

!=== Executable statements:

nx = cfg_spatial(cfg_igd)%n_X; ny = cfg_spatial(cfg_igd)%n_Y

which_vtype: select case(what_vtype)
case (2)  ! "regular" 2D variable

   call write_GrADS_dfile(nx, ny, 1, vl_result_t2%p2D, file_handle)

case (3)  ! "regular" 3D variable

   naz = cfg_spatial(cfg_igd)%n_AZ
   vl_reorder_t3%p3D = reshape(vl_result_t3%p3D, (/nx,ny,naz/), order=(/3,1,2/))
   call write_GrADS_dfile(nx, ny, naz, vl_reorder_t3%p3D, file_handle)

case (4)  ! SBPS 2D variable

   do ip=1,cfg_spatial(cfg_igd)%n_P
      call write_GrADS_dfile(nx, ny, 1, vl_result_t4%p3D(:,:,ip), file_handle)
   end do

case (5)  ! SBPS 3D variable

   nsz = cfg_spatial(cfg_igd)%n_SZ; np = cfg_spatial(cfg_igd)%n_P
   vl_reorder_t5%p4D = reshape(vl_result_t5%p4D, (/nx,ny,nsz,np/),  &
        order=(/3,1,2,4/))
   do ip=1,np
      call write_GrADS_dfile(nx, ny, nsz, vl_reorder_t5%p4D(:,:,:,ip),  &
           file_handle)
   end do

case default

   write (*, '(/a,i1,a)') 'ERROR: unknown variable type ( = ', what_vtype, ' ).'
   call exception(procedure_name, FATAL)

end select which_vtype

end subroutine postp_GrADS_varout


end module postp_GrADS_mod
