!```````````````````````````````````````````````````````````````````
module postp_NetCDF_mod

!+

use manage_NetCDF_classic_mod, only: mNc_file_metadata_obj

implicit none

type(mNc_file_metadata_obj), save :: output_file_obj


contains


!```````````````````````````````````````````````````````````````````
subroutine postp_NetCDF_driv

!+ Post-processing driver for NetCDF format output

use univ_kind_defs_mod, only: f4, f8
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_integer_time_mod, only: time_obj, fp_seconds_eqv_Time
use univ_exception_mod, only: exception, FATAL
use univ_std_fileIO_mod, only: find_file_IOunit, mkdir_p, determine_IO_reclen
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN
use univ_namelist_utils_mod, only: Time_to_gen_time_strrep
use manage_NetCDF_classic_mod, only: mNc_sync_with_file,  &
     mNc_recycle_f_md_obj, mNc_MAX_NAME_LEN
use body_specific_constants_mod, only: mnsdy
use postp_main_memory_mod, only: cfg_f, cfg_f_inc, output_pp_file_pfx,  &
     ztrnsfm, cfg_n_ov, state_file_name, model_sim_ET, cfg_vname, nxp, nyp,  &
     nvar_read, what_vtype, cfg_vtype, var_want_to_read, nvar_want,  &
     cfg_vlevs, cfg_vdesc, cfg_vunits, what_vlevs, what_vdesc, what_vunits,  &
     what_ofile, cfg_vofile, n_state_files, grid_idx_list, cfg_n_g,  &
     PRAMS_vdata, cfg_spatial, use_full_az_scr, use_full_xy_scr,  &
     use_full_azxy_scr, use_full_sz_scr, full_az_scr_needed,  &
     full_xy_scr_needed, full_azxy_scr_needed, full_sz_scr_needed
use postp_varlib_mod, only: postp_varlib
use postp_driv_mod, only: manage_PRAMS_vdata, manage_cfg_v_memory
use postp_getdata_mod, only: count_vars_to_read, fill_PRAMS_data, get_PRAMS_data
use main_memory_mod, only: module_tps, start_UT, mnsdy_Time, morbpd_Time

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_NetCDF_driv'

integer, parameter :: maxnGrADS_files = 3

integer :: igd, iv, ifn, lnoutp, lnstp, IO_unit, iq, rfcnt, nc, i, e, ip,  &
     IO_reclen
character(len=MAX_FILE_PATH_LEN) :: s_dfn, nc_dfn
character(len=MAX_FILE_PATH_LEN), dimension(maxnGrADS_files) :: g_hfn
character(len=8), dimension(maxnGrADS_files) :: file_handle
character(len=7) :: cgrid_a
character(len=9) :: cgrid_nc
character(len=3) :: c_ip
character(len=13), dimension(maxnGrADS_files) :: cgrid_gh
integer, dimension(maxnGrADS_files) :: n_g_hfn_vars
character(len=40) :: gen_time_strrep
type(time_obj) :: tmp_Time
logical, dimension(maxnGrADS_files) :: file_used
logical, allocatable, dimension(:,:) :: var_in_file
real(f4), allocatable, dimension(:,:) :: reclen_test
character(len=mNc_MAX_NAME_LEN) :: tmp_mNc_str
real(f4) :: UT
real(f8) :: tinc

!=== Executable statements:

! Make initial pass through all grids, files, and variables; record what
!  variables need to be read in and where they may be found:

nvar_read = 0   ! Start out by assuming that no vars need read in
use_full_az_scr = .FALSE.; use_full_xy_scr = .FALSE.
use_full_azxy_scr = .FALSE.; use_full_sz_scr = .FALSE.

do i=1,cfg_n_g  ! Loop over all desired grids
   igd = grid_idx_list(i)
   do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         ! What state variable(s) does this variable require? 
         call postp_varlib(igd, i, ifn, cfg_vname(iv), .TRUE.)
         use_full_az_scr(i) = ( use_full_az_scr(i) .or. full_az_scr_needed )
         use_full_xy_scr(i) = ( use_full_xy_scr(i) .or. full_xy_scr_needed )
         use_full_azxy_scr(i) = (use_full_azxy_scr(i) .or. full_azxy_scr_needed)
         use_full_sz_scr(i) = ( use_full_sz_scr(i) .or. full_sz_scr_needed )

         ! Need to configure variable(s) for reading?
         call count_vars_to_read(igd, ifn)

      end do

      ! Need topo values for 'rtgt' calc/write in `postp_NetCDF_f_auxv_tind`
      !  and/or the ZTRNSFM=2 interpolation in `postp_varlib`: 
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TOPT'/)
      call count_vars_to_read(igd, ifn)
   end do
end do

! Allocate some data storage:
allocate(PRAMS_vdata(nvar_read), var_in_file(cfg_n_ov,maxnGrADS_files), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

var_in_file = .FALSE.   ! Initialize logical array

! For GrADS-format header files:
file_handle = (/ 'z_wstag ', 'z_SBPS  ', 'z_normal' /)

! Main pass; read in data, calculate and output POSTP variable values:

i = index(output_pp_file_pfx, '/', back=.TRUE.)-1  ! Determine dir name length
call mkdir_p(output_pp_file_pfx(1:i))  ! Create any needed directories

lnoutp = len_trim(output_pp_file_pfx)  ! Length of string

do i=1,cfg_n_g   ! Loop over all desired grids
   igd = grid_idx_list(i)

   call manage_PRAMS_vdata(igd)    ! Initialize PRAMS_vdata for this grid
   call manage_cfg_v_memory(igd, i) ! Init cfg-dimensioned multi-use storage

   write (cgrid_a, '(a1,i2.2,a4)') 'g', igd, '.vfm'  ! State file grid suffix

   ! Create output file grid suffixes:
   if (ztrnsfm == 1) then
      write (cgrid_nc, '(a4,i2.2,a3)') '-S-g', igd, '.nc'
      do ifn=1,maxnGrADS_files
         write (cgrid_gh(ifn), '(a4,i2.2,a7)') '-S-g', igd, '.nc.ctl'
      end do
   else if (ztrnsfm == 2) then
      write (cgrid_nc, '(a4,i2.2,a3)') '-C-g', igd, '.nc'
      do ifn=1,maxnGrADS_files
         write (cgrid_gh(ifn), '(a4,i2.2,a7)') '-C-g', igd, '.nc.ctl'
      end do
   end if

   nc_dfn = output_pp_file_pfx(1:lnoutp)//cgrid_nc
   do ifn=1,maxnGrADS_files
      g_hfn(ifn) = output_pp_file_pfx(1:lnoutp)//'-'//  &
           trim(file_handle(ifn))//cgrid_gh(ifn)  ! GrADS header file names
   end do

   rfcnt = 0; n_g_hfn_vars = 0
   do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
      rfcnt = rfcnt+1      ! File number (relative) counter

      tmp_Time = fp_seconds_eqv_Time(model_sim_ET(ifn), module_tps)
      call Time_to_gen_time_strrep(tmp_Time, gen_time_strrep,  &
           len(gen_time_strrep), mnsdy_Time, morbpd_Time,  &
           short_unit_style=.TRUE.)
      UT = mod(real(model_sim_ET(ifn)/mnsdy*24._f8)+start_UT, 24._f4)
      write (*, '(a,i5,a,i2,3a,f6.2,a)') 'Postprocessing < ', rfcnt,  &
           ' > grid #', igd, ', time = [', trim(gen_time_strrep), '] [',  &
           UT, ' UT] ...'

      lnstp = len_trim(state_file_name(ifn))-7
      s_dfn = state_file_name(ifn)(1:lnstp)//cgrid_a    ! State file name

      IO_unit = find_file_IOunit()

      allocate(reclen_test(nxp(igd,ifn),nyp(igd,ifn)), stat=e)
      if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
           'reclen')
      IO_reclen = determine_IO_reclen(reclen_test)
      deallocate(reclen_test, stat=e)
      if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
           'reclen')

      ! Open state data file:
      open(IO_unit, file=s_dfn(1:lnstp+7), status='old', form='unformatted',  &
           access='direct', recl=IO_reclen)

      ! Read in all *NEEDED* PRAMS data for this grid/file; fill data array:
      call fill_PRAMS_data(IO_unit, ifn, igd, nxp(igd,ifn)*nyp(igd,ifn))

      close(IO_unit)   ! Close state file

      ! Compute values of configured variables; output only select data:

      if (ifn == cfg_f(1)) then

         ! Perform an initial first pass without any file output, in order to
         !  determine output var metadata:
         do iv=1,cfg_n_ov    ! Loop over # of variables to process
            ! Read/compute var data:
            call postp_varlib(igd, i, ifn, cfg_vname(iv), .FALSE.)

            ! Store/calculate variable descriptors:
            cfg_vtype(iv) = what_vtype    ! Store var types, etc.
            cfg_vlevs(iv) = what_vlevs    !
            cfg_vdesc(iv) = what_vdesc    !
            cfg_vunits(iv) = what_vunits  !
            
            if (trim(cfg_vname(iv)) == 'w') then
               what_ofile = 'z_wstag'
            else if (cfg_vtype(iv) == 5) then
               what_ofile = 'z_SBPS'
            else
               what_ofile = 'z_normal'
            end if

            do iq=1,maxnGrADS_files
               if (file_handle(iq) == what_ofile) then
                  n_g_hfn_vars(iq) = n_g_hfn_vars(iq)+1
                  exit
               end if
            end do
            if (iq > maxnGrADS_files) call exception(procedure_name,  &
                 FATAL, 'Incorrect ofile name')
            cfg_vofile(iv) = iq  ! Output file# this variable was written to
            var_in_file(iv,iq) = .TRUE.
         end do

         ! Define NetCDF-format output file:
         call postp_NetCDF_f_define(nc_dfn, i)

         ! Ascertain which GrADS header(s) need to be written:
         file_used = .FALSE.
         do iv=1,cfg_n_ov    ! Loop over # of variables processed
            file_used(cfg_vofile(iv)) = .TRUE.
         end do

         ! Try to determine the time increment between field snapshots:
         if (n_state_files > cfg_f_inc) then   ! If enough files
            tinc = model_sim_ET(cfg_f(1)+cfg_f_inc)-model_sim_ET(cfg_f(1)) 
         else
            tinc = model_sim_ET(cfg_f(1))   ! Best possible right now
         end if

         ! Write 'z_wstag' GrADS descriptor file?
         iq = 1
         if (file_used(iq)) then
            open (IO_unit, file=trim(g_hfn(iq)), status='unknown',  &
                 form='formatted')
            call postp_NetCDF_wGhfile(IO_unit, nc_dfn, 'sigma_zm', tinc,  &
                 mnsdy, UT, n_g_hfn_vars(iq))
            do iv=1,cfg_n_ov    ! Loop over # of variables to process
               if (var_in_file(iv,iq)) then
                  nc = len_trim(cfg_vname(iv)); e = min(15, nc)
                  write (IO_unit, '(3a)') cfg_vname(iv)(1:nc), '=>',  &
                       cfg_vname(iv)(1:e)
               end if
            end do
            write (IO_unit, '(a)') 'ENDVARS'
            close (IO_unit)
            write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
                 g_hfn(iq)(1:len_trim(g_hfn(iq)))
         end if

         ! Write 'z_SBPS' GrADS descriptor file?
         iq = 2
         if (file_used(iq)) then
            open (IO_unit, file=trim(g_hfn(iq)), status='unknown',  &
                 form='formatted')
            call postp_NetCDF_wGhfile(IO_unit, nc_dfn, 'z_SBPS', tinc,  &
                 mnsdy, UT, n_g_hfn_vars(iq))
            do iv=1,cfg_n_ov    ! Loop over # of variables to process
               if (var_in_file(iv,iq)) then
                  do ip=1,cfg_spatial(i)%n_P
                     write (c_ip, '(a2,i1.1)') '_p', ip
                     tmp_mNc_str = trim(cfg_vname(iv))//c_ip
                     nc = len_trim(tmp_mNc_str)
                     e = min(12, max(1, nc-3)) ! Truncate, keeping patch ID
                     write (IO_unit, '(3a)') tmp_mNc_str(1:nc), '=>',  &
                          tmp_mNc_str(1:e)//c_ip
                  end do
               end if
            end do
            write (IO_unit, '(a)') 'ENDVARS'
            close (IO_unit)
            write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
                 g_hfn(iq)(1:len_trim(g_hfn(iq)))
         end if

         ! Write 'z_normal' GrADS descriptor file?
         iq = 3
         if (file_used(iq)) then
            open (IO_unit, file=trim(g_hfn(iq)), status='unknown',  &
                 form='formatted')
            call postp_NetCDF_wGhfile(IO_unit, nc_dfn, 'sigma_zt', tinc,  &
                 mnsdy, UT, n_g_hfn_vars(iq))
            do iv=1,cfg_n_ov    ! Loop over # of variables to process
               if (var_in_file(iv,iq)) then
                  if (cfg_vtype(iv) == 4) then
                     do ip=1,cfg_spatial(i)%n_P
                        write (c_ip, '(a2,i1.1)') '_p', ip
                        tmp_mNc_str = trim(cfg_vname(iv))//c_ip
                        nc = len_trim(tmp_mNc_str)
                        e = min(12, max(1, nc-3)) ! Truncate, keeping patch ID
                        write (IO_unit, '(3a)') tmp_mNc_str(1:nc), '=>',  &
                             tmp_mNc_str(1:e)//c_ip
                     end do
                  else
                     nc = len_trim(cfg_vname(iv)); e = min(15, nc)
                     write (IO_unit, '(3a)') cfg_vname(iv)(1:nc), '=>',  &
                          cfg_vname(iv)(1:e)
                  end if
               end if
            end do
            write (IO_unit, '(a)') 'ENDVARS'
            close (IO_unit)
            write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
                 g_hfn(iq)(1:len_trim(g_hfn(iq)))
         end if
      
         call postp_NetCDF_f_auxv_tind(igd, i, ifn)      
      end if

      call postp_NetCDF_f_auxv_tdep(i, ifn, rfcnt, UT)
 
      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         ! Read/compute var data:
         call postp_varlib(igd, i, ifn, cfg_vname(iv), .FALSE.)

         ! Output variable:
         call postp_NetCDF_varout(i, iv, rfcnt)

      end do

   end do

   ! Conclude GrADS file writes:

   write (*, '(/)')
   do iv=1,maxnGrADS_files           
      if (file_used(iv)) then    ! Display GrADS header file name
         write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
              g_hfn(iv)(1:len_trim(g_hfn(iv)))
      end if
   end do

   write (*, '(/,2a)') 'Wrote NetCDF-format file:  ', trim(nc_dfn)
   write (*, '(/)')

   call mNc_sync_with_file(output_file_obj)   ! Optional

   ! Finalize file, free object memory:
   call mNc_recycle_f_md_obj(output_file_obj)

end do

! Recycle data structures:
call manage_PRAMS_vdata(-1)
call manage_cfg_v_memory(-1, -1) 

! Deallocate work array:
deallocate(PRAMS_vdata, var_in_file, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

end subroutine postp_NetCDF_driv


!```````````````````````````````````````````````````````````````````
subroutine postp_NetCDF_f_define(nc_fpath, cfg_igd)

!+ Defines a NetCDF-format file for postp output

use univ_kind_defs_mod, only: f4, f8
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use manage_NetCDF_classic_mod, only: mNc_MAX_NAME_LEN, mNc_file_md_Att_obj,  &
     mNc_define_f_md_Att_obj, mNc_populate_f_md_obj_w, mNc_FLOAT_type,  &
     mNc_recycle_f_md_Att_obj, mNc_DOUBLE_type, mNc_f_md_Att_populated
use postp_main_memory_mod, only: cfg_vtype, undef_value_f4, cfg_vname,  &
     cfg_vdesc, cfg_vunits, cfg_n_ov, cfg_spatial
use main_memory_mod, only: PRAMS_version

implicit none

!=== Argument declarations:

integer, intent(IN) :: cfg_igd
character(len=*), intent(IN) :: nc_fpath

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_NetCDF_f_define'

integer :: e, x_di, y_di, azt_di, sz_di, p_di, iv, i, ia, ic, i_d1, i_d2,   &
     t_di, azm_di, ip
logical :: az_dim_present, p_dim_present, sz_dim_present
real(f4) :: tmp_f4
real(f8) :: tmp_f8

! Items for storage/processing of NetCDF metadata:
integer :: n_dims, n_vars, n_g_Atts, max_var_ndims, max_var_nAtts
character(len=mNc_MAX_NAME_LEN), allocatable, dimension(:) :: dim_name,  &
     var_name
integer, allocatable, dimension(:) :: dimv, var_type, var_ndims, var_nAtts
integer, allocatable, dimension(:,:) :: var_dimids
type(mNc_file_md_Att_obj), allocatable, dimension(:) :: g_Att
type(mNc_file_md_Att_obj), allocatable, dimension(:,:) :: var_Att
character(len=mNc_MAX_NAME_LEN) :: tmp_mNc_str
character(len=3) :: c_ip

!=== Executable statements:

az_dim_present = count(cfg_vtype == 3) > 0
p_dim_present = count(cfg_vtype > 3) > 0
sz_dim_present = count(cfg_vtype == 5) > 0

! Total number of dimensions, variables, and global attributes:
n_dims = 3+count((/az_dim_present, az_dim_present, p_dim_present,  &
     sz_dim_present/))
n_vars = 5+count((/az_dim_present, az_dim_present, sz_dim_present/))+  &
     count(cfg_vtype == 2)+count(cfg_vtype == 3)+  &
     (count(cfg_vtype == 4)+count(cfg_vtype == 5))*cfg_spatial(cfg_igd)%n_P
n_g_Atts = 1

! Allocate and define "dimensions":
allocate(dim_name(n_dims), dimv(n_dims), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
     'dim_n_dims')
i = 1; dim_name(i) = 'time'; dimv(i) = -1; t_di = i
i = 2; dim_name(i) = 'x'; dimv(i) = cfg_spatial(cfg_igd)%n_X; x_di = i
i = 3; dim_name(i) = 'y'; dimv(i) = cfg_spatial(cfg_igd)%n_Y; y_di = i
if (az_dim_present) then
   i = i+1; dim_name(i) = 'sigma_zt'; dimv(i) = cfg_spatial(cfg_igd)%n_AZ
   azt_di = i
   i = i+1; dim_name(i) = 'sigma_zm'; dimv(i) = cfg_spatial(cfg_igd)%n_AZ
   azm_di = i
end if
if (p_dim_present) then
   i = i+1; dim_name(i) = 'patches'; dimv(i) = cfg_spatial(cfg_igd)%n_P
   p_di = i
end if
if (sz_dim_present) then
   i = i+1; dim_name(i) = 'z_SBPS'; dimv(i) = cfg_spatial(cfg_igd)%n_SZ
   sz_di = i
end if

! Allocate some variable- and global_attribute-related items:
allocate(var_name(n_vars), var_type(n_vars), var_ndims(n_vars),  &
     var_nAtts(n_vars), stat=e) 
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
     'dim_n_vars')
allocate(g_Att(n_g_Atts), stat=e) 
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
     'dim_n_g_Atts')

! Define "global attributes":
i = 1; tmp_mNc_str = trim(PRAMS_version); ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(g_Att(i), 'PRAMS_version', i, ic, tmp_mNc_str)

! Allocate variable-related items:
max_var_ndims = 5; max_var_nAtts = 4
allocate(var_dimids(max_var_ndims, n_vars), stat=e) 
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
     'dim_max_var_ndims')
allocate(var_Att(max_var_nAtts, n_vars), stat=e) 
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
     'dim_max_var_nAtts')

! Define "variables":

i = 1; var_name(i) = 'time'; var_type(i) = mNc_DOUBLE_type
var_ndims(i) = 1; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ t_di /)
ia = 1; tmp_mNc_str = 'elapsed time (in run)'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 's'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
tmp_f8 = real(undef_value_f4, kind=f8)
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ tmp_f8 /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ tmp_f8 /))

i = 2; var_name(i) = 'x'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 1; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ x_di /)
ia = 1; tmp_mNc_str = 'x grid-direction'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 'gridpoint'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
tmp_f4 = undef_value_f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ tmp_f4 /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ tmp_f4 /))

i = 3; var_name(i) = 'y'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 1; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ y_di /)
ia = 1; tmp_mNc_str = 'y grid-direction'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 'gridpoint'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
tmp_f4 = undef_value_f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ tmp_f4 /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ tmp_f4 /))

i = 4; var_name(i) = 'UT'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 1; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ t_di /)
ia = 1; tmp_mNc_str = 'UT'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 'Mars-hours'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
tmp_f4 = undef_value_f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ tmp_f4 /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ tmp_f4 /))

i = 5; var_name(i) = 'rtgt'; var_type(i) = mNc_FLOAT_type
var_ndims(i) = 2; var_nAtts(i) = 4
var_dimids(1:var_ndims(i),i) = (/ x_di, y_di /)
ia = 1; tmp_mNc_str = 'sigma_z to z conversion factor'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
ia = 2; tmp_mNc_str = 'none'; ic = len_trim(tmp_mNc_str)
call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
tmp_f4 = undef_value_f4
ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia, 1,  &
     (/ tmp_f4 /))
ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
     (/ tmp_f4 /))

if (az_dim_present) then
   i = i+1; var_name(i) = 'sigma_zt'; var_type(i) = mNc_FLOAT_type
   var_ndims(i) = 1; var_nAtts(i) = 4
   var_dimids(1:var_ndims(i),i) = (/ azt_di /)
   ia = 1; tmp_mNc_str = 'sigma_zt (raw T-pt vertical grid)'
   ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
   ia = 2; tmp_mNc_str = 'm'; ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
   tmp_f4 = undef_value_f4
   ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
        1, (/ tmp_f4 /))
   ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
        (/ tmp_f4 /))

   i = i+1; var_name(i) = 'sigma_zm'; var_type(i) = mNc_FLOAT_type
   var_ndims(i) = 1; var_nAtts(i) = 4
   var_dimids(1:var_ndims(i),i) = (/ azm_di /)
   ia = 1; tmp_mNc_str = 'sigma_zm (raw w-pt vertical grid)'
   ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
   ia = 2; tmp_mNc_str = 'm'; ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
   tmp_f4 = undef_value_f4
   ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
        1, (/ tmp_f4 /))
   ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
        (/ tmp_f4 /))
end if

if (sz_dim_present) then
   i = i+1; var_name(i) = 'z_SBPS'; var_type(i) = mNc_FLOAT_type
   var_ndims(i) = 1; var_nAtts(i) = 4
   var_dimids(1:var_ndims(i),i) = (/ sz_di /)
   ia = 1; tmp_mNc_str = 'SBPS vertical grid (positive is up)'; ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic, tmp_mNc_str)
   ia = 2; tmp_mNc_str = 'm'; ic = len_trim(tmp_mNc_str)
   call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
   tmp_f4 = undef_value_f4
   ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value', ia,  &
        1, (/ tmp_f4 /))
   ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia, 1,  &
        (/ tmp_f4 /))
end if

do iv=1,cfg_n_ov    ! Loop over # of variables to process
   select case (cfg_vtype(iv))
   case (2)
      i = i+1; var_name(i) = cfg_vname(iv); var_type(i) = mNc_FLOAT_type
      var_nAtts(i) = 4; var_ndims(i) = 3
      var_dimids(1:var_ndims(i),i) = (/ x_di, y_di, t_di /)
      ia = 1; tmp_mNc_str = cfg_vdesc(iv)(1:len_trim(cfg_vdesc(iv))-1)
      ic = len_trim(tmp_mNc_str)
      call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
           tmp_mNc_str)
      ia = 2; tmp_mNc_str = cfg_vunits(iv)(1:len_trim(cfg_vunits(iv))-1)
      ic = len_trim(tmp_mNc_str)
      call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
      tmp_f4 = undef_value_f4
      ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value',  &
           ia, 1, (/ tmp_f4 /))
      ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia,  &
           1, (/ tmp_f4 /))
   case (3)
      i = i+1; var_name(i) = cfg_vname(iv); var_type(i) = mNc_FLOAT_type
      var_nAtts(i) = 4; var_ndims(i) = 4
      if (trim(cfg_vname(iv)) == 'w') then
         var_dimids(1:var_ndims(i),i) = (/ azm_di, x_di, y_di, t_di /)
      else
         var_dimids(1:var_ndims(i),i) = (/ azt_di, x_di, y_di, t_di /)
      end if
      ia = 1; tmp_mNc_str = cfg_vdesc(iv)(1:len_trim(cfg_vdesc(iv))-1)
      ic = len_trim(tmp_mNc_str)
      call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
           tmp_mNc_str)
      ia = 2; tmp_mNc_str = cfg_vunits(iv)(1:len_trim(cfg_vunits(iv))-1)
      ic = len_trim(tmp_mNc_str)
      call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic, tmp_mNc_str)
      tmp_f4 = undef_value_f4
      ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value',  &
           ia, 1, (/ tmp_f4 /))
      ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue', ia,  &
           1, (/ tmp_f4 /))
   case (4)
      do ip=1,dimv(p_di)
         write (c_ip, '(a2,i1.1)') '_p', ip
         i = i+1; var_name(i) = trim(cfg_vname(iv))//c_ip
         var_type(i) = mNc_FLOAT_type; var_nAtts(i) = 4; var_ndims(i) = 3
         var_dimids(1:var_ndims(i),i) = (/ x_di, y_di, t_di /)
         ia = 1; c_ip(1:1) = ' '
         tmp_mNc_str = cfg_vdesc(iv)(1:len_trim(cfg_vdesc(iv))-1)//c_ip
         ic = len_trim(tmp_mNc_str)
         call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
              tmp_mNc_str)
         ia = 2; tmp_mNc_str = cfg_vunits(iv)(1:len_trim(cfg_vunits(iv))-1)
         ic = len_trim(tmp_mNc_str)
         call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic,  &
              tmp_mNc_str)
         tmp_f4 = undef_value_f4
         ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value',  &
              ia, 1, (/ tmp_f4 /))
         ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue',  &
              ia, 1, (/ tmp_f4 /))
      end do
   case (5)
      do ip=1,dimv(p_di)
         write (c_ip, '(a2,i1.1)') '_p', ip
         i = i+1; var_name(i) = trim(cfg_vname(iv))//c_ip
         var_type(i) = mNc_FLOAT_type; var_nAtts(i) = 4; var_ndims(i) = 4
         var_dimids(1:var_ndims(i),i) = (/ sz_di, x_di, y_di, t_di /)
         ia = 1; c_ip(1:1) = ' '
         tmp_mNc_str = cfg_vdesc(iv)(1:len_trim(cfg_vdesc(iv))-1)//c_ip
         ic = len_trim(tmp_mNc_str)
         call mNc_define_f_md_Att_obj(var_Att(ia,i), 'long_name', ia, ic,  &
              tmp_mNc_str)
         ia = 2; tmp_mNc_str = cfg_vunits(iv)(1:len_trim(cfg_vunits(iv))-1)
         ic = len_trim(tmp_mNc_str)
         call mNc_define_f_md_Att_obj(var_Att(ia,i), 'units', ia, ic,  &
              tmp_mNc_str)
         tmp_f4 = undef_value_f4
         ia = 3; call mNc_define_f_md_Att_obj(var_Att(ia,i), 'missing_value',  &
              ia, 1, (/ tmp_f4 /))
         ia = 4; call mNc_define_f_md_Att_obj(var_Att(ia,i), '_FillValue',  &
              ia, 1, (/ tmp_f4 /))
      end do
   end select
end do

! Begin writing the output file:
call mNc_populate_f_md_obj_w(nc_fpath, output_file_obj, n_dims, n_vars,  &
     n_g_Atts, dim_name, dimv, g_Att, var_name, var_type, var_ndims,  &
     var_nAtts, var_dimids, var_Att)

! Must "recycle" the attribute objects before reusing:
do i_d2=1,size(var_Att, 2)
   do i_d1=1,size(var_Att, 1)
      if (mNc_f_md_Att_populated(var_Att(i_d1,i_d2)))  &
           call mNc_recycle_f_md_Att_obj(var_Att(i_d1,i_d2))
   end do
end do
do i_d1=1,size(g_Att, 1)
   if (mNc_f_md_Att_populated(g_Att(i_d1))) call mNc_recycle_f_md_Att_obj(  &
        g_Att(i_d1))
end do

deallocate(dim_name, dimv, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_n_dims')
deallocate(var_name, var_type, var_ndims, var_nAtts, stat=e) 
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_n_vars')
deallocate(g_Att, stat=e) 
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_n_g_Atts')    
deallocate(var_dimids, stat=e) 
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_max_var_ndims')
deallocate(var_Att, stat=e) 
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
     'dim_max_var_nAtts')

end subroutine postp_NetCDF_f_define


!```````````````````````````````````````````````````````````````````
subroutine postp_NetCDF_f_auxv_tdep(cfg_igd, cfg_ifn, tgt_ifn, UT)

!+ Writes time-dependent auxiliary variable data (not postprocessed fields) to
!   NetCDF-format file

use univ_kind_defs_mod, only: f4
!%use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use manage_NetCDF_classic_wv_mod, only: mNc_write_var
use postp_main_memory_mod, only: model_sim_ET

implicit none

!=== Argument declarations:

integer, intent(IN) :: cfg_igd, tgt_ifn, cfg_ifn
real(f4), intent(IN) :: UT

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'postp_NetCDF_f_auxv_tdep'

!=== Executable statements:

call mNc_write_var(output_file_obj, 'time', model_sim_ET(cfg_ifn),  &
     opt_start=(/tgt_ifn/), opt_count=(/1/))

call mNc_write_var(output_file_obj, 'UT', UT,  &
     opt_start=(/tgt_ifn/), opt_count=(/1/))

end subroutine postp_NetCDF_f_auxv_tdep


!```````````````````````````````````````````````````````````````````
subroutine postp_NetCDF_f_auxv_tind(tgt_igd, cfg_igd, cfg_ifn)

!+ Writes time-independent auxiliary variable data (not postprocessed fields) to
!   NetCDF-format file

use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use manage_NetCDF_classic_mod, only: mNc_var_idx
use manage_NetCDF_classic_wv_mod, only: mNc_write_var
use postp_main_memory_mod, only: cfg_spatial, vl_scr1_t2, zt, zm, SBPS_Tlev,  &
     vctr1_sz, vctr1_az
use postp_comp_mod, only: comp_rtgX
use postp_getdata_mod, only: get_PRAMS_data

implicit none

!=== Argument declarations:

integer, intent(IN) :: tgt_igd, cfg_igd, cfg_ifn

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_NetCDF_f_auxv_tind'

integer :: i, e
real(f4), allocatable, dimension(:) :: x_gp, y_gp

!=== Executable statements:

allocate(x_gp(cfg_spatial(cfg_igd)%n_X), y_gp(cfg_spatial(cfg_igd)%n_Y), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

do i=1,size(x_gp, 1)
   x_gp(i) = real(i, kind=f4)
end do
do i=1,size(y_gp, 1)
   y_gp(i) = real(i, kind=f4)
end do

call mNc_write_var(output_file_obj, 'x', x_gp)
call mNc_write_var(output_file_obj, 'y', y_gp)

do i=1,cfg_spatial(cfg_igd)%n_AZ
   vctr1_az%p1D(i) = zt(cfg_spatial(cfg_igd)%i_AZ(i),tgt_igd,cfg_ifn)
end do
if (mNc_var_idx(output_file_obj, 'sigma_zt', neg1_when_not_found=.TRUE.) > 0)  &
     call mNc_write_var(output_file_obj, 'sigma_zt', vctr1_az%p1D)

do i=1,cfg_spatial(cfg_igd)%n_AZ
   vctr1_az%p1D(i) = zm(cfg_spatial(cfg_igd)%i_AZ(i),tgt_igd,cfg_ifn)
end do
if (mNc_var_idx(output_file_obj, 'sigma_zm', neg1_when_not_found=.TRUE.) > 0)  &
     call mNc_write_var(output_file_obj, 'sigma_zm', vctr1_az%p1D)

do i=1,cfg_spatial(cfg_igd)%n_SZ
   vctr1_sz%p1D(i) = SBPS_Tlev(cfg_spatial(cfg_igd)%i_SZ(i),cfg_ifn)
end do
if (mNc_var_idx(output_file_obj, 'z_SBPS', neg1_when_not_found=.TRUE.) > 0)  &
     call mNc_write_var(output_file_obj, 'z_SBPS', vctr1_sz%p1D)

call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, cfg_ifn)
call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2
call mNc_write_var(output_file_obj, 'rtgt', vl_scr1_t2%p2D)

deallocate(x_gp, y_gp, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

end subroutine postp_NetCDF_f_auxv_tind


!```````````````````````````````````````````````````````````````````
subroutine postp_NetCDF_varout(cfg_igd, cfg_iv, tgt_ifn)

!+ Writes postprocessed variable data to NetCDF-format file

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL
use manage_NetCDF_classic_mod, only: mNc_MAX_NAME_LEN
use manage_NetCDF_classic_wv_mod, only: mNc_write_var
use postp_main_memory_mod, only: what_vtype, vl_result_t2, vl_result_t3,  &
     vl_result_t4, vl_result_t5, cfg_spatial, cfg_vname

implicit none

!=== Argument declarations:

integer, intent(IN) :: cfg_igd, tgt_ifn, cfg_iv

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_NetCDF_varout'

integer :: nx, ny, naz, nsz, ip
character(len=3) :: c_ip
character(len=mNc_MAX_NAME_LEN) :: tmp_mNc_str

!=== Executable statements:

nx = cfg_spatial(cfg_igd)%n_X; ny = cfg_spatial(cfg_igd)%n_Y

which_vtype: select case(what_vtype)
case (2)  ! "regular" 2D variable

   call mNc_write_var(output_file_obj, cfg_vname(cfg_iv), vl_result_t2%p2D,  &
        opt_start=(/1,1,tgt_ifn/), opt_count=(/nx,ny,1/))

case (3)  ! "regular" 3D variable

   naz = cfg_spatial(cfg_igd)%n_AZ
   call mNc_write_var(output_file_obj, cfg_vname(cfg_iv), vl_result_t3%p3D,  &
        opt_start=(/1,1,1,tgt_ifn/), opt_count=(/naz,nx,ny,1/))

case (4)  ! SBPS 2D variable

   do ip=1,cfg_spatial(cfg_igd)%n_P
      write (c_ip, '(a2,i1.1)') '_p', ip
      tmp_mNc_str = trim(cfg_vname(cfg_iv))//c_ip
      call mNc_write_var(output_file_obj, tmp_mNc_str,  &
           vl_result_t4%p3D(:,:,ip),  &
           opt_start=(/1,1,tgt_ifn/), opt_count=(/nx,ny,1/))
   end do

case (5)  ! SBPS 3D variable

   nsz = cfg_spatial(cfg_igd)%n_SZ
   do ip=1,cfg_spatial(cfg_igd)%n_P
      write (c_ip, '(a2,i1.1)') '_p', ip
      tmp_mNc_str = trim(cfg_vname(cfg_iv))//c_ip
      call mNc_write_var(output_file_obj, tmp_mNc_str,  &
           vl_result_t5%p4D(:,:,:,ip),  &
           opt_start=(/1,1,1,tgt_ifn/), opt_count=(/nsz,nx,ny,1/))
   end do

case default

   write (*, '(/a,i1,a)') 'ERROR: unknown variable type ( = ', what_vtype, ' ).'
   call exception(procedure_name, FATAL)

end select which_vtype

end subroutine postp_NetCDF_varout


!```````````````````````````````````````````````````````````````````
subroutine postp_NetCDF_wGhfile(IO_unit, nc_dfn, z_var_str, tinc_sec,  &
     dp_sec, UT, n_g_hfn_v)

!+

use univ_kind_defs_mod, only: f4, f8
use postp_main_memory_mod, only: cfg_n_f

implicit none

!=== Argument declarations:

integer, intent(IN) :: IO_unit, n_g_hfn_v
character(len=*), intent(IN) :: nc_dfn, z_var_str
real(f8), intent(IN) :: tinc_sec, dp_sec
real(f4), intent(IN) :: UT

!== Local declarations:

integer, parameter :: i_hours_per_dp = 24, i_minutes_per_hour = 60,  &
     i_minutes_per_dp = i_hours_per_dp*i_minutes_per_hour

real(f8), parameter :: hours_per_dp = i_hours_per_dp,  &
                       minutes_per_hour = i_minutes_per_hour,  &
                       minutes_per_dp = i_minutes_per_dp

integer :: i, ihh, imin
real(f8) :: t_dp, t_hour, t_min
character(len=2) :: t_unit_str
character(len=4) :: t_inc_str

!=== Executable statements:

i = max(1, index(nc_dfn, '/', back=.TRUE.)+1) ! Remove any dir(s)
write (IO_unit, '(2a)') 'DSET ^', nc_dfn(i:len_trim(nc_dfn))

write (IO_unit, '(a)') 'DTYPE netcdf'
write (IO_unit, '(a)') 'TITLE generated by postp'
write (IO_unit, '(a)') 'UNDEF 1.e+34 missing_value'
write (IO_unit, '(a)') 'XDEF x'
write (IO_unit, '(a)') 'YDEF y'
write (IO_unit, '(2a)') 'ZDEF ', z_var_str

! Create TDEF entry:

t_dp = dp_sec   ! body-day [s]
t_hour = t_dp/hours_per_dp     ! body-hour [s]
t_min = t_dp/minutes_per_dp   ! body-minute [s]

i = max(1, nint(tinc_sec/t_min)); t_unit_str = 'mn'
if (i > 99) then
   i = max(1, nint(tinc_sec/t_hour))
   t_unit_str = 'hr'
   if (i >= i_hours_per_dp) then
      i = max(1, nint(tinc_sec/t_dp))
      t_unit_str = 'dy'
   end if
end if
write (t_inc_str, '(i2.2,a2)') i, t_unit_str    ! Approximate time interval

ihh = int(UT)
imin = int((UT-real(ihh, kind=f4))*real(minutes_per_hour, kind=f4))
do while (imin >= i_minutes_per_hour) 
   imin = imin-i_minutes_per_hour
   ihh = ihh+1
end do
do while (ihh >= i_hours_per_dp) 
   ihh = ihh-i_hours_per_dp
end do

write (IO_unit, '(a,i5,a,i2.2,a1,i2.2,2a)') 'TDEF time ', cfg_n_f,  &
     ' linear ', ihh, ':', imin, 'Z04jul1997 ', t_inc_str

write (IO_unit, '(a,i3)') 'VARS ', n_g_hfn_v

end subroutine postp_NetCDF_wGhfile


end module postp_NetCDF_mod
