!```````````````````````````````````````````````````````````````````
module postp_OpenDX_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine postp_OpenDX_driv

!+ Post-processing driver for OpenDX format output

use univ_kind_defs_mod, only: f4, f8
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_integer_time_Mod, only: time_obj, fp_seconds_eqv_Time
use univ_exception_mod, only: exception, FATAL
use univ_std_fileIO_mod, only: find_file_IOunit, mkdir_p, determine_IO_reclen
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN
use univ_namelist_utils_mod, only: Time_to_gen_time_strrep
use body_specific_constants_mod, only: mnsdy
use postp_main_memory_mod, only: cfg_f, cfg_f_inc, cfg_n_ov, state_file_name,  &
     model_sim_ET, cfg_n_f, cfg_vname, nxp, nyp, nvar_read, what_vtype,  &
     cfg_vtype, zt, vctr1_az, var_want_to_read, nvar_want, cfg_vlevs,  &
     cfg_vdesc, cfg_vunits, what_vlevs, what_vdesc, what_vunits, what_ofile,  &
     SBPS_Tlev, ztop, pfact_xa, pfact_ya,  &
     pfact_za, pfact_xb, pfact_yb, pfact_zb, grid_idx_list, vctr1_sz,  &
     output_pp_file_pfx, cfg_n_g, PRAMS_vdata, vl_result_t3, vl_scr1_t2,  &
     cfg_spatial, use_full_az_scr, use_full_xy_scr, use_full_azxy_scr,  &
     use_full_sz_scr, full_az_scr_needed, full_xy_scr_needed,  &
     full_azxy_scr_needed, full_sz_scr_needed
use univ_OpenDX_format_mod, only: OpenDX_init_mod, OpenDX_init_pofobj,  &
     OpenDX_chk_pofobj, OpenDX_set_pofobj, OpenDX_C_Wpos,  &
     OpenDX_C_init_dfobj, OpenDX_Csclr_Whdr, OpenDX_end_dfobj,  &
     OpenDX_end_mod, OpenDX_Cvect_Whdr, OpenDX_Cvect_Wdat
use main_memory_mod, only: module_tps, start_UT, mnsdy_Time, morbpd_Time
use postp_varlib_mod, only: postp_varlib
use postp_driv_mod, only: manage_PRAMS_vdata, manage_cfg_v_memory
use postp_getdata_mod, only: count_vars_to_read, fill_PRAMS_data, get_PRAMS_data

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_OpenDX_driv'

integer :: igd, iv, ifn, lnoutp, lnstp, in_unit, e, ic, rfcnt, IO_reclen,  &
     max_nzr, max_nxr, max_nyr, nzr, cpof_ID, iu, wvect_iv, i
character(len=MAX_FILE_PATH_LEN) :: dfn, hfn, pfn, s_dfn
character(len=7) :: cgrid_a
character(len=8) :: csfx_pof
character(len=9) :: cgrid_o
character(len=40) :: gen_time_strrep
type(time_obj) :: tmp_Time
logical :: is_sfc_var, pof_not_set, do_wvect
real(f4), allocatable, dimension(:,:,:) :: pos_x, pos_y, pos_z
real(f4), allocatable, dimension(:,:) :: reclen_test
real(f4), allocatable, dimension(:) :: time_tags
real(f4), allocatable, dimension(:,:,:,:) :: wvect
integer, allocatable, dimension(:) :: iu_list

!=== Executable statements:

! Make initial pass through all grids, files, and variables; record what
!  variables need to be read in and where they may be found:

nvar_read = 0   ! Start out by assuming that no vars need read in
use_full_az_scr = .FALSE.; use_full_xy_scr = .FALSE.
use_full_azxy_scr = .FALSE.; use_full_sz_scr = .FALSE.

do_wvect = .FALSE.   ! Initially assume that full wind vector data not desired
wvect_iv = 0

do i=1,cfg_n_g   ! Loop over all desired grids
   igd = grid_idx_list(i)
   do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files

      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         if (cfg_vname(iv) /= 'wind_vector') then

            ! What state variable(s) does this variable require? 
            call postp_varlib(igd, i, ifn, cfg_vname(iv), .TRUE.)
            use_full_az_scr(i) = ( use_full_az_scr(i) .or. full_az_scr_needed )
            use_full_xy_scr(i) = ( use_full_xy_scr(i) .or. full_xy_scr_needed )
            use_full_azxy_scr(i) = (use_full_azxy_scr(i) .or. full_azxy_scr_needed)
            use_full_sz_scr(i) = ( use_full_sz_scr(i) .or. full_sz_scr_needed )

            ! Check if var(s) is configured for reading; if not, configure:
            call count_vars_to_read(igd, ifn)

         else      !  Full wind vector is desired

            wvect_iv = iv; do_wvect = .TRUE.
            call postp_varlib(igd, i, ifn, 'u_avg', .TRUE.)
            use_full_az_scr(i) = ( use_full_az_scr(i) .or. full_az_scr_needed )
            use_full_xy_scr(i) = ( use_full_xy_scr(i) .or. full_xy_scr_needed )
            use_full_azxy_scr(i) = (use_full_azxy_scr(i) .or. full_azxy_scr_needed)
            use_full_sz_scr(i) = ( use_full_sz_scr(i) .or. full_sz_scr_needed )
            call count_vars_to_read(igd, ifn)
            call postp_varlib(igd, i, ifn, 'v_avg', .TRUE.)
            use_full_az_scr(i) = ( use_full_az_scr(i) .or. full_az_scr_needed )
            use_full_xy_scr(i) = ( use_full_xy_scr(i) .or. full_xy_scr_needed )
            use_full_azxy_scr(i) = (use_full_azxy_scr(i) .or. full_azxy_scr_needed)
            use_full_sz_scr(i) = ( use_full_sz_scr(i) .or. full_sz_scr_needed )
            call count_vars_to_read(igd, ifn)
            call postp_varlib(igd, i, ifn, 'w_avg', .TRUE.)
            use_full_az_scr(i) = ( use_full_az_scr(i) .or. full_az_scr_needed )
            use_full_xy_scr(i) = ( use_full_xy_scr(i) .or. full_xy_scr_needed )
            use_full_azxy_scr(i) = (use_full_azxy_scr(i) .or. full_azxy_scr_needed)
            use_full_sz_scr(i) = ( use_full_sz_scr(i) .or. full_sz_scr_needed )
            call count_vars_to_read(igd, ifn)

         end if

      end do

      ! Always need topo values:
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TOPT'/)
      call count_vars_to_read(igd, ifn)

   end do
end do

! Allocate some data storage:
allocate(PRAMS_vdata(nvar_read), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

! Args: Number of output data files (one for each var); Currently 4 possible
!  variable types, so 4 possible position sets:
call OpenDX_init_mod(cfg_n_ov, 4)

max_nzr = -1; max_nxr = -1; max_nyr = -1
do i=1,cfg_n_g
   igd = grid_idx_list(i)
   max_nzr = max(max_nzr, cfg_spatial(i)%n_AZ, cfg_spatial(i)%n_SZ)
   max_nxr = max(max_nxr, cfg_spatial(i)%n_X)
   max_nyr = max(max_nyr, cfg_spatial(i)%n_Y)
end do

allocate(pos_x(max_nxr,max_nyr,max_nzr), pos_y(max_nxr,max_nyr,max_nzr),  &
     pos_z(max_nxr,max_nyr,max_nzr), time_tags(cfg_n_f), iu_list(cfg_n_ov),  &
     wvect(max_nxr,max_nyr,max_nzr,3), stat=e)
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '2')

! Calculate time tags for descriptor file:
call calc_time_tags(cfg_n_f, model_sim_ET(cfg_f(1):cfg_f(2):cfg_f_inc),  &
     time_tags)

! Main pass; read in data, calculate and output POSTP variable values:

i = index(output_pp_file_pfx, '/', back=.TRUE.)-1  ! Determine dir name length
call mkdir_p(output_pp_file_pfx(1:i))  ! Create any needed directories

lnoutp = len_trim(output_pp_file_pfx)  ! Length of string

do i=1,cfg_n_g  ! Loop over all desired grids
   igd = grid_idx_list(i)

   call manage_PRAMS_vdata(igd)    ! Initialize PRAMS_vdata for this grid
   call manage_cfg_v_memory(igd, i)  ! Init cfg-dimensioned multi-use storage

   call OpenDX_init_pofobj   ! (Re)initialize position object(s) for this grid

   write (cgrid_a, '(a1,i2.2,a4)') 'g', igd, '.vfm'  ! State file grid suffix

   write (cgrid_o, '(a7,i2.2)') '-sigC-g', igd   ! Output file grid suffix 

   rfcnt = 0
   do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
      rfcnt = rfcnt+1      ! File number (relative) counter

      tmp_Time = fp_seconds_eqv_Time(model_sim_ET(ifn), module_tps)
      call Time_to_gen_time_strrep(tmp_Time, gen_time_strrep,  &
           len(gen_time_strrep), mnsdy_Time, morbpd_Time,  &
           short_unit_style=.TRUE.)
      write (*, '(a,i5,a,i2,3a,f6.2,a)') 'Postprocessing < ', rfcnt,  &
           ' > grid #', igd, ', time = [', trim(gen_time_strrep), '] [',  &
         mod(real(model_sim_ET(ifn)/mnsdy*24._f8, kind=f4)+start_UT, 24._f4),  &
           ' UT] ...'

      lnstp = len_trim(state_file_name(ifn))-7
      s_dfn = state_file_name(ifn)(1:lnstp)//cgrid_a    ! State file name

      ! Open state data file:
      in_unit = find_file_IOunit()
      allocate(reclen_test(nxp(igd,ifn),nyp(igd,ifn)), stat=e)
      if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
           'reclen')
      IO_reclen = determine_IO_reclen(reclen_test)
      deallocate(reclen_test, stat=e)
      if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name,  &
           'reclen')
      open(in_unit, file=s_dfn(1:lnstp+7), status='old', form='unformatted',  &
           access='direct', recl=IO_reclen)

      ! Read in all *NEEDED* PRAMS data for this grid/file; fill data array:
      call fill_PRAMS_data(in_unit, ifn, igd, nxp(igd,ifn)*nyp(igd,ifn))

      close(in_unit)   ! Close state file

      ! Compute values of configured variables; output data in desired manner:

      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         ! Read/compute var data:

         if (iv /= wvect_iv) then
            call postp_varlib(igd, i, ifn, cfg_vname(iv), .FALSE.)
         else
            call postp_varlib(igd, i, ifn, 'u_avg', .FALSE.)
            call fill_Cwvect(cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
                 cfg_spatial(i)%n_AZ, 3, vl_result_t3%p3D, wvect, 1)
            call postp_varlib(igd, i, ifn, 'v_avg', .FALSE.)
            call fill_Cwvect(cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
                 cfg_spatial(i)%n_AZ, 3, vl_result_t3%p3D, wvect, 2)
            call postp_varlib(igd, i, ifn, 'w_avg', .FALSE.)
            call fill_Cwvect(cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
                 cfg_spatial(i)%n_AZ, 3, vl_result_t3%p3D, wvect, 3)
         end if

         ! Store/calculate variable descriptors:

         nzr = max(what_vlevs, 1)

         if (ifn == cfg_f(1)) then
            cfg_vtype(iv) = what_vtype    ! Store var types, etc.
            cfg_vlevs(iv) = what_vlevs    !
            cfg_vdesc(iv) = what_vdesc    !
            cfg_vunits(iv) = what_vunits  !

            ! Determine if the corresponding positions file has already been
            !  set/written; if so, return its ID#; if not, set/write it:

            call OpenDX_chk_pofobj(what_vtype, pof_not_set, cpof_ID)

            if (pof_not_set) then

               ! Create positions output file name:
               write (csfx_pof, '(a2,i2.2,a4)') '-p', what_vtype, '.pos'
               pfn = output_pp_file_pfx(1:lnoutp)//cgrid_o//csfx_pof

               ! Setup positions file object:
               iu = find_file_IOunit()
               call OpenDX_set_pofobj(what_vtype, cfg_spatial(i)%n_X,  &
                    cfg_spatial(i)%n_Y, nzr, pfn, iu)

               ! Calculate position values for output:

               is_sfc_var = ( what_vlevs == 0 ) ! This is a "sfc-valid" (2D) var

               ! Need topo values:
               call get_PRAMS_data(vl_scr1_t2, 'TOPT', igd, i, ifn)
               if (what_ofile(1:3) == 'atm') then            
                  do ic=1,cfg_spatial(i)%n_AZ
                     vctr1_az%p1D(ic) = zt(cfg_spatial(i)%i_AZ(ic),igd,1)
                  end do
                  call calc_Cpos(cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
                       nzr, is_sfc_var, vctr1_az%p1D, vl_scr1_t2%p2D,  &
                       .FALSE., ztop, pos_x, pos_y, pos_z, pfact_xa,  &
                       pfact_ya, pfact_za)
               else if (what_ofile(1:3) == 'sfc') then
                  do ic=1,cfg_spatial(i)%n_SZ
                     vctr1_sz%p1D(ic) = SBPS_Tlev(cfg_spatial(i)%i_SZ(ic),1)
                  end do
                  call calc_Cpos(cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
                       nzr, is_sfc_var, vctr1_sz%p1D, vl_scr1_t2%p2D, .TRUE.,  &
                       ztop, pos_x, pos_y, pos_z, pfact_xb, pfact_yb, pfact_zb)
               else
                  call exception(procedure_name, FATAL,  &
                       'Incorrect ofile name in varlib')
               end if

               ! Write positions data to file:

               call OpenDX_C_Wpos(cpof_ID, cfg_spatial(i)%n_X,  &
                    cfg_spatial(i)%n_Y, nzr, pos_x, pos_y, pos_z)

            end if

            ! Create output file names:
            ic = len_trim(cfg_vname(iv))
            dfn = output_pp_file_pfx(1:lnoutp)//'-'//cfg_vname(iv)(1:ic)//  &
                 cgrid_o//'.dat'
            hfn = output_pp_file_pfx(1:lnoutp)//'-'//cfg_vname(iv)(1:ic)//  &
                 cgrid_o//'.dx'

            ! Initialize OpenDX datafile object for this variable:
            iu_list(iv) = find_file_IOunit()
            if (iv /= wvect_iv) then
               call OpenDX_C_init_dfobj(iv, cfg_spatial(i)%n_X,  &
                    cfg_spatial(i)%n_Y, nzr,  &
                    cfg_n_f, dfn, hfn, cfg_vname(iv), what_vdesc,  &
                    what_vunits, cpof_ID, time_tags, iu_list(iv), 0)
            else
               call OpenDX_C_init_dfobj(iv, cfg_spatial(i)%n_X,  &
                    cfg_spatial(i)%n_Y, nzr,  &
                    cfg_n_f, dfn, hfn, cfg_vname(iv), what_vdesc,  &
                    what_vunits, cpof_ID, time_tags, iu_list(iv), 3)
            end if

            ! Write OpenDX descriptor file:
            iu = find_file_IOunit()
            if (iv /= wvect_iv) then
               call OpenDX_Csclr_Whdr(iv, iu)
            else
               call OpenDX_Cvect_Whdr(iv, iu)
            end if

         end if

         ! Output variable:
         if (iv /= wvect_iv) then
            call postp_OpenDX_varout(cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y,  &
                 nzr, 1, iv, rfcnt) 
         else
            call OpenDX_Cvect_Wdat(iv, cfg_spatial(i)%n_X,  &
                 cfg_spatial(i)%n_Y, nzr, 3, wvect, rfcnt)
         end if

      end do

   end do

   ! Conclude OpenDX data file writes:

   call OpenDX_end_dfobj
     
end do

! Exit OpenDX output module:
call OpenDX_end_mod

! Recycle data structures:
call manage_PRAMS_vdata(-1)
call manage_cfg_v_memory(-1, -1) 

! Deallocate work arrays:

deallocate(PRAMS_vdata, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

deallocate(pos_x, pos_y, pos_z, time_tags, wvect, stat=e)
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '2')

end subroutine postp_OpenDX_driv


!```````````````````````````````````````````````````````````````````
subroutine postp_OpenDX_varout(nx, ny, nz, n5, dfobj_ID, rfcnt)

!+ Writes postprocessed variable data to OpenDX format file

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_OpenDX_format_mod, only: OpenDX_Csclr_Wdat
use postp_main_memory_mod, only: what_vtype, vl_reorder_t3,  &
     vl_reorder_t5, vl_result_t2, vl_result_t3, vl_result_t4, vl_result_t5

implicit none

!=== Argument declarations:

integer, intent(IN) :: nx, ny, nz, n5, dfobj_ID, rfcnt

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_OpenDX_varout'

integer :: ip

!=== Executable statements:

which_vtype: select case(what_vtype)

case (2)  ! "regular" 2D variable

   call OpenDX_Csclr_Wdat(dfobj_ID, nx, ny, nz, vl_result_t2%p2D, rfcnt)

case (3)  ! "regular" 3D variable

   vl_reorder_t3%p3D = reshape(vl_result_t3%p3D, (/nx,ny,nz/), order=(/3,1,2/))
   call OpenDX_Csclr_Wdat(dfobj_ID, nx, ny, nz, vl_reorder_t3%p3D, rfcnt)

case (4)  ! SBPS 2D variable

   do ip=1,n5
      call OpenDX_Csclr_Wdat(dfobj_ID, nx, ny, nz, vl_result_t4%p3D(:,:,ip), &
           rfcnt)
   end do

case (5)  ! SBPS 3D variable

   do ip=1,n5
      vl_reorder_t5%p4D = reshape(vl_result_t5%p4D, (/nx,ny,nz,n5/),  &
           order=(/3,1,2,4/))
      call OpenDX_Csclr_Wdat(dfobj_ID, nx, ny, nz,  &
           vl_reorder_t5%p4D(:,:,:,ip), rfcnt)
   end do

case default

   write (*,'(/a,i1,a)') 'ERROR: unknown variable type ( = ',what_vtype,' ).' 
   call exception(procedure_name, FATAL)

end select which_vtype

end subroutine postp_OpenDX_varout


!```````````````````````````````````````````````````````````````````
subroutine calc_Cpos(nx,ny,nz,is_sfc_var,gzlev,topo,no_rtg,ztop,pos_x,pos_y  &
     ,pos_z,pfact_x,pfact_y,pfact_z)

!+ Calculates data point position vector(s) (Cartesian coords only)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ZERO_f4, ONE_f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: nx,ny,nz
real(f4), intent(IN) :: ztop,pfact_x,pfact_y,pfact_z
real(f4), dimension(nz), intent(IN) :: gzlev
logical, intent(IN) :: no_rtg,is_sfc_var
real(f4), dimension(nx,ny), intent(IN) :: topo

real(f4), dimension(nx,ny,nz), intent(OUT) :: pos_x,pos_y,pos_z

!== Local declarations:

integer :: i,j,k
real(f4) :: loc_x,loc_y,loc_z,rtg,ctopo

!=== Executable statements:

loc_y = 0.
if(is_sfc_var) then   ! pos_z(*) = 0.
   loc_z = ZERO_f4
   do j=1,ny
      loc_y = loc_y+ONE_f4
      loc_x = ZERO_f4
      do i=1,nx
         loc_x = loc_x+ONE_f4
         do k=1,nz
            pos_x(i,j,k) = loc_x*pfact_x
            pos_y(i,j,k) = loc_y*pfact_y
            pos_z(i,j,k) = loc_z        
         enddo
      enddo
   enddo
elseif(no_rtg) then   ! Vertical coord DOES NOT vary as f(x,y)
   do j=1,ny
      loc_y = loc_y+ONE_f4
      loc_x = ZERO_f4
      do i=1,nx
         loc_x = loc_x+ONE_f4
         ctopo = topo(i,j)
         do k=1,nz
            loc_z = gzlev(k)
            pos_x(i,j,k) = loc_x*pfact_x
            pos_y(i,j,k) = loc_y*pfact_y
            pos_z(i,j,k) = loc_z*pfact_z+ctopo
         enddo
      enddo
   enddo
else           ! Vertical coord DOES vary as f(x,y)
   do j=1,ny
      loc_y = loc_y+ONE_f4
      loc_x = ZERO_f4
      do i=1,nx
         loc_x = loc_x+ONE_f4
         ctopo = topo(i,j)
         rtg = ONE_f4-ctopo/ztop
         do k=1,nz
            loc_z = gzlev(k)*rtg
            pos_x(i,j,k) = loc_x*pfact_x
            pos_y(i,j,k) = loc_y*pfact_y
            pos_z(i,j,k) = loc_z*pfact_z+ctopo
        enddo
      enddo
   enddo
endif

end subroutine calc_Cpos


!```````````````````````````````````````````````````````````````````
subroutine fill_Cwvect(nx,ny,nz,ndim,wcomp,wvect,tgt_dim)

!+ Fills output wind vector array with component data (Cartesian coords only)

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: nx,ny,nz,ndim,tgt_dim
real(f4), dimension(nx,ny,nz), intent(IN) :: wcomp

real(f4), dimension(nx,ny,nz,ndim), intent(OUT) :: wvect

!== Local declarations:

integer :: i,j,k

!=== Executable statements:

do k=1,nz
   do j=1,ny
      do i=1,nx
         wvect(i,j,k,tgt_dim) = wcomp(i,j,k)
      enddo
   enddo
enddo

end subroutine fill_Cwvect


!```````````````````````````````````````````````````````````````````
subroutine calc_time_tags(mt, native_time, time_tags)

!+ Converts native time data to user-defined units

use univ_kind_defs_mod, only: f4, f8
use postp_main_memory_mod, only: ttag_fmt
use body_specific_constants_mod, only: mnsdy

implicit none

!=== Argument declarations:

integer, intent(IN) :: mt
real(f8), dimension(mt), intent(IN) :: native_time

real(f4), dimension(mt), intent(OUT) :: time_tags

!== Local declarations:

integer :: it
real(f8) :: dtmp

!=== Executable statements:

if(ttag_fmt == 'sec') then
   do it=1,mt
      time_tags(it) = real(native_time(it), kind=f4)
   enddo
elseif(ttag_fmt == 'Mmn') then
   dtmp = mnsdy/(24._f8*60._f8)
   do it=1,mt
      time_tags(it) = real(native_time(it)/dtmp, kind=f4)
   enddo
elseif(ttag_fmt == 'Mhr') then
   dtmp = mnsdy/24._f8
   do it=1,mt
      time_tags(it) = real(native_time(it)/dtmp, kind=f4)
   enddo
elseif(ttag_fmt == 'sol') then
   dtmp = mnsdy
   do it=1,mt
      time_tags(it) = real(native_time(it)/dtmp, kind=f4)
   enddo
endif

end subroutine calc_time_tags


end module postp_OpenDX_mod
