!```````````````````````````````````````````````````````````````````
module postp_driv_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine manage_PRAMS_vdata(igd)

!+ Manages 'PRAMS_vdata' memory structure

use postp_main_memory_mod, only: nzg, nzs, nxp, nyp, nzp, npatch, nvar_read,  &
     var2read_type, PRAMS_vdata, init_vdata_obj, recycle_vdata_obj, cfg_f

implicit none

!=== Argument declarations:

integer, intent(IN) :: igd   ! If -1, recycle all elements

!== Local declarations:

integer :: iv, tgt_f
integer, dimension(2) :: dims_t2
integer, dimension(3) :: dims_t3, dims_t4
integer, dimension(4) :: dims_t5

!=== Executable statements:

if (igd == -1) then    ! Recycle all elements

   do iv=1,nvar_read
      if (PRAMS_vdata(iv)%is_populated) call recycle_vdata_obj(PRAMS_vdata(iv))
   end do

else         ! (Re)initialize all elements

   tgt_f = cfg_f(1)

   ! Relevant dimensions for various variable types (and this grid):
   dims_t2 = (/ nxp(igd,tgt_f), nyp(igd,tgt_f) /)
   dims_t3 = (/ nzp(igd,tgt_f), nxp(igd,tgt_f), nyp(igd,tgt_f) /)
   dims_t4 = (/ nxp(igd,tgt_f), nyp(igd,tgt_f), npatch(tgt_f) /)
   dims_t5 = (/ nzg(tgt_f)+nzs(tgt_f), nxp(igd,tgt_f),  &
        nyp(igd,tgt_f), npatch(tgt_f) /)

   ! Prepare PRAMS_vdata elements for use with the given grid:
   do iv=1,nvar_read
      if (PRAMS_vdata(iv)%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(PRAMS_vdata(iv))
      
      select case (var2read_type(iv))
      case(2)
         call init_vdata_obj(PRAMS_vdata(iv), dims_t2)
      case(3)
         call init_vdata_obj(PRAMS_vdata(iv), dims_t3)
      case(4)
         call init_vdata_obj(PRAMS_vdata(iv), dims_t4)
      case(5)
         call init_vdata_obj(PRAMS_vdata(iv), dims_t5)
      end select
   end do

end if

end subroutine manage_PRAMS_vdata


!```````````````````````````````````````````````````````````````````
subroutine manage_cfg_v_memory(tgt_igd, cfg_igd)

!+ Manages cfg-dimensioned memory structures

use postp_main_memory_mod, only: vl_result_t2, vl_result_t3, vl_result_t4,  &
     vl_result_t5, init_vdata_obj, recycle_vdata_obj, vl_reorder_t2,  &
     vl_reorder_t3, vl_reorder_t4, vl_reorder_t5, vl_scr1_t2, vl_scr2_t2,  &
     vl_scr1_t3, vl_scr2_t3, vl_scr3_t3, vl_scr4_t3, vl_scr5_t3, vl_scr6_t3,  &
     vl_scr7_t3, vl_scr1_t4, vl_scr2_t4, vl_scr3_t4, vl_scr4_t4, vl_scr3_t2,  &
     vl_scr1_t5, vl_scr2_t5, vl_scr3_t5, vctr1_az, vctr2_az, vctr3_az,  &
     vctr4_az, vctr1_x, vctr1_y, vctr1_sz, cfg_spatial, pp_action, nzp,  &
     vl_scr1_faz_t3, vl_scr2_faz_t3, vl_scr3_faz_t3, vl_scr4_faz_t3,  &
     use_full_az_scr, cfg_f, use_full_sz_scr, vl_scr1_fsz_t5, vl_scr2_fsz_t5,  &
     vl_scr1_fxy_t3, vl_scr2_fxy_t3, use_full_xy_scr, nxp, nyp, nzg, nzs,  &
     use_full_azxy_scr, vl_scr1_fazxy_t3, vl_scr2_fazxy_t3, vl_scr3_fsz_t5,  &
     vl_scr4_fsz_t5

implicit none

!=== Argument declarations:

integer, intent(IN) :: tgt_igd,  &
                       cfg_igd   ! If < 0, recycle all elements

!== Local declarations:

integer :: tgt_f
integer, dimension(2) :: dims_t2
integer, dimension(3) :: dims_t3, dims_t4
integer, dimension(4) :: dims_t5

!=== Executable statements:

if (cfg_igd < 0) then    ! Recycle all

   if (vl_result_t2%is_populated) call recycle_vdata_obj(vl_result_t2)
   if (vl_result_t3%is_populated) call recycle_vdata_obj(vl_result_t3)
   if (vl_result_t4%is_populated) call recycle_vdata_obj(vl_result_t4)
   if (vl_result_t5%is_populated) call recycle_vdata_obj(vl_result_t5)

   if (pp_action /= 'NETCDF') then
      if (vl_reorder_t2%is_populated) call recycle_vdata_obj(vl_reorder_t2)
      if (vl_reorder_t3%is_populated) call recycle_vdata_obj(vl_reorder_t3)
      if (vl_reorder_t4%is_populated) call recycle_vdata_obj(vl_reorder_t4)
      if (vl_reorder_t5%is_populated) call recycle_vdata_obj(vl_reorder_t5)
   end if

   if (vl_scr1_t2%is_populated) call recycle_vdata_obj(vl_scr1_t2)
   if (vl_scr2_t2%is_populated) call recycle_vdata_obj(vl_scr2_t2)
   if (vl_scr3_t2%is_populated) call recycle_vdata_obj(vl_scr3_t2)

   if (vl_scr1_t3%is_populated) call recycle_vdata_obj(vl_scr1_t3)
   if (vl_scr2_t3%is_populated) call recycle_vdata_obj(vl_scr2_t3)
   if (vl_scr3_t3%is_populated) call recycle_vdata_obj(vl_scr3_t3)
   if (vl_scr4_t3%is_populated) call recycle_vdata_obj(vl_scr4_t3)
   if (vl_scr5_t3%is_populated) call recycle_vdata_obj(vl_scr5_t3)
   if (vl_scr6_t3%is_populated) call recycle_vdata_obj(vl_scr6_t3)
   if (vl_scr7_t3%is_populated) call recycle_vdata_obj(vl_scr7_t3)

   if (vl_scr1_faz_t3%is_populated) call recycle_vdata_obj(vl_scr1_faz_t3)
   if (vl_scr2_faz_t3%is_populated) call recycle_vdata_obj(vl_scr2_faz_t3)
   if (vl_scr3_faz_t3%is_populated) call recycle_vdata_obj(vl_scr3_faz_t3)
   if (vl_scr4_faz_t3%is_populated) call recycle_vdata_obj(vl_scr4_faz_t3)

   if (vl_scr1_fxy_t3%is_populated) call recycle_vdata_obj(vl_scr1_fxy_t3)
   if (vl_scr2_fxy_t3%is_populated) call recycle_vdata_obj(vl_scr2_fxy_t3)

   if (vl_scr1_fazxy_t3%is_populated) call recycle_vdata_obj(vl_scr1_fazxy_t3)
   if (vl_scr2_fazxy_t3%is_populated) call recycle_vdata_obj(vl_scr2_fazxy_t3)

   if (vl_scr1_fsz_t5%is_populated) call recycle_vdata_obj(vl_scr1_fsz_t5)
   if (vl_scr2_fsz_t5%is_populated) call recycle_vdata_obj(vl_scr2_fsz_t5)
   if (vl_scr3_fsz_t5%is_populated) call recycle_vdata_obj(vl_scr3_fsz_t5)
   if (vl_scr4_fsz_t5%is_populated) call recycle_vdata_obj(vl_scr4_fsz_t5)

   if (vl_scr1_t4%is_populated) call recycle_vdata_obj(vl_scr1_t4)
   if (vl_scr2_t4%is_populated) call recycle_vdata_obj(vl_scr2_t4)
   if (vl_scr3_t4%is_populated) call recycle_vdata_obj(vl_scr3_t4)
   if (vl_scr4_t4%is_populated) call recycle_vdata_obj(vl_scr4_t4)

   if (vl_scr1_t5%is_populated) call recycle_vdata_obj(vl_scr1_t5)
   if (vl_scr2_t5%is_populated) call recycle_vdata_obj(vl_scr2_t5)
   if (vl_scr3_t5%is_populated) call recycle_vdata_obj(vl_scr3_t5)

   if (vctr1_az%is_populated) call recycle_vdata_obj(vctr1_az)
   if (vctr2_az%is_populated) call recycle_vdata_obj(vctr2_az)
   if (vctr3_az%is_populated) call recycle_vdata_obj(vctr3_az)
   if (vctr4_az%is_populated) call recycle_vdata_obj(vctr4_az)
   if (vctr1_x%is_populated) call recycle_vdata_obj(vctr1_x)
   if (vctr1_y%is_populated) call recycle_vdata_obj(vctr1_y)
   if (vctr1_sz%is_populated) call recycle_vdata_obj(vctr1_sz)

else         ! (Re)initialize all elements

   tgt_f = 1

   ! Relevant dimensions for various variable types (and this grid):
   dims_t2 = (/ cfg_spatial(cfg_igd)%n_X, cfg_spatial(cfg_igd)%n_Y /)
   dims_t3 = (/ cfg_spatial(cfg_igd)%n_AZ, cfg_spatial(cfg_igd)%n_X,  &
        cfg_spatial(cfg_igd)%n_Y /)
   dims_t4 = (/ cfg_spatial(cfg_igd)%n_X, cfg_spatial(cfg_igd)%n_Y,  &
        cfg_spatial(cfg_igd)%n_P /)
   dims_t5 = (/ cfg_spatial(cfg_igd)%n_SZ, cfg_spatial(cfg_igd)%n_X,  &
        cfg_spatial(cfg_igd)%n_Y, cfg_spatial(cfg_igd)%n_P /)

   if (vl_result_t2%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_result_t2)
   call init_vdata_obj(vl_result_t2, dims_t2)

   if (vl_result_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_result_t3)
   call init_vdata_obj(vl_result_t3, dims_t3)

   if (vl_result_t4%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_result_t4)
   call init_vdata_obj(vl_result_t4, dims_t4)

   if (vl_result_t5%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_result_t5)
   call init_vdata_obj(vl_result_t5, dims_t5)

   if (vl_scr1_t2%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr1_t2)
   call init_vdata_obj(vl_scr1_t2, dims_t2)

   if (vl_scr2_t2%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr2_t2)
   call init_vdata_obj(vl_scr2_t2, dims_t2)

   if (vl_scr3_t2%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr3_t2)
   call init_vdata_obj(vl_scr3_t2, dims_t2)

   if (vl_scr1_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr1_t3)
   call init_vdata_obj(vl_scr1_t3, dims_t3)

   if (vl_scr2_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr2_t3)
   call init_vdata_obj(vl_scr2_t3, dims_t3)

   if (vl_scr3_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr3_t3)
   call init_vdata_obj(vl_scr3_t3, dims_t3)

   if (vl_scr4_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr4_t3)
   call init_vdata_obj(vl_scr4_t3, dims_t3)

   if (vl_scr5_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr5_t3)
   call init_vdata_obj(vl_scr5_t3, dims_t3)

   if (vl_scr6_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr6_t3)
   call init_vdata_obj(vl_scr6_t3, dims_t3)

   if (vl_scr7_t3%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr7_t3)
   call init_vdata_obj(vl_scr7_t3, dims_t3)

   if (use_full_az_scr(cfg_igd)) then
      dims_t3 = (/ nzp(tgt_igd,cfg_f(1)), cfg_spatial(cfg_igd)%n_X,  &
           cfg_spatial(cfg_igd)%n_Y /)

      if (vl_scr1_faz_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr1_faz_t3)
      call init_vdata_obj(vl_scr1_faz_t3, dims_t3)

      if (vl_scr2_faz_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr2_faz_t3)
      call init_vdata_obj(vl_scr2_faz_t3, dims_t3)

      if (vl_scr3_faz_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr3_faz_t3)
      call init_vdata_obj(vl_scr3_faz_t3, dims_t3)

      if (vl_scr4_faz_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr4_faz_t3)
      call init_vdata_obj(vl_scr4_faz_t3, dims_t3)
   end if

   if (use_full_xy_scr(cfg_igd)) then
      dims_t3 = (/ cfg_spatial(cfg_igd)%n_AZ, nxp(tgt_igd,cfg_f(1)),  &
           nyp(tgt_igd,cfg_f(1)) /)

      if (vl_scr1_fxy_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr1_fxy_t3)
      call init_vdata_obj(vl_scr1_fxy_t3, dims_t3)

      if (vl_scr2_fxy_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr2_fxy_t3)
      call init_vdata_obj(vl_scr2_fxy_t3, dims_t3)
   end if

   if (use_full_azxy_scr(cfg_igd)) then
      dims_t3 = (/ nzp(tgt_igd,cfg_f(1)), nxp(tgt_igd,cfg_f(1)),  &
           nyp(tgt_igd,cfg_f(1)) /)

      if (vl_scr1_fazxy_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr1_fazxy_t3)
      call init_vdata_obj(vl_scr1_fazxy_t3, dims_t3)

      if (vl_scr2_fazxy_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr2_fazxy_t3)
      call init_vdata_obj(vl_scr2_fazxy_t3, dims_t3)
   end if

   if (vl_scr1_t4%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr1_t4)
   call init_vdata_obj(vl_scr1_t4, dims_t4)

   if (vl_scr2_t4%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr2_t4)
   call init_vdata_obj(vl_scr2_t4, dims_t4)

   if (vl_scr3_t4%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr3_t4)
   call init_vdata_obj(vl_scr3_t4, dims_t4)

   if (vl_scr4_t4%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr4_t4)
   call init_vdata_obj(vl_scr4_t4, dims_t4)

   if (vl_scr1_t5%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr1_t5)
   call init_vdata_obj(vl_scr1_t5, dims_t5)

   if (vl_scr2_t5%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr2_t5)
   call init_vdata_obj(vl_scr2_t5, dims_t5)

   if (vl_scr3_t5%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vl_scr3_t5)
   call init_vdata_obj(vl_scr3_t5, dims_t5)

   if (vctr1_az%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vctr1_az)
   call init_vdata_obj(vctr1_az, (/ cfg_spatial(cfg_igd)%n_AZ /))

   if (vctr2_az%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vctr2_az)
   call init_vdata_obj(vctr2_az, (/ cfg_spatial(cfg_igd)%n_AZ /))

   if (vctr3_az%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vctr3_az)
   call init_vdata_obj(vctr3_az, (/ cfg_spatial(cfg_igd)%n_AZ /))

   if (vctr4_az%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vctr4_az)
   call init_vdata_obj(vctr4_az, (/ cfg_spatial(cfg_igd)%n_AZ /))

   if (vctr1_x%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vctr1_x)
   call init_vdata_obj(vctr1_x, (/ cfg_spatial(cfg_igd)%n_X /))

   if (vctr1_y%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vctr1_y)
   call init_vdata_obj(vctr1_y, (/ cfg_spatial(cfg_igd)%n_Y /))

   if (vctr1_sz%is_populated)  &    ! Recycle object first, if needed
        call recycle_vdata_obj(vctr1_sz)
   call init_vdata_obj(vctr1_sz, (/ cfg_spatial(cfg_igd)%n_SZ /))

   if (use_full_sz_scr(cfg_igd)) then
      dims_t5 = (/ nzg(cfg_f(1))+nzs(cfg_f(1)), cfg_spatial(cfg_igd)%n_X,  &
           cfg_spatial(cfg_igd)%n_Y, cfg_spatial(cfg_igd)%n_P /)

      if (vl_scr1_fsz_t5%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr1_fsz_t5)
      call init_vdata_obj(vl_scr1_fsz_t5, dims_t5)

      if (vl_scr2_fsz_t5%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr2_fsz_t5)
      call init_vdata_obj(vl_scr2_fsz_t5, dims_t5)

      if (vl_scr3_fsz_t5%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr3_fsz_t5)
      call init_vdata_obj(vl_scr3_fsz_t5, dims_t5)

      if (vl_scr4_fsz_t5%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_scr4_fsz_t5)
      call init_vdata_obj(vl_scr4_fsz_t5, dims_t5)
   end if

   if (pp_action /= 'NETCDF') then

      ! Relevant dimensions for various variable types (and this grid):
      dims_t3 = (/ cfg_spatial(cfg_igd)%n_X, cfg_spatial(cfg_igd)%n_Y,  &
           cfg_spatial(cfg_igd)%n_AZ /)
      dims_t5 = (/ cfg_spatial(cfg_igd)%n_X, cfg_spatial(cfg_igd)%n_Y,  &
           cfg_spatial(cfg_igd)%n_SZ, cfg_spatial(cfg_igd)%n_P /)

      if (vl_reorder_t2%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_reorder_t2)
      call init_vdata_obj(vl_reorder_t2, dims_t2)

      if (vl_reorder_t3%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_reorder_t3)
      call init_vdata_obj(vl_reorder_t3, dims_t3)

      if (vl_reorder_t4%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_reorder_t4)
      call init_vdata_obj(vl_reorder_t4, dims_t4)

      if (vl_reorder_t5%is_populated)  &    ! Recycle object first, if needed
           call recycle_vdata_obj(vl_reorder_t5)
      call init_vdata_obj(vl_reorder_t5, dims_t5)

   end if
     
end if

end subroutine manage_cfg_v_memory


end module postp_driv_mod
