!```````````````````````````````````````````````````````````````````
module postp_fldext_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine postp_fldext_driv

!+ Post-processing driver for `extract_fldpt` format output

use univ_kind_defs_mod, only: f4, f8
use univ_GrADS_format_mod, only: close_GrADS_dfile, open_GrADS_dfile,  &
     write_GrADS_hfile, init_GrADS_format, exit_GrADS_format
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_integer_time_mod, only: time_obj, fp_seconds_eqv_Time,  &
     Time_eqv_fp_seconds, operator(*), operator(-), operator(/)
use univ_exception_mod, only: exception, FATAL
use univ_std_fileIO_mod, only: find_file_IOunit, mkdir_p, determine_IO_reclen
use univ_convert_units_mod, only: rad_to_deg_f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN, MAX_FILE_PATH_LEN
use univ_namelist_utils_mod, only: Time_to_gen_time_strrep
use body_specific_constants_mod, only: mnsdy, solar_mean_broadband, R_air,  &
     cp, p00, g_Xoid, radius, morbpd, smaxis, mass_bodysys,  &
     massratio_Sun2bodysys, s_rho, s_hc
use postp_main_memory_mod, only: cfg_n_g, cfg_f, cfg_f_inc,  &
     output_pp_file_pfx, ztrnsfm, cfg_n_ov, state_file_name, model_sim_ET,  &
     cfg_n_f, cfg_vname, nxp, nyp, nvar_read, what_vtype, cfg_vtype, zt,  &
     var_want_to_read, nvar_want, nzp, nzs, nzg, cfg_vlevs, cfg_vdesc,  &
     cfg_vunits, what_vlevs, what_vdesc, what_vunits, what_ofile, cfg_vofile,  &
     SBPS_Tlev, n_state_files, n_state_f_grids, xm, xt, ym, yt, zm, zt, dzm,  &
     dzt, naddsc, ztop, map_projection, plat, plon, grid_idx_list, maxnxp,  &
     maxnyp, maxnzp, maxnzg, maxnzs, maxnpa, undef_value_f4, Ls_of_periapse,  &
     orbital_eccentricity, axial_obliquity, AU_to_m_knob, PRAMS_vdata,  &
     cfg_spatial, vctr1_az, vctr1_sz, npatch, use_full_az_scr,  &
     use_full_xy_scr, use_full_azxy_scr, use_full_sz_scr, full_az_scr_needed,  &
     full_xy_scr_needed, full_azxy_scr_needed, full_sz_scr_needed
use main_memory_mod, only: current_sim_ET, module_tps, mnsdy_Time,  &
     orbital_info_table, PRAMS_version, start_UT, target_Ls, morbpd_Time
use simple_astron_scheme_mod, only: sas_t_obj_recycle, populate_sas_table,  &
     set_sloc_sas_table, get_current_sas_vals
use postp_varlib_mod, only: postp_varlib
use postp_getdata_mod, only: fill_PRAMS_data, get_PRAMS_data,  &
     count_vars_to_read
use postp_driv_mod, only: manage_PRAMS_vdata, manage_cfg_v_memory

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_fldext_driv'

integer, parameter :: maxnGrADS_files = 2

integer :: igd, iv, ifn, lnoutp, lnstp, in_unit, e, iq, ic, rfcnt,  &
     nc, mdftxt_unit, mdfbin_unit, mdfbin_rec, n1, n2, n3, rgcnt, idx, err,  &
     t_body_days, i, IO_reclen
character(len=MAX_FILE_PATH_LEN) :: s_dfn, strtmp
character(len=MAX_FILE_PATH_LEN), dimension(maxnGrADS_files) :: g_dfn, g_hfn
character(len=8), dimension(maxnGrADS_files) :: file_handle
character(len=7) :: cgrid_a
character(len=10), dimension(maxnGrADS_files) :: cgrid_gd, cgrid_gh
character(len=40) :: gen_time_strrep
type(time_obj) :: tmp_Time1, tmp_Time2
logical, dimension(maxnGrADS_files) :: file_used
logical, allocatable, dimension(:,:) :: var_in_file
real(f4), allocatable, dimension(:,:) :: reclen_test
real(f4) :: tinc, sec, ftmp, curr_UMST, curr_EOT, curr_Ls
real(f8) :: tmp1_f8, tmp2_f8, curr_UMST_f8, curr_EOT_f8, curr_Ls_f8
logical :: found
integer, dimension(2) :: itmp_d1

!=== Executable statements:

! Initialize orbital info table:
call populate_sas_table(orbital_info_table, morbpd, 1000._f8, smaxis,  &
     real(AU_to_m_knob, kind=f8), mass_bodysys, massratio_Sun2bodysys,  &
     real(orbital_eccentricity, kind=f8), real(Ls_of_periapse, kind=f8),  &
     real(axial_obliquity, kind=f8))

! Find and record starting location in orbital/time table:
tmp1_f8 = real(target_Ls, kind=f8); tmp2_f8 = real(start_UT, kind=f8)
call set_sloc_sas_table(orbital_info_table, tmp2_f8, mnsdy, tmp1_f8)
target_Ls = real(tmp1_f8, kind=f4)

! Begin writing "metadata" header files:

i = index(output_pp_file_pfx, '/', back=.TRUE.)-1  ! Determine dir name length
call mkdir_p(output_pp_file_pfx(1:i))  ! Create any needed directories

lnoutp = len_trim(output_pp_file_pfx)  ! Length of string

strtmp = output_pp_file_pfx(1:lnoutp)//'-header.txt'
ic = len_trim(strtmp)
mdftxt_unit = find_file_IOunit()
open(mdftxt_unit, file=strtmp(1:ic), form='formatted', status='unknown')

strtmp = output_pp_file_pfx(1:lnoutp)//'-header.bin'
ic = len_trim(strtmp)
mdfbin_unit = find_file_IOunit()
IO_reclen = determine_IO_reclen(ftmp)
open(mdfbin_unit, file=strtmp(1:ic), form='unformatted', status='unknown',  &
     access='direct', recl=IO_reclen)
mdfbin_rec = 0

rfcnt = 0
do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
   rfcnt = rfcnt+1
end do
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, rfcnt, 'w')  ! Total # of snapshots
idx = 5; call binda_in0_rw0(mdfbin_unit, mdfbin_rec, idx, 'w')  ! Metadata records/snapshot

write (mdftxt_unit, '(2a)') 'PRAMS version used: ',  &
     PRAMS_version(1:len_trim(PRAMS_version)) 
write (mdftxt_unit, '(a,i5)') 'Total # of snapshots in this data (sub)set: ',  &
     rfcnt 
write (mdftxt_unit, '(/a)') 'Grids included in this data (sub)set:' 

rgcnt = 0
write (mdftxt_unit, '(/a)')  &
     ' Grid#  NX:curr(full)  NY:curr(full)  NZ:curr(full)     DX [km]'
do i=1,cfg_n_g   ! Loop over all desired grids
   igd = grid_idx_list(i)
   rgcnt = rgcnt+1
   write (mdftxt_unit, '(3x,i2,3(3x,i4,1x,a,i4,a),9x,f7.3)') igd,  &
        cfg_spatial(i)%n_X, '(', nxp(igd,1), ')',  &
        cfg_spatial(i)%n_Y, '(', nyp(igd,1), ')', cfg_spatial(i)%n_AZ, '(',   &
        nzp(igd,1), ')', (xt(2,igd,1)-xt(1,igd,1))/1000._f4
end do
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, rgcnt, 'w')  ! Total # of grids
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, maxnxp, 'w')
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, maxnyp, 'w')
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, maxnzp, 'w')
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, maxnzg, 'w')
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, maxnzs, 'w')
call binda_in0_rw0(mdfbin_unit, mdfbin_rec, maxnpa, 'w')
 
! Require 'topo' to be specified as an output variable:
found = .FALSE.
do iv=1,cfg_n_ov
   if (cfg_vname(iv)(1:4) == 'topo') found = .TRUE.
end do
if (.NOT.found) call exception(procedure_name, FATAL,  &
     'Mode requires topo to be an output variable.')

do iv=1,cfg_n_ov   ! Loop over # of variables to process

   ! Make initial pass through all grids, files, and variables; record what
   !  variables need to be read in and where they may be found:

   nvar_read = 0   ! Start out by assuming that no vars need read in
   use_full_az_scr = .FALSE.; use_full_xy_scr = .FALSE.
   use_full_azxy_scr = .FALSE.; use_full_sz_scr = .FALSE.

   do i=1,cfg_n_g   ! Loop over all desired grids
      igd = grid_idx_list(i)
      do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
!$$      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         ! What state variable(s) does this variable require? 
         call postp_varlib(igd, i, ifn, cfg_vname(iv), .TRUE.)
         use_full_az_scr(i) = ( use_full_az_scr(i) .or. full_az_scr_needed )
         use_full_xy_scr(i) = ( use_full_xy_scr(i) .or. full_xy_scr_needed )
         use_full_azxy_scr(i) = (use_full_azxy_scr(i) .or. full_azxy_scr_needed)
         use_full_sz_scr(i) = ( use_full_sz_scr(i) .or. full_sz_scr_needed )

         ! Check if var(s) has been configured for reading; if not, configure:
         call count_vars_to_read(igd, ifn)

!$$      end do
         if (ztrnsfm == 2) then  ! If Cartesian interp desired, need topo values
            nvar_want = 1
            var_want_to_read(1:nvar_want) = (/'TOPT'/)
            call count_vars_to_read(igd, ifn)
         end if
      end do
   end do

   ! Allocate some data storage:
   allocate(PRAMS_vdata(nvar_read), var_in_file(cfg_n_ov,maxnGrADS_files),  &
        stat=e)
   if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

   var_in_file = .FALSE.   ! Initialize logical array

   ! Initialize GrADS format module:

   file_handle(1) = 'atm'; file_handle(2) = 'sfc'
   call init_GrADS_format(maxnGrADS_files, file_handle)

   ! Main pass; read in data, calculate and output POSTP variable values:

   lnoutp = len_trim(output_pp_file_pfx)  ! Length of string

   do i=1,cfg_n_g   ! Loop over all desired grids
      igd = grid_idx_list(i)

      call manage_PRAMS_vdata(igd)    ! Initialize PRAMS_vdata for this grid
      call manage_cfg_v_memory(igd, i)  ! Init cfg-dimensioned multi-use storage

      ! Write per-grid metadata - START
      if (iv == 1) then
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, igd, 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, n_state_f_grids(1), 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, nxp(igd,1), 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, nyp(igd,1), 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, nzp(igd,1), 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, nzs(1), 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, nzg(1), 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, npatch(1), 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, naddsc(1), 'w')

         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, cfg_spatial(i)%n_X, 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, cfg_spatial(i)%n_Y, 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, cfg_spatial(i)%n_AZ, 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, cfg_spatial(i)%n_SZ, 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, cfg_spatial(i)%n_P, 'w')

         idx = 2
         itmp_d1 = (/ cfg_spatial(i)%i_X(1),  &
              cfg_spatial(i)%i_X(cfg_spatial(i)%n_X) /)
         call binda_in1_rw0(mdfbin_unit, mdfbin_rec, itmp_d1, idx, 'w')
         itmp_d1 = (/ cfg_spatial(i)%i_Y(1),  &
              cfg_spatial(i)%i_Y(cfg_spatial(i)%n_Y) /)
         call binda_in1_rw0(mdfbin_unit, mdfbin_rec, itmp_d1, idx, 'w')
         itmp_d1 = (/ cfg_spatial(i)%i_AZ(1),  &
              cfg_spatial(i)%i_AZ(cfg_spatial(i)%n_AZ) /)
         call binda_in1_rw0(mdfbin_unit, mdfbin_rec, itmp_d1, idx, 'w')
         itmp_d1 = (/ cfg_spatial(i)%i_SZ(1),  &
              cfg_spatial(i)%i_SZ(cfg_spatial(i)%n_SZ) /)
         call binda_in1_rw0(mdfbin_unit, mdfbin_rec, itmp_d1, idx, 'w')
         itmp_d1 = (/ cfg_spatial(i)%i_P(1),  &
              cfg_spatial(i)%i_P(cfg_spatial(i)%n_P) /)
         call binda_in1_rw0(mdfbin_unit, mdfbin_rec, itmp_d1, idx, 'w')

         idx = max(1, cfg_spatial(i)%i_X(min(2, cfg_spatial(i)%n_X))-  &
              cfg_spatial(i)%i_X(1))
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, idx, 'w')
         idx = max(1, cfg_spatial(i)%i_Y(min(2, cfg_spatial(i)%n_Y))-  &
              cfg_spatial(i)%i_Y(1))
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, idx, 'w')
         idx = max(1, cfg_spatial(i)%i_AZ(min(2, cfg_spatial(i)%n_AZ))-  &
              cfg_spatial(i)%i_AZ(1))
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, idx, 'w')
         idx = max(1, cfg_spatial(i)%i_SZ(min(2, cfg_spatial(i)%n_SZ))-  &
              cfg_spatial(i)%i_SZ(1))
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, idx, 'w')
         idx = max(1, cfg_spatial(i)%i_P(min(2, cfg_spatial(i)%n_P))-  &
              cfg_spatial(i)%i_P(1))
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, idx, 'w')
      
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, start_UT, 'w')
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ztop, 'w')
         call binda_in0_rw0(mdfbin_unit, mdfbin_rec, map_projection, 'w')
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, plat(igd,1), 'w')
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, plon(igd,1), 'w')

         n2 = nxp(igd,1); n3 = nyp(igd,1); n1 = nzp(igd,1)
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, xm(1:n2,igd,1), n2, 'w')
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, xt(1:n2,igd,1), n2, 'w')
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, ym(1:n3,igd,1), n3, 'w')
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, yt(1:n3,igd,1), n3, 'w')
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, zm(1:n1,igd,1), n1, 'w')
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, zt(1:n1,igd,1), n1, 'w')
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, dzm(1:n1,igd,1), n1, 'w')
         call binda_fn1_rw0(mdfbin_unit, mdfbin_rec, dzt(1:n1,igd,1), n1, 'w')

         ftmp = real(mnsdy, kind=f4)
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = R_air
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = cp
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = p00
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = s_rho
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = s_hc
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = g_Xoid
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = radius
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
         ftmp = solar_mean_broadband
         call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, ftmp, 'w')
      end if
      ! Write per-grid metadata - END

      write (cgrid_a, '(a1,i2.2,a4)') 'g', igd, '.vfm'  ! State file grid suffix

      if (ztrnsfm == 1) then
         do ifn=1,maxnGrADS_files
            write (cgrid_gd(ifn), '(a4,i2.2,a4)') '-S-g', igd, '.dat' ! GrADS grid 
            write (cgrid_gh(ifn), '(a4,i2.2,a4)') '-S-g', igd, '.ctl' ! suffixes
         end do
      else if (ztrnsfm == 2) then
         do ifn=1,maxnGrADS_files
            write (cgrid_gd(ifn), '(a4,i2.2,a4)') '-C-g', igd, '.dat' ! GrADS grid
            write (cgrid_gh(ifn), '(a4,i2.2,a4)') '-C-g', igd, '.ctl' ! suffixes
         end do
      end if

      rfcnt = 0
      do ifn=cfg_f(1),cfg_f(2),cfg_f_inc   ! Loop over desired files
         rfcnt = rfcnt+1      ! File number (relative) counter

         curr_UMST = mod(real(model_sim_ET(ifn)/mnsdy*24._f8, kind=f4)+  &
              start_UT, 24._f4)

         tmp_Time1 = fp_seconds_eqv_Time(model_sim_ET(ifn), module_tps)
         call Time_to_gen_time_strrep(tmp_Time1, gen_time_strrep,  &
              len(gen_time_strrep), mnsdy_Time, morbpd_Time,  &
              short_unit_style=.TRUE.)
         write (*, '(a,i5,a,i2,3a,f6.2,a)') 'Postprocessing < ', rfcnt,  &
              ' > grid #', igd, ', time = [', trim(gen_time_strrep), '] [',  &
              curr_UMST, ' UT] ...'

         ! Write per-snapshot metadata - START
         if (iv == 1 .and. igd == grid_idx_list(1)) then
            tmp_Time1 = fp_seconds_eqv_Time(model_sim_ET(ifn), module_tps)
            t_body_days = tmp_Time1/mnsdy_Time
            tmp_Time2 = tmp_Time1-(t_body_days*mnsdy_Time)
            sec = real(Time_eqv_fp_seconds(tmp_Time2), kind=f4)

            current_sim_ET = model_sim_ET(ifn)
            call get_current_sas_vals(orbital_info_table, current_sim_ET,  &
                 curr_UMST_f8, curr_Ls_f8, curr_EOT_f8, tmp1_f8, tmp2_f8)
            curr_UMST = real(curr_UMST_f8, kind=f4)
            curr_Ls = real(curr_Ls_f8, kind=f4)
            curr_EOT = real(curr_EOT_f8, kind=f4)*rad_to_deg_f4*24._f4/360._f4
!^
!^     write (*,'(//,a,f8.4,a,f7.4,a,f7.4/)')  &
!^          '****** PRAMS curr Ls = ',curr_Ls,', and UT = ',curr_UMST  &
!^         ,', EOT = ',curr_EOT      
!^
            if (rfcnt == 1) then 
               write (mdftxt_unit, '(/a)')  &
                    'Snap#      Ls     sol  sec_in_sol   UMST   EOTadj_Mhrs'
            end if
            write (mdftxt_unit, '(i5.5,a,f8.4,2x,i4,3x,f7.1,2(3x,f7.4))')  &
                 rfcnt, ': ', curr_Ls, t_body_days, sec, curr_UMST, curr_EOT 

            call binda_in0_rw0(mdfbin_unit, mdfbin_rec, t_body_days, 'w')
            call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, sec, 'w')
            call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, curr_Ls, 'w')
            call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, curr_UMST, 'w')
            call binda_fn0_rw0(mdfbin_unit, mdfbin_rec, curr_EOT, 'w')
         end if
         ! Write per-snapshot metadata - END

         lnstp = len_trim(state_file_name(ifn))-7
         s_dfn = state_file_name(ifn)(1:lnstp)//cgrid_a    ! State file name

         in_unit = find_file_IOunit(start_IOunit=mdfbin_unit+20)

         allocate(reclen_test(nxp(igd,ifn),nyp(igd,ifn)), stat=e)
         if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
              'reclen')
         IO_reclen = determine_IO_reclen(reclen_test)
         deallocate(reclen_test, stat=e)
         if (e /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'reclen')

         ! Open state data file:
         open(in_unit, file=s_dfn(1:lnstp+7), status='old',  &
              form='unformatted', access='direct', recl=IO_reclen)

         ! Read in all *NEEDED* PRAMS data for this grid/file; fill data array:
         call fill_PRAMS_data(in_unit, ifn, igd, nxp(igd,ifn)*nyp(igd,ifn))

         ! Compute values of configured vars; output data in desired manner:

!$$      do iv=1,cfg_n_ov    ! Loop over # of variables to process

         ! Read/compute var data:
         call postp_varlib(igd, i, ifn, cfg_vname(iv), .FALSE.)

         ! Store/calculate variable descriptors:
         if (ifn == cfg_f(1)) then
            cfg_vtype(iv) = what_vtype    ! Store var types, etc.
            cfg_vlevs(iv) = what_vlevs    !
            cfg_vdesc(iv) = what_vdesc    !
            cfg_vunits(iv) = what_vunits  !
            
            iq = 1
            do while(file_handle(iq) /= what_ofile)
               iq = iq+1
               if (iq > maxnGrADS_files) call exception(procedure_name,  &
                    FATAL, 'Incorrect ofile name in varlib')
            end do
            cfg_vofile(iv) = iq  ! Output file# this variable was written to

            var_in_file(iv,iq) = .TRUE.

            ! GrADS file names:
            ic = len_trim(cfg_vname(iv))
            g_dfn(:) = output_pp_file_pfx(1:lnoutp)//'blah_d'
            g_hfn(:) = output_pp_file_pfx(1:lnoutp)//'blah_h'
            g_dfn(iq) = output_pp_file_pfx(1:lnoutp)//'-'//cfg_vname(iv)(1:ic)//cgrid_gd(1)
            g_hfn(iq) = output_pp_file_pfx(1:lnoutp)//'-'//cfg_vname(iv)(1:ic)//cgrid_gh(1)
   
            ! Open GrADS data file(s) for writing:
            call open_GrADS_dfile(g_dfn(1), cfg_spatial(i)%n_X,  &
                 cfg_spatial(i)%n_Y, 'W', file_handle(1))
            call open_GrADS_dfile(g_dfn(2), cfg_spatial(i)%n_X,  &
                 cfg_spatial(i)%n_Y, 'W', file_handle(2))

         end if

         ! Output variable:
         call postp_fldext_varout(i, file_handle(cfg_vofile(iv))) 

!$$      end do

         close(in_unit)   ! Close state file

         if (ifn == cfg_f(1)) then  ! Write GrADS file descriptors after first
                                    !  file processed

            ! Ascertain which GrADS datafile(s) was(were) actually written to:

            file_used(:) = .FALSE.
!$$         do iv=1,cfg_n_ov    ! Loop over # of variables processed
            file_used(cfg_vofile(iv)) = .TRUE.
!$$         end do
         
            ! Write GrADS descriptor file(s):
         
            write (*, '(/)')
            do iq=1,maxnGrADS_files

               if (file_used(iq)) then    ! Write GrADS header file
                  if (n_state_files > cfg_f_inc) then   ! If enough files
                     tinc = real(model_sim_ET(cfg_f(1)+cfg_f_inc)-  &
                          model_sim_ET(cfg_f(1)), kind=f4)
                  else
                     tinc = real(model_sim_ET(cfg_f(1)), kind=f4)   ! Best possible right now
                  end if
                  if (iq == 1) then
                     do ic=1,cfg_spatial(i)%n_AZ
                        vctr1_az%p1D(ic) =  &
                             zt(cfg_spatial(i)%i_AZ(ic),igd,cfg_f(1))
                     end do
                     call write_GrADS_hfile(g_dfn(iq), g_hfn(iq),  &
                          cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y, cfg_n_f,  &
                          undef_value_f4, 0, 'green', vctr1_az%p1D,  &
                          model_sim_ET(cfg_f(1)), real(tinc, kind=f8), &
                          start_UT, cfg_vname, cfg_vlevs, cfg_vdesc,  &
                          cfg_vunits, var_in_file(1:cfg_n_ov,iq), mnsdy)
                  else if (iq == 2) then
                     do ic=1,cfg_spatial(i)%n_SZ
                        vctr1_sz%p1D(ic) =  &
                             SBPS_Tlev(cfg_spatial(i)%i_SZ(ic),cfg_f(1))
                     end do
                     call write_GrADS_hfile(g_dfn(iq), g_hfn(iq),  &
                          cfg_spatial(i)%n_X, cfg_spatial(i)%n_Y, cfg_n_f,  &
                          undef_value_f4, 0, 'green', vctr1_sz%p1D,  &
                          model_sim_ET(cfg_f(1)), real(tinc, kind=f8),  &
                          start_UT, cfg_vname, cfg_vlevs, cfg_vdesc,  &
                          cfg_vunits, var_in_file(1:cfg_n_ov,iq), mnsdy)
                  end if
                  write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
                       g_hfn(iq)(1:len_trim(g_hfn(iq)))
               end if
            
            end do
            write (*, '(/)')

         end if

      end do

      ! Conclude GrADS file writes:
         
      write (*, '(/)')
      do iq=1,maxnGrADS_files
      
         ! Close GrADS data file:
         call close_GrADS_dfile(file_handle(iq))
            
         if (file_used(iq)) then    ! Display GrADS header file name
            write (*, '(2a)') 'Wrote GrADS descriptor file:  ',  &
                 g_hfn(iq)(1:len_trim(g_hfn(iq)))
         else  ! Remove unused GrADS datafiles
            nc = len_trim(g_dfn(iq))
            call univ_remove_f(nc, g_dfn(iq), err)
         end if

      end do
      write (*, '(/)')

   end do

   ! Exit GrADS format module:

   call exit_GrADS_format

   ! Recycle data structures:
   call manage_PRAMS_vdata(-1)
   call manage_cfg_v_memory(-1, -1) 

   ! Deallocate work array:
   deallocate(PRAMS_vdata, var_in_file, stat=e)
   if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

   if (iv == 1) write (mdftxt_unit, '(/a/,i4,/)') 'Fields available:', cfg_n_ov
   write (mdftxt_unit, '(a,2x,i1,2x,i4,2x,a,2x,a)')  &
        cfg_vname(iv)(1:len_trim(cfg_vname(iv))),  &
        cfg_vtype(iv),cfg_vlevs(iv),cfg_vdesc(iv)(1:len_trim(cfg_vdesc(iv))),  &
        cfg_vunits(iv)(1:len_trim(cfg_vunits(iv)))

end do

close(mdftxt_unit)
close(mdfbin_unit)

call sas_t_obj_recycle(orbital_info_table)    ! Deallocate orbital info table

end subroutine postp_fldext_driv


!```````````````````````````````````````````````````````````````````
subroutine postp_fldext_varout(cfg_igd, file_handle)

!+ Writes postprocessed variable data to GrADS format file

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_GrADS_format_mod, only: write_GrADS_dfile
use univ_exception_mod, only: exception, FATAL
use postp_main_memory_mod, only: what_vtype, vl_reorder_t3,  &
     vl_reorder_t5, vl_result_t2, vl_result_t3, vl_result_t4, vl_result_t5,  &
     cfg_spatial

implicit none

!=== Argument declarations:

integer, intent(IN) :: cfg_igd
character(len=8), intent(IN) :: file_handle

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_fldext_varout'

integer :: nx, ny, naz, nsz, ip, np

!=== Executable statements:

nx = cfg_spatial(cfg_igd)%n_X; ny = cfg_spatial(cfg_igd)%n_Y

which_vtype: select case(what_vtype)
case (2)  ! "regular" 2D variable

   call write_GrADS_dfile(nx, ny, 1, vl_result_t2%p2D, file_handle)

case (3)  ! "regular" 3D variable

   naz = cfg_spatial(cfg_igd)%n_AZ
   vl_reorder_t3%p3D = reshape(vl_result_t3%p3D, (/nx,ny,naz/), order=(/3,1,2/))
   call write_GrADS_dfile(nx, ny, naz, vl_reorder_t3%p3D, file_handle)

case (4)  ! SBPS 2D variable

   do ip=1,cfg_spatial(cfg_igd)%n_P
      call write_GrADS_dfile(nx, ny, 1, vl_result_t4%p3D(:,:,ip), file_handle)
   end do

case (5)  ! SBPS 3D variable

   nsz = cfg_spatial(cfg_igd)%n_SZ
   np = cfg_spatial(cfg_igd)%n_P
   vl_reorder_t5%p4D = reshape(vl_result_t5%p4D, (/nx,ny,nsz,np/),  &
        order=(/3,1,2,4/))
   do ip=1,np
      call write_GrADS_dfile(nx, ny, nsz, vl_reorder_t5%p4D(:,:,:,ip),  &
           file_handle)
   end do

case default

   write (*, '(/a,i1,a)') 'ERROR: unknown variable type ( = ', what_vtype, ' ).'
   call exception(procedure_name, FATAL)

end select which_vtype

end subroutine postp_fldext_varout


!```````````````````````````````````````````````````````````````````
subroutine binda_fn0_rw0(IO_unit, rec_cnt, var, r_or_w)

!+ Write "scalar" floating point value to direct-access binary file

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: IO_unit
character(len=1), intent(IN) :: r_or_w

integer, intent(INOUT) :: rec_cnt
real(f4), intent(INOUT) :: var

!== Local declarations:

! NONE

!=== Executable statements:

rec_cnt = rec_cnt+1
if (r_or_w == 'w' .or. r_or_w == 'W') then
   write (IO_unit, rec=rec_cnt) var
else if (r_or_w == 'r' .or. r_or_w == 'R') then
   read (IO_unit, rec=rec_cnt) var
end if

end subroutine binda_fn0_rw0


!```````````````````````````````````````````````````````````````````
subroutine binda_fn1_rw0(IO_unit, rec_cnt, var1, dim1, r_or_w)

!+ Write "scalar" integer value to direct-access binary file

use univ_kind_defs_mod, only: f4

implicit none

!=== Argument declarations:

integer, intent(IN) :: IO_unit, dim1
character(len=1), intent(IN) :: r_or_w

integer, intent(INOUT) :: rec_cnt
real(f4), dimension(dim1), intent(INOUT) :: var1

!== Local declarations:

integer :: iq

!=== Executable statements:

if (r_or_w == 'w' .or. r_or_w == 'W') then
   do iq=1,dim1
      rec_cnt = rec_cnt+1
      write (IO_unit, rec=rec_cnt) var1(iq)
   end do
else if (r_or_w == 'r' .or. r_or_w == 'R') then
   do iq=1,dim1
      rec_cnt = rec_cnt+1
      read (IO_unit, rec=rec_cnt) var1(iq)
   end do
end if

end subroutine binda_fn1_rw0


!```````````````````````````````````````````````````````````````````
subroutine binda_in0_rw0(IO_unit, rec_cnt, var0, r_or_w)

!+ Write "scalar" integer value to direct-access binary file

implicit none

!=== Argument declarations:

integer, intent(IN) :: IO_unit
character(len=1), intent(IN) :: r_or_w

integer, intent(INOUT) :: rec_cnt, var0

!== Local declarations:

! NONE

!=== Executable statements:

rec_cnt = rec_cnt+1
if (r_or_w == 'w' .or. r_or_w == 'W') then
   write (IO_unit, rec=rec_cnt) var0
else if (r_or_w == 'r' .or. r_or_w == 'R') then
   read (IO_unit, rec=rec_cnt) var0
end if

end subroutine binda_in0_rw0


!```````````````````````````````````````````````````````````````````
subroutine binda_in1_rw0(IO_unit, rec_cnt, var1, dim1, r_or_w)

!+ Write "scalar" integer value to direct-access binary file

implicit none

!=== Argument declarations:

integer, intent(IN) :: IO_unit, dim1
character(len=1), intent(IN) :: r_or_w

integer, intent(INOUT) :: rec_cnt
integer, dimension(dim1), intent(INOUT) :: var1

!== Local declarations:

integer :: iq

!=== Executable statements:

if (r_or_w == 'w' .or. r_or_w == 'W') then
   do iq=1,dim1
      rec_cnt = rec_cnt+1
      write (IO_unit, rec=rec_cnt) var1(iq)
   end do
else if (r_or_w == 'r' .or. r_or_w == 'R') then
   do iq=1,dim1
      rec_cnt = rec_cnt+1
      read (IO_unit, rec=rec_cnt) var1(iq)
   end do
end if

end subroutine binda_in1_rw0


end module postp_fldext_mod
