!```````````````````````````````````````````````````````````````````
module postp_getdata_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine fill_PRAMS_data(IO_unit, ifn, igd, IO_rec_elemlen)

!+ Reads all *NEEDED* PRAMS variables into storage structure(s)

use univ_kind_defs_mod, only: f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_exception_mod, only: exception, FATAL
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use postp_main_memory_mod, only: nvar_read, var2read_type, var2read_iv,  &
     var_brec, var_erec, var2read_wsize, var2read_grid, PRAMS_vdata,  &
     in_standalone_mode
use main_memory_mod, only: a

implicit none

!=== Argument declarations:

integer, intent(IN) :: IO_unit, ifn, igd
integer, optional, intent(IN) :: IO_rec_elemlen

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'fill_PRAMS_data'

integer :: iv, iva, ivt, recsizew_f, e, ie, ib, ir, n_p, ivbr, iver, ip,  &
     n_vr_per_p
real(f4), allocatable, dimension(:) :: tmp

!=== Executable statements:

if (in_standalone_mode) then        !=== Read from PRAMS state files

   if ((.not. present(IO_rec_elemlen)) .or. IO_rec_elemlen <= 0)  &
        call exception(procedure_name, FATAL,  &
        'Optional argument must be present and have a reasonable value.')
   recsizew_f = IO_rec_elemlen

   ivt = 2
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
         
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p2D)

         ! Read in data:
         ir = var_brec(iva,ifn)
         read (IO_unit, rec=ir) PRAMS_vdata(iv)%p2D
      end if
   end do

   ivt = 3
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
         
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p3D)

         ! Allocate temporary storage, if needed:
         if (.NOT.allocated(tmp)) then
            allocate(tmp(var2read_wsize(iv)), stat=e)
            if (e /= 0) call illegal_alloc_stop(for_allocation,  &
                 procedure_name, '3')
         end if

         ! Read in data:
         ib = 1
         do ir=var_brec(iva,ifn),var_erec(iva,ifn)
            ie = ib+recsizew_f-1
            read (IO_unit, rec=ir) tmp(ib:ie)
            ib = ie+1
         end do

         ! Reshape data:
         PRAMS_vdata(iv)%p3D = reshape(tmp, shape(PRAMS_vdata(iv)%p3D))
      end if
   end do
   if (allocated(tmp)) then
      deallocate(tmp, stat=e)
      if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '3')
   end if

   ivt = 4
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
      
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p3D)

         n_p = size(PRAMS_vdata(iv)%p3D, 3)
         ivbr = var_brec(iva,ifn)
         n_vr_per_p = (var_erec(iva,ifn)-ivbr+1)/n_p
         do ip=1,n_p
            ! Read in data:
            iver = ivbr+n_vr_per_p-1
            do ir=ivbr,iver
               read (IO_unit, rec=ir) PRAMS_vdata(iv)%p3D(:,:,ip)
            end do
            ivbr = iver+1
         end do
      end if
   end do

   ivt = 5
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
      
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p4D)

         ! Allocate temporary storage, if needed:
         if (.NOT.allocated(tmp)) then
            n_p = size(PRAMS_vdata(iv)%p4D, 4)
            allocate(tmp(var2read_wsize(iv)/n_p), stat=e)
            if (e /= 0) call illegal_alloc_stop(for_allocation,  &
                 procedure_name, '5')
         end if

         ivbr = var_brec(iva,ifn)
         n_vr_per_p = (var_erec(iva,ifn)-ivbr+1)/n_p
         ib = 1
         do ip=1,n_p
            ! Read in data:
            iver = ivbr+n_vr_per_p-1
            do ir=ivbr,iver
               ie = ib+recsizew_f-1
               read (IO_unit, rec=ir) tmp(ib:ie)
               ib = ie+1
            end do
            ivbr = iver+1

            ! Reshape data:
            PRAMS_vdata(iv)%p4D(:,:,:,ip) = reshape(tmp,  &
                 (/ size(PRAMS_vdata(iv)%p4D, 1),  &
                 size(PRAMS_vdata(iv)%p4D, 2), size(PRAMS_vdata(iv)%p4D, 3) /))
         end do
      end if
   end do

else                 !=== Read from in-memory PRAMS arrays

   ivt = 2
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
         
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p2D)

         ! Copy and reshape data:
         PRAMS_vdata(iv)%p2D =  &
              reshape(a(var_brec(iva,ifn):var_erec(iva,ifn)),  &
              shape(PRAMS_vdata(iv)%p2D))
      end if
   end do

   ivt = 3
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
         
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p3D)

         ! Copy and reshape data:
         PRAMS_vdata(iv)%p3D =  &
              reshape(a(var_brec(iva,ifn):var_erec(iva,ifn)),  &
              shape(PRAMS_vdata(iv)%p3D))
      end if
   end do

   ivt = 4
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
      
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p3D)

         ! Copy and reshape data:
         PRAMS_vdata(iv)%p3D =  &
              reshape(a(var_brec(iva,ifn):var_erec(iva,ifn)),  &
              shape(PRAMS_vdata(iv)%p3D))
      end if
   end do

   ivt = 5
   do iv=1,nvar_read   ! Loop over all PRAMS variables to read
      if (var2read_type(iv) == ivt) then   ! Variable of type ivt found
         if (var2read_grid(iv) /= igd) cycle   ! If true, skip this grid
      
         iva = var2read_iv(iv,ifn)
         var2read_wsize(iv) = size(PRAMS_vdata(iv)%p4D)

         ! Copy and reshape data:
         PRAMS_vdata(iv)%p4D =  &
              reshape(a(var_brec(iva,ifn):var_erec(iva,ifn)),  &
              shape(PRAMS_vdata(iv)%p4D))
      end if
   end do

end if

if (allocated(tmp)) then
   deallocate(tmp, stat=e)
   if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '5')
end if

end subroutine fill_PRAMS_data


!```````````````````````````````````````````````````````````````````
subroutine get_PRAMS_data(output_data, tgt_vname, tgt_igd, cfg_igd, ifn,  &
     full_az, full_xy, full_azxy, full_sz)

!+ Copies data from PRAMS variable array to work array

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use postp_main_memory_mod, only: var_to_read, var2read_type, nvar_read,  &
     var2read_grid, vdata_obj, PRAMS_vdata, cfg_spatial, have_full_az, nzp,  &
     have_full_xy, nxp, nyp, have_full_azxy, have_full_sz, nzg, nzs

implicit none

!=== Argument declarations:

integer, intent(IN) :: tgt_igd, cfg_igd, ifn
character(len=*), intent(IN) :: tgt_vname
type(vdata_obj), intent(INOUT) :: output_data
logical, optional, intent(IN) :: full_az, full_xy, full_azxy, full_sz

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'get_PRAMS_data'

integer :: ltv, iv, i, j, k, ip, ii, jj, kk, ipip, vtype_offset 

!=== Executable statements:

vtype_offset = 0
if (present(full_az)) then
   if (full_az .and. .NOT.have_full_az(cfg_igd)) vtype_offset = 10
end if
if (present(full_xy)) then
   if (full_xy .and. .NOT.have_full_xy(cfg_igd)) vtype_offset = 20
end if
if (present(full_azxy)) then
   if (full_azxy .and. .NOT.have_full_azxy(cfg_igd)) vtype_offset = 30
end if
if (present(full_sz)) then
   if (full_sz .and. .NOT.have_full_sz(cfg_igd)) vtype_offset = 40
end if

ltv = len_trim(tgt_vname)

findv_loop: do iv=1,nvar_read

   if (var_to_read(iv)(1:len_trim(var_to_read(iv))) ==  &  ! Is this the var?
        tgt_vname(1:ltv) .and. var2read_grid(iv) == tgt_igd) then

      ! Copy data from storage variable to work array:

      vtype: select case(var2read_type(iv)+vtype_offset)
      case (2)
         do j=1,cfg_spatial(cfg_igd)%n_Y
            jj = cfg_spatial(cfg_igd)%i_Y(j)
            do i=1,cfg_spatial(cfg_igd)%n_X
               ii = cfg_spatial(cfg_igd)%i_X(i)
               output_data%p2D(i,j) = PRAMS_vdata(iv)%p2D(ii,jj)
            end do
         end do
      case (3)
         do j=1,cfg_spatial(cfg_igd)%n_Y
            jj = cfg_spatial(cfg_igd)%i_Y(j)
            do i=1,cfg_spatial(cfg_igd)%n_X
               ii = cfg_spatial(cfg_igd)%i_X(i)
               do k=1,cfg_spatial(cfg_igd)%n_AZ
                  kk = cfg_spatial(cfg_igd)%i_AZ(k)
                  output_data%p3D(k,i,j) = PRAMS_vdata(iv)%p3D(kk,ii,jj)
               end do
            end do
         end do
      case (4)
         do ip=1,cfg_spatial(cfg_igd)%n_P
            ipip = cfg_spatial(cfg_igd)%i_P(ip)
            do j=1,cfg_spatial(cfg_igd)%n_Y
               jj = cfg_spatial(cfg_igd)%i_Y(j)
               do i=1,cfg_spatial(cfg_igd)%n_X
                  ii = cfg_spatial(cfg_igd)%i_X(i)
                  output_data%p3D(i,j,ip) = PRAMS_vdata(iv)%p3D(ii,jj,ipip)
               end do
            end do
         end do
      case (5)
         do ip=1,cfg_spatial(cfg_igd)%n_P
            ipip = cfg_spatial(cfg_igd)%i_P(ip)
            do j=1,cfg_spatial(cfg_igd)%n_Y
               jj = cfg_spatial(cfg_igd)%i_Y(j)
               do i=1,cfg_spatial(cfg_igd)%n_X
                  ii = cfg_spatial(cfg_igd)%i_X(i)
                  do k=1,cfg_spatial(cfg_igd)%n_SZ
                     kk = cfg_spatial(cfg_igd)%i_SZ(k)
                     output_data%p4D(k,i,j,ip) =  &
                          PRAMS_vdata(iv)%p4D(kk,ii,jj,ipip)
                  end do
               end do
            end do
         end do
      case (13)   ! t3, full_az
         do j=1,cfg_spatial(cfg_igd)%n_Y
            jj = cfg_spatial(cfg_igd)%i_Y(j)
            do i=1,cfg_spatial(cfg_igd)%n_X
               ii = cfg_spatial(cfg_igd)%i_X(i)
               do k=1,nzp(tgt_igd,ifn)
                  output_data%p3D(k,i,j) = PRAMS_vdata(iv)%p3D(k,ii,jj)
               end do
            end do
         end do
      case (23)   ! t3, full_xy
         do j=1,nyp(tgt_igd,ifn)
            do i=1,nxp(tgt_igd,ifn)
               do k=1,cfg_spatial(cfg_igd)%n_AZ
                  kk = cfg_spatial(cfg_igd)%i_AZ(k)
                  output_data%p3D(k,i,j) = PRAMS_vdata(iv)%p3D(kk,i,j)
               end do
            end do
         end do
      case (33)   ! t3, full_azxy
         do j=1,nyp(tgt_igd,ifn)
            do i=1,nxp(tgt_igd,ifn)
               do k=1,nzp(tgt_igd,ifn)
                  output_data%p3D(k,i,j) = PRAMS_vdata(iv)%p3D(k,i,j)
               end do
            end do
         end do
      case (45)   ! t5, full_sz
         do ip=1,cfg_spatial(cfg_igd)%n_P
            ipip = cfg_spatial(cfg_igd)%i_P(ip)
            do j=1,cfg_spatial(cfg_igd)%n_Y
               jj = cfg_spatial(cfg_igd)%i_Y(j)
               do i=1,cfg_spatial(cfg_igd)%n_X
                  ii = cfg_spatial(cfg_igd)%i_X(i)
                  do k=1,nzg(ifn)+nzs(ifn)
                     output_data%p4D(k,i,j,ip) =  &
                          PRAMS_vdata(iv)%p4D(k,ii,jj,ipip)
                  end do
               end do
            end do
         end do
      case default
         call exception(procedure_name, FATAL,  &
              'Variable type unknown for '//tgt_vname(1:ltv))
      end select vtype

      exit findv_loop     ! Done with loop
   end if

end do findv_loop

if (iv == nvar_read+1) then
   call exception(procedure_name, FATAL,  &
        'Failed to identify variable '//tgt_vname(1:ltv)//' for data extraction.')
end if

end subroutine get_PRAMS_data


!```````````````````````````````````````````````````````````````````
subroutine count_vars_to_read(tgt_igd, ifn)

!+ Check if var(s) has been configured for reading; if not, configure 

use univ_exception_mod, only: exception, FATAL
use postp_main_memory_mod, only: n_state_f_vars, var_name, var_nlen,  &
     var_grid, nvar_want, nvar_read, var_want_to_read, var_to_read,  &
     var2read_type, var2read_iv, var_type, var2read_grid, var2read_currfidx

implicit none

!=== Argument declarations:

integer, intent(IN) :: tgt_igd, ifn

!== Local declarations:

integer :: iv, tgt_iv, lv, lv1, var_repeat_idx
logical :: var_found

!=== Executable statements:

do tgt_iv=1,nvar_want   ! Loop over state vars needed for this variable

   lv = len_trim(var_want_to_read(tgt_iv))

   ! Has the variable we want already been configured for reading?

   var_repeat_idx = 0
   do iv=1,nvar_read
      lv1 = len_trim(var_to_read(iv))
      if (var_want_to_read(tgt_iv)(1:lv) == var_to_read(iv)(1:lv1) .and.  &
           var2read_grid(iv) == tgt_igd) then
         var_repeat_idx = iv   ! Already configured this variable
         exit
      end if
   end do
   
   ! No?  Then is the var we want in this state file?

   var_found = .FALSE.
   findv_loop: do iv=1,n_state_f_vars(ifn) ! Loop over all vars in state file

      ! Is this the variable we want?

      if (var_name(iv,ifn)(1:var_nlen(iv,ifn)) ==   &
         var_want_to_read(tgt_iv)(1:lv) .and. var_grid(iv,ifn) == tgt_igd) then

         if (var_repeat_idx == 0) then
            nvar_read = nvar_read+1
            var_to_read(nvar_read) = var_want_to_read(tgt_iv)(1:lv)
            var2read_type(nvar_read) = var_type(iv,ifn)
            var2read_grid(nvar_read) = var_grid(iv,ifn)
            var2read_currfidx(nvar_read) = ifn
            var2read_iv(nvar_read,ifn) = iv
         else if (var2read_currfidx(nvar_read) /= ifn) then
            var2read_iv(var_repeat_idx,ifn) = iv
         end if

         var_found = .TRUE.
         exit findv_loop     ! Done with loop

      end if

   end do findv_loop

   if (.NOT.var_found) then
      call exception('count_vars_to_read', FATAL,  &
           'Variable '//var_want_to_read(tgt_iv)(1:lv)//' not available in this run.')
   end if

end do

end subroutine count_vars_to_read


!```````````````````````````````````````````````````````````````````
subroutine PRAMS_Cartpost_interp(var, v1, v2, v3, v4, topo, ztop, zt)

!+ Interpolates sigma-z data to Cartesian levels (includes topo height)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4
use interpolation_mod, only: zV_interp
use postp_main_memory_mod, only: undef_value_f4

implicit none

!=== Argument declarations:

real(f4), dimension(:,:), intent(IN) :: topo
real(f4), dimension(:), intent(IN) :: zt
real(f4), intent(IN) :: ztop

real(f4), dimension(:,:,:), intent(INOUT) :: var
real(f4), dimension(:), intent(INOUT) :: v1, v2, v3, v4

!== Local declarations:

integer :: i, j
real(f4) :: reftopo

!=== Executable statements:

reftopo = minval(topo)    ! Lowest topography value on grid
v3 = reftopo+zt*(ONE_f4-reftopo/ztop)  ! Z above geoid to interp to

do j=1,size(var, 3)
   do i=1,size(var, 2)
      v1 = var(:,i,j)   ! Vector to be interpolated
      v2 = topo(i,j)+zt*(ONE_f4-topo(i,j)/ztop) ! Z levs of above vect
      call zV_interp(size(var, 1), v1, v2, size(var, 1), v4, v3, topo(i,j))
      where (v4 > undef_value_f4*1.e-3_f4) v4 = undef_value_f4   ! Mask interpolation artifacts
      var(:,i,j) = v4   ! Interpolated values
   end do
end do

end subroutine PRAMS_Cartpost_interp


end module postp_getdata_mod
