!```````````````````````````````````````````````````````````````````
program postp

!+ Post-processing package for PRAMS

use univ_std_fileIO_mod, only: find_file_IOunit
use univ_maxlen_string_mod, only: MAX_FILE_PATH_LEN !%, MAX_PROC_NAME_LEN
use postp_main_memory_mod, only: manage_misc_memory, in_standalone_mode,  &
     dust_present, H2Oi_present, CO2i_present
use postp_init_mod, only: postp_init, postp_init_dcfg
use postp_driver_mod, only: postprocess_data
use postp_namelist_mod, only: postp_nml_fn
use postp_namelist_read_mod, only: postp_read_namelist

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

!%character(len=MAX_PROC_NAME_LEN), parameter ::  &
!%     procedure_name = 'postp'

integer :: IO_unit
logical :: file_exists

character(len=MAX_FILE_PATH_LEN) :: namelist_fn
namelist /postp_args/ namelist_fn

!=== Executable statements:

in_standalone_mode = .TRUE.

! Set namelist defaults:
namelist_fn = 'POSTP_IN'

! Find available IO unit:
IO_unit = find_file_IOunit()

! Read initial "command-line" namelist:

inquire(file=postp_nml_fn, exist=file_exists)
if (.NOT.file_exists) call usage

open(IO_unit, file=postp_nml_fn, status='old', form='formatted')

read (IO_unit, nml=postp_args); close(IO_unit)

! Read primary namelist:
call postp_read_namelist(postp_nml_fn)

! Common initialization tasks:
call postp_init
call postp_init_dcfg

! Postprocess data; write to file(s):
call postprocess_data

! Deallocate arrays:
call manage_misc_memory(dim_n_state_files=.FALSE., dim_max_n_sf_v=.FALSE.,  &
     dim_max_n_sf_g=.FALSE., dim_max_spatial=.FALSE., dim_cfg_n_g=.FALSE.,  &
     dim_cfg_n_ov=.FALSE.)
if (dust_present .or. H2Oi_present .or. CO2i_present)  &
     call manage_misc_memory(dim_aerosols=.FALSE.)

stop '*** Normal completion ***'

contains


!```````````````````````````````````````````````````````````````````
subroutine usage

!+ Prints postp command-line usage information

use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

! NONE

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'usage'

!=== Executable statements:

write (*, '(/,a,/)')  &
     'USAGE:  Must be started from run_postp, or another similar helper script.'
call exception(procedure_name, FATAL)

end subroutine usage


end program postp
