!```````````````````````````````````````````````````````````````````
module postp_main_memory_mod

!+

use univ_kind_defs_mod, only: f8, f4
use univ_maxlen_string_mod, only: MAX_FILE_PATH_LEN, MAX_PROC_NAME_LEN,  &
     MAX_FILE_NAME_LEN
use namelist_maxdims_mod, only: nml__strtoklen_max, nml_fn_strtoklen_max
use main_memory_mod, only: simulation_description, udv_f4 => undef_value_f4,  &
     chk_udv_f4 => check_undef_val_f4

implicit none

! Is postp being run as a stand-alone program?
logical :: in_standalone_mode

!== Namelist parameters, sorted by namelist block:

! &pp_basic_control :
character(len=nml__strtoklen_max) :: pp_description, pp_action
character(len=nml_fn_strtoklen_max) :: input_state_file_pfx,  &
     output_pp_file_pfx, default_input_data_dir
character(len=nml_fn_strtoklen_max), allocatable, dimension(:) :: dcfg
real(f8) :: time1, time2
integer :: cfg_f_inc

! &pp_vars :
! NONE

! &pp_spatial_config :
integer :: ztrnsfm

! &pp_OPENDX_config :
real(f4) :: pfact_xa, pfact_ya, pfact_za, pfact_xb, pfact_yb, pfact_zb
character(len=nml__strtoklen_max) :: ttag_fmt

type dcfg_obj
   integer :: n_X, n_Y, n_AZ, n_SZ, n_P
   integer, pointer, dimension(:) :: i_X => NULL(), i_Y => NULL(),  &
        i_AZ => NULL(), i_SZ => NULL(), i_P => NULL(), i_fX => NULL(),  &
        i_fY => NULL(), i_fAZ => NULL(), i_fSZ => NULL(), i_fP => NULL()
end type dcfg_obj
type(dcfg_obj), allocatable, dimension(:) :: cfg_spatial   ! {cfg_n_g}

real(f8), dimension(2) :: cfg_t
integer, dimension(2) :: cfg_f
integer, allocatable, dimension(:) :: grid_idx_list
integer :: cfg_n_f, cfg_n_g
integer :: cfg_n_ov, n_dcfg
character(len=MAX_PROC_NAME_LEN), allocatable, dimension(:) ::  &
     cfg_vname  ! {cfg_n_ov}

real(f4), parameter :: undef_value_f4 = udv_f4,  &
                       check_undef_val_f4 = chk_udv_f4
 
!------------- Logical control variables:

logical :: full_az_scr_needed, full_xy_scr_needed, full_azxy_scr_needed,  &
     full_sz_scr_needed
logical, allocatable, dimension(:) :: have_full_az, use_full_az_scr,  &
     have_full_xy, use_full_xy_scr, have_full_azxy, use_full_azxy_scr,  &
     have_full_sz, use_full_sz_scr

!---------------------- State file info variables:

integer :: n_state_files, max_n_sf_v, maxnzp, maxnyp, maxnxp,  &
     maxnzg, maxnzs, maxnpa, max_n_sf_g, min_n_sf_g, dust_op_case, RTc_tog

character(len=MAX_FILE_PATH_LEN), allocatable, dimension(:) ::  &
     state_file_name  ! {n_state_files}
integer, allocatable, dimension(:,:) :: var_brec, var_erec,  &
     var_type, var_grid, var_nlen                    ! {nvmax,n_state_files}
character(len=16), allocatable, dimension(:,:) ::  &
     var_name  ! {nvmax,n_state_files}
integer, allocatable, dimension(:) :: n_state_f_vars  ! {n_state_files}

!------------------ Variable descriptor variables:

integer, allocatable, dimension(:) :: cfg_vtype, cfg_vlevs,  &
     cfg_vofile                   ! {cfg_n_ov}
character(len=MAX_FILE_NAME_LEN), allocatable, dimension(:) :: cfg_vdesc,  &
     cfg_vunits    ! {cfg_n_ov}
integer :: what_vlevs
character(len=MAX_FILE_NAME_LEN) :: what_vdesc, what_vunits
character(len=8) :: what_ofile

!------------------ Data retrieval variables:

logical :: var_data_is_waiting
integer :: what_vtype

integer, parameter :: var_name_len = 8

integer :: nvar_want
character(len=var_name_len), allocatable, dimension(:) ::  &
     var_want_to_read  ! {nvmax}

integer :: nvar_read
character(len=var_name_len), allocatable, dimension(:) :: var_to_read  ! {nvmax}
integer, allocatable, dimension(:) :: var2read_type,  &   ! {nvmax}
     var2read_idx, var2read_wsize, var2read_grid, var2read_currfidx
integer, allocatable, dimension(:,:) :: var2read_iv  ! {nvmax,n_state_files}

integer :: scr1, scr2, scr3, scr4, scr5, scr6, scrtopo, scr7, scr8, scrdim

integer :: output_array_idx

real(f4), allocatable, dimension(:) :: ppa

real(f4), pointer, dimension(:,:) :: p1_2D => NULL(), p2_2D => NULL(),  &
     p3_2D => NULL(), p4_2D => NULL(), p5_2D => NULL(), p6_2D => NULL()
real(f4), pointer, dimension(:,:,:) :: p1_2Ds => NULL(), p2_2Ds => NULL(),  &
     p3_2Ds => NULL(), p4_2Ds => NULL(), p5_2Ds => NULL(), p6_2Ds => NULL(),  &
     p1_3D => NULL(), p2_3D => NULL(), p3_3D => NULL(), p4_3D => NULL(),  &
     p5_3D => NULL(), p6_3D => NULL()
real(f4), pointer, dimension(:,:,:,:) :: p1_3Ds => NULL(), p2_3Ds => NULL(),  &
     p3_3Ds => NULL(), p4_3Ds => NULL(), p5_3Ds => NULL(), p6_3Ds => NULL()

type vdata_obj
   logical :: is_populated = .FALSE.
   integer :: n_dims
   real(f4), pointer, dimension(:) :: p1D => NULL()
   real(f4), pointer, dimension(:,:) :: p2D => NULL()
   real(f4), pointer, dimension(:,:,:) :: p3D => NULL()
   real(f4), pointer, dimension(:,:,:,:) :: p4D => NULL()
end type vdata_obj

type(vdata_obj), allocatable, dimension(:) :: PRAMS_vdata

type(vdata_obj), save ::  &
     vl_result_t2, vl_result_t3, vl_result_t4, vl_result_t5,  &
     vl_reorder_t2, vl_reorder_t3, vl_reorder_t4, vl_reorder_t5,  &
     vl_scr1_t2, vl_scr2_t2, vl_scr3_t2,  &
     vl_scr1_t3, vl_scr2_t3, vl_scr3_t3, vl_scr4_t3, vl_scr5_t3, vl_scr6_t3,  &
     vl_scr7_t3,  &
     vl_scr1_faz_t3, vl_scr2_faz_t3, vl_scr3_faz_t3, vl_scr4_faz_t3,  &
     vl_scr1_fxy_t3, vl_scr2_fxy_t3,  &
     vl_scr1_fazxy_t3, vl_scr2_fazxy_t3,  &
     vl_scr1_fsz_t5, vl_scr2_fsz_t5, vl_scr3_fsz_t5, vl_scr4_fsz_t5,  &
     vl_scr1_t4, vl_scr2_t4, vl_scr3_t4, vl_scr4_t4,  &
     vl_scr1_t5, vl_scr2_t5, vl_scr3_t5,  &
     vctr1_az, vctr2_az, vctr3_az, vctr4_az, vctr1_x, vctr1_y, vctr1_sz

!---------------- Variables to hold PRAMS "common info":

integer, allocatable, dimension(:) :: n_state_f_grids, nzg, nzs,  &  ! {n_state_files}
     naddsc, npatch
real(f8), allocatable, dimension(:) :: model_sim_ET   ! {n_state_files}
real(f4) :: ztop, Ls_of_periapse, orbital_eccentricity,  &
     axial_obliquity, AU_to_m_knob
real(f4), allocatable, dimension(:) :: Ls_curr, UT_curr,  &  ! {n_state_files}
     EOT_curr, sdec_curr, sdist_curr
integer, allocatable, dimension(:,:) :: nxp,  &
     nyp, nzp         ! [ngmax,n_state_files]
real(f4), allocatable, dimension(:,:) :: plat, plon  ! {ngmax,n_state_files}
real(f4), allocatable, dimension(:,:,:) :: dzm,  &
     dzt, zt, zm              ! {maxnzp,ngmax,n_state_files}
real(f4), allocatable, dimension(:,:,:) :: xm,  &
     xt         ! [maxnxp,ngmax,n_state_files]
real(f4), allocatable, dimension(:,:,:) :: ym,  &
     yt         ! {maxnyp,ngmax,n_state_files]
real(f4), allocatable, dimension(:,:) ::  &
     SBPS_Tlev     ! {maxnzg+maxnzs,n_state_files}
integer :: map_projection

integer :: nelements, ngroups, max_nbins
integer, allocatable, dimension(:) :: nbins, ienconc,  &  ! {ngroups}
     ishape, eshape
real(f4), allocatable, dimension(:) :: rhoelem   ! {nelements}
character(len=30), allocatable, dimension(:) :: groupname    ! {ngroups}
character(len=60), allocatable, dimension(:) :: elemname  ! {nelements}
real(f4), allocatable, dimension(:,:) :: rlow,  & ! {max_nbins,ngroups}
     r, rup, rmass, vol
real(f4), allocatable, dimension(:,:,:) ::  &
     rhop                     ! {maxnzp,max_nbins,ngroups}

type ustar_t_table
   real(f4) :: ust_tab_T_min, ust_tab_T_max, ust_tab_T_inc,  &
        ust_tab_rhof_min, ust_tab_rhof_max, ust_tab_rhof_inc
   integer :: ust_tab_nent_T, ust_tab_nent_rhof
   real(f4), pointer, dimension(:,:) :: ustar_t_min => NULL()
   logical :: initialized = .FALSE.
end type ustar_t_table

real(f4), parameter :: rho_basalt = 3000._f4,  &  ! [kg/m^3]
     rho_andesite = 2800._f4,  &  ! [kg/m^3]
     rho_dust_agglom = 870._f4  ! [kg/m^3]; Bridges et al. 2010, bulk JSC-Mars 1
type(ustar_t_table), save :: ustar_t_table_basalt,  &
     ustar_t_table_andesite, ustar_t_table_dust_agglom

integer :: dust_ig, dustnc_ie, H2Oi_ig, H2Oinc_ie, CO2i_ig, CO2inc_ie
logical :: dust_present, H2Oi_present, CO2i_present

real(f4), allocatable, dimension(:) :: dust_qe_vis,  &   ! {max_nbins}
     dust_xsarea_fact, H2Oi_qe_vis, H2Oi_xsarea_fact, CO2i_qe_vis,  &
     CO2i_xsarea_fact, dust_qe_IR, H2Oi_qe_IR, CO2i_qe_IR


contains


!```````````````````````````````````````````````````````````````````
subroutine manage_misc_memory(dim_n_state_files, dim_max_n_sf_v,  &
     dim_max_n_sf_g, dim_max_spatial, dim_aerosols, dim_cfg_n_g, dim_n_dcfg,  &
     dim_cfg_n_ov)

!+ Allocates or deallocates miscellaneous arrays with various dimensions

use univ_IO_mem_status_mod, only: illegal_alloc_stop
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN

implicit none

!=== Argument declarations:

logical, optional, intent(IN) :: dim_n_state_files, dim_max_n_sf_v,  &
     dim_max_n_sf_g, dim_max_spatial, dim_aerosols, dim_cfg_n_g, dim_n_dcfg,  &
     dim_cfg_n_ov

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'manage_misc_memory'

integer :: e

!=== Executable statements:

if (present(dim_n_state_files)) then
   if (dim_n_state_files) then
      allocate(state_file_name(n_state_files), n_state_f_vars(n_state_files),  &
           n_state_f_grids(n_state_files), nzg(n_state_files),  &
           nzs(n_state_files), npatch(n_state_files), naddsc(n_state_files),  &
           model_sim_ET(n_state_files), Ls_curr(n_state_files),  &
           UT_curr(n_state_files), EOT_curr(n_state_files),  &
           sdec_curr(n_state_files), sdist_curr(n_state_files), stat=e) 
   else
      deallocate(state_file_name, n_state_f_vars, n_state_f_grids, nzg, nzs,  &
           npatch, naddsc, model_sim_ET, Ls_curr, UT_curr, EOT_curr,  &
           sdec_curr, sdist_curr, stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_n_state_files, procedure_name,  &
        'dim_n_state_files')
end if

if (present(dim_max_n_sf_v)) then
   if (dim_max_n_sf_v) then
      allocate(var_brec(max_n_sf_v,n_state_files),  &
           var_erec(max_n_sf_v,n_state_files),  &
           var_type(max_n_sf_v,n_state_files),  &
           var_grid(max_n_sf_v,n_state_files),  &
           var_name(max_n_sf_v,n_state_files),  &
           var_nlen(max_n_sf_v,n_state_files), var_want_to_read(max_n_sf_v),  &
           var_to_read(max_n_sf_v), var2read_idx(max_n_sf_v),  &
           var2read_type(max_n_sf_v), var2read_wsize(max_n_sf_v),  &
           var2read_iv(max_n_sf_v,n_state_files), var2read_grid(max_n_sf_v),  &
           var2read_currfidx(max_n_sf_v), stat=e) 
   else
      deallocate(var_brec, var_erec, var_type, var_grid, var_name, var_nlen,  &
           var_want_to_read, var_to_read, var2read_idx, var2read_type,  &
           var2read_wsize, var2read_iv, var2read_grid, var2read_currfidx,  &
           stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_max_n_sf_v, procedure_name,  &
        'dim_max_n_sf_v')
end if

if (present(dim_max_n_sf_g)) then
   if (dim_max_n_sf_g) then
      allocate(nxp(max_n_sf_g,n_state_files),  &
           nyp(max_n_sf_g,n_state_files), nzp(max_n_sf_g,n_state_files),  &
           plat(max_n_sf_g,n_state_files), plon(max_n_sf_g,n_state_files),  &
           grid_idx_list(max_n_sf_g), stat=e) 
   else
      deallocate(nxp, nyp, nzp, plat, plon, stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_max_n_sf_g, procedure_name,  &
        'dim_max_n_sf_g')
end if

if (present(dim_max_spatial)) then
   if (dim_max_spatial) then
      allocate(dzm(maxnzp,max_n_sf_g,n_state_files),  &
           dzt(maxnzp,max_n_sf_g,n_state_files),  &
           zm(maxnzp,max_n_sf_g,n_state_files),  &
           zt(maxnzp,max_n_sf_g,n_state_files),  &
           xm(maxnxp,max_n_sf_g,n_state_files),  &
           xt(maxnxp,max_n_sf_g,n_state_files),  &
           ym(maxnyp,max_n_sf_g,n_state_files),  &
           yt(maxnyp,max_n_sf_g,n_state_files),  &
           SBPS_Tlev(maxnzg+maxnzs,n_state_files), stat=e) 
   else
      deallocate(dzm, dzt, zm, zt, xm, xt, ym, yt, SBPS_Tlev, stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_max_spatial, procedure_name,  &
        'dim_max_spatial')
end if

if (present(dim_aerosols)) then
   if (dim_aerosols) then
      allocate(nbins(ngroups), rhoelem(nelements), groupname(ngroups),  &
           elemname(nelements), rlow(max_nbins,ngroups),  &
           r(max_nbins,ngroups), rup(max_nbins,ngroups),  &
           rmass(max_nbins,ngroups), ienconc(ngroups),  &
           dust_qe_vis(max_nbins), dust_xsarea_fact(max_nbins),  &
           H2Oi_qe_vis(max_nbins), H2Oi_xsarea_fact(max_nbins),  &
           CO2i_qe_vis(max_nbins), CO2i_xsarea_fact(max_nbins),  &
           dust_qe_IR(max_nbins), H2Oi_qe_IR(max_nbins),  &
           CO2i_qe_IR(max_nbins), ishape(ngroups), eshape(ngroups),  &
           rhop(maxnzp,max_nbins,ngroups), vol(max_nbins,ngroups), stat=e)
   else
      deallocate(nbins, rhoelem, groupname, elemname, rlow, r, rup, rmass,  &
           ienconc, dust_qe_vis, dust_xsarea_fact, H2Oi_qe_vis,  &
           H2Oi_xsarea_fact, CO2i_qe_vis, CO2i_xsarea_fact, dust_qe_IR,  &
           H2Oi_qe_IR, CO2i_qe_IR, ishape, eshape, rhop, vol, stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_aerosols, procedure_name,  &
        'dim_aerosols')
end if

if (present(dim_cfg_n_g)) then
   if (dim_cfg_n_g) then
      allocate(cfg_spatial(cfg_n_g), have_full_az(cfg_n_g),  &
           use_full_az_scr(cfg_n_g), have_full_xy(cfg_n_g),  &
           use_full_xy_scr(cfg_n_g), have_full_azxy(cfg_n_g),  &
           use_full_azxy_scr(cfg_n_g), have_full_sz(cfg_n_g),  &
           use_full_sz_scr(cfg_n_g), stat=e) 
   else
      deallocate(cfg_spatial, have_full_az, use_full_az_scr, have_full_xy,  &
           use_full_xy_scr, have_full_azxy, use_full_azxy_scr, have_full_sz,  &
           use_full_sz_scr, stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_cfg_n_g, procedure_name,  &
        'dim_cfg_n_g')
end if

if (present(dim_cfg_n_ov)) then
   if (dim_cfg_n_ov) then
      allocate(cfg_vname(cfg_n_ov), cfg_vtype(cfg_n_ov), cfg_vlevs(cfg_n_ov),  &
           cfg_vdesc(cfg_n_ov), cfg_vunits(cfg_n_ov), cfg_vofile(cfg_n_ov),  &
           stat=e) 
   else
      deallocate(cfg_vname, cfg_vtype, cfg_vlevs, cfg_vdesc, cfg_vunits,  &
           cfg_vofile, stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_cfg_n_ov, procedure_name,  &
        'dim_cfg_n_ov')
end if

if (present(dim_n_dcfg)) then
   if (dim_n_dcfg) then
      allocate(dcfg(n_dcfg), stat=e) 
   else
      deallocate(dcfg, stat=e)
   end if
   if (e /= 0) call illegal_alloc_stop(dim_n_dcfg, procedure_name,  &
        'dim_n_dcfg')
end if

end subroutine manage_misc_memory


!```````````````````````````````````````````````````````````````````
subroutine init_vdata_obj(obj, dims)

!+

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(vdata_obj), intent(INOUT) :: obj
integer, dimension(:), intent(IN) :: dims

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'init_vdata_obj'

integer :: e

!=== Executable statements:

if (obj%is_populated) call exception(procedure_name, FATAL,  &
     'Object must not be already populated.')

obj%n_dims = size(dims)

select case (obj%n_dims)
case(1)
   allocate(obj%p1D(dims(1)), stat=e)
case(2)
   allocate(obj%p2D(dims(1),dims(2)), stat=e)
case(3)
   allocate(obj%p3D(dims(1),dims(2),dims(3)), stat=e)
case(4)
   allocate(obj%p4D(dims(1),dims(2),dims(3),dims(4)), stat=e)
end select
if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name, '1')

obj%is_populated = .TRUE.

end subroutine init_vdata_obj


!```````````````````````````````````````````````````````````````````
subroutine recycle_vdata_obj(obj)

!+

use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_deallocation
use univ_exception_mod, only: exception, FATAL

implicit none

!=== Argument declarations:

type(vdata_obj), intent(INOUT) :: obj

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'recycle_vdata_obj'

integer :: e

!=== Executable statements:

if (.NOT.obj%is_populated) call exception(procedure_name, FATAL,  &
     'Object must already be populated.')

select case (obj%n_dims)
case(1)
   deallocate(obj%p1D, stat=e); nullify(obj%p1D)
case(2)
   deallocate(obj%p2D, stat=e); nullify(obj%p2D)
case(3)
   deallocate(obj%p3D, stat=e); nullify(obj%p3D)
case(4)
   deallocate(obj%p4D, stat=e); nullify(obj%p4D)
end select
if (e /= 0) call illegal_alloc_stop(for_deallocation, procedure_name, '1')

obj%is_populated = .FALSE.

end subroutine recycle_vdata_obj


end module postp_main_memory_mod
