!```````````````````````````````````````````````````````````````````
module postp_namelist_read_mod

!+ Module that contains routines and variables related to reading the postp
!   namelist (dependent on both common and body-specific modules)

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine postp_read_namelist(namelist_fn)

!+ Main driver for reading postp namelist 

use univ_std_fileIO_mod, only: find_file_IOunit
use univ_exception_mod, only: exception, FATAL
use univ_maxlen_string_mod, only: MAX_MESSAGE_LEN, MAX_PROC_NAME_LEN
use postp_namelist_mod, only: proc_nml_pp_basic_control, nml_info_check,  &
     proc_nml_pp_vars, proc_nml_pp_spatial_config, proc_nml_pp_OPENDX_config
use postp_namelist_bspec_mod, only: proc_nml_pp_basic_ctl_bspec,  &
     nml_info_check_bspec

implicit none

!=== Argument declarations:

character(len=*), intent(IN) :: namelist_fn

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_read_namelist'

integer :: namelist_fh, cn
character(len=MAX_MESSAGE_LEN) :: msg
logical :: file_exists

!=== Executable statements:

! Obtain an available file unit:
namelist_fh = find_file_IOunit()

! Check whether the namelist file exists; if not, abort program execution:
cn = len_trim(namelist_fn)
inquire(file=namelist_fn(1:cn), exist=file_exists)
if (.NOT.file_exists) then
   write(msg, '(3a)') 'Input postp namelist file (', namelist_fn(1:cn),  &
        ') not found.'
   call exception(procedure_name, FATAL, trim(msg))
end if

! Open namelist file:
open(namelist_fh, file=namelist_fn(1:cn), status='old', form='formatted')

call nml_info_check(namelist_fh)
call nml_info_check_bspec(namelist_fh)

!== Read/process additional namelist blocks:

call proc_nml_pp_basic_control(namelist_fh)

call proc_nml_pp_basic_ctl_bspec(namelist_fh)

call proc_nml_pp_vars(namelist_fh)

call proc_nml_pp_spatial_config(namelist_fh)

call proc_nml_pp_OPENDX_config(namelist_fh)

close(namelist_fh)  ! Close file

end subroutine postp_read_namelist


end module postp_namelist_read_mod
