!~ 08.01.2007
!# TIM

!###################################################################
!          ** PRAMS Ingestion and Preprocessing Package **
!###################################################################

!```````````````````````````````````````````````````````````````````
subroutine intp_lsubs(nsolpf,ntpsl,ntimes,sol_Ls,lsubs)

!~ 08.01.2007
!# TIM

!+ Interpolates Ls for each MGCM timestep (extrapolates values for the majority
!   of timestep in the final sol)

implicit none

integer :: nsolpf,ntpsl,ntimes
real, dimension(nsolpf) :: sol_Ls
real, dimension(ntimes) :: lsubs

integer :: isol,its,iq
real :: slope


! Check if 359 -> 1 deg Ls boundary is present; if so, create "continuous
!   function" that may be easily handled:

do isol=2,nsolpf
   if(sol_Ls(isol-1) > sol_Ls(isol)) then   ! Boundary present
      sol_Ls(isol) = sol_Ls(isol)+360.
   endif
enddo

! Put interpolated Ls values in array:

do isol=1,nsolpf

   ! If isol == nsolpf, use slope for nsolpf-1:
   if(isol /= nsolpf) slope = (sol_Ls(isol+1)-sol_Ls(isol))/real(ntpsl)
   
   do its=1,ntpsl

      iq = (isol-1)*ntpsl+its
      lsubs(iq) = sol_Ls(isol)+(slope*(its-1))

      if(lsubs(iq) >= 360.) then     ! Make sure Ls range is 0.->359.99999
         lsubs(iq) = lsubs(iq)-360.
      endif

   enddo

enddo

return
end subroutine intp_lsubs


!```````````````````````````````````````````````````````````````````
subroutine intp_lsubs_flex(nsolpf,ntpsl,ntimes,sol_Ls,lsubs)

!~ 08.01.2007
!# TIM

!+ Interpolates Ls for each MGCM timestep (extrapolates values for the majority
!   of timestep in the final sol; capable of handling non-constant 'ntpsl')

implicit none

! {intent IN}:
integer :: nsolpf,ntimes
integer, dimension(nsolpf) :: ntpsl
real, dimension(nsolpf) :: sol_Ls

! {intent OUT}:
real, dimension(ntimes) :: lsubs

integer :: isol,its,iq
real :: slope


! Check if 359 -> 1 deg Ls boundary is present; if so, create "continuous
!   function" that may be easily handled:

do isol=2,nsolpf
   if(sol_Ls(isol-1) > sol_Ls(isol)) then   ! Boundary present
      sol_Ls(isol) = sol_Ls(isol)+360.
   endif
enddo

! Put interpolated Ls values in array:

iq = 0
do isol=1,nsolpf

   ! If isol == nsolpf, use slope for nsolpf-1:
   if(isol /= nsolpf) slope = (sol_Ls(isol+1)-sol_Ls(isol))/real(ntpsl(isol))
   
   do its=1,ntpsl(isol)

      iq = iq+1
      lsubs(iq) = sol_Ls(isol)+(slope*(its-1))

      if(lsubs(iq) >= 360.) then     ! Make sure Ls range is 0.->359.99999
         lsubs(iq) = lsubs(iq)-360.
      endif

   enddo

enddo

return
end subroutine intp_lsubs_flex


!```````````````````````````````````````````````````````````````````
subroutine intp_lsubs_new(nsolpf,ntpsl,ntimes,sol_Ls,lsubs)

!~ 08.01.2007
!# TIM

!+ Interpolates Ls for each MGCM timestep (no extrapolation, as N+1 pts are
!   given for N intervals)

implicit none

integer :: nsolpf,ntpsl,ntimes
real, dimension(nsolpf+1) :: sol_Ls
real, dimension(ntimes) :: lsubs

integer :: isol,its,iq
real :: slope


! Check if 359 -> 1 deg Ls boundary is present; if so, create "continuous
!   function" that may be easily handled:

do isol=2,nsolpf+1
   if(sol_Ls(isol-1) > sol_Ls(isol)) then   ! Boundary present
      sol_Ls(isol) = sol_Ls(isol)+360.
   endif
enddo

! Put interpolated Ls values in array:

do isol=1,nsolpf

   slope = (sol_Ls(isol+1)-sol_Ls(isol))/real(ntpsl)
 
   do its=1,ntpsl

      iq = (isol-1)*ntpsl+its
      lsubs(iq) = sol_Ls(isol)+(slope*(its-1))

      if(lsubs(iq) >= 360.) then     ! Make sure Ls range is 0.->359.99999
         lsubs(iq) = lsubs(iq)-360.
      endif

   enddo

enddo

return
end subroutine intp_lsubs_new

