!~ 08.01.2007
!# TIM

!###################################################################
!          ** PRAMS Ingestion and Preprocessing Package **
!###################################################################

!```````````````````````````````````````````````````````````````````
subroutine regriduvC(nlon,nlat,nlat1,nlyr,u_m,v_m,u_a,v_a)

!~ 08.01.2007
!# TIM

!+ Interpolate winds off of staggered grid (from Arakawa "C" to "A" grid)

implicit none
      
integer :: nlat,nlon,nlyr,nlat1
real, dimension(nlon,nlat,nlyr) :: u_m,v_m
real, dimension(nlon,nlat1,nlyr) :: u_a,v_a

integer :: i,j,k
real :: sum

! Interpolate winds from staggered grid (Arakawa-C) to "A" grid
!   (valid for NASA Ames MGCM grid): 

! V-component:

do k=1,nlyr

   do i=1,nlon
      do j=1,nlat-1
         v_a(i,j+1,k) = 0.5*(v_m(i,j,k)+v_m(i,j+1,k))  ! Fills v_a(:,2:nlat,:)
      enddo
      v_a(i,1,k) = v_m(i,1,k)-0.5*(v_m(i,2,k)-v_m(i,1,k))            ! S "pole"
      v_a(i,nlat1,k) = v_m(i,nlat,k)+0.5*(v_m(i,nlat,k)-v_m(i,nlat-1,k)) ! N
   enddo

   sum = 0.
   do i=1,nlon
      sum = sum+v_a(i,1,k) 
   enddo
   v_a(1:nlon,1,k) = sum/real(nlon)    ! Use average for South Pole value

   sum = 0.
   do i=1,nlon
      sum = sum+v_a(i,nlat1,k) 
   enddo
   v_a(1:nlon,nlat1,k) = sum/real(nlon)    ! Use average for North Pole value

enddo

! U-component:

do k=1,nlyr
   do j=1,nlat-1     ! Fill u_a(:,2:nlat,:)
      do i=2,nlon
         u_a(i,j+1,k) = 0.5*(u_m(i,j,k)+u_m(i-1,j,k))
      enddo
      u_a(1,j+1,k) = 0.5*(u_m(nlon,j,k)+u_m(1,j,k))  ! Handle wrapped u value
   enddo
   
   do i=1,nlon
      u_a(i,1,k) = u_a(i,2,k)-(u_a(i,3,k)-u_a(i,2,k))                ! S "pole"
      u_a(i,nlat1,k) = u_a(i,nlat,k)+(u_a(i,nlat,k)-u_a(i,nlat-1,k)) ! N "pole"
   enddo

   sum = 0.
   do i=1,nlon
      sum = sum+u_a(i,1,k) 
   enddo
   u_a(1:nlon,1,k) = sum/real(nlon)    ! Use average for South Pole value

   sum = 0.
   do i=1,nlon
      sum = sum+u_a(i,nlat1,k) 
   enddo
   u_a(1:nlon,nlat1,k) = sum/real(nlon)    ! Use average for North Pole value
   
enddo

return
end subroutine regriduvC


!```````````````````````````````````````````````````````````````````
subroutine regriduvC_new(nlon,nlat,nlat1,nlyr,u_m,v_m,u_a,v_a)

!~ 08.01.2007
!# TIM

!+ Interpolate winds off of staggered grid (from Arakawa "C" to "A" grid)

implicit none
      
integer :: nlat,nlon,nlyr,nlat1
real, dimension(nlon,nlat-1,nlyr) :: u_m
real, dimension(nlon,nlat,nlyr) :: v_m
real, dimension(nlon,nlat1,nlyr) :: u_a,v_a

integer :: i,j,k
real :: sum


! Interpolate winds from staggered grid (Arakawa-C) to "A" grid
!   (valid for NASA Ames MGCM grid): 

! V-component:

do k=1,nlyr

   do i=1,nlon
      do j=1,nlat-1
         v_a(i,j+1,k) = 0.5*(v_m(i,j,k)+v_m(i,j+1,k))  ! Fills v_a(:,2:nlat,:)
      enddo
      v_a(i,1,k) = v_m(i,1,k)-0.5*(v_m(i,2,k)-v_m(i,1,k))            ! S "pole"
      v_a(i,nlat1,k) = v_m(i,nlat,k)+0.5*(v_m(i,nlat,k)-v_m(i,nlat-1,k)) ! N
   enddo

   sum = 0.
   do i=1,nlon
      sum = sum+v_a(i,1,k) 
   enddo
   v_a(1:nlon,1,k) = sum/real(nlon)    ! Use average for South Pole value

   sum = 0.
   do i=1,nlon
      sum = sum+v_a(i,nlat1,k) 
   enddo
   v_a(1:nlon,nlat1,k) = sum/real(nlon)    ! Use average for North Pole value

enddo

! U-component:

do k=1,nlyr
   do j=1,nlat-1     ! Fill u_a(:,2:nlat,:)
      do i=2,nlon
         u_a(i,j+1,k) = 0.5*(u_m(i,j,k)+u_m(i-1,j,k))
      enddo
      u_a(1,j+1,k) = 0.5*(u_m(nlon,j,k)+u_m(1,j,k))  ! Handle wrapped u value
   enddo
   
   do i=1,nlon
      u_a(i,1,k) = u_a(i,2,k)-(u_a(i,3,k)-u_a(i,2,k))                ! S "pole"
      u_a(i,nlat1,k) = u_a(i,nlat,k)+(u_a(i,nlat,k)-u_a(i,nlat-1,k)) ! N "pole"
   enddo

   sum = 0.
   do i=1,nlon
      sum = sum+u_a(i,1,k) 
   enddo
   u_a(1:nlon,1,k) = sum/real(nlon)    ! Use average for South Pole value

   sum = 0.
   do i=1,nlon
      sum = sum+u_a(i,nlat1,k) 
   enddo
   u_a(1:nlon,nlat1,k) = sum/real(nlon)    ! Use average for North Pole value
   
enddo

return
end subroutine regriduvC_new



