!~ 10.04.2007
!# TIM

!###################################################################
!          ** PRAMS Ingestion and Preprocessing Package **
!###################################################################

!```````````````````````````````````````````````````````````````````
subroutine siglevs_calcC(nlon,nlat,nlat1,nlyr,mlevls,sp_m,p_a,t_a,gpz_m  &
     ,gpz_a,dsigs,sptrop,rgas,g)

!+ Calculates vertical pressure levels (C-grid)

!~ 17.03.2012 (last modified; TIM)

use main_memory_mod, only: undef_value_f4

implicit none

integer :: nlat,nlon,nlyr,mlevls,nlat1
real, dimension(nlon,nlat) :: sp_m
real, dimension(nlon,nlat,nlyr) :: gpz_m
real, dimension(nlon,nlat1,nlyr) :: p_a,gpz_a,t_a
real, dimension(nlyr) :: dsigs
real :: sptrop,rgas,g

integer :: klev,i,j,l,k
real, allocatable :: sig(:),avepres(:)
real :: H,tbot,pbot,gzsum,tavg,sum1,sum2

allocate(sig(mlevls),avepres(mlevls))       ! Allocate work arrays
      
! Calculate MGCM vertical pressure field from surface pressure and sigma 
!   levels:

! Calculate pressures on sigma levels:

do i=1,nlon

   do j=1,nlat-1     ! Fills gpz_a(:,2:nlat,:) and p_a(:,2:nlat,:)
      sig(1) = undef_value_f4
      sig(2) = undef_value_f4
      sig(3) = 0.
      avepres(1) = undef_value_f4
      avepres(2) = undef_value_f4
      avepres(3) = sptrop

      ! Set MGCM sigma values for odd-k levels (boundary of layers)

      do l=1,nlyr
         klev = 2*l+3
         sig(klev) = sig(klev-2)+dsigs(l)
         avepres(klev) = sig(klev)*(sp_m(i,j)-sptrop)+sptrop
      enddo

      ! Set MGCM sigma values for even-k levels (middle of layers)

      do k=4,mlevls-1,2 
         sig(k) = 0.5*(sig(k+1)+sig(k-1))
         avepres(k) = sig(k)*(sp_m(i,j)-sptrop)+sptrop
      enddo

      ! Fill array with the pressures at the middle of levels over a grid point

      do k = 4,mlevls-1,2
         p_a(i,j+1,((mlevls-1)-k+4-2)/2) = avepres(k)*100.  ! Convert to Pa
      enddo

      ! Compute geopotential near surface from surface pressure

      gzsum = 0.
      pbot = sp_m(i,j)*100.  ! Begin integration at surface pressure
      tbot = t_a(i,j+1,1)   ! Not at all the best value for this, but oh well
      do k=1,nlyr   ! Loop over all MGCM layers 
         tavg = (tbot+t_a(i,j+1,k))*0.5 ! Average tempk for the current layer
         H = (-rgas/g)*tavg*log(p_a(i,j+1,k)/pbot)  ! Delta-geopotential height
         gzsum = gzsum+H       ! Geopotential height of current layer *AGL*
         gpz_a(i,j+1,k) = gzsum  !   (Ignore whatever gpz MGCM has)
         pbot = p_a(i,j+1,k)   ! These will become the lower boundary for the
         tbot = t_a(i,j+1,k)   !   next iteration
      enddo

   enddo

   do k=1,nlyr  
      ! South "pole" values:
      gpz_a(i,1,k) = gpz_a(i,2,k)-(gpz_a(i,3,k)-gpz_a(i,2,k))
      p_a(i,1,k) = p_a(i,2,k)-(p_a(i,3,k)-p_a(i,2,k))          
      ! North "pole" values:
      gpz_a(i,nlat1,k) = gpz_a(i,nlat,k)+(gpz_a(i,nlat,k)-gpz_a(i,nlat-1,k))
      p_a(i,nlat1,k) = p_a(i,nlat,k)+(p_a(i,nlat,k)-p_a(i,nlat-1,k)) 
   enddo

enddo

do k=1,nlyr
   
   sum1 = 0.
   sum2 = 0.
   do i=1,nlon
      sum1 = sum1+gpz_a(i,1,k) 
      sum2 = sum2+p_a(i,1,k) 
   enddo
   gpz_a(1:nlon,1,k) = sum1/real(nlon)    ! Use average for South Pole value
   p_a(1:nlon,1,k) = sum2/real(nlon)      !
   
   sum1 = 0.
   sum2 = 0.
   do i=1,nlon
      sum1 = sum1+gpz_a(i,nlat1,k) 
      sum2 = sum2+p_a(i,nlat1,k) 
   enddo
   gpz_a(1:nlon,nlat1,k) = sum1/real(nlon)  ! Use average for North Pole value
   p_a(1:nlon,nlat1,k) = sum2/real(nlon)    !

enddo

deallocate(sig,avepres)                   ! Deallocate work arrays
   
return
end subroutine siglevs_calcC


!```````````````````````````````````````````````````````````````````
subroutine siglevs_calcC_new(nlon,nlat,nlat1,nlyr,mlevls,sp_m,p_a,t_a  &
     ,gpz_a,dsigs,sptrop,rgas,g)

!~ 17.03.2012 (last modified; TIM)

!+ Calculates vertical pressure levels (C-grid)

use main_memory_mod, only: undef_value_f4

implicit none

integer :: nlat,nlon,nlyr,mlevls,nlat1
real, dimension(nlon,nlat-1) :: sp_m
real, dimension(nlon,nlat1,nlyr) :: p_a,gpz_a,t_a
real, dimension(nlyr) :: dsigs
real :: sptrop,rgas,g

integer :: klev,i,j,l,k
real, allocatable, dimension(:) :: sig,avepres      ! [mlevls]
real :: H,tbot,pbot,gzsum,tavg,sum1,sum2


allocate(sig(mlevls),avepres(mlevls))       ! Allocate work arrays
      
! Calculate MGCM vertical pressure field from surface pressure and sigma 
!   levels:

! Calculate pressures on sigma levels:

do i=1,nlon

   do j=1,nlat-1     ! Fills gpz_a(:,2:nlat,:) and p_a(:,2:nlat,:)
      sig(1) = undef_value_f4
      sig(2) = undef_value_f4
      sig(3) = 0.
      avepres(1) = undef_value_f4
      avepres(2) = undef_value_f4
      avepres(3) = sptrop

      ! Set MGCM sigma values for odd-k levels (boundary of layers)

      do l=1,nlyr
         klev = 2*l+3
         sig(klev) = sig(klev-2)+dsigs(l)
         avepres(klev) = sig(klev)*(sp_m(i,j)-sptrop)+sptrop
      enddo

      ! Set MGCM sigma values for even-k levels (middle of layers)

      do k=4,mlevls-1,2 
         sig(k) = 0.5*(sig(k+1)+sig(k-1))
         avepres(k) = sig(k)*(sp_m(i,j)-sptrop)+sptrop
      enddo

      ! Fill array with the pressures at the middle of levels over a grid point

      do k = 4,mlevls-1,2
         p_a(i,j+1,((mlevls-1)-k+4-2)/2) = avepres(k)*100.  ! Convert to Pa
      enddo

      ! Compute geopotential near surface from surface pressure

      gzsum = 0.
      pbot = sp_m(i,j)*100.  ! Begin integration at surface pressure
      tbot = t_a(i,j+1,1)   ! Not at all the best value for this, but oh well
      do k=1,nlyr   ! Loop over all MGCM layers 
         tavg = (tbot+t_a(i,j+1,k))*0.5 ! Average tempk for the current layer
         H = (-rgas/g)*tavg*log(p_a(i,j+1,k)/pbot)  ! Delta-geopotential height
         gzsum = gzsum+H       ! Geopotential height of current layer *AGL*
         gpz_a(i,j+1,k) = gzsum  !   (Ignore whatever gpz MGCM has)
         pbot = p_a(i,j+1,k)   ! These will become the lower boundary for the
         tbot = t_a(i,j+1,k)   !   next iteration
      enddo

   enddo

   do k=1,nlyr  
      ! South "pole" values:
      gpz_a(i,1,k) = gpz_a(i,2,k)-(gpz_a(i,3,k)-gpz_a(i,2,k))
      p_a(i,1,k) = p_a(i,2,k)-(p_a(i,3,k)-p_a(i,2,k))          
      ! North "pole" values:
      gpz_a(i,nlat1,k) = gpz_a(i,nlat,k)+(gpz_a(i,nlat,k)-gpz_a(i,nlat-1,k))
      p_a(i,nlat1,k) = p_a(i,nlat,k)+(p_a(i,nlat,k)-p_a(i,nlat-1,k)) 
   enddo

enddo

do k=1,nlyr
   
   sum1 = 0.
   sum2 = 0.
   do i=1,nlon
      sum1 = sum1+gpz_a(i,1,k) 
      sum2 = sum2+p_a(i,1,k) 
   enddo
   gpz_a(1:nlon,1,k) = sum1/real(nlon)    ! Use average for South Pole value
   p_a(1:nlon,1,k) = sum2/real(nlon)      !
   
   sum1 = 0.
   sum2 = 0.
   do i=1,nlon
      sum1 = sum1+gpz_a(i,nlat1,k) 
      sum2 = sum2+p_a(i,nlat1,k) 
   enddo
   gpz_a(1:nlon,nlat1,k) = sum1/real(nlon)  ! Use average for North Pole value
   p_a(1:nlon,nlat1,k) = sum2/real(nlon)    !

enddo

deallocate(sig,avepres)                   ! Deallocate work arrays
   
return
end subroutine siglevs_calcC_new
