!```````````````````````````````````````````````````````````````````
module postp_varlib_mod

!+

implicit none


contains


!```````````````````````````````````````````````````````````````````
subroutine postp_varlib(tgt_igd, cfg_igd, ifn, var_name, this_is_init_pass)

!+ Library of post-processor output variables (and their computation method)

use univ_kind_defs_mod, only: f4
use univ_math_constants_mod, only: ONE_f4, ZERO_f4
use univ_maxlen_string_mod, only: MAX_PROC_NAME_LEN
use univ_IO_mem_status_mod, only: illegal_alloc_stop, for_allocation,  &
     for_deallocation
use univ_exception_mod, only: exception, FATAL
use univ_convert_units_mod, only: rad_to_deg_f4
use postp_main_memory_mod, only: what_vtype, nvar_want, var_want_to_read,  &
     nzp, ztrnsfm, ztop, zt, what_vdesc, what_vunits, what_vlevs, what_ofile,  &
     xt, yt, plat, plon, nbins, dust_ig, H2Oi_ig, CO2i_ig,  &
     dust_present, H2Oi_present, CO2i_present, zm, ustar_t_table_basalt,  &
     rho_basalt, ustar_t_table_dust_agglom, rho_dust_agglom,  &
     ustar_t_table_andesite, rho_andesite, Ls_curr, vl_result_t2,  &
     vl_result_t3, vl_result_t4, vl_result_t5, vl_scr1_t2, vl_scr1_t3,  &
     vl_scr2_t3, vl_scr3_t3, vl_scr4_t3, vl_scr5_t3, vl_scr6_t3, vl_scr7_t3,  &
     vl_scr1_t4, vl_scr2_t4, vl_scr3_t4, vl_scr4_t4, vl_scr2_t2, vl_scr3_t2,  &
     vl_scr1_t5, vl_scr2_t5, vctr1_az, vctr2_az, vctr3_az, vctr4_az,  &
     cfg_spatial, full_az_scr_needed, vl_scr1_faz_t3, vl_scr2_faz_t3,  &
     vl_scr3_faz_t3, var_name_len, full_xy_scr_needed, full_azxy_scr_needed,  &
     vl_scr1_fazxy_t3, vl_scr2_fazxy_t3, vl_scr4_faz_t3, full_sz_scr_needed,  &
     vl_scr1_fsz_t5, vl_scr2_fsz_t5, vl_scr3_fsz_t5, vl_scr4_fsz_t5,  &
     have_full_az
use aermic_rad_mem_mod, only: mvar_dust_params, ustar_t_min, ust_tab_nent_T,  &
     ust_tab_nent_rhof
use postp_getdata_mod, only: get_PRAMS_data, PRAMS_Cartpost_interp
use postp_comp_mod, only: comp_modlyr_z, comp_rtgx, comp_co2tsat_sfc,  &
     comp_press, comp_titvar_sfc, comp_sbpstempk, comp_2daer_optdepth,  &
     comp_co2tsat_lyr, comp_3daer_variance, comp_3daer_mnrad,  &
     comp_3daer_optdepth, comp_aer_vfall, comp_3d_esum, comp_aer_numc2massc,  &
     comp_aer_coremc_kludge, comp_tempk, comp_density, comp_mass_mr,  &
     comp_aer_pshell_massc, comp_ustar_neutral, comp_speed, comp_avgv,  &
     comp_avgu, comp_richnum, comp_sfc_t2th, comp_moreduct_x,  &
     comp_tN_rotate, comp_avgw, comp_h2ov_cond_coldepth, comp_vertsum,  &
     comp_sfc_stress, comp_density_lin, comp_pert_over_base,  &
     comp_supsat_co2i, comp_gas_vapor_press, comp_supsat_H2Ov,  &
     comp_nh_pressp, comp_ustar_t, comp_dir, comp_sfc_air_density,  &
     comp_sfc_value_extrp, comp_brunt_vaisala, comp_aer_coremc_kludge_2D,  &
     comp_aer_numc2massc_2D, comp_sfc_press, comp_sfclyr_density,  &
     comp_shf_sfclyr
use aemic_drvint_mod, only: pflift_ustar_t_table
use za_vardust_spec_mod, only: za_vardust_spec_init, za_vardust_spec
use namelist_bspec_mod, only: default_input_data_dir_bspec
use body_specific_constants_mod, only: cpi

implicit none

!=== Argument declarations:

integer, intent(IN) :: tgt_igd, ifn, cfg_igd
character(len=*), intent(IN) :: var_name
logical, intent(IN) :: this_is_init_pass

!== Local declarations:

character(len=MAX_PROC_NAME_LEN), parameter ::  &
     procedure_name = 'postp_varlib'

logical :: varname_is_unknown, tN_requested
integer :: lenvn, ig, e, j, ios, wv_tog, atype, vkind, vsubkind, nb, ntype,  &
     ib, ic, ninv, inv, btype
character(len=80) :: ctmp
character(len=6), dimension(3) :: ctmp6_1
character(len=3) :: wv_desc
real(f4) :: ftmp, X_lev

!=== Executable statements:

full_az_scr_needed = .FALSE.; full_xy_scr_needed = .FALSE.
full_azxy_scr_needed = .FALSE.; full_sz_scr_needed = .FALSE.

lenvn = len_trim(var_name)

!## theta
if (var_name(1:lenvn) == 'theta') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'THETA'/)
   else
      call get_PRAMS_data(vl_result_t3, 'THETA', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = '(dry) atm potential temperature;'
      what_vunits = 'K;'
    end if

!## thp
else if (var_name(1:lenvn) == 'thp') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'THP'/)
   else
      call get_PRAMS_data(vl_result_t3, 'THP', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm potential temperature;'
      what_vunits = 'K;'
    end if

!## tempk
else if (var_name(1:lenvn) == 'tempk') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'THETA','PP   ','PI0  '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

      call comp_tempk(vl_result_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = '(dry) atm temperature;'
      what_vunits = 'K;'
   end if

!## density
else if (var_name(1:lenvn) == 'density') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'THETA','PP   ','PI0  '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

      call comp_density(vl_result_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = '(dry) atm density (nonlinearized);'
      what_vunits = 'kg/m^3;'
   end if

!## sfclyr_density
else if (var_name(1:lenvn) == 'sfclyr_density') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'THETA','PP   ','PI0  '/)
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)

      call comp_sfclyr_density(vl_result_t2%p2D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D)

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'estimated sfclyr (dry) air density (nonlinearized);'
      what_vunits = 'kg/m^3;'
   end if

!## density_lin
else if (var_name(1:lenvn) == 'density_lin') then

   if (this_is_init_pass) then
      nvar_want = 5
      var_want_to_read(1:nvar_want) = (/'THETA','PP   ','DN0  ','PI0  '  &
           ,'TH0  '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'TH0', tgt_igd, cfg_igd, ifn)
      call comp_pert_over_base(vl_scr3_t3%p3D, vl_scr1_t3%p3D,  &
           vl_scr2_t3%p3D, 0)    ! scr3_t3 is (theta'/theta0)
      call get_PRAMS_data(vl_scr1_t3, 'PP', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call comp_pert_over_base(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
           vl_scr2_t3%p3D, 1)      ! scr4_t3 is (pi'/pi0)
      call get_PRAMS_data(vl_scr1_t3, 'DN0', tgt_igd, cfg_igd, ifn)
      call comp_density_lin(vl_result_t3%p3D, vl_scr1_t3%p3D, vl_scr4_t3%p3D,  &
           vl_scr3_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = '(dry) atm density (linearized);'
      what_vunits = 'kg/m^3;'
   end if

!## press
else if (var_name(1:lenvn) == 'press') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'PI0','PP '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr1_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'

      call comp_press(vl_result_t3%p3D, vl_scr1_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm pressure;'
      what_vunits = 'Pa;'
   end if

!## pi
else if (var_name(1:lenvn) == 'pi') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'PI0','PP '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_result_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm normalized Exner function;'
      what_vunits = 'J/(K.kg);'
   endif

!## exner_f
else if (var_name(1:lenvn) == 'exner_f') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'PI0','PP '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_result_t3%p3D = (vl_scr1_t3%p3D+vl_scr2_t3%p3D)*cpi ! pi = (pi0+pi')/cp

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm Exner function;'
      what_vunits = 'none;'
   end if

!## nh_pressp
else if (var_name(1:lenvn) == 'nh_pressp') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   ','TOPT '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      call comp_press(vl_scr3_faz_t3%p3D, vl_scr2_faz_t3%p3D) ! p (scr3_faz_t3)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      call comp_nh_pressp(vl_result_t3%p3D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D, vl_scr3_faz_t3%p3D, vl_scr1_t2%p2D, tgt_igd,  &
           ifn, cfg_spatial(cfg_igd)%i_AZ)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm NH pressure pert;'
      what_vunits = 'Pa;'
   end if

!## brunt_vaisala_f
else if (var_name(1:lenvn) == 'brunt_vaisala_f') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   ','TOPT '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      call comp_brunt_vaisala(vl_result_t3%p3D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D, tgt_igd, ifn,  &
           cfg_spatial(cfg_igd)%i_AZ)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'Brunt-Vaisala frequency;'
      what_vunits = '1/s;'
   end if

!## pi0
else if (var_name(1:lenvn) == 'pi0') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'PI0'/)
   else
      call get_PRAMS_data(vl_result_t3, 'PI0', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm normalized Exner function (refstate);'
      what_vunits = 'J/(K.kg);'
   end if

!## dn0
else if (var_name(1:lenvn) == 'dn0') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'DN0'/)
   else
      call get_PRAMS_data(vl_result_t3, 'DN0', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = '(dry) atm density (refstate);'
      what_vunits = 'kg/m^3;'
   end if

!## th0
else if (var_name(1:lenvn) == 'th0') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TH0'/)
   else
      call get_PRAMS_data(vl_result_t3, 'TH0', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm potential temperature (refstate);'
      what_vunits = 'K;'
   end if

!## H2Ov_mc
else if (var_name(1:lenvn) == 'H2Ov_mc') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'RV'/)
   else
      call get_PRAMS_data(vl_result_t3, 'RV', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm water vapor mass_conc;'
      what_vunits = 'g/cm^3;'
   end if

!## H2Ov_mr
else if (var_name(1:lenvn) == 'H2Ov_mr') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   ','RV   '/)
  else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

      ! Calc air density [kg/m^3]; put in scr3_t3:
      call comp_density(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

      call get_PRAMS_data(vl_scr1_t3, 'RV', tgt_igd, cfg_igd, ifn)

      call comp_mass_mr(vl_result_t3%p3D, vl_scr1_t3%p3D, vl_scr3_t3%p3D,  &
           .TRUE.)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'atm water vapor mixing ratio;'
      what_vunits = 'g/g;'
   end if

!## H2Ot_mc, H2Ot_mr, H2Ot_pcp_cdepth
else if (var_name(1:5) == 'H2Ot_') then

   if (var_name(6:7) == 'mc') then
      vsubkind = 1; full_az_scr_needed = .FALSE.
   else if (var_name(6:7) == 'mr') then
      vsubkind = 2; full_az_scr_needed = .FALSE.
   else if (var_name(6:15) == 'pcp_cdepth') then
      vsubkind = 3; full_az_scr_needed = .TRUE.
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   if (H2Oi_present) then
      if (CO2i_present) then
         atype = 2    ! Water ice contribution in H2Oi and CO2i aerosols
      else
         atype = 1   ! Water ice contribution only in H2Oi aerosol
      end if
   else
      atype = 0   ! No water ice aerosol contribution
   end if

   if (atype == 0) then
      if (this_is_init_pass) then
         nvar_want = 1
         var_want_to_read(1:nvar_want) = (/'RV'/)
      end if
   else if (atype == 1) then
      nb = nbins(H2Oi_ig)
      ninv = 2
      ctmp6_1(1) = 'H2OINC'; ctmp6_1(2) = 'H2ODCR'
      if (this_is_init_pass) then      
         nvar_want = ninv*nb+1
         do inv=1,ninv
            do ib=1,nb      ! # bins
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv),ib  ! Actual var name
               var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
            end do
         end do
         var_want_to_read(ninv*nb+1) = 'RV'
      end if
   else if (atype == 2) then
      ninv = 2+1
      ctmp6_1(1) = 'H2OINC'; ctmp6_1(2) = 'H2ODCR'; ctmp6_1(3) = 'CO2WCR'
      if (this_is_init_pass) then      
         nvar_want = 2*nbins(H2Oi_ig)+1*nbins(CO2i_ig)+1
         nb = nbins(H2Oi_ig)
         do inv=1,2
            do ib=1,nb    ! # bins
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv),ib  ! Actual var name
               var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
            end do
         end do
         nb = nbins(CO2i_ig)
         do inv=3,3
            do ib=1,nb    ! # bins
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv),ib  ! Actual variable name
               var_want_to_read(2*nbins(H2Oi_ig)+ib) = ctmp(1:var_name_len)
            end do
         end do
         nb = 2*nbins(H2Oi_ig)+1*nbins(CO2i_ig)
         var_want_to_read(nb+1) = 'RV'
      end if
   endif
   if (vsubkind == 2) then
      if (atype == 0) then
         if (this_is_init_pass) then         
            nvar_want = 4
            var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   ','RV   '/)
         end if
      else
         if (this_is_init_pass) then
            var_want_to_read(nvar_want+1) = 'THETA'
            var_want_to_read(nvar_want+2) = 'PI0'
            var_want_to_read(nvar_want+3) = 'PP'
            nvar_want = nvar_want+3
         end if
      end if
   else if (vsubkind == 3) then
      if (this_is_init_pass) then
         var_want_to_read(nvar_want+1) = 'TOPT'
         nvar_want = nvar_want+1
      end if
   end if

   if (.NOT.this_is_init_pass) then
      call get_PRAMS_data(vl_scr3_t3, 'RV', tgt_igd, cfg_igd, ifn)
      if (full_az_scr_needed) call get_PRAMS_data(vl_scr3_faz_t3, 'RV',  &
           tgt_igd, cfg_igd, ifn, full_az=.TRUE.)

      if (atype == 1 .or. atype == 2) then
         do ib=1,nbins(H2Oi_ig)     ! Sum over all bins
            write (ctmp, '(a6,i2.2)') ctmp6_1(1),ib  ! Particle # conc
            call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
            if (full_az_scr_needed) call get_PRAMS_data(vl_scr1_faz_t3, ctmp,  &
                 tgt_igd, cfg_igd, ifn, full_az=.TRUE.)
            write (ctmp, '(a6,i2.2)') ctmp6_1(2),ib  ! "Core" mass conc
            call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
            if (full_az_scr_needed) call get_PRAMS_data(vl_scr2_faz_t3, ctmp,  &
                 tgt_igd, cfg_igd, ifn, full_az=.TRUE.)
          
            ! Compute primary shell mass conc (add to scr3_t3):
            call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr2_t3%p3D,  &
                 vl_scr1_t3%p3D, vl_scr3_t3%p3D, 2, ib, 1,  &
                 cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
            if (full_az_scr_needed) call comp_aer_pshell_massc(1,  &
                 vl_scr2_faz_t3%p3D, vl_scr2_faz_t3%p3D, vl_scr1_faz_t3%p3D,   &
                 vl_scr3_faz_t3%p3D, 2, ib, 1, cfg_spatial(cfg_igd)%i_fAZ,  &
                 nzp(tgt_igd,ifn))
         end do
      end if

      if (atype == 2) then
         do ib=1,nbins(CO2i_ig)     ! Sum over all bins
            write (ctmp, '(a6,i2.2)') ctmp6_1(3),ib  ! Actual variable name
            call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
            if (full_az_scr_needed) call get_PRAMS_data(vl_scr2_faz_t3, ctmp,  &
                 tgt_igd, cfg_igd, ifn, full_az=.TRUE.)
            ! Sum aerosol mass concentrations (starting with value of scr3_t3):
            call comp_3D_esum(vl_scr3_t3%p3D, vl_scr2_t3%p3D, 1,  &
                 cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
            if (full_az_scr_needed) call comp_3D_esum(vl_scr3_faz_t3%p3D,  &
                 vl_scr2_faz_t3%p3D, 1, cfg_spatial(cfg_igd)%i_fAZ,  &
                 nzp(tgt_igd,ifn))
         end do
      end if

      vl_result_t3%p3D = vl_scr3_t3%p3D
      what_vdesc = 'total atm water substance mass_conc;'
      what_vunits = 'g/cm^3;'

      if (vsubkind == 2) then     ! Compute mixing ratio

         call get_PRAMS_data(vl_scr4_t3, 'PI0', tgt_igd, cfg_igd, ifn)
         call get_PRAMS_data(vl_scr5_t3, 'PP', tgt_igd, cfg_igd, ifn)
         vl_scr5_t3%p3D = vl_scr4_t3%p3D+vl_scr5_t3%p3D   ! pi = pi0+pi'
         call get_PRAMS_data(vl_scr4_t3, 'THETA', tgt_igd, cfg_igd, ifn)

         ! Calc air density [kg/m^3]; put in scr2_t3:
         call comp_density(vl_scr2_t3%p3D, vl_scr4_t3%p3D, vl_scr5_t3%p3D)

         ! Calculate mass mixing ratio:
         call comp_mass_mr(vl_result_t3%p3D, vl_scr3_t3%p3D, vl_scr2_t3%p3D,  &
              .TRUE.)

         what_vdesc = 'total atm water substance mixing ratio;'
         what_vunits = 'g/g;'

      else if (vsubkind == 3) then

         call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
         call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

         ! Compute precipitable water (if condensed) depth in column:
         call comp_H2Ov_cond_coldepth(vl_scr3_faz_t3%p3D, vl_result_t2%p2D,  &
              vl_scr1_t2%p2D, zm(:,tgt_igd,ifn))

         what_vdesc = 'precip H2O_tot in column (cond. depth);'
         what_vunits = '[cm];'

      end if

      if (vsubkind <= 2) then
         what_vtype = 3
         what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      else if (vsubkind == 3) then
         what_vtype = 2
         what_vlevs = 0
      end if
      what_ofile = 'atm'

   end if

!## sspct_H2Oi
else if (var_name(1:lenvn) == 'sspct_H2Oi') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'THETA','RV   ','PI0  ','PP   '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)
      ! Compute air temperature (put in scr3_t3):
      call comp_tempk(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

      ! Extract RV (put in scr1_t3):
      call get_PRAMS_data(vl_scr1_t3, 'RV', tgt_igd, cfg_igd, ifn)

      ! Compute saturation vapor pressure over ice (put in scr2_t3); scr4_t3 is
      !  a dummy array in this case:
      call comp_gas_vapor_press(vl_scr3_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr4_t3%p3D, 'W')

      ! Compute supersaturation:
      call comp_supsat_H2Ov(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_result_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'supersaturation of H2O over ice;'
      what_vunits = 'pct of supersaturation;'
   end if

!## sspct_CO2i
else if (var_name(1:lenvn) == 'sspct_CO2i') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'SCLR001','THETA  ','PI0    ',  &
           'PP     '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)
      ! Compute air density (put in scr5_t3):
      call comp_density(vl_scr5_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)
      ! Compute air temperature (put in scr3_t3):
      call comp_tempk(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

      ! Get CO2 gas fraction (put in scr1_t3)::
      call get_PRAMS_data(vl_scr1_t3, 'SCLR001', tgt_igd, cfg_igd, ifn)

      ! Compute saturation vapor pressure over ice (put in scr2_t3); scr4_t3 is
      !  a dummy array in this case:
      call comp_gas_vapor_press(vl_scr3_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr4_t3%p3D, 'C')

      ! Compute supersaturation:
      call comp_supsat_CO2i(vl_scr5_t3%p3D, vl_scr1_t3%p3D, vl_scr3_t3%p3D,  &
           vl_scr2_t3%p3D, vl_result_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'supersaturation of CO2 over ice;'
      what_vunits = 'pct of supersaturation;'
   end if

!## vap_prss_H2Oi
else if (var_name(1:lenvn) == 'vap_prss_H2Oi') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)
      ! Compute air temperature (put in scr3_t3):
      call comp_tempk(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

      ! Compute saturation vapor pressure over ice; scr4_t3 is
      !  a dummy array in this case:
      call comp_gas_vapor_press(vl_scr3_t3%p3D, vl_result_t3%p3D,  &
           vl_scr4_t3%p3D, 'W')

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'H2O sat vapor pressure over ice;'
      what_vunits = 'Pa;'
   end if

!## vap_prss_CO2i
else if (var_name(1:lenvn) == 'vap_prss_CO2i') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)
      ! Compute air temperature (put in scr3_t3):
      call comp_tempk(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

      ! Compute saturation vapor pressure over ice; scr4_t3 is
      !  a dummy array in this case:
      call comp_gas_vapor_press(vl_scr3_t3%p3D, vl_result_t3%p3D,  &
           vl_scr4_t3%p3D, 'C')

       what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'CO2 sat vapor pressure over ice;'
      what_vunits = 'Pa;'
   end if

!## tempk_Xm
else if (var_name(1:6) == 'tempk_' .and. var_name(lenvn:lenvn) == 'm') then

   if (this_is_init_pass) then
      nvar_want = 9
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK','UP      ','VP      ',  &
           'PI0     ','PP      ','THETA   ','TSTAR   ','Z0NET   ','TOPT    '/)
      full_az_scr_needed = .TRUE.; full_azxy_scr_needed = .TRUE.
   else
      ctmp = ';'; ctmp = var_name(7:lenvn-1); ic = index(ctmp, '_')
      if (ic > 0) ctmp(ic:ic) = '.'
      read (ctmp, *, iostat=ios) X_lev
      if (ios /= 0) call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' not found in varlib.')

      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      ! Calculate wind speed (put in scr3_faz_t3):
      call comp_speed(vl_scr1_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
           vl_scr3_faz_t3%p3D)

      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_t4, 'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      ! Convert sfc temperature to potential temp (put in scr2_t4; s2D):
      call comp_sfc_T2th(vl_scr2_t4%p3D, vl_scr1_t4%p3D, vl_scr2_faz_t3%p3D)

      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr1_t4, 'TSTAR', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr3_t4, 'Z0NET', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      ! Calculate MO-reduced tempk:
      !  (NOTE: scr4_t3, scr4_t4, scr2_t2 are used as dummy arrays)
      call comp_MOreduct_X(vl_scr3_faz_t3%p3D, vl_scr4_t3%p3D,  &
           vl_scr4_t3%p3D, vl_scr1_faz_t3%p3D, vl_scr2_t4%p3D,  &
           vl_scr4_t4%p3D, vl_scr1_t4%p3D, vl_scr3_t4%p3D,  &
           vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D, zt(2,tgt_igd,ifn),  &
           ztop, X_lev, vl_result_t2%p2D, vl_scr2_t2%p2D, vl_scr2_t2%p2D, 'T')

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      ctmp = ';'
      write (ctmp, '(a1,f7.2,a8)') '(', X_lev, ' m AGL);'
      what_vdesc = 'MO-reduced atm temperature '//trim(ctmp)
      what_vunits = 'K;'
   end if

!## dn_zsum
else if (var_name(1:lenvn) == 'dn_zsum') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   ','TOPT '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call comp_density(vl_scr3_faz_t3%p3D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D)   ! Air density in scr3_t3

      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      call comp_vertsum(vl_result_t2%p2D, vl_scr3_faz_t3%p3D, vl_scr1_t2%p2D,  &
           2, nzp(tgt_igd,ifn)-1, tgt_igd, ifn)

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'atm mass in unit column;'  ! uses nonlinearized density
      what_vunits = 'kg;'
   end if

!## dn_lin_zsum
else if (var_name(1:lenvn) == 'dn_lin_zsum') then

   if (this_is_init_pass) then
      nvar_want = 6
      var_want_to_read(1:nvar_want) = (/'THETA','PP   ','TOPT ','TH0  '  &
           ,'PI0  ','DN0  '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2 

      call get_PRAMS_data(vl_scr2_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr3_faz_t3, 'TH0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call comp_pert_over_base(vl_scr4_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
           vl_scr3_faz_t3%p3D, 0)    ! scr4_faz_t3 is (theta'/theta0)

      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr3_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call comp_pert_over_base(vl_scr1_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
           vl_scr3_faz_t3%p3D, 0)    ! scr1_faz_t3 is (pi'/pi0)

      call get_PRAMS_data(vl_scr2_faz_t3, 'DN0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)

      call comp_density_lin(vl_scr3_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
           vl_scr1_faz_t3%p3D, vl_scr4_faz_t3%p3D)  ! put in scr3_faz_t3

      call comp_vertsum(vl_result_t2%p2D, vl_scr3_faz_t3%p3D, vl_scr1_t2%p2D,  &
           2, nzp(tgt_igd,ifn)-1, tgt_igd, ifn)

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'atm mass in unit col (lin);'  ! uses linearized density
      what_vunits = 'kg;'
   end if

!## sfc_stress
else if (var_name(1:lenvn) == 'sfc_stress') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) =  &
           (/'SFCTEMPK', 'USTAR   ', 'PI0     ', 'PP      '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_t4, 'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'

      ! Surface air density (put in scr2_t4; s2D): 
      call comp_sfc_air_density(vl_scr2_t4%p3D, vl_scr1_t4%p3D,  &
           vl_scr2_faz_t3%p3D)

      call get_PRAMS_data(vl_scr1_t4, 'USTAR', tgt_igd, cfg_igd, ifn)

      call comp_sfc_stress(vl_scr1_t4%p3D, vl_scr2_t4%p3D, vl_result_t4%p3D)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'aerodynamic sfc shear stress;'
      what_vunits = 'Pa;'
   endif

!## sfc_stress_t_bs, ustar_t_bs
else if (var_name(1:lenvn) == 'sfc_stress_t_bs' .or.  &
     var_name(1:lenvn) == 'ustar_t_bs') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK', 'PI0     ','PP      '/)
      full_az_scr_needed = .TRUE.

      ! Generate ustar_t lookup table for basalt particles:
      if (.NOT. ustar_t_table_basalt%initialized) then
         ustar_t_table_basalt%ust_tab_T_min = 50._f4   !
         ustar_t_table_basalt%ust_tab_T_max = 350._f4  ! [K]
         ustar_t_table_basalt%ust_tab_T_inc = ONE_f4   !
         ustar_t_table_basalt%ust_tab_rhof_min = 1.e-4_f4   !
         ustar_t_table_basalt%ust_tab_rhof_max = 0.2_f4     ! [kg/m^3]
         ustar_t_table_basalt%ust_tab_rhof_inc = 1.e-4_f4   !
         ustar_t_table_basalt%ust_tab_nent_T =  &
              nint((ustar_t_table_basalt%ust_tab_T_max-  &
              ustar_t_table_basalt%ust_tab_T_min)/  &
              ustar_t_table_basalt%ust_tab_T_inc)
         ustar_t_table_basalt%ust_tab_nent_rhof =  &
              nint((ustar_t_table_basalt%ust_tab_rhof_max-  &
              ustar_t_table_basalt%ust_tab_rhof_min)/  &
              ustar_t_table_basalt%ust_tab_rhof_inc)
         allocate(ustar_t_table_basalt%ustar_t_min(  &
              ustar_t_table_basalt%ust_tab_nent_rhof,  &
              ustar_t_table_basalt%ust_tab_nent_T), stat=e)
         if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
              'sfc_stress_t_bs_1')

         ! Kludge: set/allocate these external module variables:
         ust_tab_nent_rhof = ustar_t_table_basalt%ust_tab_nent_rhof
         ust_tab_nent_T = ustar_t_table_basalt%ust_tab_nent_T
         allocate(ustar_t_min(ustar_t_table_basalt%ust_tab_nent_rhof,  &
              ustar_t_table_basalt%ust_tab_nent_T), stat=e)
         if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
              'sfc_stress_t_bs_2')

         write (*, '(a)') 'Initializing basalt saltation threshold table ...'
         call pflift_ustar_t_table(rho_basalt,  &
              ustar_t_table_basalt%ust_tab_T_min,  &
              ustar_t_table_basalt%ust_tab_T_inc,  &
              ustar_t_table_basalt%ust_tab_rhof_min,  &
              ustar_t_table_basalt%ust_tab_rhof_inc)

         ! Kludge: 
         ustar_t_table_basalt%ustar_t_min = ustar_t_min
         deallocate(ustar_t_min, stat=e)
         if (e /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'sfc_stress_t_bs_1')

         ustar_t_table_basalt%initialized = .TRUE.
      end if

   else
      call get_PRAMS_data(vl_scr1_t4,'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'

      ! Surface air density (put in scr2_t4; s2D): 
      call comp_sfc_air_density(vl_scr2_t4%p3D, vl_scr1_t4%p3D,  &
           vl_scr2_faz_t3%p3D)

      ! ustar_t (put in scr3_t4; s2D):
      call comp_ustar_t(vl_scr1_t4%p3D, vl_scr2_t4%p3D, ustar_t_table_basalt,  &
           vl_scr3_t4%p3D)

      if (var_name(1:lenvn) == 'ustar_t_bs') then
         vl_result_t4%p3D = vl_scr3_t4%p3D
         what_vdesc = 'min_salt_thr u* (basalt 3000 kg/m^3);'
         what_vunits = 'm/s;'
      else
         call comp_sfc_stress(vl_scr3_t4%p3D, vl_scr2_t4%p3D, vl_result_t4%p3D)
         what_vdesc = 'min_salt_thr sfc_stress (basalt 3000 kg/m^3);'
         what_vunits = 'Pa;'
      end if

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
   end if

!## sfc_stress_t_an, ustar_t_an
else if (var_name(1:lenvn) == 'sfc_stress_t_an' .or.  &
     var_name(1:lenvn) == 'ustar_t_an') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK', 'PI0     ','PP      '/)
      full_az_scr_needed = .TRUE.

      ! Generate ustar_t lookup table for andesite particles:
      if (.NOT. ustar_t_table_andesite%initialized) then
         ustar_t_table_andesite%ust_tab_T_min = 50._f4   !
         ustar_t_table_andesite%ust_tab_T_max = 350._f4  ! [K]
         ustar_t_table_andesite%ust_tab_T_inc = ONE_f4   !
         ustar_t_table_andesite%ust_tab_rhof_min = 1.e-4_f4   !
         ustar_t_table_andesite%ust_tab_rhof_max = 0.2_f4     ! [kg/m^3]
         ustar_t_table_andesite%ust_tab_rhof_inc = 1.e-4_f4   !
         ustar_t_table_andesite%ust_tab_nent_T =  &
              nint((ustar_t_table_andesite%ust_tab_T_max-  &
              ustar_t_table_andesite%ust_tab_T_min)/  &
              ustar_t_table_andesite%ust_tab_T_inc)
         ustar_t_table_andesite%ust_tab_nent_rhof =  &
              nint((ustar_t_table_andesite%ust_tab_rhof_max-  &
              ustar_t_table_andesite%ust_tab_rhof_min)/  &
              ustar_t_table_andesite%ust_tab_rhof_inc)
         allocate(ustar_t_table_andesite%ustar_t_min(  &
              ustar_t_table_andesite%ust_tab_nent_rhof,  &
              ustar_t_table_andesite%ust_tab_nent_T), stat=e)
         if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
              'sfc_stress_t_an_1')

         ! Kludge: set/allocate these external module variables:
         ust_tab_nent_rhof = ustar_t_table_andesite%ust_tab_nent_rhof
         ust_tab_nent_T = ustar_t_table_andesite%ust_tab_nent_T
         allocate(ustar_t_min(ustar_t_table_andesite%ust_tab_nent_rhof,  &
              ustar_t_table_andesite%ust_tab_nent_T), stat=e)
         if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
              'sfc_stress_t_an_2')

         write (*, '(a)') 'Initializing andesite saltation threshold table ...'
         call pflift_ustar_t_table(rho_andesite,  &
              ustar_t_table_andesite%ust_tab_T_min,  &
              ustar_t_table_andesite%ust_tab_T_inc,  &
              ustar_t_table_andesite%ust_tab_rhof_min,  &
              ustar_t_table_andesite%ust_tab_rhof_inc)

         ! Kludge: 
         ustar_t_table_andesite%ustar_t_min(:,:) = ustar_t_min(:,:)
         deallocate(ustar_t_min, stat=e)
         if (e /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'sfc_stress_t_an_1')

         ustar_t_table_andesite%initialized = .TRUE.
      end if

   else
      call get_PRAMS_data(vl_scr1_t4,'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'

      ! Surface air density (put in scr2_t4; s2D): 
      call comp_sfc_air_density(vl_scr2_t4%p3D, vl_scr1_t4%p3D,  &
           vl_scr2_faz_t3%p3D)

      ! ustar_t (put in scr3_t4; s2D):
      call comp_ustar_t(vl_scr1_t4%p3D, vl_scr2_t4%p3D,  &
           ustar_t_table_andesite, vl_scr3_t4%p3D)

      if (var_name(1:lenvn) == 'ustar_t_an') then
                  vl_result_t4%p3D = vl_scr3_t4%p3D
         what_vdesc = 'min_salt_thr u* (andesite 2800 kg/m^3);'
         what_vunits = 'm/s;'
      else
         call comp_sfc_stress(vl_scr3_t4%p3D, vl_scr2_t4%p3D, vl_result_t4%p3D)
         what_vdesc = 'min_salt_thr sfc_stress (andesite 2800 kg/m^3);'
         what_vunits = 'Pa;'
      end if

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
   end if

!## sfc_stress_t_ag, ustar_t_ag
else if (var_name(1:lenvn) == 'sfc_stress_t_ag' .or.  &
     var_name(1:lenvn) == 'ustar_t_ag') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK', 'PI0     ','PP      '/)
      full_az_scr_needed = .TRUE.

      ! Generate ustar_t lookup table for dust_agglom particles:
      if (.NOT. ustar_t_table_dust_agglom%initialized) then
         ustar_t_table_dust_agglom%ust_tab_T_min = 50._f4   !
         ustar_t_table_dust_agglom%ust_tab_T_max = 350._f4  ! [K]
         ustar_t_table_dust_agglom%ust_tab_T_inc = ONE_f4   !
         ustar_t_table_dust_agglom%ust_tab_rhof_min = 1.e-4_f4   !
         ustar_t_table_dust_agglom%ust_tab_rhof_max = 0.2_f4     ! [kg/m^3]
         ustar_t_table_dust_agglom%ust_tab_rhof_inc = 1.e-4_f4   !
         ustar_t_table_dust_agglom%ust_tab_nent_T =  &
              nint((ustar_t_table_dust_agglom%ust_tab_T_max-  &
              ustar_t_table_dust_agglom%ust_tab_T_min)/  &
              ustar_t_table_dust_agglom%ust_tab_T_inc)
         ustar_t_table_dust_agglom%ust_tab_nent_rhof =  &
              nint((ustar_t_table_dust_agglom%ust_tab_rhof_max-  &
              ustar_t_table_dust_agglom%ust_tab_rhof_min)/  &
              ustar_t_table_dust_agglom%ust_tab_rhof_inc)
         allocate(ustar_t_table_dust_agglom%ustar_t_min(  &
              ustar_t_table_dust_agglom%ust_tab_nent_rhof,  &
              ustar_t_table_dust_agglom%ust_tab_nent_T), stat=e)
         if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
              'sfc_stress_t_ag_1')

         ! Kludge: set/allocate these external module variables:
         ust_tab_nent_rhof = ustar_t_table_dust_agglom%ust_tab_nent_rhof
         ust_tab_nent_T = ustar_t_table_dust_agglom%ust_tab_nent_T
         allocate(ustar_t_min(ustar_t_table_dust_agglom%ust_tab_nent_rhof,  &
              ustar_t_table_dust_agglom%ust_tab_nent_T), stat=e)
         if (e /= 0) call illegal_alloc_stop(for_allocation, procedure_name,  &
              'sfc_stress_t_ag_2')

         write (*, '(a)') 'Initializing dust agglomerate saltation threshold table ...'
         call pflift_ustar_t_table(rho_dust_agglom,  &
              ustar_t_table_dust_agglom%ust_tab_T_min,  &
              ustar_t_table_dust_agglom%ust_tab_T_inc,  &
              ustar_t_table_dust_agglom%ust_tab_rhof_min,  &
              ustar_t_table_dust_agglom%ust_tab_rhof_inc)

         ! Kludge: 
         ustar_t_table_dust_agglom%ustar_t_min(:,:) = ustar_t_min(:,:)
         deallocate(ustar_t_min, stat=e)
         if (e /= 0) call illegal_alloc_stop(for_deallocation,  &
              procedure_name, 'sfc_stress_t_ag_1')

         ustar_t_table_dust_agglom%initialized = .TRUE.
      end if

   else
      call get_PRAMS_data(vl_scr1_t4,'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'

      ! Surface air density (put in scr2_t4; s2D): 
      call comp_sfc_air_density(vl_scr2_t4%p3D, vl_scr1_t4%p3D,  &
           vl_scr2_faz_t3%p3D)

      ! ustar_t (put in scr3_t4; s2D):
      call comp_ustar_t(vl_scr1_t4%p3D, vl_scr2_t4%p3D,  &
           ustar_t_table_dust_agglom, vl_scr3_t4%p3D)

      if (var_name(1:lenvn) == 'ustar_t_ag') then
         vl_result_t4%p3D = vl_scr3_t4%p3D
         what_vdesc = 'min_salt_thr u* (dust agglom 870 kg/m^3);'
         what_vunits = 'm/s;'
      else
         call comp_sfc_stress(vl_scr3_t4%p3D, vl_scr2_t4%p3D, vl_result_t4%p3D)
         what_vdesc = 'min_salt_thr sfc_stress (dust agglom 870 kg/m^3);'
         what_vunits = 'Pa;'
      end if

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
   end if

!## H2Ov_pcp_cdepth
else if (var_name(1:lenvn) == 'H2Ov_pcp_cdepth') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'TOPT','RV  '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      call get_PRAMS_data(vl_scr1_faz_t3, 'RV', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.) 

      ! Compute precipitable water vapor (if condensed) depth in column:
      call comp_H2Ov_cond_coldepth(vl_scr1_faz_t3%p3D, vl_result_t2%p2D,  &
           vl_scr1_t2%p2D, zm(:,tgt_igd,ifn))

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'precip H2O_vap in column (cond. depth);'
      what_vunits = 'cm;'
   end if

!## u
else if (var_name(1:lenvn) == 'u') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'UP'/)
   else
      call get_PRAMS_data(vl_result_t3, 'UP', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'raw u (x-dir sigma-z wind component);'
      what_vunits = 'm/s;'
   endif

!## v
else if (var_name(1:lenvn) == 'v') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'VP'/)
   else
      call get_PRAMS_data(vl_result_t3, 'VP', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'raw v (y-dir sigma-z wind component);'
      what_vunits = 'm/s;'
   end if

!## u_avg
else if (var_name(1:lenvn) == 'u_avg') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'UP'/)
      full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)

      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_result_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'avg u (wind component at T-point);'
      what_vunits = 'm/s;'
   end if

!## v_avg
else if (var_name(1:lenvn) == 'v_avg') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'VP'/)
      full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)

      call comp_avgv(vl_scr1_fazxy_t3%p3D, vl_result_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'avg v (wind component at T-point);'
      what_vunits = 'm/s;'
   end if

!## u_avg_tN
else if (var_name(1:lenvn) == 'u_avg_tN') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'UP','VP'/)
      full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)

      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      call comp_tN_rotate(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           xt(:,tgt_igd,ifn), yt(:,tgt_igd,ifn), cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y, plat(tgt_igd,ifn), plon(tgt_igd,ifn))
      vl_result_t3%p3D = vl_scr1_t3%p3D

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'avg u (at T-point, true north representation);'
      what_vunits = 'm/s;'
   end if

!## v_avg_tN
else if (var_name(1:lenvn) == 'v_avg_tN') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'UP','VP'/)
      full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)

      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      call comp_tN_rotate(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           xt(:,tgt_igd,ifn), yt(:,tgt_igd,ifn), cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y, plat(tgt_igd,ifn), plon(tgt_igd,ifn))
      vl_result_t3%p3D = vl_scr2_t3%p3D

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'avg v (at T-point, true north representation);'
      what_vunits = 'm/s;'
   end if

!## Vh_mag_avg
else if (var_name(1:lenvn) == 'Vh_mag_avg') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'UP','VP'/)
      full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)

      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_speed(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_result_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'wind speed (at T-pt, model horiz);'
      what_vunits = 'm/s;'
   end if

!## Vh_dir_avg_tN
else if (var_name(1:lenvn) == 'Vh_dir_avg_tN') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'UP','VP'/)
      full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)

      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      call comp_tN_rotate(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           xt(:,tgt_igd,ifn), yt(:,tgt_igd,ifn), cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y, plat(tgt_igd,ifn), plon(tgt_igd,ifn))

      call comp_dir(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_result_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'wind dir (at T-pt, model horiz, true north representation);'
      what_vunits = 'deg_from_N;'
   end if

!## w
else if (var_name(1:lenvn) == 'w') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'WP'/)
   else
      call get_PRAMS_data(vl_result_t3, 'WP', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'raw w (z-dir wind component);'
      what_vunits = 'm/s;'
   end if

!## w_avg
else if (var_name(1:lenvn) == 'w_avg') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'WP'/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'WP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)

      call comp_avgw(vl_scr1_faz_t3%p3D, vl_result_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_AZ)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'avg w (wind component at T-point);'
      what_vunits = 'm/s;'
   end if

!## u_avg_Xm, u_avg_tN_Xm
else if (var_name(1:6) == 'u_avg_' .and. var_name(lenvn:lenvn) == 'm') then

   if (this_is_init_pass) then
      nvar_want = 9
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK','UP      ','VP      '  &
           ,'PI0     ','PP      ','THETA   ','USTAR   ','Z0NET   ','TOPT    '/)
      full_az_scr_needed = .TRUE.; full_azxy_scr_needed = .TRUE.
   else
      tN_requested = ( index(var_name(7:lenvn-1), 'tN_') > 0 )
      ctmp = ';'
      if (tN_requested) then
         ctmp = var_name(10:lenvn-1)
      else
         ctmp = var_name(7:lenvn-1)
      end if
      ic = index(ctmp, '_')
      if (ic > 0) ctmp(ic:ic) = '.'
      read (ctmp, *, iostat=ios) X_lev

      if (ios /= 0) call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' not found in varlib.')

      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      if (tN_requested) then
         ! Rotate both wind components to true north (lat/lon) orientation 
         !  (can be done here since the rotation does not change the wind
         !  magnitude):
         call comp_tN_rotate(vl_scr1_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
              xt(:,tgt_igd,ifn), yt(:,tgt_igd,ifn), cfg_spatial(cfg_igd)%i_X,  &
              cfg_spatial(cfg_igd)%i_Y, plat(tgt_igd,ifn), plon(tgt_igd,ifn))
      end if

      ! Calculate wind speed (put in scr3_faz_t3):
      call comp_speed(vl_scr1_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
           vl_scr3_faz_t3%p3D)

      call get_PRAMS_data(vl_scr1_t4, 'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr4_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr4_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      ! Convert sfc temperature to potential temp (put in scr2_t4; s2D):
      call comp_sfc_T2th(vl_scr2_t4%p3D, vl_scr1_t4%p3D, vl_scr2_faz_t3%p3D)

      call get_PRAMS_data(vl_scr2_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr1_t4, 'USTAR', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr3_t4, 'Z0NET', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      ! Calculate MO-reduced u_avg:
      !  (NOTE: scr4_t3, scr4_t4, scr2_t2 are used as dummy arrays)
      call comp_MOreduct_X(vl_scr3_faz_t3%p3D, vl_scr1_faz_t3%p3D,  &
           vl_scr4_t3%p3D, vl_scr2_faz_t3%p3D, vl_scr2_t4%p3D,  &
           vl_scr1_t4%p3D, vl_scr4_t4%p3D, vl_scr3_t4%p3D, vl_scr4_t3%p3D,  &
           vl_scr1_t2%p2D, zt(2,tgt_igd,ifn), ztop, X_lev, vl_scr2_t2%p2D,  &
           vl_result_t2%p2D, vl_scr2_t2%p2D, 'U')

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      ctmp = ';'
      write (ctmp, '(a1,f7.2,a8)') '(', X_lev, ' m AGL);'
      if (tN_requested) then
         what_vdesc = 'MO-reduced atm u_avg, true north representation '//  &
              trim(ctmp)
      else
         what_vdesc = 'MO-reduced atm u_avg '//trim(ctmp)
      end if
      what_vunits = 'm/s;'
   end if

!## v_avg_Xm, v_avg_tN_Xm
else if (var_name(1:6) == 'v_avg_' .and. var_name(lenvn:lenvn) == 'm') then

   if (this_is_init_pass) then
      nvar_want = 9
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK','UP      ','VP      '  &
           ,'PI0     ','PP      ','THETA   ','USTAR   ','Z0NET   ','TOPT    '/)
      full_az_scr_needed = .TRUE.; full_azxy_scr_needed = .TRUE.
   else
      tN_requested = ( index(var_name(7:lenvn-1), 'tN_') > 0 )
      ctmp = ';'
      if (tN_requested) then
         ctmp = var_name(10:lenvn-1)
      else
         ctmp = var_name(7:lenvn-1)
      end if
      ic = index(ctmp, '_')
      if (ic > 0) ctmp(ic:ic) = '.'
      read (ctmp, *, iostat=ios) X_lev
      if (ios /= 0) call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' not found in varlib.')

      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      if (tN_requested) then
         ! Rotate both wind components to "true north" (lat/lon) orientation 
         !  (can be done here since the rotation does not change the wind
         !  magnitude):
         call comp_tN_rotate(vl_scr1_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
              xt(:,tgt_igd,ifn), yt(:,tgt_igd,ifn), cfg_spatial(cfg_igd)%i_X,  &
              cfg_spatial(cfg_igd)%i_Y, plat(tgt_igd,ifn), plon(tgt_igd,ifn))
      end if

      ! Calculate wind speed (put in scr3_faz_t3):
      call comp_speed(vl_scr1_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
           vl_scr3_faz_t3%p3D)

      call get_PRAMS_data(vl_scr1_t4, 'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr4_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr1_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr1_faz_t3%p3D = vl_scr4_faz_t3%p3D+vl_scr1_faz_t3%p3D  ! pi = pi0+pi'
      ! Convert sfc temperature to potential temp (put in scr2_t4; s2D):
      call comp_sfc_T2th(vl_scr2_t4%p3D, vl_scr1_t4%p3D, vl_scr1_faz_t3%p3D)

      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr1_t4, 'USTAR', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr3_t4, 'Z0NET', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      ! Calculate MO-reduced v_avg:
      !  (NOTE: scr4_t3, scr4_t4, scr2_t2 are used as dummy arrays)
      call comp_MOreduct_X(vl_scr3_faz_t3%p3D, vl_scr4_t3%p3D,  &
           vl_scr2_faz_t3%p3D, vl_scr1_faz_t3%p3D, vl_scr2_t4%p3D,  &
           vl_scr1_t4%p3D, vl_scr4_t4%p3D, vl_scr3_t4%p3D, vl_scr4_t3%p3D,  &
           vl_scr1_t2%p2D, zt(2,tgt_igd,ifn), ztop, X_lev, vl_scr2_t2%p2D,  &
           vl_scr2_t2%p2D, vl_result_t2%p2D, 'V')

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      ctmp = ';'
      write (ctmp, '(a1,f7.2,a8)') '(', X_lev, ' m AGL);'
      if (tN_requested) then
         what_vdesc = 'MO-reduced atm v_avg, true north representation'//  &
              trim(ctmp)
      else
         what_vdesc = 'MO-reduced atm v_avg '//trim(ctmp)
      end if
      what_vunits = 'm/s;'
   end if

!## hkm, hkh
else if (var_name(1:lenvn) == 'hkm' .or. var_name(1:lenvn) == 'hkh') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'HKM','DN0'/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'HKM', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'DN0', tgt_igd, cfg_igd, ifn)

      ! Remove reference air density weighting:
      vl_result_t3%p3D = vl_scr1_t3%p3D/vl_scr2_t3%p3D

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      if (var_name(1:lenvn) == 'hkm') then
         what_vdesc = 'momentum (horizontal) diffusion coeff (K);'
      else
         what_vdesc = 'scalar (horizontal) diffusion coeff (K);'
      end if
      what_vunits = 'm^2/s;'
   end if

!## vkm
else if (var_name(1:lenvn) == 'vkm') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'VKM'/)
   else
      call get_PRAMS_data(vl_result_t3, 'VKM', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'momentum (vertical) diffusion coeff (K);'
      what_vunits = 'm^2/s;'
   end if

!## vkh
else if (var_name(1:lenvn) == 'vkh') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'VKH'/)
   else
      call get_PRAMS_data(vl_result_t3, 'VKH', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'scalar (vertical) diffusion coeff (K);'
      what_vunits = 'm^2/s;'
   end if

!## sgs_tke
else if (var_name(1:lenvn) == 'sgs_tke') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TKEP'/)
   else
      call get_PRAMS_data(vl_result_t3, 'TKEP', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'subgrid-scale TKE (scheme-specific interpretation);'
      what_vunits = 'm^2/s^2;'
   end if

!## richnum
else if (var_name(1:lenvn) == 'richnum') then

   if (this_is_init_pass) then
      nvar_want = 9
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK','UP      ','VP      ',  &
           'PI0     ','PP      ','THETA   ','TSTAR   ','Z0NET   ','TOPT    '/)
      full_az_scr_needed = .TRUE.; full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
 
      call get_PRAMS_data(vl_scr1_t4, 'SFCTEMPK', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr3_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr4_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr3_faz_t3%p3D = vl_scr3_faz_t3%p3D+vl_scr4_faz_t3%p3D  ! pi = pi0+pi'
      ! Convert sfc temperature to potential temp (put in scr2_t4; s2D):
      call comp_sfc_T2th(vl_scr2_t4%p3D, vl_scr1_t4%p3D, vl_scr3_faz_t3%p3D)

      call get_PRAMS_data(vl_scr4_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      ! Calculate bulk Richardson number:
      call comp_richnum(cfg_spatial(cfg_igd)%i_AZ, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D, vl_scr4_faz_t3%p3D, vl_scr2_t4%p3D,  &
           vl_scr1_t2%p2D, zt(:,tgt_igd,ifn), ztop, vl_result_t3%p3D,  &
           vl_scr3_faz_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'bulk Richardson number;'
      what_vunits = 'none;'
   end if

!## wptp_sfclyr
else if (var_name(1:lenvn) == 'wptp_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TFZ'/)
   else
      call get_PRAMS_data(vl_result_t2, 'TFZ', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer covariance of w and potential temperature;'
      what_vunits = '(K.m)/s;'
   end if

!## shf_sfclyr
else if (var_name(1:lenvn) == 'shf_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'TFZ  ','THETA','PP   ','PI0  '/)
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)

      call get_PRAMS_data(vl_scr1_t2, 'TFZ', tgt_igd, cfg_igd, ifn)

      call comp_shf_sfclyr(vl_result_t2%p2D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer sensible heat flux;'
      what_vunits = 'W/m^2;'
   end if

!## wpwp_sfclyr
else if (var_name(1:lenvn) == 'wpwp_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'WFZ'/)
   else
      call get_PRAMS_data(vl_result_t2, 'WFZ', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer covariance of w and vertical wind (w);'
      what_vunits = 'm^2/s^2;'
   end if

!## upwp_sfclyr
else if (var_name(1:lenvn) == 'upwp_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'UW'/)
   else
      call get_PRAMS_data(vl_result_t2, 'UW', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer covariance of w and u-wind;'
      what_vunits = 'm^2/s^2;'
   end if

!## umomf_sfclyr
else if (var_name(1:lenvn) == 'umomf_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'UW   ','THETA','PP   ','PI0  '/)
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)

      call comp_sfclyr_density(vl_scr1_t2%p2D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D)  ! sfclyr_density in scr1_t2

      call get_PRAMS_data(vl_scr2_t2, 'UW', tgt_igd, cfg_igd, ifn)

      vl_result_t2%p2D = vl_scr1_t2%p2D*vl_scr2_t2%p2D

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer turbulent flux of u-momentum;'
      what_vunits = 'N/m^2;'
   end if

!## vmomf_sfclyr
else if (var_name(1:lenvn) == 'vmomf_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'VW   ','THETA','PP   ','PI0  '/)
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr2_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'
      call get_PRAMS_data(vl_scr1_faz_t3, 'THETA', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)

      call comp_sfclyr_density(vl_scr1_t2%p2D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_faz_t3%p3D)  ! sfclyr_density in scr1_t2

      call get_PRAMS_data(vl_scr2_t2, 'VW', tgt_igd, cfg_igd, ifn)

      vl_result_t2%p2D = vl_scr1_t2%p2D*vl_scr2_t2%p2D

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer turbulent flux of v-momentum;'
      what_vunits = 'N/m^2;'
   end if

!## vpwp_sfclyr
else if (var_name(1:lenvn) == 'vpwp_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'VW'/)
   else
      call get_PRAMS_data(vl_result_t2, 'VW', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer covariance of w and v-wind;'
      what_vunits = 'm^2/s^2;'
   end if

!## wprp_sfclyr
else if (var_name(1:lenvn) == 'wprp_sfclyr') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'QFZ'/)
   else
      call get_PRAMS_data(vl_result_t2, 'QFZ', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfclayer covariance of w and minor atmospheric specie (r);'
      what_vunits = '(Mg/kg_air)*(m/s);'
   end if

!## ustar
else if (var_name(1:lenvn) == 'ustar') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'USTAR'/)
   else
      call get_PRAMS_data(vl_result_t4, 'USTAR', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'u* (sfclyr wind turbflux scale; friction velocity);'
      what_vunits = 'm/s;'
   end if

!## tstar
else if (var_name(1:lenvn) == 'tstar') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TSTAR'/)
   else
      call get_PRAMS_data(vl_result_t4, 'TSTAR', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'theta* (sfclyr potential temperature turbflux scale);'
      what_vunits = 'K;'
   end if

!## rstar
else if (var_name(1:lenvn) == 'rstar') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'RSTAR'/)
   else
      call get_PRAMS_data(vl_result_t4, 'RSTAR', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'r* (sfclyr minor specie turbflux scale);'
      what_vunits = 'g/cm^3;'
   end if

!## ustar_nstab
else if (var_name(1:lenvn) == 'ustar_nstab') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'UP   ','VP   ','Z0NET','TOPT '/)
      full_az_scr_needed = .TRUE.; full_azxy_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fazxy_t3, 'UP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)
      call get_PRAMS_data(vl_scr2_fazxy_t3, 'VP', tgt_igd, cfg_igd, ifn,  &
           full_azxy=.TRUE.)

      call comp_avgu(vl_scr1_fazxy_t3%p3D, vl_scr1_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)
      call comp_avgv(vl_scr2_fazxy_t3%p3D, vl_scr2_faz_t3%p3D,  &
           cfg_spatial(cfg_igd)%i_fAZ, cfg_spatial(cfg_igd)%i_X,  &
           cfg_spatial(cfg_igd)%i_Y)

      ! Calculate wind speed (put in scr3_faz_t3):
      call comp_speed(vl_scr1_faz_t3%p3D, vl_scr2_faz_t3%p3D,  &
           vl_scr3_faz_t3%p3D)

      call get_PRAMS_data(vl_scr1_t4, 'Z0NET', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      ! Calculate neutral stability ustar:
      call comp_ustar_neutral(vl_scr3_faz_t3%p3D, vl_scr1_t4%p3D,  &
           vl_scr1_t2%p2D, zt(2,tgt_igd,ifn), ztop, vl_result_t4%p3D)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'u* (sfclyr turbflux scale; neutral stability);'
      what_vunits = 'm/s;'
   end if

!## sclrXXX
else if (var_name(1:4) == 'sclr') then
   ctmp = 'SCLR'//var_name(5:7)    ! Actual variable name

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/ctmp(1:var_name_len)/)
   else
      call get_PRAMS_data(vl_result_t3, ctmp, tgt_igd, cfg_igd, ifn)
 
      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'scalar tracer (#'//var_name(5:7)//');'
      what_vunits = '????;'
   end if

!## adust_pa_mc_bXX, adust_pa_mc_sum, adust_pa_nc_bXX, adust_pa_nc_sum,
!## adust_pa_mr_bXX, adust_pa_mr_sum, adust_pa_vf_bXX, adust_pa_vf_sum,
!## aH2Oi_pa_mc_bXX, aH2Oi_pa_mc_sum, aH2Oi_pa_nc_bXX, aH2Oi_pa_nc_sum,
!## aH2Oi_pa_mr_bXX, aH2Oi_pa_mr_sum, aH2Oi_pa_vf_bXX, aH2Oi_pa_vf_sum,
!## aH2Oi_df_mc_bXX, aH2Oi_df_mc_sum, aH2Oi_df_mc_suf, aH2Oi_df_mr_bXX,
!## aH2Oi_df_mr_sum, aH2Oi_df_mr_suf, aH2Oi_wf_mc_bXX, aH2Oi_wf_mc_sum,
!## aH2Oi_wf_mc_suf, aH2Oi_wf_mr_bXX, aH2Oi_wf_mr_sum, aH2Oi_wf_mr_suf,
!## aH2Oi_cr_rt_bXX, aH2Oi_cr_rt_sum,
!## aCO2i_pa_mc_bXX, aCO2i_pa_mc_sum, aCO2i_pa_nc_bXX, aCO2i_pa_nc_sum,
!## aCO2i_pa_mr_bXX, aCO2i_pa_mr_sum, aCO2i_pa_vf_bXX, aCO2i_pa_vf_sum,
!## aCO2i_df_mc_bXX, aCO2i_df_mc_sum, aCO2i_df_mr_bXX, aCO2i_df_mr_sum,
!## aCO2i_wf_mc_bXX, aCO2i_wf_mc_sum, aCO2i_wf_mr_bXX, aCO2i_wf_mr_sum,
!## aCO2i_cf_mc_bXX, aCO2i_cf_mc_sum, aCO2i_cf_mr_bXX, aCO2i_cf_mr_sum,
!## aCO2i_cr_rt_bXX, aCO2i_cr_rt_sum 
else if (var_name(1:6) == 'adust_' .or. var_name(1:6) == 'aH2Oi_' .or.  &
     var_name(1:6) == 'aCO2i_') then   ! Aerosol particle variables

   what_vdesc = ';'
   if (var_name(2:5) == 'dust') then
      what_vdesc = 'dust: '; ic = 6
   else if (var_name(2:5) == 'H2Oi') then
      what_vdesc = 'H2O_ice: '; ic = 9
   else if (var_name(2:5) == 'CO2i') then
      what_vdesc = 'CO2_ice: '; ic = 9
   end if

   if (var_name(13:13) == 'b') then
      vsubkind = 1
      what_vdesc = what_vdesc(1:ic)//'(bin#'//var_name(14:15)//') '
      read (var_name(14:15), *) ib   ! Bin to process
      ic = ic+9
   else if (var_name(13:15) == 'sum') then
      vsubkind = 2
      what_vdesc = what_vdesc(1:ic)//'(sum bins) '
      ic = ic+11
   else if (var_name(13:15) == 'suf') then
      vsubkind = 3
      what_vdesc = what_vdesc(1:ic)//'(sum fix) '
      ic = ic+10
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   if (var_name(10:11) == 'nc') then
      vkind = 1; what_vunits = '#/cm^3;'
      what_vdesc = what_vdesc(1:ic)//'#conc '
      ic = ic+6
   else if (var_name(10:11) == 'mc') then
      vkind = 2; what_vunits = 'g/cm^3;'
      what_vdesc = what_vdesc(1:ic)//'mconc '
      ic = ic+6
   else if (var_name(10:11) == 'mr') then
      vkind = 3; what_vunits = 'g/g;'
      what_vdesc = what_vdesc(1:ic)//'mixrat '
      ic = ic+7
   else if (var_name(10:11) == 'vf') then
      vkind = 4; what_vunits = 'cm/s;'
      what_vdesc = what_vdesc(1:ic)//'vfall '
      ic = ic+6
   else if (var_name(10:11) == 'rt') then
      vkind = 5; what_vunits = 'none;'
      what_vdesc = what_vdesc(1:ic)//'core_rat '
      ic = ic+9
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   if (var_name(2:5) == 'dust') then
      if (.NOT.dust_present) call exception(procedure_name, FATAL,  &
           'Dust unavailable in this run for var '//var_name(1:lenvn))
      if (var_name(7:8) == 'pa') then
         ninv = 1; ctmp6_1(1) = 'DUSTNC'; btype = 1
      end if
      nb = nbins(dust_ig)
      atype = 1
      ntype = 0
      ig = dust_ig
   else if (var_name(2:5) == 'H2Oi') then
      if (.NOT.H2Oi_present) call exception(procedure_name, FATAL,  &
           'H2O ice aerosol unavailable in this run for var '//var_name(1:lenvn))
      if (var_name(7:8) == 'pa') then
         ninv = 1    ! Only need #conc and rmass
         ctmp6_1(1) = 'H2OINC'; btype = 1 
      else if (vkind == 1) then
         call exception(procedure_name, FATAL,  &
              'Variable '//var_name(1:lenvn)//' unknown.')
      else if (vkind == 4 .and. vsubkind == 1) then
         call exception(procedure_name, FATAL,  &
              'Variable '//var_name(1:lenvn)//' unknown.')
      else
         if (var_name(7:8) == 'df') then
            if (vsubkind == 3) then
               ninv = 2; ctmp6_1(2) = 'H2ODCR'
               ctmp6_1(1) = 'H2OINC'
            else
               ninv = 1    ! Only need H2Od_mconc
               ctmp6_1(1) = 'H2ODCR'
            end if
            btype = 2
         else if (var_name(7:8) == 'wf') then
            ninv = 2    ! Need #conc, H2Od_mconc, and rmass
            ctmp6_1(1) = 'H2OINC'; ctmp6_1(2) = 'H2ODCR'
            btype = 3
         else if (var_name(7:8) == 'cr' .and. vkind == 5) then
            ninv = 2    ! Need #conc, H2Od_mconc, and rmass
            ctmp6_1(1) = 'H2OINC'; ctmp6_1(2) = 'H2ODCR'
            btype = 3
         else
            call exception(procedure_name, FATAL,  &
                 'Variable '//var_name(1:lenvn)//' unknown.')
         end if
      end if
      nb = nbins(H2Oi_ig)
      atype = 2
      ntype = 0
      ig = H2Oi_ig
   else if (var_name(2:5) == 'CO2i') then
      if (.NOT.CO2i_present) call exception(procedure_name, FATAL,  &
           'CO2 ice aerosol unavailable in this run for var '//var_name(1:lenvn))
      if (var_name(7:8) == 'pa') then
         ninv = 1    ! Only need #conc and rmass
         ctmp6_1(1) = 'CO2INC'; btype = 1
      else if (vkind == 1) then
         call exception(procedure_name, FATAL,  &
              'Variable '//var_name(1:lenvn)//' unknown.')
      else if (vkind == 4 .and. vsubkind == 1) then
         call exception(procedure_name, FATAL,  &
              'Variable '//var_name(1:lenvn)//' unknown.')
      else
         if (var_name(7:8) == 'df') then
            ninv = 1    ! Only need CO2d_mconc
            ctmp6_1(1) = 'CO2DCR'; btype = 2
         else if (var_name(7:8) == 'wf') then
            ninv = 1    ! Only need CO2w_mconc
            ctmp6_1(1) = 'CO2WCR'; btype = 2
         else if (var_name(7:8) == 'cf') then
            ninv = 3    ! Need #conc, CO2d_mconc, CO2w_mconc, and rmass
            ctmp6_1(1) = 'CO2INC'; ctmp6_1(2) = 'CO2DCR'; ctmp6_1(3) = 'CO2WCR'
            btype = 4
         else if (var_name(7:8) == 'cr' .and. vkind == 5) then
            ninv = 3    ! Need #conc, CO2d_mconc, CO2w_mconc, and rmass
            ctmp6_1(1) = 'CO2INC'; ctmp6_1(2) = 'CO2DCR'; ctmp6_1(3) = 'CO2WCR'
            btype = 4
         else
            call exception(procedure_name, FATAL,  &
                 'Variable '//var_name(1:lenvn)//' unknown.')
         end if
      end if
      nb = nbins(CO2i_ig)
      atype = 3
      ntype = 0
      ig = CO2i_ig
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   what_vdesc = what_vdesc(1:ic)//';'   ! Finish description string

   if (vsubkind == 1) then     ! Process a single bin only

      if (vkind == 1) then    ! # concentration
         if (this_is_init_pass) then
            nvar_want = ninv
            do inv=1,ninv
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib ! Actual variable name
               var_want_to_read(inv) = ctmp(1:var_name_len)
            end do
         else
            write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
            call get_PRAMS_data(vl_result_t3, ctmp, tgt_igd, cfg_igd, ifn)
         end if

      else if (vkind == 2) then    ! mass concentration
         if (this_is_init_pass) then
            nvar_want = ninv
            do inv=1,ninv
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib ! Actual variable name
               var_want_to_read(inv) = ctmp(1:var_name_len)
            end do
         else

            if (btype == 1) then    ! Need to convert #conc to mass
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               ! Convert from #conc to mconc:
               call comp_aer_numc2massc(vl_result_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Init
               call comp_aer_numc2massc(vl_result_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Calc

            else if (btype == 2) then    
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
               call get_PRAMS_data(vl_result_t3, ctmp, tgt_igd, cfg_igd, ifn)

            else if (btype == 3) then
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core" mass conc
               call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
               ! Compute primary shell mass conc:
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr2_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_result_t3%p3D, atype, ib, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr2_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_result_t3%p3D, atype, ib, 1,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Calc

            else if (btype == 4) then
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core1" mass conc
               call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
               write (ctmp, '(a6,i2.2)') ctmp6_1(3), ib  ! "Core2" mass conc
               call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
               ! Compute primary shell mass conc:
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_result_t3%p3D, atype, ib, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_result_t3%p3D, atype, ib, 1,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Calc
            end if

         end if

      else if (vkind == 3) then    ! mass mixing ratio
         if (this_is_init_pass) then
            nvar_want = ninv+3
            do inv=1,ninv
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib ! Actual variable name
               var_want_to_read(inv) = ctmp(1:var_name_len)
            end do
            var_want_to_read(ninv+1:ninv+3) = (/'THETA','PI0  ','PP   '/)
         else
            if (btype == 1) then    ! Need to convert #conc to mass
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               ! Convert from #conc to mconc (put in scr4_t3):
               call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Init
               call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Calc

            else if (btype == 2) then    
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
               call get_PRAMS_data(vl_scr4_t3, ctmp, tgt_igd, cfg_igd, ifn)

            else if (btype == 3) then
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core" mass conc
               call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
               ! Compute primary shell mass conc (put in scr4_t3):
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr2_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_scr4_t3%p3D, atype, ib, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr2_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_scr4_t3%p3D, atype, ib, 1,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Calc

            else if (btype == 4) then
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core1" mass conc
               call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
               write (ctmp, '(a6,i2.2)') ctmp6_1(3), ib  ! "Core2" mass conc
               call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
               ! Compute primary shell mass conc (put in scr4_t3):
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_scr4_t3%p3D, atype, ib, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_scr4_t3%p3D, atype, ib, 1,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Calc

            end if

            call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
            call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
            vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
            call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

            ! Calc air density [kg/m^3]; put in scr3_t3:
            call comp_density(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

            ! Calculate mass mixing ratio:
            call comp_mass_mr(vl_result_t3%p3D, vl_scr4_t3%p3D,   &
                 vl_scr3_t3%p3D, .TRUE.)
  
         end if

      else if (vkind == 4) then    ! Particle fall speed
         if (this_is_init_pass) then
            nvar_want = 3
            var_want_to_read(1:nvar_want) = (/'THETA','PI0  ','PP   '/)
         else
            call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
            call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
            vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
            call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

            ! Calc air density [kg/m^3]; put in scr3_t3:
            call comp_density(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)
            ! Calc air temperature [K]; put in scr4_t3:
            call comp_tempk(vl_scr4_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

            ! Calculate fall speed:
            call comp_aer_vfall(vl_scr3_t3%p3D, vl_scr4_t3%p3D, ig, ib,  &
                 vl_result_t3%p3D)

         end if

      else if (vkind == 5) then    ! Particle core mass ratio
         if (this_is_init_pass) then
            nvar_want = ninv
            do inv=1,ninv
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib ! Actual variable name
               var_want_to_read(inv) = ctmp(1:var_name_len)
            end do
         else
            if (btype == 3) then
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

               ! Convert from #conc to mconc (put in scr2_t3):
               call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Init
               call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Calc

               write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core" mass conc
               call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)

               vl_result_t3%p3D = min(ONE_f4, max(ZERO_f4, vl_scr3_t3%p3D/  &
                    max(10._f4*tiny(ftmp), vl_scr2_t3%p3D)))

            else if (btype == 4) then
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               ! Convert from #conc to mconc (put in scr2_t3):
               call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Init
               call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Calc

               write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core1" mass conc
               call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)
               write (ctmp, '(a6,i2.2)') ctmp6_1(3), ib  ! "Core2" mass conc
               call get_PRAMS_data(vl_scr4_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
               vl_result_t3%p3D = min(ONE_f4, max(ZERO_f4,  &
                    (vl_scr3_t3%p3D+vl_scr4_t3%p3D)/  &
                    max(10._f4*tiny(ftmp), vl_scr2_t3%p3D)))
            end if

         end if

      end if

   else if (vsubkind == 2) then      ! Sum of all bins' contributions

      if (vkind == 1) then    ! # concentration
         if (this_is_init_pass) then
            nvar_want = ninv*nb
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
         else
            ! Initialize summing array:
            call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 0,  &
                 cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
            
            do ib=1,nb     ! Sum over all bins
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               ! Sum # concentrations of all bins:
               call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 1,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
            end do
         end if

      else if (vkind == 2) then    ! mass concentration
         if (this_is_init_pass) then
            nvar_want = ninv*nb
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
         else
            if (btype == 1) then    ! Need to convert #conc to mass
               call comp_aer_numc2massc(vl_result_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Initialize summing array
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc:
                  call comp_aer_numc2massc(vl_result_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc
               end do

            else if (btype == 2) then
               call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Sum # concentrations of all bins (put in scr2):
                  call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

            else if (btype == 3) then
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr2_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_result_t3%p3D, atype, 1, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Init sum
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp,'(a6, i2.2)') ctmp6_1(2), ib  ! "Core" mass conc
                  call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
                  ! Compute primary shell mass conc (put in scr3):
                  call comp_aer_pshell_massc(1, vl_scr2_t3%p3D,  &
                       vl_scr2_t3%p3D, vl_scr1_t3%p3D, vl_result_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Calc
               end do

            else if (btype == 4) then
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_result_t3%p3D, atype, 1, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core1" mass conc
                  call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(3), ib  ! "Core2" mass conc
                  call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
                  ! Compute primary shell mass conc (put in scr4):
                  call comp_aer_pshell_massc(1, vl_scr2_t3%p3D,  &
                       vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_result_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Calc
               end do

            end if

         end if

      else if (vkind == 3) then    ! mass mixing ratio
         if (this_is_init_pass) then
            nvar_want = ninv*nb+3
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
            var_want_to_read(ninv*nb+1:ninv*nb+3) = (/'THETA','PI0  ','PP   '/)
         else
            if (btype == 1) then    ! Need to convert #conc to mass
               call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,atype,  &
                    1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Init
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr4_t3):
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc
               end do

            else if (btype == 2) then    
               call comp_3D_esum(vl_scr4_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Sum # concentrations of all bins (put in scr4_t3):
                  call comp_3D_esum(vl_scr4_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

            else if (btype == 3) then
               call comp_aer_pshell_massc(1, vl_scr2_t3%p3D, vl_scr2_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_scr4_t3%p3D, atype, 1, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core" mass conc
                  call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
                  ! Compute primary shell mass conc (put in scr4):
                  call comp_aer_pshell_massc(1, vl_scr2_t3%p3D,  &
                       vl_scr2_t3%p3D, vl_scr1_t3%p3D, vl_scr4_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Calc
               end do

            else if (btype == 4) then
               call comp_aer_pshell_massc(2, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_scr4_t3%p3D, atype, 1, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core1" mass conc
                  call get_PRAMS_data(vl_scr2_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(3), ib  ! "Core2" mass conc
                  call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
                  ! Compute primary shell mass conc (put in scr4):
                  call comp_aer_pshell_massc(2, vl_scr2_t3%p3D,  &
                       vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr4_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Calc
               end do

            end if

            call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
            call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
            vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
            call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

            ! Calc air density [kg/m^3]; put in scr5_t3:
            call comp_density(vl_scr5_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

            ! Calculate mass mixing ratio:
            call comp_mass_mr(vl_result_t3%p3D, vl_scr4_t3%p3D,  &
                 vl_scr5_t3%p3D, .TRUE.)

         end if

      else if (vkind == 4) then    ! Particle fall speed (mass-weighted)
         if (this_is_init_pass) then
            nvar_want = ninv*nb+3
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
            var_want_to_read(ninv*nb+1:ninv*nb+3) = (/'THETA','PI0  ','PP   '/)
         else
            call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
            call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
            vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
            call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

            ! Calc air density [kg/m^3]; put in scr3_t3:
            call comp_density(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)
            ! Calc air temperature [K]; put in scr4_t3:
            call comp_tempk(vl_scr4_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

            vl_result_t3%p3D = ZERO_f4

            if (btype == 1) then    ! Need to convert #conc to mass
               call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Initialize summing array
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

                  ! Convert from #conc to mconc (put in scr2_t3):
                  call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc
                  call comp_aer_numc2massc(vl_scr5_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Init scr5_t3
                  call comp_aer_numc2massc(vl_scr5_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc scr5_t3

                  ! Calculate fall speed (put in scr1_t3):
                  call comp_aer_vfall(vl_scr3_t3%p3D, vl_scr4_t3%p3D, ig, ib,  &
                       vl_scr1_t3%p3D)

                  vl_result_t3%p3D = vl_result_t3%p3D+  &  ! Mass-weight vfall 
                       (vl_scr5_t3%p3D*vl_scr1_t3%p3D)
               end do

            else if (btype == 2) then
               call comp_3D_esum(vl_scr2_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))    
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

                  ! Sum # concentrations of all bins (put in scr2):
                  call comp_3D_esum(vl_scr2_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))

                  ! Calculate fall speed (put in scr5_t3):
                  call comp_aer_vfall(vl_scr3_t3%p3D, vl_scr4_t3%p3D, ig, ib,  &
                       vl_scr5_t3%p3D)

                  vl_result_t3%p3D = vl_result_t3%p3D+  &  ! Mass-weight vfall 
                       (vl_scr1_t3%p3D*vl_scr5_t3%p3D)
               end do

            else if (btype == 3) then
               call comp_aer_pshell_massc(1, vl_scr1_t3%p3D, vl_scr1_t3%p3D,  &
                    vl_scr6_t3%p3D, vl_scr2_t3%p3D, atype, 1, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Init sum
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr6_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core" mass conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
                  ! Compute primary shell mass conc (put in scr2_t3):
                  call comp_aer_pshell_massc(1, vl_scr1_t3%p3D,  &
                       vl_scr1_t3%p3D, vl_scr6_t3%p3D, vl_scr2_t3%p3D, atype,  &
                       ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn)) ! Calc

                  ! Sum (put in scr5_t3):
                  call comp_aer_pshell_massc(1, vl_scr1_t3%p3D,  &
                       vl_scr1_t3%p3D, vl_scr6_t3%p3D, vl_scr5_t3%p3D, atype,  &
                       1, 0, cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn)) ! Init
                  call comp_aer_pshell_massc(1, vl_scr1_t3%p3D,  &
                       vl_scr1_t3%p3D, vl_scr6_t3%p3D, vl_scr5_t3%p3D, atype,  &
                       ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Calc

                  ! Calculate fall speed (put in scr1_t3):
                  call comp_aer_vfall(vl_scr3_t3%p3D, vl_scr4_t3%p3D, ig, ib,  &
                       vl_scr1_t3%p3D)

                  vl_result_t3%p3D = vl_result_t3%p3D+  &  ! Mass-weight vfall 
                       (vl_scr5_t3%p3D*vl_scr1_t3%p3D)
               end do

            else if (btype == 4) then
               call comp_aer_pshell_massc(2, vl_scr6_t3%p3D, vl_scr7_t3%p3D,  &
                    vl_scr1_t3%p3D, vl_scr2_t3%p3D, atype, 1, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Init
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core1" mass conc
                  call get_PRAMS_data(vl_scr6_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  write (ctmp, '(a6,i2.2)') ctmp6_1(3), ib  ! "Core2" mass conc
                  call get_PRAMS_data(vl_scr7_t3, ctmp, tgt_igd, cfg_igd, ifn)
               
                  ! Compute primary shell mass conc (put in scr4_t3):
                  call comp_aer_pshell_massc(2, vl_scr6_t3%p3D,  &
                       vl_scr7_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D, atype,  &
                       ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Calc
                  call comp_aer_pshell_massc(2, vl_scr6_t3%p3D,  &
                       vl_scr7_t3%p3D, vl_scr1_t3%p3D, vl_scr5_t3%p3D, atype,  &
                       1, 0, cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn)) ! Init
                  call comp_aer_pshell_massc(2, vl_scr6_t3%p3D,  &
                       vl_scr7_t3%p3D, vl_scr1_t3%p3D, vl_scr5_t3%p3D, atype,  &
                       ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Calc

                  ! Calculate fall speed (put in scr1_t3):
                  call comp_aer_vfall(vl_scr3_t3%p3D, vl_scr4_t3%p3D, ig, ib,  &
                       vl_scr1_t3%p3D)

                  vl_result_t3%p3D = vl_result_t3%p3D+  &  ! Mass-weight vfall 
                       (vl_scr5_t3%p3D*vl_scr1_t3%p3D)
               end do

            end if

            ! Avg mass-weighted vfall:
            vl_result_t3%p3D = vl_result_t3%p3D/  &
                 max(1.e-30_f4, vl_scr2_t3%p3D)
         end if

      else if (vkind == 5) then    ! Particle core mass ratio
         if (this_is_init_pass) then
            nvar_want = ninv*nb
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
         else
            if (btype == 1) then    ! Need to convert #conc to mass
               call comp_aer_numc2massc(vl_result_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))   ! Initialize summing array
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc:
                  call comp_aer_numc2massc(vl_result_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc
               end do

            else if (btype == 2) then
               call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))    
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Sum # concentrations of all bins:
                  call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

            else if (btype == 3) then
               call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Init
               call comp_3D_esum(vl_scr4_t3%p3D, vl_scr3_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr2_t3):
                  call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc

                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core" mass conc
                  call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Sum # concentrations of all bins (put in scr4_t3):
                  call comp_3D_esum(vl_scr4_t3%p3D, vl_scr3_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

               vl_result_t3%p3D = min(ONE_f4, max(ZERO_f4, vl_scr4_t3%p3D/  & 
                    max(10._f4*tiny(ftmp), vl_scr2_t3%p3D)))

            else if (btype == 4) then
               call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Init
               call comp_3D_esum(vl_scr4_t3%p3D, vl_scr3_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               call comp_3D_esum(vl_scr6_t3%p3D, vl_scr5_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Particle # conc
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  call comp_aer_numc2massc(vl_scr2_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc
                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! "Core1" mass conc
                  call get_PRAMS_data(vl_scr3_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  call comp_3D_esum(vl_scr4_t3%p3D, vl_scr3_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
                  write (ctmp, '(a6,i2.2)') ctmp6_1(3), ib  ! "Core2" mass conc
                  call get_PRAMS_data(vl_scr5_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  call comp_3D_esum(vl_scr6_t3%p3D, vl_scr5_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

               vl_result_t3%p3D = min(ONE_f4, max(ZERO_f4, (vl_scr4_t3%p3D+  & 
                    vl_scr6_t3%p3D)/max(10._f4*tiny(ftmp), vl_scr2_t3%p3D)))

            end if

         end if

      end if

   else if (vsubkind == 3) then ! "Fixed version" sum of all bins' contributions
! DELETE ME WHEN aerosol model is fixed!!!!!!
      if (vkind == 2) then    ! mass concentration
         if (this_is_init_pass) then
            nvar_want = ninv*nb
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
         else
            call comp_aer_numc2massc(vl_scr3_t3%p3D, vl_scr1_t3%p3D, atype,  &
                 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                 nzp(tgt_igd,ifn))   ! Initialize summing array
            do ib=1,nb
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               ! Convert from #conc to mconc (put in scr3_t3):
               call comp_aer_numc2massc(vl_scr3_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Calc
            end do
            ftmp = log10(maxval(vl_scr3_t3%p3D))

            if (btype == 2) then

               call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))    
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr4_t3):
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Init
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc

                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  
                  ! Calculate clipped massconc for this bin; put in (scr1_t3)
                  call comp_aer_coremc_kludge(ftmp, vl_scr3_t3%p3D,  &
                       vl_scr4_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
                  
                  ! Sum # concentrations of all bins:
                  call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

            else if (btype == 3) then

               call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr4_t3):
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Init
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc

                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  
                  ! Calculate clipped massconc for this bin; put in (scr1_t3):
                  call comp_aer_coremc_kludge(ftmp, vl_scr3_t3%p3D,  &
                       vl_scr4_t3%p3D, vl_scr1_t3%p3D, 2,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
                  
                  ! Sum # concentrations of all bins:
                  call comp_3D_esum(vl_result_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

            end if
         end if

      else if (vkind == 3) then    ! mass mixing ratio
         if (this_is_init_pass) then
            nvar_want = ninv*nb+3
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
            var_want_to_read(ninv*nb+1:ninv*nb+3) = (/'THETA','PI0  ','PP   '/)
         else

            call comp_aer_numc2massc(vl_scr3_t3%p3D, vl_scr1_t3%p3D, atype,  &
                 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                 nzp(tgt_igd,ifn))   ! Initialize summing array
            do ib=1,nb
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               ! Convert from #conc to mconc (put in scr3_t3):
               call comp_aer_numc2massc(vl_scr3_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Calc
            end do
            ftmp = log10(maxval(vl_scr3_t3%p3D))

            if (btype == 2) then    

               call comp_3D_esum(vl_scr6_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))    
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr4_t3):
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Init
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc

                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  
                  ! Calculate clipped massconc for this bin; put in (scr1_t3):
                  call comp_aer_coremc_kludge(ftmp, vl_scr3_t3%p3D,  &
                       vl_scr4_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
                  
                  ! Sum # concentrations of all bins (put in scr6_t3):
                  call comp_3D_esum(vl_scr6_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

            else if (btype == 3) then
               call comp_3D_esum(vl_scr6_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))    
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr4_t3):
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Init
                  call comp_aer_numc2massc(vl_scr4_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc

                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  
                  ! Calculate clipped massconc for this bin; put in (scr1_t3):
                  call comp_aer_coremc_kludge(ftmp, vl_scr3_t3%p3D,  &
                       vl_scr4_t3%p3D, vl_scr1_t3%p3D, 2,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
                  
                  ! Sum # concentrations of all bins (put in scr6_t3):
                  call comp_3D_esum(vl_scr6_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))
               end do

            end if

            call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
            call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
            vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
            call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

            ! Calc air density [kg/m^3]; put in scr5_t3:
            call comp_density(vl_scr5_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

            ! Calculate mass mixing ratio:
            call comp_mass_mr(vl_result_t3%p3D, vl_scr6_t3%p3D,  &
                 vl_scr5_t3%p3D, .TRUE.)

         end if

      else if (vkind == 4) then    ! Particle fall speed (mass-weighted)
         if (this_is_init_pass) then
            nvar_want = ninv*nb+3
            do inv=1,ninv
               do ib=1,nb    ! # bins
                  write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
                  var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
               end do
            end do
            var_want_to_read(ninv*nb+1:ninv*nb+3) = (/'THETA','PI0  ','PP   '/)
         else
            call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
            call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
            vl_scr2_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'
            call get_PRAMS_data(vl_scr1_t3, 'THETA', tgt_igd, cfg_igd, ifn)

            ! Calc air density [kg/m^3]; put in scr3_t3:
            call comp_density(vl_scr3_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)
            ! Calc air temperature [K]; put in scr4_t3:
            call comp_tempk(vl_scr4_t3%p3D, vl_scr1_t3%p3D, vl_scr2_t3%p3D)

            call comp_aer_numc2massc(vl_scr7_t3%p3D, vl_scr1_t3%p3D, atype,  &
                 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                 nzp(tgt_igd,ifn))   ! Initialize summing array
            do ib=1,nb
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
               call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
               ! Convert from #conc to mconc (put in scr7_t3):
               call comp_aer_numc2massc(vl_scr7_t3%p3D, vl_scr1_t3%p3D,  &
                    atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                    nzp(tgt_igd,ifn))  ! Calc
            end do
            ftmp = log10(maxval(vl_scr7_t3%p3D))

            vl_result_t3%p3D = ZERO_f4

            if (btype == 2) then
               call comp_3D_esum(vl_scr2_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))    
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr6_t3):
                  call comp_aer_numc2massc(vl_scr6_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Init
                  call comp_aer_numc2massc(vl_scr6_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc

                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  
                  ! Calculate clipped massconc for this bin; put in (scr1_t3)
                  call comp_aer_coremc_kludge(ftmp, vl_scr7_t3%p3D,  &
                       vl_scr6_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))

                  ! Sum # concentrations of all bins (put in scr2_t3):
                  call comp_3D_esum(vl_scr2_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))

                  ! Calculate fall speed (put in scr5_t3):
                  call comp_aer_vfall(vl_scr3_t3%p3D, vl_scr4_t3%p3D, ig, ib,  &
                       vl_scr5_t3%p3D)

                  vl_result_t3%p3D = vl_result_t3%p3D+  &  ! Mass-weight vfall 
                       (vl_scr1_t3%p3D*vl_scr5_t3%p3D)
               end do

            else if (btype == 3) then
               call comp_3D_esum(vl_scr2_t3%p3D, vl_scr1_t3%p3D, 0,  &
                    cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))    
               do ib=1,nb
                  write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  ! Convert from #conc to mconc (put in scr6_t3):
                  call comp_aer_numc2massc(vl_scr6_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, 1, 0, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))   ! Init
                  call comp_aer_numc2massc(vl_scr6_t3%p3D, vl_scr1_t3%p3D,  &
                       atype, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
                       nzp(tgt_igd,ifn))  ! Calc

                  write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! Actual var name
                  call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)
                  
                  ! Calculate clipped massconc for this bin; put in (scr1_t3)
                  call comp_aer_coremc_kludge(ftmp, vl_scr7_t3%p3D,  &
                       vl_scr6_t3%p3D, vl_scr1_t3%p3D, 2,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))

                  ! Sum # concentrations of all bins (put in scr2_t3):
                  call comp_3D_esum(vl_scr2_t3%p3D, vl_scr1_t3%p3D, 1,  &
                       cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))

                  ! Calculate fall speed (put in scr5_t3):
                  call comp_aer_vfall(vl_scr3_t3%p3D, vl_scr4_t3%p3D, ig, ib,  &
                       vl_scr5_t3%p3D)

                  vl_result_t3%p3D = vl_result_t3%p3D+  &  ! Mass-weight vfall 
                       (vl_scr1_t3%p3D*vl_scr5_t3%p3D)
               end do

            end if

            ! Avg mass-weighted vfall:
            vl_result_t3%p3D = vl_result_t3%p3D/  &
                 max(1.e-30_f4, vl_scr2_t3%p3D)

         end if

      end if
      
   end if

   what_vtype = 3
   what_ofile = 'atm'
   what_vlevs = cfg_spatial(cfg_igd)%n_AZ

!## dust_cum_od_vis, dust_cum_od_IR, dust_lyr_od_vis, dust_lyr_od_IR
else if (var_name(1:12) == 'dust_cum_od_' .or.  &
     var_name(1:12) == 'dust_lyr_od_') then
   
   if (var_name(13:lenvn) == 'vis') then
      wv_tog = 1; wv_desc = 'vis'
   else if (var_name(13:lenvn) == 'IR') then
      wv_tog = 2; wv_desc = 'IR'
   else
      call exception(procedure_name, FATAL,  &
           'Variable unknown -- '//var_name(1:lenvn))
   end if

   if (.NOT.dust_present) call exception(procedure_name, FATAL,  &
           'Dust unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(dust_ig)+1
      do ib=1,nbins(dust_ig)    ! # bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'DUSTNC'//ctmp(1:2)    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
      var_want_to_read(nbins(dust_ig)+1) = 'TOPT'
      full_az_scr_needed = .TRUE.
   else

      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      vl_result_t3%p3D = ZERO_f4 ! Initialize optical depth array

      do ib=1,nbins(dust_ig)     ! Sum over all bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'DUSTNC'//ctmp(1:2)    ! Actual variable name

         call get_PRAMS_data(vl_scr2_faz_t3, ctmp, tgt_igd, cfg_igd, ifn,  &
              full_az=.TRUE.)

         call comp_3Daer_optdepth(vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D,  &
              zt(:,tgt_igd,ifn), zm(:,tgt_igd,ifn), vl_result_t3%p3D, 1, ib,  &
              wv_tog, var_name(6:8), vctr1_az%p1D, cfg_spatial(cfg_igd)%i_AZ)
      end do

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      if (var_name(6:8) == 'cum') then
         what_vdesc = 'cum. column dust 3D '//wv_desc//' optical depth;'
         what_vunits = 'none;'
      else
         what_vdesc = 'layer dust 3D '//wv_desc//' opacity;'
         what_vunits = '1/m;'
      end if
   end if

!## H2Oi_cum_od_vis, H2Oi_cum_od_IR, H2Oi_lyr_od_vis, H2Oi_lyr_od_IR
else if (var_name(1:12) == 'H2Oi_cum_od_' .or.  &
     var_name(1:12) == 'H2Oi_lyr_od_') then
   
   if (var_name(13:lenvn) == 'vis') then
      wv_tog = 1; wv_desc = 'vis'
   elseif(var_name(13:lenvn) == 'IR') then
      wv_tog = 2; wv_desc = 'IR'
   else
      call exception(procedure_name, FATAL,  &
           'Variable unknown -- '//var_name(1:lenvn))
   end if

   full_az_scr_needed = .TRUE.

   if (.NOT.H2Oi_present) call exception(procedure_name, FATAL,  &
        'H2Oi unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(H2Oi_ig)+1
      do ib=1,nbins(H2Oi_ig)    ! # bins
         write (ctmp, '(a6,i2.2)') 'H2OINC', ib     ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
      var_want_to_read(nbins(H2Oi_ig)+1) = 'TOPT'
   else

      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      vl_result_t3%p3D = ZERO_f4 ! Initialize optical depth array

      do ib=1,nbins(H2Oi_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'H2OINC', ib   ! Actual variable name

         call get_PRAMS_data(vl_scr2_faz_t3, ctmp, tgt_igd, cfg_igd, ifn,  &
              full_az=.TRUE.)

         call comp_3Daer_optdepth(vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D,  &
              zt(:,tgt_igd,ifn), zm(:,tgt_igd,ifn), vl_result_t3%p3D, 2, ib,  &
              wv_tog, var_name(6:8), vctr1_az%p1D, cfg_spatial(cfg_igd)%i_AZ)
      end do

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      if (var_name(6:8) == 'cum') then
         what_vdesc = 'cum. column H2O_ice 3D '//wv_desc//' optical depth;'
         what_vunits = 'none;'
      else
         what_vdesc = 'layer H2O_ice 3D '//wv_desc//' opacity;'
         what_vunits = '1/m;'
      end if
   end if

!## CO2i_cum_od_vis, CO2i_cum_od_IR, CO2i_lyr_od_vis, CO2i_lyr_od_IR
else if (var_name(1:12) == 'CO2i_cum_od_' .or.  &
     var_name(1:12) == 'CO2i_lyr_od_') then
   
   if (var_name(13:lenvn) == 'vis') then
      wv_tog = 1; wv_desc = 'vis'
   elseif(var_name(13:lenvn) == 'IR') then
      wv_tog = 2; wv_desc = 'IR'
   else
      call exception(procedure_name, FATAL,  &
           'Variable unknown -- '//var_name(1:lenvn))
   end if

   full_az_scr_needed = .TRUE.

   if (.NOT.CO2i_present) call exception(procedure_name, FATAL,  &
        'CO2i unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(CO2i_ig)+1
      do ib=1,nbins(CO2i_ig)    ! # bins
         write (ctmp, '(a6,i2.2)') 'CO2INC', ib    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
      var_want_to_read(nbins(CO2i_ig)+1) = 'TOPT'
   else

      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      vl_result_t3%p3D = ZERO_f4 ! Initialize optical depth array

      do ib=1,nbins(CO2i_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'CO2INC', ib    ! Actual variable name

         call get_PRAMS_data(vl_scr2_faz_t3, ctmp, tgt_igd, cfg_igd, ifn,  &
              full_az=.TRUE.)

         call comp_3Daer_optdepth(vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D,  &
              zt(:,tgt_igd,ifn), zm(:,tgt_igd,ifn), vl_result_t3%p3D, 3, ib,  &
              wv_tog, var_name(6:8), vctr1_az%p1D, cfg_spatial(cfg_igd)%i_AZ)
      end do

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      if (var_name(6:8) == 'cum') then
         what_vdesc = 'cum. column CO2_ice 3D '//wv_desc//' optical depth;'
         what_vunits = 'none;'
      else
         what_vdesc = 'layer CO2_ice 3D '//wv_desc//' opacity;'
         what_vunits = '1/m;'
      end if
   end if

!## adust_mnrad
else if (var_name(1:lenvn) == 'adust_mnrad') then
   
   if (.NOT.dust_present) call exception(procedure_name, FATAL,  &
        'Dust unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(dust_ig)
      do ib=1,nbins(dust_ig)    ! # bins
         write (ctmp, '(a6,i2.2)') 'DUSTNC', ib    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
   else

      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D, vl_scr3_t3%p3D,  &
           1, 0, 0, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(dust_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'DUSTNC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D,  &
              vl_scr3_t3%p3D, 1, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
              nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D, vl_scr3_t3%p3D,  &
           1, 0, 2, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Finish mean radius

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'total dust 3D mean radius;'
      what_vunits = 'cm;'
   end if

!## aH2Oi_mnrad
else if (var_name(1:lenvn) == 'aH2Oi_mnrad') then
   
   if (.NOT.H2Oi_present) call exception(procedure_name, FATAL,  &
        'H2Oi unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(H2Oi_ig)
      do ib=1,nbins(H2Oi_ig)    ! # bins
         write (ctmp, '(a6,i2.2)') 'H2OINC', ib   ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
   else

      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D, vl_scr3_t3%p3D,  &
           2, 0, 0, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(H2Oi_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'H2OINC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D,  &
              vl_scr3_t3%p3D, 2, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
              nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D, vl_scr3_t3%p3D,  &
           2, 0, 2, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Finish mean radius

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'total H2O_ice 3D mean radius;'
      what_vunits = 'cm;'
   end if

!## aCO2i_mnrad
else if (var_name(1:lenvn) == 'aCO2i_mnrad') then
   
   if (.NOT.CO2i_present) call exception(procedure_name, FATAL,  &
        'CO2i unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(CO2i_ig)
      do ib=1,nbins(CO2i_ig)    ! # bins
         write (ctmp, '(a6,i2.2)') 'CO2INC', ib    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
   else

      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D, vl_scr3_t3%p3D,  &
           3, 0, 0, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(CO2i_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'CO2INC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D,  &
              vl_scr3_t3%p3D, 3, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
              nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_result_t3%p3D, vl_scr3_t3%p3D,  &
           3, 0, 2, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Finish mean radius

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'total CO2_ice 3D mean radius;'
      what_vunits = 'cm;'
   end if

!## adust_variance
else if (var_name(1:lenvn) == 'adust_variance') then
   
   if (.NOT.dust_present) call exception(procedure_name, FATAL,  &
        'Dust unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(dust_ig)
      do ib=1,nbins(dust_ig)    ! # bins
         write (ctmp, '(a6,i2.2)') 'DUSTNC', ib   ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
   else

      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
           1, 0, 0, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(dust_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'DUSTNC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
              vl_scr3_t3%p3D, 1, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
              nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
           1, 0, 2, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Finish mean radius; tot_num in scr3_t3

      call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 1, 0, 0,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(dust_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'DUSTNC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 1, ib, 1,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 1, 0, 2,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Finish variance

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'total dust 3D variance;'
      what_vunits = 'cm^2;'
   end if

!## aH2Oi_variance
else if (var_name(1:lenvn) == 'aH2Oi_variance') then
   
   if (.NOT.H2Oi_present) call exception(procedure_name, FATAL,  &
        'H2Oi unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(H2Oi_ig)
      do ib=1,nbins(H2Oi_ig)    ! # bins
         write (ctmp,'(a6, i2.2)') 'H2OINC', ib    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
   else

      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
           2, 0, 0, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(H2Oi_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'H2OINC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
              vl_scr3_t3%p3D, 2, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
              nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
           2, 0, 2, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Finish mean radius; tot_num in scr3_t3

      call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 2, 0, 0,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(H2Oi_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'H2OINC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 2, ib, 1,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 2, 0, 2,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Finish variance

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'total H2O_ice 3D variance;'
      what_vunits = 'cm^2;'
   end if

!## aCO2i_variance
else if (var_name(1:lenvn) == 'aCO2i_variance') then
   
   if (.NOT.CO2i_present) call exception(procedure_name, FATAL,  &
        'CO2i unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(CO2i_ig)
      do ib=1,nbins(CO2i_ig)    ! # bins
         write (ctmp, '(a6,i2.2)') 'CO2INC', ib    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
   else

      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
           3, 0, 0, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(CO2i_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'CO2INC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
              vl_scr3_t3%p3D, 3, ib, 1, cfg_spatial(cfg_igd)%i_AZ,  &
              nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_mnrad(vl_scr1_t3%p3D, vl_scr2_t3%p3D, vl_scr3_t3%p3D,  &
           3, 0, 2, cfg_spatial(cfg_igd)%i_AZ,  &
           nzp(tgt_igd,ifn))   ! Finish mean radius; tot_num in scr3

      call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 3, 0, 0,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))   ! Initialize arrays
      do ib=1,nbins(CO2i_ig)     ! Sum over all bins
         write (ctmp, '(a6,i2.2)') 'CO2INC', ib    ! Actual variable name
         
         call get_PRAMS_data(vl_scr1_t3, ctmp, tgt_igd, cfg_igd, ifn)

         call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 3, ib, 1,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Sum
      end do
      call comp_3Daer_variance(vl_scr1_t3%p3D, vl_scr2_t3%p3D,  &
           vl_scr3_t3%p3D, vl_result_t3%p3D, 3, 0, 2,  &
           cfg_spatial(cfg_igd)%i_AZ, nzp(tgt_igd,ifn))  ! Finish variance

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'total CO2_ice 3D variance;'
      what_vunits = 'cm^2;'
   end if

!## lyr_Tsat_CO2
else if (var_name(1:lenvn) == 'lyr_Tsat_CO2') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'PI0','PP '/)
   else
      call get_PRAMS_data(vl_scr1_t3, 'PI0', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t3, 'PP', tgt_igd, cfg_igd, ifn)
      vl_scr1_t3%p3D = vl_scr1_t3%p3D+vl_scr2_t3%p3D   ! pi = pi0+pi'

      call comp_press(vl_scr2_t3%p3D, vl_scr1_t3%p3D)  ! Press in scr2_t3

      call comp_CO2Tsat_lyr(vl_result_t3%p3D, vl_scr2_t3%p3D)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'lyr CO2 saturation temp;'
      what_vunits = 'K;'
   endif

!## sdust_pa_rs_bXX, sdust_pa_rs_sum, sdust_pa_fi_bXX, sdust_pa_fi_sum,
!## sdust_pa_ff_bXX, sdust_pa_ff_sum, sdust_pa_aj_bXX, sdust_pa_aj_sum,
!## sdust_pa_ag_bXX, sdust_pa_ag_sum, 
!## sH2Oi_pa_rs_bXX, sH2Oi_pa_rs_sum, sH2Oi_pa_fi_bXX, sH2Oi_pa_fi_sum,
!## sH2Oi_pa_ff_bXX, sH2Oi_pa_ff_sum, sH2Oi_pa_aj_bXX, sH2Oi_pa_aj_sum,
!## sH2Oi_pa_ag_bXX, sH2Oi_pa_ag_sum,
!## sH2Oi_df_rs_bXX, sH2Oi_df_rs_sum, sH2Oi_df_rs_suf, sH2Oi_df_fi_bXX,
!## sH2Oi_df_fi_sum, sH2Oi_df_fi_suf, sH2Oi_df_ff_bXX, sH2Oi_df_ff_sum,
!## sH2Oi_df_ff_suf, sH2Oi_df_aj_bXX, sH2Oi_df_aj_sum, sH2Oi_df_aj_suf,
!## sH2Oi_df_ag_bXX, sH2Oi_df_ag_sum, sH2Oi_df_ag_suf,
!## sCO2i_pa_rs_bXX, sCO2i_pa_rs_sum, sCO2i_pa_fi_bXX, sCO2i_pa_fi_sum,
!## sCO2i_pa_ff_bXX, sCO2i_pa_ff_sum, sCO2i_pa_aj_bXX, sCO2i_pa_aj_sum,
!## sCO2i_pa_ag_bXX, sCO2i_pa_ag_sum,
!## sCO2i_df_rs_bXX, sCO2i_df_rs_sum, sCO2i_df_fi_bXX, sCO2i_df_fi_sum,
!## sCO2i_df_ff_bXX, sCO2i_df_ff_sum, sCO2i_df_aj_bXX, sCO2i_df_aj_sum,
!## sCO2i_df_ag_bXX, sCO2i_df_ag_sum,
!## sCO2i_wf_rs_bXX, sCO2i_wf_rs_sum, sCO2i_wf_fi_bXX, sCO2i_wf_fi_sum,
!## sCO2i_wf_ff_bXX, sCO2i_wf_ff_sum, sCO2i_wf_aj_bXX, sCO2i_wf_aj_sum,
!## sCO2i_wf_ag_bXX, sCO2i_wf_ag_sum
else if (var_name(1:6) == 'sdust_' .or. var_name(1:6) == 'sH2Oi_' .or.  &
     var_name(1:6) == 'sCO2i_') then

   what_vdesc = ';'
   if (var_name(2:5) == 'dust') then
      what_vdesc = 'dust '; ic = 5
   else if (var_name(2:5) == 'H2Oi') then
      what_vdesc = 'H2O_ice '; ic = 8
   else if (var_name(2:5) == 'CO2i') then
      what_vdesc = 'CO2_ice '; ic = 8
   end if

   if (var_name(13:13) == 'b') then
      vsubkind = 1
      what_vdesc = what_vdesc(1:ic)//'(b#'//var_name(14:15)//') '
      read (var_name(14:15), *) ib   ! Bin to process
      ic = ic+7
   else if (var_name(13:15) == 'sum') then
      vsubkind = 2
      what_vdesc = what_vdesc(1:ic)//'(sum bins) '
      ic = ic+11
   else if (var_name(13:15) == 'suf') then
      vsubkind = 3
      what_vdesc = what_vdesc(1:ic)//'(sum fix) '
      ic = ic+10
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   if (var_name(10:11) == 'rs') then
      vkind = 1; ctmp6_1(1) = '    RS'
      what_vdesc = what_vdesc(1:ic)//'reservoir: '; ic = ic+11
      what_vunits = ' /cm^2;'
   else if (var_name(10:11) == 'fi') then
      vkind = 2; ctmp6_1(1) = '    FI'
      what_vdesc = what_vdesc(1:ic)//'inj flux: '; ic = ic+10
      what_vunits = ' /(s.cm^2);'
   else if (var_name(10:11) == 'ff') then
      vkind = 3; ctmp6_1(1) = '    FF'
      what_vdesc = what_vdesc(1:ic)//'fall flux: '; ic = ic+11
      what_vunits = ' /(s.cm^2);'
   else if (var_name(10:11) == 'aj') then
      vkind = 4; ctmp6_1(1) = '    AJ'
      what_vdesc = what_vdesc(1:ic)//'accum inj: '; ic = ic+11
      what_vunits = ' /cm^2;'
   else if (var_name(10:11) == 'ag') then
      vkind = 5; ctmp6_1(1) = '    AG'
      what_vdesc = what_vdesc(1:ic)//'accum fall: '; ic = ic+12
      what_vunits = ' /cm^2;'
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   varname_is_unknown = .FALSE.
   if (var_name(2:5) == 'dust') then
      if (.NOT.dust_present) call exception(procedure_name, FATAL,  &
           'Dust unavailable in this run for var '//var_name(1:lenvn))
      if (var_name(7:8) == 'pa') then
         btype = 1; what_vunits(1:1) = '#'
         what_vdesc = what_vdesc(1:ic)//'dust # '; ic = ic+7
         ninv = 1; ctmp6_1(1)(1:4) = 'DUST';
      else
         varname_is_unknown = .TRUE.
      end if
      nb = nbins(dust_ig)
      atype = 1
      ig = dust_ig
   else if (var_name(2:5) == 'H2Oi') then
      if (.NOT.H2Oi_present) call exception(procedure_name, FATAL,  &
          'H2O ice aerosol unavailable in this run for var '//var_name(1:lenvn))
      if (var_name(7:8) == 'pa') then
         btype = 1; what_vunits(1:1) = '#'
         what_vdesc = what_vdesc(1:ic)//'H2Oi # '; ic = ic+7
         ninv = 1; ctmp6_1(1)(1:4) = 'H2OI';
      else if (var_name(7:8) == 'df') then
         btype = 2; what_vunits(1:1) = 'g'
         what_vdesc = what_vdesc(1:ic)//'dust core mass '; ic = ic+15
         if (vsubkind == 3) then          
            ninv = 2; ctmp6_1(1)(1:4) = 'H2OI'
            ctmp6_1(2) = ctmp6_1(1); ctmp6_1(2)(4:4) = 'D'          
         else
            ninv = 1; ctmp6_1(1)(1:4) = 'H2OD'
         end if
      else
         varname_is_unknown = .TRUE.
      end if
      nb = nbins(H2Oi_ig)
      atype = 2
      ig = H2Oi_ig
   else if (var_name(2:5) == 'CO2i') then
      if (.NOT.CO2i_present) call exception(procedure_name, FATAL,  &
          'CO2 ice aerosol unavailable in this run for var '//var_name(1:lenvn))
      if (var_name(7:8) == 'pa') then
         btype = 1; what_vunits(1:1) = '#'
         what_vdesc = what_vdesc(1:ic)//'CO2i # '; ic = ic+7
         ninv = 1; ctmp6_1(1)(1:4) = 'CO2I';
      else if (var_name(7:8) == 'df') then
         btype = 2; what_vunits(1:1) = 'g'
         what_vdesc = what_vdesc(1:ic)//'dust core mass '; ic = ic+15
         if (vsubkind == 3) then          
            ninv = 2; ctmp6_1(1)(1:4) = 'CO2I'
            ctmp6_1(2) = ctmp6_1(1); ctmp6_1(2)(4:4) = 'D'          
         else
            ninv = 1; ctmp6_1(1)(1:4) = 'CO2D'
         end if
      else if (var_name(7:8) == 'wf') then
         btype = 2; what_vunits(1:1) = 'g'
         what_vdesc = what_vdesc(1:ic)//'H2Oi core mass '; ic = ic+15
         ninv = 1; ctmp6_1(1)(1:4) = 'CO2W'
      else
         varname_is_unknown = .TRUE.
      end if
      nb = nbins(CO2i_ig)
      atype = 3
      ig = CO2i_ig
   else
      varname_is_unknown = .TRUE.
   end if

   if (varname_is_unknown) call exception(procedure_name, FATAL,  &
        'Variable '//var_name(1:lenvn)//' unknown.')

   what_vdesc = what_vdesc(1:ic)//';'   ! Finish description string

   if (vsubkind == 1) then     ! Process a single bin only

      if (this_is_init_pass) then
         nvar_want = ninv
         do inv=1,ninv
            write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib ! Actual variable name
            var_want_to_read(inv) = ctmp(1:var_name_len)
         end do
      else
         write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
         call get_PRAMS_data(vl_result_t2, ctmp, tgt_igd, cfg_igd, ifn)
      end if

   else if (vsubkind == 2) then      ! Sum of all bins' contributions

      if (this_is_init_pass) then
         nvar_want = ninv*nb
         do inv=1,ninv
            do ib=1,nb    ! # bins
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
               var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
            end do
         end do
      else        
         vl_result_t2%p2D = ZERO_f4  ! Initialize summing array
         do ib=1,nb     ! Sum over all bins
            write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual variable name
            call get_PRAMS_data(vl_scr1_t2, ctmp, tgt_igd, cfg_igd, ifn)
            ! Sum # concentrations of all bins:
            vl_result_t2%p2D = vl_result_t2%p2D+vl_scr1_t2%p2D
         end do
      end if

   else if (vsubkind == 3) then ! "Fixed version" sum of all bins' contributions
! DELETE ME WHEN aerosol model is fixed!!!!!!

      if (this_is_init_pass) then
         nvar_want = ninv*nb
         do inv=1,ninv
            do ib=1,nb    ! # bins
               write (ctmp, '(a6,i2.2)') ctmp6_1(inv), ib  ! Actual var name
               var_want_to_read((inv-1)*nb+ib) = ctmp(1:var_name_len)
            end do
         end do
      else
         call comp_aer_numc2massc_2D(vl_scr3_t2%p2D, vl_scr1_t2%p2D, atype,  &
              1, 0)   ! Initialize summing array
         do ib=1,nb
            write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
            call get_PRAMS_data(vl_scr1_t2, ctmp, tgt_igd, cfg_igd, ifn)
            ! Convert from #conc to mconc (put in scr3_t2):
            call comp_aer_numc2massc_2D(vl_scr3_t2%p2D, vl_scr1_t2%p2D,  &
                 atype, ib, 1)  ! Calc
         end do
         ftmp = log10(maxval(vl_scr3_t2%p2D))

         if (btype == 2) then
            vl_result_t2%p2D = ZERO_f4    
            do ib=1,nb
               write (ctmp, '(a6,i2.2)') ctmp6_1(1), ib  ! Actual var name
               call get_PRAMS_data(vl_scr1_t2, ctmp, tgt_igd, cfg_igd, ifn)
               ! Convert from #conc to mconc (put in scr2_t2):
               call comp_aer_numc2massc_2D(vl_scr2_t2%p2D, vl_scr1_t2%p2D,  &
                    atype, 1, 0)   ! Init
               call comp_aer_numc2massc_2D(vl_scr2_t2%p2D, vl_scr1_t2%p2D,  &
                    atype, ib, 1)  ! Calc
               
               write (ctmp, '(a6,i2.2)') ctmp6_1(2), ib  ! Actual var name
               call get_PRAMS_data(vl_scr1_t2, ctmp, tgt_igd, cfg_igd, ifn)
                  
               ! Calculate clipped massconc for this bin; put in (scr1_t2)
               call comp_aer_coremc_kludge_2D(ftmp, vl_scr3_t2%p2D,  &
                    vl_scr2_t2%p2D, vl_scr1_t2%p2D, 1)
                  
               ! Sum # concentrations of all bins:
               vl_result_t2%p2D = vl_result_t2%p2D+vl_scr1_t2%p2D
            end do
         end if
      end if

   end if

   what_vtype = 2
   what_ofile = 'sfc'
   what_vlevs = 0

!## dust_sfc_od_vis, dust_sfc_od_IR
else if (var_name(1:12) == 'dust_sfc_od_') then
   
   if (var_name(13:lenvn) == 'vis') then
      wv_tog = 1;  wv_desc = 'vis'
   else if (var_name(13:lenvn) == 'IR') then
      wv_tog = 2; wv_desc = 'IR'
   else
      call exception(procedure_name, FATAL,  &
           'Variable unknown -- '//var_name(1:lenvn))
   end if

   if (.NOT.dust_present) call exception(procedure_name, FATAL,  &
        'Dust unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(dust_ig)+1
      do ib=1,nbins(dust_ig)    ! # bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'DUSTNC'//ctmp(1:2)    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
      var_want_to_read(nbins(dust_ig)+1) = 'TOPT'
      full_az_scr_needed = .TRUE.
   else

      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      vl_result_t2%p2D = ZERO_f4 ! Initialize optical depth array

      do ib=1,nbins(dust_ig)     ! Sum over all bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'DUSTNC'//ctmp(1:2)    ! Actual variable name

         call get_PRAMS_data(vl_scr2_faz_t3, ctmp, tgt_igd, cfg_igd, ifn,  &
              full_az=.TRUE.)

         call comp_2Daer_optdepth(vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D,  &
              zt(:,tgt_igd,ifn), zm(:,tgt_igd,ifn), vl_result_t2%p2D, 1, ib,  &
              wv_tog)
      end do

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'sfc dust '//wv_desc//' optical depth (column);'
      what_vunits = 'none;'
   end if

!## H2Oi_sfc_od_vis, H2Oi_sfc_od_IR
else if (var_name(1:12) == 'H2Oi_sfc_od_') then
   
   if (var_name(13:lenvn) == 'vis') then
      wv_tog = 1; wv_desc = 'vis'
   else if (var_name(13:lenvn) == 'IR') then
      wv_tog = 2; wv_desc = 'IR'
   else
      call exception(procedure_name, FATAL,  &
           'Variable unknown -- '//var_name(1:lenvn))
   end if

   if (.NOT.H2Oi_present) call exception(procedure_name, FATAL,  &
        'H2Oi unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(H2Oi_ig)+1
      do ib=1,nbins(H2Oi_ig)    ! # bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'H2OINC'//ctmp(1:2)    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
      var_want_to_read(nbins(H2Oi_ig)+1) = 'TOPT'
      full_az_scr_needed = .TRUE.
   else

      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      vl_result_t2%p2D = ZERO_f4 ! Initialize optical depth array

      do ib=1,nbins(H2Oi_ig)     ! Sum over all bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'H2OINC'//ctmp(1:2)    ! Actual variable name

         call get_PRAMS_data(vl_scr2_faz_t3, ctmp, tgt_igd, cfg_igd, ifn,  &
              full_az=.TRUE.)

         call comp_2Daer_optdepth(vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D,  &
              zt(:,tgt_igd,ifn), zm(:,tgt_igd,ifn), vl_result_t2%p2D, 1, ib,  &
              wv_tog)
      end do

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'sfc H2O_ice '//wv_desc//' optical depth (column);'
      what_vunits = 'none;'
   end if

!## CO2i_sfc_od_vis, CO2i_sfc_od_IR
else if (var_name(1:12) == 'CO2i_sfc_od_') then
   
   if (var_name(13:lenvn) == 'vis') then
      wv_tog = 1; wv_desc = 'vis'
   else if (var_name(13:lenvn) == 'IR') then
      wv_tog = 2; wv_desc = 'IR'
   else
      call exception(procedure_name, FATAL,  &
           'Variable unknown -- '//var_name(1:lenvn))
   end if

   if (.NOT.CO2i_present) call exception(procedure_name, FATAL,  &
        'CO2i unavailable in this run for variable '//var_name(1:lenvn))

   if (this_is_init_pass) then
      nvar_want = nbins(CO2i_ig)+1
      do ib=1,nbins(CO2i_ig)    ! # bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'CO2INC'//ctmp(1:2)    ! Actual variable name
         var_want_to_read(ib) = ctmp(1:var_name_len)
      end do
      var_want_to_read(nbins(CO2i_ig)+1) = 'TOPT'
      full_az_scr_needed = .TRUE.
   else

      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      vl_result_t2%p2D = ZERO_f4 ! Initialize optical depth array

      do ib=1,nbins(CO2i_ig)     ! Sum over all bins
         write (ctmp, '(i2.2)') ib
         ctmp = 'CO2INC'//ctmp(1:2)    ! Actual variable name

         call get_PRAMS_data(vl_scr2_faz_t3, ctmp, tgt_igd, cfg_igd, ifn,  &
              full_az=.TRUE.)

         call comp_2Daer_optdepth(vl_scr2_faz_t3%p3D, vl_scr1_t2%p2D,  &
              zt(:,tgt_igd,ifn), zm(:,tgt_igd,ifn), vl_result_t2%p2D, 1, ib,  &
              wv_tog)
      end do

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'sfc CO2_ice '//wv_desc//' optical depth (column);'
      what_vunits = 'none;'
   end if

!## fthrd
else if (var_name(1:lenvn) == 'fthrd') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'FTHRD'/)
   else
      call get_PRAMS_data(vl_result_t3, 'FTHRD', tgt_igd, cfg_igd, ifn)

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vdesc = 'radiative heating rate;'
      what_vunits = 'K/s;'
   end if

!## rlongup
else if (var_name(1:lenvn) == 'rlongup') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'RLONGUP'/)
   else
      call get_PRAMS_data(vl_result_t2, 'RLONGUP', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'longwave flux upwards (from sfc);'
      what_vunits = 'W/m^2;'
   end if

!## rlong
else if (var_name(1:lenvn) == 'rlong') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'RLONG'/)
   else
      call get_PRAMS_data(vl_result_t2, 'RLONG', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'longwave flux downwards (at sfc);'
      what_vunits = 'W/m^2;'
   end if

!## rshort
else if (var_name(1:lenvn) == 'rshort') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'RSHORT'/)
   else
      call get_PRAMS_data(vl_result_t2, 'RSHORT', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface shortwave flux;'
      what_vunits = 'W/m^2;'
   end if

!## vartau_cnu, vartau_tau
else if (var_name(1:7) == 'vartau_') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'GLAT'/)
   else
      call get_PRAMS_data(vl_scr1_t2, 'GLAT', tgt_igd, cfg_igd, ifn)

      ! Conrath-nu in scr3_t2, tau in scr2_t2:
      call mvar_dust_params(cfg_spatial(cfg_igd)%n_AZ,  &
           cfg_spatial(cfg_igd)%n_X, cfg_spatial(cfg_igd)%n_Y, 1,  &
           cfg_spatial(cfg_igd)%n_X, 1, cfg_spatial(cfg_igd)%n_Y,  &
           Ls_curr(ifn), vl_scr1_t2%p2D, vl_scr2_t2%p2D, vl_scr3_t2%p2D)

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0

      if (var_name(8:lenvn) == 'cnu') then
         vl_result_t2%p2D = vl_scr3_t2%p2D
         what_vdesc = 'Conrath-nu parameter;'
         what_vunits = 'none;'
      else if (var_name(8:lenvn) == 'tau') then
         vl_result_t2%p2D = vl_scr2_t2%p2D
         what_vdesc = 'visible dust opacity (ref to 6.1 hPa);'
         what_vunits = 'none;'
      else
         call exception(procedure_name, FATAL,  &
              'Variable '//var_name(1:lenvn)//' not found in varlib.')
      end if

   end if

!## vtau_hzt, vtau_tau
else if (var_name(1:5) == 'vtau_') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'GLAT'/)
      call za_vardust_spec_init(default_input_data_dir_bspec)
   else
      call get_PRAMS_data(vl_scr1_t2, 'GLAT', tgt_igd, cfg_igd, ifn)

      ! Haze-top in scr3_t2, tau (9 um) in scr2_t2:
      do j=1,cfg_spatial(cfg_igd)%n_Y
         call za_vardust_spec(cfg_spatial(cfg_igd)%n_X, Ls_curr(ifn),  &
              vl_scr1_t2%p2D(:,j), vl_scr2_t2%p2D(:,j), vl_scr3_t2%p2D(:,j))
      end do

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0

      if (var_name(6:lenvn) == 'hzt') then
         vl_result_t2%p2D = vl_scr3_t2%p2D
         what_vdesc = 'haze-top parameter;'
         what_vunits = 'km;'
      else if (var_name(6:lenvn) == 'tau') then
         vl_result_t2%p2D = vl_scr2_t2%p2D
         what_vdesc = '9 um dust opacity (ref to 6.1 hPa);'
         what_vunits = 'none;'
      else
         call exception(procedure_name, FATAL,  &
              'Variable '//var_name(1:lenvn)//' not found in varlib.')
      end if

   end if

!## tslope_angle
else if (var_name(1:lenvn) == 'tslope_angle') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TSLOPANG'/)
   else
      call get_PRAMS_data(vl_scr1_t2, 'TSLOPANG', tgt_igd, cfg_igd, ifn)
 
      vl_result_t2%p2D = vl_scr1_t2%p2D*rad_to_deg_f4
 
      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'topographic slope angle (from horizontal);'
      what_vunits = 'deg;'
   end if

!## cosz
else if (var_name(1:lenvn) == 'cosz') then
 
   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'COSZ'/)
   else
      call get_PRAMS_data(vl_result_t2, 'COSZ', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'cosine of the zenith angle;'
      what_vunits = 'none;'
   end if

!## sbps_thickness
else if (var_name(1:lenvn) == 'sbps_thickness') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'SBPSLTHK'/)
   else
      call get_PRAMS_data(vl_result_t5, 'SBPSLTHK', tgt_igd, cfg_igd, ifn)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'thickness of each SBPS layer;'
      what_vunits = 'm;'
   end if

!## sbps_q
else if (var_name(1:lenvn) == 'sbps_q') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'SBPSQ'/)
   else
      call get_PRAMS_data(vl_result_t5, 'SBPSQ', tgt_igd, cfg_igd, ifn)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'SBPS thermal energy (layer);'
      what_vunits = 'J/m^2;'
   end if

!## sbps_H2Oi_fraction
else if (var_name(1:lenvn) == 'sbps_H2Oi_fraction') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'SBSH2OIF'/)
   else
      call get_PRAMS_data(vl_result_t5, 'SBSH2OIF', tgt_igd, cfg_igd, ifn)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'Solid-phase H2O fraction for each SBPS layer;'
      what_vunits = 'none;'
   end if

!## sbps_H2O_substance
else if (var_name(1:lenvn) == 'sbps_H2O_substance') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'SBPSH2OA'/)
   else
      call get_PRAMS_data(vl_result_t5, 'SBPSH2OA', tgt_igd, cfg_igd, ifn)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'Amount of H2O substance in each SBPS layer;'
      what_vunits = 'kg/m^3;'
   end if

!## sbps_tempk
else if (var_name(1:lenvn) == 'sbps_tempk') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SBPSDENS','SBPSTMCP','SBPSQ   '/)
   else
      call get_PRAMS_data(vl_scr1_t5, 'SBPSDENS', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_scr2_t5, 'SBPSTMCP', tgt_igd, cfg_igd, ifn)
      call get_PRAMS_data(vl_result_t5, 'SBPSQ', tgt_igd, cfg_igd, ifn)

      call comp_SBPStempk(vl_result_t5%p4D, vl_scr1_t5%p4D, vl_scr2_t5%p4D)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'SBPS temperature (layer);'
      what_vunits = 'K;'
   end if

!## sbps_density
else if (var_name(1:lenvn) == 'sbps_density') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SBPSDENS'/)
   else
      call get_PRAMS_data(vl_result_t5, 'SBPSDENS', tgt_igd, cfg_igd, ifn)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'SBPS density (layer);'
      what_vunits = 'kg/m^3;'
   end if

!## sbps_thrmcp
else if (var_name(1:lenvn) == 'sbps_thrmcp') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SBPSTMCP'/)
   else
      call get_PRAMS_data(vl_result_t5, 'SBPSTMCP', tgt_igd, cfg_igd, ifn)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'SBPS heat capacity (layer);'
      what_vunits = 'J/(kg.K);'
   end if

!## sbps_thrmcond
else if (var_name(1:lenvn) == 'sbps_thrmcond') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SBPSTMCD'/)
   else
      call get_PRAMS_data(vl_result_t5, 'SBPSTMCD', tgt_igd, cfg_igd, ifn)

      what_vtype = 5
      what_ofile = 'sfc'
      what_vlevs = cfg_spatial(cfg_igd)%n_SZ
      what_vdesc = 'SBPS thermal conductivity (layer);'
      what_vunits = 'W/(m.K);'
   end if

!## topo
else if (var_name(1:lenvn) == 'topo') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TOPT'/)
   else
      call get_PRAMS_data(vl_result_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface topography height;'
      what_vunits = 'm;'
   end if

!## albedt
else if (var_name(1:lenvn) == 'albedt') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'ALBEDT'/)
   else
      call get_PRAMS_data(vl_result_t2, 'ALBEDT', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface albedo (time-varying);'  
      what_vunits = 'none;'
   end if

!## emisst
else if (var_name(1:lenvn) == 'emisst') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'EMISST'/)
   else
      call get_PRAMS_data(vl_result_t2, 'EMISST', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface emissivity (time-varying);'  
      what_vunits = 'none;'
   end if

!## albd_static
else if (var_name(1:lenvn) == 'albd_static') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'ALBDSTAT'/)
   else
      call get_PRAMS_data(vl_result_t4, 'ALBDSTAT', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface albedo (time-invariant);'  
      what_vunits = 'none;'
   end if

!## emiss_static
else if (var_name(1:lenvn) == 'emiss_static') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'EMISSTAT'/)
   else
      call get_PRAMS_data(vl_result_t4, 'EMISSTAT', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface emissivity (time-invariant);'  
      what_vunits = 'none;'
   end if

!## z0net
else if (var_name(1:lenvn) == 'z0net') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'Z0NET'/)
   else
      call get_PRAMS_data(vl_result_t4, 'Z0NET', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'net aerodynamic sfc roughness;'
      what_vunits = 'm;'
   end if

!## ti_static
else if (var_name(1:lenvn) == 'ti_static') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TISTAT'/)
   else
      call get_PRAMS_data(vl_result_t4, 'TISTAT', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface thermal inertia (time-invariant);'
      what_vunits = 'J/((s^0.5).K.(m^2));'
   end if

!## ti_tvar_sfc
else if (var_name(1:lenvn) == 'ti_tvar_sfc') then

   if (this_is_init_pass) then
      nvar_want = 4
      var_want_to_read(1:nvar_want) = (/'SBPSDENS','SBPSTMCP','SBPSTMCD',  &
           'SBPSLTHK'/)
      full_sz_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_fsz_t5, 'SBPSDENS', tgt_igd, cfg_igd, ifn,  &
           full_sz=.TRUE.)
      call get_PRAMS_data(vl_scr2_fsz_t5, 'SBPSTMCP', tgt_igd, cfg_igd, ifn,  &
           full_sz=.TRUE.)
      call get_PRAMS_data(vl_scr3_fsz_t5, 'SBPSTMCD', tgt_igd, cfg_igd, ifn,  &
           full_sz=.TRUE.)
      call get_PRAMS_data(vl_scr4_fsz_t5, 'SBPSLTHK', tgt_igd, cfg_igd, ifn,  &
           full_sz=.TRUE.)

      call comp_TItvar_sfc(vl_result_t4%p3D, vl_scr1_fsz_t5%p4D,  &
           vl_scr2_fsz_t5%p4D, vl_scr3_fsz_t5%p4D, vl_scr4_fsz_t5%p4D)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'surface thermal inertia (time-varying);'
      what_vunits = 'J/((s^0.5).K.(m^2));'
   end if

!## sfc_tempk
else if (var_name(1:lenvn) == 'sfc_tempk') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK'/)
   else
      call get_PRAMS_data(vl_result_t4, 'SFCTEMPK', tgt_igd, cfg_igd, ifn)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'sfc (ice or mineral) temperature;'
      what_vunits = 'K;'
   end if

!## sfc_press
else if (var_name(1:lenvn) == 'sfc_press') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'PI0','PP '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr1_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'

      call comp_sfc_press(vl_result_t2%p2D, vl_scr1_faz_t3%p3D)

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'surface air pressure;'
      what_vunits = 'Pa;'
   end if

!## sfc_Tsat_CO2
else if (var_name(1:lenvn) == 'sfc_Tsat_CO2') then

   if (this_is_init_pass) then
      nvar_want = 2
      var_want_to_read(1:nvar_want) = (/'PI0','PP '/)
      full_az_scr_needed = .TRUE.
   else
      call get_PRAMS_data(vl_scr1_faz_t3, 'PI0', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      call get_PRAMS_data(vl_scr2_faz_t3, 'PP', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)
      vl_scr1_faz_t3%p3D = vl_scr1_faz_t3%p3D+vl_scr2_faz_t3%p3D  ! pi = pi0+pi'

      call comp_sfc_press(vl_scr2_t2%p2D, vl_scr1_faz_t3%p3D) ! sfc_p in scr2_t2

      call comp_CO2Tsat_sfc(vl_result_t2%p2D, vl_scr2_t2%p2D)

      what_vtype = 2
      what_ofile = 'atm'
      what_vlevs = 0
      what_vdesc = 'sfc CO2 saturation temp;'
      what_vunits = 'K;'
   end if

!## sfc_sspct_H2Oi, sfc_sspct_H2Ol
else if (var_name(1:lenvn) == 'sfc_sspct_H2Oi' .or.  &
     var_name(1:lenvn) == 'sfc_sspct_H2Ol') then

   if (this_is_init_pass) then
      nvar_want = 3
      var_want_to_read(1:nvar_want) = (/'SFCTEMPK','TOPT    ','RV      '/)
      full_az_scr_needed = .TRUE.
   else
      ! Ground temperature (put in scr1_t4):
      call get_PRAMS_data(vl_scr1_t4, 'SFCTEMPK', tgt_igd, cfg_igd, ifn)

      ! Extract RV (put in scr1_t3):
      call get_PRAMS_data(vl_scr1_faz_t3, 'RV', tgt_igd, cfg_igd, ifn,  &
           full_az=.TRUE.)

      ! TOPT (put in scr1_t2):
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      ! Layer midpoint height AGL (put in scr4_faz_t3):
      call comp_modlyr_Z(vl_scr4_faz_t3%p3D, vl_scr1_t2%p2D, vl_scr1_t2%p2D,  &
           zt(:,tgt_igd,ifn), 'AGL', cfg_spatial(cfg_igd)%i_fAZ)

      ! Approximate sfc value of RV (via linear extrapolation; put in scr2_t4):
      call comp_sfc_value_extrp(vl_scr4_faz_t3%p3D, vl_scr1_faz_t3%p3D,  &
           vl_scr2_t4%p3D, 1)

      if (var_name(lenvn:lenvn) == 'i') then
         ! Compute sfc saturation vapor pressure over ice (put in scr3_t4);
         !  result_t4 is a dummy array in this case:
         call comp_gas_vapor_press(vl_scr1_t4%p3D, vl_scr3_t4%p3D,  &
              vl_result_t4%p3D, 'W')
         what_vdesc = 'sfc H2O supersaturation over ice;'
      else if (var_name(lenvn:lenvn) == 'l') then
         ! Compute sfc saturation vapor pressure over liquid (put in scr3_t4);
         !  result_t4 is a dummy array in this case:
         call comp_gas_vapor_press(vl_scr1_t4%p3D, vl_result_t4%p3D,  &
              vl_scr3_t4%p3D, 'W')
         what_vdesc = 'sfc H2O supersaturation over liquid;'
      end if

      ! Compute supersaturation (put in scr4):
      call comp_supsat_H2Ov(vl_scr1_t4%p3D, vl_scr2_t4%p3D,  &
           vl_scr3_t4%p3D, vl_result_t4%p3D)

      what_vtype = 4
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vunits = 'pct of supersaturation;'
   end if

!## glat
else if (var_name(1:lenvn) == 'glat') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'GLAT'/)
   else
      call get_PRAMS_data(vl_result_t2, 'GLAT', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'grid point latitude;'
      what_vunits = 'deg;'
   end if

!## glon
else if (var_name(1:lenvn) == 'glon') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'GLON'/)
   else
      call get_PRAMS_data(vl_result_t2, 'GLON', tgt_igd, cfg_igd, ifn)

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'grid point longitude;'
      what_vunits = 'deg;'
   end if

!## z_lyrmid_AGL, z_lyrmid_XOID, z_lyrbnd_AGL, z_lyrbnd_XOID
else if (var_name(1:2) == 'z_') then

   if (var_name(3:8) == 'lyrmid') then
      atype = 1
   else if (var_name(3:8) == 'lyrbnd') then
      atype = 2
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   if (var_name(9:12) == '_AGL') then
      vkind = 1
   else if (var_name(9:13) == '_XOID') then
      vkind = 2
   else
      call exception(procedure_name, FATAL,  &
           'Variable '//var_name(1:lenvn)//' unknown.')
   end if

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'TOPT'/)
   else
      call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
      call comp_rtgX(vl_scr1_t2%p2D)   ! calc rtgt; put in scr1_t2

      call get_PRAMS_data(vl_scr2_t2, 'TOPT', tgt_igd, cfg_igd, ifn)

      if (atype == 1) then
         if (vkind == 1) then
            call comp_modlyr_Z(vl_result_t3%p3D, vl_scr2_t2%p2D,  &
                 vl_scr1_t2%p2D, zt(:,tgt_igd,ifn), 'AGL',  &
                 cfg_spatial(cfg_igd)%i_AZ)
            what_vdesc = 'atm model lyr midpt height (AGL);'
         else if (vkind == 2) then
            call comp_modlyr_Z(vl_result_t3%p3D, vl_scr2_t2%p2D,  &
                 vl_scr1_t2%p2D, zt(:,tgt_igd,ifn), 'XOID',  &
                 cfg_spatial(cfg_igd)%i_AZ)
            what_vdesc = 'atm model lyr midpt, referenced to Xoid surface;'
         end if
      else if (atype == 2) then
         if (vkind == 1) then
            call comp_modlyr_Z(vl_result_t3%p3D, vl_scr2_t2%p2D,  &
                 vl_scr1_t2%p2D, zm(:,tgt_igd,ifn), 'AGL',  &
                 cfg_spatial(cfg_igd)%i_AZ)
            what_vdesc = 'atm model lyr lower_bnd height (AGL);'
         else if (vkind == 2) then
            call comp_modlyr_Z(vl_result_t3%p3D, vl_scr2_t2%p2D,  &
                 vl_scr1_t2%p2D, zm(:,tgt_igd,ifn), 'XOID',  &
                 cfg_spatial(cfg_igd)%i_AZ)
            what_vdesc = 'atm model lyr lower_bnd, referenced to Xoid surface;'
         end if
      end if

      what_vtype = 3
      what_ofile = 'atm'
      what_vlevs = cfg_spatial(cfg_igd)%n_AZ
      what_vunits = 'm;'
   end if

!## dXt_body
else if (var_name(1:lenvn) == 'dXt_body') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'DXT'/)
   else
      call get_PRAMS_data(vl_scr1_t2, 'DXT', tgt_igd, cfg_igd, ifn)

      vl_result_t2%p2D = ONE_f4/vl_scr1_t2%p2D

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'T-pt delta-X on the surface of the body;'
      what_vunits = 'm;'
   end if

!## dYt_body
else if (var_name(1:lenvn) == 'dYt_body') then

   if (this_is_init_pass) then
      nvar_want = 1
      var_want_to_read(1:nvar_want) = (/'DYT'/)
   else
      call get_PRAMS_data(vl_scr1_t2, 'DYT', tgt_igd, cfg_igd, ifn)

      vl_result_t2%p2D = ONE_f4/vl_scr1_t2%p2D

      what_vtype = 2
      what_ofile = 'sfc'
      what_vlevs = 0
      what_vdesc = 'T-pt delta-Y on the surface of the body;'
      what_vunits = 'm;'
   end if

else

   call exception(procedure_name, FATAL,  &
        'Variable '//var_name(1:lenvn)//' not found in varlib.')

end if

! Interpolate result values to Cartesian surfaces (in the vertical) if desired:
!  (NOTE: only for 3D atmospheric fields with a full AZ range configured)
if (.NOT.this_is_init_pass .and. ztrnsfm == 2 .and. what_vtype == 3 .and.  &
     have_full_az(cfg_igd)) then
   call get_PRAMS_data(vl_scr1_t2, 'TOPT', tgt_igd, cfg_igd, ifn)
   call PRAMS_Cartpost_interp(vl_result_t3%p3D, vctr1_az%p1D, vctr2_az%p1D,  &
        vctr3_az%p1D, vctr4_az%p1D, vl_scr1_t2%p2D, ztop,  &
        zt(:,tgt_igd,ifn))
end if

end subroutine postp_varlib


end module postp_varlib_mod
